/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.commands;

import java.awt.Color;
import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.manipulators.*;

/**
 * Parses color expressions.
 */
public class ColorExprParser
{
    /**
     * Parses color expressions.
     */
    public Color parseColor(String expr)
    {
        expr = expr.trim();

        // defined color?
        int colorIndex = ColorTable.getIndex(expr);
        if(colorIndex != -1){
            return ColorTable.getColor(colorIndex);
        }

        // RGB value?
        TupleFloatExprParser tep = new TupleFloatExprParser();
        float[] rgb = tep.parse(expr, "[", "]", 3);
        if(rgb != null){
            if( rgb[0] >=0 && rgb[0] <= 255 &&
                rgb[1] >=0 && rgb[1] <= 255 &&
                rgb[2] >=0 && rgb[2] <= 255 )
            {
                return new Color( (int)rgb[0], (int)rgb[1], (int)rgb[2] );
            }
        }

        return null;
    }

    /**
     * Parses color expressions with the target object.
     */
    public ColorExpr parse(String expr)
    {
        ColorExpr colorExpr = new ColorExpr();

        expr = expr.toLowerCase().trim();
        int index = expr.indexOf(" ");

        String color = null;
        if(index != -1){
            String obj = expr.substring(0, index);
            color = expr.substring(index).trim();

            if( obj.equals("atoms") || obj.equals("atom") ){
                colorExpr.setTarget(ColorExpr.TARGET_ATOM);
            }
            else if( obj.equals("bonds") || obj.equals("bond") ){
                colorExpr.setTarget(ColorExpr.TARGET_BOND);
            }
            else if( obj.equals("backbone") ){
                colorExpr.setTarget(ColorExpr.TARGET_BACKBONE);
            }
            else if( obj.equals("hbonds") || obj.equals("hbond") ){
                colorExpr.setTarget(ColorExpr.TARGET_HBOND);
            }
            else if( obj.equals("ssbonds") || obj.equals("ssbond") ){
                colorExpr.setTarget(ColorExpr.TARGET_SSBOND);
            }
            else if( obj.equals("ribbons") || obj.equals("ribbon") || 
                     obj.equals("cartoons") || obj.equals("cartoon") )
            {
                colorExpr.setTarget(ColorExpr.TARGET_RESIDUE);
            }
            else if( obj.equals("ribbonback") || obj.equals("cartoonback") ){
                colorExpr.setTarget(ColorExpr.TARGET_RESIDUE_BACK);
            }
            else{
                colorExpr.setTarget(ColorExpr.TARGET_ATOM);
                color = expr;
            }
        }
        else{
            colorExpr.setTarget(ColorExpr.TARGET_ATOM);
            color = expr;
        }

        Color col = this.parseColor(color);
        if(col != null){
            colorExpr.setScheme(ColorExpr.SCHEME_RGB);
            colorExpr.setColor(col);
            return colorExpr;
        }

        // defined scheme?
        if(colorExpr.getTarget() == ColorExpr.TARGET_ATOM){
            if( color.equals("amino") ){
                colorExpr.setScheme(ColorExpr.SCHEME_AMINO);
                return colorExpr;
            }
            else if( color.equals("chain") ){
                colorExpr.setScheme(ColorExpr.SCHEME_CHAIN);
                return colorExpr;
            }
            else if( color.equals("charge") ){
                colorExpr.setScheme(ColorExpr.SCHEME_CHARGE);
                return colorExpr;
            }
            else if( color.equals("cpk") ){
                colorExpr.setScheme(ColorExpr.SCHEME_CPK);
                return colorExpr;
            }
            else if( color.equals("group") ){
                colorExpr.setScheme(ColorExpr.SCHEME_GROUP);
                return colorExpr;
            }
            else if( color.equals("shapely") ){
                colorExpr.setScheme(ColorExpr.SCHEME_SHAPELY);
                return colorExpr;
            }
            else if( color.equals("structure") ){
                colorExpr.setScheme(ColorExpr.SCHEME_STRUCTURE);
                return colorExpr;
            }
            else if( color.equals("temperature") ){
                colorExpr.setScheme(ColorExpr.SCHEME_TEMPERATURE);
                return colorExpr;
            }
        }
        else if( colorExpr.getTarget() == ColorExpr.TARGET_HBOND 
                 && color.equals("type") )
        {
            colorExpr.setScheme(ColorExpr.SCHEME_TYPE);
            return colorExpr;
        }
        else if( color.equals("none") ){
            colorExpr.setScheme(ColorExpr.SCHEME_DEFAULT);
            return colorExpr;
        }

        if ((colorExpr.getTarget() == ColorExpr.TARGET_ATOM
                || colorExpr.getTarget() == ColorExpr.TARGET_RESIDUE
                || colorExpr.getTarget() == ColorExpr.TARGET_RESIDUE_BACK)
                && color.equals("dssp")) {
            colorExpr.setScheme(ColorExpr.SCHEME_DSSP);
            return colorExpr;
        }
        return null;
    }
}
