/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.commands;

import java.util.ArrayList;

/**
 * Represents command.
 */
public class Command
{
    private String[] _names;

    private ArrayList _children;
    private ArrayList _possibles;

    private Command _default;

    private int _fixedPosition;

    String message;
    String errorMessage;
    boolean useUsage;
    
    boolean requiresPreExecute;

    /**
     * Constructor.
     */
    public Command(String[] names)
    {
        _names = names;
        
        _children  = new ArrayList();
        _possibles = new ArrayList();
    }

    /**
     * Parse command name.
     */
    public boolean parseName(String commandName)
    {
        commandName = commandName.toLowerCase();

        for(int i=0; i<_names.length; i++){
            if( _names[i].equals(commandName) ) return true;
        }
        return false;
    }

    //--------------------------------------------------------------------
    // Internal accessor
    //--------------------------------------------------------------------
    String[] getNames()
    {
        return _names;
    }

    void addChild(Command child)
    {
        _children.add(child);
    }

    int numChildren()
    {
        return _children.size();
    }
    
    Command getChild(int index)
    {
        return (Command)_children.get(index);
    }

    void setDefault(Command defaultChild)
    {
        _default = defaultChild;
    }

    Command getDefault()
    {
        return _default;
    }

    int getFixedPosition()
    {
        return _fixedPosition;
    }

    ArrayList getPossibles()
    {
        return _possibles;
    }

    //--------------------------------------------------------------------
    // Pre-Execute
    //--------------------------------------------------------------------
    public void preExecute(CommandTokenizer tokenizer, Context context)
    {
        message = null;
        errorMessage = null;
        if(!requiresPreExecute) return;        

        if(_children.size() == 0){
            this.preExecuteCommand(tokenizer, context);
        }
        else{
            Command next = null;
            if( tokenizer.hasMoreTokens() ){
                String nextToken = tokenizer.nextToken();
                for(int i=0; i<_children.size(); i++){
                    Command child = (Command)_children.get(i);
                    if( child.parseName(nextToken) ){
                        next = child;
                        break;
                    }
                }
                if(next == null) tokenizer.movePrevious();
            }
            if(next == null) next = _default;

            if(next != null){
                next.preExecute(tokenizer, context);
                message = next.message;
                errorMessage = next.errorMessage;
            }
        }
    }

    private void preExecuteCommand(CommandTokenizer tokenizer, Context context)
    {
        CommandBehavior.preExecute(this, tokenizer, context);
    }

    //--------------------------------------------------------------------
    // Execute
    //--------------------------------------------------------------------
    public void execute(CommandTokenizer tokenizer, Context context)
    {
        message = null;
        errorMessage = null;

        if(_children.size() == 0){
            this.executeCommand(tokenizer, context);
        }
        else{
            Command next = null;
            if( tokenizer.hasMoreTokens() ){
                String nextToken = tokenizer.nextToken();
                for(int i=0; i<_children.size(); i++){
                    Command child = (Command)_children.get(i);
                    if( child.parseName(nextToken) ){
                        next = child;
                        break;
                    }
                }
                if(next == null) tokenizer.movePrevious();
            }
            if(next == null) next = _default;

            if(next == null){
                errorMessage = CommandBehavior.INVALID_COMMAND;
            }
            else{
                next.execute(tokenizer, context);
                message = next.message;
                errorMessage = next.errorMessage;
            }
        }
    }

    private void executeCommand(CommandTokenizer tokenizer, Context context)
    {
        CommandBehavior.execute(this, tokenizer, context);
    }

    //--------------------------------------------------------------------
    // Complement
    //--------------------------------------------------------------------
    /**
     * Complement Command.
     */
    public void complement(CommandTokenizer tokenizer, Context context)
    {
        _possibles.clear();

        _fixedPosition = tokenizer.getPosition();
        String nextToken = tokenizer.nextToken();

        if( !tokenizer.hasMoreTokens() ){
            this.complementCommand(nextToken, context);
            // handle default node
            if(_default != null){
                tokenizer.movePrevious();
                _default.complement(tokenizer, context);
                _possibles.addAll( _default.getPossibles() );
            }
        }
        else{
            Command next = null;
            for(int i=0; i<_children.size(); i++){
                Command child = (Command)_children.get(i);
                if( child.parseName(nextToken) ){
                    next = child;
                    break;
                }
            }
            // handle default node
            if(next == null){
                next = _default;
                tokenizer.movePrevious();
            }

            if(next != null){
                next.complement(tokenizer, context);
                _possibles.addAll( next.getPossibles() );
                _fixedPosition = next.getFixedPosition();
            }
        }
    }

    private void complementCommand(String token, Context context)
    {
        token = token.toLowerCase();
        int len = token.length();

        for(int i=0; i<_children.size(); i++){
            Command child = (Command)_children.get(i);
            String[] names = child.getNames();
            for(int j=0; j<names.length; j++){
                if( len == 0 || 
                    names[j].length() >= len && 
                    names[j].substring(0, len).equals(token) )
                {
                    _possibles.add(names[j] + " ");
                    break;
                }
            }
        }
    }

    //--------------------------------------------------------------------
    // Usage
    //--------------------------------------------------------------------
    /**
     * Returns the collection of usages.
     */
    public ArrayList getUsage(CommandTokenizer tokenizer)
    {
        ArrayList list = new ArrayList();

        if(useUsage || _children.size() == 0){
            Usage[] usages = CommandUsage.create(this);
            if(usages != null){
                for(int i=0; i<usages.length; i++) list.add(usages[i]);
            }
            return list;
        }
        else{
            Command next = null;
            if( tokenizer.hasMoreTokens() ){
                String nextToken = tokenizer.nextToken();
                for(int i=0; i<_children.size(); i++){
                    Command child = (Command)_children.get(i);
                    if( child.parseName(nextToken) ){
                        next = child;
                        break;
                    }
                }
                if(next == null){
                    next = _default;
                    tokenizer.movePrevious();
                }
            }

            if(next == null){
                while(tokenizer.hasMoreTokens()) tokenizer.nextToken();
                if(_default != null) list.addAll( this.handleUsage(_default, tokenizer) );
                for(int i=0; i<_children.size(); i++){
                    Command child = (Command)_children.get(i);
                    if(child == _default) continue;
                    list.addAll( this.handleUsage(child, tokenizer) );
                }
            }
            else{
                list.addAll( this.handleUsage(next, tokenizer) );
            }
        }

        return list;
    }

    private ArrayList handleUsage(Command next, CommandTokenizer tokenizer)
    {
        ArrayList list = next.getUsage(tokenizer);
        for(int i=0; i<list.size(); i++){
            Usage usage = (Usage)list.get(i);
            String arg = next.getNames()[0];
            if(next == _default && !arg.equals("")) arg = "{" + arg + "}";
            usage.setArgument(arg + " " + usage.getArgument());
        }
        return list;
    }
}

