/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.commands;

import java.awt.Color;
import java.awt.Dimension;
import java.text.*;
import java.util.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.manipulators.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.mouse.*;
import org.pdbj.viewer.plus.*;
import org.pdbj.viewer.site.*;

/**
 * Command behavior.
 */
public class CommandBehavior extends CommandList
{
    public static final String INVALID_COMMAND = "Invalid Command Syntax.";
    public static final String INVALID_FILE_ID = "Invalid File ID.";
    public static final String INVALID_VALUE   = "Invalid Value.";

    private static String[] _trueTokens = {"on", "true"};
    private static String[] _falseTokens = {"off", "false"};

    private static final int AXIS_X = 1;
    private static final int AXIS_Y = 2;
    private static final int AXIS_Z = 3;

    /**
     * Pre-executes.
     */
    public static void preExecute(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] names = command.getNames();

        if (names == PDBJ_DESCRIBE)    prePdbj(command, tokenizer, context);
        else if(names == PDBJ_EXECUTE) prePdbj(command, tokenizer, context);
        else if(names == SELECT)       preSelect(command, tokenizer, context);
        else if(names == SELECTVERTEX) preSelect(command, tokenizer, context);
        else if(names == SHOW_GODATA)  preShowGOData(command, tokenizer, context);
        else if(names == SHOW_SITE)    preShowSite(command, tokenizer, context);
        else if(names == SHOW_XPS3)    preShowXps3(command, tokenizer, context);
    }
    
    /**
     * Executes.
     */
    public static void execute(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] names = command.getNames();
        
        if     (names == ADD_BACKBONE)   addBackbone(command, tokenizer, context);
        else if(names == ANIM_FORWARD)   animForward(command, tokenizer, context);
        else if(names == ANIM_BACKWARD)  animBackward(command, tokenizer, context);
        else if(names == ANIM_STOP)      animStop(command, tokenizer, context);
        else if(names == ANIMFRAME)      animframe(command, tokenizer, context);
        else if(names == ANIMSELECT)     animselect(command, tokenizer, context);
        else if(names == ANIMMODE_ONCE)  animmodeOnce(command, tokenizer, context);
        else if(names == ANIMMODE_LOOP)  animmodeLoop(command, tokenizer, context);
        else if(names == ANIMMODE_SWING) animmodeSwing(command, tokenizer, context);
        else if(names == ANIMRANGE)      animrange(command, tokenizer, context);
        else if(names == ANIMSPEED)      animspeed(command, tokenizer, context);
        else if(names == ANIMSTEP)       animstep(command, tokenizer, context);
        else if(names == BACKBONE)       backbone(command, tokenizer, context);
        else if(names == BACKGROUND)     background(command, tokenizer, context);
        else if(names == BALL_AND_STICK) ballAndStick(command, tokenizer, context);
        else if(names == CARTOON)        cartoon(command, tokenizer, context);
        else if(names == CENTER)         center(command, tokenizer, context);
        else if(names == COLOR)          color(command, tokenizer, context);
        else if(names == COLORVERTEX)    colorvertex(command, tokenizer, context);
        else if(names == CPK)            spacefill(command, tokenizer, context);
        else if(names == DEFINE)         define(command, tokenizer, context);
        else if(names == DISPLAY)        display(command, tokenizer, context);
        else if(names == DISPLAYATOM)    displayatom(command, tokenizer, context);
        else if(names == DISPLAYVERTEX)  displayvertex(command, tokenizer, context);
        else if(names == ECHO)           echo(command, tokenizer, context);
        else if(names == EXIT)           exit(command, tokenizer, context);
        else if(names == FIT)            fit(command, tokenizer, context);
        else if(names == FSELECT)        fselect(command, tokenizer, context);
        else if(names == HBOND)          hbond(command, tokenizer, context);
        else if(names == LOAD_EFSITE)    loadEfsite(command, tokenizer, context);
        else if(names == LOAD_PDBML)     loadPDBml(command, tokenizer, context);
        else if(names == LOAD_PDB)       loadPDB(command, tokenizer, context);
        else if(names == LOAD_POLYGON)   loadPolygon(command, tokenizer, context);
        else if(names == LOAD_POLYGON_CGO) loadPolygonCgo(command, tokenizer, context);
        else if(names == LOAD_ANIMATION) loadAnimation(command, tokenizer, context);
        else if(names == LOAD_FTP)       loadFtp(command, tokenizer, context);
        else if(names == PAUSE)          pause(command, tokenizer, context);
        else if(names == PDBJ_DESCRIBE)  pdbjDescribe(command, tokenizer, context);
        else if(names == PDBJ_EXECUTE)   pdbjExecute(command, tokenizer, context);
        else if(names == QUIT)           exit(command, tokenizer, context);
        else if(names == REFRESH)        refresh(command, tokenizer, context);
        else if(names == RESET_DEFAULT)  reset(command, tokenizer, context);
        else if(names == RESET_CARTOON)  resetCartoon(command, tokenizer, context);
        else if(names == RESET_TRANSPARENCY)
            resetTransparency(command, tokenizer, context, ViewerController.POLYGON_ALL); 
        else if(names == RESET_POINT_TRANSPARENCY)
            resetTransparency(command, tokenizer, context, ViewerController.POLYGON_POINTS);
        else if(names == RESET_LINE_TRANSPARENCY)
            resetTransparency(command, tokenizer, context, ViewerController.POLYGON_LINES);
        else if(names == RESET_TRIANGLE_TRANSPARENCY)
            resetTransparency(command, tokenizer, context, ViewerController.POLYGON_TRIANGLES);
        else if(names == RESET_QUAD_TRANSPARENCY)
            resetTransparency(command, tokenizer, context, ViewerController.POLYGON_QUADS);
        else if(names == RESET_POLYLINE_TRANSPARENCY)
            resetTransparency(command, tokenizer, context, ViewerController.POLYGON_POLYLINES);
        else if(names == RESET_POINT_SIZE)
            resetPolygonSize(command, tokenizer, context, ViewerController.POLYGON_POINTS);
        else if(names == RESET_LINE_WIDTH)
            resetPolygonSize(command, tokenizer, context, ViewerController.POLYGON_LINES);
        else if(names == RESET_POLYLINE_WIDTH)
            resetPolygonSize(command, tokenizer, context, ViewerController.POLYGON_POLYLINES);
        else if(names == RESET_PICKRADIUS) resetPickradius(command, tokenizer, context);
        else if(names == RIBBON)         ribbon(command, tokenizer, context);
        else if(names == ROTATE_X)       rotate(command, tokenizer, context, AXIS_X);
        else if(names == ROTATE_Y)       rotate(command, tokenizer, context, AXIS_Y);
        else if(names == ROTATE_Z)       rotate(command, tokenizer, context, AXIS_Z);
        else if(names == SAVE_PDB)       savePDB(command, tokenizer, context);
        else if(names == SAVE_SCRIPT)    saveScript(command, tokenizer, context);
        else if(names == SAVE_PNG)       savePng(command, tokenizer, context);
        else if(names == SAVE_JPEG)      saveJpeg(command, tokenizer, context);
        else if(names == SCRIPT)         script(command, tokenizer, context);
        else if(names == SELECT)         select(command, tokenizer, context);
        else if(names == SELECTVERTEX)   selectvertex(command, tokenizer, context);
        else if(names == SET_ADJUSTVIEW_ON)          setAdjustviewOn(command, tokenizer, context);
        else if(names == SET_ADJUSTVIEW_OFF)         setAdjustviewOff(command, tokenizer, context);
        else if(names == SET_AMBIENT)                setAmbient(command, tokenizer, context);
        else if(names == SET_BACKGROUND)             background(command, tokenizer, context);
        else if(names == SET_BONDMODE_AND)           setBondmodeAnd(command, tokenizer, context);
        else if(names == SET_BONDMODE_OR)            setBondmodeOr(command, tokenizer, context);
        else if(names == SET_CARTOON)                setCartoon(command, tokenizer, context);
        else if(names == SET_CARTOON_LOOP_TUBE_ON)   setCartoonLoopTubeOn(command, tokenizer, context);
        else if(names == SET_CARTOON_LOOP_TUBE_OFF)  setCartoonLoopTubeOff(command, tokenizer, context);
        else if(names == SET_CARTOON_ROUND_ON)       setCartoonRoundOn(command, tokenizer, context);
        else if(names == SET_CARTOON_ROUND_OFF)      setCartoonRoundOff(command, tokenizer, context);
        else if(names == SET_CENTER)                 setCenter(command, tokenizer, context);
        else if(names == SET_DIFFUSE1)               setDiffuse1(command, tokenizer, context);
        else if(names == SET_DIFFUSE2)               setDiffuse2(command, tokenizer, context);
        else if(names == SET_DIFFUSE1_DIRECTION)     setDiffuse1Direction(command, tokenizer, context);
        else if(names == SET_DIFFUSE2_DIRECTION)     setDiffuse2Direction(command, tokenizer, context);
        else if(names == SET_DRAWLEVEL)              setDrawlevel(command, tokenizer, context);
        else if(names == SET_EFSITE_URL)             setEfsiteUrl(command, tokenizer, context);
        else if(names == SET_EXT_SITE_URL)           setExtSiteUrl(command, tokenizer, context);
        else if(names == SET_HBOND_BACKBONE)         setHBondBackbone(command, tokenizer, context);
        else if(names == SET_HBOND_SIDECHAIN)        setHBondSidechain(command, tokenizer, context);
        else if(names == SET_HETERO_ON)              setHeteroOn(command, tokenizer, context);
        else if(names == SET_HETERO_OFF)             setHeteroOff(command, tokenizer, context);
        else if(names == SET_HYDROGEN_ON)            setHydrogenOn(command, tokenizer, context);
        else if(names == SET_HYDROGEN_OFF)           setHydrogenOff(command, tokenizer, context);
        else if(names == SET_IMAGESIZE)              setImagesize(command, tokenizer, context);
        else if(names == SET_LOADCENTER_ON)          setLoadcenterOn(command, tokenizer, context);
        else if(names == SET_LOADCENTER_OFF)         setLoadcenterOff(command, tokenizer, context);
        else if(names == SET_PDBML_NOATOM_URL)       setPdbmlNoatomUrl(command, tokenizer, context);
        else if(names == SET_PDBML_EXTATOM_URL)      setPdbmlExtatomUrl(command, tokenizer, context);
        else if(names == SET_PDBML_PLUS_URL)         setPdbmlPlusUrl(command, tokenizer, context);
        else if(names == SET_PICK_OFF)               setPickOff(command, tokenizer, context);
        else if(names == SET_PICK_IDENT)             setPickIdent(command, tokenizer, context);
        else if(names == SET_PICK_COORD)             setPickCoord(command, tokenizer, context);
        else if(names == SET_PICK_DISTANCE)          setPickDistance(command, tokenizer, context);
        else if(names == SET_PICK_CENTER)            setPickCenter(command, tokenizer, context);
        else if(names == SET_PICK_SELECT)            setPickSelect(command, tokenizer, context);
        else if(names == SET_PICKRADIUS)             setPickradius(command, tokenizer, context);
        else if(names == SET_PROJECTION_PERSPECTIVE) setProjectionPerspective(command, tokenizer, context);
        else if(names == SET_PROJECTION_PARALLEL)    setProjectionParallel(command, tokenizer, context);
        else if(names == SET_RIBBONBACK)             setRibbonback(command, tokenizer, context);
        else if(names == SET_SPECPOWER)              setSpecpower(command, tokenizer, context);
        else if(names == SET_SPECULAR)               setSpecular(command, tokenizer, context);
        else if(names == SET_SSBOND_BACKBONE)        setSSBondBackbone(command, tokenizer, context);
        else if(names == SET_SSBOND_SIDECHAIN)       setSSBondSidechain(command, tokenizer, context);
        else if(names == SET_STEREO)                 stereo(command, tokenizer, context);
        else if(names == SET_TRANSPARENCY)
            setTransparency(command, tokenizer, context, ViewerController.POLYGON_ALL);
        else if(names == SET_POINT_TRANSPARENCY)
            setTransparency(command, tokenizer, context, ViewerController.POLYGON_POINTS);
        else if(names == SET_LINE_TRANSPARENCY)
            setTransparency(command, tokenizer, context, ViewerController.POLYGON_LINES);
        else if(names == SET_TRIANGLE_TRANSPARENCY)
            setTransparency(command, tokenizer, context, ViewerController.POLYGON_TRIANGLES);
        else if(names == SET_QUAD_TRANSPARENCY)
            setTransparency(command, tokenizer, context, ViewerController.POLYGON_QUADS);
        else if(names == SET_POLYLINE_TRANSPARENCY)
            setTransparency(command, tokenizer, context, ViewerController.POLYGON_POLYLINES);
        else if(names == SET_POINT_SIZE)
            setPolygonSize(command, tokenizer, context, ViewerController.POLYGON_POINTS);
        else if(names == SET_LINE_WIDTH)
            setPolygonSize(command, tokenizer, context, ViewerController.POLYGON_LINES);
        else if(names == SET_POLYLINE_WIDTH)
            setPolygonSize(command, tokenizer, context, ViewerController.POLYGON_POLYLINES);
        else if(names == SET_VIEWPOINT)              setViewpoint(command, tokenizer, context);
        else if(names == SHOW_EFSITE_URL)  showEfsiteUrl(command, tokenizer, context);
        else if(names == SHOW_EXT_SITE_URL) showExtSiteUrl(command, tokenizer, context);
        else if(names == SHOW_GODATA)      showGOData(command, tokenizer, context);
        else if(names == SHOW_IMAGESIZE)   showImagesize(command, tokenizer, context);
        else if(names == SHOW_INFORMATION) showInformation(command, tokenizer, context);
        else if(names == SHOW_PDBML_URL)   showPdbmlUrl(command, tokenizer, context);
        else if(names == SHOW_SITE)        showSite(command, tokenizer, context);
        else if(names == SHOW_TRANSFORM)   showTransform(command, tokenizer, context);
        else if(names == SHOW_VIEWPOINT)   showViewpoint(command, tokenizer, context);
        else if(names == SHOW_XPS3)        showXps3(command, tokenizer, context);
        else if(names == SLAB)             slab(command, tokenizer, context);
        else if(names == SPACEFILL)        spacefill(command, tokenizer, context);
        else if(names == SSBOND)           ssbond(command, tokenizer, context);
        else if(names == STEREO)           stereo(command, tokenizer, context);
        else if(names == STRUCTURE)        structure(command, tokenizer, context);
        else if(names == TRACE)            trace(command, tokenizer, context);
        else if(names == TRANSLATE_X)      translate(command, tokenizer, context, AXIS_X);
        else if(names == TRANSLATE_Y)      translate(command, tokenizer, context, AXIS_Y);
        else if(names == TRANSLATE_Z)      translate(command, tokenizer, context, AXIS_Z);
        else if(names == WIREFRAME)        wireframe(command, tokenizer, context);
        else if(names == ZAP)              zap(command, tokenizer, context);
        else if(names == ZOOM)             zoom(command, tokenizer, context);
    }
    
    //--------------------------------------------------------------------
    // Execute
    //--------------------------------------------------------------------
    private static void addBackbone(
            Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        if (args.length == 0) {
            StringBuffer sb = new StringBuffer();
            sb.append("amino:CA,C,O,N\n");
            sb.append("nucleic:P,OP1,OP2,O5',C5',C4',O4',C3',O3',C2',O2',C1'");
            List residues = context.getBackboneAtomSet().getDefinedResidueNames();
            if (residues != null) {
                Collections.sort(residues);
                Iterator it = residues.iterator();
                while (it.hasNext()) {
                    String residue = (String) it.next();
                    sb.append("\n").append(residue).append(":");
                    sb.append(context.getBackboneAtomSet().getBackboneAtomsFormatted(residue));
                }
                command.message = sb.toString();
            }
        }
        else if (args.length == 1) {
            context.getBackboneAtomSet().clearBackboneAtoms(args[0]);            
        }
        else if (args.length == 2) {
            context.getBackboneAtomSet().setBackboneAtoms(args[0], args[1].split(","));
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animForward(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") ){
            context.getAnimationController().start(false);
        }
        else if( isIntToken(arg) ){
            int step = Integer.parseInt(arg);
            if(step > 0){
                context.getAnimationController().move(step, false);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animBackward(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") ){
            context.getAnimationController().start(true);
        }
        else if( isIntToken(arg) ){
            int step = Integer.parseInt(arg);
            if(step > 0){
                context.getAnimationController().move(step, true);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animStop(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") ){
            context.getAnimationController().stop();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animframe(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);

        boolean err = false;
        int file = 0;
        int frame = 0;

        if(args.length > 0 && isIntToken(args[0])){
            frame = Integer.parseInt(args[0]);
            if(args.length == 1){
                file = 0;
            }
            else if(args.length == 2 && isIntToken(args[1])){
                file = Integer.parseInt(args[1]);
            }
            else err = true;
        }
        else err = true;

        if(!err){
            if(frame > 0){
                if(file == 0){
                    context.getAnimationController().setFrame(frame);
                }
                else{
                    boolean result = context.getAnimationController().setFrame(frame, file);
                    if(!result) command.errorMessage = INVALID_FILE_ID;
                }
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animselect(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        boolean exclusive = true;
        boolean enable = true;

        if( tokenizer.hasMoreTokens() ){
            String opt = tokenizer.nextToken().toLowerCase();
            if( opt.equals("-a") ){
                exclusive = false;
                arg = currentSubstring(tokenizer);
            }
            else if( opt.equals("-r") ){
                exclusive = false;
                enable = false;
                arg = currentSubstring(tokenizer);
            }
        }

        int[] ids = splitInt(arg, ',');
        if(ids != null){
            boolean result = false;
            if(exclusive){
                result = context.getAnimationController().select(ids);
            }
            else{
                result = context.getAnimationController().select(enable, ids);
            }
            if(!result) command.errorMessage = INVALID_FILE_ID;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animmodeOnce(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getAnimationController().setOnceMode();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animmodeLoop(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getAnimationController().setLoopMode();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animmodeSwing(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getAnimationController().setSwingMode();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animrange(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);

        boolean err = false;
        int file = 0;
        int init = 0;
        int term = 0;

        if(args.length > 0)
        {
            int index = args[0].indexOf("-");
            if(index != -1){
                String initExpr = args[0].substring(0, index);
                String termExpr = args[0].substring(index+1);
                if( isIntToken(initExpr) && isIntToken(termExpr) ){
                    init = Integer.parseInt(initExpr);
                    term = Integer.parseInt(termExpr);
                    if(init <= 0 || term < init) init = term = 0;
                }
            }

            if(args.length == 1){
                file = 0;
            }
            else if(args.length == 2 && isIntToken(args[1])){
                file = Integer.parseInt(args[1]);
            }
            else err = true;
        }
        else err = true;

        if(!err){
            if(init > 0){
                if(file == 0){
                    context.getAnimationController().setRange(init, term);
                }
                else{
                    boolean result = context.getAnimationController().setRange(init, term, file);
                    if(!result) command.errorMessage = INVALID_FILE_ID;
                }
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animspeed(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( isIntToken(arg) ){
            int speed = Integer.parseInt(arg);
            if(speed >= AnimationController.MIN_SPEED && speed <= AnimationController.MAX_SPEED){
                context.getAnimationController().setSpeed(speed);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void animstep(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);

        boolean err = false;
        int file = 0;
        int step = 0;

        if(args.length > 0 && isIntToken(args[0])){
            step = Integer.parseInt(args[0]);
            if(args.length == 1){
                file = 0;
            }
            else if(args.length == 2 && isIntToken(args[1])){
                file = Integer.parseInt(args[1]);
            }
            else err = true;
        }
        else err = true;

        if(!err){
            if(step > 0){
                if(file == 0){
                    context.getAnimationController().setStep(step);
                }
                else{
                    boolean result = context.getAnimationController().setStep(step, file);
                    if(!result) command.errorMessage = INVALID_FILE_ID;
                }
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void backbone(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().backbone(0);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().backboneOff();
        }
        else if( isFloatToken(arg) ){
            float radius = parseFloat(arg);
            if(radius >= 0 && radius <= ViewerController.MAX_BOND_RADIUS){
                context.getViewerController().backbone(radius);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void background(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        ColorExprParser parser = new ColorExprParser();
        Color color = parser.parseColor(arg);

        if(color != null){
            context.getViewerController().setBackground(color);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void ballAndStick(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().spacefill(0.5f);
            context.getViewerController().wireframe(0.2f);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().spacefillOff();
            context.getViewerController().wireframeOff();
        }
        else {
            String[] args = getArgs(tokenizer);            
            if (args.length == 2) {
                try {
                    float br = parseFloat(args[0]);
                    float sr = parseFloat(args[1]);
                    if (br >= 0 && br <= ViewerController.MAX_BALL_RADIUS
                            && sr >= 0 && sr <= ViewerController.MAX_BOND_RADIUS) {
                        context.getViewerController().spacefill(br);
                        context.getViewerController().wireframe(sr);
                    }
                    else command.errorMessage = INVALID_VALUE;
                } catch (NumberFormatException e) {
                    command.errorMessage = INVALID_COMMAND;
                }
            }
            else command.errorMessage = INVALID_COMMAND;
        }
    }

    private static void cartoon(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().cartoon();
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().cartoonOff();
        }
        else if( isFloatToken(arg) ){
            float width = parseFloat(arg);
            if(width >= 0 && width <= ViewerController.MAX_RIBBON_WIDTH){
                context.getViewerController().cartoon(width);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void center(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        boolean rot = true;
        boolean scr = true;

        if( tokenizer.hasMoreTokens() ){
            String opt = tokenizer.nextToken().toLowerCase();
            if( opt.equals("-r") ){
                scr = false;
                arg = currentSubstring(tokenizer);
            }
            else if( opt.equals("-s") ){
                rot = false;
                arg = currentSubstring(tokenizer);
            }
        }

        Vector3f v = new Vector3f();
        boolean mov = false;
        boolean err = false;

        if( arg.equals("") ){
            context.getViewerController().getViewStatus().getCurrentCenter(v);
        }
        else{
            SelectExprParser parser = new SelectExprParser();
            SelectExpr expr = parser.parse(arg, context, true);

            if(expr != null){
                float[] val = context.getViewerController().findCenter(expr);
                if(val != null){
                    v = new Vector3f(val[0], val[1], val[2]);
                }
                else return;
            }
            else{
                TupleFloatExprParser tep = new TupleFloatExprParser();

                if(rot){
                    float[] val = tep.parse(arg, "[", "]", 3, true);
                    if(val != null){
                        v = new Vector3f(val[0], val[1], val[2]);
                        mov = true;
                    }
                    else{
                        val = tep.parse(arg, "{", "}", 3);
                        if(val != null){
                            v = new Vector3f(val[0], val[1], val[2]);
                        }
                        else err = true;
                    }
                }
                else{
                    float[] val = tep.parse(arg, "[", "]", 2, true);
                    if(val != null){
                        v = new Vector3f(val[0], val[1], 0);
                        mov = true;
                    }
                    else{
                        val = tep.parse(arg, "{", "}", 2);
                        if(val != null){
                            v = new Vector3f(val[0], val[1], 0);
                        }
                        else err = true;
                    }
                }
            }
        }

        if(!err){
            if(!mov){
                if(rot) context.getViewerController().setRotationCenter(v);
                if(scr) context.getViewerController().setScreenCenter(v.x, v.y);
            }
            else{
                if(rot) context.getViewerController().moveRotationCenter(v);
                if(scr) context.getViewerController().moveScreenCenter(v.x, v.y);
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void color(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        ColorExprParser parser = new ColorExprParser();
        ColorExpr expr = parser.parse(arg);

        if(expr != null){
            context.getViewerController().setColor(expr);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void colorvertex(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("none") ){
            context.getViewerController().resetColorVertex();
        }
        else{
            ColorExprParser parser = new ColorExprParser();
            Color color = parser.parseColor(arg);

            if(color != null){
                context.getViewerController().setColorVertex(color);
            }
            else command.errorMessage = INVALID_COMMAND;
        }
    }

    private static void define(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        if( tokenizer.hasMoreTokens() ){
            String name = tokenizer.nextToken().toLowerCase();
            String expr = currentSubstring(tokenizer);

            SelectExprParser parser = new SelectExprParser();
            if(parser.parse(expr, context, false) != null){
                if( !context.getUserAtomSet().add(name, expr) ){
                    command.errorMessage = "Invalid Set Name.";
                }
            }
            else command.errorMessage = INVALID_COMMAND;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void display(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.toLowerCase().equals("all") ){
            context.getViewerController().displayAllFile(true);
            return;
        }
        else if( arg.toLowerCase().equals("none") ){
            context.getViewerController().displayAllFile(false);
            return;
        }

        boolean exclusive = true;
        boolean enable = true;

        if( tokenizer.hasMoreTokens() ){
            String opt = tokenizer.nextToken().toLowerCase();
            if( opt.equals("-a") ){
                exclusive = false;
                arg = currentSubstring(tokenizer);
            }
            else if( opt.equals("-r") ){
                exclusive = false;
                enable = false;
                arg = currentSubstring(tokenizer);
            }
        }

        int[] ids = splitInt(arg, ',');
        if(ids != null){
            boolean result = false;
            if(exclusive){
                result = context.getViewerController().displayFile(ids);
            }
            else{
                result = context.getViewerController().displayFile(enable, ids);
            }
            if(!result) command.errorMessage = INVALID_FILE_ID;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void displayatom(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().displayAtom(true);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().displayAtom(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void displayvertex(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().displayVertex(true);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().displayVertex(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void echo(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( !arg.equals("") ) command.message = dropQuot(arg);
    }

    private static void exit(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            //if( !context.getGuiController().isScriptMode() )
            {
                context.getGuiController().exit();
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void fit(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        
        int srcId = -1;
        int dstId = -1;
        
        if(args.length == 2){
            try{
                dstId = Integer.parseInt(args[0]);
                srcId = Integer.parseInt(args[1]);
            }
            catch(NumberFormatException e){
            }
        }
        if(srcId > 0 && dstId > 0){
            boolean success = context.getViewerController().copyTransform(srcId, dstId);
            if(!success) command.errorMessage = INVALID_FILE_ID;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void fselect(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.toLowerCase().equals("all") ){
            context.getViewerController().selectAllFile(true);
            return;
        }
        else if( arg.toLowerCase().equals("none") ){
            context.getViewerController().selectAllFile(false);
            return;
        }

        boolean exclusive = true;
        boolean enable = true;

        if( tokenizer.hasMoreTokens() ){
            String opt = tokenizer.nextToken().toLowerCase();
            if( opt.equals("-a") ){
                exclusive = false;
                arg = currentSubstring(tokenizer);
            }
            else if( opt.equals("-r") ){
                exclusive = false;
                enable = false;
                arg = currentSubstring(tokenizer);
            }
        }

        int[] ids = splitInt(arg, ',');
        if(ids != null){
            boolean result = false;
            if(exclusive){
                result = context.getViewerController().selectFile(ids);
            }
            else{
                result = context.getViewerController().selectFile(enable, ids);
            }
            if(!result) command.errorMessage = INVALID_FILE_ID;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void hbond(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().hBond(0);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().hBondOff();
        }
        else if( isFloatToken(arg) ){
            float radius = parseFloat(arg);
            if(radius >= 0 && radius <= ViewerController.MAX_BOND_RADIUS){
                context.getViewerController().hBond(radius);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static boolean parseLoad(String[] args)
    {
        if(args.length == 1){
            return true;
        }
        else if(args.length == 3 && args[1].toLowerCase().equals("fit") && isIntToken(args[2]) ){
            return true;
        }
        return false;
    }
    
    private static FileStatus getCopy(int id, Context context)
    {
        ArrayList list = context.getViewerController().getFileStatusList();
        for(int i=0; i<list.size(); i++){
            FileStatus status = (FileStatus)list.get(i);
            if(status.getId() == id) return status;
        }
        return null;
    }

    private static void loadEfsite(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( !arg.equals("") ){
            if (arg.length() >= 4) {
                context.getGuiController().loadEfsite(arg);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void loadPDBml(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        FileStatus copy = null;
        if( parseLoad(args) ){
            if(args.length == 3){
                copy = getCopy( Integer.parseInt(args[2]), context);
            }
            if(args.length == 3 && copy == null){
                command.errorMessage = INVALID_FILE_ID;
            }
            else{
                context.getGuiController().loadPDBmlFile(args[0], copy);
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void loadPDB(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        FileStatus copy = null;
        if( parseLoad(args) ){
            if(args.length == 3){
                copy = getCopy( Integer.parseInt(args[2]), context);
            }
            if(args.length == 3 && copy == null){
                command.errorMessage = INVALID_FILE_ID;
            }
            else{
                context.getGuiController().loadPDBFile(args[0], copy);
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void loadPolygon(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        FileStatus copy = null;
        if( parseLoad(args) ){
            if(args.length == 3){
                copy = getCopy( Integer.parseInt(args[2]), context);
            }
            if(args.length == 3 && copy == null){
                command.errorMessage = INVALID_FILE_ID;
            }
            else{
                context.getGuiController().loadPolygonFile(args[0], copy);
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void loadPolygonCgo(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        FileStatus copy = null;
        if( parseLoad(args) ){
            if(args.length == 3){
                copy = getCopy( Integer.parseInt(args[2]), context);
            }
            if(args.length == 3 && copy == null){
                command.errorMessage = INVALID_FILE_ID;
            }
            else{
                context.getGuiController().loadPolygonCgoFile(args[0], copy);
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void loadAnimation(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        FileStatus copy = null;
        if( parseLoad(args) ){
            if(args.length == 3){
                copy = getCopy( Integer.parseInt(args[2]), context);
            }
            if(args.length == 3 && copy == null){
                command.errorMessage = INVALID_FILE_ID;
            }
            else{
                context.getGuiController().loadAnimationFile(args[0], copy);
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void loadFtp(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        FileStatus copy = null;
        if( parseLoad(args) ){
            String code = args[0].toLowerCase();
            if( checkPdb(code) ){
                if(args.length == 3){
                    copy = getCopy( Integer.parseInt(args[2]), context);
                }
                if(args.length == 3 && copy == null){
                    command.errorMessage = INVALID_FILE_ID;
                }
                else{
                    context.getGuiController().loadPDBmlFileFtp(code, copy);
                }
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static boolean checkPdb(String pdbCode)
    {
        if(pdbCode.length() != 4) return false;
        pdbCode = pdbCode.toLowerCase();
        for(int i=0; i<4; i++){
            char c = pdbCode.charAt(i);
            if( (c < '0' || c > '9' ) && ( c < 'a' || c > 'z') ) return false;
        }
        return true;
    }

    private static void pause(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getGuiController().pauseScript();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void pdbjDescribe(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);

        int fileId = 0;
        String name = null;
        if (args.length == 1) {
            if (isIntToken(args[0])) {
                fileId = Integer.parseInt(args[0]);
            }
        } else if (args.length == 2) {
            if (isIntToken(args[0])) {
                fileId = Integer.parseInt(args[0]);
                name = args[1];
            }
        }

        if (fileId > 0) {
            FileStatus file = context.getViewerController().findFile(fileId);
            if (file != null && file.getType() != FileStatus.POLYGON) {
                MoleculeFile moleculeFile = file.getMoleculeFile();
                if (!moleculeFile.isAnimation()) {
                    Function function = moleculeFile.getFunction();
                    StringBuffer buffer = new StringBuffer();
                    if (name == null) {
                        for (Iterator it = function.iteratorJvCommands(); it.hasNext();) {
                            JvCommand jvCommand = (JvCommand) it.next();
                            buffer.append(jvCommand.getName() + " ");
                        }
                    } else {
                        JvCommand jvCommand = function.getJvCommand(name);
                        if (jvCommand != null) {
                            buffer.append(jvCommand.getDescription());
                        } else {
                            command.errorMessage = "not found jV_command '" + name + "'.";
                        }
                    }
                    String message = buffer.toString().trim();
                    if( !message.equals("") ) command.message = message;
                }
            }
            
        } else command.errorMessage = INVALID_COMMAND;
    }

    private static void pdbjExecute(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);

        int fileId = 0;
        String name = null;
        if (args.length == 2) {
            if (isIntToken(args[0])) {
                fileId = Integer.parseInt(args[0]);
                name = args[1];
            }
        }

        if (fileId > 0 && name != null) {
            FileStatus file = context.getViewerController().findFile(fileId);
            if (file != null && file.getType() != FileStatus.POLYGON) {
                MoleculeFile moleculeFile = file.getMoleculeFile();
                if (!moleculeFile.isAnimation()) {
                    Function function = moleculeFile.getFunction();
                    JvCommand jvCommand = function.getJvCommand(name);
                    if (jvCommand != null) {
                        context.getGuiController().pdbjExecute(fileId, name);
                    } else {
                        command.errorMessage = "not found jV_command '" + name + "'.";
                    }
                }
            }

        } else command.errorMessage = INVALID_COMMAND;
    }

    private static void refresh(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().refresh();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void reset(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().reset();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void resetCartoon(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetCartoonHeight();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void resetTransparency(
        Command command, CommandTokenizer tokenizer, Context context, int target)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetTransparency(target);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void resetPolygonSize(
        Command command, CommandTokenizer tokenizer, Context context, int target)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetPolygonSize(target);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void resetPickradius(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getMouseBehavior().setUsePickRadius(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void ribbon(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().ribbon();
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().ribbonOff();
        }
        else if( isFloatToken(arg) ){
            float width = parseFloat(arg);
            if(width >= 0 && width <= ViewerController.MAX_RIBBON_WIDTH){
                context.getViewerController().ribbon(width);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void rotate(
        Command command, CommandTokenizer tokenizer, Context context, int axis)
    {
        String arg = currentSubstring(tokenizer);

        if( isFloatToken(arg) ){
            float angle = (float)( Float.parseFloat(arg)*Math.PI/180 );
            if(axis == AXIS_X){
                context.getViewerController().rotateX(angle);
            }
            else if(axis == AXIS_Y){
                context.getViewerController().rotateY(angle);
            }
            else if(axis == AXIS_Z){
                context.getViewerController().rotateZ(angle);
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void savePDB(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( !arg.equals("") ){
            context.getGuiController().savePDBFile(arg);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void saveScript(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( !arg.equals("") ){
            context.getGuiController().saveScriptFile(arg);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void savePng(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);

        float ratio = 0;

        if(args.length == 1){
            ratio = 1;
        }
        /*
        else if(args.length == 2){
            if( isFloatToken(args[1]) ) ratio = Float.parseFloat(args[1]);
        }
        */

        if(ratio > 0){
            context.getGuiController().savePngFile(args[0], ratio);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void saveJpeg(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);

        float ratio = 0;
        float quality = 0;

        if(args.length == 1){
            ratio = 1;
            quality = 0.75f;
        }
        else if(args.length == 2){
            ratio = 1;
            if( isFloatToken(args[1]) ) quality = Float.parseFloat(args[1]);
            /*
            if( isFloatToken(args[1]) ) ratio = Float.parseFloat(args[1]);
            quality = 0.75f;
            */
        }
        /*
        else if(args.length == 3){
            if( isFloatToken(args[1]) ) ratio = Float.parseFloat(args[1]);
            if( isFloatToken(args[2]) ) quality = Float.parseFloat(args[2]);
        }
        */

        if(ratio > 0 && quality > 0 && quality <= 1){
            context.getGuiController().saveJpegFile(args[0], ratio, quality);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void script(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( !arg.equals("") ){
            context.getGuiController().loadScriptFile(arg);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void select(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        int cnt = 0;
        if( arg.equals("") ){
            cnt = context.getViewerController().select(false);
        }
        else if( arg.equalsIgnoreCase("all") ){
            cnt = context.getViewerController().select(true);            
        }
        else if( arg.equalsIgnoreCase("none") ){
            context.getViewerController().selectOff();            
        }
        else{
            SelectExprParser parser = new SelectExprParser();
            SelectExpr expr = parser.parse(arg, context, true);
            if(expr != null){
                cnt = context.getViewerController().select(expr);
            }
            else{
                command.errorMessage = INVALID_COMMAND;
                return;
            }
        }

        if( !context.getGuiController().isScriptMode() ){
            if(cnt == 0){
                command.message = "No Atoms Selected.";
            }
            else if(cnt == 1){
                command.message = cnt + " Atom Selected.";
            }
            else{
                command.message = cnt + " Atoms Selected.";
            }
        }
    }

    private static void selectvertex(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg1 = currentSubstring(tokenizer);
        String arg2 = arg1.toLowerCase();

        int cnt = 0;
        if( arg2.equals("") || arg2.equals("all") ){
            cnt = context.getViewerController().selectVertexAll();
        }
        else if( arg2.equals("none") ){
            cnt = context.getViewerController().selectVertexNone();            
        }
        else{
            SelectExprParser parser = new SelectExprParser(true);
            SelectExpr expr = parser.parse(arg1, context, true);
            if(expr != null){
                cnt = context.getViewerController().selectVertex(expr);
            }
            else{
                command.errorMessage = INVALID_COMMAND;
                return;
            }
        }

        if( !context.getGuiController().isScriptMode() ){
            if(cnt == 0){
                command.message = "No Vertices Selected.";
            }
            else if(cnt == 1){
                command.message = cnt + " Vertex Selected.";
            }
            else{
                command.message = cnt + " Vertices Selected.";
            }
        }
    }

    private static void setAdjustviewOn(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setAdjustView(true);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setAdjustviewOff(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setAdjustView(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setAmbient(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetAmbient();
        }
        else if( isFloatToken(arg) ){
            float value = Float.parseFloat(arg);
            if (value >= 0 && value <= 100)
            {
                context.getViewerController().setAmbient(value/100);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setBondmodeAnd(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().getSelectionStatus().setBondMode(
                SelectionStatus.BOND_AND);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setBondmodeOr(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().getSelectionStatus().setBondMode(
                SelectionStatus.BOND_OR);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setCartoon(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetCartoonHeight();
        }
        else if( isFloatToken(arg) ){
            float height = parseFloat(arg);
            if( height >= ViewerController.MIN_CARTOON_HEIGHT && 
                height <= ViewerController.MAX_CARTOON_HEIGHT )
            {
                context.getViewerController().setCartoonHeight(height);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setCartoonLoopTubeOn(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setCartoonLoopTube(true);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setCartoonLoopTubeOff(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setCartoonLoopTube(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setCartoonRoundOn(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setCartoonRound(true);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setCartoonRoundOff(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setCartoonRound(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setCenter(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        float[] val = null;
        SelectExprParser parser = new SelectExprParser();
        SelectExpr expr = parser.parse(arg, context, true);

        if(expr != null){
            val = context.getViewerController().findCenter(expr);
            if(val == null) return;
        }
        else{
            TupleFloatExprParser tep = new TupleFloatExprParser();
            val = tep.parse(arg, "[", "]", 3, false);
        }

        if(val != null){
            Vector3f center = new Vector3f(val[0], val[1], val[2]);
            context.getViewerController().setCenter(center);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setDiffuse1(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetDiffuse1();
        }
        else if( isFloatToken(arg) ){
            float value = Float.parseFloat(arg);
            if (value >= 0 && value <= 100)
            {
                context.getViewerController().setDiffuse1(value/100);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setDiffuse2(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetDiffuse2();
        }
        else if( isFloatToken(arg) ){
            float value = Float.parseFloat(arg);
            if (value >= 0 && value <= 100)
            {
                context.getViewerController().setDiffuse2(value/100);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setDiffuse1Direction(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetDiffuseDir1();
        }
        else {
            TupleFloatExprParser tep = new TupleFloatExprParser();
            float[] val = tep.parse(arg, "[", "]", 3, false);
            if(val != null){
                context.getViewerController().setDiffuseDir1(val[0], val[1], val[2]);
            }
            else command.errorMessage = INVALID_COMMAND;
        }
    }

    private static void setDiffuse2Direction(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetDiffuseDir2();
        }
        else {
            TupleFloatExprParser tep = new TupleFloatExprParser();
            float[] val = tep.parse(arg, "[", "]", 3, false);
            if(val != null){
                context.getViewerController().setDiffuseDir2(val[0], val[1], val[2]);
            }
            else command.errorMessage = INVALID_COMMAND;
        }
    }
    
    private static void setDrawlevel(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( isIntToken(arg) ){
            int level = Integer.parseInt(arg);
            if(level >= ViewerController.MIN_DRAWLEVEL){
                context.getViewerController().setDrawlevel(level);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setEfsiteUrl(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( !arg.equals("") ){
            if (!context.getViewerController().getEfsiteUrl().setUrl(arg)) {
                command.errorMessage = INVALID_VALUE;
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setExtSiteUrl(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        if (args.length == 2) {
            if (!context.getViewerController().getSiteUrlMap().addUrl(
                    args[0].toLowerCase(), args[1])) {
                command.errorMessage = INVALID_VALUE;
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setHBondBackbone(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setHBondBackbone(true);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setHBondSidechain(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setHBondBackbone(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setHeteroOn(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().getSelectionStatus().setHetero(true);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setHeteroOff(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().getSelectionStatus().setHetero(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setHydrogenOn(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().getSelectionStatus().setHydrogen(true);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setHydrogenOff(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().getSelectionStatus().setHydrogen(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setImagesize(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        
        if (args.length == 2) {
            try {
                int w = Integer.parseInt(args[0]);
                int h = Integer.parseInt(args[1]);
                if (w >= 10 && h >= 10) {
                    Dimension d1 = new Dimension(w, h);
                    Dimension d2 = context.getGuiController().setImageSize(d1);
                    if (d2 != null) {
                        if (d2.width != w || d2.height != h) {
                            command.message = "set to " + d2.width + " x " + d2.height;
                        }
                    }
                } else command.errorMessage = INVALID_VALUE;
            } catch (NumberFormatException e) {
                command.errorMessage = INVALID_COMMAND;
            }
        } else command.errorMessage = INVALID_COMMAND;
    }

    private static void setLoadcenterOn(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setLoadCenter(true);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setLoadcenterOff(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setLoadCenter(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPdbmlNoatomUrl(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( !arg.equals("") ){
            if (!context.getViewerController().getPDBmlUrl().setUrlNoatom(arg)) {
                command.errorMessage = INVALID_VALUE;
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPdbmlExtatomUrl(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( !arg.equals("") ){
            if (!context.getViewerController().getPDBmlUrl().setUrlExtatom(arg)) {
                command.errorMessage = INVALID_VALUE;
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPdbmlPlusUrl(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( !arg.equals("") ){
            if (!context.getViewerController().getPDBmlUrl().setUrlFunction(arg)) {
                command.errorMessage = INVALID_VALUE;
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPickOff(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getMouseBehavior().setPickingMode(MouseBehavior.PICKING_OFF);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPickIdent(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getMouseBehavior().setPickingMode(MouseBehavior.PICKING_ID);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPickCoord(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getMouseBehavior().setPickingMode(MouseBehavior.PICKING_COORDINATES);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPickDistance(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getMouseBehavior().setPickingMode(MouseBehavior.PICKING_DISTANCE);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPickCenter(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getMouseBehavior().setPickingMode(MouseBehavior.PICKING_CENTER);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPickSelect(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getMouseBehavior().setPickingMode(MouseBehavior.PICKING_SELECT);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPickradius(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( isFloatToken(arg) ){
            float radius = parseFloat(arg);
            if(radius > 0) 
            {
                context.getMouseBehavior().setUsePickRadius(true);
                context.getMouseBehavior().setPickRadius(radius);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setProjectionPerspective(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setPerspectiveProjection();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setProjectionParallel(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") ){
            context.getViewerController().setParallelProjection();
        }
        else if( isFloatToken(arg) ){
            float size = parseFloat(arg);
            if(size > 0){
                context.getViewerController().setParallelProjection(size/2);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setRibbonback(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("none") ){
            context.getViewerController().resetColorRibbonback();
        }
        else{
            ColorExprParser parser = new ColorExprParser();
            Color color = parser.parseColor(arg);

            if(color != null){
                context.getViewerController().setColorRibbonback(color);
            }
            else command.errorMessage = INVALID_COMMAND;
        }
    }

    private static void setSpecpower(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setShininess(
                ViewerController.DEFAULT_SHININESS);
        }
        else if( isFloatToken(arg) ){
            float power =  Float.parseFloat(arg)/100;
            if(power >= 0 && power <= 1){
                context.getViewerController().setShininess(1+127*power);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setSpecular(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( isTrueToken(arg) ){
            context.getViewerController().setSpecular(true);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().setSpecular(false);
        }
        else if( isFloatToken(arg) ){
            float c = Float.parseFloat(arg)/100;
            if(c >= 0 && c <= 1){
                context.getViewerController().setSpecularColor(c, c, c, 1);
                context.getViewerController().setSpecular(true);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setSSBondBackbone(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setSSBondBackbone(true);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setSSBondSidechain(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().setSSBondBackbone(false);
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setTransparency(
        Command command, CommandTokenizer tokenizer, Context context, int target)
    {
        String arg = currentSubstring(tokenizer);

        if( isFloatToken(arg) ){
            float transparency = Float.parseFloat(arg);
            if( transparency >= 0 && transparency <= 1 )
            {
                context.getViewerController().setTransparency(target, transparency);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setPolygonSize(
        Command command, CommandTokenizer tokenizer, Context context, int target)
    {
        String arg = currentSubstring(tokenizer);

        if( isFloatToken(arg) ){
            float size = Float.parseFloat(arg);
            if(size > 0)
            {
                context.getViewerController().setPolygonSize(target, size);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void setViewpoint(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        
        TupleFloatExprParser tep = new TupleFloatExprParser();
        float[] val = tep.parse(arg, "{", "}", 3, false);        
        if (val != null) {
            context.getViewerController().setViewpoint(val[0], val[1], val[2]);
        } else command.errorMessage = INVALID_COMMAND;
    }

    private static void showEfsiteUrl(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( arg.equals("") ){
            command.message = "efsite=" + context.getViewerController().getEfsiteUrl().getUrl();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void showExtSiteUrl(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( arg.equals("") ){
            StringBuffer buffer = new StringBuffer();
            Map map = context.getViewerController().getSiteUrlMap().getMap();
            Iterator it = map.keySet().iterator();
            boolean first = true;
            while (it.hasNext()) {
                if (!first) {
                    buffer.append("\n");
                }
                String key = (String) it.next();
                String val = (String) map.get(key);
                buffer.append(key + "=" + val);
                first = false;
            }
            String message = buffer.toString();
            if (!message.equals("")) {
                command.message = message;
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void showGOData(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        StringBuffer buffer = new StringBuffer();

        ArrayList list = context.getViewerController().getFileStatusList();
        for(int i=0; i<list.size(); i++){
            FileStatus file = (FileStatus)list.get(i);
            if(file.getType() != FileStatus.POLYGON && file.isSelected()){
                MoleculeFile moleculeFile = file.getMoleculeFile();
                if(!moleculeFile.isAnimation()){
                    buffer.append("File " + moleculeFile.getFileId() + ":\n");
                    Function function = moleculeFile.getFunction();
                    if(function != null && function != Function.UNDEFINED){
                        for(int j=0; j<function.numGORefs(); j++){
                            buffer.append("  " + function.getGORef(j).toString() + "\n");
                        }
                    }
                }
            }
        }
        
        String message = buffer.toString().trim();
        if( !message.equals("") ) command.message = message;
    }

    private static void showImagesize(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            Dimension d = context.getGuiController().getImageSize();
            command.message = d.width + " x " + d.height;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void showInformation(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            String[] infos = context.getViewerController().getInformation();
    
            StringBuffer sb = new StringBuffer();
            for(int i=0; i<infos.length; i++){
                sb.append(infos[i]);
            }
            String message = sb.toString().trim();
            if( !message.equals("") ) command.message = message;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void showPdbmlUrl(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( arg.equals("") ){
            StringBuffer buffer = new StringBuffer();
            buffer.append("pdbml_noatom=");
            buffer.append(context.getViewerController().getPDBmlUrl().getUrlNoatom());
            buffer.append("\n");
            buffer.append("pdbml_extatom=");
            buffer.append(context.getViewerController().getPDBmlUrl().getUrlExtatom());
            buffer.append("\n");
            buffer.append("pdbml_plus=");
            buffer.append(context.getViewerController().getPDBmlUrl().getUrlFunction());
            command.message = buffer.toString();
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void showSite(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        HashMap siteUrlMap = context.getGuiController().getSiteUrlMap();
        if (siteUrlMap == null) {
            return;
        }
        Set keys = siteUrlMap.keySet();
        String[] prefix = new String[keys.size()];
        keys.toArray(prefix);
        Arrays.sort(prefix);

        if (arg.equals("")) {
            StringBuffer buffer = new StringBuffer();
            for (int i = 0; i < prefix.length; i++) {
                if (i != 0) {
                    buffer.append("\n");
                }
                buffer.append(prefix[i] + "=" + siteUrlMap.get(prefix[i]));
            }
            String message = buffer.toString();
            if (!message.equals("")) {
                command.message = message;
            }
            return;
        }
        
        String[] path = new String[3];
        for (int i = 0; i < prefix.length; i++) {
            if (arg.equalsIgnoreCase(prefix[i])) {
                path[0] = prefix[i];
            } else if (arg.toLowerCase().startsWith(prefix[i] + ":")) {
                String s = arg.substring(prefix[i].length() + 1);
                if (s.equals("")) {
                    path[0] = prefix[i];
                    path[1] = "";
                    path[2] = "";
                } else {
                    int index = s.indexOf(":");
                    if (index == -1) {
                        path[0] = prefix[i];
                        path[1] = s;
                        path[2] = "";
                    } else if (s.indexOf(":", index + 1) == -1) {
                        path[0] = prefix[i];
                        path[1] = s.substring(0, index);
                        path[2] = s.substring(index + 1);
                    }
                }
            }
            if (path[0] != null) {
                String message = getSiteContents(path, context);
                if (!message.equals("")) {
                    command.message = message;
                }
                return;
            }
        }
        command.errorMessage = INVALID_COMMAND;
    }
    
    private static String getSiteContents(String[] path, Context context) {
        StringBuffer buffer = new StringBuffer();
        List list = context.getViewerController().getFileStatusList();
        Iterator it = list.iterator();
        while (it.hasNext()) {
            FileStatus file = (FileStatus) it.next();
            if (file.getType() == FileStatus.POLYGON || !file.isSelected()) {
                continue;
            }
            MoleculeFile moleculeFile = file.getMoleculeFile();
            if (moleculeFile.isAnimation()) {
                continue;
            }
            buffer.append("File " + moleculeFile.getFileId() + ":\n");
            SiteList siteList = moleculeFile.getSiteList(path[0]);
            if (siteList == null) {
                continue;
            }
            if (path[1] == null && path[2] == null) {
                createSiteKeywords(siteList, buffer);
            } else {
                createSiteContents(siteList, path, buffer);
            }
        }
        return buffer.toString().trim();
    }
    
    private static void createSiteKeywords(SiteList siteList,
            StringBuffer buffer) {
        String title1 = "db";
        String title2 = "category";
        int max1 = title1.length();
        int max2 = title2.length();

        HashMap map = new HashMap();
        List item1 = new ArrayList();
        List item2 = new ArrayList();
        Iterator it = siteList.iterator();
        while (it.hasNext()) {
            Site site = (Site) it.next();
            String db = site.getDb();
            String category = site.getCategory();
            String key = db + " " + category;
            if (map.get(key) == null) {
                map.put(key, "");
                item1.add(db);
                item2.add(category);
                max1 = (db.length() > max1) ? db.length() : max1;
                max2 = (category.length() > max2) ? category.length() : max2;
            }
        }
        
        int size = map.size();
        if (size != 0) {
            buffer.append("  ");
            appendKeyword(buffer, title1, max1);
            buffer.append("  ");
            appendKeyword(buffer, title2, max2);
            buffer.append("\n");
            buffer.append("  ");
            int num = max1 + max2 + 2;
            for (int i = 0; i < num; i++){
                buffer.append("-");
            }
            buffer.append("\n");
            for (int i = 0; i < size; i++) {
                String db = (String) item1.get(i);
                String category = (String) item2.get(i);
                buffer.append("  ");
                appendKeyword(buffer, db, max1);
                buffer.append("  ");
                appendKeyword(buffer, category, max2);
                buffer.append("\n");
            }
        }
    }

    private static void createSiteContents(SiteList siteList, String[] path,
            StringBuffer buffer) {
        Iterator it = siteList.iterator();
        while (it.hasNext()) {
            Site site = (Site) it.next();
            String db = site.getDb();
            String category = site.getCategory();
            if (!path[1].equals("") && !path[1].equals(db)) {
                continue;
            }
            if (!path[2].equals("") && !path[2].equals(category)) {
                continue;
            }
            buffer.append("  id=" + site.getId());
            buffer.append("  db=" + db);
            buffer.append("  category=" + category);
            buffer.append("  region=" + getRegion(site, "") + "\n");
            buffer.append("     description=" + site.getDescription() + "\n");
        }
    }
    
    private static String getRegion(Site site, String blankChain) {
        StringBuffer buffer = new StringBuffer();
        Iterator it = site.iterator();
        boolean first = true;
        while (it.hasNext()) {
            if (!first) {
                buffer.append(",");
            }
            Region region = (Region) it.next();
            String chain = region.getChainId();
            String beg = region.getBegSeqId();
            String end = region.getEndSeqId();
            if (chain.equals("")) {
                chain = blankChain;
            }
            buffer.append(chain);
            if (!beg.equals("")) {
                if (!chain.equals("")) {
                    buffer.append(":");
                }
                if (beg.equals(end)) {
                    buffer.append(beg);
                } else {
                    buffer.append(beg + "-" + end);
                }
            }
            first = false;
        }
        return buffer.toString();
    }

    private static void showTransform(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            StringBuffer sb = new StringBuffer();
    
            ArrayList list = context.getViewerController().getMoleculeFiles();
            for(int i=0; i<list.size(); i++){
                MoleculeFile file = (MoleculeFile)list.get(i);
                if( file.isSelected() )
                {
                    Matrix4f matrix = file.getMolecule(0).getImage().getTransform();
    
                    // determine center
                    SphereUtil util = new SphereUtil();
                    for(int m=0; m<file.numMolecules(); m++){
                        Molecule molecule = file.getMolecule(m);
                        Vector3f center = new Vector3f(
                            molecule.getCenterX(), molecule.getCenterY(), molecule.getCenterZ()
                        );
                        util.addSphere(center, molecule.getRadius());
                    }
                    Sphere sphere = util.getSphere();
                    sb.append("File " + file.getFileId() + "\n");
                    showMatrix(matrix, sphere.center(), sb);
                }
            }
            String message = sb.toString().trim();
            if( !message.equals("") ) command.message = message;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void showMatrix(
        Matrix4f matrix, Vector3f center, StringBuffer buffer)
    {
        NumberFormat format = NumberFormat.getNumberInstance();
        format.setMaximumFractionDigits(4);
        format.setMinimumFractionDigits(4);
        format.setGroupingUsed(false);

        String[][] values = new String[3][3];
        values[0][0] = format.format(matrix.m11);
        values[0][1] = format.format(matrix.m12);
        values[0][2] = format.format(matrix.m13);
        values[1][0] = format.format(matrix.m21);
        values[1][1] = format.format(matrix.m22);
        values[1][2] = format.format(matrix.m23);
        values[2][0] = format.format(matrix.m31);
        values[2][1] = format.format(matrix.m32);
        values[2][2] = format.format(matrix.m33);

        int[] len = new int[3];
        for(int i=0; i<3; i++){
            len[i] = values[0][i].length();
            for(int j=1; j<3; j++){
                if(values[j][i].length() > len[i]) len[i] = values[j][i].length();
            }
        }

        buffer.append(" rotation\n");
        for(int i=0; i<3; i++){
            buffer.append("  ");
            for(int j=0; j<3; j++){
                for(int k=0; k<len[j]-values[i][j].length(); k++){
                    buffer.append(" ");
                }
                buffer.append(values[i][j] + " ");
            }
            buffer.append("\n");
        }

        Matrix4f m = new Matrix4f();
        m.translation(center);
        m.multiply(matrix, m);
        
        float[] amount = {m.m14-center.x, m.m24-center.y, m.m34-center.z};
        buffer.append(" translation\n");
        buffer.append("  ");
        for(int i=0; i<3; i++){
            buffer.append(format.format(amount[i]) + " ");
        }
        buffer.append("\n");
    }

    private static void showViewpoint(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);
        if( arg.equals("") ){
            Vector3f v = new Vector3f();
            context.getViewerController().getViewStatus().getCameraPosition(v);
            StringBuffer sb = new StringBuffer();
            NumberFormat format = NumberFormat.getNumberInstance();
            format.setMaximumFractionDigits(4);
            format.setMinimumFractionDigits(4);
            format.setGroupingUsed(false);
            sb.append("  ");
            sb.append(format.format(v.x)).append(" ");
            sb.append(format.format(v.y)).append(" ");
            sb.append(format.format(v.z));
            String message = sb.toString();
            if( !message.equals("") ) command.message = message;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void showXps3(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        StringBuffer buffer = new StringBuffer();

        ArrayList list = context.getViewerController().getFileStatusList();
        for(int i=0; i<list.size(); i++){
            FileStatus file = (FileStatus)list.get(i);
            if(file.getType() != FileStatus.POLYGON && file.isSelected()){
                MoleculeFile moleculeFile = file.getMoleculeFile();
                if(!moleculeFile.isAnimation()){
                    buffer.append("File " + moleculeFile.getFileId() + ":\n");
                    Function function = moleculeFile.getFunction();
                    if(function != null && function != Function.UNDEFINED){
                        addKeywords(function, buffer);
                    }
                }
            }
        }
        
        String message = buffer.toString().trim();
        if( !message.equals("") ) command.message = message;
    }

    private static void addKeywords(Function function, StringBuffer buffer)
    {
        String title1 = "type";
        String title2 = "subtype";

        int max1 = title1.length();
        int max2 = title2.length();

        ArrayList listType1 = new ArrayList();
        ArrayList listType2 = new ArrayList();
        for(int i=0; i<function.numStructSites(); i++){
            StructSite structSite = function.getStructSite(i);
            String type1 = fixType(structSite.getInfoType());
            String type2 = fixType(structSite.getInfoSubType());
            if(addKeyword(listType1, listType2, type1, type2)){
                if(type1.length() > max1){
                    max1 = type1.length();
                }
                if(type2.length() > max2){
                    max2 = type2.length();
                }
            }
        }

        if(listType1.size() > 0){
            buffer.append("  ");
            appendKeyword(buffer, title1, max1);
            buffer.append("  ");
            appendKeyword(buffer, title2, max2);
            buffer.append("\n");
            buffer.append("  ");
            int num = max1 + max2 + 2;
            for(int i=0; i<num; i++){
                buffer.append("-");
            }
            buffer.append("\n");
            for(int i=0; i<listType1.size(); i++){
                String type1 = (String) listType1.get(i);
                String type2 = (String) listType2.get(i);
                buffer.append("  ");
                appendKeyword(buffer, type1, max1);
                buffer.append("  ");
                appendKeyword(buffer, type2, max2);
                buffer.append("\n");
            }
        }
    }

    private static void appendKeyword(StringBuffer buffer, String type, int max)
    {
        buffer.append(type);
        int num = max - type.length();
        for(int i=0; i<num; i++){
            buffer.append(" ");
        }
    }

    private static boolean addKeyword(
        List list1, List list2, String type1, String type2)
    {
        if(type1.equals("") && type2.equals("")){
            return false;
        }

        for(int i=0; i<list1.size(); i++){
            if(type1.equals((String) list1.get(i))){
                if(type2.equals((String) list2.get(i))){
                    return false;
                }
            }
        }

        list1.add(type1);
        list2.add(type2);
        return true;
    }

    private static String fixType(String type)
    {
        return (type == null) ? "" : type;
    }

    private static void slab(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().slab();
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().slabOff();
        }
        else{
            boolean val = false;
            if( tokenizer.hasMoreTokens() ){
                String opt = tokenizer.nextToken().toLowerCase();
                if( opt.equals("-v") ){
                    val = true;
                    arg = currentSubstring(tokenizer);
                }
            }

            if( isFloatToken(arg) ){
                if(val){
                    context.getViewerController().slabValue( Float.parseFloat(arg) );
                }
                else{
                    float ratio = Float.parseFloat(arg)/100;
                    if(ratio >= 0 && ratio <= 1){
                        context.getViewerController().slabRatio(ratio);
                    }
                    else command.errorMessage = INVALID_VALUE;
                }
            }
            else command.errorMessage = INVALID_COMMAND;
        }
    }

    private static void spacefill(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().spacefillVdW();
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().spacefillOff();
        }
        else if( arg.equals("temperature") ){
            context.getViewerController().spacefillTemp();
        }
        else if( isFloatToken(arg) ){
            float radius = parseFloat(arg);
            if(radius >= 0 && radius <= ViewerController.MAX_BALL_RADIUS){
                context.getViewerController().spacefill(radius);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void ssbond(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().ssBond(0);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().ssBondOff();
        }
        else if( isFloatToken(arg) ){
            float radius = parseFloat(arg);
            if(radius >= 0 && radius <= ViewerController.MAX_BOND_RADIUS){
                context.getViewerController().ssBond(radius);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void stereo(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        String action = null;
        float angle = 0;
        ViewStatus status = context.getViewerController().getViewStatus();
        if (args.length == 0) {
            angle = Math.abs(status.getStereoAngle());
            if (status.isStereo()) {
                if (status.getStereoAngle() >= 0) {
                    action = "on";
                    angle *= -1;
                } else {
                    action = "redblue";                    
                }
            } else if (status.isAnaglyphOff()) {
                action = "on";
            } else if (status.isAnaglyphRedBlue()) {
                action = "redcyan";
            } else if (status.isAnaglyphRedCyan()) {
                action = "redgreen";
            } else {
                action = "off";
            }            
        } else if (args.length == 1) {
            if (isFloatToken(args[0])) {
                action = "on";
                angle = Float.parseFloat(args[0])/2;
            } else if (isTrueToken(args[0]) || isFalseToken(args[0])
                    || args[0].equalsIgnoreCase("redblue")
                    || args[0].equalsIgnoreCase("redcyan")
                    || args[0].equalsIgnoreCase("redgreen")) {
                action = args[0];
                angle = status.getStereoAngle();
            }
        } else if (args.length == 2) {
            if ((args[0].equalsIgnoreCase("redblue")
                    || args[0].equalsIgnoreCase("redcyan")
                    || args[0].equalsIgnoreCase("redgreen"))
                    && isFloatToken(args[1])) {
                action = args[0];
                angle = Float.parseFloat(args[1])/2;
            }
        }
        if (action == null) {
            command.errorMessage = INVALID_COMMAND;
            return;
        }

        if (context.getGuiController().isNewt() && isTrueToken(action)) {
            command.errorMessage = "Stereo view is unavailable in this environment.";
            return;
        }

        context.getViewerController().setStereoAngle(angle);
        if (isTrueToken(action)) {
            context.getGuiController().initializeStereo();
            context.getViewerController().stereo();
            context.getGuiController().setupStereo();
        } else {
            context.getViewerController().stereoOff();
            context.getGuiController().setupStereo();
            if (isFalseToken(action)) {
                context.getViewerController().anaglyphOff();
            } else if (action.equalsIgnoreCase("redblue")) {
                context.getViewerController().anaglyphRedBlue();
            } else if (action.equalsIgnoreCase("redcyan")) {
                context.getViewerController().anaglyphRedCyan();
            } else {
                context.getViewerController().anaglyphRedGreen();
            }
        }
    }

    private static void structure(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().structure();
    
            if( !context.getGuiController().isScriptMode() )
            {
                StringBuffer sb = new StringBuffer();
                ArrayList list = context.getViewerController().getMoleculeFiles();
                for(int i=0; i<list.size(); i++)
                {
                    MoleculeFile file = (MoleculeFile)list.get(i);
                    if( !file.isSelected() ) continue;
    
                    sb.append("File " +file.getFileId() + "\n");
                    for(int m=0; m<file.numMolecules(); m++)
                    {
                        Molecule molecule = file.getMolecule(m);
                        if(molecule.getModelId() != Molecule.NO_MODEL_ID){
                            sb.append("  Model " + molecule.getModelId() + ":  ");
                        }
                        sb.append("  Helices ... " + molecule.getHelixCount());
                        sb.append("  Strands ... " + molecule.getStrandCount());
                        sb.append("  Turns ... " + molecule.getTurnCount() + "\r\n");
                    }
                }
    
                String message = sb.toString().trim();
                if( !message.equals("") ) command.message = message;
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }
    private static void trace(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().trace(0.3f);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().traceOff();
        }
        else if( isFloatToken(arg) ){
            float radius = parseFloat(arg);
            if(radius >= 0 && radius <= ViewerController.MAX_BOND_RADIUS){
                context.getViewerController().trace(radius);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void translate(
        Command command, CommandTokenizer tokenizer, Context context, int axis)
    {
        String arg = currentSubstring(tokenizer);

        if( isFloatToken(arg) ){
            float amount = Float.parseFloat(arg);
            if(axis == AXIS_X){
                context.getViewerController().translate(amount, 0, 0);
            }
            else if(axis == AXIS_Y){
                context.getViewerController().translate(0, amount, 0);
            }
            else if(axis == AXIS_Z){
                context.getViewerController().translate(0, 0, amount);
            }
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void wireframe(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        if( arg.equals("") || isTrueToken(arg) ){
            context.getViewerController().wireframe(0);
        }
        else if( isFalseToken(arg) ){
            context.getViewerController().wireframeOff();
        }
        else if( isFloatToken(arg) ){
            float radius = parseFloat(arg);
            if(radius >= 0 && radius <= ViewerController.MAX_BOND_RADIUS){
                context.getViewerController().wireframe(radius);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    private static void zap(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().closeAllFile();
        }
        else{
            int[] ids = splitInt(arg, ',');
            if(ids != null){
                boolean result = context.getViewerController().closeFile(ids);
                if(!result) command.errorMessage = INVALID_FILE_ID;
            }
            else command.errorMessage = INVALID_COMMAND;
        }
    }

    private static void zoom(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer);

        if( arg.equals("") ){
            context.getViewerController().resetZoom();
        }
        else if( isFloatToken(arg) ){
            float ratio =  Float.parseFloat(arg)/100;
            if(ratio >= ViewerController.MIN_ZOOM_RATIO){
                context.getViewerController().zoom(ratio);
            }
            else command.errorMessage = INVALID_VALUE;
        }
        else command.errorMessage = INVALID_COMMAND;
    }

    //--------------------------------------------------------------------
    // Pre-Execute
    //--------------------------------------------------------------------
    private static void prePdbj(Command command, CommandTokenizer tokenizer, Context context)
    {
        String[] args = getArgs(tokenizer);
        if (args.length > 0)
        {
            if (isIntToken(args[0])) {
                FileStatus file = context.getViewerController().findFile(Integer.parseInt(args[0]));
                if(file != null && file.getType() != FileStatus.POLYGON)
                {
                    MoleculeFile moleculeFile = file.getMoleculeFile();
                    if( !moleculeFile.isAnimation() )
                    {
                        if(moleculeFile.getFunction() == null){
                            context.getGuiController().loadFunctionFtp(file);
                        }
                    }
                }
            }
        }
    }

    private static void preSelect(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = " " + currentSubstring(tokenizer).toLowerCase() + " ";

        StringBuffer buffer = new StringBuffer();
        buffer.append(arg);

        // append user set
        UserAtomSet userAtomSet = context.getUserAtomSet();
        Iterator it1 = userAtomSet.iteratorName();
        while (it1.hasNext()) {
            String name = (String) it1.next();
            if (testUserSet(name, arg)) {
                buffer.append(userAtomSet.valueOf(name) + " ");
            }
        }

        arg = buffer.toString();

        if (testPrefixColon("pdbj", arg) || testPrefixColon("xps3", arg)) {
            loadFunction(context, false);
        }

        HashMap siteUrlMap = context.getGuiController().getSiteUrlMap();
        if (siteUrlMap == null) {
            return;
        }
        Iterator it2 = siteUrlMap.keySet().iterator();
        while (it2.hasNext()) {
            String prefix = (String) it2.next();
            if (testPrefix(prefix, arg)) {
                loadSiteList(prefix, context, false);
            }
        }
    }
    
    private static boolean testPrefixColon(String prefix, String arg)
    {
        String target = prefix + ":";
        int index = arg.indexOf(target);
        while (index != -1) {
            if (arg.charAt(index - 1) != ':') {
                return true;
            }
            index = arg.indexOf(target, index + 1);
        }
        return false;
    }
    
    private static boolean testPrefix(String prefix, String arg)
    {
        int index = arg.indexOf(prefix);
        while (index != -1) {
            if (arg.charAt(index - 1) != ':') {
                char next = arg.charAt(index + prefix.length());
                if (next == ':' || testDelim(next)) {
                    return true;
                }
            }
            index = arg.indexOf(prefix, index + 1);
        }
        return false;
    }
    
    private static boolean testUserSet(String name, String arg) 
    {
        int index = arg.indexOf(name);
        while (index != -1) {
            char prev = arg.charAt(index - 1);
            char next = arg.charAt(index + name.length());
            if (testDelim(prev) && testDelim(next)) {
                return true;
            }
            index = arg.indexOf(name, index + 1);
        }
        return false;
    }
    
    private static boolean testDelim(char ch)
    {
        if (ch == ' ') {
            return true;
        } else if (ch == '=') {
            return true;
        } else if (ch == '<') {
            return true;
        } else if (ch == '>') {
            return true;
        } else if (ch == '!') {
            return true;
        } else if (ch == '/') {
            return true;
        } else if (ch == '(') {
            return true;
        } else if (ch == ')') {
            return true;
        } else if (ch == '&') {
            return true;
        } else if (ch == '|') {
            return true;
        } else if (ch == ',') {
            return true;
        } else if (ch == '~') {
            return true;
        }
        return false;
    }

    private static void preShowGOData(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        loadFunction(context, true);
    }

    private static void preShowXps3(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        loadFunction(context, true);
    }
    
    private static void preShowSite(
        Command command, CommandTokenizer tokenizer, Context context)
    {
        String arg = currentSubstring(tokenizer).toLowerCase();

        HashMap siteUrlMap = context.getGuiController().getSiteUrlMap();
        if (siteUrlMap == null) {
            return;
        }
        Iterator it = siteUrlMap.keySet().iterator();
        while (it.hasNext()) {
            String prefix = (String) it.next();
            if (arg.equals(prefix) || arg.startsWith(prefix + ":")) {
                loadSiteList(prefix, context, true);
            }
        }
    }
        
    private static void loadFunction(Context context, boolean selectedOnly)
    {
        ArrayList list = context.getViewerController().getFileStatusList();
        for(int i=0; i<list.size(); i++)
        {
            FileStatus file = (FileStatus)list.get(i);
            if(file.getType() != FileStatus.POLYGON)
            {
                if(!selectedOnly || file.isSelected())
                {
                    MoleculeFile moleculeFile = file.getMoleculeFile();
                    if( !moleculeFile.isAnimation() )
                    {
                        if(moleculeFile.getFunction() == null){
                            context.getGuiController().loadFunctionFtp(file);
                        }
                    }
                }
            }
        }
    }
    
    private static void loadSiteList(String prefix, Context context, 
            boolean selectedOnly) {
        ArrayList list = context.getViewerController().getFileStatusList();
        for (int i = 0; i < list.size(); i++) {
            FileStatus file = (FileStatus)list.get(i);
            if (file.getType() == FileStatus.POLYGON) {
                continue;
            }
            MoleculeFile moleculeFile = file.getMoleculeFile();
            if(moleculeFile.isAnimation()) {
                continue;
            }
            if (!selectedOnly || file.isSelected()) {
                if (moleculeFile.getSiteList(prefix) == null) {
                    context.getGuiController().loadSiteList(file, prefix);
                }
            }
        }
    }

    //--------------------------------------------------------------------
    // Utils for command parse
    //--------------------------------------------------------------------
    private static String currentSubstring(CommandTokenizer tokenizer)
    {
        if( tokenizer.hasMoreTokens() ){
            return tokenizer.getCommand().substring(tokenizer.getPosition()).trim();
        }
        else return "";
    }

    private static String[] getArgs(CommandTokenizer tokenizer)
    {
        int cnt = tokenizer.countTokens();
        if(tokenizer.endsWithSpace() && cnt !=0 ) cnt--;

        String[] args = new String[cnt];
        for(int i=0; i<cnt; i++){
            args[i] = tokenizer.nextToken();
        }

        return args;
    }

    private static boolean isTrueToken(String token)
    {
        token = token.toLowerCase();
        for(int i=0; i<_trueTokens.length; i++){
            if( _trueTokens[i].equals(token) ) return true;
        }
        return false;
    }

    private static boolean isFalseToken(String token)
    {
        token = token.toLowerCase();
        for(int i=0; i<_falseTokens.length; i++){
            if( _falseTokens[i].equals(token) ) return true;
        }
        return false;
    }

    private static boolean isFloatToken(String token)
    {
        try{
            Float.parseFloat(token);
        }
        catch(NumberFormatException e){
            return false;
        }
        return true;
    }

    private static boolean isIntToken(String token)
    {
        try{
            Integer.parseInt(token);
        }
        catch(NumberFormatException e){
            return false;
        }
        return true;
    }

    private static int[] splitInt(String expr, char delimiter)
    {
        ArrayList list = new ArrayList();
        int index = expr.indexOf(delimiter);
        while(index != -1){
            list.add( expr.substring(0, index) );
            expr = expr.substring(index+1);
            index = expr.indexOf(delimiter);
        }
        list.add(expr);

        int[] ids = new int[ list.size() ];
        try{
            for(int i=0; i<list.size(); i++){
                ids[i] = Integer.parseInt( ( (String)list.get(i) ).trim() );
            }
            return ids;
        }
        catch(NumberFormatException e){
            return null;
        }
    }

    private static String dropQuot(String command)
    {
        int len = command.length();
        if(len < 2) return command;

        char init = command.charAt(0);
        char term = command.charAt(len-1);
        if((init == '\"' && term == '\"') || (init == '\'' && term == '\'')){
            return command.substring(1, len-1).trim();
        }
        else{
            return command;
        }
    }
    
    private static float parseFloat(String token)
    {
        int index = token.indexOf(".");
        float val = Float.parseFloat(token);
        
        return (index == -1) ? val/250 : val;
    }
}
