/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.commands;

import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.mouse.*;

/**
 * Command executor.
 */
public class CommandExecutor
{
    private Context _context;
    private Command _root;

    private MessageArea _messageArea;
    private boolean     _messageEnabled;
    
    private String _currMessage;

    /**
     * Sets the message area.
     */
    public void setMessageArea(MessageArea messageArea)
    {
        _messageArea = messageArea;
    }

    /**
     * Returns the message area.
     */
    public MessageArea getMessageArea()
    {
        return _messageArea;
    }

    /**
     * Sets whether the message is showed.
     */
    public void setMessageEnabled(boolean enable)
    {
        _messageEnabled = enable;
    }

    /**
     * Returns whether the message is showed.
     */
    public boolean getMessageEnabled()
    {
        return _messageEnabled;
    }

    /**
     * Constructor.
     */
    public CommandExecutor(
        ViewerController viewerController, MouseBehavior mouseBehavior,
        GuiController guiController, AnimationController animController)
    {
        _context = new Context(
            viewerController, mouseBehavior, guiController, animController
        );
        _root = CommandTree.createRoot();

        _messageEnabled = true;
    }

    /**
     * Creates tokenizer.
     */
    public CommandTokenizer createTokenizer(String command)
    {
        if(command == null) return null;

        command = dropComment(command).trim();
        if( command.equals("") ) return null;

        return new CommandTokenizer(command);
    }
    
    /**
     * Executes command with a thread other than the swing event dispatching thread
     * before execution with the swing event dispatching thread.
     */
    public void preExecute(String command)
    {
        this.preExecute( this.createTokenizer(command) );
    }

    /**
     * Executes command with a thread other than the swing event dispatching thread
     * before execution with the swing event dispatching thread.
     */
    public void preExecute(CommandTokenizer tokenizer)
    {
        if(tokenizer == null) return;
        
        _root.preExecute(tokenizer, _context);

        if(_root.errorMessage != null){
            _currMessage = _root.errorMessage;
        }
        else if(_root.message != null){
            _currMessage = _root.message;
        }
    }

    /**
     * Executes command with the swing event dispatching thread.
     */
    public void execute(String command) throws RecoverableException
    {
        this.execute( this.createTokenizer(command) );
    }

    /**
     * Executes command with the swing event dispatching thread.
     */
    public void execute(CommandTokenizer tokenizer) throws RecoverableException
    {
        if(tokenizer == null) return;
        
        _root.execute(tokenizer, _context);

        if(_currMessage != null){
            this.showMessage(_currMessage);
            _currMessage = null;
        }

        if(_root.errorMessage != null){
            this.showMessage( _root.errorMessage );
        }
        else if(_root.message != null){
            this.showMessage( _root.message );
        }
    }

    /**
     * Complements command.
     */
    public CommandComplement complement(String command)
    {
        CommandTokenizer tokenizer = new CommandTokenizer(command);
        _root.complement(tokenizer, _context);

        CommandComplement complement = new CommandComplement();
        complement.setOriginal(command);
        complement.setFixedPosition( _root.getFixedPosition() );
        complement.setPossibles( _root.getPossibles() );

        return complement;
    }

    //--------------------------------------------------------------------
    // Message
    //--------------------------------------------------------------------
    private void showMessage(String message)
    {
        if(_messageArea != null && _messageEnabled){
            _messageArea.appendMessage(message);
        }
    }

    //--------------------------------------------------------------------
    // Utils
    //--------------------------------------------------------------------
    private static String dropComment(String command)
    {
        int index = command.indexOf("#");
        
        if(index == -1){
            return command;
        }
        else{
            return command.substring(0,index);
        }
    }
}
