/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.commands;

import java.util.*;
import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.manipulators.*;

/**
 * Parses select expressions.
 */
public class SelectExprParser
{
    private String[] _tokens;
    private int _index;
    private boolean _vertex;
    private boolean _filter;

    private String[] _prefix;

    private HashMap _siteUrlMap;
    private BackboneAtomSet _backboneAtomSet;
    private UserAtomSet _userAtomSet;

    /**
     * Constructor.
     */
    public SelectExprParser()
    {
    }

    /**
     * Constructor.
     */
    public SelectExprParser(boolean vertex)
    {
        _vertex = vertex;
    }

    /**
     * Parses select expressions.
     */
    public SelectExpr parse(String expr, Context context, boolean useUserAtomSet) 
    {
        _siteUrlMap = context.getGuiController().getSiteUrlMap();
        _backboneAtomSet = context.getBackboneAtomSet();
        if (useUserAtomSet) {
            _userAtomSet = context.getUserAtomSet();
        }
        return parse(expr);
    }

    public SelectExpr parseAsFilter(String expr) {
        _filter = true;
        _backboneAtomSet = new BackboneAtomSet();
        return parse(expr);
    }

    private SelectExpr parse(String expr)
    {
        if(expr == null) return null;

        StringSplitter splitter = new StringSplitter();
        String[] delimiters = new String[]{
            " ", "==", "<>", "!=", "/=", "<=", ">=", "=", "<", ">",
            "(", ")", "&&", "&", "||", "|", ",", "!", "~"
        };
        String[] tokens = splitter.splitTrim(expr, delimiters);

        if(tokens.length == 0) return null;

        // handle user atom set
        if(_userAtomSet != null)
        {
            ArrayList list = new ArrayList();
            for(int i=0; i<tokens.length; i++){
                String userExpr = _userAtomSet.valueOf(tokens[i].toLowerCase());
                if(userExpr != null){
                    String currExpr = "(" + userExpr + ")";
                    String[] array = splitter.splitTrim(currExpr, delimiters);
                    for(int j=0; j<array.length; j++){
                        list.add(array[j]);
                    }
                }
                else{
                    list.add(tokens[i]);
                }
            }
            _tokens = new String[list.size()];
            for(int i=0; i<list.size(); i++){
                _tokens[i] = (String)list.get(i);
            }            
        }
        else{
            _tokens = tokens;
        }

        if (_siteUrlMap != null) {
            Set keys = _siteUrlMap.keySet();
            _prefix = new String[keys.size()];
            keys.toArray(_prefix);
        } else {
            _prefix = new String[0];
        }

        _index = 0;

        SelectExpr selectExpr = null;
        try{
            selectExpr = this.parseOrExpr();
        }
        catch(IndexOutOfBoundsException e){
            return null;
        }
        catch(InvalidSyntaxException e){
            return null;
        }

        if(_index == _tokens.length){
            return selectExpr;
        }
        else{
            return null;
        }        
    }

    //--------------------------------------------------------------------
    // Parse
    //--------------------------------------------------------------------
    private SelectExpr parseOrExpr() throws InvalidSyntaxException
    {
        SelectExpr expr = this.parseAndExpr();
        if(expr == null) return null;

        while(_index != _tokens.length)
        {
            String token = _tokens[_index++];
            if( isOrToken(token) ){
                SelectExpr expr1 = expr;
                SelectExpr expr2 = this.parseAndExpr();
                if(expr2 == null) return null;
    
                expr = new OrExpr();
                ((OrExpr)expr).setLeftChild(expr1);
                ((OrExpr)expr).setRightChild(expr2);
            }
            else{
                _index--;
                break;
            }
        }

        return expr;
    }

    private SelectExpr parseAndExpr() throws InvalidSyntaxException
    {
        SelectExpr expr = this.parseTerminalExpr();
        if(expr == null) return null;

        while(_index != _tokens.length)
        {
            String token = _tokens[_index++];
            if( isAndToken(token) ){
                SelectExpr expr1 = expr;
                SelectExpr expr2 = this.parseTerminalExpr();
                if(expr2 == null) return null;
    
                expr = new AndExpr();
                ((AndExpr)expr).setLeftChild(expr1);
                ((AndExpr)expr).setRightChild(expr2);
            }
            else{
                _index--;
                break;
            }
        }

        return expr;
    }

    private SelectExpr parseTerminalExpr() throws InvalidSyntaxException
    {
        SelectExpr expr = null;

        String token = _tokens[_index++];
        if( token.equals("(") ){
            expr = this.parseOrExpr();
            return ( _tokens[_index++].equals(")") ) ? expr : null;
        }

        if( isNotToken(token) ){
            SelectExpr expr2 = this.parseTerminalExpr();
            if(expr2 != null){
                expr = new NotExpr();
                ((NotExpr)expr).setChild(expr2);
            }
            return expr;
        }

        return this.parseExpr(token);
    }

    private SelectExpr parseExpr(String firstToken) throws InvalidSyntaxException
    {
        if(_vertex){
            return this.parseVertexExpr(firstToken);
        }
        else{
            return this.parseAtomExpr(firstToken);
        }
    }

    //--------------------------------------------------------------------
    // Parse atoms
    //--------------------------------------------------------------------
    private SelectExpr parseAtomExpr(String firstToken) throws InvalidSyntaxException
    {
        SelectExpr expr = null;

        expr = this.parsePredefinedSetExpr(firstToken);
        if(expr != null) {
            if (_filter) {
                PredefinedSetExpr pexpr = (PredefinedSetExpr) expr;
                switch (pexpr.getIndex()) {
                case PredefinedSetExpr.INDEX_BONDED:
                    throw new InvalidFilterException("Invalid filter: bonded");
                case PredefinedSetExpr.INDEX_HELIX:
                    throw new InvalidFilterException("Invalid filter: helix");
                case PredefinedSetExpr.INDEX_SHEET:
                    throw new InvalidFilterException("Invalid filter: sheet");
                case PredefinedSetExpr.INDEX_TURN:
                    throw new InvalidFilterException("Invalid filter: turn");
                }
            }
            return expr;
        }

        expr = this.parseElementExpr(firstToken);
        if(expr != null) return expr;

        expr = this.parseResidueRangeExpr(firstToken);
        if(expr != null) return expr;

        expr = this.parseComparisonExpr(firstToken);
        if(expr != null) return expr;

        expr = this.parseWithinExpr(firstToken);
        if(expr != null) {
            if (_filter) {
                throw new InvalidFilterException("Invalid filter: within");
            }
            return expr;
        }

        /*
        expr = this.parseXpsssExpr(firstToken);
        if(expr != null) {
            if (_filter) {
                throw new InvalidFilterException("PDBj expression (pdbj:) cannot be used as a filter.");
            }
            return expr;
        }

        expr = this.parseSiteExpr(firstToken);
        if(expr != null) return expr;
        */

        expr = this.parsePrimitiveExpr(firstToken);
        return expr;
    }

    private SelectExpr parsePredefinedSetExpr(String token)
    {
        int index = PredefinedSetExpr.getIndex(token);
        if(index != -1){
            return new PredefinedSetExpr(index, _backboneAtomSet);
        }
        else{
            return null;
        }
    }

    private SelectExpr parseElementExpr(String token)
    {
        int atomicNo = ElementTable.getAtomicNo(token);
        if(atomicNo != 0){
            return new ElementExpr(atomicNo);
        }
        else{
            return null;
        }
    }

    private SelectExpr parseResidueRangeExpr(String token)
    {
        String chainId = "*";

        // find chain separator
        int index = token.indexOf(":");
        if(index != -1)
        {
            if(index == token.length()-1)
            {
                token = token.substring(0, index);
            }
            else
            {
                chainId = token.substring(index + 1);
                token = token.substring(0, index);
            }
        }
        else{
            char ch = token.charAt( token.length() - 1 );
            if( testChainId(ch) ){
                chainId = "" + ch;
                token = token.substring(0, token.length() - 1);
            }
            else if( ch < '0' || ch > '9' ) return null;
        }
        

        // find separator
        index = token.indexOf("-");
        if(index == 0){
            index = token.indexOf("-", 1);
        }

        try{
            if(index == -1){
                int id = Integer.parseInt(token);
                return new ResidueRangeExpr(id, id, chainId);
            }
            else if(index == token.length() - 1){
                return null;
            }
            else{
                int init = Integer.parseInt( token.substring(0, index) );
                int term = Integer.parseInt( token.substring(index+1) );
                return new ResidueRangeExpr(init, term, chainId);
            }
        }
        catch(NumberFormatException e){
            return null;
        }
    }

    private boolean testChainId(char ch)
    {
        return ( ( ch >= 'a' && ch <= 'z' ) || ( ch >= 'A' && ch <= 'Z' )
                 || ch == '*' || ch == '?' );
    }

    private SelectExpr parseComparisonExpr(String firstToken)
        throws InvalidSyntaxException
    {
        int index = AtomComparisonExpr.getIndex(firstToken);

        if(index != -1){
            String op = _tokens[_index++];
            int opType  = this.getOperatorType(op);
            float value = 0;
            try{
                value = Float.parseFloat( _tokens[_index++] );
            }
            catch(NumberFormatException e){
                throw new InvalidSyntaxException();
            }

            return new AtomComparisonExpr(index, opType, value);
        }
        else{
            return null;
        }
    }
    
    private int getOperatorType(String op) throws InvalidSyntaxException
    {
        if( op.equals("==") || op.equals("=") ){
            return ComparisonExpr.OP_EQUAL;
        }
        else if( op.equals("<>") || op.equals("!=") || op.equals("/=") ){
            return ComparisonExpr.OP_NOT_EQUAL;
        }
        else if( op.equals(">") ){
            return ComparisonExpr.OP_LARGER;
        }
        else if( op.equals(">=") ){
            return ComparisonExpr.OP_LARGER_EQUAL;
        }
        else if( op.equals("<") ){
            return ComparisonExpr.OP_SMALLER;
        }
        else if( op.equals("<=") ){
            return ComparisonExpr.OP_SMALLER_EQUAL;
        }
        else{
            throw new InvalidSyntaxException();
        }
    }

    private SelectExpr parseWithinExpr(String firstToken) throws InvalidSyntaxException
    {
        if( firstToken.toLowerCase().equals("within") ){
            SelectExpr expr = null;
            float distance = 0;
            boolean err = false;

            try{
                if( !_tokens[_index++].equals("(") ) err = true;
                distance = parseFloat( _tokens[_index++] );
                if( !_tokens[_index++].equals(",") ) err = true;
                expr = this.parseOrExpr();
                if(expr == null) err = true;
                if( !_tokens[_index++].equals(")") ) err = true;    
            }
            catch(NumberFormatException e){
                err = true;
            }

            if(!err){
                return new AtomWithinExpr(expr, distance);
            }
            else{
                throw new InvalidSyntaxException();
            }
        }
        else{
            return null;
        }
    }

    /*
    private SelectExpr parseXpsssExpr(String token)
    {
        String lowerToken = token.toLowerCase();
        if (lowerToken.startsWith("pdbj:") || lowerToken.startsWith("xps3:")) {
            return new XpsssExpr( token.substring(5) );
        }
        else{
            return null;
        }
    }

    private SelectExpr parseSiteExpr(String token)
            throws InvalidSyntaxException {
        for (int i = 0; i < _prefix.length; i++) {
            if (token.toLowerCase().startsWith(_prefix[i])) {
                String s = token.substring(_prefix[i].length());
                if (!s.equals("") && !s.startsWith(":")) {
                    continue;
                } else if (!s.equals("")) {
                    s = s.substring(1);
                }
                String db = "";
                String category = "";
                int index = s.indexOf(":");
                if (index == -1) {
                    db = s;
                } else {
                    db = s.substring(0, index);
                    if (s.indexOf(":", index + 1) != -1) {
                        throw new InvalidSyntaxException();
                    } else {
                        category = s.substring(index + 1);
                    }
                }
                return new SiteExpr(_prefix[i], db, category);
            }
        }
        return null;
    }
    */
    
    private SelectExpr parsePrimitiveExpr(String token)
    {
        return PrimitiveExprParser.parse(token);
    }

    //--------------------------------------------------------------------
    // Parse vertex
    //--------------------------------------------------------------------
    private SelectExpr parseVertexExpr(String firstToken) throws InvalidSyntaxException
    {
        SelectExpr expr = null;

        expr = this.parseVertexPredefinedSetExpr(firstToken);
        if(expr != null) return expr;

        expr = this.parseVertexDistanceExpr(firstToken);
        if(expr != null) return expr;

        expr = this.parseVertexComparisonExpr(firstToken);
        return expr;
    }

    private SelectExpr parseVertexPredefinedSetExpr(String firstToken) throws InvalidSyntaxException
    {
        int index = VertexPredefinedSetExpr.getIndex(firstToken);
        if(index != -1){
            return new VertexPredefinedSetExpr(index);
        }
        else{
            return null;
        }
    }

    private SelectExpr parseVertexDistanceExpr(String firstToken) throws InvalidSyntaxException
    {
        boolean within = false;

        if( firstToken.toLowerCase().equals("within") ){
            within = true;
        }
        else if( !firstToken.toLowerCase().equals("box") ){
            return null;
        }

        SelectExpr expr = null;
        float[] coord = null;
        boolean transform = false;
        float distance = 0;
        boolean err = false;

        try{
            if( !_tokens[_index++].equals("(") ) err = true;
            distance = parseFloat( _tokens[_index++] );
            if( !_tokens[_index++].equals(",") ) err = true;
            
            char bracket = _tokens[_index].charAt(0);
            if(bracket == '['){
                transform = true;
                coord = this.getCoord("[", "]");
                if(coord == null) err = true;
            }
            else if(bracket == '{'){
                coord = this.getCoord("{", "}");
                if(coord == null) err = true;
            }
            else{
                _vertex = false;
                expr = this.parseOrExpr();
                _vertex = true;
                if(expr == null) err = true;
            }

            if( !_tokens[_index++].equals(")") ) err = true;    
        }
        catch(NumberFormatException e){
            err = true;
        }

        if(!err){
            if(expr != null){
                if(within){
                    return new VertexWithinExpr(expr, distance);
                }
                else{
                    return new VertexBoxExpr(expr, distance);
                }
            }
            else{
                if(within){
                    return new VertexWithinExpr(coord[0], coord[1], coord[2], transform, distance);
                }
                else{
                    return new VertexBoxExpr(coord[0], coord[1], coord[2], transform, distance);
                }
            }
        }
        else{
            throw new InvalidSyntaxException();
        }
    }
    
    private float[] getCoord(String initBracket, String termBracket)
    {
        StringBuffer buffer = new StringBuffer();
        String token = _tokens[_index++];
        while( !token.endsWith(termBracket) ){
            buffer.append(token);
            token = _tokens[_index++];
        }
        buffer.append(token);

        TupleFloatExprParser parser = new TupleFloatExprParser();
        return parser.parse(buffer.toString(), initBracket, termBracket, 3);
    }

    private SelectExpr parseVertexComparisonExpr(String firstToken)
        throws InvalidSyntaxException
    {
        int index = VertexComparisonExpr.getIndex(firstToken);

        String op = _tokens[_index++];
        int opType  = this.getOperatorType(op);

        float value = 0;
        try{
            value = Float.parseFloat( _tokens[_index++] );
        }
        catch(NumberFormatException e){
            throw new InvalidSyntaxException();
        }

        if(index != -1){
            return new VertexComparisonExpr(index, opType, value);
        }
        else{
            return new VertexComparisonExpr(firstToken, opType, value);
        }
    }

    //--------------------------------------------------------------------
    // Utility
    //--------------------------------------------------------------------
    private static boolean isOrToken(String token)
    {
        return ( token.toLowerCase().equals("or") || token.equals(",") ||
                 token.equals("|") || token.equals("||") );
    }

    private static boolean isAndToken(String token)
    {
        return ( token.toLowerCase().equals("and") || token.equals("&") || 
                 token.equals("&&") );
    }

    private static boolean isNotToken(String token)
    {
        return ( token.toLowerCase().equals("not") || token.equals("!") || 
                 token.equals("~") );
    }

    private static float parseFloat(String token)
    {
        int index = token.indexOf(".");
        float val = Float.parseFloat(token);
        
        return (index == -1) ? val/250 : val;
    }

    //--------------------------------------------------------------------
    // Exception
    //--------------------------------------------------------------------
    public static class InvalidFilterException extends RuntimeException {

        private static final long serialVersionUID = 1L;

        public InvalidFilterException(String message) {
            super(message);
        }
    }
}
