/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.controller;

import java.util.ArrayList;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;

/**
 * Class that holds all files loaded.
 */
public class FileHolder
{
    private ArrayList _moleculeFiles;
    private ArrayList _polygonFiles;
    private ArrayList _fileStatusList;
    private ArrayList _fileStatusListeners;

    private int _maxId;

    private ArrayList _moleculeList;
    private ArrayList _polygonList;

    /**
     * Returns a collection of file status.
     */
    public ArrayList getFileStatusList()
    {
        return _fileStatusList;
    }

    /**
     * Returns a collection of molecule files.
     */
    public ArrayList getMoleculeFiles()
    {
        return _moleculeFiles;
    }

    /**
     * Returns a collection of polygon files.
     */
    public ArrayList getPolygonFiles()
    {
        return _polygonFiles;
    }

    /**
     * Constructor.
     */
    public FileHolder()
    {
        _moleculeFiles  = new ArrayList();
        _polygonFiles   = new ArrayList();
        _fileStatusList = new ArrayList();
        _fileStatusListeners = new ArrayList();

        _moleculeList = new ArrayList();
        _polygonList  = new ArrayList();
    }

    /**
     * Appends file status listener.
     */
    public void addFileStatusListener(FileStatusListener fileStatusListener)
    {
        _fileStatusListeners.add(fileStatusListener);
    }

    /**
     * Removes file status listener.
     */
    public void removeFileStatusListener(FileStatusListener fileStatusListener)
    {
        _fileStatusListeners.remove(fileStatusListener);
    }

    /**
     * Appends the molecule file.
     */
    public void add(MoleculeFile moleculeFile)
    {
        moleculeFile.setFileId(++_maxId);
        _moleculeFiles.add(moleculeFile);

        FileStatus status = new FileStatus(moleculeFile);
        _fileStatusList.add(status);

        this.fireFileAdded();
    }

    /**
     * Appends the polygon file.
     */
    public void add(PolygonFile polygonFile)
    {
        polygonFile.setFileId(++_maxId);
        _polygonFiles.add(polygonFile);

        FileStatus status = new FileStatus(polygonFile);
        _fileStatusList.add(status);

        this.fireFileAdded();
    }

    /**
     * Removes the molecule file.
     */
    public void remove(MoleculeFile moleculeFile)
    {
        int index = _moleculeFiles.indexOf(moleculeFile);
        if(index != -1){
            _moleculeFiles.remove(index);
            this.adjustMaxId();
            this.removeStatus( moleculeFile.getFileId() );
        }

        this.fireFileRemoved();
    }

    /**
     * Removes the polygon file.
     */
    public void remove(PolygonFile polygonFile)
    {
        int index = _polygonFiles.indexOf(polygonFile);
        if(index != -1){
            _polygonFiles.remove(index);
            this.adjustMaxId();
            this.removeStatus( polygonFile.getFileId() );
        }

        this.fireFileRemoved();
    }

    /**
     * Returns the molecule file specified by ID.
     */
    public MoleculeFile findMoleculeFile(int fileId)
    {
        for(int i=0; i<_moleculeFiles.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)_moleculeFiles.get(i);
            if(file.getFileId() == fileId) return file;
        }
        return null;
    }

    /**
     * Returns the polygon file specified by ID.
     */
    public PolygonFile findPolygonFile(int fileId)
    {
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            if(file.getFileId() == fileId) return file;
        }
        return null;
    }

    /**
     * Returns the file specified by ID.
     */
    public FileStatus findFile(int fileId)
    {
        for(int i=0; i<_fileStatusList.size(); i++)
        {
            FileStatus file = (FileStatus)_fileStatusList.get(i);
            if(file.getId() == fileId) return file;
        }
        return null;
    }

    /**
     * Sets specified files.
     * Returns false for invaid ids.
     */
    public boolean find(
        ArrayList moleculeFiles, ArrayList polygonFiles, int[] fileIds)
    {
        moleculeFiles.clear();
        polygonFiles.clear();
        if(fileIds.length == 0) return false;

        for(int i=0; i<fileIds.length; i++)
        {
            int id = fileIds[i];
            
            MoleculeFile moleculeFile = this.findMoleculeFile(id);
            if(moleculeFile != null){
                moleculeFiles.add(moleculeFile);
                continue;
            }

            PolygonFile polygonFile = this.findPolygonFile(id);
            if(polygonFile != null){
                polygonFiles.add(polygonFile);
                continue;
            }

            moleculeFiles.clear();
            polygonFiles.clear();
            return false;
        }
        return true;
    }

    /**
     * Change selection of files.
     */
    public boolean selectFile(boolean enable, int[] fileIds)
    {
        if( this.find(_moleculeList, _polygonList, fileIds) )
        {
            for(int i=0; i<_moleculeList.size(); i++)
            {
                ( (MoleculeFile)_moleculeList.get(i) ).setSelected(enable);
            }
            for(int i=0; i<_polygonList.size(); i++)
            {
                ( (PolygonFile)_polygonList.get(i) ).setSelected(enable);
            }
            this.fireStatusChanged();
            this.clearList();
            return true;
        }
        else return false;
    }

    /**
     * Selects files.
     */
    public boolean selectFile(int[] fileIds)
    {
        if( this.find(_moleculeList, _polygonList, fileIds) )
        {
            for(int i=0; i<_moleculeFiles.size(); i++)
            {
                ( (MoleculeFile)_moleculeFiles.get(i) ).setSelected(false);
            }
            for(int i=0; i<_polygonFiles.size(); i++)
            {
                ( (PolygonFile)_polygonFiles.get(i) ).setSelected(false);
            }
            for(int i=0; i<_moleculeList.size(); i++)
            {
                ( (MoleculeFile)_moleculeList.get(i) ).setSelected(true);
            }
            for(int i=0; i<_polygonList.size(); i++)
            {
                ( (PolygonFile)_polygonList.get(i) ).setSelected(true);
            }
            this.fireStatusChanged();
            this.clearList();
            return true;
        }
        else return false;
    }

    /**
     * Selects all files.
     */
    public void selectAllFile(boolean enable)
    {
        for(int i=0; i<_moleculeFiles.size(); i++)
        {
            ( (MoleculeFile)_moleculeFiles.get(i) ).setSelected(enable);
        }
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            ( (PolygonFile)_polygonFiles.get(i) ).setSelected(enable);
        }
        this.fireStatusChanged();    
    }

    /**
     * Change display of files.
     */
    public boolean displayFile(boolean enable, int[] fileIds)
    {
        if( this.find(_moleculeList, _polygonList, fileIds) )
        {
            for(int i=0; i<_moleculeList.size(); i++)
            {
                ( (MoleculeFile)_moleculeList.get(i) ).setDisplayed(enable);
            }
            for(int i=0; i<_polygonList.size(); i++)
            {
                ( (PolygonFile)_polygonList.get(i) ).setDisplayed(enable);
            }
            this.fireStatusChanged();
            this.clearList();
            return true;
        }
        else return false;
    }

    /**
     * Displays files.
     */
    public boolean displayFile(int[] fileIds)
    {
        if( this.find(_moleculeList, _polygonList, fileIds) )
        {
            for(int i=0; i<_moleculeFiles.size(); i++)
            {
                ( (MoleculeFile)_moleculeFiles.get(i) ).setDisplayed(false);
            }
            for(int i=0; i<_polygonFiles.size(); i++)
            {
                ( (PolygonFile)_polygonFiles.get(i) ).setDisplayed(false);
            }
            for(int i=0; i<_moleculeList.size(); i++)
            {
                ( (MoleculeFile)_moleculeList.get(i) ).setDisplayed(true);
            }
            for(int i=0; i<_polygonList.size(); i++)
            {
                ( (PolygonFile)_polygonList.get(i) ).setDisplayed(true);
            }
            this.fireStatusChanged();
            this.clearList();
            return true;
        }
        else return false;
    }

    /**
     * Displays all files.
     */
    public void displayAllFile(boolean enable)
    {
        for(int i=0; i<_moleculeFiles.size(); i++)
        {
            ( (MoleculeFile)_moleculeFiles.get(i) ).setDisplayed(enable);
        }
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            ( (PolygonFile)_polygonFiles.get(i) ).setDisplayed(enable);
        }
        this.fireStatusChanged();    
    }

    /**
     * Notifies select-atom.
     */
    public void notifySelectAtom() {
        this.fireStatusChanged();    
    }

    /**
     * Notifies select-vertex.
     */
    public void notifySelectVertex() {
        this.fireStatusChanged();    
    }

    //--------------------------------------------------------------------
    // Clear list
    //--------------------------------------------------------------------
    private void clearList()
    {
        _moleculeList.clear();
        _polygonList.clear();
    }

    //--------------------------------------------------------------------
    // Adjust max ID
    //--------------------------------------------------------------------
    private void adjustMaxId()
    {
        _maxId = 0;
        for(int i=0; i<_moleculeFiles.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)_moleculeFiles.get(i);
            if(file.getFileId() > _maxId) _maxId = file.getFileId();
        }
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            if(file.getFileId() > _maxId) _maxId = file.getFileId();
        }        
    }

    //--------------------------------------------------------------------
    // Handle status list
    //--------------------------------------------------------------------
    private void removeStatus(int fileId)
    {
        for(int i=0; i<_fileStatusList.size(); i++)
        {
            FileStatus status = (FileStatus)_fileStatusList.get(i);
            if(status.getId() == fileId){
                _fileStatusList.remove(status);
                return;
            }
        }
    }

    //--------------------------------------------------------------------
    // Call listeners
    //--------------------------------------------------------------------
    private void fireFileAdded()
    {
        for(int i=0; i<_fileStatusListeners.size(); i++)
        {
            FileStatusListener listener
                = (FileStatusListener)_fileStatusListeners.get(i);
            listener.fileAdded(_fileStatusList);
        }
    }

    private void fireFileRemoved()
    {
        for(int i=0; i<_fileStatusListeners.size(); i++)
        {
            FileStatusListener listener
                = (FileStatusListener)_fileStatusListeners.get(i);
            listener.fileRemoved(_fileStatusList);
        }
    }

    private void fireStatusChanged()
    {
        for(int i=0; i<_fileStatusListeners.size(); i++)
        {
            FileStatusListener listener
                = (FileStatusListener)_fileStatusListeners.get(i);
            listener.statusChanged(_fileStatusList);
        }
    }
}
