/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.controller;

import java.io.File;
import java.net.URL;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.moleculebuilders.*;
import org.pdbj.viewer.plus.*;
import org.pdbj.viewer.polygons.*;
import org.pdbj.viewer.site.*;

/**
 * File loader.
 */
public class FileLoader
{
    /**
     * Load a PDBML file on local file system.
     */
    public static MoleculeFile loadPDBmlFile(File file, MessageListener listener, PDBmlUrl pdbmlUrl)
        throws RecoverableException
    {
        PDBmlReader reader = new PDBmlReader(listener, pdbmlUrl);
        return reader.read(file);
    }

    /**
     * Load a PDBML file through network.
     */
    public static MoleculeFile loadPDBmlFile(URL url, MessageListener listener, PDBmlUrl pdbmlUrl)
        throws RecoverableException
    {
        PDBmlReader reader = new PDBmlReader(listener, pdbmlUrl);
        return reader.read(url);
    }

    /**
     * Load a PDBML file by FTP.
     */
    public static MoleculeFile loadPDBmlFileFTP(String pdbCode, MessageListener listener, PDBmlUrl pdbmlUrl)
        throws RecoverableException
    {
        PDBmlReader reader = new PDBmlReader(listener, pdbmlUrl);
        return reader.readFTP(pdbCode);
    }

    /**
     * Load a PDB file on local file system.
     */
    public static MoleculeFile loadPDBFile(File file) throws RecoverableException
    {
        PDBReader reader = new PDBReader(null, false);
        return reader.read(file, null);
    }

    /**
     * Load a PDB file on local file system.
     */
    public static MoleculeFile loadPDBFile(File file,
            MoleculeFileBuilder.AtomFilter filter) throws RecoverableException
    {
        PDBReader reader = new PDBReader(null, false);
        return reader.read(file, filter);
    }

    /**
     * Load a PDB file through network.
     */
    public static MoleculeFile loadPDBFile(URL url, MessageListener listener)
            throws RecoverableException
    {
        PDBReader reader = new PDBReader(listener, false);
        return reader.read(url);
    }
    
    /**
     * Load an animation file on local file system.
     */
    public static MoleculeFile loadAnimationFile(File file, MessageListener listener)
            throws RecoverableException
    {
        PDBReader reader = new PDBReader(listener, true);
        MoleculeFile moleculeFile = reader.read(file, null);
        moleculeFile.setAnimation(true);
        return moleculeFile;
    }

    /**
     * Load an animation file through network.
     */
    public static MoleculeFile loadAnimationFile(URL url, MessageListener listener)
            throws RecoverableException
    {
        PDBReader reader = new PDBReader(listener, true);
        MoleculeFile moleculeFile = reader.read(url);
        moleculeFile.setAnimation(true);
        return moleculeFile;
    }
    
    /**
     * Load a polygon file on local file system.
     */
    public static PolygonFile loadPolygonFile(File file) throws RecoverableException
    {
        PolygonFileReader reader = new PolygonFileReader(null);
        return reader.read(file);
    }

    /**
     * Load a polygon file through network.
     */
    public static PolygonFile loadPolygonFile(URL url, MessageListener listener)
        throws RecoverableException
    {
        PolygonFileReader reader = new PolygonFileReader(listener);
        return reader.read(url);
    }
    
    /**
     * Load a polygon CGO file on local file system.
     */
    public static PolygonFile loadPolygonCgoFile(File file) throws RecoverableException
    {
        CgoReader reader = new CgoReader(null);
        return reader.read(file);
    }

    /**
     * Load a polygon CGO file through network.
     */
    public static PolygonFile loadPolygonCgoFile(URL url, MessageListener listener)
        throws RecoverableException
    {
        CgoReader reader = new CgoReader(listener);
        return reader.read(url);
    }

    /**
     * Load a function file by FTP.
     */
    public static Function loadFunctionFileFTP(String pdbCode, MessageListener listener, PDBmlUrl pdbmlUrl)
        throws RecoverableException
    {
        try {
            FunctionReader reader = new FunctionReader(listener, pdbmlUrl);
            return reader.readFTP(pdbCode);
        } catch (java.security.AccessControlException e) {
            throw new RecoverableException("cannot access to " + pdbmlUrl.getHostFunction());
        }
    }

    /**
     * Load a site list file.
     */
    public static SiteList loadSiteList(URL url, MessageListener listener)
            throws RecoverableException {
        SiteListReader reader = new SiteListReader(listener);
        return reader.read(url);
    }
}
