/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.controller;

import java.util.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;

/**
 * Transform handler.
 */
class TransformHandler
{
    private FileHolder _fileHolder;
    private ViewParameters _parameters;

    private Vector3f _vector1 = new Vector3f();
    private Matrix4f _matrix1 = new Matrix4f();
    private Matrix4f _matrix2 = new Matrix4f();
    
    /**
     * Constructor.
     */
    TransformHandler(FileHolder fileHolder, ViewParameters parameters)
    {
        _fileHolder = fileHolder;
        _parameters = parameters;
    }

    /**
     * Rotates by the specified rotation matrix.
     */
    void rotate(Matrix4f rotation)
    {
        _vector1.set( _parameters.getRotationCenter() );
        _vector1.multiply(-1);

        _matrix1.translation(_vector1);
        _matrix2.translation( _parameters.getRotationCenter() );

        _matrix1.multiply(rotation, _matrix1);
        _matrix1.multiply(_matrix2, _matrix1);

        this.transform(_matrix1, false);
    }

    /**
     * Translates by the specified vector.
     */
    void translate(Vector3f translation)
    {
        _matrix1.translation(translation);
        this.transform(_matrix1, false);
    }

    /**
     * Transforms images.
     */
    void transform(Matrix4f transform, boolean all)
    {
        ArrayList moleculeList = _fileHolder.getMoleculeFiles();
        for(int i=0; i<moleculeList.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)moleculeList.get(i);
            if( all || file.isSelected() )
            {
                int num = file.isTrajectory() ? 1 : file.numMolecules();
                for(int m=0; m<num; m++){
                    Molecule molecule = file.getMolecule(m);
                    Matrix4f matrix = molecule.getImage().getTransform();
                    matrix.multiply(transform, matrix);
                }
            }
        }

        ArrayList polygonList = _fileHolder.getPolygonFiles();
        for(int i=0; i<polygonList.size(); i++)
        {
            PolygonFile file = (PolygonFile)polygonList.get(i);
            if( all || file.isSelected() )
            {
                Matrix4f matrix = file.getImage().getTransform();
                matrix.multiply(transform, matrix);
            }
        }    

        // transform current center
        Vector4f v = new Vector4f(_parameters.getCurrentCenter(), 1);
        v.transform(transform);
        _parameters.getCurrentCenter().set(v.x, v.y, v.z);
    }

    /**
     * Resets transforms.
     */
    void resetTransform(boolean all)
    {
        Matrix4f matrix = new Matrix4f();
        
        ArrayList moleculeList = _fileHolder.getMoleculeFiles();
        for(int i=0; i<moleculeList.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)moleculeList.get(i);
            if( all || file.isSelected() ){
                matrix.translation( file.offset() );
                for(int m=0; m<file.numMolecules(); m++){
                    Molecule molecule = file.getMolecule(m);
                    molecule.getImage().getTransform().set(matrix);
                }
            }
        }

        ArrayList polygonList = _fileHolder.getPolygonFiles();
        for(int i=0; i<polygonList.size(); i++)
        {
            PolygonFile file = (PolygonFile)polygonList.get(i);
            if( all || file.isSelected() ){
                matrix.translation( file.offset() );
                file.getImage().getTransform().set(matrix);
            }
        }    
    }

    /**
     * Resets the transform of a molecule file.
     */
    void resetTransform(MoleculeFile file)
    {
        Matrix4f matrix = new Matrix4f();
        matrix.translation( file.offset() );
        for(int m=0; m<file.numMolecules(); m++){
            Molecule molecule = file.getMolecule(m);
            molecule.getImage().getTransform().set(matrix);
        }
    }

    /**
     * Resets the transform of a polygon file.
     */
    void resetTransform(PolygonFile file)
    {
        Matrix4f matrix = new Matrix4f();
        matrix.translation( file.offset() );
        file.getImage().getTransform().set(matrix);
    }

    /**
     * Copies the transform.
     */
    boolean copyTransform(int srcFileId, int dstFileId)
    {
        FileStatus srcFile = null;
        FileStatus dstFile = null;
        ArrayList list = _fileHolder.getFileStatusList();
        for(int i=0; i<list.size(); i++){
            FileStatus status = (FileStatus)list.get(i);
            if(status.getId() == srcFileId){
                srcFile = status;
            }
            else if(status.getId() == dstFileId){
                dstFile = status;
            }
        }
        if(srcFile == null || dstFile == null) return false;
        
        Matrix4f transform = null;
        if(srcFile.getType() != FileStatus.POLYGON){
            MoleculeFile file = srcFile.getMoleculeFile();
            transform = file.getMolecule(0).getImage().getTransform();
        }
        else{
            PolygonFile file = srcFile.getPolygonFile();
            transform = file.getImage().getTransform();
        }
        if(dstFile.getType() != FileStatus.POLYGON){
            MoleculeFile file = dstFile.getMoleculeFile();
            for(int m=0; m<file.numMolecules(); m++){
                file.getMolecule(m).getImage().getTransform().set(transform);
            }
        }
        else{
            PolygonFile file = dstFile.getPolygonFile();
            file.getImage().getTransform().set(transform);
        }
        
        return true;
    }


    /**
     * Sets up the default sphere.
     */
    void setupDefaultSphere()
    {
        Sphere sphere = this.createSphere(true, false);
        _parameters.getDefaultCenter().set( sphere.center() );
        _parameters.setDefaultRadius( sphere.getRadius() );
    }

    /**
     * Sets up the current sphere.
     */
    void setupCurrentSphere()
    {
        Sphere sphere = this.createSphere(false, true);
        _parameters.getCurrentCenter().set( sphere.center() );
        _parameters.setCurrentRadius( sphere.getRadius() );
    }

    /**
     * Sets up sphere.
     */
    Sphere createSphere(boolean all, boolean transform)
    {
        SphereUtil util = new SphereUtil();

        ArrayList moleculeList = _fileHolder.getMoleculeFiles();
        for(int i=0; i<moleculeList.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)moleculeList.get(i);
            if( all || file.isSelected() )
            {
                Vector3f center = new Vector3f( file.center() );
                if(transform){
                    center.transform( file.getMolecule(0).getImage().getTransform() );
                }
                else{
                    center.add( file.offset() );
                }
                util.addSphere(center, file.getRadius());
            }
        }
        
        ArrayList polygonList = _fileHolder.getPolygonFiles();
        for(int i=0; i<polygonList.size(); i++)
        {
            PolygonFile file = (PolygonFile)polygonList.get(i);
            if (file.getRadius() == -1) {
                continue;
            }

            if( all || file.isSelected() )
            {
                Vector3f center = new Vector3f( file.center() );
                if(transform){
                    center.transform( file.getImage().getTransform() );
                }
                else{
                    center.add( file.offset() );
                }
                util.addSphere(center, file.getRadius());
            }
        }

        return util.getSphere();
    }
}
