/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.controller;

import java.awt.Color;
import java.util.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.renderer.*;

/**
 * Represents status of view.
 */
public class ViewStatus
{
    private ViewSettings   _viewSettings;
    private ViewParameters _viewParameters;

    private List _viewStatusListeners;

    /**
     * Constructor.
     */
    public ViewStatus(ViewSettings viewSettings, ViewParameters viewParameters)
    {
        _viewSettings   = viewSettings;
        _viewParameters = viewParameters;

        _viewStatusListeners = new ArrayList();
    }

    /**
     * Adds view status listener.
     */
    public void addViewStatusListener(ViewStatusListener viewStatusListener) {
        _viewStatusListeners.add(viewStatusListener);
    }

    //--------------------------------------------------------------------
    // renderer
    //--------------------------------------------------------------------
    /**
     * Sets the passed vector a copy of the camera position.
     */
    public void getCameraPosition(Vector3f cameraPosition)
    {
        cameraPosition.set( _viewSettings.getCameraPosition() );
    }

    /**
     * Returns the field of view in radians.
     */
    public float getFieldAngle()
    {
        return _viewSettings.getFieldAngle();
    }

    /**
     * Returns the near clipping plane distance.
     */
    public float getNearClipping()
    {
        return _viewSettings.getNearClipping();
    }

    /**
     * Returns the far clipping plane distance.
     */
    public float getFarClipping()
    {
        return _viewSettings.getFarClipping();
    }

    /**
     * Returns the background color.
     */
    public Color getBGColor()
    {
        Color bgColor = _viewSettings.getBGColor() ;
        return new Color(
            bgColor.getRed(), 
            bgColor.getGreen(), 
            bgColor.getBlue(), 
            bgColor.getAlpha() 
        );
    }

    /**
     * Returns whether the stereo mode.
     */
    public boolean isStereo()
    {
        return _viewSettings.isStereo();
    }

    /**
     * Returns whether the anaglyph is off.
     */
    public boolean isAnaglyphOff() {
        return _viewSettings.getAnaglyphMode() == ViewSettings.ANAGLYPH_OFF;
    }

    /**
     * Returns whether the red-blue anaglyph mode.
     */
    public boolean isAnaglyphRedBlue() {
        return _viewSettings.getAnaglyphMode() == ViewSettings.ANAGLYPH_REDBLUE;
    }

    /**
     * Returns whether the red-cyan anaglyph mode.
     */
    public boolean isAnaglyphRedCyan() {
        return _viewSettings.getAnaglyphMode() == ViewSettings.ANAGLYPH_REDCYAN;
    }

    /**
     * Returns whether the red-green anaglyph mode.
     */
    public boolean isAnaglyphRedGreen() {
        return _viewSettings.getAnaglyphMode() == ViewSettings.ANAGLYPH_REDGREEN;
    }

    /**
     * Sets the passed vector a copy of the stereo center.
     */
    public void getStereoCenter(Vector3f center)
    {
        center.set( _viewSettings.getStereoCenter() );
    }

    /**
     * Returns the stereo angle in degrees.
     */
    public float getStereoAngle()
    {
        float angle = _viewSettings.getStereoAngle();
        return (float)(angle*180/Math.PI);
    }

    /**
     * Returns the stereo angle in radians.
     */
    public float getStereoAngleRadians()
    {
        return _viewSettings.getStereoAngle();
    }

    /**
     * Returns whether specular is enabled.
     */
    public boolean getSpecular()
    {
        return _viewSettings.getSpecular();
    }

    /**
     * Returns the specular color.
     */
    public float[] getSpecularColor()
    {
        float[] color1 = _viewSettings.getSpecularColor();
        float[] color2 = new float[4];
        for(int i=0; i<4; i++) color2[i] = color1[i];
        
        return color2;
    }
    
    /**
     * Returns the shininess.
     */
    public float getShininess()
    {
        return _viewSettings.getShininess();
    }

    /**
    * Returns whether parallel projection.
    */
    public boolean isParallelProjection()
    {
        return _viewSettings.isParallelProjection();
    }

    /**
    * Returns the view size in parallel projection.
    */
    public float getViewSize()
    {
        return _viewSettings.getViewSize();
    }

    /**
     * Returns the strength of ambient light.
     */
    public float getAmbient()
    {
        return _viewSettings.getAmbient();
    }

    /**
     * Returns the strength of diffuse light 1.
     */
    public float getDiffuse1()
    {
        return _viewSettings.getDiffuse1();
    }

    /**
     * Returns the strength of diffuse light 2.
     */
    public float getDiffuse2()
    {
        return _viewSettings.getDiffuse2();
    }

    /**
     * Returns the direction of diffuse light 1.
     */
    public float[] getDiffuseDir1()
    {
        float[] curr = _viewSettings.getDiffuseDir1();
        float[] dir = new float[3];
        dir[0] = curr[0];
        dir[1] = curr[1];
        dir[2] = curr[2];
        return dir;
    }

    /**
     * Returns the direction of diffuse light 2.
     */
    public float[] getDiffuseDir2()
    {
        float[] curr = _viewSettings.getDiffuseDir2();
        float[] dir = new float[3];
        dir[0] = curr[0];
        dir[1] = curr[1];
        dir[2] = curr[2];
        return dir;
    }

    //--------------------------------------------------------------------
    // controller
    //--------------------------------------------------------------------
    /**
     * Sets the passed vector a copy of the default camera position.
     */
    public void getDefaultCameraPosition(Vector3f defaultCameraPosition)
    {
        defaultCameraPosition.set( _viewParameters.getDefaultCameraPosition() );
    }

    /**
     * Sets the passed vector a copy of the default center.
     */
    public void getDefaultCenter(Vector3f defaultCenter)
    {
        defaultCenter.set( _viewParameters.getDefaultCenter() );
    }

    /**
     * Returns the default radius.
     */
    public float getDefaultRadius()
    {
        return _viewParameters.getDefaultRadius();
    }

    /**
     * Sets the passed vector a copy of the current center.
     */
    public void getCurrentCenter(Vector3f currentCenter)
    {
        currentCenter.set( _viewParameters.getCurrentCenter() );
    }

    /**
     * Returns the current radius.
     */
    public float getCurrentRadius()
    {
        return _viewParameters.getCurrentRadius();
    }

    /**
     * Sets the passed vector a copy of the rotation center.
     */
    public void getRotationCenter(Vector3f rotationCenter)
    {
        rotationCenter.set( _viewParameters.getRotationCenter() );
    }

    /**
     * Returns whether the slab mode.
     */
    public boolean isSlab()
    {
        return _viewParameters.isSlab();
    }

    /**
     * Returns the z coordinate of the slab plane.
     */
    public float getSlabValue()
    {
        return _viewParameters.getSlabValue();
    }

    /**
     * Returns whether to load to center.
     */
    public boolean isLoadCenter()
    {
        return _viewParameters.isLoadCenter();
    }

    /**
     * Returns whether to adjust view.
     */
    public boolean isAdjustView()
    {
        return _viewParameters.isAdjustView();
    }

    void setSlab(boolean slab) {
        _viewParameters.setSlab(slab);
        notifySlabChanged();
    }

    void setSlabValue(float slabValue) {
        _viewParameters.setSlabValue(slabValue);
        notifySlabChanged();
    }

    private void notifySlabChanged() {
        for (Iterator it = _viewStatusListeners.iterator(); it.hasNext();) {
            ((ViewStatusListener) it.next()).slabChanged(this);
        }
    }
}
