/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.controller;

import java.awt.*;
import java.awt.image.*;
import java.util.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.images.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;
import org.pdbj.viewer.manipulators.*;
import org.pdbj.viewer.renderer.*;

/**
 * Controls the viewer.
 */
public class ViewerController implements ViewerConstants, PolygonSelection
{
    private MoleculeManipulator _moleculeManipulator;
    private PolygonManipulator  _polygonManipulator;

    private ColorManager  _colorManager;
    private FileHolder    _fileHolder;
    private ImageRenderer _renderer;

    private TransformHandler _transformHandler;

    private ViewParameters _parameters;
    private ViewStatus     _viewStatus;

    private float _currViewSize;

    private PDBmlUrl _pdbmlUrl;
    private EfsiteUrl _efsiteUrl;
    private SiteUrlMap _siteUrlMap;

    private Vector3f _vector1 = new Vector3f();

    /**
    * Returns the color manager.
    */
    public ColorManager getColorManager()
    {
        return _colorManager;
    }

    /**
    * Returns the view status.
    */
    public ViewStatus getViewStatus()
    {
        return _viewStatus;
    }

    //--------------------------------------------------------------------
    // Constructor & Finalizer
    //--------------------------------------------------------------------
    /**
    * Constructor.
    */
    public ViewerController(RendererListener listener, boolean newt) throws Throwable
    {
        _colorManager = new ColorManager();
        _fileHolder   = new FileHolder();

        _moleculeManipulator = new MoleculeManipulator();
        _moleculeManipulator.setColorManager(_colorManager);
        
        _polygonManipulator = new PolygonManipulator();

        // if renderer is not initialized, ViewerController is not instantiated
        try {
            _renderer = new ImageRendererImpl(newt);
            _renderer.initialize(_colorManager, listener);
            
        } catch (Throwable t) {
            throw t;
        }

        this.initializeStatus();

        _parameters = new ViewParameters();
        _viewStatus = new ViewStatus(_renderer.getViewSettings(), _parameters);
        _transformHandler = new TransformHandler(_fileHolder, _parameters);

        _pdbmlUrl = new PDBmlUrl();
        _efsiteUrl = new EfsiteUrl();
        _siteUrlMap = new SiteUrlMap();
    }

    /**
    * Dispose.
    */
    public void dispose()
    {
        if (_renderer != null)
        {
            // clear screen
            this.closeAllFile();
            this.setBackground(ViewSettings.DEFAULT_BACKGROUND);

            _renderer.dispose();
        }
    }

    //--------------------------------------------------------------------
    // Renderer
    //--------------------------------------------------------------------
    /**
    * Returns the left canvas.
    */
    public Canvas getCanvasL()
    {
        return _renderer.getCanvasL();
    }

    /**
    * Returns the right canvas.
    */
    public Canvas getCanvasR()
    {
        return _renderer.getCanvasR();
    }

    /**
    * Tests whether the graphics is ready.
    */
    public boolean graphicsReady()
    {
        return _renderer.graphicsReady();
    }

    /**
    * Tests whether the renderer is running.
    */
    public boolean rendererRunning()
    {
        return _renderer.running();
    }

    /**
    * Sets the draw level.
    */
    public void setDrawlevel(int drawlevel)
    {
        if(drawlevel >= MIN_DRAWLEVEL){
            _renderer.getViewSettings().setDrawlevel(drawlevel);
            _renderer.setDrawlevelChanged();
            this.refresh();
        }
    }

    /**
    * Returns the draw level.
    */
    public int getDrawlevel()
    {
        return _renderer.getViewSettings().getDrawlevel();
    }

    /**
    * Sets the background color.
    */
    public void setBackground(Color color)
    {
        _renderer.getViewSettings().setBGColor(color);
        _renderer.setBackgroundChanged();
        this.refresh();
    }

    /**
    * Sets whether the specular is enabled.
    */
    public void setSpecular(boolean enable)
    {
        _renderer.getViewSettings().setSpecular(enable);
        this.refresh();
    }

    /**
    * Sets the specular color.
    */
    public void setSpecularColor(float r, float g, float b, float a)
    {
        _renderer.getViewSettings().setSpecularColor(r, g, b, a);
        this.refresh();
    }

    /**
    * Sets the shininess (from 1 to 128).
    */
    public void setShininess(float shininess)
    {
        if(shininess >= 1 && shininess <= 128){
            _renderer.getViewSettings().setShininess(shininess);
            this.refresh();
        }
    }

    /**
    * Resets the shininess.
    */
    public void resetShininess()
    {
        _renderer.getViewSettings().setShininess(DEFAULT_SHININESS);
        this.refresh();
    }

    /**
     * Initialize stereo.
     */
    public void initializeStereo()
    {
        _renderer.initializeStereo();
    }

    /**
    * Enables the stereo.
    */
    public void stereo()
    {
        _renderer.setStereo(true);
        _renderer.setStereoCenter(_parameters.getRotationCenter());
        _renderer.setAnaglyphMode(ViewSettings.ANAGLYPH_OFF);
        this.refresh();
    }

    /**
    * Enables the stereo with the specified separation angle.
    */
    public void stereo(float angle)
    {
        this.setStereoAngle(angle);
        this.stereo();
    }

    /**
    * Sets the stereo angle.
    */
    public void setStereoAngle(float angle)
    {
        _renderer.setStereoAngle((float) (angle * Math.PI / 180));
    }

    /**
    * Disables the stereo.
    */
    public void stereoOff()
    {
        _renderer.setStereo(false);
        this.refresh();
    }

    /**
    * Enables the red-blue anaglyph.
    */
    public void anaglyphRedBlue() {
        _renderer.setStereo(false);
        _renderer.setStereoCenter(_parameters.getRotationCenter());
        _renderer.setAnaglyphMode(ViewSettings.ANAGLYPH_REDBLUE);
        this.refresh();
    }

    /**
    * Enables the red-cyan anaglyph.
    */
    public void anaglyphRedCyan() {
        _renderer.setStereo(false);
        _renderer.setStereoCenter(_parameters.getRotationCenter());
        _renderer.setAnaglyphMode(ViewSettings.ANAGLYPH_REDCYAN);
        this.refresh();
    }

    /**
    * Enables the red-green anaglyph.
    */
    public void anaglyphRedGreen() {
        _renderer.setStereo(false);
        _renderer.setStereoCenter(_parameters.getRotationCenter());
        _renderer.setAnaglyphMode(ViewSettings.ANAGLYPH_REDGREEN);
        this.refresh();
    }

    /**
    * Disables the anaglyph.
    */
    public void anaglyphOff() {
        _renderer.setAnaglyphMode(ViewSettings.ANAGLYPH_OFF);
        this.refresh();
    }

    /**
     * Sets the strength of ambient light.
     */
    public void setAmbient(float ambient)
    {
        _renderer.getViewSettings().setAmbient(ambient);
        _renderer.setLightChanged();
        this.refresh();
    }

    /**
     * Resets the strength of ambient light.
     */
    public void resetAmbient()
    {
        setAmbient(DEFAULT_AMBIENT);
    }

    /**
     * Sets the strength of diffuse light 1.
     */
    public void setDiffuse1(float diffuse1)
    {
        _renderer.getViewSettings().setDiffuse1(diffuse1);
        _renderer.setLightChanged();
        this.refresh();
    }

    /**
     * Resets the strength of diffuse light 1.
     */
    public void resetDiffuse1()
    {
        setDiffuse1(DEFAULT_DIFFUSE1);
    }

    /**
     * Sets the strength of diffuse light 2.
     */
    public void setDiffuse2(float diffuse2)
    {
        _renderer.getViewSettings().setDiffuse2(diffuse2);
        _renderer.setLightChanged();
        this.refresh();
    }

    /**
     * Resets the strength of diffuse light 2.
     */
    public void resetDiffuse2()
    {
        setDiffuse2(DEFAULT_DIFFUSE2);
    }

    /**
     * Sets the direction of diffuse light 1.
     */
    public void setDiffuseDir1(float x, float y, float z)
    {
        _renderer.getViewSettings().setDiffuseDir1(x, y, z);
        _renderer.setLightChanged();
        this.refresh();
    }

    /**
     * Resets the direction of diffuse light 1.
     */
    public void resetDiffuseDir1()
    {
        setDiffuseDir1(DEFAULT_DIFFUSE1_X, DEFAULT_DIFFUSE1_Y, DEFAULT_DIFFUSE1_Z);
    }

    /**
     * Sets the direction of diffuse light 2.
     */
    public void setDiffuseDir2(float x, float y, float z)
    {
        _renderer.getViewSettings().setDiffuseDir2(x, y, z);
        _renderer.setLightChanged();
        this.refresh();
    }

    /**
     * Resets the direction of diffuse light 2.
     */
    public void resetDiffuseDir2()
    {
        setDiffuseDir2(DEFAULT_DIFFUSE2_X, DEFAULT_DIFFUSE2_Y, DEFAULT_DIFFUSE2_Z);
    }

    /**
    * Starts the renderer.
    */
    public void startRenderer()
    {
        _renderer.start();
        this.refresh();
    }

    /**
    * Stops the renderer.
    */
    public void stopRenderer()
    {
        _renderer.stop();
    }

    /**
     * Refresh image.
     */
    public void refresh()
    {
        _renderer.render();
    }

    /**
     * Returns the current image.
     */
    public BufferedImage getImage()
    {
        return _renderer.getImage();
    }

    /**
     * Destroy GLWindow.
     */
    public void destroyGLWindow()
    {
        _renderer.destroyGLWindow();
    }

    //--------------------------------------------------------------------
    // File holder
    //--------------------------------------------------------------------
    /**
     * Returns a collection of file status.
     */
    public ArrayList getFileStatusList()
    {
        return _fileHolder.getFileStatusList();
    }

    /**
     * Returns a collection of molecule files.
     */
    public ArrayList getMoleculeFiles()
    {
        return _fileHolder.getMoleculeFiles();
    }

    /**
     * Returns a collection of polygon files.
     */
    public ArrayList getPolygonFiles()
    {
        return _fileHolder.getPolygonFiles();
    }

    /**
     * Appends file status listener.
     */
    public void addFileStatusListener(FileStatusListener fileStatusListener)
    {
        _fileHolder.addFileStatusListener(fileStatusListener);
    }

    /**
     * Removes file status listener.
     */
    public void removeFileStatusListener(FileStatusListener fileStatusListener)
    {
        _fileHolder.removeFileStatusListener(fileStatusListener);
    }

    /**
     * Returns the file specified by ID.
     */
    public FileStatus findFile(int fileId)
    {
        return _fileHolder.findFile(fileId);
    }

    /**
     * Change selection of files.
     */
    public boolean selectFile(boolean enable, int[] fileIds)
    {
        if( _fileHolder.selectFile(enable, fileIds) ){
            this.setupCurrentSphere();
            return true;
        }
        else return false;
    }

    /**
     * Selects files.
     */
    public boolean selectFile(int[] fileIds)
    {
        if( _fileHolder.selectFile(fileIds) ){
            this.setupCurrentSphere();
            return true;
        }
        else return false;
    }

    /**
     * Selects all files.
     */
    public void selectAllFile(boolean enable)
    {
        _fileHolder.selectAllFile(enable);
        this.setupCurrentSphere();
    }

    /**
     * Change display of files.
     */
    public boolean displayFile(boolean enable, int[] fileIds)
    {
        if( _fileHolder.displayFile(enable, fileIds) ){
            this.refresh();
            return true;
        }
        else return false;
    }

    /**
     * Displays files.
     */
    public boolean displayFile(int[] fileIds)
    {
        if( _fileHolder.displayFile(fileIds) ){
            this.refresh();
            return true;
        }
        else return false;
    }

    /**
     * Displays all files.
     */
    public void displayAllFile(boolean enable)
    {
        _fileHolder.displayAllFile(enable);
        this.refresh();
    }

    //--------------------------------------------------------------------
    // File open & close
    //--------------------------------------------------------------------
    /**
     * Load a molecule file and returns file ID assigned.
     */
    public int loadMoleculeFile(
        MoleculeFile moleculeFile, boolean defaultImage, FileStatus copy)
    {
        if(moleculeFile != null){
            this.add(moleculeFile, defaultImage, copy);
            this.refresh();
            return moleculeFile.getFileId();
        }
        else return -1;
    }

    /**
     * Load an polygon file and returns file ID assigned.
     */
    public int loadPolygonFile(PolygonFile polygonFile, FileStatus copy)
    {
        if(polygonFile != null){
            this.add(polygonFile, copy);
            this.refresh();
            return polygonFile.getFileId();
        }
        else return -1;
    }

    /**
     * Close all files.
     */
    public void closeAllFile()
    {
        ArrayList moleculeFiles = new ArrayList(_fileHolder.getMoleculeFiles());
        ArrayList polygonFiles = new ArrayList(_fileHolder.getPolygonFiles());

        this.close(moleculeFiles, polygonFiles);
        this.refresh();
    }

    /**
     * Close the specified files.
     */
    public boolean closeFile(int[] fileIds)
    {
        ArrayList moleculeFiles = new ArrayList();
        ArrayList polygonFiles = new ArrayList();

        if( _fileHolder.find(moleculeFiles, polygonFiles, fileIds) ){
            this.close(moleculeFiles, polygonFiles);
            this.refresh();
            return true;
        }
        else return false;
    }

    /**
     * Returns information of selected molecule files.
     */
    public String[] getInformation()
    {
        ArrayList listInfo = new ArrayList();

        ArrayList list = _fileHolder.getMoleculeFiles();
        for(int i=0; i<list.size(); i++){
            MoleculeFile file = (MoleculeFile)list.get(i);
            if( file.isSelected() ) listInfo.add( file.getInformation() );
        }
        return (String[])listInfo.toArray( new String[0] );
    }

    /**
     * Returns information of the molecule file specified by file ID.
     */
    public String getInformation(int fileId)
    {
        ArrayList list = _fileHolder.getMoleculeFiles();
        for(int i=0; i<list.size(); i++){
            MoleculeFile file = (MoleculeFile)list.get(i);
            if( file.getFileId() == fileId ) return file.getInformation();
        }
        return null;
    }

    //--------------------------------------------------------------------
    // Molecule manipulator
    //--------------------------------------------------------------------
    /**
     * Returns the selection status.
     */
    public SelectionStatus getSelectionStatus()
    {
        return _moleculeManipulator.getSelectionStatus();
    }

    /**
     * Returns the center coordinates of atoms.
     */
    public float[] findCenter(SelectExpr expr)
    {
        Vector4f center = _moleculeManipulator.findCenter(expr);
        if(center != null){
            return new float[]{center.x, center.y, center.z};
        }
        else return null;
    }

    /**
     * Selects atoms.
     */
    public int select(boolean all)
    {
        int cnt = _moleculeManipulator.select(all);
        _fileHolder.notifySelectAtom();
        return cnt;
    }

    /**
     * Selects atoms.
     */
    public int select(SelectExpr expr)
    {
        int cnt = _moleculeManipulator.select(expr);
        _fileHolder.notifySelectAtom();
        return cnt;
    }

    /**
     * Selects none.
     */
    public void selectOff()
    {
        _moleculeManipulator.selectOff();
        _fileHolder.notifySelectAtom();
    }

    /**
    * displays atoms.
    */
    public void displayAtom(boolean enable)
    {
        _moleculeManipulator.display(enable);
        this.refresh();
    }

    /**
     * Enable wireframes.
     */
    public void wireframe(float radius)
    {
        if(radius >= 0 && radius <= MAX_BOND_RADIUS){
            _moleculeManipulator.wireframe(radius);
            this.refresh();
        }
    }

    /**
     * Disable wireframes.
     */
    public void wireframeOff()
    {
        _moleculeManipulator.wireframeOff();
        this.refresh();
    }

    /**
     * Enable spacefills.
     */
    public void spacefill(float radius)
    {
        if(radius > 0 && radius <= MAX_BALL_RADIUS){
            _moleculeManipulator.spacefill(radius);
            this.refresh();
        }
        else if(radius == 0){
            this.spacefillOff();
        }
    }

    /**
     * Enable spacefills with Van der Waals radius.
     */
    public void spacefillVdW()
    {
        _moleculeManipulator.spacefillVdW();
        this.refresh();
    }

    /**
     * Enable spacefills using temperature factor.
     */
    public void spacefillTemp()
    {
        _moleculeManipulator.spacefillTemp();
        this.refresh();
    }

    /**
     * Disable spacefills.
     */
    public void spacefillOff()
    {
        _moleculeManipulator.spacefillOff();
        this.refresh();
    }

    /**
     * Enable backbones.
     */
    public void backbone(float radius)
    {
        if(radius >= 0 && radius <= MAX_BOND_RADIUS){
            _moleculeManipulator.backbone(radius);
            this.refresh();
        }
    }

    /**
     * Disable backbones.
     */
    public void backboneOff()
    {
        _moleculeManipulator.backboneOff();
        this.refresh();
    }

    /**
     * Enable hydrogen bonds.
     */
    public void hBond(float radius)
    {
        if(radius >= 0 && radius <= MAX_BOND_RADIUS){
            _moleculeManipulator.hBond(radius);
            this.refresh();
        }
    }

    /**
     * Disable hydrogen bonds.
     */
    public void hBondOff()
    {
        _moleculeManipulator.hBondOff();
        this.refresh();
    }

    /**
     * Switches the hydrogen bonds.
     */
    public void setHBondBackbone(boolean enable)
    {
        _moleculeManipulator.setHBondBackbone(enable);
        this.refresh();
    }

    /**
     * Enable disulfide bonds.
     */
    public void ssBond(float radius)
    {
        if(radius >= 0 && radius <= MAX_BOND_RADIUS){
            _moleculeManipulator.ssBond(radius);
            this.refresh();
        }
    }

    /**
     * Disable disulfide bonds.
     */
    public void ssBondOff()
    {
        _moleculeManipulator.ssBondOff();
        this.refresh();
    }

    /**
     * Switches the disulfide bonds.
     */
    public void setSSBondBackbone(boolean enable)
    {
        _moleculeManipulator.setSSBondBackbone(enable);
        this.refresh();
    }

    /**
     * Enable ribbons with default width.
     */
    public void ribbon()
    {
        _moleculeManipulator.ribbon();
        this.refresh();
    }

    /**
     * Enable ribbons.
     */
    public void ribbon(float width)
    {
        if(width > 0 && width <= MAX_RIBBON_WIDTH){
            _moleculeManipulator.ribbon(width);
            this.refresh();
        }
        else if(width == 0){
            this.ribbonOff();
        }
    }

    /**
     * Disable ribbons.
     */
    public void ribbonOff()
    {
        _moleculeManipulator.ribbonOff();
        this.refresh();
    }

    /**
     * Enable cartoons with default width.
     */
    public void cartoon()
    {
        _moleculeManipulator.cartoon();
        this.refresh();
    }

    /**
     * Enable cartoons.
     */
    public void cartoon(float width)
    {
        if(width > 0 && width <= MAX_RIBBON_WIDTH){
            _moleculeManipulator.cartoon(width);
            this.refresh();
        }
        else if(width == 0){
            this.cartoonOff();
        }
    }

    /**
     * Disable cartoons.
     */
    public void cartoonOff()
    {
        _moleculeManipulator.cartoonOff();
        this.refresh();
    }

    /**
     * Enable trace.
     */
    public void trace(float radius)
    {
        if(radius > 0 && radius <= MAX_BOND_RADIUS){
            _moleculeManipulator.trace(radius);
            this.refresh();
        }
        else if(radius == 0){
            this.traceOff();
        }
    }

    /**
     * Disable trace.
     */
    public void traceOff()
    {
        _moleculeManipulator.traceOff();
        this.refresh();
    }

    /**
     * Eliminates hetero atoms.
     */
    public void eliminateHeteroAtoms()
    {
        _moleculeManipulator.eliminateHeteroAtoms();
        this.refresh();
    }
    
    /**
     * Eliminates hydrogens.
     */
    public void eliminateHydrogens()
    {
        _moleculeManipulator.eliminateHydrogens();
        this.refresh();
    }
    
    /**
     * Sets the height of cartoon model.
     */
    public void setCartoonHeight(float height)
    {
        if(height >= MIN_CARTOON_HEIGHT && height <= MAX_CARTOON_HEIGHT){
            _moleculeManipulator.setCartoonHeight(height);
            this.refresh();
        }
    }

    /**
     * Resets the height of cartoon model.
     */
    public void resetCartoonHeight()
    {
        this.setCartoonHeight(MoleculeImage.DEFAULT_CARTOON_HEIGHT);
    }

    /**
     * Sets the back color of ribbon/cartoon model.
     */
    public void setColorRibbonback(Color color)
    {
        _moleculeManipulator.setResidueBackColor(color);
        this.refresh();
    }

    /**
     * Resets the back color of ribbon/cartoon model.
     */
    public void resetColorRibbonback()
    {
        _moleculeManipulator.resetResidueBackColor();
        this.refresh();
    }

    /**
     * Switches the loop cartoon.
     */
    public void setCartoonLoopTube(boolean enable)
    {
        _moleculeManipulator.setCartoonLoopTube(enable);
        this.refresh();
    }

    /**
     * Switches the cartoon round.
     */
    public void setCartoonRound(boolean enable)
    {
        _moleculeManipulator.setCartoonRound(enable);
        this.refresh();
    }

    /**
     * Sets colors.
     */
    public void setColor(ColorExpr expr)
    {
        _moleculeManipulator.setColor(expr);
        this.refresh();
    }

    //--------------------------------------------------------------------
    // Polygons
    //--------------------------------------------------------------------
    /**
     * Selects vertices.
     */
    public int selectVertex(SelectExpr expr)
    {
        int cnt = _polygonManipulator.select(expr);
        _fileHolder.notifySelectVertex();
        return cnt;
    }

    /**
    * Selects all vertices.
    */
    public int selectVertexAll()
    {
        int cnt = _polygonManipulator.selectAll();
        _fileHolder.notifySelectVertex();
        return cnt;
    }

    /**
    * Selects no vertices.
    */
    public int selectVertexNone()
    {
        int cnt = _polygonManipulator.selectNone();
        _fileHolder.notifySelectVertex();
        return cnt;
    }

    /**
    * displays vertices.
    */
    public void displayVertex(boolean enable)
    {
        _polygonManipulator.display(enable);
        this.refresh();
    }

    /**
     * Sets the transparency of polygons.
     */
    public void setTransparency(int target, float transparency)
    {
        if(transparency >= 0 && transparency <= 1){
            _polygonManipulator.setTransparency(target, transparency);
            this.refresh();
        }
    }

    /**
     * Resets the transparency of polygons.
     */
    public void resetTransparency(int target)
    {
        _polygonManipulator.resetTransparency(target);
        this.refresh();        
    }

    /**
     * Sets the size of polygons (point size and line width).
     */
    public void setPolygonSize(int target, float size)
    {
        if(size > 0){
            _polygonManipulator.setPolygonSize(target, size);
            this.refresh();        
        }
    }

    /**
     * Resets the size of polygons (point size and line width).
     */
    public void resetPolygonSize(int target)
    {
        _polygonManipulator.resetPolygonSize(target);
        this.refresh();        
    }

    /**
     * Sets the color of vertices.
     */
    public void setColorVertex(Color color)
    {
        _polygonManipulator.setColor(color);
        this.refresh();        
    }

    /**
     * Resets the color of vertices.
     */
    public void resetColorVertex()
    {
        _polygonManipulator.resetColor();
        this.refresh();        
    }

    //--------------------------------------------------------------------
    // Transform
    //--------------------------------------------------------------------
    /**
     * Rotate about x-axis by the specified angle (in radians).
     */
    public void rotateX(float angle)
    {
        Matrix4f rotation = new Matrix4f();
        rotation.rotationX(angle);
        this.rotate(rotation);
    }

    /**
     * Rotate about y-axis by the specified angle (in radians).
     */
    public void rotateY(float angle)
    {
        Matrix4f rotation = new Matrix4f();
        rotation.rotationY(angle);
        this.rotate(rotation);
    }

    /**
     * Rotate about z-axis by the specified angle (in radians).
     */
    public void rotateZ(float angle)
    {
        Matrix4f rotation = new Matrix4f();
        rotation.rotationZ(angle);
        this.rotate(rotation);
    }

    /**
     * Rotate by the specified rotation matrix.
     */
    public void rotate(Matrix4f rotation)
    {
        _transformHandler.rotate(rotation);
        this.refresh();
    }

    /**
     * Translate from the current state.
     */
    public void translate(float x, float y, float z)
    {
        Vector3f translation = new Vector3f(x, y, z);
        this.translate(translation);
    }

    /**
     * Translate from the current state.
     */
    public void translate(Vector3f translation)
    {
        _transformHandler.translate(translation);
        this.moveRotationCenter(translation);
        this.refresh();
    }

    /**
     * Sets the center of rotation.
     */
    public void setRotationCenter(Vector3f rotationCenter)
    {
        _parameters.getRotationCenter().set(rotationCenter);
        _renderer.setStereoCenter( _parameters.getRotationCenter() );
    }

    /**
     * Moves the center of rotation.
     */
    public void moveRotationCenter(Vector3f displacement)
    {
        Vector3f center = _parameters.getRotationCenter();
        center.add(displacement);
        _renderer.setStereoCenter( _parameters.getRotationCenter() );
    }

    /**
     * Resets the center of rotation.
     */
    public void resetRotationCenter()
    {
        this.setRotationCenter( _parameters.getDefaultCenter() );
    }

    /**
     * Sets the center of files.
     */
    public void setCenter(Vector3f center)
    {
        Vector3f offset = new Vector3f();
        ArrayList moleculeList = _fileHolder.getMoleculeFiles();
        for(int i=0; i<moleculeList.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)moleculeList.get(i);
            if( file.isSelected() ){
                offset.sub( center, file.center() );
                file.offset().set(offset);
            }
        }
        ArrayList polygonList = _fileHolder.getPolygonFiles();
        for(int i=0; i<polygonList.size(); i++)
        {
            PolygonFile file = (PolygonFile)polygonList.get(i);
            if (file.getRadius() == -1) {
                continue;
            }
            if( file.isSelected() ){
                offset.sub( center, file.center() );
                file.offset().set(offset);
            }
        }
        this.setupDefaultView();
    }

    /**
     * Copies the transform.
     */
    public boolean copyTransform(int srcFileId, int dstFileId)
    {
        if( _transformHandler.copyTransform(srcFileId, dstFileId) ){
            this.refresh();
            return true;
        }
        else return false;
    }

    //--------------------------------------------------------------------
    // View
    //--------------------------------------------------------------------
    /**
     * Sets the viewpoint.
     */
    public void setViewpoint(float x, float y, float z)
    {
        _vector1.set(x, y, z);
        this.adjustCameraPosition(_vector1);
        this.refresh(); 
    }

    /**
     * Zoom from the current state.
     */
    public void zoom(float ratio)
    {
        if(ratio >= MIN_ZOOM_RATIO)
        {
            _vector1.set( _renderer.getViewSettings().getCameraPosition() );
            float centerZ = _parameters.getCurrentCenter().z;

            if(_vector1.z > centerZ){
                _vector1.z = (_vector1.z + (ratio - 1)*centerZ)/ratio;
                this.adjustCameraPosition(_vector1);
                this.adjustFrontClip(); 

                if( _renderer.getViewSettings().isParallelProjection() )
                {
                    float size = _renderer.getViewSettings().getViewSize();
                    size /= ratio;
                    setupViewSize(size);
                    //_renderer.getViewSettings().setViewSize(size);
                }

                this.refresh(); 
            }
        }
    }

    /**
     * Resets the magnification of zoom.
     */
    public void resetZoom()
    {
        _vector1.set( _renderer.getViewSettings().getCameraPosition() );
        _vector1.z = _parameters.getDefaultCameraPosition().z;
        this.adjustCameraPosition(_vector1);
        this.adjustFrontClip(); 

        if( _renderer.getViewSettings().isParallelProjection() ){
            setupViewSize();
            //_renderer.getViewSettings().setViewSize( this.getParallelSize() );
        }
        this.refresh(); 
    }

    /**
     * Moves the center of screen.
     */
    public void moveScreenCenter(float x, float y)
    {
        Vector3f camera = _renderer.getViewSettings().getCameraPosition();
        this.setScreenCenter(camera.x + x, camera.y + y);
    }

    /**
     * Sets the center of screen.
     */
    public void setScreenCenter(float x, float y)
    {
        _vector1.x = x;
        _vector1.y = y;
        _vector1.z = _renderer.getViewSettings().getCameraPosition().z;
        this.adjustCameraPosition(_vector1);
        this.refresh();
    }

    /**
     * Resets the center of screen.
     */
    public void resetScreenCenter()
    {
        Vector3f center = _parameters.getDefaultCenter();
        this.setScreenCenter(center.x, center.y); 
    }

    /**
     * Enables slab mode.
     */
    public void slab()
    {
        this.slabRatio(0.5f);
    }

    /**
     * Enables slab mode with the specified ratio.
     */
    public void slabRatio(float ratio)
    {
        if(ratio >= 0 && ratio <= 1){
            float cz = _parameters.getCurrentCenter().z;
            float r  = _parameters.getCurrentRadius();
            float slabZ = cz - (1 - 2*ratio)*r;
            this.slabValue(slabZ);
        }
    }

    /**
     * Enables slab mode with the specified z coordinate.
     */
    public void slabValue(float value)
    {
        _viewStatus.setSlab(true);
        _viewStatus.setSlabValue(value);
        this.adjustFrontClip();
        this.refresh();
    }

    /**
     * Moves slab plane.
     */
    public void moveSlabPlane(float amount)
    {
        if( _parameters.isSlab() ){
            _viewStatus.setSlabValue(_parameters.getSlabValue() + amount);
            this.adjustFrontClip();
            this.refresh();
        }
    }

    /**
     * Disables slab mode.
     */
    public void slabOff()
    {
        _viewStatus.setSlab(false);
        _viewStatus.setSlabValue(ViewParameters.SLABVALUE_UNDEFINED);
        this.adjustFrontClip();
        this.refresh();        
    }

    /**
     * Sets using the parallel projection with a default size.
     */
    public void setParallelProjection()
    {
        setParallelProjection(getParallelSize());
    }

    /**
     * Sets using the parallel projection with the specified half size.
     */
    public void setParallelProjection(float size)
    {
        _renderer.getViewSettings().setParallelProjection(true);
        setupViewSize(size);
        //_renderer.getViewSettings().setViewSize(size);
        _renderer.setPerspectiveChanged();
        this.refresh();
        _currViewSize = size;
    }

    /**
     * Sets using the parallel projection.
     */
    public void setPerspectiveProjection()
    {
        _renderer.getViewSettings().setParallelProjection(false);
        _renderer.setPerspectiveChanged();
        this.refresh();
    }

    /**
     * Modifies the view size.
     */
    public void modifyViewSize(float changeAmount, boolean refresh) {
        _currViewSize += changeAmount;
        setupViewSize(_currViewSize);
        _renderer.setPerspectiveChanged();
        if (refresh) {
            refresh();
        }
    }

    private float getParallelSize()
    {
        float cz = _renderer.getViewSettings().getCameraPosition().z;
        float cc = _parameters.getCurrentCenter().z;
        float angle = _renderer.getViewSettings().getFieldAngle();
        float size = (cz - cc)*(float)Math.tan(angle/2);
        //if(size < 0) size *= -1;
        //if(size < 1) size = 1;
        return size;
    }

    private void setupViewSize() {
        setupViewSize(getParallelSize());
    }

    private void setupViewSize(float size) {
        _currViewSize = size;
        if (size < 0.1f) {
            size = 0.1f;
        }
        _renderer.getViewSettings().setViewSize(size);
    }
    
    /**
     * Sets whether to load to center.
     */
    public void setLoadCenter(boolean enable)
    {
        _parameters.setLoadCenter(enable);
    }

    /**
     * Sets whether to adjust view.
     */
    public void setAdjustView(boolean enable)
    {
        _parameters.setAdjustView(enable);
    }

    //--------------------------------------------------------------------
    // Reset
    //--------------------------------------------------------------------
    /**
     * Resets transform of all files, center of rotation and view point.
     */
    public void reset()
    {
        _transformHandler.resetTransform(true);

        this.setRotationCenter( _parameters.getDefaultCenter() );
        this.adjustCameraPosition( _parameters.getDefaultCameraPosition() );
        this.setupCurrentSphere();
        this.adjustFrontClip();
        if( _renderer.getViewSettings().isParallelProjection() ){
            setupViewSize();
            //_renderer.getViewSettings().setViewSize( this.getParallelSize() );
        }
        this.refresh();
    }

    //--------------------------------------------------------------------
    // Secondary structure
    //--------------------------------------------------------------------
    /**
     * Determines the secondary structure.
     */
    public void structure()
    {
        ArrayList moleculeList = _fileHolder.getMoleculeFiles();
        for(int i=0; i<moleculeList.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)moleculeList.get(i);
            if( !file.isSelected() ) continue;
            for(int m=0; m<file.numMolecules(); m++)
            {
                Molecule molecule = file.getMolecule(m);
                if(molecule.getStructure() != Molecule.STRUCT_CALCULATION &&
                        molecule.getStructure() != Molecule.STRUCT_GROMACS){
                    MoleculeUtil util = new MoleculeUtil();
                    util.determineStructure(molecule);                    
                }
            }
        }
    }

    //--------------------------------------------------------------------
    // External site
    //--------------------------------------------------------------------
    /**
     * Sets PDBML URL.
     */
    public void setPDBmlUrl(PDBmlUrl pdbmlUrl) {
        _pdbmlUrl = pdbmlUrl;
    }

    /**
     * Returns PDBML URL.
     */
    public PDBmlUrl getPDBmlUrl() {
        return _pdbmlUrl;
    }
    
    /**
     * Sets eF-site URL.
     */
    public void setEfsiteUrl(EfsiteUrl efsiteUrl) {
        _efsiteUrl = efsiteUrl;
    }

    /**
     * Returns eF-site URL.
     */
    public EfsiteUrl getEfsiteUrl() {
        return _efsiteUrl;
    }

    /**
     * Sets external site URL map.
     */
    public void setSiteUrlMap(SiteUrlMap siteUrlMap) {
        _siteUrlMap = siteUrlMap;
    }

    /**
     * Returns external site URL map.
     */
    public SiteUrlMap getSiteUrlMap() {
        return _siteUrlMap;
    }
    
    //--------------------------------------------------------------------
    // Initialize status
    //--------------------------------------------------------------------
    private void initializeStatus()
    {
        // initialize ViewSettings
        ViewSettings settings = _renderer.getViewSettings(); 
        settings.setCameraPosition( new Vector3f(0, 0, 0) );
        settings.setCameraTarget( new Vector3f(0, 0, -1) );
        settings.setCameraUpVector( new Vector3f(0, 1, 0) );
        settings.setFieldAngle(DEFAULT_FIELD_ANGLE);
        settings.setNearClipping(DEFAULT_NEAR_CLIP);
        settings.setFarClipping(DEFAULT_FAR_CLIP);
        settings.setBGColor(ViewSettings.DEFAULT_BACKGROUND);

        float c = DEFAULT_SPECULAR_RATIO;
        settings.setSpecular(false);
        settings.setSpecularColor(c, c, c, 1.0f);
        settings.setShininess(DEFAULT_SHININESS);

        settings.setDrawlevel(DEFAULT_DRAWLEVEL);

        settings.setAmbient(DEFAULT_AMBIENT);
        settings.setDiffuse1(DEFAULT_DIFFUSE1);
        settings.setDiffuse2(DEFAULT_DIFFUSE2);
        settings.setDiffuseDir1(DEFAULT_DIFFUSE1_X, DEFAULT_DIFFUSE1_Y, DEFAULT_DIFFUSE1_Z);
        settings.setDiffuseDir2(DEFAULT_DIFFUSE2_X, DEFAULT_DIFFUSE2_Y, DEFAULT_DIFFUSE2_Z);

        _renderer.setStereoAngle( (float)(DEFAULT_STEREO_ANGLE*Math.PI/180) );

        _renderer.setLightChanged();
        _renderer.setCameraChanged();
        _renderer.setPerspectiveChanged();
        _renderer.setBackgroundChanged();
        _renderer.setDrawlevelChanged();
    }

    //--------------------------------------------------------------------
    // Handle view
    //--------------------------------------------------------------------
    private void adjustFrontClip()
    {
        if( _parameters.isSlab() ){
            float vz = _renderer.getViewSettings().getCameraPosition().z;
            float sz = _parameters.getSlabValue();
            float clip = vz - sz;
            if(clip >= DEFAULT_NEAR_CLIP){
                _renderer.getViewSettings().setNearClipping(clip);
            }
        }
        else{
            _renderer.getViewSettings().setNearClipping(DEFAULT_NEAR_CLIP);
        }
        _renderer.setPerspectiveChanged();
    }

    private void adjustCameraPosition(Vector3f position)
    {
        _renderer.getViewSettings().getCameraPosition().set(position);
        _renderer.getViewSettings().getCameraTarget().set(
            position.x, position.y, position.z-1
        );
        _renderer.setCameraChanged();
    }

    private void adjustViewOpen()
    {
        this.setupDefaultView();
        this.setupCurrentSphere();

        if( _parameters.isAdjustView() )
        {
            // camera position
            this.adjustCameraPosition( _parameters.getDefaultCameraPosition() );

            // slab off
            _viewStatus.setSlab(false);
            _viewStatus.setSlabValue(ViewParameters.SLABVALUE_UNDEFINED);
            this.adjustFrontClip();

            // set rotation center
            if(_fileHolder.getFileStatusList().size() == 1){
                _parameters.getRotationCenter().set( _parameters.getDefaultCenter() );
            }

            // parallel projection
            if( _renderer.getViewSettings().isParallelProjection() ){
                setupViewSize();
            }
        }
    }

    private void adjustViewClose()
    {
        this.setupDefaultView();
        this.setupCurrentSphere();

        // initialize
        if(_parameters.isAdjustView() && _fileHolder.getFileStatusList().size() == 0)
        {
            _viewStatus.setSlab(false);
            _viewStatus.setSlabValue(ViewParameters.SLABVALUE_UNDEFINED);
            this.adjustFrontClip();

            this.adjustCameraPosition( new Vector3f(0, 0, 0) );

            _parameters.getDefaultCameraPosition().set( new Vector3f(0, 0, 0) );
            _parameters.getRotationCenter().set(_parameters.getDefaultCenter());
        }
    }

    private void setupDefaultView()
    {
        this.setupDefaultSphere();

        Vector3f center = _parameters.getDefaultCenter();
        float radius    = _parameters.getDefaultRadius();
        float dist = radius/(DEFAULT_RATIO * (float)Math.sin(DEFAULT_FIELD_ANGLE/2));

        Vector3f camera = new Vector3f(center.x, center.y, center.z + dist);
        _parameters.getDefaultCameraPosition().set(camera);
    }

    private void setupDefaultSphere()
    {
        _transformHandler.setupDefaultSphere();
    }

    private void setupCurrentSphere()
    {
        _transformHandler.setupCurrentSphere();
    }

    //--------------------------------------------------------------------
    // Handle files
    //--------------------------------------------------------------------
    private void close(ArrayList moleculeFiles, ArrayList polygonFiles)
    {
        for(int i=0; i<moleculeFiles.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)moleculeFiles.get(i);
            _renderer.remove(file);
            _moleculeManipulator.remove(file);
            _polygonManipulator.remove(file);
            _fileHolder.remove(file);
        }
        for(int i=0; i<polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)polygonFiles.get(i);

            if (file.getRadius() != -1) {
                _renderer.remove(file);
                _polygonManipulator.remove(file);
            }

            _fileHolder.remove(file);
        }

        this.adjustViewClose();
    }

    private void add(MoleculeFile moleculeFile, boolean defaultImage, FileStatus copy)
    {
        if( _parameters.isLoadCenter() ){
            Vector3f offset = new Vector3f();
            offset.sub( this.getLoadCenter(), moleculeFile.center() );
            moleculeFile.offset().set(offset);
            _transformHandler.resetTransform(moleculeFile);
        }

        moleculeFile.setSelected(true);
        moleculeFile.setDisplayed(true);

        this.initMoleculeFile(moleculeFile, defaultImage);

        _fileHolder.add(moleculeFile);
        _moleculeManipulator.add(moleculeFile);
        _polygonManipulator.add(moleculeFile);
        _renderer.add(moleculeFile);
        
        if(copy != null){
            int srcId = copy.getId();
            int dstId = moleculeFile.getFileId();
            _transformHandler.copyTransform(srcId, dstId);
        }

        this.adjustViewOpen();
    }

    private void add(PolygonFile polygonFile, FileStatus copy)
    {
        boolean valid = (polygonFile.getRadius() != -1);
        
        if( valid && _parameters.isLoadCenter() ){
            Vector3f offset = new Vector3f();
            offset.sub( this.getLoadCenter(), polygonFile.center() );
            polygonFile.offset().set(offset);
            _transformHandler.resetTransform(polygonFile);
        }

        polygonFile.setSelected(true);
        polygonFile.setDisplayed(true);

        _fileHolder.add(polygonFile);

        if (valid) {
            _renderer.add(polygonFile);
            _polygonManipulator.add(polygonFile);
        }

        if(copy != null){
            int srcId = copy.getId();
            int dstId = polygonFile.getFileId();
            _transformHandler.copyTransform(srcId, dstId);
        }

        this.adjustViewOpen();
    }

    private void initMoleculeFile(MoleculeFile file, boolean defaultImage)
    {
        MoleculeManipulator manipulator = new MoleculeManipulator();
        manipulator.setColorManager(_colorManager);
        manipulator.add(file);

        // color cpk
        ColorExpr expr = new ColorExpr();
        expr.setTarget(ColorExpr.TARGET_ATOM);
        expr.setScheme(ColorExpr.SCHEME_CPK);
        manipulator.setColor(expr);

        // handle AltLoc
        for(int m=0; m<file.numMolecules(); m++)
        {
        Molecule molecule = file.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            chain.getResidue(re).selectFirstAltLoc();
        }}}

        // cartoon back color
        int backColorIndex = _colorManager.getIndex(new Color(125, 125, 125));
        for(int m=0; m<file.numMolecules(); m++)
        {
            file.getMolecule(m).getImage().setResidueBackColorIndex(backColorIndex);
        }

        if(defaultImage)
        {
            manipulator.wireframe(0);
    
            // backbone if no bonds
            boolean backbone = false;
            for(int m=0; m<file.numMolecules(); m++){
                Molecule molecule = file.getMolecule(m);
                if(molecule.numBonds() == 0) backbone = true;
            }
            if(backbone) manipulator.backbone(0.4f);
        }
    }
    
    private Vector3f getLoadCenter()
    {
        Sphere sphere = _transformHandler.createSphere(true, true);
        return sphere.center();
    }
}
