/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import org.pdbj.viewer.molecules.*;

/**
 * Animation File.
 */
public class AnimationFile
{
    public static final int ONCE  = 0;
    public static final int LOOP  = 1;
    public static final int SWING = 2;

    private MoleculeFile _file;

    private int _mode;
    private int _step;

    private int _init;
    private int _term;

    private boolean _reverse;
    private boolean _running;
    private boolean _selected;

    private int _numFrames;

    private boolean _initialized;
    private boolean _ignore;

    private AnimationHandler _handler;
    
    private JCheckBox _checkBoxSelect;
    private JSlider   _sliderFrame;
    private JSpinner  _spinnerFrame;
    private JSpinner  _spinnerStep;
    
    private boolean _ctrl;

    /**
     * Returns the file ID.
     */
    public int getFileId()
    {
        return _file.getFileId();
    }

    /**
     * Returns the number of frames.
     */
    public int numFrames()
    {
        return _numFrames;
    }

    /**
     * Returns whether it is reversed.
     */
    public boolean isReverse()
    {
        return _reverse;
    }

    /**
     * Returns whether it is running.
     */
    public boolean isRunning()
    {
        return _running;
    }

    /**
     * Returns whether it is selected.
     */
    public boolean isSelected()
    {
        return _selected;
    }

    /**
     * Returns the mode.
     */
    public int getMode()
    {
        return _mode;
    }

    /**
     * Returns the step.
     */
    public int getStep()
    {
        return _step;
    }

    /**
     * Returns the initial frame of the current range.
     */
    public int getInitRange()
    {
        return _init;
    }

    /**
     * Returns the terminal frame of the current range.
     */
    public int getTermRange()
    {
        return _term;
    }

    /**
     * Sets whether it is reversed.
     */
    public void setReverse(boolean reverse)
    {
        _reverse = reverse;
    }

    /**
     * Sets whether it is running.
     */
    public void setRunning(boolean running)
    {
        _running = running;
    }

    /**
     * Sets whether it is selected.
     */
    public void setSelected(boolean selected)
    {
        _selected = selected;
        this.updateComponent();
    }

    /**
     * Sets the mode.
     */
    public void setMode(int mode)
    {
        _mode = mode;
    }

    /**
     * Sets the step.
     */
    public void setStep(int step)
    {
        _step = step;
        this.updateComponent();
    }

    /**
     * Sets the range.
     */
    public void setRange(int init, int term)
    {
        if(init >= 0 && term >= init)
        {
            if(init >= _numFrames){
                _init = _term = _numFrames - 1;
            }
            else if(term >= _numFrames){
                _init = init;
                _term = _numFrames - 1;
            }
            else{
                _init = init;
                _term = term;
            }
        }
    }

    /**
     * Returns the select checkbox.
     */
    public JCheckBox getCheckBoxSelect()
    {
        return _checkBoxSelect;
    }

    /**
     * Returns the frame slider.
     */
    public JSlider getSliderFrame()
    {
        return _sliderFrame;
    }

    /**
     * Returns the frame spinner.
     */
    public JSpinner getSpinnerFrame()
    {
        return _spinnerFrame;
    }

    /**
     * Returns the step spinner.
     */
    public JSpinner getSpinnerStep()
    {
        return _spinnerStep;
    }

    /**
     * Constructor.
     */
    public AnimationFile(MoleculeFile file, AnimationHandler handler)
    {
        _file = file;
        _handler = handler;

        _step = 1;

        if(_file != null){
            _numFrames = _file.numMolecules();
            _init = 0;
            _term = _numFrames - 1;
        }
    }

    /**
     * Sets the current frame.
     */
    public void setFrame(int frame)
    {
        _file.setCurrentFrame(frame);

        this.updateComponent();
    }

    /**
     * Sets the current frame to the first one.
     */
    public void firstFrame()
    {
        this.setFrame(0);
    }

    /**
     * Moves forward by the current step.
     */
    public void stepForward()
    {
        this.stepForward(_step);
    }

    /**
     * Moves forward by the specified step.
     */
    public void stepForward(int step)
    {
        if(step > 0){
            int frame = _file.getCurrentFrame() + step;
            if(frame >= _numFrames) frame = _numFrames - 1;
 
            this.setFrame(frame);
        }
    }

    /**
     * Moves frame.
     */
    public void move()
    {
        int frame = _file.getCurrentFrame();

        if(frame < _init){
            frame = _init;
        }
        else if(frame > _term){
            frame = _term;
        }

        if(_reverse){
            frame -= _step;
            if(frame < _init){
                if(_mode == ONCE){
                    frame = _init;
                    _running = false;
                }
                else if(_mode == LOOP){
                    frame = frame + (_term - _init + 1);
                }
                else if(_mode == SWING){
                    frame = _init - (frame - _init) - 1;
                    _reverse = false;
                }
            }
        }
        else{
            frame += _step;
            if(frame > _term){
                if(_mode == ONCE){
                    frame = _term;
                    _running = false;
                }
                else if(_mode == LOOP){
                    frame = frame - (_term - _init + 1);
                }
                else if(_mode == SWING){
                    frame = _term - (frame - _term) + 1;
                    _reverse = true;
                }
            }
        }
        if(frame >= _init && frame <= _term){
            this.setFrame(frame);
        }
        else{
            _running = false;            
        }
    }

    /**
     * Moves backward by the current step.
     */
    public void stepBackward()
    {
        this.stepBackward(_step);
    }

    /**
     * Moves backward by the specified step.
     */
    public void stepBackward(int step)
    {
        if(step > 0){
            int frame = _file.getCurrentFrame() - step;
            if(frame < 0) frame = 0;
 
            this.setFrame(frame);
        }
    }

    /**
     * Sets the current frame to the last one.
     */
    public void lastFrame()
    {
        this.setFrame(_numFrames - 1);
    }

    /**
     * Initializes component.
     */
    public void initializeComponent()
    {
        if(!_initialized)
        {
            this.createComponent();
            _initialized = true;
            
            if(_file != null) this.updateComponent();
        }
    }

    //--------------------------------------------------------------------
    // Create component
    //--------------------------------------------------------------------
    private void updateComponent()
    {
        if(_initialized){
            _ignore = true;
            int frame = _file.getCurrentFrame();
            _sliderFrame.setValue(frame + 1);
            _spinnerFrame.setValue( new Integer(frame + 1) );
            _spinnerStep.setValue( new Integer(_step) );
            _checkBoxSelect.setSelected(_selected);
            _ignore = false;
        }
    }
    
    private void createComponent()
    {
        _checkBoxSelect = new JCheckBox();
        _checkBoxSelect.setSize(40, 20);
        _checkBoxSelect.setText("");
        
        _spinnerFrame = new JSpinner( new SpinnerNumberModel(1, 1, 1, 1) );
        _spinnerFrame.setSize(54, 20);

        _sliderFrame = new JSlider();
        _sliderFrame.setSize(300, 20);
        _sliderFrame.setMinimum(1);
        _sliderFrame.setMaximum(1);
        _sliderFrame.setValue(1);
        _sliderFrame.setSnapToTicks(true);
        
        _spinnerStep = new JSpinner( new SpinnerNumberModel(1, 1, 1, 1) );
        _spinnerStep.setSize(50, 20);

        if(_file != null)
        {
            _checkBoxSelect.setText(_file.getFileId() + ".");
            _checkBoxSelect.addItemListener(
                new ItemListener(){
                    public void itemStateChanged(ItemEvent e){
                        checkBoxSelect_itemStateChanged(e);
                    }
                }
            );
            // add KeyListener
            _checkBoxSelect.addMouseListener(
                new MouseAdapter(){
                    public void mousePressed(MouseEvent e){
                        checkBoxSelect_mousePressed(e);
                    }
                    public void mouseReleased(MouseEvent e){
                        checkBoxSelect_mouseReleased(e);
                    }
                }
            );
            
            SpinnerNumberModel model = null;
            
            model = (SpinnerNumberModel)_spinnerFrame.getModel();
            model.setMaximum( new Integer( _file.numMolecules() ) );
            _spinnerFrame.addChangeListener(
                new ChangeListener(){
                    public void stateChanged(ChangeEvent e){
                        spinnerFrame_stateChanged(e);
                    }
                }
            );

            _sliderFrame.setMaximum( _file.numMolecules() );
            _sliderFrame.addChangeListener(
                new ChangeListener(){
                    public void stateChanged(ChangeEvent e){
                        sliderFrame_stateChanged(e);
                    }
                }
            );

            model = (SpinnerNumberModel)_spinnerStep.getModel();
            model.setMaximum( new Integer( _file.numMolecules() ) );
            _spinnerStep.addChangeListener(
                new ChangeListener(){
                    public void stateChanged(ChangeEvent e){
                        spinnerStep_stateChanged(e);
                    }
                }
            );
        }
        else
        {
            _checkBoxSelect.setEnabled(false);
            _spinnerFrame.setEnabled(false);
            _sliderFrame.setEnabled(false);
            _spinnerStep.setEnabled(false);
        }
    }

    //--------------------------------------------------------------------
    // Event handler
    //--------------------------------------------------------------------
    private void checkBoxSelect_itemStateChanged(ItemEvent e)
    {
        if(!_ignore){
            _selected = ( e.getStateChange() == ItemEvent.SELECTED );
            if(_selected && _ctrl){
                _handler.select( new int[]{_file.getFileId()} );
            }
        }
    }

    private void checkBoxSelect_mousePressed(MouseEvent e)
    {
        if( (e.getModifiers() & MouseEvent.CTRL_MASK) != 0)
        {
            _ctrl = true;
        }
    }

    private void checkBoxSelect_mouseReleased(MouseEvent e)
    {
        _ctrl = false;
    }

    private void spinnerFrame_stateChanged(ChangeEvent e)
    {
        if(!_ignore){
            int frame = ((Integer)_spinnerFrame.getValue()).intValue() - 1;
            this.setFrame(frame);
            _handler.refreshImage();
        }
    }

    private void sliderFrame_stateChanged(ChangeEvent e)
    {
        if(!_ignore){
            int frame = (int)_sliderFrame.getValue() - 1;
            this.setFrame(frame);
            _handler.refreshImage();
        }
    }

    private void spinnerStep_stateChanged(ChangeEvent e)
    {
        if(!_ignore){
            _step = ((Integer)_spinnerStep.getValue()).intValue();
        }
    }
}
