/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import javax.swing.*;
import javax.swing.event.*;

/**
 * Animation control form.
 */
public class AnimationFrame extends JFrame
    implements ActionListener, ItemListener, ChangeListener
{
    private static final long serialVersionUID = 1L;

    private AnimationHandler _handler;
    
    private AnimationFile _empty;

    private JButton _buttonInit;
    private JButton _buttonReverse;
    private JButton _buttonPrev;
    private JButton _buttonStop;
    private JButton _buttonNext;
    private JButton _buttonStart;
    private JButton _buttonTerm;
    
    private JLabel  _labelSpeed;
    private JSlider _sliderSpeed;
    private JLabel  _labelSelect;
    private JLabel  _labelFrame;
    private JLabel  _labelStep;

    private JCheckBox _checkBoxLoop;
    private JCheckBox _checkBoxSwing;
    
    private boolean _ignore;

    /**
     * Constructor.
     */
    public AnimationFrame(AnimationHandler handler)
    {
        _handler = handler;

        _handler.setAnimationFrame(this);

        _empty = new AnimationFile(null, _handler);
        _empty.initializeComponent();

        this.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

        ImageIcon icon = new ImageIcon(getClass().getClassLoader().getResource(GuiConstants.ICON_IMAGE));
        this.setIconImage(icon.getImage());

        this.createComponent();
        this.updateComponent();
        this.updateStatus();
    }

    /**
     * Sets this dialog at the center of the passed component.
     */
    public void setCenterTo(Component component)
    {
        Point ownerLocation = component.getLocation();
        Dimension ownerSize = component.getSize();

        Dimension size = this.getSize();
        this.setLocation(
            ownerLocation.x + (ownerSize.width - size.width)/2, 
            ownerLocation.y + (ownerSize.height - size.height)/2
        );
    }

    //--------------------------------------------------------------------
    // Event handler
    //--------------------------------------------------------------------
    public void actionPerformed(ActionEvent e)
    {
        if(e.getSource() == _buttonInit){
            _handler.firstFrame();
        }
        else if(e.getSource() == _buttonReverse){
            _handler.start(true);
        }
        else if(e.getSource() == _buttonPrev){
            _handler.stepBackward();
        }
        else if(e.getSource() == _buttonStop){
            _handler.stop();
        }
        else if(e.getSource() == _buttonNext){
            _handler.stepForward();
        }
        else if(e.getSource() == _buttonStart){
            _handler.start(false);
        }
        else if(e.getSource() == _buttonTerm){
            _handler.lastFrame();
        }
    }

    public void stateChanged(ChangeEvent e)
    {
        if(_ignore) return;
        
        if(e.getSource() == _sliderSpeed){
            _handler.setSpeed( _sliderSpeed.getValue() );
        }
    }

    public void itemStateChanged(ItemEvent e)
    {
        if(_ignore) return;
        
        if(e.getSource() == _checkBoxLoop){
            if(e.getStateChange() == ItemEvent.SELECTED){
                _checkBoxSwing.setSelected(false);
                _handler.setMode(AnimationFile.LOOP);
            }
            else{
                _handler.setMode(AnimationFile.ONCE);
            }
        }
        else if(e.getSource() == _checkBoxSwing){
            if(e.getStateChange() == ItemEvent.SELECTED){
                _checkBoxLoop.setSelected(false);
                _handler.setMode(AnimationFile.SWING);
            }
            else{
                _handler.setMode(AnimationFile.ONCE);
            }
        }
    }

    //--------------------------------------------------------------------
    // Handle component
    //--------------------------------------------------------------------
    void updateStatus()
    {
        _ignore = true;
        
        _sliderSpeed.setValue(_handler.getSpeed());

        int mode = _handler.getMode();
        if(mode == AnimationFile.ONCE){
            _checkBoxLoop.setSelected(false);
            _checkBoxSwing.setSelected(false);
        }
        else if(mode == AnimationFile.LOOP){
            _checkBoxLoop.setSelected(true);
            _checkBoxSwing.setSelected(false);
        }
        else{
            _checkBoxLoop.setSelected(false);
            _checkBoxSwing.setSelected(true);
        }
        
        _ignore = false;
    }
    
    void updateComponent()
    {
        Container container = this.getContentPane();
        container.removeAll();

        int num = _handler.numFiles() - 1;
        if(num < 0) num = 0;

        this.setSize(496, 232 + 40*num);

        container.add(_labelSelect);
        container.add(_labelFrame);
        container.add(_labelStep);

        int y = 40;
        for(int i=0; i<_handler.numFiles(); i++)
        {
            AnimationFile file = _handler.getFile(i);
            file.initializeComponent();
            
            file.getCheckBoxSelect().setLocation(18, y);
            file.getSpinnerFrame().setLocation(62, y);
            file.getSliderFrame().setLocation(118, y);
            file.getSpinnerStep().setLocation(420, y);

            container.add( file.getCheckBoxSelect() );
            container.add( file.getSpinnerFrame() );
            container.add( file.getSliderFrame() );
            container.add( file.getSpinnerStep() );

            y += 40;            
        }
        if(y == 40){
            _empty.getCheckBoxSelect().setLocation(18, y);
            _empty.getSpinnerFrame().setLocation(62, y);
            _empty.getSliderFrame().setLocation(118, y);
            _empty.getSpinnerStep().setLocation(420, y);

            container.add( _empty.getCheckBoxSelect() );
            container.add( _empty.getSpinnerFrame() );
            container.add( _empty.getSliderFrame() );
            container.add( _empty.getSpinnerStep() );

            y += 40;
        }

        y += 10;
        int x = 100;
        int dx = 42;
        _buttonInit.setLocation(x, y); x += dx;
        _buttonReverse.setLocation(x, y); x += dx;
        _buttonPrev.setLocation(x, y); x += dx;
        _buttonStop.setLocation(x, y); x += dx;
        _buttonNext.setLocation(x, y); x += dx;
        _buttonStart.setLocation(x, y); x += dx;
        _buttonTerm.setLocation(x, y);

        y += 48;
        _labelSpeed.setLocation(115, y);
        _sliderSpeed.setLocation(155, y);
        _checkBoxLoop.setLocation(260, y);
        _checkBoxSwing.setLocation(330, y);

        container.add(_buttonInit);
        container.add(_buttonReverse);
        container.add(_buttonPrev);
        container.add(_buttonStop);
        container.add(_buttonNext);
        container.add(_buttonStart);
        container.add(_buttonTerm);

        container.add(_labelSpeed);
        container.add(_sliderSpeed);
        container.add(_checkBoxLoop);
        container.add(_checkBoxSwing);
        
        if( this.isVisible() ){
            this.repaint();
            this.setVisible(true);
        }
        
        _buttonStop.requestFocus();
    }

    private void createComponent()
    {
        this.setTitle("jV animation");
        this.setResizable(false);

        Container container = this.getContentPane();
        container.setLayout(null);
        
        String images = "org/pdbj/viewer/gui/images/";
        Dimension buttonSize = new Dimension(38, 30);
        
        _buttonInit = new JButton();
        _buttonInit.setSize(buttonSize);
        _buttonInit.addActionListener(this);
        this.setButtonIcon(_buttonInit, images + "button1.bmp", "|<");
        
        _buttonReverse = new JButton();
        _buttonReverse.setSize(buttonSize);
        _buttonReverse.addActionListener(this);
        this.setButtonIcon(_buttonReverse, images + "button2.bmp", "<");
        
        _buttonPrev = new JButton();
        _buttonPrev.setSize(buttonSize);
        _buttonPrev.addActionListener(this);
        this.setButtonIcon(_buttonPrev, images + "button3.bmp", "<|");

        _buttonStop = new JButton();
        _buttonStop.setSize(buttonSize);
        _buttonStop.addActionListener(this);
        this.setButtonIcon(_buttonStop, images + "button4.bmp", "||");

        _buttonNext = new JButton();
        _buttonNext.setSize(buttonSize);
        _buttonNext.addActionListener(this);
        this.setButtonIcon(_buttonNext, images + "button5.bmp", "|>");

        _buttonStart = new JButton();
        _buttonStart.setSize(buttonSize);
        _buttonStart.addActionListener(this);
        this.setButtonIcon(_buttonStart, images + "button6.bmp", ">");

        _buttonTerm = new JButton();
        _buttonTerm.setSize(buttonSize);
        _buttonTerm.addActionListener(this);
        this.setButtonIcon(_buttonTerm, images + "button7.bmp", ">|");
        
        _labelSpeed = new JLabel("Speed");
        _labelSpeed.setSize(40, 20);
        
        _sliderSpeed = new JSlider();
        _sliderSpeed.setSize(80, 20);
        _sliderSpeed.setMinimum(AnimationHandler.MIN_SPEED);
        _sliderSpeed.setMaximum(AnimationHandler.MAX_SPEED);
        _sliderSpeed.addChangeListener(this);
        
        _labelSelect = new JLabel("Select");
        _labelSelect.setSize(40, 20);
        _labelSelect.setLocation(16, 16);
        
        _labelFrame = new JLabel("Frame");
        _labelFrame.setSize(40, 20);
        _labelFrame.setLocation(66, 16);
        
        _labelStep = new JLabel("Step");
        _labelStep.setSize(40, 20);
        _labelStep.setLocation(424, 16);
        
        _checkBoxLoop = new JCheckBox("Loop");
        _checkBoxLoop.setSize(70, 20);
        _checkBoxLoop.addItemListener(this);

        _checkBoxSwing = new JCheckBox("Swing");
        _checkBoxSwing.setSize(70, 20);
        _checkBoxSwing.addItemListener(this);
    }

    private void setButtonIcon(JButton button, String imagePath, String alt)
    {
        URL imageURL = getClass().getClassLoader().getResource(imagePath);
        if(imageURL != null){
            button.setHorizontalAlignment(SwingConstants.CENTER);
            button.setIcon( new ImageIcon(imageURL) );
        }
        else{
            Font font = new Font("DialogInput", Font.PLAIN, 12);
            button.setFont(font);
            button.setMargin( new Insets(0,0,4,0) );
            button.setHorizontalAlignment(SwingConstants.CENTER);
            button.setText(alt);
        }
    }
}
