/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.event.*;
import java.util.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.commands.*;

/**
 * Animation handler.
 */
public class AnimationHandler implements FileStatusListener, AnimationController
{    
    private javax.swing.Timer _timer;

    private ViewerController _viewerController;
    private ArrayList _files;

    private int _mode;
    private int _speed;
    
    private AnimationFrame _frame;

    /**
     * Constructor.
     */
    public AnimationHandler(ViewerController viewerController)
    {
        _viewerController = viewerController;

        _viewerController.addFileStatusListener(this);

        _files = new ArrayList();

        _mode = AnimationFile.ONCE;
        _speed = 10;

        _timer = new javax.swing.Timer(
            this.getDelay(_speed),
            new ActionListener(){
                public void actionPerformed(ActionEvent e){
                    timer_actionPerformed(e);
                }
            }
        );
    }

    /**
     * Returns the number of files.
     */
    public int numFiles()
    {
        return _files.size();
    }

    /**
     * Returns the specified file.
     */
    public AnimationFile getFile(int index)
    {
        return (AnimationFile)_files.get(index);
    }

    /**
     * Sets the mode.
     */
    public void setMode(int mode)
    {
        _mode = mode;

        for(int i=0; i<_files.size(); i++){
            AnimationFile file = (AnimationFile)_files.get(i);
            file.setMode(mode);
        }

        if(_frame != null) _frame.updateStatus();
    }

    /**
     * Returns the mode.
     */
    public int getMode()
    {
        return _mode;
    }

    /**
     * Returns the speed.
     */
    public int getSpeed()
    {
        return _speed;
    }

    /**
     * Sets the current frame to the first one.
     */
    public void firstFrame()
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if( file.isSelected() ){
                file.setRunning(false);
                file.firstFrame();
            }
        }
        _viewerController.refresh();
    }

    /**
     * Steps forward.
     */
    public void stepForward()
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if( file.isSelected() ){
                file.setRunning(false);
                file.stepForward();
            }
        }
        _viewerController.refresh();
    }

    /**
     * Steps backward.
     */
    public void stepBackward()
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if( file.isSelected() ){
                file.setRunning(false);
                file.stepBackward();
            }
        }
        _viewerController.refresh();
    }

    /**
     * Sets the current frame to the last one.
     */
    public void lastFrame()
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if( file.isSelected() ){
                file.setRunning(false);
                file.lastFrame();
            }
        }
        _viewerController.refresh();
    }

    /**
     * Refreshes image.
     */
    public void refreshImage()
    {
        _viewerController.refresh();
    }

    /**
     * Stops all animation.
     */
    public void stopAll()
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            file.setRunning(false);
        }
    }

    void setAnimationFrame(AnimationFrame frame)
    {
        _frame = frame;
    }

    //--------------------------------------------------------------------
    // Implements AnimationController
    //--------------------------------------------------------------------
    /**
     * Starts animation.
     */
    public void start(boolean reverse)
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if( file.isSelected() ){
                file.setReverse(reverse);
                file.setRunning(true);
            }
        }
        if( !_timer.isRunning() ) _timer.start();
    }

    /**
     * Stops animation.
     */
    public void stop()
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if( file.isSelected() ){
                file.setRunning(false);
            }
        }
    }

    /**
     * Moves frmae.
     */
    public void move(int frame, boolean reverse)
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if( file.isSelected() ){
                file.setRunning(false);
                if(reverse) file.stepBackward(frame);
                else        file.stepForward(frame);
            }
        }
        _viewerController.refresh();
    }

    /**
     * Selects files.
     */
    public boolean select(int[] fileIds)
    {
        AnimationFile[] files = this.find(fileIds);
        if(files == null) return false;

        for(int i=0; i<_files.size(); i++){
            AnimationFile file = (AnimationFile)_files.get(i);
            file.setSelected(false);
        }        
        for(int i=0; i<files.length; i++){
            files[i].setSelected(true);
        }
        return true;
    }

    /**
     * Changes selection status.
     */
    public boolean select(boolean enable, int[] fileIds)
    {
        AnimationFile[] files = this.find(fileIds);
        if(files == null) return false;

        for(int i=0; i<files.length; i++){
            files[i].setSelected(enable);
        }
        return true;
    }

    /**
     * Sets 'once' mode.
     */
    public void setOnceMode()
    {
        this.setMode(AnimationFile.ONCE);
    }

    /**
     * Sets 'loop' mode.
     */
    public void setLoopMode()
    {
        this.setMode(AnimationFile.LOOP);
    }

    /**
     * Sets 'swing' mode.
     */
    public void setSwingMode()
    {
        this.setMode(AnimationFile.SWING);
    }

    /**
     * Sets the speed.
     */
    public void setSpeed(int speed)
    {
        if(speed >= MIN_SPEED && speed <= MAX_SPEED)
        {
            _speed = speed;
            _timer.setDelay( this.getDelay(_speed) );
        
            if(_frame != null) _frame.updateStatus();
        }
    }

    /**
     * Sets the step for all files.
     */
    public void setStep(int step)
    {
        for(int i=0; i<_files.size(); i++){
            AnimationFile file = (AnimationFile)_files.get(i);
            file.setStep(step);
        }
    }

    /**
     * Sets the step for the specified file.
     */
    public boolean setStep(int step, int fileId)
    {
        AnimationFile file = this.find(fileId);
        if(file == null) return false;
        
        file.setStep(step);
        return true;
    }

    /**
     * Sets the frame for all files.
     */
    public void setFrame(int frame)
    {
        for(int i=0; i<_files.size(); i++){
            AnimationFile file = (AnimationFile)_files.get(i);
            if(frame > file.numFrames()){
                file.lastFrame();
            }
            else{
                file.setFrame(frame-1);
            }
        }
        _viewerController.refresh();
    }

    /**
     * Sets the frame for the specified file.
     */
    public boolean setFrame(int frame, int fileId)
    {
        AnimationFile file = this.find(fileId);
        if(file == null) return false;

        if(frame > file.numFrames()){
            file.lastFrame();
        }
        else{
            file.setFrame(frame-1);
        }
        _viewerController.refresh();
        return true;
    }

    /**
     * Sets the range for all files.
     */
    public void setRange(int init, int term)
    {
        for(int i=0; i<_files.size(); i++){
            AnimationFile file = (AnimationFile)_files.get(i);
            file.setRange(init-1, term-1);
        }
    }

    /**
     * Sets the range for the specified file.
     */
    public boolean setRange(int init, int term, int fileId)
    {
        AnimationFile file = this.find(fileId);
        if(file == null) return false;

        file.setRange(init-1, term-1);
        return true;
    }

    //--------------------------------------------------------------------
    // Implements FileStatusListener
    //--------------------------------------------------------------------
    /**
     * Receives addition of files.
     */
    public void fileAdded(ArrayList fileStatusList)
    {
        ArrayList moleculeFiles = _viewerController.getMoleculeFiles();
        for(int i=0; i<moleculeFiles.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)moleculeFiles.get(i);
            if(file.isAnimation() && this.find( file.getFileId() ) == null)
            {
                AnimationFile animFile = new AnimationFile(file, this);
                animFile.setSelected(true);
                animFile.setMode(_mode);
                this.add(animFile);
            }
        }

        if(_frame != null) _frame.updateComponent();
    }

    /**
     * Receives remove of files.
     */
    public void fileRemoved(ArrayList fileStatusList)
    {
        ArrayList removedList = new ArrayList();

        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            boolean exist = false;

            ArrayList moleculeFiles = _viewerController.getMoleculeFiles();
            for(int j=0; j<moleculeFiles.size(); j++)
            {
                MoleculeFile moleculeFile = (MoleculeFile)moleculeFiles.get(j);
                if( file.getFileId() == moleculeFile.getFileId() ){
                    exist = true;
                    break;
                }
            }
            if(!exist) removedList.add(file);
        }

        for(int i=0; i<removedList.size(); i++)
        {
            AnimationFile file = (AnimationFile)removedList.get(i);
            this.remove(file);
        }

        if(_frame != null) _frame.updateComponent();
    }

    /**
     * Receives a change of status.
     */
    public void statusChanged(ArrayList fileStatusList)
    {
    }

    //--------------------------------------------------------------------
    // Handle timer
    //--------------------------------------------------------------------
    private void timer_actionPerformed(ActionEvent e)
    {
        boolean running = false;

        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if( file.isRunning() ){
                file.move();
                running = true;
            }
        }

        if(running){
            _viewerController.refresh();
        }
        else{
            _timer.stop();
        }
    }

    private int getDelay(int index)
    {
        if(index < 8){
            return ( 990 - index * 100 );
        }
        else{
            return ( 190 - (index - 8) * 15 );
        }
    }

    //--------------------------------------------------------------------
    // Handle file
    //--------------------------------------------------------------------
    private void add(AnimationFile file)
    {
        _files.add(file);
    }

    private void remove(AnimationFile file)
    {
        _files.remove(file);
    }

    private AnimationFile find(int fileId)
    {
        for(int i=0; i<_files.size(); i++)
        {
            AnimationFile file = (AnimationFile)_files.get(i);
            if(file.getFileId() == fileId) return file;
        }
        return null;
    }

    private AnimationFile[] find(int[] fileIds)
    {
        AnimationFile[] files = new AnimationFile[fileIds.length];

        for(int i=0; i<fileIds.length; i++)
        {
            AnimationFile file = this.find(fileIds[i]);
            if(file == null) return null;
            
            files[i] = file;
        }
        return files;
    }
}
