/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.ArrayList;
import java.util.StringTokenizer;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.commands.*;

/**
 * Command area.
 */
public class CommandArea extends JScrollPane implements MessageArea
{
    private static final long serialVersionUID = 1L;

    //protected static final String PROMPT = "jV> ";
    protected static final String PROMPT = "> ";

    protected static final int NUMBER_OF_ROWS = 6;
    protected static final int MAX_HISTORY    = 100;

    private CommandHandler _handler;
    private GuiWrapper _guiWrapper;

    private boolean _active;
    private boolean _editable;
    private boolean _prompting;

    private ArrayList _history;

    private int _currStart;
    private int _index;

    private CommandComplement _complement;
    private int _complementPosition;

    private JTextArea _textArea;
    
    /**
    * Sets the command handler.
    */
    public void setCommandHandler(CommandHandler handler)
    {
        _handler = handler;
    }

    /**
    * Returns the command handler.
    */
    public CommandHandler getCommandHandler()
    {
        return _handler;
    }

    /**
    * Sets whether it is editable.
    * The method works only when this is active.
    */
    public void setEditable(boolean editable)
    {
        if(_active){
            _editable = editable;
        }
    }

    /**
    * Tests whether it is editable.
    */
    public boolean isEditable()
    {
        return _editable;
    }

    /**
    * Sets whether active.
    */
    public void setActive(boolean active)
    {
        _active = active;
        
        if(_active){
            _editable = true;
            if(!_prompting) this.showPrompt();
        }
        else{
            _editable = false;
            if(_prompting){
                this.newLine();
                _prompting = false;
            }
        }
    }

    /**
    * Returns whether active.
    */
    public boolean isActive()
    {
        return _active;
    }

    /**
    * Overrides requestFocus().
    */
    public void requestFocus()
    {
        _textArea.requestFocus();
    }

    /**
    * Constructor.
    */
    public CommandArea(GuiWrapper guiWrapper)
    {
        _guiWrapper = guiWrapper;

        _history = new ArrayList();

        this.createComponent();
    }

    /**
     * Appends the message.
     */
    public void appendMessage(String message)
    {
        boolean origPrompting = _prompting;

        if(_prompting){
            this.newLine();
            _prompting = false;
        }

        _textArea.append(message);
        this.newLine();

        if(origPrompting && _editable){
            this.showPrompt();
        }
    }

    //--------------------------------------------------------------------
    // Create Component
    //--------------------------------------------------------------------
    private void createComponent()
    {
        _textArea = new JTextArea();

        _textArea.setLineWrap(true);
        _textArea.setRows(NUMBER_OF_ROWS);
        //_textArea.setFont( new Font("DialogInput", Font.PLAIN, 12) );
        _textArea.setFont( new Font("Monospaced", Font.PLAIN, 12) );

        this.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        this.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

        this.getViewport().add(_textArea);

        // add KeyListener
        _textArea.addKeyListener(
            new KeyAdapter(){
                public void keyTyped(KeyEvent e){
                    handle_keyTyped(e);
                }
                public void keyPressed(KeyEvent e){
                    handle_keyPressed(e);
                }
            }
        );

        // add MouseListener
        _textArea.addMouseListener(
            new MouseAdapter(){
                public void mouseClicked(MouseEvent e){
                    handle_mouseClicked(e);
                }
                public void mousePressed(MouseEvent e){
                    handle_mousePressed(e);
                }
            }
        );
        
        // add MouseMotionListener
        _textArea.addMouseMotionListener(
            new MouseMotionAdapter(){
                public void mouseDragged(MouseEvent e){
                    handle_mouseDragged(e);
                }
            }
        );

        // add ComponentListener
        _textArea.addComponentListener(
            new ComponentAdapter(){
                public void componentResized(ComponentEvent e){
                    handle_componentResized(e);
                }
            }
        );
    }

    //--------------------------------------------------------------------
    // Event handler
    //--------------------------------------------------------------------
    private void handle_mouseDragged(MouseEvent e)
    {
        this.fixCaretPosition();
    }

    private void handle_mouseClicked(MouseEvent e)
    {
        this.fixCaretPosition();
    }

    private void handle_mousePressed(MouseEvent e)
    {
        this.fixCaretPosition();
    }

    private void handle_keyPressed(KeyEvent e)
    {
        if(e.getKeyCode() != KeyEvent.VK_TAB){
            _complement = null;
        }
        
        if(!_editable){
            e.consume();
        }
        else if(e.getKeyCode() == KeyEvent.VK_UP){
            this.showPrevious();
            e.consume();
        }
        else if(e.getKeyCode() == KeyEvent.VK_DOWN){
            this.showNext();
            e.consume();
        }
        else if(e.getKeyCode() == KeyEvent.VK_LEFT){
            if(_textArea.getCaretPosition() == _currStart){
                e.consume();
            }
        }
        else if(e.getKeyCode() == KeyEvent.VK_TAB){
            e.consume();
            this.complement();
        }
        else if(e.getKeyCode() == KeyEvent.VK_BACK_SPACE){
            if(_textArea.getCaretPosition() == _currStart){
                e.consume();
            }
        }
        else if(e.getKeyCode() == KeyEvent.VK_HOME){
            e.consume();
        }
        else if( (e.getModifiersEx() & InputEvent.CTRL_DOWN_MASK) == InputEvent.CTRL_DOWN_MASK )
        {
            if(e.getKeyCode() == KeyEvent.VK_P){
                this.showPrevious();
            }
            else if(e.getKeyCode() == KeyEvent.VK_N){
                this.showNext();
            }
            else if(e.getKeyCode() == KeyEvent.VK_F){
                int position = _textArea.getCaretPosition();
                if(position != _textArea.getText().length()){
                    _textArea.setCaretPosition(position+1);
                }
            }
            else if(e.getKeyCode() == KeyEvent.VK_B){
                int position = _textArea.getCaretPosition();
                if(position != _currStart){
                    _textArea.setCaretPosition(position-1);
                }
            }
        }
    }

    private void handle_keyTyped(KeyEvent e)
    {
        int end = _textArea.getSelectionEnd();
        if( _textArea.getSelectionStart() != end ){
            _textArea.setSelectionStart(end);
        }

        if(!_editable){
            e.consume();
            _guiWrapper.commandKeyTyped();
        }
        else if(e.getKeyChar() == KeyEvent.VK_ENTER){
            e.consume();
            this.execute();
        }
        else if(e.getKeyChar() == KeyEvent.VK_BACK_SPACE){
            if(_textArea.getCaretPosition() == _currStart){
                e.consume();
            }
        }
    }

    private void handle_componentResized(ComponentEvent e)
    {
        int caretPosition = _textArea.getCaretPosition();

        // enforce refresh
        _textArea.setText( _textArea.getText() );
        _textArea.setCaretPosition(caretPosition);
        _textArea.invalidate();
    }
    
    //--------------------------------------------------------------------
    // Execute
    //--------------------------------------------------------------------
    private void execute()
    {
        String command = this.extractCommand();

        this.showCommand(command);
        this.setActive(false);

        if( !command.equals("") ){
            if(_handler != null){
                _handler.executeSingleCommand(command);
            }
            this.addHistory(command);
        }
        else this.setActive(true);
    }

    private String extractCommand()
    {
        String tmp = _textArea.getText().substring(_currStart);

        StringTokenizer st = new StringTokenizer(tmp, "\n\r");
        StringBuffer sb = new StringBuffer();
        while(st.hasMoreTokens()){
            sb.append(st.nextToken());
        }
        return sb.toString().trim();
    }

    //--------------------------------------------------------------------
    // Complement
    //--------------------------------------------------------------------
    private void complement()
    {
        String wholeCommand = _textArea.getText().substring(_currStart);
        int caret = _textArea.getCaretPosition();
        int length = caret - _currStart;

        String command = wholeCommand.substring(0, length);
        String rest = wholeCommand.substring(length);

        boolean first = false;
        if( _complement == null || caret != _complementPosition)
        {
            _complement = _handler.complementCommand(command);
            first = true;
            _complementPosition = caret;
        }

        if(_complement.getPossibles().size() == 0){
            return;
        }
        else{
            String[] possibles = (String[])_complement.getPossibles().toArray( new String[0] );
            String common = this.getCommonString(possibles);
            int fix = _complement.getFixedPosition();
            if(common != null && fix + common.length() > length){
                String expr = command.substring(0, fix) + common;
                this.cutoffText(_currStart);
                _textArea.append(expr);
                _textArea.append(rest);
                _textArea.setCaretPosition(_currStart + expr.length());                
            }
            else if(!first){
                this.newLine();
                for(int i=0; i<possibles.length; i++){
                    _textArea.append(possibles[i] + "  ");
                }
                this.newLine();
                this.showPrompt();
                int curr = _currStart;
                _textArea.append(wholeCommand);
                _textArea.setCaretPosition(curr + length);
                _complementPosition = curr + length;
            }
        }
    }
    
    private String getCommonString(String[] possibles)
    {
        String s = possibles[0];
        int num = s.length();
        for(int i=1; num!= 0 && i<possibles.length; i++)
        {
            int k = 0;
            int max = possibles[i].length();
            while(k < num && k < max && s.charAt(k) == possibles[i].charAt(k)) k++;
            num = k;
        }
        
        return (num == 0) ? null : s.substring(0, num);
    }

    //--------------------------------------------------------------------
    // Handle text
    //--------------------------------------------------------------------
    private void newLine()
    {
        _textArea.append("\n");
    }

    private void showPrompt()
    {
        _textArea.append(PROMPT);

        _currStart = _textArea.getText().length();
        _prompting = true;

        _textArea.setCaretPosition(_currStart);
    }

    private void showCommand(String command)
    {
        this.cutoffText(_currStart);
        _textArea.append(command);
    }

    private void cutoffText(int position)
    {
        _textArea.replaceRange("", position, _textArea.getText().length());
    }

    private void fixCaretPosition()
    {
        if( _textArea.getCaretPosition() < _currStart )
        {
            _textArea.setCaretPosition( _textArea.getText().length() );
        }
    }

    //--------------------------------------------------------------------
    // Command history
    //--------------------------------------------------------------------
    protected void addHistory(String command)
    {
        if(_history.size() == MAX_HISTORY){
            _history.remove(0);
        }

        _history.add(command);
        _index = _history.size();
    }

    protected void showPrevious()
    {
        if(_index > 0){
            _index--;
            this.showCommand( (String)_history.get(_index) );
        }
    }

    protected void showNext()
    {
        if(_index != _history.size()){
            _index++;
            if(_index < _history.size()){
                this.showCommand( (String)_history.get(_index) );
            }
            else{
                this.showCommand("");
            }
        }
    }
}
