/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.util.*;

/**
 * Command queue.
 */
public class CommandQueue
{
    private static final CommandType _default = new CommandType(CommandType.DEFAULT);
    
    private boolean _active;

    private LinkedList _queue;
    private LinkedList _queueType;

    private ScriptController _scriptController;
    private JvCommandController _jvCommandController;

    /**
     * Constructor.
     */
    public CommandQueue()
    {
        _queue = new LinkedList();
        _queueType = new LinkedList();
    }
    
    /**
     * Sets the script controller.
     */
    public void setScriptController(ScriptController scriptController) {
        _scriptController = scriptController;
    }

    /**
     * Sets the jV command controller.
     */
    public void setJvCommandController(JvCommandController jvCommandController) {
        _jvCommandController = jvCommandController;
    }
    
    /**
     * Sets whether active.
     */
    public synchronized void setActive(boolean enable)
    {
        _active = enable;
        if(_active){
            this.notifyAll();
        }
    }

    /**
     * Retrieves the first command in the queue.
     */
    public synchronized void getCommand(CommandHolder holder) throws InterruptedException
    {
        while(!_active || _queue.size() <= 0)
        {
            this.wait();
        }

        _active = false;

        holder.command = (String)_queue.removeFirst();
        holder.type = (CommandType)_queueType.removeFirst();
    }

    /**
     * Appends a command with the default type in the queue.
     */
    public synchronized void putCommand(String command)
    {
        _queue.addLast(command);
        _queueType.addLast(_default);
        
        this.notifyAll();
    }

    /**
     * Appends a command in the queue.
     */
    public synchronized void putCommand(String command, CommandType type)
    {
        _queue.addLast(command);
        _queueType.addLast(type);
        
        this.notifyAll();
    }

    /**
     * Moves to the next script.
     */
    public synchronized void nextScript()
    {
        if( _queue.size() == 0 && _scriptController.hasMoreCommand() )
        {
            _queue.addLast( _scriptController.nextCommand() );
            _queueType.addLast(_default);

            this.notifyAll();
        }
    }

    /**
     * Moves to the next jV command.
     */
    public synchronized void nextJvCommand()
    {
        if( _queue.size() == 0 && _jvCommandController.hasNext() )
        {
            _queue.addLast( _jvCommandController.next() );
            _queueType.addLast(_default);

            this.notifyAll();
        }
    }
}

