/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import org.pdbj.viewer.controller.*;

/**
 * File status control.
 */
class FileStatusControl extends JPanel
{
    private static final long serialVersionUID = 1L;

    static final int HEIGHT  = 25;

    private FileStatusPanel _parent;
    
    private int _fileId;
    private int _fileType;

    JCheckBox checkSelectFile;
    JCheckBox checkDisplay;
    JCheckBox checkSelectComponent;
    JLabel labelName;

    private boolean _handleEvent;

    /**
     * Returns the file ID.
     */
    int getFileId()
    {
        return _fileId;
    }
    
    /**
     * Constructor.
     */
    FileStatusControl(FileStatusPanel parent, FileStatus status)
    {
        _parent  = parent;
        _fileId = status.getId();
        _fileType = status.getType();

        createComponent(status);
        updateStatus(status);

        _handleEvent = true;
    }

    /**
     * Updates the file status.
     */
    void updateStatus(FileStatus status)
    {
        _handleEvent = false;
        checkSelectFile.setSelected(status.isSelected());
        checkDisplay.setSelected(status.isDisplayed());
        checkSelectComponent.setSelected(status.isComponentSelected());
        _handleEvent = true;
    }

    //--------------------------------------------------------------------
    // Create component
    //--------------------------------------------------------------------
    private void createComponent(FileStatus status)
    {
        checkSelectFile = new JCheckBox();
        checkSelectFile.addItemListener(
            new ItemListener(){
                public void itemStateChanged(ItemEvent e){
                    checkBox_itemStateChanged(e);
                }
            }
        );

        checkDisplay = new JCheckBox();
        checkDisplay.addItemListener(
            new ItemListener(){
                public void itemStateChanged(ItemEvent e){
                    checkBox_itemStateChanged(e);
                }
            }
        );
        checkSelectComponent = new JCheckBox();
        checkSelectComponent.addItemListener(
            new ItemListener(){
                public void itemStateChanged(ItemEvent e){
                    checkBox_itemStateChanged(e);
                }
            }
        );

        labelName = new JLabel();
        labelName.setFont( new Font("SansSerif", Font.PLAIN, 12) );
        labelName.setText( _fileId + ".  " + getName(status) );

        this.setLayout( new BoxLayout(this, BoxLayout.X_AXIS) );

        int w1 = getWidthDiff(_parent._labelSelectFile, checkSelectFile);
        int w2 = getWidthDiff(_parent._labelDisplay, checkDisplay);
        int w3 = getWidthDiff(_parent._labelSelectComponent, checkSelectComponent);

        this.add( Box.createRigidArea( new Dimension(getInterval(5, 0, w1), HEIGHT) ) );
        this.add( checkSelectFile );
        this.add( Box.createRigidArea( new Dimension(getInterval(6, w1, w2), HEIGHT) ) );
        this.add( checkDisplay );
        this.add( Box.createRigidArea( new Dimension(getInterval(8, w2, w3), HEIGHT) ) );
        this.add( checkSelectComponent );
        this.add( Box.createRigidArea( new Dimension(4, HEIGHT) ) );
        this.add( labelName );
        this.add( Box.createRigidArea( new Dimension(6, HEIGHT) ) );
        this.add( Box.createHorizontalGlue() );
        
        checkSelectFile.setFocusable(false);
        checkDisplay.setFocusable(false);
        checkSelectComponent.setFocusable(false);
    }
    
    private String getName(FileStatus status) {
        String path = status.getPath();
        int index = path.lastIndexOf("\\");
        if(index == -1) index = path.lastIndexOf("/");
        return path.substring(index + 1);
    }

    private int getWidthDiff(Component c1, Component c2) {
        return c1.getPreferredSize().width - c2.getPreferredSize().width;
    }

    private int getInterval(int d, int w1, int w2) {
        int v = d + w1 / 2 + w2 / 2;
        return v > 0 ? v : 0;
    }

    //--------------------------------------------------------------------
    // Event handler
    //--------------------------------------------------------------------
    private void checkBox_itemStateChanged(ItemEvent e)
    {
        if(!_handleEvent) return;

        if(_parent.getViewerController() != null)
        {
            JCheckBox checkBox = (JCheckBox)e.getSource();

            if( _parent.isOperatable() ){
                if(checkBox == checkSelectFile){
                    _parent.getViewerController().selectFile(
                        checkBox.isSelected(), new int[]{_fileId}
                    );
                }
                else if (checkBox == checkDisplay){
                    _parent.getViewerController().displayFile(
                        checkBox.isSelected(),  new int[]{_fileId}
                    );
                }
                else if (checkBox == checkSelectComponent){
                    String command = buildCommand(checkBox.isSelected());
                    _parent.getCommandHandler().executeSingleCommand(command);
                }
            }
            else{
                _handleEvent = false;
                checkBox.setSelected( !checkBox.isSelected() );
                _handleEvent = true;
            }
        }

        if(_parent.getGuiWrapper() != null){
            _parent.getGuiWrapper().focusMoved();
        }
    }

    private String buildCommand(boolean enable) {
        String command = _fileType == FileStatus.POLYGON ? "selectvertex" : "select";
        if (enable) {
            return command + " selected || file=" + _fileId;
        } else {
            return command + " selected && file!=" + _fileId;
        }
    }
}
