/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.util.ArrayList;
import javax.swing.*;
import javax.swing.border.*;
import org.pdbj.viewer.controller.*;

/**
 * File status panel.
 */
public class FileStatusPanel extends JScrollPane implements FileStatusListener
{
    private static final long serialVersionUID = 1L;

    private ViewerController _controller;
    private GuiWrapper _guiWrapper;
    private CommandHandler _commandHandler;

    private boolean _operatable;

    private JPanel _panel;

    private JPanel _panelNoFiles;
    private JPanel _panelFiles;

    private JLabel _labelNoFiles;
    JLabel _labelSelectFile;
    JLabel _labelDisplay;
    JLabel _labelSelectComponent;

    private ArrayList _controls;
    
    /**
     * Sets this size.
     */
    public void setSize(int width, int height)
    {
        super.setSize(width, height);
        _panel.setSize(width, height);
    }

    /**
     * Sets this size.
     */
    public void setSize(Dimension size)
    {
        super.setSize(size);
        _panel.setSize(size);
    }

    /**
     * Returns the viewer controller.
     */
    public ViewerController getViewerController()
    {
        return _controller;
    }

    /**
     * Sets the GUI wrapper.
     */
    public void setGuiWrapper(GuiWrapper guiWrapper)
    {
        _guiWrapper = guiWrapper;
    }

    /**
     * Returns the GUI wrapper.
     */
    public GuiWrapper getGuiWrapper()
    {
        return _guiWrapper;
    }

    /**
     * Sets the command handler.
     */
    public void setCommandHandler(CommandHandler commandHandler)
    {
        _commandHandler = commandHandler;
    }

    /**
     * Returns the command handler.
     */
    public CommandHandler getCommandHandler()
    {
        return _commandHandler;
    }

    /**
     * Sets whether operatable.
     */
    public void setOperatable(boolean operatable)
    {
        _operatable = operatable;
    }

    /**
     * Returns whether operatable.
     */
    public boolean isOperatable()
    {
        return _operatable;
    }

    /**
     * Constructor.
     */
    public FileStatusPanel(ViewerController controller)
    {
        _controller = controller;
        _controller.addFileStatusListener(this);

        _controls = new ArrayList();

        this.createComponent();
        this.updateFile(null);

        _operatable = true;
    }

    /**
     * Refresh.
     */
    public void refresh()
    {
        _panel.repaint();
        this.setViewportView(_panel);
    }

    //--------------------------------------------------------------------
    // Implements FileStatusListener
    //--------------------------------------------------------------------
    /**
     * Receives addition of files.
     */
    public void fileAdded(ArrayList fileStatusList)
    {
        this.updateFile(fileStatusList);
    }

    /**
     * Receives remove of files.
     */
    public void fileRemoved(ArrayList fileStatusList)
    {
        this.updateFile(fileStatusList);
    }

    /**
     * Receives a change of status.
     */
    public void statusChanged(ArrayList fileStatusList)
    {
        this.updateStatus(fileStatusList);
    }

    //--------------------------------------------------------------------
    // Create component
    //--------------------------------------------------------------------
    private void createComponent()
    {        
        _panel = new JPanel();
        _panel.setLayout( new BoxLayout(_panel, BoxLayout.Y_AXIS) );
        
        _labelNoFiles = new JLabel();
        _labelNoFiles.setFont( new Font("SansSerif", Font.PLAIN, 12) );
        _labelNoFiles.setText("No Files");

        _labelSelectFile = new JLabel();
        _labelSelectFile.setFont( new Font("Serif", Font.PLAIN, 11) );
        _labelSelectFile.setText("Mouse");

        _labelDisplay = new JLabel();
        _labelDisplay.setFont( new Font("Serif", Font.PLAIN, 11) );
        _labelDisplay.setText("Disp");

        _labelSelectComponent = new JLabel();
        _labelSelectComponent.setFont( new Font("Serif", Font.PLAIN, 11) );
        _labelSelectComponent.setText("Sel");

        _panelNoFiles = new JPanel();
        _panelNoFiles.setLayout( new BoxLayout(_panelNoFiles, BoxLayout.X_AXIS) );
        _panelNoFiles.add( Box.createRigidArea( new Dimension(10, 25) ) );
        _panelNoFiles.add(_labelNoFiles);
        _panelNoFiles.add( Box.createHorizontalGlue() );

        _panelFiles = new JPanel();
        _panelFiles.setLayout( new BoxLayout(_panelFiles, BoxLayout.X_AXIS) );
        _panelFiles.add( Box.createRigidArea( new Dimension(5, 25) ) );
        _panelFiles.add(_labelSelectFile);
        _panelFiles.add( Box.createRigidArea( new Dimension(6, 25) ) );
        _panelFiles.add(_labelDisplay);
        _panelFiles.add( Box.createRigidArea( new Dimension(8, 25) ) );
        _panelFiles.add(_labelSelectComponent);
        _panelFiles.add( Box.createHorizontalGlue() );

        this.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        this.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);

        this.setViewportView(_panel);
        //this.setBorder( BorderFactory.createEmptyBorder() );
        this.setBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));
    }

    //--------------------------------------------------------------------
    // Update
    //--------------------------------------------------------------------
    private void updateFile(ArrayList statusList)
    {
        _panel.removeAll();

        if(statusList == null || statusList.size() == 0){
            _panel.add(_panelNoFiles);
        }
        else{
            _panel.add(_panelFiles);
            _controls.clear();
            for (int i = 0; i < statusList.size(); i++) {
                FileStatus status = (FileStatus)statusList.get(i);
                FileStatusControl control = new FileStatusControl(this, status);
                _controls.add(control);
                _panel.add(control);
            }
        }
        
        // enforce refresh
        this.refresh();
    }

    private void updateStatus(ArrayList statusList)
    {
        for(int i=0; i<statusList.size(); i++)
        {
            FileStatus status = (FileStatus)statusList.get(i);

            FileStatusControl control = (FileStatusControl)_controls.get(i);
            control.updateStatus(status);
        }
    }
}
