/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.io.File;
import java.io.FilenameFilter;
import javax.swing.filechooser.FileFilter;

/**
 * Represents file type.
 */
public class FileType extends FileFilter implements FilenameFilter
{
    public static final FileType PDBML       = new FileType(0);
    public static final FileType PDB         = new FileType(1);
    public static final FileType POLYGON     = new FileType(2);
    public static final FileType SCRIPT      = new FileType(3);
    public static final FileType PNG         = new FileType(4);
    public static final FileType JPEG        = new FileType(5);
    public static final FileType ANIMATION   = new FileType(6);
    public static final FileType FTP_PDBML   = new FileType(7);
    public static final FileType FUNCTION    = new FileType(8);
    public static final FileType SITELIST    = new FileType(9);
    public static final FileType POLYGON_CGO = new FileType(10);

    private static final String[][] EXTENSIONS =
        {
            {"xml", "gz"},          // PDBML
            {"pdb", "ent", "gz"},   // PDB
            {},                     // POLYGON
            {},                     // SCRIPT
            {"png"},                // PNG
            {"jpg", "jpeg"},        // JPEG
            {"pdb", "ent", "gz"},   // ANIMATION
            {},                     // FTP_PDBML
            {"xml", "gz"},          // FUNCTION
            {},                     // SITELIST
            {}                      // POLYGON_CGO
        };

    private static final String[] DISPLAY_NAME =
    {
        "PDBML",
        "PDB",
        "Polygon",
        "Script",
        "PNG",
        "JPEG",
        "Animation",
        "",
        "",
        "",
        "Polygon CGO"
    };

    private int _type;

    /**
     * Returns whether or not to accept any files.
     */
    public boolean acceptAll()
    {
        return (EXTENSIONS[_type].length == 0);
    }

    /**
     * Returns file extensions.
     */
    public String[] getExtensions()
    {
        return EXTENSIONS[_type];
    }

    /**
     * Returns the display name.
     */
    public String getDisplayName()
    {
        return DISPLAY_NAME[_type];
    }

    /**
     * Fixes the file extension.
     */
    public File fixExtension(File file)
    {
        if( this.acceptAll() ) return file;

        String name = file.getName();
        int index = name.lastIndexOf(".");
        String ext = "";
        if(index != -1){
            ext = name.substring(index+1).toLowerCase();
        }
        
        String[] exts = this.getExtensions();
        boolean toAdd = true;
        for(int i=0; i<exts.length; i++)
        {
            if(exts[i].equals("gz")) continue;
            if(exts[i].equals(ext)) toAdd = false;
        }
        
        if(toAdd){
            name += "." + exts[0];
            return new File(file.getParentFile(), name);
        }
        else{
            return file;
        }
    }
    
    //--------------------------------------------------------------------
    // Implements FileFilter
    //--------------------------------------------------------------------
    /**
     * Returns a description.
     */
    public String getDescription()
    {
        if( EXTENSIONS[_type].length == 0 ){
            return "all files (*.*)";
        }
        else{
            StringBuffer sb = new StringBuffer();
            for(int i=0; i<EXTENSIONS[_type].length; i++){
                sb.append("*." + EXTENSIONS[_type][i] + ", ");
            }
            return sb.substring(0, sb.length()-2);
        }
    }

    /**
     * Returns whether the file is accepted or not.
     */
    public boolean accept(File file)
    {
        if(file.isDirectory()) return true;

        return acceptFile(file.getName());
    }

    //--------------------------------------------------------------------
    // Implements FilenameFilter
    //--------------------------------------------------------------------
    /**
     * Returns whether the file is accepted or not.
     */
    public boolean accept(File dir, String name)
    {
        return acceptFile(name);
    }

    private boolean acceptFile(String fileName) {
        if(EXTENSIONS[_type].length == 0){
            return true;
        }
        else{
            int index = fileName.lastIndexOf(".");
            if(index != -1){
                String ext = fileName.substring(index+1).toLowerCase();
                for(int i=0; i<EXTENSIONS[_type].length; i++){
                    if(ext.equals(EXTENSIONS[_type][i])) return true;
                }
            }
            return false;
        }        
    }

    //--------------------------------------------------------------------
    // private constructor
    //--------------------------------------------------------------------
    private FileType(int type)
    {
        _type = type;
    }
}
