/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.util.*;
import org.pdbj.viewer.commands.*;
import org.pdbj.viewer.plus.*;

/**
 * jV command controller.
 */
public class JvCommandController {

    private Function _function;
    private int _fileId;

    private List _jvCommands;
    private int _depth;

    /**
     * Constructor.
     */
    public JvCommandController() {
        _jvCommands = new ArrayList();
        _depth = -1;
    }

    /**
     * Clear.
     */
    public void clear() {
        _function = null;
        _fileId = 0;
        _jvCommands.clear();
        _depth = -1;
    }

    /**
     * Start.
     */
    public void start(Function function, String name, int fileId) {
        _function = function;
        _fileId = fileId;

        _function.resetCommands();
        _jvCommands.clear();
        _jvCommands.add(_function.getJvCommand(name));
        _depth = 0;
    }

    /**
     * Returns whether this has next command.
     */
    public boolean hasNext() {
        return _depth >= 0;
    }

    /**
     * Returns the next command.
     */
    public String next() {
        JvCommand jvCommand = (JvCommand) _jvCommands.get(_depth);
        String command = nextCommand(jvCommand);
        prepareNext();
        return processCommand(command);
    }

    private String nextCommand(JvCommand jvCommand) {
        String command = jvCommand.next();
        if (isJvCommand(command)) {
            String name = getCommandArgs(command);
            if (name == null) {
                return "echo \"" + CommandBehavior.INVALID_COMMAND + "\"";
            }
            JvCommand nestedJvCommand = _function.getJvCommand(name);
            if (nestedJvCommand == null) {
                return "echo \"not found jV_command '" + name + "'.\"";
            } else if (_jvCommands.contains(nestedJvCommand)) {
                return "echo \"ignore circular call of jV_command '" + name + "'.\"";
            } else {
                _jvCommands.add(nestedJvCommand);
                _depth++;
                return nextCommand(nestedJvCommand);
            }
        } else {
            return command;
        }
    }

    private void prepareNext() {
        boolean hasNext = false;
        while (!hasNext && _depth >= 0) {
            JvCommand jvCommand = (JvCommand) _jvCommands.get(_depth);
            if (jvCommand.hasNext()) {
                hasNext = true;
            } else {
                _jvCommands.remove(_depth);
                _depth--;
            }
        }
    }

    private String processCommand(String command) {
        String commandName = getCommandName(command);
        if (isInvalid(commandName)) {
            return "echo \"ignore '" + commandName + "' command.\"";
        } else if (isSelectCommand(commandName)) {
            String args = getCommandArgs(command);
            if (args == null) {
                return commandName + " file=" + _fileId;
            } else if (args.equalsIgnoreCase("all")){
                return commandName + " file=" + _fileId;
            } else if (args.equalsIgnoreCase("none")){
                return command;
            } else {
                return commandName + " (" + args + ") & file=" + _fileId;
            }
        } else if (isDescribeCommand(commandName)) {
            String args = getCommandArgs(command);
            if (args != null) {
                return commandName + " " + _fileId + " " + args;
            } else {
                return commandName + " " + _fileId;
            }
        } else {
            return command;
        }
    }

    private boolean isJvCommand(String command) {
        return commandEquals(getCommandName(command), CommandList.PDBJ_EXECUTE);
    }

    private boolean isDescribeCommand(String commandName) {
        return commandEquals(commandName, CommandList.PDBJ_DESCRIBE);
    }

    private boolean isSelectCommand(String commandName) {
        return commandEquals(commandName, CommandList.SELECT);
    }

    private boolean isInvalid(String commandName) {
        if (commandEquals(commandName, CommandList.EXIT)) {
            return true;            
        }
        if (commandEquals(commandName, CommandList.QUIT)) {
            return true;            
        }
        if (commandEquals(commandName, CommandList.LOAD)) {
            return true;            
        }
        if (commandEquals(commandName, CommandList.SAVE)) {
            return true;            
        }
        if (commandEquals(commandName, CommandList.WRITE)) {
            return true;            
        }
        if (commandEquals(commandName, CommandList.SCRIPT)) {
            return true;            
        }
        if (commandEquals(commandName, CommandList.ZAP)) {
            return true;            
        }
        return false;
    }

    private boolean commandEquals(String commandName, String[] commands) {
        for (int i = 0; i < commands.length; i++) {
            if (commandName.equalsIgnoreCase(commands[i])) {
                return true;
            }
        }
        return false;
    }

    private String getCommandName(String command) {
        return command.split("\\s+")[0];
    }

    private String getCommandArgs(String command) {
        String[] tokens = command.split("\\s+");
        if (tokens.length == 1) {
            return null;
        }
        String args = command.substring(tokens[0].length() + 1).trim();
        return args.length() > 0 ? args : null;
    }
}
