/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.util.*;
import javax.swing.*;
import org.pdbj.viewer.utils.*;

/**
 * Message dialog
 */
public class MessageDialog extends JDialog implements MessageListener
{
    private static final long serialVersionUID = 1L;

    private boolean _active;

    private int _num;
    private int _width;
    
    private Component _owner;

    private ArrayList _labels;

    private int _minHeight;

    private static final int MIN_HEIGHT = 44;
    private static final int PITCH = 24;

    /**
     * Constructor.
     */
    public MessageDialog(Component owner, String title)
    {
        this(owner, title, 0);
    }

    /**
     * Constructor.
     */
    public MessageDialog(Component owner, String title, int addHeight)
    {
        _owner = owner;
        _width = 400;
        _minHeight = MIN_HEIGHT + addHeight;

        this.createComponent();
        this.setTitle(title);
        this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        
        _labels = new ArrayList();
    }

    /**
     * Sets the width.
     */
    public void setWidth(int width)
    {
        _width = width;
        
        // resize
        this.setSize(_width, this.getHeight());
    }

    /**
     * Returns whethrer this is active.
     */
    public boolean isActive()
    {
        return _active;
    }

    /**
     * Begins message.
     */
    public void begin(int numLines)
    {
        if(!_active)
        {
            Container container = getContentPane();
            container.removeAll();
            _labels.clear();
            
            _num = numLines;

            this.setSize(_width, _minHeight + PITCH * _num);
            this.adjustLocation();
            this.setVisible(true);

            _active = true;
        }
    }

    /**
     * Appends a message.
     */
    public void appendMessage(String message)
    {
        int cnt = _labels.size();
        
        Font font = new Font("DialogInput", Font.PLAIN, 12);
        
        JLabel label = new JLabel(message);
        label.setFont(font);
        label.setLocation(24, 6 + PITCH * cnt);
        label.setSize( label.getPreferredSize() );

        _labels.add(label);
        cnt++;

        if(cnt > _num){
            this.setSize(_width, _minHeight + PITCH * cnt);
            _num = cnt;
        }

        Container container = getContentPane();
        container.add(label);

        this.refresh();        
    }

    /**
     * Clears message.
     */
    public void clear()
    {
        Container container = getContentPane();
        container.removeAll();
        _labels.clear();        

        this.refresh();        
    }

    /**
     * Sets the message.
     */
    public void setMessage(String message, int index)
    {
        JLabel label = (JLabel)_labels.get(index);
        label.setText(message);
        label.setSize( label.getPreferredSize() );
        label.invalidate();
        this.refresh();
    }

    /**
     * Ends message.
     */
    public void end()
    {
        this.setVisible(false);
        _active = false;

        // handle GUI menu trouble on Mac
        if (_owner instanceof java.awt.Window) {
            _owner.setVisible(true);
        }
    }

    //--------------------------------------------------------------------
    // Refresh
    //--------------------------------------------------------------------
    private void refresh()
    {
        this.repaint();
        // enforce repaint
        if( this.isVisible() ) this.setVisible(true);
    }

    //--------------------------------------------------------------------
    // Create component
    //--------------------------------------------------------------------
    private void createComponent()
    {
        this.setResizable(false);

        Container container = getContentPane();
        container.setLayout(null);
    }

    private void adjustLocation()
    {
        //Point ownerLocation = _owner.getLocation();
        Point ownerLocation = _owner.getLocationOnScreen();
        Dimension ownerSize = _owner.getSize();

        Dimension size = this.getSize();    

        this.setLocation(
            ownerLocation.x + (ownerSize.width - size.width)/2, 
            ownerLocation.y + (ownerSize.height - size.height)/2
        );
    }
}
