/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import javax.swing.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.plus.*;
import org.pdbj.viewer.site.*;
import org.pdbj.viewer.controller.*;

/**
 * Class that represents a thread reading files. Not safe in muti-thread.
 */
class ReaderThread extends Thread implements MessageListener
{
    private ReaderContext _context;

    private ListenerTask    _listenerTask;
    private ReaderFrameTask _frameTask;

    /**
     * Constructor.
     */
    ReaderThread(GuiWrapper guiWrapper, ViewerController controller)
    {
        _context = new ReaderContext();

        _context.guiWrapper = guiWrapper;
        _context.controller = controller;

        _listenerTask = new ListenerTask(_context);
        _frameTask = new ReaderFrameTask(_context);
    }

    /**
     * Sets up file attribute.
     */
    void setupFile(
        FileType fileType, FileLocator locator, boolean toHistory, boolean defaultImage,
        MessageDialog messageDialog, FileStatus copy, PDBmlUrl pdbmlUrl)
    {
        _context.fileType = fileType;
        _context.locator = locator;
        _context.toHistory = toHistory;
        _context.image = defaultImage;
        _context.messageDialog = messageDialog;
        _context.copy = copy;
        _context.pdbmlUrl = pdbmlUrl;
        
        _context.endCommand = (fileType != FileType.FUNCTION 
                && fileType != FileType.SITELIST);
    }
    
    void setMoleculeFile(MoleculeFile moleculeFile)
    {
        _context.moleculeFile = moleculeFile;
    }

    void setSiteListKey(String key) {
        _context.siteListKey = key;
    }

    void setEfsiteData(EfsiteData efsiteData) {
        _context.efsiteData = efsiteData;
    }

    //--------------------------------------------------------------------
    // Implements Runnable
    //--------------------------------------------------------------------
    /**
     * Run.
     */
    public void run()
    {
        try{
            this.read();
        }
        catch(RecoverableException e){
            _context.recoverableException = e;
        }
        catch(OutOfMemoryError e){
            _context.fatalException = ExceptionUtil.handleOutOfMemoryError(e);
        }
        catch(Throwable t){
            _context.fatalException = ExceptionUtil.handleThrowable(t);
        }

        if(_context.recoverableException != null)
        {
            if (_context.efsiteData != null) {
                _context.efsiteData.incrementError();
            }
            _frameTask.setStatus(ReaderFrameTask.RECOVERABLE_ERROR);
            this.invokeFrame();
        }
        else if(_context.fatalException != null)
        {
            _frameTask.setStatus(ReaderFrameTask.FATAL_ERROR);
            this.invokeFrame();
        }
    }

    private void read() throws RecoverableException
    {
        FileType fileType = _context.fileType;
        FileLocator locator = _context.locator;
        PDBmlUrl pdbmlUrl = _context.pdbmlUrl;

        if(fileType == FileType.PDBML)
        {
            if( locator.isLocalFile() )
            {
                _context.moleculeFile = FileLoader.loadPDBmlFile(locator.getFile(), this, pdbmlUrl);
            }
            else if( locator.isRemoteFile() )
            {
                _context.moleculeFile = FileLoader.loadPDBmlFile(locator.getURL(), this, pdbmlUrl);
            }
            else return;

            _frameTask.setStatus(ReaderFrameTask.LOAD_MOLECULE);
        }
        else if(fileType == FileType.FTP_PDBML)
        {
            if( locator.isPDBCode() )
            {
                _context.moleculeFile = FileLoader.loadPDBmlFileFTP(locator.getPDBCode(), this, pdbmlUrl);
            }
            else return;

            _frameTask.setStatus(ReaderFrameTask.LOAD_MOLECULE);
        }
        else if(fileType == FileType.PDB)
        {
            if( locator.isLocalFile() )
            {
                _context.moleculeFile = FileLoader.loadPDBFile(locator.getFile());
            }
            else if( locator.isRemoteFile() )
            {
                if (_context.efsiteData != null) {
                    _context.efsiteData.setMoleculeFile(FileLoader.loadPDBFile(locator.getURL(), this));
                } else {
                    _context.moleculeFile = FileLoader.loadPDBFile(locator.getURL(), this);
                }
            }
            else return;

            _frameTask.setStatus(ReaderFrameTask.LOAD_MOLECULE);
        }
        else if(fileType == FileType.POLYGON)
        {
            if( locator.isLocalFile() )
            {
                _context.polygonFile = FileLoader.loadPolygonFile(locator.getFile());
            }
            else if( locator.isRemoteFile() )
            {
                if (_context.efsiteData != null) {
                    _context.efsiteData.setPolygonFile(FileLoader.loadPolygonFile(locator.getURL(), this));
                } else {
                    _context.polygonFile = FileLoader.loadPolygonFile(locator.getURL(), this);
                }
            }
            else return;

            _frameTask.setStatus(ReaderFrameTask.LOAD_POLYGON);
        }
        else if(fileType == FileType.POLYGON_CGO)
        {
            if( locator.isLocalFile() )
            {
                _context.polygonFile = FileLoader.loadPolygonCgoFile(locator.getFile());
            }
            else if( locator.isRemoteFile() )
            {
                _context.polygonFile = FileLoader.loadPolygonCgoFile(locator.getURL(), this);
            }
            else return;

            _frameTask.setStatus(ReaderFrameTask.LOAD_POLYGON);
        }
        else if(fileType == FileType.ANIMATION)
        {
            if( locator.isLocalFile() )
            {
                _context.moleculeFile = FileLoader.loadAnimationFile(locator.getFile(), this);
            }
            else if( locator.isRemoteFile() )
            {
                _context.moleculeFile = FileLoader.loadAnimationFile(locator.getURL(), this);
            }
            else return;

            _frameTask.setStatus(ReaderFrameTask.LOAD_MOLECULE);
        }
        else if(fileType == FileType.FUNCTION)
        {
            _context.moleculeFile.setFunction(Function.UNDEFINED);

            if( locator.isPDBCode() )
            {
                Function function = FileLoader.loadFunctionFileFTP(locator.getPDBCode(), this, pdbmlUrl);
                _context.moleculeFile.setFunction(function);
            }
            
            return;
        } else if (fileType == FileType.SITELIST) {
            String key = _context.siteListKey;
            _context.moleculeFile.addSiteList(key, SiteList.UNDEFINED);
            SiteList siteList = FileLoader.loadSiteList(locator.getURL(), this);
            _context.moleculeFile.addSiteList(key, siteList);
            return;
        } else {
            return;
        }

        this.invokeFrame();
    }

    //--------------------------------------------------------------------
    // Implements MessageListener
    //--------------------------------------------------------------------
    /**
     * Returns whethrer this is active.
     */
    public boolean isActive()
    {
        if(_context.messageDialog != null){
            return _context.messageDialog.isActive();
        }
        else return false;
    }

    /**
     * Begins message.
     */
    public void begin(int numLines)
    {
        if(_context.messageDialog != null){
            _context.initSize = numLines;
            _listenerTask.setStatus(ListenerTask.BEGIN);
            this.invokeListener();
        }
    }

    /**
     * Appends a message.
     */
    public void appendMessage(String message)
    {
        if(_context.messageDialog != null){
            _context.message = message;
            _listenerTask.setStatus(ListenerTask.ADD_MESSAGE);
            this.invokeListener();
        }
    }

    /**
     * Sets the message.
     */
    public void setMessage(String message, int index)
    {
        if(_context.messageDialog != null){
            _context.index = index;
            _context.message = message;
            _listenerTask.setStatus(ListenerTask.SET_MESSAGE);
            this.invokeListener();
        }
    }

    /**
     * Clears message.
     */
    public void clear()
    {
        if(_context.messageDialog != null){
            _listenerTask.setStatus(ListenerTask.CLEAR);
            this.invokeListener();
        }
    }

    /**
     * Ends message.
     */
    public void end()
    {
        if(_context.messageDialog != null){
            _listenerTask.setStatus(ListenerTask.END);
            this.invokeListener();
        }
    }

    //--------------------------------------------------------------------
    // Invoke swing
    //--------------------------------------------------------------------
    private void invokeListener()
    {
        try{
            SwingUtilities.invokeAndWait(_listenerTask);
        }
        catch(Exception e){
            e.printStackTrace();
        }
    }

    private void invokeFrame()
    {
        try{
            SwingUtilities.invokeAndWait(_frameTask);
        }
        catch(Exception e){
            e.printStackTrace();
        }
    }
}
