/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.io.*;
import java.net.*;
import java.util.ArrayList;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;

/**
 * Controls script files.
 */
public class ScriptController
{
    private ScriptFile _scriptFile;
    private boolean _pausing;

    /**
     * Returns the script file.
     */
    public ScriptFile getFile()
    {
        return _scriptFile;
    }

    /**
     * Returns whether there exists more commands.
     */
    public boolean hasMoreCommand()
    {
        if(_scriptFile != null){
            return _scriptFile.hasMoreCommand();
        }
        else return false;
    }

    /**
     * Returns the next command.
     */
    public String nextCommand()
    {
        return _scriptFile.nextCommand();
    }
    
    /**
     * Sets whether the execution is pausing.
     */
    public void setPausing(boolean pausing)
    {
        _pausing = pausing;
    }

    /**
     * Returns whether the execution is pausing.
     */
    public boolean isPausing()
    {
        return _pausing;
    }

    /**
     * Loads a script file.
     */
    public void load(File file) throws RecoverableException
    {
        this.loadFile( file.getAbsolutePath(), new ResourceRequest(file) );
    }

    /**
     * Loads a script file.
     */
    public void load(URL url) throws RecoverableException
    {
        this.loadFile( url.toString(), new ResourceRequest(url) );
    }

    //--------------------------------------------------------------------
    // Load
    //--------------------------------------------------------------------
    private void loadFile(String path, ResourceRequest request) throws RecoverableException
    {
        try{
            _scriptFile = this.build(path, request);
        }
        catch(FileNotFoundException e){
            throw new RecoverableException(ErrorMessages.CANT_FIND_SCRIPT_FILE);
        }
        catch(UnknownHostException e){
            throw ExceptionUtil.handleUnknownHostException(e);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
    }

    private ScriptFile build(String path, ResourceRequest request) throws IOException
    {
        String[] commands = null;
        BufferedReader reader = null;

        try{
            reader = new BufferedReader( new InputStreamReader( request.openStream() ) );

            ArrayList list = new ArrayList();

            String s = null;
            while( (s = reader.readLine()) != null) list.add(s);

            commands = (String[])list.toArray(new String[0]);
        }
        finally{
            if(reader != null) reader.close();
            request.dispose();
        }

        return new ScriptFile(path, commands);
    }
}
