/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.io.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;

/**
 * Class that reads and writes UrlConfig.
 */
public class UrlConfigHandler extends DefaultHandler
{
    private SAXNodeHandler _handler;

    private SAXNode _remote_file;
    private SAXNode _params;
    private SAXNode _param;

    private boolean      _pickupText;
    private StringBuffer _buffer;

    private UrlConfig _config;

    /**
     * Reads a UrlConfig.
     */
    public UrlConfig read(File file) throws RecoverableException
    {
        if(_handler == null) this.setupHandler();
        
        _config = null;

        try{
            this.readConfig(file);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }        

        return _config;
    }

    /**
     * Writes a UrlConfig.
     */
    public void write(UrlConfig config, File file) throws RecoverableException
    {
        try{
            this.writeConfig(config, file);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }        
    }

    //--------------------------------------------------------------------
    // Write
    //--------------------------------------------------------------------
    private void writeConfig(UrlConfig config, File file) throws IOException
    {
        String  url = config.getUrl();
        boolean attachParam = config.isAttachParameter();
        String  delimiter = config.getDelimiter();

        BufferedWriter writer = new BufferedWriter( new FileWriter(file) );
        
        writeLine(writer, "<?xml version=\"1.0\"?>");
        writeLine(writer, "");
        writeLine(writer, "<remote_file url=\"" + url + "\">");
        writeLine(writer, "  <params attach=\"" + attachParam + "\" coordinates_delimiter=\"" + delimiter + "\">");
        for(int i=0; i<config.numParameters(); i++)
        {
            String name  = config.getName(i);
            String value = config.getValue(i);
            String type  = config.getType(i);
            writeLine(writer, "    <param name=\"" + name + "\" value=\"" + value + "\" type=\"" + type + "\"/>");
        }
        writeLine(writer, "  </params>");
        writeLine(writer, "</remote_file>");
        
        writer.close();
    }

    private void writeLine(BufferedWriter writer, String line) throws IOException
    {
        writer.write(line);
        writer.newLine();
    }

    //--------------------------------------------------------------------
    // Read
    //--------------------------------------------------------------------
    private void readConfig(File file) throws IOException, RecoverableException
    {
        ResourceRequest request = new ResourceRequest(file);

        InputStream stream = null;
        try{
            stream = request.openStream();
            this.parse(stream);            
        }
        finally{
            if(stream != null) stream.close();
            request.dispose();
        }
    }

    //--------------------------------------------------------------------
    // Parse
    //--------------------------------------------------------------------
    private void parse(InputStream stream) throws IOException, RecoverableException
    {
        try{
            SAXParserFactory spFactory = SAXParserFactory.newInstance();
            spFactory.setNamespaceAware(false);

            SAXParser parser = spFactory.newSAXParser();
            parser.parse(stream, this);
        }
        catch(SAXException e){
            throw ExceptionUtil.handleSAXException(e);
        }
        catch(ParserConfigurationException e){
            throw ExceptionUtil.handleParserConfigurationException(e);
        }
    }

    //-------------------------------------------------------------------------
    // SAX Handler
    //-------------------------------------------------------------------------
    /**
     * Start document.
     */
    public void startDocument() throws SAXException
    {
        _handler.initialize();
        _buffer = new StringBuffer();
        _config = new UrlConfig();
    }

    /**
     * Start element.
     */
    public void startElement(
        String uri, String localName, String qName, Attributes attributes)
        throws SAXException
    {
        _handler.enter(qName);

        _pickupText = false;

        SAXNode curr = _handler.getCurrentNode();
        if(curr != null)
        {
            if(curr == _remote_file)
            {
                String url = attributes.getValue("url");
                if(url != null) _config.setUrl(url);
            }
            else if(curr == _params)
            {
                String attach = attributes.getValue("attach");
                String delimiter = attributes.getValue("coordinates_delimiter");
                if( attach != null && attach.equals("false") ){
                    _config.setAttachParameter(false);
                }
                if( delimiter != null  ){
                    _config.setDelimiter(delimiter);
                }
            }
            else if(curr == _param)
            {
                String name  = attributes.getValue("name");
                String value = attributes.getValue("value");
                String type  = attributes.getValue("type");
                if( name != null && !name.equals("") )
                {
                    if(value == null) value = "";
                    type = this.fixType(type);
                    _config.addParameter(name, value, type);
                }
            }

            _pickupText = true;
            _buffer.setLength(0);
        }
    }

    /**
     * Characters.
     */
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        if(_pickupText){
            _buffer.append( new String(ch, start, length) );
        }
    }

    /**
     * End element.
     */
    public void endElement(String uri, String localName, String qName)
        throws SAXException
    {
        _handler.leave();
    }

    /**
     * End document.
     */
    public void endDocument() throws SAXException
    {
    }

    //--------------------------------------------------------------------
    // Set up SAXNodeHandler
    //--------------------------------------------------------------------
    private void setupHandler()
    {
        _remote_file = new SAXNode("remote_file");
        _params = _remote_file.createNode("params");
        _param = _params.createNode("param");

        _handler = new SAXNodeHandler(_remote_file);
    }

    //--------------------------------------------------------------------
    // Util
    //--------------------------------------------------------------------
    private String fixType(String type)
    {
        if( !type.equals(UrlConfig.TYPE_TEXT) &&
            !type.equals(UrlConfig.TYPE_ATOM_ID) &&
            !type.equals(UrlConfig.TYPE_ATOM) &&
            !type.equals(UrlConfig.TYPE_RESIDUE) &&
            !type.equals(UrlConfig.TYPE_ORIG_COORD) &&
            !type.equals(UrlConfig.TYPE_CURR_COORD) )
        {
            return UrlConfig.TYPE_TEXT;
        }
        else return type;
    }
}
