/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.mouse.*;

/**
 * URL dialog.
 */
public class UrlDialog extends JDialog implements ActionListener, ItemListener
{
    private static final long serialVersionUID = 1L;

    private static final String[] TYPES = {
        UrlConfig.TYPE_TEXT, 
        UrlConfig.TYPE_ATOM_ID, 
        UrlConfig.TYPE_ATOM, 
        UrlConfig.TYPE_RESIDUE, 
        UrlConfig.TYPE_ORIG_COORD, 
        UrlConfig.TYPE_CURR_COORD
    };
    
    private GuiWrapper _guiWrapper;
    private FileType _fileType;

    private UrlConfig _config;
    private File _currDir;

    private JTextField _textURL;
    private JTextField _textCell;
    private JTable _tableParams;
    private DefaultTableModel _tableModel;
    private JScrollPane _scrollPane;
    private JCheckBox _checkboxParams;
    private JTextField _textDelimiter;
    private JTextField _textFit;

    private JButton _buttonAdd;
    private JButton _buttonRemove;
    private JButton _buttonClear;
    private JButton _buttonOK;
    private JButton _buttonCancel;
    private JButton _buttonSubmit;

    private JButton _buttonLoad;
    private JButton _buttonSave;
    
    /**
     * Constructor.
     */
    public UrlDialog(Frame owner, String title, boolean modal)
    {
        super(owner, title, modal);
        this.createComponent();
        this.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

        _config = new UrlConfig();
        _currDir = new File("./");
    }

    /**
     * Sets the GUI wrapper.
     */
    public void setGuiWrapper(GuiWrapper guiWrapper)
    {
        _guiWrapper = guiWrapper;
    }

    /**
     * Sets this dialog at the center of the passed component.
     */
    public void setCenterTo(Component component)
    {
        Point ownerLocation = component.getLocation();
        Dimension ownerSize = component.getSize();

        Dimension size = this.getSize();
        this.setLocation(
            ownerLocation.x + (ownerSize.width - size.width)/2, 
            ownerLocation.y + (ownerSize.height - size.height)/2
        );
    }

    /**
     * Sets the file type.
     */
    public void setFileType(FileType fileType)
    {
        _fileType = fileType;
    }

    /**
     * Returns the file type.
     */
    public FileType getFileType()
    {
        return _fileType;
    }

    /**
     * Returns the URL.
     */
    public String getUrl(String encode) throws UnsupportedEncodingException
    {
        String url = _config.getUrl();

        if( _config.isAttachParameter() )
        {
            int index = url.indexOf("?");
            if(index != -1) url = url.substring(0, index);
            String params = _config.encodeParameters(encode);
            if( !params.equals("") ) url += "?" + params;
        }

        return url;
    }

    /**
     * Returns the ID to be fit.
     */
    public int getFitTo()
    {
        int id = -1;
        try{
            String text = _textFit.getText().trim();
            if( !text.equals("") ) id = Integer.parseInt(text);
        }
        catch(NumberFormatException e){
        }
        return id;
    }

    /**
     * Sets the pick result.
     */
    public void setPickResult(PickResult pickResult)
    {
        if( !_checkboxParams.isSelected() ) return;
        
        int index = _tableParams.getSelectedRow();
        if(index == -1) return;

        String value = null;
        String type = (String)_tableModel.getValueAt(index, 2);    
        if( type.equals(UrlConfig.TYPE_ATOM_ID) ){
            value = "" + pickResult.getAtomId();
        }
        else
        {
            String delim = _textDelimiter.getText().trim();
            if( delim.equals("") ) delim = " ";

            if( type.equals(UrlConfig.TYPE_ATOM) ){
                String chain = pickResult.getChainId().trim();
                String residue = "" + pickResult.getResidueId();
                String atom = "" + pickResult.getAtomId();
                value = chain + delim + residue + delim + atom;
            }
            else if( type.equals(UrlConfig.TYPE_RESIDUE) ){
                String chain = pickResult.getChainId().trim();
                String residue = "" + pickResult.getResidueId();
                value = chain + delim + residue;
            }
            else if( type.equals(UrlConfig.TYPE_ORIG_COORD) ){
                String[] coords = pickResult.toStringOrigCoordinates(3);
                value = "" + coords[0] + delim + coords[1] + delim + coords[2];
            }
            else if( type.equals(UrlConfig.TYPE_CURR_COORD) ){
                String[] coords = pickResult.toStringCoordinates(3);
                value = "" + coords[0] + delim + coords[1] + delim + coords[2];
            }
            else return;
        }
        
        _tableModel.setValueAt(value, index, 1);

        if( _tableParams.isEditing() )
        {
            int row = _tableParams.getEditingRow();
            int col = _tableParams.getEditingColumn();
            if(row == index && col == 1){
                _textCell.setText(value);
            }
        }
    }

    //--------------------------------------------------------------------
    // Implements ActionListener
    //--------------------------------------------------------------------
    /**
     * Called when a button action is performed.
     */
    public void actionPerformed(ActionEvent e)
    {
        if(e.getSource() == _buttonOK)
        {
            this.execute();
            this.setVisible(false);
        }
        else if(e.getSource() == _buttonSubmit)
        {
            this.execute();
        }
        else if(e.getSource() == _buttonCancel)
        {
            this.setVisible(false);
        }
        else if(e.getSource() == _buttonAdd){
            this.addParam();
        }
        else if(e.getSource() == _buttonRemove){
            this.removeParam();
        }
        else if(e.getSource() == _buttonClear){
            this.clearParam();
        }
        else if(e.getSource() == _buttonLoad){
            File file = this.chooseFile(false);
            if(file != null){
                this.loadConfigFile(file);
            }
        }
        else if(e.getSource() == _buttonSave){
            File file = this.chooseFile(true);
            if(file != null){
                this.saveConfigFile(file);
            }
        }
    }

    //--------------------------------------------------------------------
    // Implements ItemListener
    //--------------------------------------------------------------------
    /**
     * Called when an item state is changed.
     */
    public void itemStateChanged(ItemEvent e)
    {
        AbstractButton item = (AbstractButton)e.getSource();
        this.setEnableParam( item.isSelected() );
    }

    //--------------------------------------------------------------------
    // ComponentListener
    //--------------------------------------------------------------------
    private void dialog_shown(ComponentEvent e)
    {
        this.readConfig();
    }

    //--------------------------------------------------------------------
    // Handle config
    //--------------------------------------------------------------------
    private void readConfig()
    {
        _textURL.setText( _config.getUrl() );
        _checkboxParams.setSelected( _config.isAttachParameter() );
        _textDelimiter.setText( _config.getDelimiter() );

        this.setEnableParam( _config.isAttachParameter() );

        while(_tableModel.getRowCount() > 0){
            _tableModel.removeRow(0);
        }

        int num = _config.numParameters();
        for(int i=0; i<num; i++)
        {
            Vector v = new Vector();
            v.addElement( _config.getName(i) );
            v.addElement( _config.getValue(i) );
            v.addElement( _config.getType(i) );
            _tableModel.addRow(v);
        }
        
        this.moveEdit();
    }
    
    private void writeConfig()
    {
        _config.setUrl( _textURL.getText().trim() );
        _config.setAttachParameter( _checkboxParams.isSelected() );
        _config.setDelimiter( _textDelimiter.getText() );
        
        this.fixEdit();
        _config.clearParameters();
        for(int i=0; i<_tableModel.getRowCount(); i++)
        {
            String name  = (String)_tableModel.getValueAt(i, 0);
            String value = (String)_tableModel.getValueAt(i, 1);
            String type  = (String)_tableModel.getValueAt(i, 2);
            if( !name.equals("") ){
                _config.addParameter(name, value, type);
            }
        }
    }
    
    private void loadConfigFile(File file)
    {
        UrlConfigHandler handler = new UrlConfigHandler();
        try{
            _config = handler.read(file);
        }
        catch(RecoverableException e){
            this.handleRecoverableException(e);
        }
        
        this.readConfig();
    }

    private void saveConfigFile(File file)
    {
        this.writeConfig();

        UrlConfigHandler handler = new UrlConfigHandler();
        try{
            handler.write(_config, file);
        }
        catch(RecoverableException e){
            this.handleRecoverableException(e);
        }
    }

    private void handleRecoverableException(RecoverableException e)
    {
        JOptionPane.showMessageDialog(
            this, e.getMessage(), "Warning", JOptionPane.WARNING_MESSAGE
        );
    }

    //--------------------------------------------------------------------
    // Execute
    //--------------------------------------------------------------------
    private void execute()
    {
        this.writeConfig();
        if(_guiWrapper != null){
            _guiWrapper.loadRemoteFile(this);
        }
    }

    //--------------------------------------------------------------------
    // Handle params
    //--------------------------------------------------------------------
    private void setEnableParam(boolean enable)
    {
        _tableParams.setEnabled(enable);
        _buttonAdd.setEnabled(enable);
        _buttonRemove.setEnabled(enable);
        _buttonClear.setEnabled(enable);
        _textDelimiter.setEnabled(enable);
        _textCell.setEnabled(enable);
    }
    
    private void addParam()
    {
        Vector v = new Vector();
        v.addElement("");
        v.addElement("");
        v.addElement("text");
        _tableModel.addRow(v);
        
        this.updateScrollBar();
    }

    private void removeParam()
    {
        int index = _tableParams.getSelectedRow();
        if(index != -1){
            _tableModel.removeRow(index);
            this.moveEdit();
        }
    }

    private void clearParam()
    {
        while(_tableModel.getRowCount() > 0 ){
            _tableModel.removeRow(0);
        }
        this.moveEdit();
    }

    private void updateScrollBar()
    {
        _scrollPane.validate();

        JScrollBar scrollBar = _scrollPane.getVerticalScrollBar();
        scrollBar.setValue( scrollBar.getMaximum() );

        _scrollPane.invalidate();
    }

    //--------------------------------------------------------------------
    // Handle cell edit
    //--------------------------------------------------------------------
    private void fixEdit()
    {
        if( _tableParams.isEditing() ){
            int row = _tableParams.getEditingRow();
            int col = _tableParams.getEditingColumn();
            if(_tableModel.getRowCount() > row){
                _tableModel.setValueAt(_textCell.getText(), row, col);
            }
        }
    }

    private void moveEdit()
    {
        if( _tableParams.isEditing() )
        {
            int row = _tableParams.getEditingRow();
            int col = _tableParams.getEditingColumn();
            if(_tableModel.getRowCount() > 0){
                if( _tableModel.getRowCount() >= row){
                    row = _tableModel.getRowCount() - 1;
                    _tableParams.setEditingRow(row);
                }
                String value = (String)_tableModel.getValueAt(row, col);
                _textCell.setText(value);
            }
            else{
                _tableParams.setEditingRow(0);
                _tableParams.setEditingColumn(0);
                _textCell.setText("");
            }
        }
    }

    //--------------------------------------------------------------------
    // File chooser
    //--------------------------------------------------------------------
    private File chooseFile(boolean save)
    {
        JFileChooser jFileChooser = new JFileChooser(_currDir);
        jFileChooser.setMultiSelectionEnabled(false);
        jFileChooser.setSelectedFile( new File(_currDir, "site.xml") );

        int option;
        if(save){
            option = jFileChooser.showSaveDialog(this);
        }
        else{
            option = jFileChooser.showOpenDialog(this);
        }

        if(option == JFileChooser.APPROVE_OPTION)
        {
            File file = jFileChooser.getSelectedFile();
            _currDir = file.getParentFile();
            return file;
        }
        return null;
    }

    //--------------------------------------------------------------------
    // Create component
    //--------------------------------------------------------------------
    private void createComponent()
    {
        int tw = 400;
        int th = 118;
        int W = tw + 130;
        int H = th + 230;
        this.setSize( new Dimension(W, H) );
        this.setResizable(false);

        Container container = this.getContentPane();
        container.setLayout(null);

        Font font = new Font("SansSerif", Font.PLAIN, 12);
        
        int x = 20;
        int y = 20;

        JLabel labelURL = new JLabel("Enter URL");
        labelURL.setFont(font);
        labelURL.setSize( labelURL.getPreferredSize() );
        labelURL.setLocation(x, y);

        y += 30;

        _textURL = new JTextField();
        _textURL.setFont(font);
        _textURL.setSize(400, 20);
        _textURL.setLocation(x, y);
        _textURL.addKeyListener(
            new KeyAdapter(){
                public void keyTyped(KeyEvent e){
                    handle_keyTyped(e);
                }
            }
        );

        //y += 30;

        JLabel labelFit = new JLabel("fit to");
        labelFit.setFont(font);
        Dimension labelFitSize = labelFit.getPreferredSize();
        labelFit.setSize(labelFitSize);
        labelFit.setLocation(x+420, y);

        _textFit = new JTextField();
        _textFit.setFont(font);
        _textFit.setSize(25, 20);
        _textFit.setLocation(x + labelFitSize.width + 430, y);

        y += 40;

        _checkboxParams = new JCheckBox("Attach Parameters");
        _checkboxParams.setFont(font);
        _checkboxParams.setSize( _checkboxParams.getPreferredSize() );
        _checkboxParams.setLocation(x, y);
        _checkboxParams.addItemListener(this);

        y += 30;

        this.createTable();
        _scrollPane.setSize(tw, th);
        _scrollPane.setLocation(x, y);

        _buttonRemove = new JButton("Remove");
        _buttonRemove.setFont(font);
        Dimension buttonSize1 = _buttonRemove.getPreferredSize();
        _buttonRemove.setSize(buttonSize1);
        _buttonRemove.setLocation(x+tw+10, y+30);
        _buttonRemove.addActionListener(this);

        _buttonAdd = new JButton(" Add ");
        _buttonAdd.setFont(font);
        _buttonAdd.setSize(buttonSize1);
        _buttonAdd.setLocation(x+tw+10, y);
        _buttonAdd.addActionListener(this);

        _buttonClear = new JButton("Clear");
        _buttonClear.setFont(font);
        _buttonClear.setSize(buttonSize1);
        _buttonClear.setLocation(x+tw+10, y+60);
        _buttonClear.addActionListener(this);

        y += (th + 10);

        JLabel labelDelim = new JLabel("delimiter :");
        labelDelim.setFont(font);
        Dimension labelSize = labelDelim.getPreferredSize();
        labelDelim.setSize(labelSize);
        labelDelim.setLocation(x, y);

        _textDelimiter = new JTextField();
        _textDelimiter.setFont(font);
        _textDelimiter.setSize(25, 20);
        _textDelimiter.setLocation(x + labelSize.width + 10, y);

        y += 30;

        _buttonCancel = new JButton("Cancel");
        _buttonCancel.setFont(font);
        Dimension buttonSize2 = _buttonCancel.getPreferredSize();
        int w = buttonSize2.width;
        _buttonCancel.setSize(buttonSize2);
        _buttonCancel.setLocation(W - 2*w - 30, y);
        _buttonCancel.addActionListener(this);

        _buttonSubmit = new JButton("Submit");
        _buttonSubmit.setFont(font);
        _buttonSubmit.setSize(buttonSize2);
        _buttonSubmit.setLocation(W - w - 20, y);
        _buttonSubmit.addActionListener(this);

        _buttonOK = new JButton("  OK  ");
        _buttonOK.setFont(font);
        _buttonOK.setSize(buttonSize2);
        _buttonOK.setLocation(W - 3*w - 40, y);
        _buttonOK.addActionListener(this);

        _buttonSave = new JButton(" Save ");
        _buttonSave.setFont(font);
        _buttonSave.setSize(buttonSize2);
        _buttonSave.setLocation(W - 4*w - 70, y);
        _buttonSave.addActionListener(this);

        _buttonLoad = new JButton(" Load ");
        _buttonLoad.setFont(font);
        _buttonLoad.setSize(buttonSize2);
        _buttonLoad.setLocation(W - 5*w - 80, y);
        _buttonLoad.addActionListener(this);

        container.add(labelURL);
        container.add(_textURL);
        container.add(labelFit);
        container.add(_textFit);
        container.add(_checkboxParams);
        container.add(_scrollPane);
        container.add(_buttonAdd);
        container.add(_buttonRemove);
        container.add(_buttonClear);
        container.add(labelDelim);
        container.add(_textDelimiter);
        container.add(_buttonLoad);
        container.add(_buttonSave);
        container.add(_buttonOK);
        container.add(_buttonCancel);
        container.add(_buttonSubmit);
        
        this.addComponentListener(
            new ComponentAdapter(){
                public void componentShown(ComponentEvent e){
                    dialog_shown(e);
                }
            }
        );
    }
    
    private void createTable()
    {
        Font font = new Font("SansSerif", Font.PLAIN, 12);

        String[] columns = {"name", "value", "type"};
        Vector columnNames = new Vector( Arrays.asList(columns) );
        Vector rowData = new Vector();

        _tableModel = new DefaultTableModel(rowData, columnNames);

        _tableParams = new JTable(_tableModel);
        _tableParams.setFont(font);
        _tableParams.getTableHeader().setFont(font);

        _textCell = new JTextField();
        DefaultCellEditor editor1 = new DefaultCellEditor(_textCell);
        _tableParams.getColumnModel().getColumn(0).setCellEditor(editor1);
        _tableParams.getColumnModel().getColumn(1).setCellEditor(editor1);

        DefaultCellEditor editor2 = new DefaultCellEditor( new JComboBox(TYPES) );
        _tableParams.getColumnModel().getColumn(2).setCellEditor(editor2);
        _tableParams.getColumnModel().getColumn(2).setMinWidth(80);
        _tableParams.getColumnModel().getColumn(2).setMaxWidth(80);

        _tableParams.setRowHeight(20);

        _scrollPane = new JScrollPane();
        _scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        _scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        _scrollPane.setViewportView(_tableParams);
        
        Color background = this.getContentPane().getBackground();
        _scrollPane.getViewport().setBackground(background);
    }

    private void handle_keyTyped(KeyEvent e)
    {
        if(e.getKeyChar() == KeyEvent.VK_ENTER)
        {
            e.consume();
            this.execute();
            this.setVisible(false);
        }
    }
}
