/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.File;
import java.net.*;
import java.util.*;
import javax.media.opengl.GLProfile;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import netscape.javascript.*;
import org.jdesktop.applet.util.JNLPAppletLauncher;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.plus.*;
import org.pdbj.viewer.renderer.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.mouse.*;
import org.pdbj.viewer.commands.*;
import org.pdbj.viewer.writers.*;

/**
 * Applet.
 */
public class ViewerApplet extends JApplet
    implements CommandHandler, FileStatusListener, GuiWrapper,
    MouseCallbackReceiver, RendererListener, MousePickListener
{
    static volatile boolean glInitialized;
    static {
        synchronized (ViewerApplet.class) {
            if (!glInitialized) {
                GLProfile.initSingleton();
                glInitialized = true;
            }
        }
    }

    private static final long serialVersionUID = 1L;

    private String _name;
    private String _mouseSendTo;
    private String _pickingReceiver;

    private boolean _signed;

    private String[] _pdbIds;
    private String[] _pdbmlURLs;
    private String[] _pdbURLs;
    private String[] _polygonURLs;
    private String[] _polygonCgoURLs;
    private String[] _animURLs;

    private LinkedList _fileLoadTasks;
    private String _initCommands;

    private String[]       _mouseTargetName;
    private ViewerApplet[] _mouseTarget;
    private Object[] _mouseTargetObject;
    private boolean        _forwardMouseEvent;

    private long _ticks;
    private static final long CLICK_INTERVAL = 200; // 0.2 second

    private boolean _useCommandArea;
    private boolean _useFileStatusPanel;
    private boolean _usePopupMenu;
    private boolean _reportMousePick;
    private boolean _fileLoadMessage;
    private boolean _callOnStart;
    private boolean _newt;

    private FileStatusPanel _fileStatusPanel;
    private ViewerStatusPanel _viewerStatusPanel;
    private CommandArea     _commandArea;

    private JSObject _window;
    private boolean _jsFailed;

    private ViewerController _controller;
    private MouseBehavior _mouseBehavior;

    private CommandQueue  _commandQueue;
    private CommandThread _commandThread;

    private CommandExecutor  _executor;
    private JvCommandController _jvCommandController;

    private AnimationHandler _animHandler;
    private AnimationFrame _animationFrame;

    private MessageDialog _messageDialog;

    // flags
    private boolean _jvCommandMode;
    private boolean _screenInitialized;
    private boolean _fileLoading;
    private boolean _stereoInitialized;

    // file chooser directory
    private File _currDir;

    // components
    private ViewerLayout _viewerLayout;
    private boolean _componentInitialized;

    // menus
    private ViewerMenu _menu;
    private PopupMenu _popupMenu;

    /**
     * Returns the name.
     */
    public String getName()
    {
        return _name;
    }

    //--------------------------------------------------------------------
    // Communicates with JavaScript
    //--------------------------------------------------------------------
    /**
     * Executes commands.
     */
    public void executeCommand(String commands)
    {
        if(commands != null){
            this.executeMultipleCommands(commands);
        }
    }

    /**
     * Loads PDB file.
     */
    public void loadPDB(String filePath)
    {
        this.addFile(FileType.PDB, filePath, -1);
    }

    /**
     * Loads PDBML file.
     */
    public void loadPDBML(String filePath)
    {
        this.addFile(FileType.PDBML, filePath, -1);
    }

    /**
     * Loads PDBML file by FTP.
     */
    public void loadFTP(String pdbId)
    {
        this.addFilePdbCode(pdbId, -1);
    }

    /**
     * Loads polygon file.
     */
    public void loadPolygon(String filePath)
    {
        this.addFile(FileType.POLYGON, filePath, -1);
    }

    /**
     * Loads polygon CGO file.
     */
    public void loadPolygonCGO(String filePath)
    {
        this.addFile(FileType.POLYGON_CGO, filePath, -1);
    }

    /**
     * Loads animation file.
     */
    public void loadAnim(String filePath)
    {
        this.addFile(FileType.ANIMATION, filePath, -1);
    }

    /**
     * Loads PDB file.
     */
    public void loadPDB(String filePath, int fitTo)
    {
        this.addFile(FileType.PDB, filePath, fitTo);
    }

    /**
     * Loads PDBML file.
     */
    public void loadPDBML(String filePath, int fitTo)
    {
        this.addFile(FileType.PDBML, filePath, fitTo);
    }

    /**
     * Loads PDBML file by FTP.
     */
    public void loadFTP(String pdbId, int fitTo)
    {
        this.addFilePdbCode(pdbId, fitTo);
    }

    /**
     * Loads polygon file.
     */
    public void loadPolygon(String filePath, int fitTo)
    {
        this.addFile(FileType.POLYGON, filePath, fitTo);
    }

    /**
     * Loads polygon CGO file.
     */
    public void loadPolygonCGO(String filePath, int fitTo)
    {
        this.addFile(FileType.POLYGON_CGO, filePath, fitTo);
    }

    /**
     * Loads animation file.
     */
    public void loadAnim(String filePath, int fitTo)
    {
        this.addFile(FileType.ANIMATION, filePath, fitTo);
    }

    //--------------------------------------------------------------------
    // Overrides Applet interface
    //--------------------------------------------------------------------
    /**
     * Init.
     */
    public void init()
    {
        super.init();

        try
        {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
            SwingUtilities.updateComponentTreeUI(this);
        }
        catch(Exception e)
        {
        }

        _fileLoadTasks = new LinkedList();

        // read parameters
        this.readParameter();

        try {
            _controller = new ViewerController(this, _newt);
        } catch (Throwable t) {
            t.printStackTrace();
            createErrorComponent("Cannot find JOGL");
            return;
        }

        _signed = isSignedApplet();

        _controller.stopRenderer();
        _controller.addFileStatusListener(this);
        
        MousePickListener pickListener = _reportMousePick ? this : null;
        _mouseBehavior = new MouseBehavior(_controller, pickListener);
        _mouseBehavior.addMouseCallbackReceiver(this);
        _forwardMouseEvent = false;

        _animHandler = new AnimationHandler(_controller);

        _executor = new CommandExecutor(
            _controller, _mouseBehavior, this, _animHandler
        );

        _jvCommandController = new JvCommandController();

        _commandQueue = new CommandQueue();
        _commandQueue.setJvCommandController(_jvCommandController);
        _commandThread = new CommandThread(this, _commandQueue, _executor);

        this.setupFiles();
        this.executeMultipleCommands(_initCommands);

        this.initializeMouseTarget();

        if(_useFileStatusPanel)
        {
            _fileStatusPanel = new FileStatusPanel(_controller);
            _fileStatusPanel.setGuiWrapper(this);
            _fileStatusPanel.setCommandHandler(this);

            _viewerStatusPanel = new ViewerStatusPanel();
            _controller.getViewStatus().addViewStatusListener(_viewerStatusPanel);
            _mouseBehavior.addMouseSettingsListener(_viewerStatusPanel);

            _viewerStatusPanel.updateAll(_controller.getViewStatus(), _mouseBehavior);
        }
        if(_useCommandArea)
        {
            this.createCommandArea();
            _commandArea.setCommandHandler(this);
            _mouseBehavior.setMessageArea(_commandArea);
            _executor.setMessageArea(_commandArea);
        }

        this.createComponent();
        if(_usePopupMenu)
        {
            this.createMenu();
        }

        if(_useCommandArea)
        {
            _commandArea.setActive(true);
            _commandArea.requestFocus();
        }
        _messageDialog = new MessageDialog(this, "Progress", 20);
    }

    public void start() {
        if (_controller != null) {
            initComponent();
        }
    }

    private void initComponent() {
        if (!_componentInitialized) {
            _viewerLayout.initialize(getContentPane());
            if (_useFileStatusPanel) {
                _fileStatusPanel.refresh();
            }
            _componentInitialized = true;
        }
    }
    
    private void initViewer()
    {
        initComponent();

        this.setCommandActive(true);
        _commandThread.start();
        _commandQueue.putCommand("", new CommandType(CommandType.START_RENDERER));

        if (_callOnStart) {
            onStart();
        }
    }

    public void destroy()
    {
        if (_controller != null)
        {
            _controller.stopRenderer();
            _animHandler.stopAll();

            _commandThread.terminate();

            if(_animationFrame != null){
                _animationFrame.dispose();
            }

            if (_messageDialog != null) {
                _messageDialog.dispose();
            }

            _controller.dispose();

            if (_newt) {
                this.removeAll();
                _controller.destroyGLWindow();
            }
        }

        super.destroy();
    }

    static boolean isSignedApplet() {
        try {
            new File("./").list();
            return true;
        } catch (java.security.AccessControlException e) {
            return false;
        }
    }

    //--------------------------------------------------------------------
    // Implements MousePickListener
    //--------------------------------------------------------------------
    /**
     * Receives mouse pick result.
     */
    public void receiveMousePick(PickResult pickResult)
    {
        prepareJSWindow();
        if(!_jsFailed && _window != null){
            String[] coord = pickResult.toStringCoordinates(4);
            Object[] args = new Object[]{
                "" + pickResult.getFileId(),
                "" + pickResult.toStringModelId(),
                "" + pickResult.getChainId(),
                "" + pickResult.getResidueId(),
                "" + pickResult.getAtomId(),
                "" + pickResult.getAltLoc(),
                "" + pickResult.getResidueName(),
                "" + pickResult.getAtomName(),
                "" + coord[0],
                "" + coord[1],
                "" + coord[2]
            };
            try{
                _window.call(_pickingReceiver, args);
            }
            catch(JSException e){
                this.handleRecoverableException( new RecoverableException(e.getMessage()) );
                _jsFailed = true;
            }
        }
    }

    //--------------------------------------------------------------------
    // Implements RendererListener
    //--------------------------------------------------------------------
    /**
    * Called when initialized.
    */
    public void screenInitialized()
    {
    }

    /**
    * Called when drawed.
    */
    public void screenDrawed()
    {
        if(!_screenInitialized){
            this.initViewer();
            _screenInitialized = true;
        }
    }

    /**
    * Called when resized.
    */
    public void screenResized(int x, int y, int width, int height)
    {
    }

    //--------------------------------------------------------------------
    // Implements CommandHandler
    //--------------------------------------------------------------------
    /**
     * Executes command.
     */
    public void executeSingleCommand(String command)
    {
        _commandQueue.putCommand(command);
    }

    /**
     * Complements command.
     */
    public CommandComplement complementCommand(String command)
    {
        return _executor.complement(command);
    }

    //--------------------------------------------------------------------
    // Implements FileStatusListener
    //--------------------------------------------------------------------
    /**
     * Receives addition of files.
     */
    public void fileAdded(ArrayList fileStatusList)
    {
        if(_usePopupMenu) _menu.updateFileControllMenu();
    }

    /**
     * Receives remove of files.
     */
    public void fileRemoved(ArrayList fileStatusList)
    {
        if(_usePopupMenu) _menu.updateFileControllMenu();
    }

    /**
     * Receives a change of status.
     */
    public void statusChanged(ArrayList fileStatusList)
    {
        if(_usePopupMenu) _menu.updateFileStatus();
    }

    //--------------------------------------------------------------------
    // Implements GUIWrapper
    //--------------------------------------------------------------------
    /**
     * Returns whether or not the script mode.
     */
    public boolean isScriptMode()
    {
        return false;
    }

    /**
     * Loads PDB file.
     */
    public void loadPDBFile(String filePath, FileStatus copy)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Loads PDBML file.
     */
    public void loadPDBmlFile(String filePath, FileStatus copy)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Loads polygon file.
     */
    public void loadPolygonFile(String filePath, FileStatus copy)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Loads polygon CGO file.
     */
    public void loadPolygonCgoFile(String filePath, FileStatus copy)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Loads script file.
     */
    public void loadScriptFile(String filePath)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Loads animation file.
     */
    public void loadAnimationFile(String filePath, FileStatus copy)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Loads PDBML file by FTP.
     */
    public void loadPDBmlFileFtp(String pdbCode, FileStatus copy)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Loads eF-site data files.
     */
    public void loadEfsite(String efsiteId)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Saves pdb file.
     */
    public void savePDBFile(String filePath)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Saves script file.
     */
    public void saveScriptFile(String filePath)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Saves PNG file.
     */
    public void savePngFile(String filePath, float ratio)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Saves JPEG file.
     */
    public void saveJpegFile(String filePath, float ratio, float quality)
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Pause execution of script.
     */
    public void pauseScript()
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Sets whether the stereo mode.
     */
    public void setupStereo()
    {
        _viewerLayout.setStereo( _controller.getViewStatus().isStereo() );
    }

    /**
     * Terminates the application.
     */
    public void exit()
    {
        this.showMessage(COMMAND_NOT_SUPPORTED);
    }

    /**
     * Returns the FTP message listener.
     */
    public MessageListener getFtpMessageListener()
    {
        return null;
    }

    /**
     * Returns whether or not the standalone application.
     */
    public boolean isStandalone()
    {
        return false;
    }

    /**
     * Loads function file by FTP.
     */
    public void loadFunctionFtp(FileStatus file)
    {
        if(file.getType() == FileStatus.POLYGON) return;

        FileType fileType = FileType.FUNCTION;
        FileLocator locator = new FileLocator(file.getMoleculeFile().getMoleculeId());

        _messageDialog.setWidth(480);
        
        ReaderThread thread = new ReaderThread(this, _controller);
        thread.setupFile(fileType, locator, false, false, _messageDialog, null,
                _controller.getPDBmlUrl());
        thread.setMoleculeFile(file.getMoleculeFile());
        thread.start();
        
        try{
            thread.join();
        }
        catch(InterruptedException e){
        }
    }

    /**
     * Returns the site URL map.
     */
    public HashMap getSiteUrlMap() {
        return null;
    }

    /**
     * Loads the site list.
     */
    public void loadSiteList(FileStatus file, String key) {
    }

    /**
     * Returns the size of the 3D-rendering panel.
     */
    public Dimension getImageSize() {
        return _viewerLayout.getImageSize();
    }

    /**
     * Requests the size of the 3D-rendering panel and 
     * returns the size actually set.
     */
    public Dimension setImageSize(Dimension d) {
        this.showMessage(COMMAND_NOT_SUPPORTED);
        return null;
    }

    /**
     * PDBj execute.
     */
    public void pdbjExecute(int fileId, String name) {
        FileStatus file = _controller.findFile(fileId);
        if (file != null && file.getType() != FileStatus.POLYGON) {
            MoleculeFile moleculeFile = file.getMoleculeFile();
            if (!moleculeFile.isAnimation()) {
                Function function = moleculeFile.getFunction();
                _jvCommandController.start(function, name, fileId);
                startJvCommand();
            }
        }
    }

    /**
     * Initialize stereo.
     */
    public void initializeStereo()
    {
        if (!_stereoInitialized) {
            _controller.initializeStereo();
            _mouseBehavior.initializeStereo();
            _viewerLayout.setCanvasR( _controller.getCanvasR() );
            _viewerLayout.initializeStereo();
            _controller.getCanvasR().addMouseListener(
                new MouseAdapter(){
                    public void mousePressed(MouseEvent e){
                        canvas_mousePressed(e);
                    }
                    public void mouseReleased(MouseEvent e){
                        canvasR_mouseReleased(e);
                    }
                }
            );

            _stereoInitialized = true;
        }
    }

    /**
     * Returns whether NEWT is used.
     */
    public boolean isNewt()
    {
        return _newt;
    }

    /**
     * Returns a viewer controller.
     */
    public ViewerController getViewerController()
    {
        return _controller;
    }

    /**
     * Returns a mouse behavior.
     */
    public MouseBehavior getMouseBehavior()
    {
        return _mouseBehavior;
    }

    /**
     * Loads a local file.
     */
    public void loadLocalFile(FileType fileType)
    {
    }

    /**
     * Loads a remote file.
     */
    public void loadRemoteFile(FileType fileType)
    {
    }

    /**
     * Loads a remote file.
     */
    public void loadRemoteFile(UrlDialog urlDialog)
    {
    }

    /**
     * Loads files specified by PDB ID.
     */
    public void loadByPDBID()
    {
    }

    /**
     * Loads files specified by eF-site ID.
     */
    public void loadByEFsiteID()
    {
    }

    /**
     * Loads a file registered in the history.
     */
    public void loadHistory(FileType fileType, FileLocator locator)
    {
    }

    /**
     * Appends a file history.
     */
    public void addFileHistory(FileType type, FileLocator locator)
    {
    }

    /**
     * Saves a file.
     */
    public void saveFile(FileType fileType)
    {
        if(_controller.getFileStatusList().size() == 0) return;

        File file = this.chooseFile(fileType, true);
        if(file != null){
            this.saveFile(fileType, file, 1, 0.75f);
        }
    }

    /**
     * Shows information of files.
     */
    public void showInformation()
    {
        String header = GuiConstants.NAME + " " + GuiConstants.VERSION + "\n\n";
        
        String[] infos = _controller.getInformation();
        for(int i=0; i<infos.length; i++)
        {
            JOptionPane.showMessageDialog(
                this, header + infos[i], "Information", JOptionPane.INFORMATION_MESSAGE,
                new ImageIcon(getClass().getClassLoader().getResource(GuiConstants.ICON_IMAGE))
            );
        }
    }

    /**
     * Shows an animation frame.
     */
    public void showAnimationFrame()
    {
        if(_animationFrame == null){
            _animationFrame = new AnimationFrame(_animHandler);
            _animationFrame.setCenterTo(this);
        }
        _animationFrame.setVisible(true);
    }

    /**
     * Shows an about dialog.
     */
    public void showAboutDialog()
    {
        StringBuffer sb = new StringBuffer();
        sb.append(GuiConstants.NAME + " " + GuiConstants.VERSION + "\n\n");
        sb.append(GuiConstants.CREDIT);

        JOptionPane.showMessageDialog(
            this, sb.toString(), "About jV", JOptionPane.INFORMATION_MESSAGE,
            new ImageIcon(getClass().getClassLoader().getResource(GuiConstants.ICON_IMAGE))
        );
    }

    /**
     * Opens the manual.
     */
    public void openManual() {
        // do nothing
    }

    /**
     * Rotates stereo mode.
     */
    public void rotateStereo()
    {
        executeSingleCommand("stereo");
    }

    /**
     * Handles a recoverable exception.
     */
    public void handleRecoverableException(RecoverableException e)
    {
        JOptionPane.showMessageDialog(
            this, e.getMessage(), "Warning", JOptionPane.WARNING_MESSAGE
        );
    }

    /**
     * Handles a fatal exception.
     */
    public void handleFatalException(FatalException e)
    {
        JOptionPane.showMessageDialog(
            this, e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE
        );
    }

    /**
     * Starts an execution of a command.
     */
    public void startCommand(String command, CommandType type)
    {
        this.startCommand(_executor.createTokenizer(command), type);
    }

    /**
     * Starts an execution of a command.
     */
    public void startCommand(CommandTokenizer tokenizer, CommandType type)
    {
        if(type.action == CommandType.APPLET_LOAD_FILE)
        {
            FileLoadTask task = (FileLoadTask)_fileLoadTasks.removeFirst();
            if(task != null) this.loadFile(task);
        }
        else
        {
            try{
                if(tokenizer != null) _executor.execute(tokenizer);
            }
            catch(RecoverableException e){
                this.handleRecoverableException(e);
            }
            catch(OutOfMemoryError e){
                this.handleFatalException( ExceptionUtil.handleOutOfMemoryError(e) );
            }
            catch(Throwable t){
                this.handleFatalException( ExceptionUtil.handleThrowable(t) );
            }

            if( type.action == CommandType.START_RENDERER) _controller.startRenderer();

            this.endCommand();
        }
    }

    /**
     * Ends an execution of a command.
     */
    public void endCommand()
    {
        if(_fileLoading){
            _fileLoading = false;
            this.setActive(true);
        }

        if (_jvCommandMode) {
            handleJvCommand();
        } else if (_commandArea != null) {
            _commandArea.setActive(true);
        }
        this.setCommandActive(true);
    }

    /**
     * Returns whether menu is active.
     */
    public boolean menuActive()
    {
        return (!_fileLoading && !_jvCommandMode);
    }

    /**
     * Called when a key is typed to start script.
     */
    public void commandKeyTyped()
    {
    }

    /**
     * Called when focus is moved.
     */
    public void focusMoved()
    {
        if(_commandArea != null) _commandArea.requestFocus();
    }

    //--------------------------------------------------------------------
    // Implements MouseCallbackReceiver
    //--------------------------------------------------------------------
    /**
     * Called when a mouse button is pressed.
     */
    public void receiveMousePressed(MouseEvent e)
    {
        this.callbackMousePressed(e);
    }
    
    /**
     * Called when the mouse pointer is dragged.
     */
    public void receiveMouseDragged(MouseEvent e)
    {
        this.callbackMouseDragged(e);
    }

    //--------------------------------------------------------------------
    // Inter-applet communication
    //--------------------------------------------------------------------
    /**
     * Receives a mouse-pressed event on the canvas of other applets.
     */
    public void receiveCanvasMousePressed(MouseEvent e)
    {
        if(_fileLoading) return;

        _mouseBehavior.handleMousePressed(e);
        if( _forwardMouseEvent && e.getSource() != _controller.getCanvasL() &&
            e.getSource() != _controller.getCanvasR() )
        {
            this.callbackMousePressed(e);
        }
    }
    
    /**
     * Receives a mouse-dragged event on the canvas of other applets.
     */
    public void receiveCanvasMouseDragged(MouseEvent e)
    {
        if(_fileLoading) return;

        _mouseBehavior.handleMouseDragged(e);
        if( _forwardMouseEvent && e.getSource() != _controller.getCanvasL() &&
            e.getSource() != _controller.getCanvasR() )
        {
            this.callbackMouseDragged(e);
        }
    }
    
    //--------------------------------------------------------------------
    // Read parameter
    //--------------------------------------------------------------------
    private void readParameter()
    {
        try{
            _name            = getParameter("name", "").trim();
            _mouseSendTo = getParameter("mouse_send_to", "").trim();
            _pickingReceiver = getParameter("picking_receiver", "receiveMousePick").trim();
    
            _pdbIds     = split( getParameter("pdbid", "").trim() );
            _pdbmlURLs   = split( getParameter("pdbmlURL", "").trim() );
            _pdbURLs     = split( getParameter("pdbURL", "").trim() );
            _polygonURLs = split( getParameter("polygonURL", "").trim() );
            _polygonCgoURLs = split( getParameter("polygon_cgoURL", "").trim() );
            _animURLs    = split( getParameter("animURL", "").trim() );
    
            _initCommands    = getParameter("commands", "").trim();
    
            String commandArea = getParameter("command_area", "false").trim();
            String fileStatus  = getParameter("file_control", "false").trim();
            String popupMenu   = getParameter("popup_menu", "true").trim();
            String mousePick   = getParameter("report_picking", "false").trim();
            String fileLoadMessage = getParameter("file_load_message", "false").trim();
            String callOnStart = getParameter("call_on_start", "false").trim();
            String newt = getParameter("newt", "false").trim();
    
            _useCommandArea = Boolean.valueOf(commandArea).booleanValue();
            _useFileStatusPanel = Boolean.valueOf(fileStatus).booleanValue();
            _usePopupMenu = Boolean.valueOf(popupMenu).booleanValue();
            _reportMousePick = Boolean.valueOf(mousePick).booleanValue();
            _fileLoadMessage = Boolean.valueOf(fileLoadMessage).booleanValue();
            _callOnStart = Boolean.valueOf(callOnStart).booleanValue();
            _newt = Boolean.valueOf(newt).booleanValue();
            if (_newt) {
                _useCommandArea = false;
                _useFileStatusPanel = false;
            }
        }
        catch(Exception e){
            getAppletContext().showStatus(e.getMessage());
        }
    }
    
    private String getParameter(String key, String defaultValue)
    {
        String value = getParameter(key);
        return (value != null) ? value : defaultValue;
    }
        
    private String[] split(String expr)
    {
        StringTokenizer st = new StringTokenizer(expr, ",");
        String[] array = new String[ st.countTokens() ];
        for(int i=0; i<array.length; i++){
            array[i] = st.nextToken();
        }
        return array;
    }
    
    //--------------------------------------------------------------------
    // Activate
    //--------------------------------------------------------------------
    private void setActive(boolean active)
    {
        _mouseBehavior.setEnabled(active);
        if(_fileStatusPanel != null) _fileStatusPanel.setOperatable(active);
        if(_commandArea != null && _commandArea.isActive()) _commandArea.setEditable(active);
        if(_animationFrame != null) _animationFrame.setEnabled(active);
    }

    private void setCommandActive(boolean enable)
    {
        _commandQueue.setActive(enable);
    }

    //--------------------------------------------------------------------
    // Mouse event
    //--------------------------------------------------------------------
    private void initializeMouseTarget()
    {
        StringTokenizer st = new StringTokenizer(_mouseSendTo, ",");
        int num = st.countTokens();
    
        _mouseTargetName = new String[num];
        _mouseTarget = new ViewerApplet[num];
        _mouseTargetObject = new Object[num];
    
        for(int i=0; i<num; i++){
            _mouseTargetName[i] = st.nextToken().trim();
        }
    }
    
    private void findMouseTarget(int index)
    {
        Applet app = getAppletContext().getApplet(_mouseTargetName[index]);
        if(app != null){
            if (app instanceof ViewerApplet) {
                _mouseTarget[index] = (ViewerApplet) app;
            } else if (app.getClass().getName().equals("org.pdbj.viewer.gui.ViewerApplet")) {
                _mouseTargetObject[index] = app;
            } else if (Util.existClass("org.jdesktop.applet.util.JNLPAppletLauncher")) {
                if (app instanceof JNLPAppletLauncher) {
                    _mouseTarget[index] = (ViewerApplet) ((JNLPAppletLauncher) app).getSubApplet();
                } else if (app.getClass().getName().equals("org.jdesktop.applet.util.JNLPAppletLauncher")) {
                    try {
                        _mouseTargetObject[index] = Util.invokeMethod(app, "getSubApplet", null, null);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
            }
        }
    }

    private void callbackMousePressed(MouseEvent e)
    {
        for(int i=0; i<_mouseTarget.length; i++)
        {
            if(_mouseTarget[i] == null && _mouseTargetObject[i] == null){
                this.findMouseTarget(i);
            }
            if(_mouseTarget[i] != null){
                _mouseTarget[i].receiveCanvasMousePressed(e);
            } else if (_mouseTargetObject[i] != null) {
                try {
                    Util.invokeMethod(_mouseTargetObject[i], "receiveCanvasMousePressed", new Class[]{MouseEvent.class}, new Object[]{e});
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
            }
        }
    }
    
    private void callbackMouseDragged(MouseEvent e)
    {
        for(int i=0; i<_mouseTarget.length; i++)
        {
            if(_mouseTarget[i] == null && _mouseTargetObject[i] == null){
                this.findMouseTarget(i);
            }
            if(_mouseTarget[i] != null){
                _mouseTarget[i].receiveCanvasMouseDragged(e);
            } else if (_mouseTargetObject[i] != null) {
                try {
                    Util.invokeMethod(_mouseTargetObject[i], "receiveCanvasMouseDragged", new Class[]{MouseEvent.class}, new Object[]{e});
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
            }
        }
    }

    //--------------------------------------------------------------------
    // Call JavaScript
    //--------------------------------------------------------------------
    private void onStart() {
        prepareJSWindow();
        if (!_jsFailed && _window != null) {
            Object[] args = new Object[] { _name };
            try {
                _window.call("jvOnStart", args);
            } catch (JSException e) {
                this.handleRecoverableException(new RecoverableException(e.getMessage()));
                _jsFailed = true;
            }
        }
    }

    private void prepareJSWindow() {
        if (!_jsFailed && _window == null) {
            try {
                _window = JSObject.getWindow(this);
            } catch (JSException e) {
                this.handleRecoverableException(new RecoverableException(e.getMessage()));
                _jsFailed = true;
            }
        }
    }

    //--------------------------------------------------------------------
    // Load files
    //--------------------------------------------------------------------
    private void addFile(FileType fileType, String fileURL, int fitTo)
    {
        URL url = this.toURL(fileURL);
        if(url != null){
            FileStatus copyTo = this.findFile(fitTo);
            _fileLoadTasks.add( new FileLoadTask(fileType, new FileLocator(url), copyTo) );
            _commandQueue.putCommand("", new CommandType(CommandType.APPLET_LOAD_FILE));
        }
    }

    private void addFilePdbCode(String pdbCode, int fitTo)
    {
        FileStatus copyTo = this.findFile(fitTo);
        _fileLoadTasks.add( new FileLoadTask(FileType.FTP_PDBML, new FileLocator(pdbCode), copyTo) );
        _commandQueue.putCommand("", new CommandType(CommandType.APPLET_LOAD_FILE));
    }

    private void setupFiles()
    {
        for(int i=0; i<_pdbIds.length; i++)
        {
            this.addFilePdbCode(_pdbIds[i], -1);
        }
        for(int i=0; i<_pdbmlURLs.length; i++)
        {
            this.addFile(FileType.PDBML, _pdbmlURLs[i], -1);
        }
        for(int i=0; i<_pdbURLs.length; i++)
        {
            this.addFile(FileType.PDB, _pdbURLs[i], -1);
        }
        for(int i=0; i<_polygonURLs.length; i++)
        {
            this.addFile(FileType.POLYGON, _polygonURLs[i], -1);
        }
        for(int i=0; i<_polygonCgoURLs.length; i++)
        {
            this.addFile(FileType.POLYGON_CGO, _polygonCgoURLs[i], -1);
        }
        for(int i=0; i<_animURLs.length; i++)
        {
            this.addFile(FileType.ANIMATION, _animURLs[i], -1);
        }
    }

    private void loadFile(FileLoadTask task)
    {
        _fileLoading = true;
        //this.setCommandActive(false);
        this.setActive(false);

        MessageDialog messageDialog = null;
        if (_fileLoadMessage) {
            messageDialog = _messageDialog;
            messageDialog.setWidth(300);
        }

        PDBmlUrl pdbmlUrl = _signed? _controller.getPDBmlUrl() : null;

        ReaderThread thread = new ReaderThread(this, _controller);
        thread.setupFile(task.fileType, task.locator, false, false, messageDialog, task.copyTo, pdbmlUrl);
        thread.start();
    }
    
    private URL toURL(String path)
    {
        URL url = null;
        try{
            url = new URL(path);
        }
        catch(MalformedURLException e){
            String message = ErrorMessages.INVALID_URL + "\n\n" + path;
            handleRecoverableException( new RecoverableException(message) );
        }
        return url;
    }
    
    private FileStatus findFile(int id)
    {
        ArrayList list = _controller.getFileStatusList();
        for(int i=0; i<list.size(); i++){
            FileStatus status = (FileStatus)list.get(i);
            if(status.getId() == id) return status;
        }
        return null;
    }

    //--------------------------------------------------------------------
    // Save files
    //--------------------------------------------------------------------

    private void saveFile(FileType fileType, File file, float ratio, float quality)
    {
        if(_controller.getFileStatusList().size() == 0) return;

        if( file.isDirectory() ){
            this.handleRecoverableException(
                new RecoverableException("Directory is Specified.")
            );
            return;
        }
    
        try{
            if(!_controller.rendererRunning()){
                _controller.startRenderer();
                _controller.stopRenderer();
            }

            BufferedImage image = _controller.getImage();
    
            if(fileType == FileType.PNG){
                ImageFileWriter.writePNG(file, image);
            }
            else if(fileType == FileType.JPEG){
                ImageFileWriter.writeJPEG(file, image, quality);
            }
        }
        catch(RecoverableException e){
            this.handleRecoverableException(e);
        }
        catch(Throwable t){
            this.handleFatalException( ExceptionUtil.handleThrowable(t) );
        }
    }

    //--------------------------------------------------------------------
    // File chooser
    //--------------------------------------------------------------------
    private File chooseFile(FileType fileType, boolean save)
    {
        if (_currDir == null) {
            _currDir = new File("./");
        }
        JFileChooser jFileChooser = new JFileChooser(_currDir);
        jFileChooser.setMultiSelectionEnabled(false);
        if( !fileType.acceptAll() ) jFileChooser.setFileFilter(fileType);

        int option;
        if(save){
            option = jFileChooser.showSaveDialog(this);
        }
        else{
            option = jFileChooser.showOpenDialog(this);
        }

        if(option == JFileChooser.APPROVE_OPTION)
        {
            File file = jFileChooser.getSelectedFile();
            _currDir = file.getParentFile();
            if(save){
                FileFilter currFilter = jFileChooser.getFileFilter();
                if(currFilter instanceof FileType){
                    FileType type = (FileType)currFilter;
                    file = type.fixExtension(file);
                }
            }
            return file;
        }
        return null;
    }

    //--------------------------------------------------------------------
    // Execute multiple commands
    //--------------------------------------------------------------------
    private void executeMultipleCommands(String commands)
    {
        StringTokenizer st = new StringTokenizer(commands, ";");
        if(st.countTokens() == 0) return;

        boolean running = _controller.rendererRunning();        
        if(running) _controller.stopRenderer();
    
        String command = st.nextToken();
        while( st.hasMoreTokens() ){
            _commandQueue.putCommand(command);
            command = st.nextToken();
        }
        if(running){
            CommandType type = new CommandType(CommandType.START_RENDERER);
            _commandQueue.putCommand(command, type);
        }
        else _commandQueue.putCommand(command);
    }

    //--------------------------------------------------------------------
    // Message
    //--------------------------------------------------------------------
    private void showMessage(String message)
    {
        if(_commandArea != null){
            _commandArea.appendMessage(message);
        }
    }

    //--------------------------------------------------------------------
    // jV command
    //--------------------------------------------------------------------
    private void startJvCommand() {
        if (_jvCommandController.hasNext()) {
            _jvCommandMode = true;
            if (_commandArea != null) {
                _commandArea.setEditable(false);
            }
            if (_fileStatusPanel != null) {
                _fileStatusPanel.setOperatable(false);
            }
            _controller.stopRenderer();
            _commandQueue.nextJvCommand();
        }
    }

    private void endJvCommand() {
        _controller.startRenderer();
        if (_fileStatusPanel != null) {
            _fileStatusPanel.setOperatable(true);
        }
        _jvCommandMode = false;            
        if (_commandArea != null) {
            _commandArea.setActive(true);
        }
    }

    private void handleJvCommand() {
        if (_jvCommandController.hasNext()) {
            _commandQueue.nextJvCommand();
        } else {
            endJvCommand();
        }
    }

    //--------------------------------------------------------------------
    // Create component
    //--------------------------------------------------------------------
    private void createComponent()
    {
        _viewerLayout = new ViewerLayoutSplitPane();
        _viewerLayout.setCanvasL( _controller.getCanvasL() );

        if(_useFileStatusPanel)
        {
            _viewerLayout.setFileStatusPanel(_fileStatusPanel);
            _viewerLayout.setViewerStatusPanel(_viewerStatusPanel);
        }
        if(_useCommandArea)
        {
            _viewerLayout.setCommandArea(_commandArea);
        }

        _viewerLayout.create( getContentPane() );
        //_viewerLayout.initialize( getContentPane() );

        // Mouse listener to canvas
        _controller.getCanvasL().addMouseListener(
            new MouseAdapter(){
                public void mousePressed(MouseEvent e){
                    canvas_mousePressed(e);
                }
                public void mouseReleased(MouseEvent e){
                    canvasL_mouseReleased(e);
                }
            }
        );
    }

    private void createCommandArea()
    {
        // to avoid a trouble in JRE 1.4.2 for Windows
        int cnt = 0;
        while(_commandArea == null && cnt < 10)
        {
            try
            {
                _commandArea = new CommandArea(this);
            }
            catch(ClassCastException e)
            {
                _commandArea = null;
                try
                {
                    Thread.sleep(200);
                }
                catch(InterruptedException ex)
                {
                }
            }
            cnt++;
        }
    }

    private void canvas_mousePressed(MouseEvent e)
    {
        if(_usePopupMenu && MouseBehavior.maskButton3(e))
        {
            _ticks = System.currentTimeMillis();
        }
    }
    
    private void canvasL_mouseReleased(MouseEvent e)
    {
        if(_usePopupMenu)
        {
            if( System.currentTimeMillis() - _ticks < CLICK_INTERVAL)
            {
                _menu.updateOptionsMenu();
                _menu.updateFileStatus();
                _popupMenu.show(_controller.getCanvasL(), e.getX(), e.getY());
            }
        }
    }

    private void canvasR_mouseReleased(MouseEvent e)
    {
        if(_usePopupMenu)
        {
            if( System.currentTimeMillis() - _ticks < CLICK_INTERVAL)
            {
                _menu.updateOptionsMenu();
                _menu.updateFileStatus();
                _popupMenu.show(_controller.getCanvasR(), e.getX(), e.getY());
            }
        }
    }

    //--------------------------------------------------------------------
    // Creates menu
    //--------------------------------------------------------------------
    private void createMenu()
    {
        _menu = new ViewerMenu(this, this, false);
        
        _popupMenu = new PopupMenu();
        
        _popupMenu.add( _menu.getFileMenu() );
        _popupMenu.add( _menu.getDisplayMenu() );
        _popupMenu.add( _menu.getColorsMenu() );
        _popupMenu.add( _menu.getOptionsMenu() );
        _popupMenu.add( _menu.getHelpMenu() );

        this.add(_popupMenu);
    }

    //--------------------------------------------------------------------
    // error screen
    //--------------------------------------------------------------------
    private void createErrorComponent(String message) {
        JLabel label = new JLabel(message);
        Container container = getContentPane();
        container.setLayout(new BorderLayout());
        container.add(label, BorderLayout.CENTER);
    }
}
