/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.media.opengl.GLProfile;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.plus.*;
import org.pdbj.viewer.renderer.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.mouse.*;
import org.pdbj.viewer.commands.*;
import org.pdbj.viewer.writers.*;

/**
 * Stand-alone application frame.
 */
public class ViewerFrame extends JFrame
    implements CommandHandler, FileStatusListener, GuiWrapper, 
    RendererListener, MousePickListener
{
    static {
        GLProfile.initSingleton();
    }

    private static final long serialVersionUID = 1L;

    private static final int DEFAULT_WIDTH  = 850;
    private static final int DEFAULT_HEIGHT = 800;

    private FileStatusPanel _fileStatusPanel;
    private ViewerStatusPanel _viewerStatusPanel;
    private CommandArea     _commandArea;

    private ViewerController _controller;
    private MouseBehavior _mouseBehavior;

    private CommandQueue  _commandQueue;
    private CommandThread _commandThread;

    private CommandExecutor  _executor;
    private ScriptController _scriptController;
    private JvCommandController _jvCommandController;

    private AnimationHandler _animHandler;
    private AnimationFrame _animationFrame;

    private MessageDialog _messageDialog;
    private UrlDialog _urlDialog;

    private StdinCommandQueue _stdinCommandQueue;
    private StdinCommandThread _stdinCommandThread;
    private StdinReaderThread _stdinReaderThread;

    // flags
    private boolean _scriptMode;
    private boolean _jvCommandMode;
    private boolean _screenInitialized;
    private boolean _fileLoading;
    private boolean _stdin;
    private boolean _mac;
    private boolean _stereoInitialized;

    // file chooser directory
    private File _currDir;

    // components
    private ViewerLayout _viewerLayout;

    // menus
    private ViewerMenu _menu;

    // command line arguments
    private String[] _initCommands;

    // constants
    private static final String PROPERTY_FILE = "properties.txt";
    private static final String ONE_MOLECULE = "Only one molecule should be selected.";

    /**
     * Returns whether the graphics is ready.
     */
    public boolean graphicsReady()
    {
        return _controller != null && _controller.graphicsReady();
    }

    //--------------------------------------------------------------------
    // Start up & Dispose
    //--------------------------------------------------------------------
    /**
     * Constructor.
     */
    public ViewerFrame(String[] initCommands, boolean stdin)
    {
        _initCommands = initCommands;
        _stdin = stdin;

        try
        {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
            SwingUtilities.updateComponentTreeUI(this);
        }
        catch(Exception e)
        {
        }

        _mac = (System.getProperty("mrj.version") != null);

        // WindowListner
        this.addWindowListener(
            new WindowAdapter(){
                public void windowClosing(WindowEvent e){
                    exit();
                }
                public void windowOpened(WindowEvent e){
                    init();
                }
            }
        );        

        // size & title
        this.setSize( new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT) );
        this.setTitle(GuiConstants.NAME);

        ImageIcon icon = new ImageIcon(getClass().getClassLoader().getResource(GuiConstants.ICON_IMAGE));
        this.setIconImage(icon.getImage());

        try {
            _controller = new ViewerController(this, false);
        } catch (Throwable t) {
            t.printStackTrace();
            return;
        }

        _controller.stopRenderer();
        _controller.addFileStatusListener(this);

        _fileStatusPanel = new FileStatusPanel(_controller);
        _fileStatusPanel.setGuiWrapper(this);
        _fileStatusPanel.setCommandHandler(this);
        
        _viewerStatusPanel = new ViewerStatusPanel();
        _controller.getViewStatus().addViewStatusListener(_viewerStatusPanel);

        _commandArea = new CommandArea(this);
        _commandArea.setCommandHandler(this);

        _mouseBehavior = new MouseBehavior(_controller, this);
        _mouseBehavior.setMessageArea(_commandArea);
        _mouseBehavior.addMouseSettingsListener(_viewerStatusPanel);

        _animHandler = new AnimationHandler(_controller);

        _executor = new CommandExecutor(
            _controller, _mouseBehavior, this, _animHandler
        );
        _executor.setMessageArea(_commandArea);

        _scriptController = new ScriptController();
        _jvCommandController = new JvCommandController();

        _messageDialog = new MessageDialog(this, "Progress");
        
        _commandQueue = new CommandQueue();
        _commandQueue.setScriptController(_scriptController);
        _commandQueue.setJvCommandController(_jvCommandController);
        _commandThread = new CommandThread(this, _commandQueue, _executor);

        if (_stdin) {
            _stdinCommandQueue = new StdinCommandQueue();
            _stdinCommandThread = new StdinCommandThread(this, _stdinCommandQueue);
            _stdinReaderThread = new StdinReaderThread(_stdinCommandQueue);
        }

        this.createMenu();
        this.createComponent();

        _menu.updateOptionsMenu();
        _viewerStatusPanel.updateAll(_controller.getViewStatus(), _mouseBehavior);

        boolean readProp = true;
        if (readProp) {
            this.readProperties();
        } else {
            _currDir = getApplicationDir();
        }

        if (_currDir == null) {
            _currDir = new File("./");
        }
    }
    
    private void init()
    {
        _viewerLayout.initialize( getContentPane() );
        
        _commandArea.setActive(true);
        _commandArea.requestFocus();

        setCommandActive(true);

        _commandThread.start();

        if (_stdin) {
            _stdinCommandThread.start();
        }
    }

    private void initViewer()
    {
        _commandQueue.putCommand("", new CommandType(CommandType.START_RENDERER));

        for(int i=0; i<_initCommands.length; i++){
            executeSingleCommand(_initCommands[i]);
        }

        if (_stdin) {
            _stdinReaderThread.start();
        }
    }

    /**
    * Dispose.
    */
    public void dispose()
    {
        if (_controller != null)
        {
            _animHandler.stopAll();

            _commandThread.terminate();
            try{
                _commandThread.join();
            }
            catch(InterruptedException e){
                e.printStackTrace();
            }

            if (_stdin) {
                _stdinCommandThread.terminate();
                try{
                    _stdinCommandThread.join();
                }
                catch(InterruptedException e){
                    e.printStackTrace();
                }
            }

            if(_animationFrame != null){
                _animationFrame.dispose();
            }

            _controller.dispose();
        }

        super.dispose();
    }

    //--------------------------------------------------------------------
    // Implements RendererListener
    //--------------------------------------------------------------------
    /**
    * Called when initialized.
    */
    public void screenInitialized()
    {
    }

    /**
    * Called when drawed.
    */
    public void screenDrawed()
    {
        if(!_screenInitialized){
            this.initViewer();
            _screenInitialized = true;
        }
    }

    /**
    * Called when resized.
    */
    public void screenResized(int x, int y, int width, int height)
    {
    }

    //--------------------------------------------------------------------
    // Implements CommandHandler
    //--------------------------------------------------------------------
    /**
     * Executes command.
     */
    public void executeSingleCommand(String command)
    {
        _commandQueue.putCommand(command);
    }

    /**
     * Complements command.
     */
    public CommandComplement complementCommand(String command)
    {
        return _executor.complement(command);
    }

    //--------------------------------------------------------------------
    // Implements FileStatusListener
    //--------------------------------------------------------------------
    /**
     * Receives addition of files.
     */
    public void fileAdded(ArrayList fileStatusList)
    {
        _menu.updateFileCloseMenu();
    }

    /**
     * Receives remove of files.
     */
    public void fileRemoved(ArrayList fileStatusList)
    {
        _menu.updateFileCloseMenu();
    }

    /**
     * Receives a change of status.
     */
    public void statusChanged(ArrayList fileStatusList)
    {
    }

    //--------------------------------------------------------------------
    // Implements GuiWrapper
    //--------------------------------------------------------------------
    /**
     * Returns whether or not the script mode.
     */
    public boolean isScriptMode()
    {
        return _scriptMode;
    }

    /**
     * Loads PDB file.
     */
    public void loadPDBFile(String filePath, FileStatus copy)
    {
        this.loadFile(FileType.PDB, FileLocator.fromPath(filePath), !_scriptMode, copy);
    }

    /**
     * Loads PDBML file.
     */
    public void loadPDBmlFile(String filePath, FileStatus copy)
    {
        this.loadFile(FileType.PDBML, FileLocator.fromPath(filePath), !_scriptMode, copy);
    }

    /**
     * Loads polygon file.
     */
    public void loadPolygonFile(String filePath, FileStatus copy)
    {
        this.loadFile(FileType.POLYGON, FileLocator.fromPath(filePath), !_scriptMode, copy);
    }

    /**
     * Loads polygon CGO file.
     */
    public void loadPolygonCgoFile(String filePath, FileStatus copy)
    {
        this.loadFile(FileType.POLYGON_CGO, FileLocator.fromPath(filePath), !_scriptMode, copy);
    }

    /**
     * Loads script file.
     */
    public void loadScriptFile(String filePath)
    {
        if(!_scriptMode){
            this.loadFile(FileType.SCRIPT, FileLocator.fromPath(filePath), true, null);
        }
    }

    /**
     * Loads animation file.
     */
    public void loadAnimationFile(String filePath, FileStatus copy)
    {
        this.loadFile(FileType.ANIMATION, FileLocator.fromPath(filePath), !_scriptMode, copy);
    }

    /**
     * Loads PDBML file by FTP.
     */
    public void loadPDBmlFileFtp(String pdbCode, FileStatus copy)
    {
        this.loadFile(FileType.FTP_PDBML, new FileLocator(pdbCode), !_scriptMode, copy);
    }

    /**
     * Loads eF-site data files.
     */
    public void loadEfsite(String efsiteId)
    {
        if (efsiteId.length() >= 4) {
            String eFsiteCode = efsiteId.substring(0, 4).toLowerCase() + efsiteId.substring(4);
            String dir = eFsiteCode.substring(1,3);
            String efsitePath = _controller.getEfsiteUrl().getUrl();
            String pdbPath = efsitePath + "pdbview/" + dir + "/" + eFsiteCode + ".pdb";
            String efvetPath = efsitePath + "efvetview/" + dir + "/" + eFsiteCode + ".efview.gz";
            EfsiteData efsiteData = new EfsiteData();
            this.loadFile(FileType.PDB, FileLocator.fromPath(pdbPath), !_scriptMode, null, efsiteData);
            this.loadFile(FileType.POLYGON, FileLocator.fromPath(efvetPath), !_scriptMode, null, efsiteData);
        }
    }

    /**
     * Saves pdb file.
     */
    public void savePDBFile(String filePath)
    {
        if(this.getSelectedMoleculeFiles().size() > 1){
            _commandArea.appendMessage(ONE_MOLECULE);
            return;
        }
        this.saveFile(FileType.PDB, new File(filePath), 0, 0);
    }

    /**
     * Saves script file.
     */
    public void saveScriptFile(String filePath)
    {
        this.saveFile(FileType.SCRIPT, new File(filePath), 0, 0);
    }

    /**
     * Saves PNG file.
     */
    public void savePngFile(String filePath, float ratio)
    {
        this.saveFile(FileType.PNG, new File(filePath), ratio, 0);
    }

    /**
     * Saves JPEG file.
     */
    public void saveJpegFile(String filePath, float ratio, float quality)
    {
        this.saveFile(FileType.JPEG, new File(filePath), ratio, quality);
    }

    /**
     * Pause execution of script.
     */
    public void pauseScript()
    {
        if(_scriptMode) _scriptController.setPausing(true);
    }

    /**
     * Sets whether the stereo mode.
     */
    public void setupStereo()
    {
        _viewerLayout.setStereo( _controller.getViewStatus().isStereo() );
    }

    /**
     * Terminates the application.
     */
    public void exit()
    {
        this.dispose();
        System.exit(0);
    }

    /**
     * Returns the FTP message listener.
     */
    public MessageListener getFtpMessageListener()
    {
        return _messageDialog;
    }

    /**
     * Returns whether or not the standalone application.
     */
    public boolean isStandalone()
    {
        return true;
    }

    /**
     * Loads function file by FTP.
     */
    public void loadFunctionFtp(FileStatus file)
    {
        if(file.getType() == FileStatus.POLYGON) return;

        FileType fileType = FileType.FUNCTION;
        FileLocator locator = new FileLocator(file.getMoleculeFile().getMoleculeId());

        _messageDialog.setWidth(480);
        
        ReaderThread thread = new ReaderThread(this, _controller);
        thread.setupFile(fileType, locator, false, false, _messageDialog, null,
                _controller.getPDBmlUrl());
        thread.setMoleculeFile(file.getMoleculeFile());
        thread.start();
        
        try{
            thread.join();
        }
        catch(InterruptedException e){
        }
    }

    /**
     * Returns the site URL map.
     */
    public HashMap getSiteUrlMap() {
        return _controller.getSiteUrlMap().getMap();
    }

    /**
     * Loads the site list.
     */
    public void loadSiteList(FileStatus file, String key) {
        if (file.getType() == FileStatus.POLYGON) {
            return;
        }

        String pdbcode = file.getMoleculeFile().getMoleculeId();
        URL url = createURL(pdbcode, key);
        if (url == null) {
            return;
        }

        FileType fileType = FileType.SITELIST;
        FileLocator locator = new FileLocator(url);
        
        _messageDialog.setWidth(480);
        
        ReaderThread thread = new ReaderThread(this, _controller);
        thread.setupFile(fileType, locator, false, false, _messageDialog, null,
                null);
        thread.setMoleculeFile(file.getMoleculeFile());
        thread.setSiteListKey(key);
        thread.start();
        
        try {
            thread.join();
        } catch (InterruptedException e) {
        }
    }
    
    private URL createURL(String pdbcode, String key) {
        String code = pdbcode.toLowerCase();
        String path = _controller.getSiteUrlMap().getUrl(key);
        if (path == null) {
            return null;
        }
        path = path.replaceAll("\\{0\\}", code.substring(0, 1));
        path = path.replaceAll("\\{1\\}", code.substring(1, 2));
        path = path.replaceAll("\\{2\\}", code.substring(2, 3));
        path = path.replaceAll("\\{3\\}", code.substring(3, 4));
        try {
            return new URL(path);
        } catch (MalformedURLException e) {
            String message = "Invalid URL: " + path;
            handleRecoverableException(new RecoverableException(message));
            return null;
        }
    }

    /**
     * Returns the size of the 3D-rendering panel.
     */
    public Dimension getImageSize() {
        return _viewerLayout.getImageSize();
    }

    /**
     * Requests the size of the 3D-rendering panel and 
     * returns the size actually set.
     */
    public Dimension setImageSize(Dimension d) {
        int w = d.width;
        int h = d.height;
        if (w % 2 != 0) {
            w -= 1;
        }
        if (w < 10) {
            w = 10;
        }
        if (h < 10) {
            h = 10;
        }

        Dimension imageSize = _viewerLayout.getImageSize();
        int iw = imageSize.width;
        int ih = imageSize.height;

        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        int sw = screenSize.width;
        int sh = screenSize.height;

        Dimension frameSize = getSize();
        int fw = frameSize.width;
        int fh = frameSize.height;
        
        int dw = w - imageSize.width;
        int dh = h - imageSize.height;
        if (fw + dw > sw) {
            dw = sw - fw;
        }
        if (fh + dh > sh) {
            dh = sh - fh;
        }

        setSize(fw + dw, fh + dh);
        setVisible(true);

        return new Dimension(iw + dw, ih + dh);
    }

    /**
     * PDBj execute.
     */
    public void pdbjExecute(int fileId, String name) {
        FileStatus file = _controller.findFile(fileId);
        if (file != null && file.getType() != FileStatus.POLYGON) {
            MoleculeFile moleculeFile = file.getMoleculeFile();
            if (!moleculeFile.isAnimation()) {
                Function function = moleculeFile.getFunction();
                _jvCommandController.start(function, name, fileId);
                startJvCommand();
            }
        }
    }

    /**
     * Initialize stereo.
     */
    public void initializeStereo()
    {
        if (!_stereoInitialized) {
            _controller.initializeStereo();
            _mouseBehavior.initializeStereo();
            _viewerLayout.setCanvasR( _controller.getCanvasR() );
            _viewerLayout.initializeStereo();
            _stereoInitialized = true;
        }
    }

    /**
     * Returns whether NEWT is used.
     */
    public boolean isNewt()
    {
        return false;
    }

    /**
     * Returns a viewer controller.
     */
    public ViewerController getViewerController()
    {
        return _controller;
    }

    /**
     * Returns a mouse behavior.
     */
    public MouseBehavior getMouseBehavior()
    {
        return _mouseBehavior;
    }

    /**
     * Loads a local file.
     */
    public void loadLocalFile(FileType fileType)
    {
        File file = this.chooseFile(fileType, false);
        if(file != null){
            this.loadFile(fileType, new FileLocator(file), true, null);
        }
    }

    /**
     * Loads a remote file.
     */
    public void loadRemoteFile(FileType fileType)
    {
        if(_urlDialog == null){
            _urlDialog = new UrlDialog(this, "", false);
            _urlDialog.setGuiWrapper(this);
            _urlDialog.setCenterTo(this);
        }
        _urlDialog.setFileType(fileType);
        String name = "";
        if(fileType == FileType.PDB)              name = "PDB";
        else if(fileType == FileType.PDBML)       name = "PDBML";
        else if(fileType == FileType.POLYGON)     name = "Polygon";
        else if(fileType == FileType.POLYGON_CGO) name = "Polygon CGO";
        else if(fileType == FileType.ANIMATION)   name = "Animation";
        else if(fileType == FileType.SCRIPT)      name = "Script";
        _urlDialog.setTitle("Open - Remote " + name);
        _urlDialog.setVisible(true);
    }

    /**
     * Loads a remote file.
     */
    public void loadRemoteFile(UrlDialog urlDialog)
    {
        boolean TO_HISTORY = true;

        try{
            String value = urlDialog.getUrl("UTF-8");
            FileType fileType = urlDialog.getFileType();
            int fitTo = urlDialog.getFitTo();
            if(value != null){
                URL url = new URL(value);
                FileLocator locator = new FileLocator(url);
                FileStatus copy = null;
                if(fitTo != -1) copy = _controller.findFile(fitTo);
                this.loadFile(fileType, locator, TO_HISTORY, copy);                    
            }
        }
        catch(UnsupportedEncodingException e){
            this.handleRecoverableException(
                new RecoverableException( e.getMessage() )
            );
        }
        catch(MalformedURLException e){
            this.handleRecoverableException(
                new RecoverableException(ErrorMessages.INVALID_URL)
            );
        }
    }

    /**
     * Loads files specified by PDB ID.
     */
    public void loadByPDBID()
    {
        boolean TO_HISTORY = true;

        InputDialog dialog = new InputDialog(
            this, "PDB ID", "Enter PDB ID", 60
        );
        String pdbCode = dialog.show("");
        if(pdbCode != null){
            pdbCode = pdbCode.toLowerCase();
            if( this.checkPDB(pdbCode) ){
                FileLocator locator = new FileLocator(pdbCode);
                this.loadFile(FileType.FTP_PDBML, locator, TO_HISTORY, null);
            }
            else{
                this.handleRecoverableException(
                    new RecoverableException("Invalid Value.")
                );
            }
        }
    }

    /**
     * Loads files specified by eF-site ID.
     */
    public void loadByEFsiteID()
    {
        boolean TO_HISTORY = true;

        InputDialog dialog = new InputDialog(
            this, "eF-site ID", "Enter eF-site ID", 180
        );
        String eFsiteCode = dialog.show("");
        if(eFsiteCode != null)
        {
            if(eFsiteCode.length() >= 4)
            {
                eFsiteCode = eFsiteCode.substring(0,4).toLowerCase() + eFsiteCode.substring(4);
                String dir = eFsiteCode.substring(1,3);
                String efsitePath = _controller.getEfsiteUrl().getUrl();
                String pdbPath = efsitePath + "pdbview/" + dir + "/" + eFsiteCode + ".pdb";
                String efvetPath = efsitePath + "efvetview/" + dir + "/" + eFsiteCode + ".efview.gz";
                EfsiteData efsiteData = new EfsiteData();
                this.loadFile(FileType.PDB, FileLocator.fromPath(pdbPath), TO_HISTORY, null, efsiteData);
                this.loadFile(FileType.POLYGON, FileLocator.fromPath(efvetPath), TO_HISTORY, null, efsiteData);
            }
            else
            {
                this.handleRecoverableException(
                    new RecoverableException("Invalid Value.")
                );
            }
        }
    }

    /**
     * Loads a file registered in the history.
     */
    public void loadHistory(FileType fileType, FileLocator locator)
    {
        this.loadFile(fileType, locator, false, null);
    }

    /**
     * Appends a file history.
     */
    public void addFileHistory(FileType type, FileLocator locator)
    {
        _menu.addFileHistory(type, locator);
    }

    /**
     * Saves a file.
     */
    public void saveFile(FileType fileType)
    {
        if(_controller.getFileStatusList().size() == 0) return;
        
        if( fileType == FileType.PDB
            && this.getSelectedMoleculeFiles().size() != 1 )
        {
            this.handleRecoverableException(
                new RecoverableException(ONE_MOLECULE)
            );
            return;
        }
        
        File file = this.chooseFile(fileType, true);
        if(file != null){
            this.saveFile(fileType, file, 1, 0.75f);
        }
    }

    /**
     * Shows information of files.
     */
    public void showInformation()
    {
        String header = GuiConstants.NAME + " " + GuiConstants.VERSION + "\n\n";
        
        String[] infos = _controller.getInformation();
        for(int i=0; i<infos.length; i++)
        {
            JOptionPane.showMessageDialog(
                this, header + infos[i], "Information", JOptionPane.INFORMATION_MESSAGE,
                new ImageIcon(getClass().getClassLoader().getResource(GuiConstants.ICON_IMAGE))
            );
        }
    }

    /**
     * Shows an animation frame.
     */
    public void showAnimationFrame()
    {
        if(_animationFrame == null){
            _animationFrame = new AnimationFrame(_animHandler);
            _animationFrame.setCenterTo(this);
        }
        _animationFrame.setVisible(true);
    }

    /**
     * Shows an about dialog.
     */
    public void showAboutDialog()
    {
        StringBuffer sb = new StringBuffer();
        sb.append(GuiConstants.NAME + " " + GuiConstants.VERSION + "\n\n");
        sb.append(GuiConstants.CREDIT);

        JOptionPane.showMessageDialog(
            this, sb.toString(), "About jV", JOptionPane.INFORMATION_MESSAGE,
            new ImageIcon(getClass().getClassLoader().getResource(GuiConstants.ICON_IMAGE))
        );
    }

    /**
     * Opens the manual.
     */
    public void openManual() {
        try {
            Util.browse(new URI("http://ef-site.hgc.jp/wiki/jV/index.php/How_to_use"));
        } catch (Exception e) {
            handleRecoverableException(new RecoverableException(e.getMessage()));
        }
    }

    /**
     * Rotates stereo mode.
     */
    public void rotateStereo()
    {
        executeSingleCommand("stereo");
    }

    /**
     * Handles a recoverable exception.
     */
    public void handleRecoverableException(RecoverableException e)
    {
        JOptionPane.showMessageDialog(
            this, e.getMessage(), "Warning", JOptionPane.WARNING_MESSAGE
        );
    }

    /**
     * Handles a fatal exception.
     */
    public void handleFatalException(FatalException e)
    {
        JOptionPane.showMessageDialog(
            this, e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE
        );

        this.exit();
    }

    /**
     * Starts an execution of a command.
     */
    public void startCommand(String command, CommandType type)
    {
        this.startCommand(_executor.createTokenizer(command), type);
    }

    /**
     * Starts an execution of a command.
     */
    public void startCommand(CommandTokenizer tokenizer, CommandType type)
    {
        try{
            if(tokenizer != null) _executor.execute(tokenizer);
        }
        catch(RecoverableException e){
            this.handleRecoverableException(e);
        }
        catch(OutOfMemoryError e){
            this.handleFatalException( ExceptionUtil.handleOutOfMemoryError(e) );
        }
        catch(Throwable t){
            this.handleFatalException( ExceptionUtil.handleThrowable(t) );
        }

        if( type.action == CommandType.START_RENDERER ) _controller.startRenderer();

        if(!_fileLoading) this.endCommand();
    }

    /**
     * Ends an execution of a command.
     */
    public void endCommand()
    {
        if(_fileLoading){
            _fileLoading = false;
            this.setActive(true);
        }

        if (_scriptMode) {
            handleScript();
        } else if (_jvCommandMode) {
            handleJvCommand();
        } else {
            _commandArea.setActive(true);
        }

        _menu.updateOptionsMenu();

        this.setCommandActive(true);
    }

    /**
     * Returns whether menu is active.
     */
    public boolean menuActive()
    {
        return (!_fileLoading && !_scriptMode && !_jvCommandMode);
    }

    /**
     * Called when a key is typed to start script.
     */
    public void commandKeyTyped()
    {
        if( _scriptController.isPausing() ) this.startScript();
    }

    /**
     * Called when focus is moved.
     */
    public void focusMoved()
    {
        _commandArea.requestFocus();
    }

    //--------------------------------------------------------------------
    // Implements MousePickListener
    //--------------------------------------------------------------------
    /**
     * Receives mouse pick result.
     */
    public void receiveMousePick(PickResult pickResult)
    {
        if(_urlDialog != null && _urlDialog.isVisible()){
            _urlDialog.setPickResult(pickResult);
        }
    }

    //--------------------------------------------------------------------
    // Activate
    //--------------------------------------------------------------------
    private void setActive(boolean active)
    {
        _mouseBehavior.setEnabled(active);
        _fileStatusPanel.setOperatable(active);
        if( _commandArea.isActive() ) _commandArea.setEditable(active);
        if(_animationFrame != null) _animationFrame.setEnabled(active);
    }

    private void setCommandActive(boolean enable)
    {
        _commandQueue.setActive(enable);
        if (_stdin) {
            _stdinCommandQueue.setActive(enable);
        }
    }

    //--------------------------------------------------------------------
    // Load files
    //--------------------------------------------------------------------
    private void loadFile(FileType fileType, FileLocator locator, boolean toHistory, FileStatus copy)
    {
        loadFile(fileType, locator, toHistory, copy, null);
    }

    private void loadFile(FileType fileType, FileLocator locator,
            boolean toHistory, FileStatus copy, EfsiteData efsiteData)
    {
        // add history only for local file
        toHistory = (toHistory && locator.isLocalFile());
        
        if(fileType == FileType.SCRIPT)
        {
            this.loadScriptFile(locator, toHistory);
        }
        else{
            if(fileType == FileType.PDBML || fileType == FileType.FTP_PDBML)
            {
                _messageDialog.setWidth(480);
            }
            else if( fileType == FileType.PDB || fileType == FileType.POLYGON ||
                     fileType == FileType.ANIMATION )
            {
                _messageDialog.setWidth(400);
            }

            _fileLoading = true;
            this.setCommandActive(false);
            this.setActive(false);
            
            ReaderThread thread = new ReaderThread(this, _controller);
            thread.setupFile(fileType, locator, toHistory, true,
                    _messageDialog, copy, _controller.getPDBmlUrl());
            thread.setEfsiteData(efsiteData);
            thread.start();
        }
    }

    private void loadScriptFile(FileLocator locator, boolean toHistory)
    {
        try{
            if( locator.isLocalFile() ){
                _scriptController.load( locator.getFile() );
            }
            else if( locator.isRemoteFile() ){
                _scriptController.load( locator.getURL() );
            }
            this.startScript();
        }
        catch(RecoverableException e){
            this.handleRecoverableException(e);
        }
        catch(Throwable t){
            this.handleFatalException( ExceptionUtil.handleThrowable(t) );
        }

        if(toHistory) this.addFileHistory(FileType.SCRIPT, locator);
    }
    
    //--------------------------------------------------------------------
    // Save files
    //--------------------------------------------------------------------
    private void saveFile(FileType fileType, File file, float ratio, float quality)
    {
        if(_controller.getFileStatusList().size() == 0) return;

        if( file.isDirectory() ){
            this.handleRecoverableException(
                new RecoverableException("Directory is Specified.")
            );
            return;
        }
    
        try{
            if(fileType == FileType.PDB){
                PDBWriter pw = new PDBWriter(_controller);
                ArrayList list = this.getSelectedMoleculeFiles();
                FileStatus status = (FileStatus)list.get(0);
                pw.write(file, status.getId());
            }
            else if(fileType == FileType.SCRIPT){
                ScriptWriter sw = new ScriptWriter(_controller, _mouseBehavior);
                sw.write(file);
            }
            else{
                if(!_controller.rendererRunning()){
                    _controller.startRenderer();
                    _controller.stopRenderer();
                }

                BufferedImage image = _controller.getImage();
        
                if(fileType == FileType.PNG){
                    ImageFileWriter.writePNG(file, image);
                }
                else if(fileType == FileType.JPEG){
                    ImageFileWriter.writeJPEG(file, image, quality);
                }
            }
        }
        catch(RecoverableException e){
            this.handleRecoverableException(e);
        }
        catch(Throwable t){
            this.handleFatalException( ExceptionUtil.handleThrowable(t) );
        }
    }
    
    //--------------------------------------------------------------------
    // File chooser
    //--------------------------------------------------------------------
    private File chooseFile(FileType fileType, boolean save)
    {
        if (!_mac) {
            JFileChooser jFileChooser = new JFileChooser(_currDir);
            jFileChooser.setMultiSelectionEnabled(false);
            if( !fileType.acceptAll() ) jFileChooser.setFileFilter(fileType);

            int option;
            if(save){
                option = jFileChooser.showSaveDialog(this);
            }
            else{
                option = jFileChooser.showOpenDialog(this);
            }

            if(option == JFileChooser.APPROVE_OPTION)
            {
                File file = jFileChooser.getSelectedFile();
                _currDir = file.getParentFile();
                if(save){
                    FileFilter currFilter = jFileChooser.getFileFilter();
                    if(currFilter instanceof FileType){
                        FileType type = (FileType)currFilter;
                        file = type.fixExtension(file);
                    }
                }
                return file;
            }
            return null;

        } else {
            FileDialog fileDialog = new FileDialog(this);
            if (save) {
                fileDialog.setMode(FileDialog.SAVE);
                fileDialog.setTitle("Save " + fileType.getDisplayName());
            } else {
                fileDialog.setMode(FileDialog.LOAD);
                fileDialog.setTitle("Open " + fileType.getDisplayName());
            }
            fileDialog.setDirectory(_currDir.getAbsolutePath());
            //fileDialog.setFilenameFilter(fileType);
            fileDialog.setVisible(true);

            File file = null;
            if (fileDialog.getFile() != null) {
                file = new File(fileDialog.getDirectory(), fileDialog.getFile());
                if (file != null) {
                    _currDir = file.getParentFile();
                    if (save) {
                        file = fileType.fixExtension(file);
                    }
                }
            }
            return file;
        }
    }

    //--------------------------------------------------------------------
    // Script
    //--------------------------------------------------------------------
    private void startScript()
    {
        if( _scriptController.hasMoreCommand() )
        {
            _scriptMode = true;

            _commandArea.setEditable(false);
            _fileStatusPanel.setOperatable(false);

            _controller.stopRenderer();
            _scriptController.setPausing(false);

            _commandQueue.nextScript();
        }
        else if(_scriptMode){
            this.endScript();
        }
    }
    
    private void endScript()
    {
        _controller.startRenderer();
        _fileStatusPanel.setOperatable(true);
        _scriptMode = false;            
        _commandArea.setActive(true);
    }
    
    private void handleScript()
    {
        if( _scriptController.isPausing() )
        {
            _controller.startRenderer();
            _commandArea.setActive(false);
        }
        else if( !_scriptController.hasMoreCommand() )
        {
            this.endScript();
        }
        else{
            _commandQueue.nextScript();
        }
    }

    //--------------------------------------------------------------------
    // jV command
    //--------------------------------------------------------------------
    private void startJvCommand() {
        if (_jvCommandController.hasNext()) {
            _jvCommandMode = true;
            _commandArea.setEditable(false);
            _fileStatusPanel.setOperatable(false);
            _controller.stopRenderer();
            _commandQueue.nextJvCommand();
        }
    }

    private void endJvCommand() {
        _controller.startRenderer();
        _fileStatusPanel.setOperatable(true);
        _jvCommandMode = false;            
        _commandArea.setActive(true);
    }

    private void handleJvCommand() {
        if (_jvCommandController.hasNext()) {
            _commandQueue.nextJvCommand();
        } else {
            endJvCommand();
        }
    }

    //--------------------------------------------------------------------
    // Read properties
    //--------------------------------------------------------------------
    private void readProperties()
    {
        File file = this.getPropertyFile();
        if(file != null)
        {
            HashMap hashMap = new HashMap();
            try
            {
                BufferedReader reader = new BufferedReader( new FileReader(file) );
                String s = null;
                while( (s = reader.readLine()) != null)
                {
                    if( s.trim().startsWith("#") ) continue;
                    int index = s.indexOf("=");
                    if(index != -1)
                    {
                        String key = s.substring(0, index).toLowerCase();
                        if (key.equals("pdbj") || key.equals("xps3")) {
                            continue;
                        }
                        if (key.indexOf(":") != -1) {
                            continue;
                        }
                        String val = s.substring(index+1);
                        hashMap.put(key, val);
                    }
                }
                reader.close();
            }
            catch(IOException e)
            {
            }
            
            String urlNoatom = (String)hashMap.get("pdbml_noatom");
            String urlExtatom = (String)hashMap.get("pdbml_extatom");
            String urlFunction = (String)hashMap.get("pdbml_plus");
            if(urlNoatom != null && urlExtatom != null)
            {
                _controller.setPDBmlUrl(new PDBmlUrl(urlNoatom, urlExtatom, urlFunction));
            }
            
            String urlEfsite = (String)hashMap.get("efsite");
            if(urlEfsite != null)
            {
                _controller.setEfsiteUrl(new EfsiteUrl(urlEfsite));
            }

            hashMap.remove("pdbml_noatom");
            hashMap.remove("pdbml_extatom");
            hashMap.remove("pdbml_plus");
            hashMap.remove("efsite");

            _controller.setSiteUrlMap(new SiteUrlMap(hashMap));
        }
    }

    private File getPropertyFile()
    {
        File dir = getApplicationDir();
        if (dir != null) {
            _currDir = dir;
            return new File(dir, PROPERTY_FILE);
        } else {
            return null;
        }
    }

    private File getApplicationDir() {
        File dir = null;
        try {
            URL url = this.getClass().getClassLoader().getResource("org/pdbj/viewer");
            if (url != null) {
                String locator = url.toString();
                if (locator.startsWith("jar:")) {
                    locator = locator.substring(4);
                }
                for (int i = 0; i < 4; i++) {
                    int index = locator.lastIndexOf("/");
                    if (index != -1) {
                        locator = locator.substring(0, index);
                    }
                }
                dir = new File(new URI(locator));
            }
        } catch(Exception e) {
            dir = null;
        }
        return dir;
    }

    //--------------------------------------------------------------------
    // Create component
    //--------------------------------------------------------------------
    private void createComponent()
    {
        _viewerLayout = new ViewerLayoutSplitPane();
        _viewerLayout.setCanvasL( _controller.getCanvasL() );
        _viewerLayout.setFileStatusPanel(_fileStatusPanel);
        _viewerLayout.setViewerStatusPanel(_viewerStatusPanel);
        _viewerLayout.setCommandArea(_commandArea);        
        _viewerLayout.create( getContentPane() );
    }

    //--------------------------------------------------------------------
    // Creates menu
    //--------------------------------------------------------------------
    private void createMenu()
    {
        _menu = new ViewerMenu(this, this, true);
        
        MenuBar menuBar = new MenuBar();
        menuBar.add( _menu.getFileMenu() );
        menuBar.add( _menu.getDisplayMenu() );
        menuBar.add( _menu.getColorsMenu() );
        menuBar.add( _menu.getOptionsMenu() );
        menuBar.add( _menu.getHelpMenu() );

        setMenuBar(menuBar);
    }

    //--------------------------------------------------------------------
    // Util
    //--------------------------------------------------------------------
    private boolean checkPDB(String pdbCode)
    {
        if(pdbCode.length() != 4) return false;
        pdbCode = pdbCode.toLowerCase();
        for(int i=0; i<4; i++){
            char c = pdbCode.charAt(i);
            if( (c < '0' || c > '9' ) && ( c < 'a' || c > 'z') ) return false;
        }
        return true;
    }

    private ArrayList getSelectedMoleculeFiles()
    {
        ArrayList list = new ArrayList();

        ArrayList statusList = _controller.getFileStatusList();
        for(int i=0; i<statusList.size(); i++)
        {
            FileStatus status = (FileStatus)statusList.get(i);
            if( status.getType() == FileStatus.PDB || 
                status.getType() == FileStatus.PDBML )
            {
                if( status.isSelected() ) list.add(status);
            }
        }
        return list;
    }
}
