/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * GUI layout with JSplitPane.
 */
public class ViewerLayoutSplitPane extends ViewerLayout
{
    private static final int MIN_W = 10;
    private static final int MIN_H = 10;

    private Component  _componentMain;

    private JPanel     _panelCanvas;
    private JSplitPane _splitPane1;
    private JSplitPane _splitPane2;

    private int _toolWidth;
    private int _commandHeight;

    /**
     * Sets whether the stereo mode.
     */
    public void setStereo(boolean stereo)
    {
        if(stereo != isStereo())
        {
            super.setStereo(stereo);

            if (stereo) {
                updateCanvas();
                setupStereo(stereo);
            } else {
                setupStereo(stereo);
                updateCanvas();
            }
        }
    }

    /**
     * Creates the components.
     */
    public void create(Container container)
    {
        createComponent(container);
    }

    /**
     * Initializes the components.
     */
    public void initialize(Container container)
    {
        int width = container.getWidth();
        int height = container.getHeight();
        
        initializeComponent(width, height);
    }

    /**
     * Initialize stereo.
     */
    public void initializeStereo()
    {
        Canvas canvasR = getCanvasR();
        canvasR.setSize(1, 1);
        canvasR.setVisible(false);
        _panelCanvas.add(canvasR);

        final CommandArea commandArea = getCommandArea();
        if(commandArea != null)
        {
            canvasR.addMouseListener(new MouseAdapter() {
                public void mouseReleased(MouseEvent e) {
                    commandArea.requestFocus();
                }
            });
        }
    }
    
    //--------------------------------------------------------------------
    // Handle component with JSplitPane
    //--------------------------------------------------------------------    
    private void createComponent(Container container)
    {
        Canvas canvasL = getCanvasL();
        
        FileStatusPanel fileStatusPanel = getFileStatusPanel();
        ViewerStatusPanel viewerStatusPanel = getViewerStatusPanel();
        final CommandArea commandArea = getCommandArea();

        // give a finite size ( to avoid a trouble in JDK 1.4.2 )
        canvasL.setSize(1, 1);

        //canvasL.setFocusable(false);
        //canvasR.setFocusable(false);
        
        _panelCanvas = new JPanel();
        _panelCanvas.setLayout(null);
        _panelCanvas.add(canvasL);
        _panelCanvas.setMinimumSize( new Dimension(MIN_W, MIN_H) );

        Component component1 = null;

        if(fileStatusPanel != null)
        {
            JPanel statusPanel = new JPanel();
            statusPanel.setLayout(new BorderLayout());
            statusPanel.add(fileStatusPanel, BorderLayout.CENTER);
            statusPanel.add(viewerStatusPanel, BorderLayout.SOUTH);

            _splitPane1 = new JSplitPane(
                JSplitPane.HORIZONTAL_SPLIT, _panelCanvas, statusPanel
            );
            _splitPane1.setResizeWeight(1);
            _splitPane1.setDividerLocation(MIN_W);
            _splitPane1.setFocusable(false);
            component1 = _splitPane1;
        }
        else
        {
            component1 = _panelCanvas;
        }
        
        if(commandArea != null)
        {
            this.createSplitPane2(component1);
            _splitPane2.setResizeWeight(1);
            _splitPane2.setDividerLocation(MIN_H);
            _splitPane2.setFocusable(false);
            _componentMain = _splitPane2;

            canvasL.addMouseListener(new MouseAdapter() {
                public void mouseReleased(MouseEvent e) {
                    commandArea.requestFocus();
                }
            });
        }
        else 
        {
            _componentMain = component1;
        }

        _panelCanvas.addComponentListener(
            new ComponentAdapter(){
                public void componentResized(ComponentEvent e){
                    panelCanvas_componentResized(e);
                }
            }
        );

        container.setLayout( new BorderLayout() );
        container.add(_componentMain, BorderLayout.CENTER);        
    }

    private void createSplitPane2(Component component1)
    {
        CommandArea commandArea = getCommandArea();

        // to avoid a trouble in JRE 1.4.2 for Windows
        int cnt = 0;
        while(_splitPane2 == null && cnt < 10)
        {
            try
            {
                _splitPane2 = new JSplitPane(
                    JSplitPane.VERTICAL_SPLIT, component1, commandArea
                );
            }
            catch(ClassCastException e)
            {
                _splitPane2 = null;
                try
                {
                    Thread.sleep(200);
                }
                catch(InterruptedException ex)
                {
                }
            }
            cnt++;
        }
    }

    private void panelCanvas_componentResized(ComponentEvent e)
    {
        if(_panelCanvas.getWidth()%2 != 0)
        {
            _panelCanvas.setSize(_panelCanvas.getWidth()-1, _panelCanvas.getHeight());
        }
        
        updateCanvas();
    }
    
    private void initializeComponent(int width, int height)
    {
        FileStatusPanel fileStatusPanel = getFileStatusPanel();
        CommandArea commandArea = getCommandArea();

        _toolWidth = (int)(0.25*width);
        _commandHeight = (int)(0.2*height);

        if(_toolWidth > TOOL_WIDTH)
        {
            _toolWidth = TOOL_WIDTH;
        }
        if(_commandHeight > COMMAND_HEIGHT)
        {
            _commandHeight = COMMAND_HEIGHT;        
        }

        int w2 = width - _toolWidth;
        int h2 = height - _commandHeight;
        
        if(w2 < MIN_W)
        {
            w2 = MIN_W;
        }
        if(h2 < MIN_H)
        {
            h2 = MIN_H;
        }
    
        if(fileStatusPanel != null)
        {
            _splitPane1.setDividerLocation(w2);
        }
        if(commandArea != null)
        {
            _splitPane2.setDividerLocation(h2);
        }
    }

    private void updateCanvas()
    {
        Canvas canvasL = getCanvasL();
        Canvas canvasR = getCanvasR();

        int w = _panelCanvas.getWidth();
        int h = _panelCanvas.getHeight();

        if( isStereo() )
        {
            int w1 = w/2;
            canvasL.setLocation(0, 0);
            canvasL.setSize(w1, h);
            canvasR.setLocation(w1, 0);
            canvasR.setSize(w - w1, h);
        }
        else
        {
            canvasL.setLocation(0, 0);
            canvasL.setSize(w, h);
        }
    }
    
    private void setupStereo(boolean stereo)
    {
        Canvas canvasR = getCanvasR();

        if( isStereo() )
        {
            canvasR.setVisible(true);
        }
        else
        {
            canvasR.setVisible(false);
        }
    }
}
