/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.mouse.*;
import org.pdbj.viewer.utils.*;

/**
 * Class that represents menus.
 */
public class ViewerMenu
{
    private GuiWrapper     _guiWrapper;
    private CommandHandler _commandHandler;

    private boolean _standalone;
    
    private Menu     _menuFile;
    private MenuItem _miFileInformation;

    // stand alone
    private Menu     _menuFileLoadLocal;
    private MenuItem _miFileLoadLocalPDBml;
    private MenuItem _miFileLoadLocalPDB;
    private MenuItem _miFileLoadLocalPolygon;
    private MenuItem _miFileLoadLocalPolygonCgo;
    private MenuItem _miFileLoadLocalScript;
    private MenuItem _miFileLoadLocalAnimation;
    private Menu     _menuFileLoadRemote;
    private MenuItem _miFileLoadRemotePDBml;
    private MenuItem _miFileLoadRemotePDB;
    private MenuItem _miFileLoadRemotePolygon;
    private MenuItem _miFileLoadRemotePolygonCgo;
    private MenuItem _miFileLoadRemoteScript;
    private MenuItem _miFileLoadRemoteAnimation;
    private MenuItem _miFileLoadRemoteFTP;
    private MenuItem _miFileLoadRemoteEFsite;
    private Menu     _menuFileClose;
    private Menu     _menuFileSave;
    private MenuItem _miFileSavePDB;
    private MenuItem _miFileSaveScript;
    private MenuItem _miFileSavePNG;
    private MenuItem _miFileSaveJPEG;
    private MenuItem _miFileExit;

    // applet
    private Menu _menuFileSelect;
    private Menu _menuFileDisplay;

    private Menu     _menuDisplay;
    private MenuItem _miDisplayWireframe;
    private MenuItem _miDisplayBackbone;
    private MenuItem _miDisplaySticks;
    private MenuItem _miDisplaySpacefill;
    private MenuItem _miDisplayBallStick;
    private MenuItem _miDisplayRibbons;
    private MenuItem _miDisplayCartoon;

    private Menu     _menuColors;
    private MenuItem _miColorsMonochrome;
    private MenuItem _miColorsCPK;
    private MenuItem _miColorsShapely;
    private MenuItem _miColorsGroup;
    private MenuItem _miColorsChain;
    private MenuItem _miColorsTemperature;
    private MenuItem _miColorsStructure;
    private MenuItem _miColorsCharge;
    private MenuItem _miColorsAmino;

    private Menu             _menuOptions;
    private CheckboxMenuItem _miOptionsHetero;
    private CheckboxMenuItem _miOptionsHydrogen;
    private CheckboxMenuItem _miOptionsSlab;
    private Menu             _menuOptionsStereo;
    private CheckboxMenuItem _miOptionsStereoOff;
    private CheckboxMenuItem _miOptionsStereoWallEyed;
    private CheckboxMenuItem _miOptionsStereoCrossEyed;
    private CheckboxMenuItem _miOptionsStereoRedBlue;
    private CheckboxMenuItem _miOptionsStereoRedCyan;
    private CheckboxMenuItem _miOptionsStereoRedGreen;
    private CheckboxMenuItem _miOptionsLoadCenter;
    private Menu             _menuOptionsPick;
    private CheckboxMenuItem _miOptionsPickOff;
    private CheckboxMenuItem _miOptionsPickId;
    private CheckboxMenuItem _miOptionsPickCoord;
    private CheckboxMenuItem _miOptionsPickDist;
    private CheckboxMenuItem _miOptionsPickCenter;
    private CheckboxMenuItem _miOptionsPickSelect;
    private CheckboxMenuItem _miOptionsAnimation;

    private Menu     _menuHelp;
    private MenuItem _miHelpAbout;
    private MenuItem _miHelpManual;

    // File History
    private static final int MAX_HISTORY = 4;
    private ArrayList _historyFileTypes;
    private ArrayList _historyFileLocators;
    private MenuItem[] _miFileHistory;

    // File-Close Menu
    private ArrayList _miFileCloseItems;
    private MenuItem _miFileCloseAll;

    // File-Select Menu
    private ArrayList _miFileSelectItems;

    // File-Display Menu
    private ArrayList _miFileDisplayItems;

    /**
     * Constructor.
     */
    public ViewerMenu(
        GuiWrapper guiWrapper, CommandHandler commandHandler, boolean standalone)
    {
        _guiWrapper = guiWrapper;
        _commandHandler = commandHandler;
        _standalone = standalone;

        this.createMenu();
    }

    /**
     * Returns the file menu.
     */
    public Menu getFileMenu()
    {
        return _menuFile;
    }

    /**
     * Returns the display menu.
     */
    public Menu getDisplayMenu()
    {
        return _menuDisplay;
    }

    /**
     * Returns the colors menu.
     */
    public Menu getColorsMenu()
    {
        return _menuColors;
    }

    /**
     * Returns the options menu.
     */
    public Menu getOptionsMenu()
    {
        return _menuOptions;
    }

    /**
     * Returns the help menu.
     */
    public Menu getHelpMenu()
    {
        return _menuHelp;
    }

    /**
     * Updates the file control menu (select and display).
     */
    public void updateFileControllMenu()
    {
        ArrayList fileStatusList = _guiWrapper.getViewerController().getFileStatusList();

        _menuFileSelect.removeAll();
        _menuFileDisplay.removeAll();

        for(int i=0; i<fileStatusList.size(); i++)
        {
            FileStatus status = (FileStatus)fileStatusList.get(i);

            // label
            String path = status.getPath();
            int index = path.lastIndexOf("/");
            String label = status.getId() + ".  " + path.substring(index + 1);

            // Select
            if(_miFileSelectItems.size() == i){
                _miFileSelectItems.add( this.createFileCheckboxItem("") );
            }
            CheckboxMenuItem miSelect = (CheckboxMenuItem)_miFileSelectItems.get(i);
            miSelect.setLabel(label);
            miSelect.setState( status.isSelected() );
            _menuFileSelect.add(miSelect);

            // Display
            if(_miFileDisplayItems.size() == i){
                _miFileDisplayItems.add( this.createFileCheckboxItem("") );
            }
            CheckboxMenuItem miDisplay = (CheckboxMenuItem)_miFileDisplayItems.get(i);
            miDisplay.setLabel(label);
            miDisplay.setState( status.isDisplayed() );
            _menuFileDisplay.add(miDisplay);
        }
    }

    /**
     * Updates the file status.
     */
    public void updateFileStatus()
    {
        ArrayList fileStatusList = _guiWrapper.getViewerController().getFileStatusList();
        for(int i=0; i<fileStatusList.size(); i++)
        {
            FileStatus status = (FileStatus)fileStatusList.get(i);
            ( (CheckboxMenuItem)_menuFileSelect.getItem(i) ).setState( status.isSelected() );
            ( (CheckboxMenuItem)_menuFileDisplay.getItem(i) ).setState( status.isDisplayed() );
        }
    }

    /**
     * Updates the file close menu.
     */
    public void updateFileCloseMenu()
    {
        ArrayList fileStatusList = _guiWrapper.getViewerController().getFileStatusList();

        _menuFileClose.removeAll();

        int cnt = fileStatusList.size();
        for(int i=0; i<cnt; i++)
        {
            FileStatus status = (FileStatus)fileStatusList.get(i);

            if(_miFileCloseItems.size() == i){
                _miFileCloseItems.add( this.createFileMenuItem("") );
            }

            MenuItem menuItem = (MenuItem)_miFileCloseItems.get(i);

            String path = status.getPath();
            int index = path.lastIndexOf("\\");
            if(index == -1) index = path.lastIndexOf("/");
            String name = path.substring(index + 1);

            menuItem.setLabel( status.getId() + ".  " + name );

            _menuFileClose.add(menuItem);
        }

        if(cnt > 1){
            _menuFileClose.addSeparator();
            _menuFileClose.add(_miFileCloseAll);            
        }
    }

    /**
     * Updates the options menu.
     */
    public void updateOptionsMenu()
    {
        ViewerController controller = _guiWrapper.getViewerController();
        MouseBehavior mouseBehavior = _guiWrapper.getMouseBehavior();

        _miOptionsHetero.setState( controller.getSelectionStatus().getHetero() );
        _miOptionsHydrogen.setState( controller.getSelectionStatus().getHydrogen() );
        _miOptionsSlab.setState( controller.getViewStatus().isSlab() );
        _miOptionsStereoOff.setState(
                !controller.getViewStatus().isStereo() && 
                controller.getViewStatus().isAnaglyphOff());
        _miOptionsStereoWallEyed.setState(
                controller.getViewStatus().isStereo() && 
                controller.getViewStatus().getStereoAngle() > 0);
        _miOptionsStereoCrossEyed.setState(
                controller.getViewStatus().isStereo() && 
                controller.getViewStatus().getStereoAngle() < 0);
        _miOptionsStereoRedBlue.setState(
                controller.getViewStatus().isAnaglyphRedBlue());
        _miOptionsStereoRedCyan.setState(
                controller.getViewStatus().isAnaglyphRedCyan());
        _miOptionsStereoRedGreen.setState(
                controller.getViewStatus().isAnaglyphRedGreen());
        if(_standalone){
            _miOptionsLoadCenter.setState( controller.getViewStatus().isLoadCenter() );
        }

        int mode = mouseBehavior.getPickingMode();

        _miOptionsPickOff.setState(mode == MouseBehavior.PICKING_OFF);
        _miOptionsPickId.setState(mode == MouseBehavior.PICKING_ID);
        _miOptionsPickCoord.setState(mode == MouseBehavior.PICKING_COORDINATES);
        _miOptionsPickDist.setState(mode == MouseBehavior.PICKING_DISTANCE);
        _miOptionsPickCenter.setState(mode == MouseBehavior.PICKING_CENTER);
        _miOptionsPickSelect.setState(mode == MouseBehavior.PICKING_SELECT);

        _miOptionsAnimation.setState(false);
    }

    /**
     * Appends a file to the history.
     */
    public void addFileHistory(FileType type, FileLocator locator)
    {
        int cnt = _historyFileTypes.size();
        int pos = _menuFile.getItemCount();

        _historyFileTypes.add(type);
        _historyFileLocators.add(locator);

        if(cnt < MAX_HISTORY){
            if(cnt == 0){
                _menuFile.insertSeparator(pos-2);
                pos++;
            }
            _menuFile.insert( _miFileHistory[cnt], pos-2);
        }
        else{
            _historyFileTypes.remove(0);
            _historyFileLocators.remove(0);
        }

        for(int i=0; i<_historyFileLocators.size(); i++)
        {
            FileLocator fileLocator = (FileLocator)_historyFileLocators.get(i);
            String name = null;
            if( fileLocator.isLocalFile() ){
                name = fileLocator.getFile().getName();
            }
            else if( fileLocator.isRemoteFile() ){
                name = fileLocator.getURL().getFile();
            }
            else{
                name = fileLocator.getPDBCode();
            }
            int index = name.lastIndexOf("/");
            if(index != -1) name = name.substring(index + 1);
            _miFileHistory[i].setLabel(name);
        }
    }
    
    //--------------------------------------------------------------------
    // Creates menu
    //--------------------------------------------------------------------
    private void createMenu()
    {
        // File Menu
        _menuFile = new Menu("File");
        
        _miFileInformation = createFileMenuItem("Information");

        if(_standalone){
            _menuFileLoadLocal = new Menu("Open - Local");
            _miFileLoadLocalPDBml      = createFileMenuItem("PDBML");
            _miFileLoadLocalPDB        = createFileMenuItem("PDB");
            _miFileLoadLocalPolygon    = createFileMenuItem("Polygon");
            _miFileLoadLocalPolygonCgo = createFileMenuItem("Polygon CGO");
            _miFileLoadLocalScript     = createFileMenuItem("Script");
            _miFileLoadLocalAnimation  = createFileMenuItem("Animation");
            _menuFileLoadLocal.add(_miFileLoadLocalPDBml);
            _menuFileLoadLocal.add(_miFileLoadLocalPDB);
            _menuFileLoadLocal.add(_miFileLoadLocalPolygon);
            _menuFileLoadLocal.add(_miFileLoadLocalPolygonCgo);
            _menuFileLoadLocal.add(_miFileLoadLocalScript);
            _menuFileLoadLocal.add(_miFileLoadLocalAnimation);
            _menuFileLoadRemote = new Menu("Open - Remote");
            _miFileLoadRemoteFTP       = createFileMenuItem("PDB ID");
            _miFileLoadRemoteEFsite    = createFileMenuItem("eF-site ID");
            _miFileLoadRemotePDBml     = createFileMenuItem("PDBML");
            _miFileLoadRemotePDB       = createFileMenuItem("PDB");
            _miFileLoadRemotePolygon   = createFileMenuItem("Polygon");
            _miFileLoadRemotePolygonCgo = createFileMenuItem("Polygon CGO");
            _miFileLoadRemoteScript    = createFileMenuItem("Script");
            _miFileLoadRemoteAnimation = createFileMenuItem("Animation");
            _menuFileLoadRemote.add(_miFileLoadRemoteFTP);
            _menuFileLoadRemote.add(_miFileLoadRemoteEFsite);
            _menuFileLoadRemote.addSeparator();
            _menuFileLoadRemote.add(_miFileLoadRemotePDBml);
            _menuFileLoadRemote.add(_miFileLoadRemotePDB);
            _menuFileLoadRemote.add(_miFileLoadRemotePolygon);
            _menuFileLoadRemote.add(_miFileLoadRemotePolygonCgo);
            _menuFileLoadRemote.add(_miFileLoadRemoteScript);
            _menuFileLoadRemote.add(_miFileLoadRemoteAnimation);
            _menuFileClose = new Menu("Close");
            _menuFileSave  = new Menu("Save");
            _miFileSavePDB    = createFileMenuItem("PDB");
            _miFileSaveScript = createFileMenuItem("Script");
            _miFileSavePNG    = createFileMenuItem("PNG");
            _miFileSaveJPEG   = createFileMenuItem("JPEG");
            _menuFileSave.add(_miFileSavePDB);
            _menuFileSave.add(_miFileSaveScript);
            _menuFileSave.add(_miFileSavePNG);
            _menuFileSave.add(_miFileSaveJPEG);
            _miFileExit = createFileMenuItem("Exit");

            _menuFile.add(_menuFileLoadLocal);
            _menuFile.add(_menuFileLoadRemote);
            _menuFile.add(_miFileInformation);
            _menuFile.add(_menuFileClose);
            _menuFile.addSeparator();
            _menuFile.add(_menuFileSave);
            _menuFile.addSeparator();
            _menuFile.add(_miFileExit);
            
            // set up close
            _miFileCloseItems = new ArrayList();
            _miFileCloseAll = this.createFileMenuItem("All");

            // set up history
            _historyFileTypes = new ArrayList();
            _historyFileLocators = new ArrayList();
            _miFileHistory = new MenuItem[MAX_HISTORY];
            for(int i=0; i<MAX_HISTORY; i++){
                _miFileHistory[i] = createFileMenuItem("");
            }
        }
        else{
            _menuFileSelect  = new Menu("Select");
            _menuFileDisplay = new Menu("Display");

            _miFileSelectItems  = new ArrayList();
            _miFileDisplayItems = new ArrayList();
        
            _menuFileSave  = new Menu("Save");
            _miFileSavePNG    = createFileMenuItem("PNG");
            _miFileSaveJPEG   = createFileMenuItem("JPEG");
            _menuFileSave.add(_miFileSavePNG);
            _menuFileSave.add(_miFileSaveJPEG);
            if (!ViewerApplet.isSignedApplet()) {
                _miFileSavePNG.setEnabled(false);
                _miFileSaveJPEG.setEnabled(false);
            }

            _menuFile.add(_menuFileSelect);
            _menuFile.add(_menuFileDisplay);
            _menuFile.add(_miFileInformation);
            _menuFile.add(_menuFileSave);
        }

        // Display Menu
        _menuDisplay = new Menu("Display");
        _miDisplayWireframe = createDisplayMenuItem("Wireframe");
        _miDisplayBackbone  = createDisplayMenuItem("Backbone");
        _miDisplaySticks    = createDisplayMenuItem("Sticks");
        _miDisplaySpacefill = createDisplayMenuItem("Spacefill");
        _miDisplayBallStick = createDisplayMenuItem("Ball&Stick");
        _miDisplayRibbons   = createDisplayMenuItem("Ribbons");
        _miDisplayCartoon   = createDisplayMenuItem("Cartoon");

        _menuDisplay.add(_miDisplayWireframe);
        _menuDisplay.add(_miDisplayBackbone);
        _menuDisplay.add(_miDisplaySticks);
        _menuDisplay.add(_miDisplaySpacefill);
        _menuDisplay.add(_miDisplayBallStick);
        _menuDisplay.add(_miDisplayRibbons);
        _menuDisplay.add(_miDisplayCartoon);

        // Colors Menu
        _menuColors = new Menu("Colors");
        _miColorsMonochrome  = createColorsMenuItem("Monochrome");
        _miColorsCPK         = createColorsMenuItem("CPK");
        _miColorsShapely     = createColorsMenuItem("Shapely");
        _miColorsGroup       = createColorsMenuItem("Group");
        _miColorsChain       = createColorsMenuItem("Chain");
        _miColorsTemperature = createColorsMenuItem("Temperature");
        _miColorsStructure   = createColorsMenuItem("Structure");
        _miColorsCharge      = createColorsMenuItem("Charge");
        _miColorsAmino       = createColorsMenuItem("Amino");

        _menuColors.add(_miColorsMonochrome);
        _menuColors.add(_miColorsCPK);
        _menuColors.add(_miColorsShapely);
        _menuColors.add(_miColorsGroup);
        _menuColors.add(_miColorsChain);
        _menuColors.add(_miColorsTemperature);
        _menuColors.add(_miColorsStructure);
        _menuColors.add(_miColorsCharge);
        _menuColors.add(_miColorsAmino);
        
        // Options Menu
        _menuOptions = new Menu("Options");
        _miOptionsHetero     = createOptionsMenuItem("Hetero Atoms");
        _miOptionsHydrogen   = createOptionsMenuItem("Hydrogens");
        _miOptionsSlab       = createOptionsMenuItem("Slab");
        _menuOptionsStereo = new Menu("Stereo");
        _miOptionsStereoOff = createOptionsMenuItem("Off");
        _miOptionsStereoWallEyed  = createOptionsMenuItem("Wall-Eyed");
        _miOptionsStereoCrossEyed = createOptionsMenuItem("Cross-Eyed");
        _miOptionsStereoRedBlue   = createOptionsMenuItem("Red-Blue");
        _miOptionsStereoRedCyan   = createOptionsMenuItem("Red-Cyan");
        _miOptionsStereoRedGreen  = createOptionsMenuItem("Red-Green");
        _miOptionsStereoWallEyed.setEnabled(!_guiWrapper.isNewt());
        _miOptionsStereoCrossEyed.setEnabled(!_guiWrapper.isNewt());
        _menuOptionsStereo.add(_miOptionsStereoOff);
        _menuOptionsStereo.add(_miOptionsStereoWallEyed);
        _menuOptionsStereo.add(_miOptionsStereoCrossEyed);
        _menuOptionsStereo.add(_miOptionsStereoRedBlue);
        _menuOptionsStereo.add(_miOptionsStereoRedCyan);
        _menuOptionsStereo.add(_miOptionsStereoRedGreen);
        _menuOptionsPick = new Menu("Pick");
        _miOptionsPickOff    = createOptionsMenuItem("Off");
        _miOptionsPickId     = createOptionsMenuItem("Ident");
        _miOptionsPickCoord  = createOptionsMenuItem("Coordinates");
        _miOptionsPickDist   = createOptionsMenuItem("Distance");
        _miOptionsPickCenter = createOptionsMenuItem("Center");
        _miOptionsPickSelect = createOptionsMenuItem("Select");
        _menuOptionsPick.add(_miOptionsPickOff);
        _menuOptionsPick.add(_miOptionsPickId);
        _menuOptionsPick.add(_miOptionsPickCoord);
        _menuOptionsPick.add(_miOptionsPickDist);
        _menuOptionsPick.add(_miOptionsPickCenter);
        _menuOptionsPick.add(_miOptionsPickSelect);
        _miOptionsAnimation  = createOptionsMenuItem("Animation");
        
        _menuOptions.add(_miOptionsHetero);
        _menuOptions.add(_miOptionsHydrogen);
        _menuOptions.add(_miOptionsSlab);
        _menuOptions.add(_menuOptionsStereo);
        if(_standalone){
            _miOptionsLoadCenter = createOptionsMenuItem("Load To Center");
            _menuOptions.add(_miOptionsLoadCenter);
        }
        _menuOptions.add(_menuOptionsPick);
        _menuOptions.add(_miOptionsAnimation);

        // Help Menu
        _menuHelp = new Menu("Help");
        _miHelpAbout = createHelpMenuItem("About jV");

        _menuHelp.add(_miHelpAbout);

        if (_standalone) {
            _miHelpManual = createHelpMenuItem("Open Manual");
            _miHelpManual.setEnabled(isDesktopSupported());
            _menuHelp.add(_miHelpManual);
        }
    }
        
    private boolean isDesktopSupported() {
        try {
            return Util.isDesktopSupported();
        } catch (Exception e) {
            return false;
        }
    }

    private MenuItem createFileMenuItem(String label)
    {
        MenuItem menuItem = new MenuItem(label);
        menuItem.addActionListener(
            new ActionListener(){
                public void actionPerformed(ActionEvent e){
                    menuFile_actionPerformed(e);
                }
            }
        );
        return menuItem;
    }

    private MenuItem createDisplayMenuItem(String label)
    {
        MenuItem menuItem = new MenuItem(label);
        menuItem.addActionListener(
            new ActionListener(){
                public void actionPerformed(ActionEvent e){
                    menuDisplay_actionPerformed(e);
                }
            }
        );
        return menuItem;
    }

    private MenuItem createColorsMenuItem(String label)
    {
        MenuItem menuItem = new MenuItem(label);
        menuItem.addActionListener(
            new ActionListener(){
                public void actionPerformed(ActionEvent e){
                    menuColors_actionPerformed(e);
                }
            }
        );
        return menuItem;
    }

    private CheckboxMenuItem createOptionsMenuItem(String label)
    {
        CheckboxMenuItem menuItem = new CheckboxMenuItem(label);
        menuItem.addItemListener(
            new ItemListener(){
                public void itemStateChanged(ItemEvent e){
                    menuOptions_itemStateChanged(e);
                }
            }
        );
        return menuItem;
    }

    private MenuItem createHelpMenuItem(String label)
    {
        MenuItem menuItem = new MenuItem(label);
        menuItem.addActionListener(
            new ActionListener(){
                public void actionPerformed(ActionEvent e){
                    menuHelp_actionPerformed(e);
                }
            }
        );
        return menuItem;
    }

    private CheckboxMenuItem createFileCheckboxItem(String label)
    {
        CheckboxMenuItem menuItem = new CheckboxMenuItem(label);
        menuItem.addItemListener(
            new ItemListener(){
                public void itemStateChanged(ItemEvent e){
                    file_itemStateChanged(e);
                }
            }
        );
        return menuItem;
    }

    //--------------------------------------------------------------------
    // menu event handlers
    //--------------------------------------------------------------------
    private void menuFile_actionPerformed(ActionEvent e)
    {
        MenuItem item = (MenuItem)e.getSource();
        if(item == _miFileExit){
            _guiWrapper.exit();
        }
        else if( !_guiWrapper.menuActive() ) return;

        ViewerController controller = _guiWrapper.getViewerController();
        
        if(item == _miFileLoadLocalPDBml)           _guiWrapper.loadLocalFile(FileType.PDBML);
        else if(item == _miFileLoadLocalPDB)        _guiWrapper.loadLocalFile(FileType.PDB);
        else if(item == _miFileLoadLocalPolygon)    _guiWrapper.loadLocalFile(FileType.POLYGON);
        else if(item == _miFileLoadLocalPolygonCgo) _guiWrapper.loadLocalFile(FileType.POLYGON_CGO);
        else if(item == _miFileLoadLocalScript)     _guiWrapper.loadLocalFile(FileType.SCRIPT);
        else if(item == _miFileLoadLocalAnimation)  _guiWrapper.loadLocalFile(FileType.ANIMATION);
        else if(item == _miFileLoadRemotePDBml)     _guiWrapper.loadRemoteFile(FileType.PDBML);
        else if(item == _miFileLoadRemotePDB)       _guiWrapper.loadRemoteFile(FileType.PDB);
        else if(item == _miFileLoadRemotePolygon)   _guiWrapper.loadRemoteFile(FileType.POLYGON);
        else if(item == _miFileLoadRemotePolygonCgo) _guiWrapper.loadRemoteFile(FileType.POLYGON_CGO);
        else if(item == _miFileLoadRemoteScript)    _guiWrapper.loadRemoteFile(FileType.SCRIPT);
        else if(item == _miFileLoadRemoteAnimation) _guiWrapper.loadRemoteFile(FileType.ANIMATION);
        else if(item == _miFileLoadRemoteFTP)       _guiWrapper.loadByPDBID();
        else if(item == _miFileLoadRemoteEFsite)    _guiWrapper.loadByEFsiteID();
        else if(item == _miFileInformation)         _guiWrapper.showInformation();
        else if(item == _miFileSavePDB)             _guiWrapper.saveFile(FileType.PDB);
        else if(item == _miFileSaveScript)          _guiWrapper.saveFile(FileType.SCRIPT);
        else if(item == _miFileSavePNG)             _guiWrapper.saveFile(FileType.PNG);
        else if(item == _miFileSaveJPEG)            _guiWrapper.saveFile(FileType.JPEG);
        else if(item == _miFileCloseAll){
            controller.closeAllFile();
        }
        else{
            for(int i=0; i<_menuFileClose.getItemCount(); i++)
            {
                if( item == _menuFileClose.getItem(i) ){
                    FileStatus status = (FileStatus)controller.getFileStatusList().get(i);
                    controller.closeFile( new int[]{status.getId()} );
                    return;
                }
            }
            for(int i=0; i<MAX_HISTORY; i++)
            {
                if( item == _miFileHistory[i] ){
                    FileType type = (FileType)_historyFileTypes.get(i);
                    FileLocator locator = (FileLocator)_historyFileLocators.get(i);
                    _guiWrapper.loadHistory(type, locator);
                    return;
                }
            }
        }
    }
    
    private void menuDisplay_actionPerformed(ActionEvent e)
    {
        if( !_guiWrapper.menuActive() ) return;
        
        ViewerController controller = _guiWrapper.getViewerController();
        controller.stopRenderer();

        MenuItem item = (MenuItem)e.getSource();
        if(item == _miDisplayWireframe){
            controller.spacefillOff();
            controller.backboneOff();        
            controller.ribbonOff();        
            controller.wireframe(0);
        }
        else if(item == _miDisplayBackbone){
            controller.wireframeOff();
            controller.spacefillOff();
            controller.ribbonOff();        
            controller.backbone(0.3f);
        }
        else if(item == _miDisplaySticks){
            controller.spacefillOff();
            controller.backboneOff();        
            controller.ribbonOff();        
            controller.wireframe(0.2f);
        }
        else if(item == _miDisplaySpacefill){
            controller.wireframeOff();
            controller.ribbonOff();        
            controller.backboneOff();
            controller.spacefillVdW();
        }
        else if(item == _miDisplayBallStick){
            controller.backboneOff();        
            controller.ribbonOff();        
            controller.spacefill(0.5f);
            controller.wireframe(0.2f);
        }
        else if(item == _miDisplayRibbons){
            controller.wireframeOff();
            controller.spacefillOff();
            controller.backboneOff();
            controller.ribbon();
        }
        else if(item == _miDisplayCartoon){
            controller.wireframeOff();
            controller.spacefillOff();
            controller.backboneOff();
            controller.cartoon();
        }

        controller.startRenderer();
    }

    private void menuColors_actionPerformed(ActionEvent e)
    {
        if( !_guiWrapper.menuActive() ) return;

        String command = null;
    
        MenuItem item = (MenuItem)e.getSource();
        if(item == _miColorsMonochrome)       command = "color white";
        else if(item == _miColorsCPK)         command = "color cpk";
        else if(item == _miColorsShapely)     command = "color shapely";
        else if(item == _miColorsGroup)       command = "color group";
        else if(item == _miColorsChain)       command = "color chain";
        else if(item == _miColorsTemperature) command = "color temperature";
        else if(item == _miColorsStructure)   command = "color structure";
        else if(item == _miColorsCharge)      command = "color charge";
        else if(item == _miColorsAmino)       command = "color amino";

        if(command != null) _commandHandler.executeSingleCommand(command);
    }

    private void menuOptions_itemStateChanged(ItemEvent e)
    {
        if( !_guiWrapper.menuActive() ) return;

        ViewerController controller = _guiWrapper.getViewerController();
        MouseBehavior mouseBehavior = _guiWrapper.getMouseBehavior();

        CheckboxMenuItem item = (CheckboxMenuItem)e.getSource();
        if(item == _miOptionsHetero){
            controller.getSelectionStatus().setHetero( item.getState() );
            if( !item.getState() ){
                controller.eliminateHeteroAtoms();
            }
            _commandHandler.executeSingleCommand("select");
        }
        else if(item == _miOptionsHydrogen){
            controller.getSelectionStatus().setHydrogen( item.getState() );
            if( !item.getState() ){
                controller.eliminateHydrogens();
            }
            _commandHandler.executeSingleCommand("select");
        }
        else if(item == _miOptionsSlab){
            if( item.getState() ) controller.slab();
            else controller.slabOff();
        }
         else if(item == _miOptionsStereoOff){
            _commandHandler.executeSingleCommand("stereo off");
        }
        else if(item == _miOptionsStereoWallEyed){
            _commandHandler.executeSingleCommand("stereo " + getSeparationAngleAbs());
        }
        else if(item == _miOptionsStereoCrossEyed){
            _commandHandler.executeSingleCommand("stereo " + (-getSeparationAngleAbs()));
        }
        else if(item == _miOptionsStereoRedBlue){
            _commandHandler.executeSingleCommand("stereo redblue " + getSeparationAngleAbs());
        }
        else if(item == _miOptionsStereoRedCyan){
            _commandHandler.executeSingleCommand("stereo redcyan " + getSeparationAngleAbs());
        }
        else if(item == _miOptionsStereoRedGreen){
            _commandHandler.executeSingleCommand("stereo redgreen " + getSeparationAngleAbs());
        }
        else if(item == _miOptionsLoadCenter){
            controller.setLoadCenter( item.getState() );
        }
        else if(item == _miOptionsPickOff){
            mouseBehavior.setPickingMode(MouseBehavior.PICKING_OFF);
        }
        else if(item == _miOptionsPickId){
            mouseBehavior.setPickingMode(MouseBehavior.PICKING_ID);
        }
        else if(item == _miOptionsPickCoord){
            mouseBehavior.setPickingMode(MouseBehavior.PICKING_COORDINATES);
        }
        else if(item == _miOptionsPickDist){
            mouseBehavior.setPickingMode(MouseBehavior.PICKING_DISTANCE);
        }
        else if(item == _miOptionsPickCenter){
            mouseBehavior.setPickingMode(MouseBehavior.PICKING_CENTER);
        }
        else if(item == _miOptionsPickSelect){
            mouseBehavior.setPickingMode(MouseBehavior.PICKING_SELECT);
        }
        else if(item == _miOptionsAnimation){
            _guiWrapper.showAnimationFrame();
        }

        this.updateOptionsMenu();
    }

    private float getSeparationAngleAbs() {
        return 2 * Math.abs(_guiWrapper.getViewerController().getViewStatus().getStereoAngle());
    }

    private void menuHelp_actionPerformed(ActionEvent e)
    {
        if( !_guiWrapper.menuActive() ) return;

        MenuItem item = (MenuItem)e.getSource();
        if (item == _miHelpAbout) {
            _guiWrapper.showAboutDialog();
        } else if (item == _miHelpManual) {
            _guiWrapper.openManual();
        }
    }

    private void file_itemStateChanged(ItemEvent e)
    {
        if( !_guiWrapper.menuActive() ) return;

        ViewerController controller = _guiWrapper.getViewerController();

        CheckboxMenuItem item = (CheckboxMenuItem)e.getSource();
        for(int i=0; i<_menuFileSelect.getItemCount(); i++)
        {
            if( item == _menuFileSelect.getItem(i) ){
                FileStatus status = (FileStatus)controller.getFileStatusList().get(i);
                controller.selectFile(item.getState(), new int[]{status.getId()});
                return;
            }
        }
        for(int i=0; i<_menuFileDisplay.getItemCount(); i++)
        {
            if( item == _menuFileDisplay.getItem(i) ){
                FileStatus status = (FileStatus)controller.getFileStatusList().get(i);
                controller.displayFile(item.getState(), new int[]{status.getId()});
                break;
            }
        }
    }
}
