/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.gui;

import java.awt.*;
import java.text.NumberFormat;
import javax.swing.*;
import javax.swing.border.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.mouse.*;

/**
 * Viewer status panel.
 */
public class ViewerStatusPanel extends JPanel implements ViewStatusListener,
        MouseSettingsListener {

    private static final long serialVersionUID = 1L;

    private JLabel _slabLabel;
    private JLabel _mousePickLabel;

    private NumberFormat _format;

    /**
     * Constructor.
     */
    public ViewerStatusPanel() {
        createComponent();

        _format = NumberFormat.getNumberInstance();
        _format.setMaximumFractionDigits(2);
        _format.setMinimumFractionDigits(2);
        _format.setGroupingUsed(false);
    }

    /**
     * Called when slab settings is changed.
     */
    public void slabChanged(ViewStatus viewStatus) {
        if (viewStatus.isSlab()) {
            float slabZ = viewStatus.getSlabValue();
            _slabLabel.setText("z = " + _format.format(slabZ));
        } else {
            _slabLabel.setText("Off");
        }
    }

    /**
     * Called when picking mode is changed.
     */
    public void pickingModeChanged(MouseBehavior mouseBehavior) {
        switch (mouseBehavior.getPickingMode()) {
        case MouseBehavior.PICKING_ID:
            _mousePickLabel.setText("Ident");
            return;
        case MouseBehavior.PICKING_COORDINATES:
            _mousePickLabel.setText("Coordinates");
            return;
        case MouseBehavior.PICKING_DISTANCE:
            _mousePickLabel.setText("Distance");
            return;
        case MouseBehavior.PICKING_CENTER:
            _mousePickLabel.setText("Center");
            return;
        case MouseBehavior.PICKING_SELECT:
            _mousePickLabel.setText("Select");
            return;
        default:
            _mousePickLabel.setText("Off");
        }
    }

    /**
     * Update all.
     */
    public void updateAll(ViewStatus viewStatus, MouseBehavior mouseBehavior) {
        slabChanged(viewStatus);
        pickingModeChanged(mouseBehavior);
    }

    private void createComponent() {
        _slabLabel = new JLabel("");
        _mousePickLabel = new JLabel("");

        JLabel label1 = new JLabel("Pick");
        JLabel label2 = new JLabel("Slab");

        Font font = new Font("SansSerif", Font.PLAIN, 12);
        _slabLabel.setFont(font);
        _mousePickLabel.setFont(font);
        label1.setFont(font);
        label2.setFont(font);

        BoxLayout layout = new BoxLayout(this, BoxLayout.X_AXIS);
        setLayout(layout);

        JPanel panelT = new JPanel();
        panelT.setLayout(new BoxLayout(panelT, BoxLayout.Y_AXIS));

        JPanel panelT1 = new JPanel();
        panelT1.setLayout(new BoxLayout(panelT1, BoxLayout.X_AXIS));
        panelT1.add(Box.createRigidArea(new Dimension(10, 20)));
        panelT1.add(label1);
        panelT1.add(new Box.Filler(new Dimension(1, 20), new Dimension(5, 20), new Dimension(20, 20)));

        JPanel panelT2 = new JPanel();
        panelT2.setLayout(new BoxLayout(panelT2, BoxLayout.X_AXIS));
        panelT2.add(Box.createRigidArea(new Dimension(10, 20)));
        panelT2.add(label2);
        panelT2.add(Box.createRigidArea(new Dimension(5, 20)));

        JPanel panelT3 = new JPanel();
        panelT3.setLayout(new BoxLayout(panelT3, BoxLayout.X_AXIS));
        panelT3.add(Box.createRigidArea(new Dimension(5, 2)));

        panelT.add(panelT1);
        panelT.add(panelT2);
        panelT.add(panelT3);

        JPanel panelS = new JPanel();
        panelS.setLayout(new BoxLayout(panelS, BoxLayout.Y_AXIS));

        JPanel panelS1 = new JPanel();
        panelS1.setLayout(new BoxLayout(panelS1, BoxLayout.X_AXIS));
        panelS1.add(Box.createRigidArea(new Dimension(5, 20)));
        panelS1.add(new JLabel(":"));

        JPanel panelS2 = new JPanel();
        panelS2.setLayout(new BoxLayout(panelS2, BoxLayout.X_AXIS));
        panelS2.add(Box.createRigidArea(new Dimension(5, 20)));
        panelS2.add(new JLabel(":"));

        JPanel panelS3 = new JPanel();
        panelS3.setLayout(new BoxLayout(panelS3, BoxLayout.X_AXIS));
        panelS3.add(Box.createRigidArea(new Dimension(5, 2)));

        panelS.add(panelS1);
        panelS.add(panelS2);
        panelS.add(panelS3);

        JPanel panelV = new JPanel();
        panelV.setLayout(new BoxLayout(panelV, BoxLayout.Y_AXIS));

        JPanel panelV1 = new JPanel();
        panelV1.setLayout(new BoxLayout(panelV1, BoxLayout.X_AXIS));
        panelV1.add(Box.createRigidArea(new Dimension(5, 20)));
        panelV1.add(_mousePickLabel);
        panelV1.add(Box.createHorizontalGlue());

        JPanel panelV2 = new JPanel();
        panelV2.setLayout(new BoxLayout(panelV2, BoxLayout.X_AXIS));
        panelV2.add(Box.createRigidArea(new Dimension(5, 20)));
        panelV2.add(_slabLabel);
        panelV2.add(Box.createHorizontalGlue());

        JPanel panelV3 = new JPanel();
        panelV3.setLayout(new BoxLayout(panelV3, BoxLayout.X_AXIS));
        panelV3.add(Box.createRigidArea(new Dimension(90, 2)));

        panelV.add(panelV1);
        panelV.add(panelV2);
        panelV.add(panelV3);

        add(panelT);
        add(panelS);
        add(panelV);

        this.setBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));
    }
}
