/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.manipulators;

import java.awt.Color;
import java.util.ArrayList;
import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.images.*;
import org.pdbj.viewer.molecules.*;

// 20040517 findCenter, cartoon height, hbond backbone
// 20060414 modify private void ribbon(), default ribbon width.

/**
* Class that manipulates molecules.
*/
public class MoleculeManipulator
{
    public static final float MAX_TEMP_FACTOR_RADIUS = 3.0f;

    private SelectionStatus _selectionStatus;
    private ColorManager    _colorManager;
    private ArrayList       _moleculeFiles;

    private MoleculeFile[] _moleculeFileArray;

    /**
    * Returns the selection status.
    */
    public SelectionStatus getSelectionStatus()
    {
        return _selectionStatus;
    }

    /**
    * Sets the color manager.
    */
    public void setColorManager(ColorManager colorManager)
    {
        _colorManager = colorManager;
    }

    /**
    * Returns the color manager.
    */
    public ColorManager getColorManager()
    {
        return _colorManager;
    }

    /**
    * Constructor.
    */
    public MoleculeManipulator()
    {
        _moleculeFiles = new ArrayList();
        _selectionStatus = new SelectionStatus();
        this.updateMolecules();
    }

    /**
     * Appends the molecule file.
     */
    public void add(MoleculeFile moleculeFile)
    {
        _moleculeFiles.add(moleculeFile);
        this.updateMolecules();
    }

    /**
     * Removes the molecule file.
     */
    public void remove(MoleculeFile moleculeFile)
    {
        _moleculeFiles.remove(moleculeFile);
        this.updateMolecules();
    }

    /**
    * Selects atoms.
    */
    public int select(boolean all)
    {
        boolean hetero   = (all || _selectionStatus.getHetero());
        boolean hydrogen = (all || _selectionStatus.getHydrogen());

        return this.select(null, true, hetero, hydrogen);
    }

    /**
    * Selects atoms.
    */
    public int select(SelectExpr expr)
    {
        return this.select(expr, false, false, false);
    }

    /**
    * Selects none.
    */
    public void selectOff()
    {
        this.select(null, false, false, false);
    }

    /**
    * Enable wireframes.
    */
    public void wireframe(float radius)
    {
        this.wireframe(true, radius);
    }

    /**
    * Disable wireframes.
    */
    public void wireframeOff()
    {
        this.wireframe(false, 0);
    }

    /**
    * Enable spacefills.
    */
    public void spacefill(float radius)
    {
        this.spacefill(true, false, false, radius);
    }

    /**
    * Enable spacefills with Van der Waals radius.
    */
    public void spacefillVdW()
    {
        this.spacefill(true, true, false, 0);
    }

    /**
    * Enable spacefills using temperature factor.
    */
    public void spacefillTemp()
    {
        this.spacefill(true, false, true, 0);
    }

    /**
    * Disable spacefills.
    */
    public void spacefillOff()
    {
        this.spacefill(false, false, false, 0);
    }

    /**
    * Enable backbones.
    */
    public void backbone(float radius)
    {
        this.backbone(true, radius);
    }

    /**
    * Disable backbones.
    */
    public void backboneOff()
    {
        this.backbone(false, 0);
    }

    /**
    * Enable hydrogen bonds.
    */
    public void hBond(float radius)
    {
        this.hBond(true, radius);
    }

    /**
    * Disable hydrogen bonds.
    */
    public void hBondOff()
    {
        this.hBond(false, 0);
    }

    /**
    * Enable disulfide bonds.
    */
    public void ssBond(float radius)
    {
        this.ssBond(true, radius);
    }

    /**
    * Disable disulfide bonds.
    */
    public void ssBondOff()
    {
        this.ssBond(false, 0);
    }

    /**
     * Enable ribbons with default width.
     */
    public void ribbon()
    {
        this.ribbon(true, 0, false, true);
    }

    /**
     * Enable ribbons.
     */
    public void ribbon(float width)
    {
        this.ribbon(true, width, false, false);
    }

    /**
     * Disable ribbons.
     */
    public void ribbonOff()
    {
        this.ribbon(false, 0, false, false);
    }

    /**
     * Enable cartoons with default width.
     */
    public void cartoon()
    {
        this.ribbon(true, 0, true, true);
    }

    /**
     * Enable cartoons.
     */
    public void cartoon(float width)
    {
        this.ribbon(true, width, true, false);
    }

    /**
     * Disable cartoons.
     */
    public void cartoonOff()
    {
        this.ribbon(false, 0, true, false);
    }

    /**
     * Enable trace.
     */
    public void trace(float radius)
    {
        this.trace(true, radius);
    }

    /**
     * Disable trace.
     */
    public void traceOff()
    {
        this.trace(false, 0);
    }

    /**
    * Sets the color of images.
    */
    public void setColor(ColorExpr expr)
    {
        int colorIndex = -1;
        if(expr.getScheme() == ColorExpr.SCHEME_RGB){
            colorIndex = _colorManager.getIndex(expr.getColor());
        }
        
        if(expr.getTarget() == ColorExpr.TARGET_ATOM){
            if(expr.getScheme() == ColorExpr.SCHEME_RGB){
                this.colorAtom(colorIndex);
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_CPK){
                this.colorCPK();
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_AMINO){
                this.colorAmino();
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_SHAPELY){
                this.colorShapely();
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_CHAIN){
                this.colorChain();
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_GROUP){
                this.colorGroup();
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_TEMPERATURE){
                this.colorTemp(false);
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_CHARGE){
                this.colorTemp(true);
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_STRUCTURE){
                this.colorStruct();
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_DSSP){
                this.colorDssp();
            }
        }
        else if(expr.getTarget() == ColorExpr.TARGET_BOND){
            if( expr.getScheme() == ColorExpr.SCHEME_RGB ||
                expr.getScheme() == ColorExpr.SCHEME_DEFAULT )
            {
                this.colorBond(colorIndex);
            }
        }
        else if(expr.getTarget() == ColorExpr.TARGET_BACKBONE){
            if( expr.getScheme() == ColorExpr.SCHEME_RGB ||
                expr.getScheme() == ColorExpr.SCHEME_DEFAULT )
            {
                this.colorBackbone(colorIndex);
            }
        }
        else if(expr.getTarget() == ColorExpr.TARGET_HBOND){
            if( expr.getScheme() == ColorExpr.SCHEME_RGB ||
                expr.getScheme() == ColorExpr.SCHEME_DEFAULT )
            {
                this.colorHBond(colorIndex);
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_TYPE){
                this.colorHBondType();
            }
        }
        else if(expr.getTarget() == ColorExpr.TARGET_SSBOND){
            if( expr.getScheme() == ColorExpr.SCHEME_RGB ||
                expr.getScheme() == ColorExpr.SCHEME_DEFAULT )
            {
                this.colorSSBond(colorIndex);
            }
        }
        else if(expr.getTarget() == ColorExpr.TARGET_RESIDUE){
            if( expr.getScheme() == ColorExpr.SCHEME_RGB ||
                expr.getScheme() == ColorExpr.SCHEME_DEFAULT )
            {
                this.colorResidue(colorIndex);
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_DSSP){
                this.colorResidueDssp();
            }
        }
        else if(expr.getTarget() == ColorExpr.TARGET_RESIDUE_BACK){
            if( expr.getScheme() == ColorExpr.SCHEME_RGB ||
                expr.getScheme() == ColorExpr.SCHEME_DEFAULT )
            {
                this.colorResidueBack(colorIndex);
            }
            else if(expr.getScheme() == ColorExpr.SCHEME_DSSP){
                this.colorResidueBackDssp();
            }
        }
    }

    /**
    * Returns the center of specified atoms.
    */
    public Vector4f findCenter(SelectExpr expr)
    {
        if(expr == null) return null;

        expr.setup(_moleculeFileArray);

        Vector4f center = new Vector4f();
        AtomPath atomPath = new AtomPath();

        Vector4f v = new Vector4f();
        v.w = 1;
        
        int cnt = 0;
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            atomPath.setMoleculeFile(molecule.getFile());
            atomPath.setMolecule(molecule);
            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                atomPath.setChain(chain);
                for(int re=0; re<chain.numResidues(); re++)
                {
                    Residue residue = chain.getResidue(re);
                    atomPath.setResidue(residue);
                    for(int at=0; at<residue.numAtoms(); at++)
                    {
                        Atom atom = residue.getAtom(at);
                        atomPath.setAtom(atom);
                        if( expr.match(atomPath) ){
                            v.x = atom.getX();
                            v.y = atom.getY();
                            v.z = atom.getZ();
                            v.transform(molecule.getImage().getTransform());
                            center.add(v);
                            cnt++;
                        }
                    }
                }
            }
        }}

        if(cnt == 0) return null;

        center.x /= cnt;
        center.y /= cnt;
        center.z /= cnt;
        center.w = 1;

        return center;
    }

    /**
     * Eliminates hetero atoms.
     */
    public void eliminateHeteroAtoms()
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
            Residue residue = chain.getResidue(re);
            for(int at=0; at<residue.numAtoms(); at++)
            {
                Atom atom = residue.getAtom(at);
                if( atom.isHetero() ){
                    atom.getImage().setModel(AtomImage.MODEL_NONE);
                }
            }}}

            if( molecule.builtBond() )
            {
                for(int b=0; b<molecule.numBonds(); b++)
                {
                    Bond bond = molecule.getBond(b);
                    Atom src = bond.getSrcAtom();
                    Atom dst = bond.getDstAtom();
                    if(src.isHetero() || dst.isHetero()){
                        bond.getImage().setModel(BondImage.MODEL_NONE);
                    }
                }    
            }
        }}
    }

    /**
     * Eliminates hydrogens.
     */
    public void eliminateHydrogens()
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
            Residue residue = chain.getResidue(re);
            for(int at=0; at<residue.numAtoms(); at++)
            {
                Atom atom = residue.getAtom(at);
                if( atom.getAtomicNo() == 1 ){
                    atom.getImage().setModel(AtomImage.MODEL_NONE);
                }
            }}}

            if( molecule.builtBond() )
            {
                for(int b=0; b<molecule.numBonds(); b++)
                {
                    Bond bond = molecule.getBond(b);
                    Atom src = bond.getSrcAtom();
                    Atom dst = bond.getDstAtom();
                    if(src.getAtomicNo() == 1 || dst.getAtomicNo() == 1){
                        bond.getImage().setModel(BondImage.MODEL_NONE);
                    }
                }    
            }
        }}
    }

    //----------------------------------------------------------------------
    // For selected files
    //----------------------------------------------------------------------
    /**
     * Switches the hydrogen bonds.
     */
    public void setHBondBackbone(boolean enable)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( molecule.getFile().isSelected() ){
                molecule.getImage().setHBondBackbone(enable);
            }
        }}
    }

    /**
     * Switches the disulfide bonds.
     */
    public void setSSBondBackbone(boolean enable)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( molecule.getFile().isSelected() ){
                molecule.getImage().setSSBondBackbone(enable);
            }
        }}
    }

    /**
     * Sets the height of cartoon model.
     */
    public void setCartoonHeight(float height)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( molecule.getFile().isSelected() ){
                molecule.getImage().setCartoonHeight(height);
                molecule.getImage().setRefreshResidue(true);
            }
        }}
    }

    /**
     * Sets the back color of ribbon/cartoon model.
     */
    public void setResidueBackColor(Color color)
    {
        int colorIndex = _colorManager.getIndex(color);
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( molecule.getFile().isSelected() ){
                molecule.getImage().setResidueBackColorIndex(colorIndex);
            }
        }}
    }

    /**
     * Resets the back color of ribbon/cartoon model.
     */
    public void resetResidueBackColor()
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( molecule.getFile().isSelected() ){
                molecule.getImage().setResidueBackColorIndex(-1);
            }
        }}
    }

    /**
     * Switches the loop cartoon.
     */
    public void setCartoonLoopTube(boolean enable)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( molecule.getFile().isSelected() ){
                molecule.getImage().setCartoonLoopTube(enable);
                for(int ch=0; ch<molecule.numChains(); ch++)
                {
                Chain chain = molecule.getChain(ch);
                for(int re=0; re<chain.numResidues(); re++)
                {
                    Residue residue = chain.getResidue(re);
                    if (enable) {
                        if (residue.getImage().getModel() == ResidueImage.MODEL_CARTOON) {
                            if (residue.isProtein()
                                    && residue.getStructure() != Residue.STRUCT_HELIX
                                    && residue.getStructure() != Residue.STRUCT_SHEET) {
                                residue.getImage().setModel(ResidueImage.MODEL_CARTOON_TUBE);
                            }
                        }
                    } else {
                        if (residue.getImage().getModel() == ResidueImage.MODEL_CARTOON_TUBE) {
                            residue.getImage().setModel(ResidueImage.MODEL_CARTOON);
                        }
                    }
                }}
                molecule.getImage().setRefreshResidue(true);
            }
        }}
    }

    /**
     * Switches the cartoon round.
     */
    public void setCartoonRound(boolean enable)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( molecule.getFile().isSelected() ){
                molecule.getImage().setCartoonRound(enable);
                molecule.getImage().setRefreshResidue(true);
            }
        }}
    }

    /**
     * displays atoms.
     */
    public void display(boolean enable) {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            for (int ch = 0; ch < molecule.numChains(); ch++) {
                Chain chain = molecule.getChain(ch);
                for (int b = 0; b < chain.numBackbones(); b++) {
                    Bond backbone = chain.getBackbone(b);
                    if (isSelected(backbone)) {
                        backbone.getImage().setDisplayed(enable);
                    }
                }
                for (int re = 0; re < chain.numResidues(); re++) {
                    Residue residue = chain.getResidue(re);
                    if (isSelected(residue)) {
                        residue.getImage().setDisplayed(enable);
                    }
                    for (int at = 0; at < residue.numAtoms(); at++) {
                        Atom atom = residue.getAtom(at);
                        if (atom.isSelected()) {
                            atom.setDisplayed(enable);
                            atom.getImage().setDisplayed(enable);
                        }
                    }
                }
            }
            for (int b = 0; b < molecule.numBonds(); b++) {
                Bond bond = molecule.getBond(b);
                if (isSelected(bond)) {
                    bond.getImage().setDisplayed(enable);
                }
            }
            for (int b = 0; b < molecule.numHBonds(); b++) {
                HBond hbond = molecule.getHBond(b);
                if (isSelected(hbond)) {
                    hbond.getImage().setDisplayed(enable);
                    hbond.getImageCA().setDisplayed(enable);
                }
            }
            for (int b = 0; b < molecule.numSSBonds(); b++) {
                SSBond ssbond = molecule.getSSBond(b);
                if (isSelected(ssbond)) {
                    ssbond.getImage().setDisplayed(enable);
                    ssbond.getImageCA().setDisplayed(enable);
                }
            }
            molecule.getImage().setRefreshResidue(true);
        }}
    }

    //----------------------------------------------------------------------
    // Select
    //----------------------------------------------------------------------
    private int select(
        SelectExpr expr, boolean enable, boolean hetero, boolean hydrogen)
    {
        if(expr != null) expr.setup(_moleculeFileArray);

        AtomPath atomPath = new AtomPath();

        int cnt = 0;
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        //for(int m=0; m<mf.numMolecules(); m++)
        //{
            //Molecule molecule = mf.getMolecule(m);
            Molecule molecule = mf.getMolecule(mf.getCurrentFrame());

            atomPath.setMoleculeFile(molecule.getFile());            
            atomPath.setMolecule(molecule);
            int currCnt = 0;
            
            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                atomPath.setChain(chain);
                for(int re=0; re<chain.numResidues(); re++)
                {
                    Residue residue = chain.getResidue(re);
                    atomPath.setResidue(residue);
                    for(int at=0; at<residue.numAtoms(); at++)
                    {
                        Atom atom = residue.getAtom(at);
                        atomPath.setAtom(atom);
                        
                        if(expr != null){
                            if(expr.match(atomPath)){
                                atom.setSelected(true);
                                currCnt++;
                            }
                            else{
                                atom.setSelected(false);
                            }
                        }
                        else if(enable){
                            if( atom.isHetero() && !hetero ||
                                atom.getAtomicNo() == 1 && !hydrogen )
                            {
                                atom.setSelected(false);
                            }
                            else{
                                atom.setSelected(true);
                                currCnt++;
                            }
                        }
                        else{
                            atom.setSelected(false);
                        }
                    }
                }
            }
            molecule.setSelectAtomCount(currCnt);
            cnt += currCnt;    
        //}
        }

        return cnt;
    }
    
    //----------------------------------------------------------------------
    // Wireframe
    //----------------------------------------------------------------------
    private void wireframe(boolean enable, float radius)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( !molecule.builtBond() ){
                MoleculeUtil util = new MoleculeUtil();
                util.buildBonds(molecule);
            }

            for(int b=0; b<molecule.numBonds(); b++)
            {
                Bond bond = molecule.getBond(b);

                if( isSelected(bond) ){
                    if(enable){
                        bond.getImage().setModel(BondImage.MODEL_WIREFRAME);
                        bond.getImage().setRadius(radius);
                    }
                    else{
                        bond.getImage().setModel(BondImage.MODEL_NONE);
                    }
                }
            }    
        }}
    }

    //----------------------------------------------------------------------
    // Spacefill
    //----------------------------------------------------------------------
    private void spacefill(
        boolean enable, boolean vdw, boolean temp, float radius)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
        Residue residue = chain.getResidue(re);
        for(int at=0; at<residue.numAtoms(); at++)
        {
            Atom atom = residue.getAtom(at);
            if( atom.isSelected() ){
                if(enable){
                    atom.getImage().setModel(AtomImage.MODEL_SPACEFILL);
                    if(vdw){
                        atom.getImage().setRadius(getVdWRadius(molecule, atom));
                    }
                    else if(temp){
                        if(atom.getTempFactor() <= 0){
                            atom.getImage().setModel(AtomImage.MODEL_NONE);
                        }
                        else{
                            float r = atom.getTempFactor();
                            if(r > MAX_TEMP_FACTOR_RADIUS) r = MAX_TEMP_FACTOR_RADIUS;
                            atom.getImage().setRadius(r);
                        }
                    }
                    else{
                        atom.getImage().setRadius(radius);
                    }
                }
                else{
                    atom.getImage().setModel(AtomImage.MODEL_NONE);
                }
            }
        }}}}}
    }

    private float getVdWRadius(Molecule molecule, Atom atom)
    {
        if(atom.getAtomicNo() == 6 && !molecule.hasHydrogens()){
            return ElementTable.VDW_CARBON;
        }
        else if(atom.getAtomicNo() == 7 && !molecule.hasHydrogens()){
            return ElementTable.VDW_NITROGEN;
        }
        else if(atom.getAtomicNo() == 8 && !molecule.hasHydrogens()){
            return ElementTable.VDW_OXYGEN;
        }
        else if(atom.getAtomicNo() == 10 && !molecule.hasHydrogens()){
            return ElementTable.VDW_SULFUR;
        }
        else{
            return ElementTable.getVdWRadius(atom.getAtomicNo());
        }
    }

    //----------------------------------------------------------------------
    // Backbone
    //----------------------------------------------------------------------
    private void backbone(boolean enable, float radius)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( !molecule.builtBackbone() ){
                MoleculeUtil util = new MoleculeUtil();
                util.buildBackbones(molecule);
            }

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                backbone(chain, enable, radius);
            }
        }}
    }

    private void backbone(Chain chain, boolean enable, float radius)
    {
        for(int b=0; b<chain.numBackbones(); b++)
        {
            Bond backbone = chain.getBackbone(b);
            if( isSelected(backbone) ){
                if(enable){
                    backbone.getImage().setModel(BondImage.MODEL_WIREFRAME);
                    backbone.getImage().setRadius(radius);
                }
                else{
                    backbone.getImage().setModel(BondImage.MODEL_NONE);
                }
            }
        }
    }

    //----------------------------------------------------------------------
    // HBond
    //----------------------------------------------------------------------
    private void hBond(boolean enable, float radius)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( !molecule.builtHBond() ){
                MoleculeUtil util = new MoleculeUtil();
                util.buildHBonds(molecule);
            }

            for(int b=0; b<molecule.numHBonds(); b++)
            {
                HBond hbond = molecule.getHBond(b);
                if( isSelected(hbond) ){
                    if(enable){
                        hbond.getImage().setModel(BondImage.MODEL_WIREFRAME);
                        hbond.getImageCA().setModel(BondImage.MODEL_WIREFRAME);
                        hbond.getImage().setRadius(radius);
                        hbond.getImageCA().setRadius(radius);
                    }
                    else{
                        hbond.getImage().setModel(BondImage.MODEL_NONE);
                        hbond.getImageCA().setModel(BondImage.MODEL_NONE);
                    }
                }
            }    
        }}
    }

    //----------------------------------------------------------------------
    // SSBond
    //----------------------------------------------------------------------
    private void ssBond(boolean enable, float radius)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if( !molecule.builtSSBond() ){
                MoleculeUtil util = new MoleculeUtil();
                util.buildSSBonds(molecule);
            }

            for(int b=0; b<molecule.numSSBonds(); b++)
            {
                SSBond ssbond = molecule.getSSBond(b);
                if( isSelected(ssbond) ){
                    if(enable){
                        ssbond.getImage().setModel(BondImage.MODEL_WIREFRAME);
                        ssbond.getImageCA().setModel(BondImage.MODEL_WIREFRAME);
                        ssbond.getImage().setRadius(radius);
                        ssbond.getImageCA().setRadius(radius);
                    }
                    else{
                        ssbond.getImage().setModel(BondImage.MODEL_NONE);
                        ssbond.getImageCA().setModel(BondImage.MODEL_NONE);
                    }
                }
            }
        }}
    }

    //----------------------------------------------------------------------
    // Ribbon & Cartoon
    //----------------------------------------------------------------------
    private void ribbon(
        boolean enable, float width, boolean cartoon, boolean typical)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if(enable){
                if(molecule.getStructure() == Molecule.STRUCT_NONE){
                    MoleculeUtil util = new MoleculeUtil();
                    util.determineStructure(molecule);
                }
                /*
                if( !molecule.builtBackbone() ){
                    MoleculeUtil util = new MoleculeUtil();
                    util.buildBackbones(molecule);
                }
                */
            }
            if (!molecule.isRibbonChecked()) {
                MoleculeUtil util = new MoleculeUtil();
                util.checkRibbon(molecule);
            }

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                if (chain.isRibbonDrawable()) {
                    ribbon(molecule, chain, enable, width, cartoon, typical);
                } else {
                    backbone(chain, enable, 0.3f);
                }
            }
        }}
    }

    private void ribbon(Molecule molecule, Chain chain, boolean enable,
            float width, boolean cartoon, boolean typical)
    {
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            if(!isSelected(residue)) continue;

            molecule.getImage().setRefreshResidue(true);
            if(enable)
            {        
                if(cartoon){
                    if (molecule.getImage().getCartoonLoopTube()
                                && residue.isProtein()
                                && residue.getStructure() != Residue.STRUCT_HELIX
                                && residue.getStructure() != Residue.STRUCT_SHEET) {
                        residue.getImage().setModel(ResidueImage.MODEL_CARTOON_TUBE);
                    } else {
                        residue.getImage().setModel(ResidueImage.MODEL_CARTOON);
                    }
                }
                else{
                    residue.getImage().setModel(ResidueImage.MODEL_RIBBON);
                }

                if(typical){
                    if( residue.isProtein() )
                    {
                        if( residue.getStructure() == Residue.STRUCT_HELIX ||
                            residue.getStructure() == Residue.STRUCT_SHEET )
                        {
                            width = 320.0f/250;
                        }
                        else{
                            width = 50.0f/250;
                        }
                    }
                    else if( residue.isNucleotide() )
                    {
                        width = 720.0f/250;
                    }
                    else{
                        width = 50.0f/250;
                    }
                }
                residue.getImage().setWidth(width);
            }
            else{
                residue.getImage().setModel(ResidueImage.MODEL_NONE);
            }
        }
    }

    //----------------------------------------------------------------------
    // Trace
    //----------------------------------------------------------------------
    private void trace(boolean enable, float radius)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            if(enable){
                if(molecule.getStructure() == Molecule.STRUCT_NONE){
                    MoleculeUtil util = new MoleculeUtil();
                    util.determineStructure(molecule);
                }
                /*
                if( !molecule.builtBackbone() ){
                    MoleculeUtil util = new MoleculeUtil();
                    util.buildBackbones(molecule);
                }
                */
            }
            if (!molecule.isRibbonChecked()) {
                MoleculeUtil util = new MoleculeUtil();
                util.checkRibbon(molecule);
            }

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                if (chain.isRibbonDrawable()) {
                    trace(molecule, chain, enable, radius);
                } else {
                    backbone(chain, enable, radius);
                }
            }
        }}
    }

    private void trace(Molecule molecule, Chain chain, boolean enable, float radius)
    {
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            if(!isSelected(residue)) continue;

            molecule.getImage().setRefreshResidue(true);
            if(enable){
                residue.getImage().setModel(ResidueImage.MODEL_TRACE);
                residue.getImage().setWidth(radius);
            }
            else{
                residue.getImage().setModel(ResidueImage.MODEL_NONE);
            }
        }
    }

    //----------------------------------------------------------------------
    // Color
    //----------------------------------------------------------------------
    private void colorAtom(int colorIndex)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
        Residue residue = chain.getResidue(re);
        for(int at=0; at<residue.numAtoms(); at++)
        {
            Atom atom = residue.getAtom(at);
            if( atom.isSelected() ){
                atom.getImage().setColorIndex(colorIndex);
            }
        }}}}}
    }

    private void colorCPK()
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
        Residue residue = chain.getResidue(re);
        for(int at=0; at<residue.numAtoms(); at++)
        {
            Atom atom = residue.getAtom(at);
            if( atom.isSelected() ){
                atom.getImage().setColorIndex(
                    _colorManager.getCPKIndex(atom.getAtomicNo()) );
            }
        }}}}}
    }

    private void colorAmino()
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            int colorIndex = _colorManager.getAminoIndex(residue.getIndex());
            for(int at=0; at<residue.numAtoms(); at++)
            {
                Atom atom = residue.getAtom(at);
                if( atom.isSelected() ){
                    atom.getImage().setColorIndex(colorIndex);
                }
            }
        }}}}
    }

    private void colorShapely()
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            int colorIndex = _colorManager.getShapelyIndex(residue.getIndex());
            for(int at=0; at<residue.numAtoms(); at++)
            {
                Atom atom = residue.getAtom(at);
                if( atom.isSelected() ){
                    atom.getImage().setColorIndex(colorIndex);
                }
            }
        }}}}
    }

    private void colorChain()
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);
            int[] colors = getSpectrumColors(molecule.numChains());

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                for(int re=0; re<chain.numResidues(); re++)
                {
                Residue residue = chain.getResidue(re);
                for(int at=0; at<residue.numAtoms(); at++)
                {
                    Atom atom = residue.getAtom(at);
                    if( atom.isSelected() ){
                        atom.getImage().setColorIndex(colors[ch]);
                    }
                }}
            }
        }}
    }

    private void colorGroup()
    {
        int numColors = 31;
        int[] colors = getSpectrumColors(numColors);

        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            int div = chain.numResidues() - 1;
            if(div == 0) div = 1;

            for(int re=0; re<chain.numResidues(); re++)
            {
                Residue residue = chain.getResidue(re);
                int index = re*(numColors-1)/div;
                for(int at=0; at<residue.numAtoms(); at++)
                {
                    Atom atom = residue.getAtom(at);
                    if( atom.isSelected() ){
                        atom.getImage().setColorIndex(colors[index]);
                    }
                }
            }
        }}}
    }

    private void colorTemp(boolean reverse)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);

            float min = molecule.getMinTempFactor();
            float max = molecule.getMaxTempFactor();
            if(molecule.hasHeteroAtoms() && _selectionStatus.getHetero()){
                float minHetero = molecule.getMinHeteroTempFactor();
                float maxHetero = molecule.getMaxHeteroTempFactor();
                if(minHetero < min) min = minHetero;
                if(maxHetero > max) max = maxHetero;
            }
            
            float d = max - min;
            int numColors = (d == 0) ? 1 : 31;
            int[] colors = getSpectrumColors(numColors);

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
            Residue residue = chain.getResidue(re);
            for(int at=0; at<residue.numAtoms(); at++)
            {
                Atom atom = residue.getAtom(at);
                if( atom.isSelected() ){
                    if(d == 0){
                        atom.getImage().setColorIndex(colors[0]);
                    }
                    else{
                        float temp = atom.getTempFactor();
                        int index = (int)( (numColors-1)*(temp-min)/d );
                        if(reverse) index = (numColors-1) - index;
                        atom.getImage().setColorIndex(colors[index]);
                    }
                }
            }}}
        }}
    }

    private void colorStruct()
    {
        int[] colors = new int[4];
        colors[0] = _colorManager.getIndex(255,255,255);
        colors[1] = _colorManager.getIndex(255,  0,128);
        colors[2] = _colorManager.getIndex(255,200,  0);
        colors[3] = _colorManager.getIndex( 96,128,255);

        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
            Molecule molecule = mf.getMolecule(m);

            if(molecule.getStructure() == Molecule.STRUCT_NONE){
                MoleculeUtil util = new MoleculeUtil();
                util.determineStructure(molecule);
            }

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
                Residue residue = chain.getResidue(re);
                int index = 0;
                if(residue.getStructure() == Residue.STRUCT_HELIX){
                    index = 1;
                }
                else if(residue.getStructure() == Residue.STRUCT_SHEET){
                    index = 2;
                }
                else if(residue.getStructure() == Residue.STRUCT_TURN){
                    index = 3;
                }
    
                for(int at=0; at<residue.numAtoms(); at++)
                {
                    Atom atom = residue.getAtom(at);
                    if( atom.isSelected() ){
                        atom.getImage().setColorIndex(colors[index]);
                    }
                }
            }
        }}}
    }

    private void colorDssp()
    {
        int nodssp = _colorManager.getIndex(new Color(0x99ccff));
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
        Residue residue = chain.getResidue(re);
        int colorIndex = residue.getColorIndexGromacs();
        for(int at=0; at<residue.numAtoms(); at++)
        {
            Atom atom = residue.getAtom(at);
            if( atom.isSelected() ){
                if (colorIndex >= 0) {
                    atom.getImage().setColorIndex(colorIndex);
                } else {
                    atom.getImage().setColorIndex(nodssp);
                }
            }
        }}}}}
    }

    private void colorResidueDssp()
    {
        int nodssp = _colorManager.getIndex(new Color(0x99ccff));
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            int colorIndex = residue.getColorIndexGromacs();
            if( isSelected(residue) ){
                if (colorIndex >= 0) {
                    residue.getImage().setColorIndex(colorIndex);
                } else {
                    residue.getImage().setColorIndex(nodssp);
                }
            }
        }}}}
    }

    private void colorResidueBackDssp()
    {
        int nodssp = _colorManager.getIndex(new Color(0x99ccff));
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            int colorIndex = residue.getColorIndexGromacs();
            if( isSelected(residue) ){
                if (colorIndex >= 0) {
                    residue.getImage().setBackColorIndex(colorIndex);
                } else {
                    residue.getImage().setBackColorIndex(nodssp);
                }
            }
        }}}}
    }

    private void colorBond(int colorIndex)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int b=0; b<molecule.numBonds(); b++)
        {
            Bond bond = molecule.getBond(b);
            if( isSelected(bond) ){
                bond.getImage().setColorIndex(colorIndex);
            }
        }}}
    }

    private void colorBackbone(int colorIndex)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int b=0; b<chain.numBackbones(); b++)
        {
            Bond backbone = chain.getBackbone(b);
            if( isSelected(backbone) ){
                backbone.getImage().setColorIndex(colorIndex);
            }
        }}}}
    }

    private void colorHBond(int colorIndex)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int b=0; b<molecule.numHBonds(); b++)
        {
            HBond hbond = molecule.getHBond(b);
            if( isSelected(hbond) ){
                hbond.getImage().setColorIndex(colorIndex);
                hbond.getImageCA().setColorIndex(colorIndex);
            }
        }}}
    }

    private void colorHBondType()
    {
        int[] colors = new int[7];
        colors[0] = _colorManager.getIndex(255,255,255);
        colors[1] = _colorManager.getIndex(255,  0,255);
        colors[2] = _colorManager.getIndex(255,  0,  0);
        colors[3] = _colorManager.getIndex(255,165,  0);
        colors[4] = _colorManager.getIndex(  0,255,255);
        colors[5] = _colorManager.getIndex(  0,255,  0);
        colors[6] = _colorManager.getIndex(255,255,  0);

        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int b=0; b<molecule.numHBonds(); b++)
        {
            HBond hbond = molecule.getHBond(b);
            if( isSelected(hbond) ){
                int index = 6;
                if(hbond.getOffset() == 2){
                    index = 0;
                }
                else if(hbond.getOffset() == 3){
                    index = 1;
                }
                else if(hbond.getOffset() == 4){
                    index = 2;
                }
                else if(hbond.getOffset() == 5){
                    index = 3;
                }
                else if(hbond.getOffset() == -3){
                    index = 4;
                }
                else if(hbond.getOffset() == -4){
                    index = 5;
                }
                hbond.getImage().setColorIndex(colors[index]);
                hbond.getImageCA().setColorIndex(colors[index]);
            }
        }}}
    }

    private void colorSSBond(int colorIndex)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int b=0; b<molecule.numSSBonds(); b++)
        {
            SSBond ssbond = molecule.getSSBond(b);
            if( isSelected(ssbond) ){
                ssbond.getImage().setColorIndex(colorIndex);
                ssbond.getImageCA().setColorIndex(colorIndex);
            }
        }}}
    }

    private void colorResidue(int colorIndex)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            if( isSelected(residue) ){
                residue.getImage().setColorIndex(colorIndex);
            }
        }}}}
    }

    private void colorResidueBack(int colorIndex)
    {
        for(int f=0; f<_moleculeFiles.size(); f++)
        {
        MoleculeFile mf = (MoleculeFile) _moleculeFiles.get(f);
        for(int m=0; m<mf.numMolecules(); m++)
        {
        Molecule molecule = mf.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            if( isSelected(residue) ){
                residue.getImage().setBackColorIndex(colorIndex);
            }
        }}}}
    }

    private int[] getSpectrumColors(int numColors)
    {
        int[] colors = new int[numColors];

        if(numColors == 1){
            colors[0] = _colorManager.getIndex(255,255,255);
            return colors;
        }

        for(int i=0; i<numColors; i++){
            int fraction = (int)(1023*i/(numColors - 1));
            int r, g, b;
            if( fraction < 256 ){
                r = 0;  g = fraction;  b = 255;
            }
            else if( fraction < 512 ){
                r = 0;  g = 255;  b = 511-fraction;
            }
            else if( fraction < 768 ){
                r = fraction-512;  g = 255;  b = 0;
            }
            else{
                r = 255;  g = 1023-fraction;  b = 0;
            }

            colors[i] = _colorManager.getIndex(r, g, b);
        }

        return colors;
    }

    //----------------------------------------------------------------------
    // Tests selected
    //----------------------------------------------------------------------
    private boolean isSelected(Bond bond)
    {
        if(_selectionStatus.getBondMode() == SelectionStatus.BOND_AND){
            if( bond.getSrcAtom().isSelected() &&
                bond.getDstAtom().isSelected()   )
            {
                return true;
            }
        }
        else{
            if( bond.getSrcAtom().isSelected() ||
                bond.getDstAtom().isSelected()   )
            {
                return true;
            }
        }
        return false;    
    }

    private boolean isSelected(Residue residue)
    {
        Atom atomCA = null;
        if (residue.isProtein()) {
            atomCA = residue.getProteinCA();
        } else if (residue.isNucleotide()) {
            atomCA = residue.getNucleotideP();
        }

        return ( atomCA != null && atomCA.isSelected() );    
    }

    //----------------------------------------------------------------------
    // Molecule list
    //----------------------------------------------------------------------
    private void updateMolecules()
    {
        _moleculeFileArray = (MoleculeFile[])_moleculeFiles.toArray( new MoleculeFile[0] );
    }
}

