/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.manipulators;

import java.awt.*;
import java.util.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;

/**
* Class that manipulates polygons.
*/
public class PolygonManipulator implements PolygonSelection
{
    private ArrayList _polygonFiles;
    private ArrayList _moleculeFiles;

    private MoleculeFile[] _moleculeFileArray;

    /**
    * Constructor.
    */
    public PolygonManipulator()
    {
        _polygonFiles = new ArrayList();
        _moleculeFiles = new ArrayList();
        this.updateMolecules();
    }

    /**
    * Appends a polygon file.
    */
    public void add(PolygonFile polygonFile)
    {
        _polygonFiles.add( polygonFile );
    }

    /**
     * Removes a polygon file.
     */
    public void remove(PolygonFile polygonFile)
    {
        _polygonFiles.remove( polygonFile );
    }

    /**
    * Appends a molecule file.
    */
    public void add(MoleculeFile moleculeFile)
    {
        _moleculeFiles.add( moleculeFile );
        this.updateMolecules();
    }

    /**
     * Removes a molecule file.
     */
    public void remove(MoleculeFile moleculeFile)
    {
        _moleculeFiles.remove( moleculeFile );
        this.updateMolecules();
    }

    /**
    * Selects all vertices.
    */
    public int selectAll()
    {
        return this.select(null, true);
    }

    /**
    * Selects vertices.
    */
    public int select(SelectExpr expr)
    {
        return this.select(expr, false);
    }

    /**
    * Selects no vertices.
    */
    public int selectNone()
    {
        return this.select(null, false);
    }

    /**
    * displays vertices.
    */
    public void display(boolean enable)
    {
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            BitSet selectFlags = file.getSelectFlags();
            BitSet displayFlags = file.getDisplayFlags();

            int num = file.getVertices().length();
            for(int j=0; j<num; j++){
                if(selectFlags.get(j)){
                    displayFlags.set(j, enable);
                }
            }

            file.getImage().setRefresh(true);
        }
    }

    /**
     * Sets the color of vertices.
     */
    public void setColor(Color color)
    {
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            if( !file.isSelected() ) continue;

            BitSet selectFlag = file.getSelectFlags();
            int[] currColors = file.getImage().getColors();
            for(int j=0; j<currColors.length; j++){
                if( selectFlag.get(j) ) currColors[j] = color.getRGB();
            }
            file.getImage().setRefresh(true);
        }
    }

    /**
     * Resets the color of vertices.
     */
    public void resetColor()
    {
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            if( !file.isSelected() ) continue;

            BitSet selectFlag = file.getSelectFlags();
            Vertices vertices = file.getVertices();
            int[] currColors = file.getImage().getColors();
            for(int j=0; j<currColors.length; j++){
                if( selectFlag.get(j) ) currColors[j] = vertices.getColor(j);
            }
            file.getImage().setRefresh(true);
        }
    }

    /**
     * Sets the transparency of polygons.
     */
    public void setTransparency(int target, float transparency)
    {
        if(transparency < 0 || transparency > 1) return;

        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            if( !file.isSelected() ) continue;

            if(target == POLYGON_POINTS || target == POLYGON_ALL){
                file.getImage().setPointTransparency(transparency);
            }
            if(target == POLYGON_LINES || target == POLYGON_ALL){
                file.getImage().setLineTransparency(transparency);
            }
            if(target == POLYGON_TRIANGLES || target == POLYGON_ALL){
                file.getImage().setTriangleTransparency(transparency);
            }
            if(target == POLYGON_QUADS || target == POLYGON_ALL){
                file.getImage().setQuadTransparency(transparency);
            }
            if(target == POLYGON_POLYLINES || target == POLYGON_ALL){
                file.getImage().setPolylineTransparency(transparency);
            }
            file.getImage().setRefresh(true);
        }
    }

    /**
     * Resets the transparency of polygons.
     */
    public void resetTransparency(int target)
    {
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            if( !file.isSelected() ) continue;

            if(target == POLYGON_POINTS || target == POLYGON_ALL){
                file.getImage().setPointTransparency( file.getPointTransparency() );
            }
            if(target == POLYGON_LINES || target == POLYGON_ALL){
                file.getImage().setLineTransparency( file.getLineTransparency() );
            }
            if(target == POLYGON_TRIANGLES || target == POLYGON_ALL){
                file.getImage().setTriangleTransparency( file.getTriangleTransparency() );
            }
            if(target == POLYGON_QUADS || target == POLYGON_ALL){
                file.getImage().setQuadTransparency( file.getQuadTransparency() );
            }
            if(target == POLYGON_POLYLINES || target == POLYGON_ALL){
                file.getImage().setPolylineTransparency( file.getPolylineTransparency() );
            }
            file.getImage().setRefresh(true);
        }
    }

    /**
     * Sets the size of polygons (point size and line width).
     */
    public void setPolygonSize(int target, float size)
    {
        if(size <= 0) return;
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            if( !file.isSelected() ) continue;

            if(target == POLYGON_POINTS){
                file.getImage().setPointSize(size);
            }
            else if(target == POLYGON_LINES){
                file.getImage().setLineWidth(size);
            }
            else if(target == POLYGON_POLYLINES){
                file.getImage().setPolylineWidth(size);
            }
        }
    }

    /**
     * Resets the size of polygons (point size and line width).
     */
    public void resetPolygonSize(int target)
    {
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            if( !file.isSelected() ) continue;

            if(target == POLYGON_POINTS){
                file.getImage().setPointSize(file.getPointSize());
            }
            else if(target == POLYGON_LINES){
                file.getImage().setLineWidth(file.getLineWidth());
            }
            else if(target == POLYGON_POLYLINES){
                file.getImage().setPolylineWidth(file.getPolylineWidth());
            }
        }
    }

    //----------------------------------------------------------------------
    // Select
    //----------------------------------------------------------------------
    private int select(SelectExpr expr, boolean enable)
    {
        if(expr != null) expr.setup(_moleculeFileArray);

        int cnt = 0;
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_polygonFiles.get(i);
            BitSet selectFlags = file.getSelectFlags();
            
            int currCnt = 0;
            int num = file.getVertices().length();
            for (int j = 0; j < num; j++)
            {
                if(expr != null){
                    if( expr.match(file, j) ){
                        selectFlags.set(j);
                        currCnt++;
                    }
                    else{
                        selectFlags.clear(j);
                    }
                }
                else if(enable){
                    selectFlags.set(j);
                    currCnt++;
                }
                else{
                    selectFlags.clear(j);
                }
            }
            file.setSelectVertexCount(currCnt);
            cnt += currCnt;    
        }

        return cnt;
    }

    //----------------------------------------------------------------------
    // Molecule list
    //----------------------------------------------------------------------
    private void updateMolecules()
    {
        _moleculeFileArray = (MoleculeFile[])_moleculeFiles.toArray( new MoleculeFile[0] );
    }
}
