/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.manipulators;

import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.molecules.*;

/**
* Class that represents predefined atom set.
*/
public class PredefinedSetExpr extends SelectExpr
{
    public static final int INDEX_AT          =  0;
    public static final int INDEX_ACIDIC      =  1;
    public static final int INDEX_ACYCLIC     =  2;
    public static final int INDEX_ALIPHATIC   =  3;
    public static final int INDEX_ALPHA       =  4;
    public static final int INDEX_AMINO       =  5;
    public static final int INDEX_AROMATIC    =  6;
    public static final int INDEX_BACKBONE    =  7;
    public static final int INDEX_BASIC       =  8;
    public static final int INDEX_BONDED      =  9;
    public static final int INDEX_BURIED      = 10;
    public static final int INDEX_CG          = 11;
    public static final int INDEX_CHARGED     = 12;
    public static final int INDEX_CYCLIC      = 13;
    public static final int INDEX_CYSTEINE    = 14;
    public static final int INDEX_HELIX       = 15;
    public static final int INDEX_HETERO      = 16;
    public static final int INDEX_HYDROGEN    = 17;
    public static final int INDEX_HYDROPHOBIC = 18;
    public static final int INDEX_IONS        = 19;
    public static final int INDEX_LARGE       = 20;
    public static final int INDEX_LIGAND      = 21;
    public static final int INDEX_MEDIUM      = 22;
    public static final int INDEX_NEGATIVE    = 23;
    public static final int INDEX_NEUTRAL     = 24;
    public static final int INDEX_NUCLEIC     = 25;
    public static final int INDEX_POLAR       = 26;
    public static final int INDEX_POSITIVE    = 27;
    public static final int INDEX_PROTEIN     = 28;
    public static final int INDEX_PURINE      = 29;
    public static final int INDEX_PYRIMIDINE  = 30;
    public static final int INDEX_SELECTED    = 31;
    public static final int INDEX_SHEET       = 32;
    public static final int INDEX_SIDECHAIN   = 33;
    public static final int INDEX_SMALL       = 34;
    public static final int INDEX_SOLVENT     = 35;
    public static final int INDEX_SURFACE     = 36;
    public static final int INDEX_TURN        = 37;
    public static final int INDEX_WATER       = 38;
    public static final int INDEX_MAINCHAIN   = 39;
    public static final int INDEX_DNA         = 40;
    public static final int INDEX_RNA         = 41;

    private static String[] _setNames
        ={
            "AT",           //  0
            "ACIDIC",       //  1
            "ACYCLIC",      //  2
            "ALIPHATIC",    //  3
            "ALPHA",        //  4
            "AMINO",        //  5
            "AROMATIC",     //  6
            "BACKBONE",     //  7
            "BASIC",        //  8
            "BONDED",       //  9
            "BURIED",       // 10
            "CG",           // 11
            "CHARGED",      // 12
            "CYCLIC",       // 13
            "CYSTEINE",     // 14
            "HELIX",        // 15
            "HETERO",       // 16
            "HYDROGEN",     // 17
            "HYDROPHOBIC",  // 18
            "IONS",         // 19
            "LARGE",        // 20
            "LIGAND",       // 21
            "MEDIUM",       // 22
            "NEGATIVE",     // 23
            "NEUTRAL",      // 24
            "NUCLEIC",      // 25
            "POLAR",        // 26
            "POSITIVE",     // 27
            "PROTEIN",      // 28
            "PURINE",       // 29
            "PYRIMIDINE",   // 30
            "SELECTED",     // 31
            "SHEET",        // 32
            "SIDECHAIN",    // 33
            "SMALL",        // 34
            "SOLVENT",      // 35
            "SURFACE",      // 36
            "TURN",         // 37
            "WATER",        // 38
            "MAINCHAIN",    // 39
            "DNA",          // 40
            "RNA"           // 41
        };

    private int _index;
    private BackboneAtomSet _backboneAtomSet;

    /**
    * Returns the number of predefined sets.
    */
    public static int counts()
    {
        return _setNames.length;
    }

    /**
    * Sets the index of predefined set.
    */
    public void setIndex(int index)
    {
        _index = index;
    }

    /**
    * Returns the index of predefined set.
    */
    public int getIndex()
    {
        return _index;
    }

    /**
    * Constructor.
    */
    public PredefinedSetExpr(int index, BackboneAtomSet backboneAtomSet)
    {
        _index = index;
        _backboneAtomSet = backboneAtomSet;
    }

    /**
    * Returns the index of the predefined set specified by name.
    */
    public static int getIndex(String setName)
    {
        String name = setName.toUpperCase();
        for(int i=0; i<_setNames.length; i++){
            if( name.equals(_setNames[i]) ) return i;
        }
        return -1;
    }

    /**
    * Returns the name of the specified set.
    */
    public static String getName(int index)
    {
        return _setNames[index];
    }

    /**
    * Tests whether the specified atom matches the expression.
    */
    public boolean match(AtomPath atomPath)
    {
        Molecule molecule = atomPath.getMolecule();
        Residue residue   = atomPath.getResidue();
        Atom atom         = atomPath.getAtom();

        switch(_index)
        {
            case INDEX_AT:
                return (residue.isDNA_A() || residue.isDNA_T());

            case INDEX_ACIDIC:
                return residue.hasProperty(ResidueTable.PROP_ACIDIC);

            case INDEX_ACYCLIC:
                return residue.hasProperty(ResidueTable.PROP_ACYCLIC);

            case INDEX_ALIPHATIC:
                return residue.hasProperty(ResidueTable.PROP_ALIPHATIC);

            case INDEX_ALPHA:
                return ( residue.isAmino() &&
                         atom.getIndex() == AtomTable.INDEX_CA          );

            case INDEX_AMINO:
                return residue.isAmino();

            case INDEX_AROMATIC:
                return residue.hasProperty(ResidueTable.PROP_AROMATIC);

            case INDEX_BACKBONE:
            case INDEX_MAINCHAIN:
                return ( ( residue.isAmino() &&
                           atom.isInGroup(AtomTable.GROUP_AMINO_BACKBONE) ) ||
                         ( residue.isNucleotide() &&
                           atom.isInGroup(AtomTable.GROUP_NUCLEIC_BACKBONE) ) ||
                         _backboneAtomSet.isBackboneAtom(residue.getName().trim(), atom.getName().trim())
                       );

            case INDEX_BASIC:
                return residue.hasProperty(ResidueTable.PROP_BASIC);

            case INDEX_BONDED:
                if( !molecule.builtBond() ){
                    MoleculeUtil util = new MoleculeUtil();
                    util.buildBonds(molecule);
                }
                return atom.isBonded();

            case INDEX_BURIED:
                return residue.hasProperty(ResidueTable.PROP_BURIED);

            case INDEX_CG:
                return (residue.isDNA_C() || residue.isDNA_G());

            case INDEX_CHARGED:
                return residue.hasProperty(ResidueTable.PROP_CHARGED);

            case INDEX_CYCLIC:
                return residue.hasProperty(ResidueTable.PROP_CYCLIC);

            case INDEX_CYSTEINE:
                if( !molecule.builtSSBond() ){
                    MoleculeUtil util = new MoleculeUtil();
                    util.buildSSBonds(molecule);
                }
                for(int i=0; i<molecule.numSSBonds(); i++){
                    SSBond ssbond = molecule.getSSBond(i);
                    if( atom == ssbond.getSrcAtom() ||
                        atom == ssbond.getDstAtom()   )
                    {
                        return true;
                    }
                }
                return false;

            case INDEX_HELIX:
                if(molecule.getStructure() == Molecule.STRUCT_NONE){
                    MoleculeUtil util = new MoleculeUtil();
                    util.determineStructure(molecule);
                }
                return (residue.getStructure() == Residue.STRUCT_HELIX);

            case INDEX_HETERO:
                return atom.isHetero();

            case INDEX_HYDROGEN:
                return (atom.getAtomicNo() == 1);

            case INDEX_HYDROPHOBIC:
                return residue.hasProperty(ResidueTable.PROP_HYDROPHOBIC);

            case INDEX_IONS:
                return residue.isInGroup(ResidueTable.GROUP_ION);

            case INDEX_LARGE:
                return residue.hasProperty(ResidueTable.PROP_LARGE);

            case INDEX_LIGAND:
                return ( atom.isHetero() &&
                         !residue.isInGroup(ResidueTable.GROUP_SOLVENT) );

            case INDEX_MEDIUM:
                return residue.hasProperty(ResidueTable.PROP_MEDIUM);

            case INDEX_NEGATIVE:
                return residue.hasProperty(ResidueTable.PROP_NEGATIVE);

            case INDEX_NEUTRAL:
                return residue.hasProperty(ResidueTable.PROP_NEUTRAL);

            case INDEX_NUCLEIC:
                return residue.isNucleotide();

            case INDEX_POLAR:
                return residue.hasProperty(ResidueTable.PROP_POLAR);

            case INDEX_POSITIVE:
                return residue.hasProperty(ResidueTable.PROP_POSITIVE);

            case INDEX_PROTEIN:
                return residue.isProtein();

            case INDEX_PURINE:
                return (residue.isDNA_A() || residue.isDNA_G() 
                        || residue.isRNA_A() || residue.isRNA_G());

            case INDEX_PYRIMIDINE:
                return (residue.isDNA_C() || residue.isDNA_T() 
                        || residue.isRNA_C() || residue.isRNA_U());

            case INDEX_SELECTED:
                return atom.isSelected();

            case INDEX_SHEET:
                if(molecule.getStructure() == Molecule.STRUCT_NONE){
                    MoleculeUtil util = new MoleculeUtil();
                    util.determineStructure(molecule);
                }
                return (residue.getStructure() == Residue.STRUCT_SHEET);

            case INDEX_SIDECHAIN:
                return ( !_backboneAtomSet.isBackboneAtom(residue.getName().trim(), atom.getName().trim()) &&
                         (( residue.isAmino() &&
                           !atom.isInGroup(AtomTable.GROUP_AMINO_BACKBONE) ) ||
                          ( residue.isNucleotide() &&
                           !atom.isInGroup(AtomTable.GROUP_NUCLEIC_BACKBONE) ) ||
                          _backboneAtomSet.isDefined(residue.getName().trim())
                         )
                       );

            case INDEX_SMALL:
                return residue.hasProperty(ResidueTable.PROP_SMALL);

            case INDEX_SOLVENT:
                return residue.isInGroup(ResidueTable.GROUP_SOLVENT);

            case INDEX_SURFACE:
                return residue.hasProperty(ResidueTable.PROP_SURFACE);

            case INDEX_TURN:
                if(molecule.getStructure() == Molecule.STRUCT_NONE){
                    MoleculeUtil util = new MoleculeUtil();
                    util.determineStructure(molecule);
                }
                return (residue.getStructure() == Residue.STRUCT_TURN);

            case INDEX_WATER:
                return residue.isInGroup(ResidueTable.GROUP_WATER);

            case INDEX_DNA:
                return residue.isDNA();

            case INDEX_RNA:
                return residue.isRNA();

            default: 
                return false;
        }
    }
}

