/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.manipulators;

import org.pdbj.viewer.molecules.*;

/**
* Class that represents primitive expressions.
*/
public class PrimitiveExpr extends SelectExpr
{
    public static final int NULL = Integer.MIN_VALUE;

    private int    _fileId;
    private int    _modelId;
    private String _chainId;
    private int    _residueId;
    private String _residueName;
    private String _atomName;
    private String _altLoc;

    /**
    * Sets the file ID.
    */
    public void setFileId(int fileId)
    {
        _fileId = fileId;
    }

    /**
    * Returns the file ID.
    */
    public int getFileId()
    {
        return _fileId;
    }

    /**
    * Sets the model ID.
    */
    public void setModelId(int modelId)
    {
        _modelId = modelId;
    }

    /**
    * Returns the model ID.
    */
    public int getModelId()
    {
        return _modelId;
    }

    /**
    * Sets the chain ID.
    */
    public void setChainId(String chainId)
    {
        _chainId = chainId;
    }

    /**
    * Returns the chain ID.
    */
    public String getChainId()
    {
        return _chainId;
    }

    /**
    * Sets the residue ID.
    */
    public void setResidueId(int residueId)
    {
        _residueId = residueId;
    }

    /**
    * Returns the residue ID.
    */
    public int getResidueId()
    {
        return _residueId;
    }

    /**
    * Sets the residue name.
    */
    public void setResidueName(String residueName)
    {
        _residueName = residueName;
    }

    /**
    * Returns the residue name.
    */
    public String getResidueName()
    {
        return _residueName;
    }

    /**
    * Sets the atom name.
    */
    public void setAtomName(String atomName)
    {
        _atomName = atomName;
    }

    /**
    * Returns the atom name.
    */
    public String getAtomName()
    {
        return _atomName;
    }

    /**
    * Sets the alternate location indicator.
    */
    public void setAltLoc(String altLoc)
    {
        _altLoc = altLoc;
    }

    /**
    * Returns the alternate location indicator.
    */
    public String getAltLoc()
    {
        return _altLoc;
    }

    /**
    * Constructor.
    */
    public PrimitiveExpr()
    {
        _fileId      = NULL;
        _modelId     = NULL;
        _chainId     = null;
        _residueId   = NULL;
        _residueName = null;
        _atomName    = null;
        _altLoc      = null;
    }

    /**
    * Tests whether the specified atom matches the expression.
    */
    public boolean match(AtomPath atomPath)
    {
        // file ID
        if(_fileId != NULL){
            if(atomPath.getMoleculeFile().getFileId() != _fileId) return false;
        }

        // model ID
        if(_modelId != NULL){
            if(atomPath.getMolecule().getModelId() != _modelId) return false;
        }

        // chain ID
        if(_chainId != null){
            if(!compareChain(_chainId, atomPath.getChain())) return false;
        }

        // residue ID
        if(_residueId != NULL){
            if(atomPath.getResidue().getId() != _residueId) return false;
        }

        // residue name
        if(_residueName != null){
            if(!compareIgnoreCase(_residueName, atomPath.getResidue().getName())) return false;
        }

        // atom name
        if(_atomName != null){
            if(!compareIgnoreCase(_atomName, atomPath.getAtom().getName())) return false;
        }

        // altLoc
        if(_altLoc != null){
            if(!compareIgnoreCase(_altLoc, atomPath.getAtom().getAltLoc())) return false;
        }

        return true;
    }

    static boolean compareChain(String expr, Chain chain)
    {
        if (expr.startsWith("a_")) {
            return compare(expr.substring(2), chain.getId());
        } else if (expr.startsWith("l_")) {
            if (chain.getLabelId().length() == 0) {
                return false;
            } else {
                return compare(expr.substring(2), chain.getLabelId());
            }
        } else {
            return compareIgnoreCase(expr, chain.getId());
        }
    }

    private static boolean compareIgnoreCase(String expr, String curr)
    {
        if(curr == null) return false;

        String str1 = expr.trim().toUpperCase();
        String str2 = curr.trim().toUpperCase();
        return compare(str1, str2);
    }

    private static boolean compare(String str1, String str2)
    {
        if(str2 == null) return false;

        int len1 = str1.length();
        int len2 = str2.length();

        if(len2 > len1){
            return false;
        }
        else if(len2 < len1){
            for(int i=len2; i<len1; i++){
                if(str1.charAt(i) != '?') return false;
            }
            str1 = str1.substring(0, len2);
        }

        for(int i=0; i<len2; i++){
            if(str1.charAt(i) != '?' && str1.charAt(i) != str2.charAt(i)){
                return false;
            }
        }

        return true;
    }
}

