/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.moleculebuilders;

import java.util.*;
import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.manipulators.*;
import org.pdbj.viewer.molecules.*;

/**
 * Creates MoleculeFile object.
 */
public class MoleculeFileBuilder
{
    private MoleculeFile _moleculeFile;

    private ArrayList _structureInfos;

    private Molecule _currMolecule;
    private Chain    _currChain;
    private Residue  _currResidue;

    private Map _modresMap;
    private Set _chemCompAminoSet;
    private Set _chemCompDnaSet;
    private Set _chemCompRnaSet;

    private AtomPath _atomPath;
    private AtomFilter _filter;

    /**
     * Constructs an instance.
     */
    public MoleculeFileBuilder()
    {
        _structureInfos = new ArrayList();
        _modresMap = new HashMap();
        _chemCompAminoSet = new HashSet();
        _chemCompDnaSet = new HashSet();
        _chemCompRnaSet = new HashSet();
        this.initialize();
    }

    /**
     * Initialize.
     */
    public void initialize()
    {
        _currMolecule = null;
        _currChain = null;
        _currResidue = null;
        _moleculeFile = new MoleculeFileImpl();
        _moleculeFile.setPDBml(false);
        _moleculeFile.setFilePath("");
        _moleculeFile.setMoleculeId("");
        _moleculeFile.setMoleculeName("");
        _moleculeFile.setClassification("");
        _moleculeFile.setDate("");
        _moleculeFile.setExperiment("");
        _structureInfos.clear();
        _modresMap.clear();
        _chemCompAminoSet.clear();
        _chemCompDnaSet.clear();
        _chemCompRnaSet.clear();
        _atomPath = new AtomPath();
        _atomPath.setMoleculeFile(_moleculeFile);
    }

    /**
     * Sets whether the source file is PDBj-ML.
     */
    public void setPDBml(boolean pdbml)
    {
        _moleculeFile.setPDBml(pdbml);
    }

    /**
     * Sets the absolute path of the source file.
     */
    public void setFilePath(String filePath)
    {
        _moleculeFile.setFilePath(filePath);
    }

    /**
     * Sets the molecule ID (PDB code).
     */
    public void setMoleculeId(String moleculeId)
    {
        _moleculeFile.setMoleculeId(moleculeId);
    }

    /**
     * Sets the molecule name.
     */
    public void setMoleculeName(String moleculeName)
    {
        _moleculeFile.setMoleculeName(moleculeName);
    }

    /**
     * Sets the classification.
     */
    public void setClassification(String classification)
    {
        _moleculeFile.setClassification(classification);
    }

    /**
     * Sets the date.
     */
    public void setDate(String date)
    {
        _moleculeFile.setDate(date);
    }

    /**
     * Sets the experiment.
     */
    public void setExperiment(String experiment)
    {
        _moleculeFile.setExperiment(experiment);
    }

    /**
     * Appends secondary structure information.
     */
    public void addStructureInfo(StructureInfo structureInfo)
    {
        _structureInfos.add(structureInfo);
    }

    /**
     * Adds MODRES information.
     */
    public void addModres(String name, String standardName)
    {
        _modresMap.put(name, standardName);
    }

    /**
     * Adds chem_comp type information.
     */
    public void addChemCompType(String name, String type)
    {
        if (type.equalsIgnoreCase("D-PEPTIDE LINKING") ||
                type.equalsIgnoreCase("D-PEPTIDE NH3 AMINO TERMINUS") ||
                type.equalsIgnoreCase("L-PEPTIDE COOH CARBOXY TERMINUS") ||
                type.equalsIgnoreCase("L-PEPTIDE LINKING") ||
                type.equalsIgnoreCase("L-PEPTIDE NH3 AMINO TERMINUS") ||
                type.equalsIgnoreCase("PEPTIDE LINKING") ||
                type.equalsIgnoreCase("PEPTIDE-LIKE")) {
            _chemCompAminoSet.add(name);
        } else if (type.equalsIgnoreCase("DNA LINKING")) {
            _chemCompDnaSet.add(name);
        } else if (type.equalsIgnoreCase("RNA LINKING")) {
            _chemCompRnaSet.add(name);
        }
    }

    public void setFilter(AtomFilter filter) {
        _filter = filter;
    }

    /**
     * Appends atom records.
     */
    public void addAtomRecord(AtomRecord atomRecord, boolean hetero)
    {
        // get fields
        String authChainId = atomRecord.getAuthAsymId();
        String labelChainId = atomRecord.getLabelAsymId();
        String resName = atomRecord.getCompId();
        String name    = atomRecord.getAtomId();
        String altLoc  = atomRecord.getAltId();
        char   iCode   = atomRecord.getInsCode();

        int serialNo, modelId, resSeq;
        float x, y, z, tempFactor;
        try{
            if( atomRecord.getMdlId().equals(AtomRecord.NO_MODEL_ID) ){
                modelId = Molecule.NO_MODEL_ID;
            }
            else{
                modelId = Integer.parseInt( atomRecord.getMdlId() );
            }
            serialNo   = Integer.parseInt( atomRecord.getId() );
            resSeq     = Integer.parseInt( atomRecord.getSeqId() );
            x          = Float.parseFloat( atomRecord.getCartnX() );
            y          = Float.parseFloat( atomRecord.getCartnY() );
            z          = Float.parseFloat( atomRecord.getCartnZ() );
            tempFactor = ( atomRecord.getBFactor().equals("") ) ? 
                            0 : Float.parseFloat( atomRecord.getBFactor() );
        }
        catch(NumberFormatException e){
            return;
        }

        // if model, chain, or residue has changed
        if(_currMolecule != null && modelId != _currMolecule.getModelId()){
            this.addMolecule();
        }
        else if (_currChain != null
                && (!authChainId.equals(_currChain.getId()) 
                        || !labelChainId.equals(_currChain.getLabelId()))) {
            this.addChain();
        }
        else if(_currResidue != null && ( resSeq != _currResidue.getId() 
                 || iCode != _currResidue.getInsCode() ) )
        {
            this.addResidue();
        }

        // create Residue
        if(_currResidue == null){
            _currResidue = new ResidueImpl();
            _currResidue.setId(resSeq);
            _currResidue.setInsCode(iCode);
            _currResidue.setName(resName);
            _atomPath.setResidue(_currResidue);
            this.determineResidueType(_currResidue);

            // create Chain
            if(_currChain == null){
                _currChain = new ChainImpl();
                _currChain.setId(authChainId);
                _currChain.setLabelId(labelChainId);
                _atomPath.setChain(_currChain);

                // create Molecule
                if(_currMolecule == null){
                    _currMolecule = new MoleculeImpl();
                    _currMolecule.setModelId(modelId);
                    _atomPath.setMolecule(_currMolecule);
                }
            }
        }

        // refers to tables
        int atomicNo = 0;
        String symbol = atomRecord.getTypeSymbol();
        if( symbol != null && !symbol.equals("") )
        {
            if( symbol.toUpperCase().equals("D") ){
                atomicNo = 1;
            }
            else{
                atomicNo = ElementTable.getAtomicNoBySymbol(symbol);
            }
        }

        // create Atom object
        Atom atom = new AtomImpl();
        atom.setId(serialNo);
        atom.setX(x);
        atom.setY(y);
        atom.setZ(z);
        atom.setTempFactor(tempFactor);
        atom.setAltLoc(altLoc);
        atom.setHetero(hetero);
        atom.setAtomicNo(atomicNo);
        atom.setSelected(true);
        atom.setDisplayed(true);
        atom.setName(name);
        _atomPath.setAtom(atom);
        this.determineAtomType(_currResidue, atom);

        if (_filter != null) {
            if (_filter.getExpr() == null
                    || _filter.getExpr().match(_atomPath)) {
                _currResidue.addAtom(atom);
                _filter.addAtomFlag(Boolean.TRUE);
            } else {
                _filter.addAtomFlag(Boolean.FALSE);
            }
        }
    }

    /**
     * Breaks the current chain.
     */
    public void breakChain() {
        if (_currChain != null) {
            this.addChain();
        }
    }

    /**
     * Returns MoleculeFile object. Subsequently, Initialize method is
     * automatically called.
     */
    public MoleculeFile getResult(boolean ver3)
    {
        if(_currMolecule != null){
            this.addMolecule();
        }

        MoleculeFile result = null;

        if(_moleculeFile.numMolecules() != 0){
            // assign structure
            for(int i=0; i<_structureInfos.size(); i++){
                StructureInfo info = (StructureInfo)_structureInfos.get(i);
                this.assignStructure(info);
            }
    
            // initialize molecule
            for(int m=0; m<_moleculeFile.numMolecules(); m++){
                Molecule molecule = _moleculeFile.getMolecule(m);
                initializeMolecule(molecule);
            }

            // center & radius
            SphereUtil util = new SphereUtil();
            for(int m=0; m<_moleculeFile.numMolecules(); m++){
                Molecule molecule = _moleculeFile.getMolecule(m);
                Vector3f center = new Vector3f(
                    molecule.getCenterX(), molecule.getCenterY(), molecule.getCenterZ()
                );
                util.addSphere(center, molecule.getRadius());
            }
            Sphere sphere = util.getSphere();
            _moleculeFile.center().set( sphere.center() );
            _moleculeFile.setRadius( sphere.getRadius() );

            // set ver3
            _moleculeFile.setVer3(ver3);

            result = _moleculeFile;
        }

        this.initialize();
        return result;
    }

    //--------------------------------------------------------------------
    // Constructs molecule
    //--------------------------------------------------------------------
    private void addResidue()
    {
        if (_currResidue.numAtoms() > 0) {
            _currChain.addResidue(_currResidue);
        }
        _currResidue = null;
    }

    private void addChain()
    {
        if(_currResidue != null) this.addResidue();

        if (_currChain.numResidues() > 0) {
            _currMolecule.addChain(_currChain);
        }
        _currChain = null;
    }

    private void addMolecule()
    {
        if(_currChain != null) this.addChain();

        _currMolecule.setFile(_moleculeFile);

        if (_currMolecule.numChains() > 0) {
            _moleculeFile.addMolecule(_currMolecule);
        }
        _currMolecule = null;
    }

    //--------------------------------------------------------------------
    // Determines type
    //--------------------------------------------------------------------
    private void determineResidueType(Residue residue) {
        boolean ver3 = _moleculeFile.isVer3();
        String rname = residue.getName();
        int rindex = 0;
        if (ver3) {
            rindex = ResidueTable.getIndexVer3(rname);
            if (rindex == ResidueTable.INDEX_OTHER) {
                String sname = (String) _modresMap.get(rname);
                if (sname != null) {
                    residue.setParentIndex(ResidueTable.getIndexVer3(sname));
                }
                if (residue.getParentIndex() == ResidueTable.INDEX_OTHER) {
                    if (_chemCompAminoSet.contains(rname)) {
                        residue.setChemCompType(Residue.CHEM_COMP_TYPE_AMINO);
                    } else if (_chemCompDnaSet.contains(rname)) {
                        residue.setChemCompType(Residue.CHEM_COMP_TYPE_DNA);
                    } else if (_chemCompRnaSet.contains(rname)) {
                        residue.setChemCompType(Residue.CHEM_COMP_TYPE_RNA);
                    }
                }
            }
        } else {
            rindex = ResidueTable.getIndex(rname);
        }
        residue.setIndex(rindex);
        if (rname.equals(ResidueTable.getName(rindex))) {
            residue.setName(null);
        }
    }

    private void determineAtomType(Residue residue, Atom atom) {
        boolean pdbml = _moleculeFile.isPDBml();
        boolean ver3 = _moleculeFile.isVer3();
        boolean solvent = residue.isInGroup(ResidueTable.GROUP_SOLVENT);
        int rindex = residue.getIndex();
        String aname = atom.getName();
        int aindex = 0;
        if (ver3) {
            aindex = AtomTable.getIndexVer3(aname);
        } else {
            aindex = AtomTable.getIndex(aname, rindex);                        
        }
        atom.setIndex(aindex);
        if (aname.equals(AtomTable.getName(aindex))) {
            atom.setName(null);
        }
        if (solvent) {
            atom.setHetero(true);
        }
        if (!pdbml && !ver3) {
            atom.setAtomicNo(AtomTable.getAtomicNo(aname, rindex));
        }
    }

    //--------------------------------------------------------------------
    // Assigns secondary structure
    //--------------------------------------------------------------------
    private void assignStructure(StructureInfo info)
    {
        int init, term;
        try{
            init = Integer.parseInt(info.getInitSeqId());
            term = Integer.parseInt(info.getTermSeqId());
        }
        catch(NumberFormatException e){
            return;
        }

        // termChainId is not used
        for(int m=0; m<_moleculeFile.numMolecules(); m++)
        {
            Molecule molecule = _moleculeFile.getMolecule(m);
            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                if( !chain.getId().equals(info.getInitChainId()) ) continue;
                for(int re=0; re<chain.numResidues(); re++)
                {
                    Residue residue = chain.getResidue(re);
                    if(residue.getId() >= init && residue.getId() <= term){
                        residue.setStructure( info.getStructure() );
                    }
                }
            }

            switch( info.getStructure() )
            {
                case Residue.STRUCT_HELIX :
                    molecule.setHelixCount( molecule.getHelixCount()+1 );
                    break;
                case Residue.STRUCT_SHEET :
                    molecule.setStrandCount( molecule.getStrandCount()+1 );
                    break;
                case Residue.STRUCT_TURN :
                    molecule.setTurnCount( molecule.getTurnCount()+1 );
                    break;
            }

            molecule.setStructure(Molecule.STRUCT_FILE);
        }
    }

    //--------------------------------------------------------------------
    // Initialize molecule
    //--------------------------------------------------------------------
    public static void initializeMolecule(Molecule molecule)
    {
        // select count, hydrogens and hetero atoms
        int atomCnt = 0;
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
                Residue residue = chain.getResidue(re);
                for(int at=0; at<residue.numAtoms(); at++)
                {
                    Atom atom = residue.getAtom(at);

                    if( atom.isSelected() ) atomCnt++;
                    if( atom.getAtomicNo() == 1) molecule.setHydrogens(true);
                    if( atom.isHetero() ) molecule.setHeteroAtoms(true);
                }
            }
        }
        molecule.setSelectAtomCount(atomCnt);

        // min & max values
        float minX, minY, minZ, minTemp, minHetTemp;
        float maxX, maxY, maxZ, maxTemp, maxHetTemp;
        minX = minY = minZ = minTemp = minHetTemp = Float.POSITIVE_INFINITY;
        maxX = maxY = maxZ = maxTemp = maxHetTemp = Float.NEGATIVE_INFINITY;

        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
                Residue residue = chain.getResidue(re);
                for(int at=0; at<residue.numAtoms(); at++)
                {
                    Atom atom = residue.getAtom(at);

                    // find min & max coordinates
                    float vdw = getVdWRadius(molecule, atom);
                    float x1 = atom.getX() - vdw;
                    float y1 = atom.getY() - vdw;
                    float z1 = atom.getZ() - vdw;
                    float x2 = atom.getX() + vdw;
                    float y2 = atom.getY() + vdw;
                    float z2 = atom.getZ() + vdw;
                    if(x1 < minX) minX = x1;
                    if(y1 < minY) minY = y1;
                    if(z1 < minZ) minZ = z1;
                    if(x2 > maxX) maxX = x2;
                    if(y2 > maxY) maxY = y2;
                    if(z2 > maxZ) maxZ = z2;

                    // find min & max temperature factor
                    float temp = atom.getTempFactor();
                    if( atom.isHetero() ){
                        if(temp < minHetTemp) minHetTemp = temp;
                        if(temp > maxHetTemp) maxHetTemp = temp;
                    }
                    else{
                        if(temp < minTemp) minTemp = temp;
                        if(temp > maxTemp) maxTemp = temp;
                    }
                }
            }
        }

        molecule.setMinX( minX );
        molecule.setMinY( minY );
        molecule.setMinZ( minZ );
        molecule.setMaxX( maxX );
        molecule.setMaxY( maxY );
        molecule.setMaxZ( maxZ );
        molecule.setCenterX( (maxX + minX)/2 );
        molecule.setCenterY( (maxY + minY)/2 );
        molecule.setCenterZ( (maxZ + minZ)/2 );

        molecule.setMinTempFactor( minTemp );
        molecule.setMaxTempFactor( maxTemp );
        molecule.setMinHeteroTempFactor( minHetTemp );
        molecule.setMaxHeteroTempFactor( maxHetTemp );

        // determine radius
        float radius = 0;
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
                Residue residue = chain.getResidue(re);
                for(int at=0; at<residue.numAtoms(); at++)
                {
                    Atom atom = residue.getAtom(at);

                    float x = atom.getX() - molecule.getCenterX();
                    float y = atom.getY() - molecule.getCenterY();
                    float z = atom.getZ() - molecule.getCenterZ();

                    float currRadius
                        = (float)Math.sqrt(x*x + y*y + z*z) 
                          + getVdWRadius(molecule, atom);
                    if(currRadius > radius) radius = currRadius;
                }
            }
        }
        molecule.setRadius(radius);
    }

    public static float getVdWRadius(Molecule molecule, Atom atom)
    {
        if( atom.getAtomicNo() == 6 && !molecule.hasHydrogens() ){
            return ElementTable.VDW_CARBON;
        }
        else if( atom.getAtomicNo() == 7 && !molecule.hasHydrogens() ){
            return ElementTable.VDW_NITROGEN;
        }
        else if( atom.getAtomicNo() == 8 && !molecule.hasHydrogens() ){
            return ElementTable.VDW_OXYGEN;
        }
        else if( atom.getAtomicNo() == 10 && !molecule.hasHydrogens() ){
            return ElementTable.VDW_SULFUR;
        }
        else{
            return ElementTable.getVdWRadius( atom.getAtomicNo() );
        }
    }

    public static class AtomFilter {

        private SelectExpr _expr;
        private List _atomFlags;

        public AtomFilter() {
            _atomFlags = new ArrayList();
        }

        public SelectExpr getExpr() {
            return _expr;
        }

        public void setExpr(SelectExpr expr) {
            _expr = expr;
        }

        public void addAtomFlag(Boolean flag) {
            _atomFlags.add(flag);
        }

        public List getAtomFlags() {
            return _atomFlags;
        }        
    }
}
