/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.moleculebuilders;

import java.io.*;
import java.net.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.molecules.*;

/**
 * Class that reads PDB files.
 */
public class PDBReader
{
    private MoleculeFileBuilder _builder;

    private String     _currentModel;
    private boolean    _moleculeNameDone;
    private AtomRecord _atomRecord;

    private MessageListener _listener;
    private boolean _anim;
    private int _frame;

    private boolean _ver3;

    /**
     * Constructor.
     */
    public PDBReader(MessageListener listener, boolean anim)
    {
        _listener = listener;
        _anim = anim;
    }

    /**
     * Reads the specified molecule file and creates the MoleculeFile object.
     */
    public MoleculeFile read(File file, MoleculeFileBuilder.AtomFilter filter)
            throws RecoverableException
    {
        _builder = new MoleculeFileBuilder();
        _builder.setPDBml(false);
        _builder.setFilePath( file.getAbsolutePath() );
        _builder.setFilter(filter);

        if(_anim && _listener != null){
            _listener.begin(1);
            _listener.appendMessage("");
        }

        try{
            this.readFile( new ResourceRequest(file) );
        }
        finally{
            if(_anim && _listener != null) _listener.end();
        }

        return _builder.getResult(_ver3);
    }

    /**
     * Reads the specified molecule file and creates the MoleculeFile object.
     */
    public MoleculeFile read(URL url) throws RecoverableException
    {
        _builder = new MoleculeFileBuilder();
        _builder.setPDBml(false);
        _builder.setFilePath( url.toString() );

        // message
        if(_listener != null){
            _listener.begin(1);
            if(_anim) _listener.appendMessage("Retrieving an animation file");
            else      _listener.appendMessage("Retrieving a PDB file");
        }

        try{
            this.readFile( new ResourceRequest(url) );
        }
        finally{
            if(_listener != null) _listener.end();
        }

        return _builder.getResult(_ver3);
    }

    //--------------------------------------------------------------------
    // Parse
    //--------------------------------------------------------------------
    private void readFile(ResourceRequest request) throws RecoverableException
    {
        try{
            this.parse(request);
        }
        catch(FileNotFoundException e){
            throw new RecoverableException(ErrorMessages.CANT_FIND_PDB_FILE);
        }
        catch(UnknownHostException e){
            throw ExceptionUtil.handleUnknownHostException(e);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
        catch(java.security.AccessControlException e)
        {
            throw new RecoverableException(ErrorMessages.CANT_FIND_PDB_FILE);
        }
    }

    private void parse(ResourceRequest request) throws IOException
    {
        BufferedReader reader = null;
        try{
            reader = new BufferedReader( new InputStreamReader( request.openStream() ) );
        
            _atomRecord = new AtomRecord();
            _currentModel = AtomRecord.NO_MODEL_ID;
            _moleculeNameDone = false;
            _frame = 0;

            _ver3 = false;

            String s = null;
            while( (s = reader.readLine()) != null )
            {
                // handle short line
                if(s.length() < 80) s = addSpace(s, 80);

                if(s.substring(0,6).equals("HEADER")){
                    processHEADER(s);
                }
                else if(s.substring(0,6).equals("COMPND")){
                    processCOMPND(s);
                }
                else if(s.substring(0,6).equals("EXPDTA")){
                    processEXPDTA(s);
                }
                else if(s.substring(0,6).equals("REMARK")){
                    processREMARK(s);
                }
                else if(s.substring(0,6).equals("MODEL ")){
                    processMODEL(s);
                }
                else if(s.substring(0,6).equals("ATOM  ")){
                    processATOM(s, false);
                }
                else if(s.substring(0,6).equals("HETATM")){
                    processATOM(s, true);
                }
                else if(s.substring(0,6).equals("HELIX ")){
                    processHELIX(s);
                }
                else if(s.substring(0,6).equals("SHEET ")){
                    processSHEET(s);
                }
                else if(s.substring(0,6).equals("TURN  ")){
                    processTURN(s);
                }
                else if(s.substring(0,6).equals("TER   ")){
                    processTER(s);
                }
                else if(s.substring(0,6).equals("MODRES")){
                    processMODRES(s);
                }
            }
        }
        finally{
            if(reader != null) reader.close();
            request.dispose();
        }
    }

    private void processHEADER(String s)
    {
        _builder.setClassification( s.substring(10,50).trim() );
        _builder.setDate( s.substring(50,59).trim() );
        _builder.setMoleculeId( s.substring(62,66).toLowerCase() );
    }

    private void processCOMPND(String s)
    {
        if( s.substring(9,10).equals(" ")
            && !s.substring(10,17).equals("MOL_ID:") )
        {
            _builder.setMoleculeName( s.substring(10,70).trim() );
            _moleculeNameDone = true;
        }
        else if( !_moleculeNameDone && s.substring(11,21).equals("MOLECULE: ") ){
            _builder.setMoleculeName( s.substring(21,70).trim() );
            _moleculeNameDone = true;
        }
    }

    private void processEXPDTA(String s)
    {
        _builder.setExperiment(s.substring(10,70).trim());
    }

    private void processREMARK(String s) {
        if (s.charAt(9) == '4'
                && s.substring(16, 39).equals("COMPLIES WITH FORMAT V.")) {
            String ver = null;
            int index = s.indexOf(".", 40);
            if (index != -1) {
                ver = s.substring(40, index);
            }
            if (ver != null) {
                try {
                    int majorVer = Integer.parseInt(ver);
                    if (majorVer >= 3) {
                        _ver3 = true;
                    }
                } catch (NumberFormatException e){
                }
            }
        }
    }

    private void processMODEL(String s)
    {
        _currentModel = s.substring(10,14).trim();        
        _frame++;
        if(_anim && _listener != null){
            _listener.setMessage("Reading frame " + _frame, 0);            
        }
    }

    private void processATOM(String s, boolean hetero)
    {
        _atomRecord.setId( s.substring(6,11).trim() );
        _atomRecord.setMdlId( _currentModel );
        _atomRecord.setAuthAsymId( s.substring(21,22) );
        _atomRecord.setSeqId( s.substring(22,26).trim() );
        _atomRecord.setAltId( s.substring(16,17).trim() );
        _atomRecord.setCompId( s.substring(17,20).toUpperCase() );
        _atomRecord.setAtomId( s.substring(12,16).toUpperCase() );
        _atomRecord.setCartnX( s.substring(30,38).trim() );
        _atomRecord.setCartnY( s.substring(38,46).trim() );
        _atomRecord.setCartnZ( s.substring(46,54).trim() );
        _atomRecord.setBFactor( s.substring(60,66).trim() );
        _atomRecord.setInsCode( s.charAt(26) );
        if (_ver3) {
            _atomRecord.setTypeSymbol(s.substring(76, 78).trim());
        }

        _builder.addAtomRecord(_atomRecord, hetero);
    }

    private void processHELIX(String s)
    {
        StructureInfo info = new StructureInfo();
        info.setInitChainId( s.substring(19,20) );
        info.setTermChainId( s.substring(31,32) );
        info.setInitSeqId( s.substring(21,25).trim() );
        info.setTermSeqId( s.substring(33,37).trim() );
        info.setStructure(Residue.STRUCT_HELIX);
        _builder.addStructureInfo(info);
    }

    private void processSHEET(String s)
    {
        StructureInfo info = new StructureInfo();
        info.setInitChainId( s.substring(21,22) );
        info.setTermChainId( s.substring(32,33) );
        info.setInitSeqId ( s.substring(22,26).trim() );
        info.setTermSeqId( s.substring(33,37).trim() );
        info.setStructure(Residue.STRUCT_SHEET);
        _builder.addStructureInfo(info);
    }

    private void processTURN(String s)
    {
        StructureInfo info = new StructureInfo();
        info.setInitChainId( s.substring(19,20) );
        info.setTermChainId( s.substring(30,31) );
        info.setInitSeqId( s.substring(20,24).trim() );
        info.setTermSeqId( s.substring(31,35).trim() );
        info.setStructure(Residue.STRUCT_TURN);
        _builder.addStructureInfo(info);
    }

    private void processTER(String s) {
        _builder.breakChain();
    }

    private void processMODRES(String s) {
        String name = s.substring(12,15);
        String sname = s.substring(24,27);
        _builder.addModres(name, sname);
    }

    private static String addSpace(String s, int column)
    {
        int len = s.length();
        if(len < column){
            StringBuffer sb = new StringBuffer();
            for(int i=0; i<column-len; i++){
                sb.append(' ');
            }
            s += sb.toString();
        }
        return s;
    }
}
