/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.moleculebuilders;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.molecules.*;

/**
 * Class that reads PDBML files.
 */
public class PDBmlReader extends DefaultHandler
{
    private static boolean AUTH = true;

    private PDBmlUrl _pdbmlUrl;

    private MoleculeFileBuilder _builder;
    private MessageListener _listener;

    private boolean _mainDone;
    private boolean _coordsDone;
    private boolean _readExt;

    private String _pdbCode;
    private String _compound;
    private String _classification;
    private String _experiment;
    private String _date;

    private boolean _ver3;

    private StructureInfo _currStructInfo;
    private AtomRecord    _atomRecord;
    private boolean       _struct;
    private boolean       _hetero;
    private String        _currId;

    private String _modresCompId;
    private String _modresParentCompId;
    private String _chemCompId;

    private boolean      _pickupText;
    private StringBuffer _buffer;

    private SAXNodeHandler _handler;

    private SAXNode _datablock;
    private SAXNode _entry;
    private SAXNode _pdbx_version;
    private SAXNode _pdbx_description;
    private SAXNode _pdbx_keywords;
    private SAXNode _date_original;
    private SAXNode _exptl_method;
    private SAXNode _struct_conf;
    private SAXNode _conf_beg_asym_id;
    private SAXNode _conf_end_asym_id;
    private SAXNode _conf_beg_seq_id;
    private SAXNode _conf_end_seq_id;
    private SAXNode _struct_sheet_range;
    private SAXNode _sheet_beg_asym_id;
    private SAXNode _sheet_end_asym_id;
    private SAXNode _sheet_beg_seq_id;
    private SAXNode _sheet_end_seq_id;
    private SAXNode _struct_mod_residue;
    private SAXNode _modres_comp_id;
    private SAXNode _modres_parent_comp_id;
    private SAXNode _chem_comp;
    private SAXNode _chem_comp_type;

    private SAXNode _atom_site;
    private SAXNode _group_PDB;
    private SAXNode _pdbx_PDB_model_num;
    private SAXNode _label_asym_id;
    private SAXNode _auth_asym_id;
    private SAXNode _seq_id;
    private SAXNode _auth_seq_id;
    private SAXNode _label_alt_id;
    private SAXNode _comp_id;
    private SAXNode _type_symbol;
    private SAXNode _atom_id;
    private SAXNode _Cartn_x;
    private SAXNode _Cartn_y;
    private SAXNode _Cartn_z;
    private SAXNode _occupancy;
    private SAXNode _B_iso_or_equiv;

    private SAXNode _atom_record;

    /**
     * Constructor.
     */
    public PDBmlReader(MessageListener listener, PDBmlUrl pdbmlUrl)
    {
        _listener = listener;
        _pdbmlUrl = pdbmlUrl;

        _buffer = new StringBuffer();
        _atomRecord = new AtomRecord();

        this.setupHandler();
    }

    /**
     * Reads the specified molecule file and creates the MoleculeFile object.
     */
    public MoleculeFile read(File file) throws RecoverableException
    {
        _builder = new MoleculeFileBuilder();
        _builder.setFilePath( file.getAbsolutePath() );

        try{
            URL url = file.toURI().toURL();
            this.readAll(url);
        }
        catch(MalformedURLException e){
            throw new RecoverableException(ErrorMessages.CANT_FIND_PDBML_FILE);
        }
        finally{
            if(_listener != null && _listener.isActive()) _listener.end();
        }

        return _builder.getResult(_ver3);
    }
    
    /**
     * Reads the specified molecule file and creates the MoleculeFile object.
     */
    public MoleculeFile read(URL url) throws RecoverableException
    {
        _builder = new MoleculeFileBuilder();
        _builder.setFilePath( url.toString() );

        // message
        if(_listener != null){
            _listener.begin(1);
            _listener.appendMessage("Retrieving a PDBML file");
        }

        try{
            this.readAll(url);
        }
        finally{
            if(_listener != null) _listener.end();
        }

        return _builder.getResult(_ver3);
    }

    /**
     * Reads the specified PDBML file by FTP and creates the MoleculeFile object.
     */
    public MoleculeFile readFTP(String pdbCode) throws RecoverableException
    {
        if(_pdbmlUrl == null) return null;
        
        _builder = new MoleculeFileBuilder();

        String name = pdbCode.toLowerCase() + "-noatom.xml";
        String path = _pdbmlUrl.getUrlNoatom() + name;

        this.initialize();

        // show message
        if(_listener != null){
            _listener.begin(2);
            _listener.appendMessage("Retrieving " + name + " from " + _pdbmlUrl.getHostNoatom());
        }

        try{
            _readExt = false;

            this.readFile( new URL(path + ".gz") );

            if(!_mainDone){
                this.readFile( new URL(path) );
            }

            if(!_mainDone){
                throw new RecoverableException(ErrorMessages.CANT_FIND_PDBML_FILE);
            }

            if( !_coordsDone && !_pdbCode.equals("") )
            {
                _readExt = true;

                this.readExtAtomFtp(pdbCode);

                if(!_coordsDone){
                    throw new RecoverableException(ErrorMessages.CANT_FIND_PDBML_EXT_FILE);
                }
            }
        }
        catch(MalformedURLException e){
            throw new RecoverableException(ErrorMessages.CANT_FIND_PDBML_FILE);
        }
        catch(UnknownHostException e){
            throw ExceptionUtil.handleUnknownHostException(e);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
        finally{
            if(_listener != null) _listener.end();
        }

        _builder.setPDBml(true);
        _builder.setFilePath(pdbCode);
        _builder.setMoleculeId(_pdbCode);
        _builder.setMoleculeName(_compound);
        _builder.setClassification(_classification);
        _builder.setDate(_date);
        _builder.setExperiment(_experiment);

        MoleculeFile moleculeFile = _builder.getResult(_ver3);
        moleculeFile.setFTP(true);

        return moleculeFile;
    }

    //--------------------------------------------------------------------
    // Read
    //--------------------------------------------------------------------
    private void initialize()
    {
        _mainDone = false;
        _coordsDone = false;

        _pdbCode        = "";
        _compound       = "";
        _classification = "";
        _experiment     = "";
        _date           = "";

        _ver3 = false;

        _modresCompId = null;
        _modresParentCompId = null;
        _chemCompId = null;
    }

    private void readAll(URL url) throws RecoverableException
    {
        this.initialize();

        try{
            _readExt = false;

            this.readFile(url);
            
            if(!_mainDone){
                throw new RecoverableException(ErrorMessages.CANT_FIND_PDBML_FILE);
            }

            if( !_coordsDone && !_pdbCode.equals("") )
            {
                _readExt = true;

                this.readExtAtom(url);

                if(!_coordsDone){
                    this.readExtAtomFtp(_pdbCode);
                }

                if(!_coordsDone){
                    throw new RecoverableException(ErrorMessages.CANT_FIND_PDBML_EXT_FILE);
                }
            }
        }
        catch(UnknownHostException e){
            throw ExceptionUtil.handleUnknownHostException(e);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
        catch(java.security.AccessControlException e)
        {
            if( ExceptionUtil.isPropertyPermission(e) ){
                throw ExceptionUtil.handlePropertyPermission();
            }
            else{
                throw new RecoverableException(ErrorMessages.CANT_FIND_PDBML_FILE);
            }
        }

        _builder.setPDBml(true);
        _builder.setMoleculeId(_pdbCode);
        _builder.setMoleculeName(_compound);
        _builder.setClassification(_classification);
        _builder.setDate(_date);
        _builder.setExperiment(_experiment);
    }

    private void readExtAtom(URL url) throws IOException, RecoverableException
    {
        String s = url.toString();

        String name = _pdbCode.toLowerCase() + "-extatom.xml";
        int idx1 = s.lastIndexOf("/");
        int idx2 = s.substring(0, idx1).lastIndexOf("/");
        String parent = s.substring(idx2 + 1, idx1);

        String path = null;
        if (parent.equals("XML-noatom")) {
            path = s.substring(0, idx2 + 1) + "XML-extatom/" + name;
        } else if (parent.equals("all-noatom")) {
            path = s.substring(0, idx2 + 1) + "all-extatom/" + name;
        } else {
            path = s.substring(0, idx1 + 1) + name;
        }

        try{
            this.readFile( new URL(path + ".gz") );
            if(!_coordsDone){
                this.readFile( new URL(path) );
            }
        }
        catch(MalformedURLException e){
        }
    }

    private void readExtAtomFtp(String pdbCode) throws IOException, RecoverableException
    {
        if(_pdbmlUrl == null) return;

        String name = pdbCode.toLowerCase() + "-extatom.xml";
        String path = _pdbmlUrl.getUrlExtatom() + name;

        // show message
        if(_listener != null){
            if( !_listener.isActive() ) _listener.begin(1);
            _listener.appendMessage("Retrieving "+ name + " from " + _pdbmlUrl.getHostExtatom());
        }

        try{
            this.readFile( new URL(path + ".gz") );
            if(!_coordsDone){
                this.readFile( new URL(path) );
            }
        }
        catch(MalformedURLException e){
        }
    }

    private void readFile(URL url) throws IOException, RecoverableException
    {
        ResourceRequest request = new ResourceRequest(url);

        InputStream stream = null;
        try{
            stream = request.openStream();
            this.parse(stream);            
        }
        catch(FileNotFoundException e){
        }
        finally{
            if(stream != null) stream.close();
            request.dispose();
        }
    }

    //--------------------------------------------------------------------
    // Parse
    //--------------------------------------------------------------------
    private void parse(InputStream stream) throws IOException, RecoverableException
    {
        try{
            SAXParserFactory spFactory = SAXParserFactory.newInstance();
            spFactory.setNamespaceAware(true);

            SAXParser parser = spFactory.newSAXParser();
            parser.parse(stream, this);
        }
        catch(SAXException e){
            throw ExceptionUtil.handleSAXException(e);
        }
        catch(ParserConfigurationException e){
            throw ExceptionUtil.handleParserConfigurationException(e);
        }
    }

    //-------------------------------------------------------------------------
    // SAX Handler
    //-------------------------------------------------------------------------
    /**
     * Start document.
     */
    public void startDocument() throws SAXException
    {
        _handler.initialize();
    }

    /**
     * Start element.
     */
    public void startElement(
        String uri, String localName, String qName, Attributes attributes)
        throws SAXException
    {
        _handler.enter(localName);

        _pickupText = false;

        SAXNode curr = _handler.getCurrentNode();
        if(curr != null)
        {
            if(curr == _datablock && _pdbCode.equals(""))
            {
                String value = attributes.getValue("datablockName");
                if(value != null && value.length() >= 4)
                {
                    _pdbCode = value.substring(0, 4).toUpperCase();
                }
            }
            else if(curr == _entry)
            {
                if( _pdbCode.equals("") ){
                    String value = attributes.getValue("id");
                    if(value != null) _pdbCode = value.toUpperCase();
                }
            }
            else if(curr == _pdbx_version)
            {
                String value = attributes.getValue("major_version");
                if (value != null) {
                    try {
                        int majorVer = Integer.parseInt(value);
                        if (majorVer >= 3) {
                            _ver3 = true;
                        }
                    } catch (NumberFormatException e) {
                    }
                }
            }
            else if(curr == _struct_conf)
            {
                _struct = true;
                _currStructInfo = new StructureInfo();

                String type = attributes.getValue("id");
                if(type == null){
                    _struct = false;
                }
                else if( type.startsWith("HELX") ){
                    _currStructInfo.setStructure(Residue.STRUCT_HELIX);
                }
                else if( type.startsWith("TURN") ){
                    _currStructInfo.setStructure(Residue.STRUCT_TURN);
                }
                else _struct = false;
            }
            else if(curr == _struct_sheet_range)
            {
                _struct = true;
                _currStructInfo = new StructureInfo();
                _currStructInfo.setStructure(Residue.STRUCT_SHEET);
            }
            else if(curr == _struct_mod_residue)
            {
                _modresCompId = null;
                _modresParentCompId = null;
            }
            else if(curr == _chem_comp)
            {
                _chemCompId = attributes.getValue("id");
            }
            else if(curr == _atom_site)
            {
                _atomRecord.clear();
                _atomRecord.setId( attributes.getValue("id") );
                _coordsDone = true;
            }
            else if(curr == _atom_record)
            {
                _currId = attributes.getValue("id");
                _coordsDone = true;
            }

            _pickupText = true;
            _buffer.setLength(0);
        }
    }

    /**
     * Characters.
     */
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        if(_pickupText){
            _buffer.append( new String(ch, start, length) );
        }
    }

    /**
     * End element.
     */
    public void endElement(String uri, String localName, String qName)
        throws SAXException
    {
        SAXNode curr = _handler.getCurrentNode();
        if(curr != null)
        {
            if(curr == _pdbx_description)
            {
                if( _compound.equals("") ) _compound = _buffer.toString();
            }
            else if(curr == _pdbx_keywords)
            {
                if( _classification.equals("") ) _classification = _buffer.toString();
            }
            else if(curr == _date_original)
            {
                if( _date.equals("") ) _date = _buffer.toString();
            }
            else if(curr == _exptl_method)
            {
                if( _experiment.equals("") ) _experiment = _buffer.toString();
            }
            else if(curr == _struct_conf || curr == _struct_sheet_range)
            {
                if(_struct) _builder.addStructureInfo(_currStructInfo);
            }
            else if(curr == _conf_beg_asym_id || curr == _sheet_beg_asym_id)
            {
                _currStructInfo.setInitChainId( _buffer.toString() );
            }
            else if(curr == _conf_end_asym_id || curr == _sheet_end_asym_id)
            {
                _currStructInfo.setTermChainId( _buffer.toString() );
            }
            else if(curr == _conf_beg_seq_id || curr == _sheet_beg_seq_id)
            {
                _currStructInfo.setInitSeqId( _buffer.toString() );
            }
            else if(curr == _conf_end_seq_id || curr == _sheet_end_seq_id)
            {
                _currStructInfo.setTermSeqId( _buffer.toString() );
            }
            else if(curr == _struct_mod_residue)
            {
                if (_modresCompId != null && _modresParentCompId != null) {
                    _builder.addModres(_modresCompId, _modresParentCompId);
                }
            }
            else if(curr == _modres_comp_id)
            {
                _modresCompId = _buffer.toString();
                if (_modresCompId.length() < 3) {
                    _modresCompId = fixResidueName(_modresCompId);
                }
            }
            else if(curr == _modres_parent_comp_id)
            {
                _modresParentCompId = _buffer.toString();
                if (_modresParentCompId.length() < 3) {
                    _modresParentCompId = fixResidueName(_modresParentCompId);
                }
            }
            else if(curr == _chem_comp_type)
            {
                if (_chemCompId != null) {
                    _builder.addChemCompType(_chemCompId, _buffer.toString());
                }
                _chemCompId = null;
            }
            else if(curr == _group_PDB)
            {
                _hetero = ( _buffer.toString().equals("HETATM") );
            }
            else if(curr == _pdbx_PDB_model_num)
            {
                _atomRecord.setMdlId( _buffer.toString() );
            }
            else if(curr == _label_asym_id)
            {
                _atomRecord.setLabelAsymId( _buffer.toString() );
            }
            else if(curr == _auth_asym_id)
            {
                _atomRecord.setAuthAsymId( _buffer.toString() );
            }
            else if(curr == _seq_id && !_hetero)
            {
                _atomRecord.setSeqId( _buffer.toString() );
            }
            else if(curr == _auth_seq_id && _hetero)
            {
                _atomRecord.setSeqId( _buffer.toString() );
            }
            else if(curr == _label_alt_id)
            {
                _atomRecord.setAltId( _buffer.toString() );
            }
            else if(curr == _comp_id)
            {
                _atomRecord.setCompId( _buffer.toString() );
            }
            else if(curr == _type_symbol)
            {
                _atomRecord.setTypeSymbol( _buffer.toString() );
            }
            else if(curr == _atom_id)
            {
                _atomRecord.setAtomId( _buffer.toString() );
            }
            else if(curr == _Cartn_x)
            {
                _atomRecord.setCartnX( _buffer.toString() );
            }
            else if(curr == _Cartn_y)
            {
                _atomRecord.setCartnY( _buffer.toString() );
            }
            else if(curr == _Cartn_z)
            {
                _atomRecord.setCartnZ( _buffer.toString() );
            }
            else if(curr == _occupancy)
            {
                _atomRecord.setOccupancy( _buffer.toString() );
            }
            else if(curr == _B_iso_or_equiv)
            {
                _atomRecord.setBFactor( _buffer.toString() );
            }
            else if(curr == _atom_site)
            {
                this.fixAtomRecord();
                _builder.addAtomRecord(_atomRecord, _hetero);
            }
            else if(curr == _atom_record)
            {
                this.processAtomRecord(_currId, _buffer.toString());
            }
        }

        _handler.leave();
    }

    /**
     * End document.
     */
    public void endDocument() throws SAXException
    {
        if(!_readExt) _mainDone = true;
    }

    //--------------------------------------------------------------------
    // Process atom records
    //--------------------------------------------------------------------
    private void processAtomRecord(String id, String record)
    {
        _atomRecord.clear();

        StringTokenizer st = new StringTokenizer(record);
        if(st.countTokens() < 18) return;

        boolean hetero = st.nextToken().equals("HETATM");

        _atomRecord.setId(id);
        _atomRecord.setMdlId( st.nextToken() );

        _atomRecord.setLabelAsymId( st.nextToken() );
        _atomRecord.setAuthAsymId( st.nextToken() );

        if(AUTH || hetero){
            st.nextToken(); // skip label_seq_id
            _atomRecord.setSeqId( st.nextToken() );
        }
        else{
            _atomRecord.setSeqId( st.nextToken() );
            st.nextToken(); // skip auth_seq_id
        }
        
        char insCode = st.nextToken().charAt(0);
        if (insCode == '?' || insCode == '.') {
            _atomRecord.setInsCode(' ');
        } else {
            _atomRecord.setInsCode(insCode);
        }

        _atomRecord.setAltId( st.nextToken() );

        if(AUTH){
            st.nextToken(); // skip label_comp_id
            _atomRecord.setCompId( st.nextToken() );
        }
        else{
            _atomRecord.setCompId( st.nextToken() );
            st.nextToken(); // skip auth_comp_id
        }

        _atomRecord.setTypeSymbol( st.nextToken() );

        if(AUTH){
            st.nextToken(); // skip label_atom_id
            _atomRecord.setAtomId( st.nextToken() );
        }
        else{
            _atomRecord.setAtomId( st.nextToken() );
            st.nextToken(); // skip auth_atom_id
        }

        _atomRecord.setCartnX( st.nextToken() );
        _atomRecord.setCartnY( st.nextToken() );
        _atomRecord.setCartnZ( st.nextToken() );
        _atomRecord.setOccupancy( st.nextToken() );
        _atomRecord.setBFactor( st.nextToken() );

        this.fixAtomRecord();

        _builder.addAtomRecord(_atomRecord, hetero);
    }

    private void fixAtomRecord()
    {
        // fix AltLoc
        String altLoc = _atomRecord.getAltId();
        if( altLoc.equals(".") || altLoc.equals("?") ){
            altLoc = "";
        }
        _atomRecord.setAltId( altLoc );

        // fix chain ID
        if( _atomRecord.getAuthAsymId().equals("?") ) _atomRecord.setAuthAsymId("");

        // fix length of comp_id
        String compId = _atomRecord.getCompId();
        if(compId.length() > 3){
            compId = compId.substring(0, 3);
        }
        else if(compId.length() < 3){
            compId = fixResidueName(compId);
        }
        _atomRecord.setCompId( compId );

        // fix length of atom_id
        String atomId = _atomRecord.getAtomId();
        String typeSymbol = _atomRecord.getTypeSymbol();
        if (atomId.length() >= 4) {
            atomId = atomId.substring(0, 4);
        } else if (atomId.length() < 4) {
            atomId = fixAtomName(atomId, typeSymbol);
        }
        _atomRecord.setAtomId(atomId);

        // fix auth_seq_id
        String seqId = _atomRecord.getSeqId();
        int length = seqId.length();
        if(length > 0){
            char last = seqId.charAt(length-1);
            if(last < '0' || last > '9'){
                _atomRecord.setSeqId( seqId.substring(0, length-1) );
            }
        }
    }
    
    private String fixResidueName(String compId) {
        String name = "  " + compId;
        return name.substring(name.length() - 3);
    }
    
    private String fixAtomName(String atomId, String typeSymbol) {
        String name = " " + atomId + "  ";

        int index = -1;
        if (typeSymbol != null && !typeSymbol.equals("")) {
            index = name.indexOf(typeSymbol);
        }

        if (index != -1) {
            if (typeSymbol.length() == 1) {
                index--;
            }
            return name.substring(index, index + 4);
        } else {
            return name.substring(0, 4);
        }
    }

    //--------------------------------------------------------------------
    // Set up SAXNodeHandler
    //--------------------------------------------------------------------
    private void setupHandler()
    {
        String TYPE = AUTH ? "auth" : "label";
        
        SAXNode root = new SAXNode("datablock");

        _datablock        = root;
        _entry            = root.createNode("entryCategory/entry");
        _pdbx_version     = root.createNode("pdbx_versionCategory/pdbx_version");
        _pdbx_description = root.createNode("entityCategory/entity/pdbx_description");
        _pdbx_keywords    = root.createNode("struct_keywordsCategory/struct_keywords/pdbx_keywords");
        _date_original    = root.createNode("database_PDB_revCategory/database_PDB_rev/date_original");
        _exptl_method     = root.createNode("exptlCategory/exptl/method");
                
        _struct_conf = root.createNode("struct_confCategory/struct_conf");
        _conf_beg_asym_id = _struct_conf.createNode("beg_" + TYPE + "_asym_id");
        _conf_end_asym_id = _struct_conf.createNode("end_" + TYPE + "_asym_id");
        _conf_beg_seq_id  = _struct_conf.createNode("beg_" + TYPE + "_seq_id");
        _conf_end_seq_id  = _struct_conf.createNode("end_" + TYPE + "_seq_id");

        _struct_sheet_range = root.createNode("struct_sheet_rangeCategory/struct_sheet_range");
        _sheet_beg_asym_id = _struct_sheet_range.createNode("beg_" + TYPE + "_asym_id");
        _sheet_end_asym_id = _struct_sheet_range.createNode("end_" + TYPE + "_asym_id");
        _sheet_beg_seq_id  = _struct_sheet_range.createNode("beg_" + TYPE + "_seq_id");
        _sheet_end_seq_id  = _struct_sheet_range.createNode("end_" + TYPE + "_seq_id");

        _struct_mod_residue = root.createNode("pdbx_struct_mod_residueCategory/pdbx_struct_mod_residue");
        _modres_comp_id = _struct_mod_residue.createNode(TYPE + "_comp_id");
        _modres_parent_comp_id = _struct_mod_residue.createNode("parent_comp_id");

        _chem_comp = root.createNode("chem_compCategory/chem_comp");
        _chem_comp_type = _chem_comp.createNode("type");

        _atom_site = root.createNode("atom_siteCategory/atom_site");
        _group_PDB          = _atom_site.createNode("group_PDB");
        _pdbx_PDB_model_num = _atom_site.createNode("pdbx_PDB_model_num");
        _label_asym_id      = _atom_site.createNode("label_asym_id");
        _auth_asym_id       = _atom_site.createNode("auth_asym_id");
        _seq_id             = _atom_site.createNode(TYPE + "_seq_id");
        _auth_seq_id        = _atom_site.createNode("auth_seq_id");
        _label_alt_id       = _atom_site.createNode("label_alt_id");
        _comp_id            = _atom_site.createNode(TYPE + "_comp_id");
        _type_symbol        = _atom_site.createNode("type_symbol");
        _atom_id            = _atom_site.createNode(TYPE + "_atom_id");
        _Cartn_x            = _atom_site.createNode("Cartn_x");
        _Cartn_y            = _atom_site.createNode("Cartn_y");
        _Cartn_z            = _atom_site.createNode("Cartn_z");
        _occupancy          = _atom_site.createNode("occupancy");
        _B_iso_or_equiv     = _atom_site.createNode("B_iso_or_equiv");

        _atom_record = root.createNode("category_atom_record/atom_record");

        _handler = new SAXNodeHandler(root);
    }
}

