/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.molecules;

import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.images.*;

/**
 * Represents atoms.
 */
public abstract class Atom extends AtomImage
{
    private int     _id;
    private int     _atomicNo;
    private int     _index;
    private String  _name;
    private float   _tempFactor;
    private String  _altLoc;
    private boolean _hetero;
    private boolean _bonded;
    private boolean _selected;
    private boolean _displayed;

    /**
     * Sets the atom ID.
     */
    public void setId(int id)
    {
        _id = id;
    }

    /**
     * Returns the atom ID.
     */
    public int getId()
    {
        return _id;
    }

    /**
     * Sets the atomic number.
     */
    public void setAtomicNo(int atomicNo)
    {
        _atomicNo = atomicNo;
    }

    /**
     * Returns the atomic number.
     */
    public int getAtomicNo()
    {
        return _atomicNo;
    }

    /**
     * Sets the index in AtomTable.
     */
    public void setIndex(int index)
    {
        _index = index;
    }

    /**
     * Returns the index in AtomTable.
     */
    public int getIndex()
    {
        return _index;
    }

    /**
     * Sets the name.
     */
    public void setName(String name)
    {
        _name = name;
    }

    /**
     * Returns the name.
     */
    public String getName()
    {
        return (_name != null) ? _name : AtomTable.getName(_index);
    }

    /**
     * Sets the x coordinate.
     */
    public abstract void setX(float x);

    /**
     * Returns the x coordinate.
     */
    public abstract float getX();

    /**
     * Sets the y coordinate.
     */
    public abstract void setY(float y);

    /**
     * Returns the y coordinate.
     */
    public abstract float getY();

    /**
     * Sets the z coordinate.
     */
    public abstract void setZ(float z);

    /**
     * Returns the z coordinate.
     */
    public abstract float getZ();

    /**
     * Sets the temperature factor.
     */
    public void setTempFactor(float tempFactor)
    {
        _tempFactor = tempFactor;
    }

    /**
     * Returns the temperature factor.
     */
    public float getTempFactor()
    {
        return _tempFactor;
    }

    /**
     * Sets the alternate location indicator.
     */
    public void setAltLoc(String altLoc)
    {
        _altLoc = altLoc;
    }

    /**
     * Returns the alternate location indicator.
     */
    public String getAltLoc()
    {
        return _altLoc;
    }

    /**
     * Sets whether the atom belongs to the hetero atom group.
     */
    public void setHetero(boolean hetero)
    {
        _hetero = hetero;
    }

    /**
     * Tests whether the atom belongs to the hetero atom group.
     */
    public boolean isHetero()
    {
        return _hetero;
    }

    /**
     * Sets whether the atom is bonded to other atoms.
     */
    public void setBonded(boolean bonded)
    {
        _bonded = bonded;
    }

    /**
     * Tests whether the atom is bonded to other atoms.
     */
    public boolean isBonded()
    {
        return _bonded;
    }

    /**
     * Sets whether the atom is selected.
     */
    public void setSelected(boolean selected)
    {
        _selected = selected;
    }

    /**
     * Tests whether the atom is selected.
     */
    public boolean isSelected()
    {
        return _selected;
    }

    /**
     * Sets whether this is displayed.
     */
    public void setDisplayed(boolean displayed)
    {
        _displayed = displayed;
    }

    /**
     * Tests whether this is displayed.
     */
    public boolean isDisplayed()
    {
        return _displayed;
    }

    /**
     * Returns the image.
     */
    public AtomImage getImage()
    {
        return this;
    }

    /**
     * Tests whether the atom belongs to the group speified by an index 
     * defined in AtomTable class.
     */
    public boolean isInGroup(int group)
    {
        return AtomTable.isInGroup(_index, group);
    }
}

