/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.molecules;

import java.util.ArrayList;
import org.pdbj.viewer.images.*;

/**
 * Represents molecules.
 */
public abstract class Molecule extends MoleculeImage
{
    public static final int NO_MODEL_ID = Integer.MIN_VALUE;

    public static final int STRUCT_NONE        = 0;
    public static final int STRUCT_FILE        = 1;
    public static final int STRUCT_CALCULATION = 2;
    public static final int STRUCT_GROMACS     = 3;

    private MoleculeFile _file;
    private int _modelId;
    private float _maxTempFactor;
    private float _minTempFactor;
    private float _maxHeteroTempFactor;
    private float _minHeteroTempFactor;
    private boolean _builtBond;
    private boolean _builtBackbone;
    private boolean _heteroAtoms;
    private boolean _hydrogens;

    private ArrayList _chains;
    private ArrayList _bonds;

    /**
     * Constructor.
     */
    public Molecule()
    {
        _modelId = NO_MODEL_ID;
        _chains  = new ArrayList();
        _bonds   = new ArrayList();
    }

    public abstract void setFrame(int frame);

    /**
     * Sets the molecule file.
     */
    public void setFile(MoleculeFile moleculeFile)
    {
        _file = moleculeFile;
    }

    /**
     * Returns the molecule file.
     */
    public MoleculeFile getFile()
    {
        return _file;
    }

    /**
     * Sets the model ID.
     */
    public void setModelId(int modelId)
    {
        _modelId = modelId;
    }

    /**
     * Returns the model ID.
     */
    public int getModelId()
    {
        return _modelId;
    }

    /**
     * Sets the x coordinate at the center of the molecule.
     */
    public abstract void setCenterX(float centerX);

    /**
     * Returns the x coordinate at the center of the molecule.
     */
    public abstract float getCenterX();

    /**
     * Sets the y coordinate at the center of the molecule.
     */
    public abstract void setCenterY(float centerY);

    /**
     * Returns the y coordinate at the center of the molecule.
     */
    public abstract float getCenterY();

    /**
     * Sets the z coordinate at the center of the molecule.
     */
    public abstract void setCenterZ(float centerZ);

    /**
     * Returns the z coordinate at the center of the molecule.
     */
    public abstract float getCenterZ();

    /**
     * Sets the radius of the molecule.
     */
    public abstract void setRadius(float radius);

    /**
     * Returns the radius of the molecule.
     */
    public abstract float getRadius();

    /**
     * Sets the minimum value of the x coordinate.
     */
    public abstract void setMinX(float minX);

    /**
     * Returns the minimum value of the x coordinate.
     */
    public abstract float getMinX();

    /**
     * Sets the minimum value of the y coordinate.
     */
    public abstract void setMinY(float minY);

    /**
     * Returns the minimum value of the y coordinate.
     */
    public abstract float getMinY();

    /**
     * Sets the minimum value of the z coordinate.
     */
    public abstract void setMinZ(float minZ);

    /**
     * Returns the minimum value of the z coordinate.
     */
    public abstract float getMinZ();

    /**
     * Sets the maximum value of the x coordinate.
     */
    public abstract void setMaxX(float maxX);

    /**
     * Returns the maximum value of the x coordinate.
     */
    public abstract float getMaxX();

    /**
     * Sets the maximum value of the y coordinate.
     */
    public abstract void setMaxY(float maxY);

    /**
     * Returns the maximum value of the y coordinate.
     */
    public abstract float getMaxY();

    /**
     * Sets the maximum value of the z coordinate.
     */
    public abstract void setMaxZ(float maxZ);

    /**
     * Returns the maximum value of the z coordinate.
     */
    public abstract float getMaxZ();

    /**
     * Sets the minimum value of the temperature factor of normal group atoms.
     */
    public void setMinTempFactor(float minTempFactor)
    {
        _minTempFactor = minTempFactor;
    }

    /**
     * Returns the minimum value of the temperature factor of normal group
     * atoms.
     */
    public float getMinTempFactor()
    {
        return _minTempFactor;
    }

    /**
     * Sets the maximum value of the temperature factor of normal group atoms.
     */
    public void setMaxTempFactor(float maxTempFactor)
    {
        _maxTempFactor = maxTempFactor;
    }

    /**
     * Returns the maximum value of the temperature factor of normal group
     * atoms.
     */
    public float getMaxTempFactor()
    {
        return _maxTempFactor;
    }

    /**
     * Sets the minimum value of the temperature factor of hetero group atoms.
     */
     public void setMinHeteroTempFactor(float minHeteroTempFactor)
    {
        _minHeteroTempFactor = minHeteroTempFactor;
    }

    /**
     * Returns the minimum value of the temperature factor of hetero group
     * atoms.
     */
    public float getMinHeteroTempFactor()
    {
        return _minHeteroTempFactor;
    }

    /**
     * Sets the maximum value of the temperature factor of hetero group atoms.
     */
     public void setMaxHeteroTempFactor(float maxHeteroTempFactor)
    {
        _maxHeteroTempFactor = maxHeteroTempFactor;
    }

    /**
     * Returns the maximum value of the temperature factor of hetero group
     * atoms.
     */
    public float getMaxHeteroTempFactor()
    {
        return _maxHeteroTempFactor;
    }

    /**
     * Sets whether bonds have been built.
     */
    public void setBuiltBond(boolean builtBond)
    {
        _builtBond = builtBond;
    }

    /**
     * Returns whether bonds have been built.
     */
    public boolean builtBond()
    {
        return _builtBond;
    }

    /**
     * Sets whether backbones have been built.
     */
    public void setBuiltBackbone(boolean builtBackbone)
    {
        _builtBackbone = builtBackbone;
    }

    /**
     * Returns whether backbones have been built.
     */
    public boolean builtBackbone()
    {
        return _builtBackbone;
    }

    /**
     * Sets whether hydrogen bonds have been built.
     */
    public abstract void setBuiltHBond(boolean builtHBond);

    /**
     * Returns whether hydrogen bonds have been built.
     */
    public abstract boolean builtHBond();

    /**
     * Sets whether disulfide bonds have been built.
     */
    public abstract void setBuiltSSBond(boolean builtSSBond);

    /**
     * Returns whether disulfide bonds have been built.
     */
    public abstract boolean builtSSBond();

    /**
     * Sets whether the ribbon model is checked for each chain.
     */
    public abstract void setRibbonChecked(boolean ribbonChecked);

    /**
     * Returns whether the ribbon model is checked for each chain.
     */
    public abstract boolean isRibbonChecked();

    /**
     * Sets the structure source.
     */
    public abstract void setStructure(int structure);

    /**
     * Returns the structure source.
     */
    public abstract int getStructure();

    /**
     * Sets the number of helices.
     */
    public abstract void setHelixCount(int helixCount);

    /**
     * Returns the number of helices.
     */
    public abstract int getHelixCount();

    /**
     * Sets the number of strands.
     */
    public abstract void setStrandCount(int strandCount);

    /**
     * Returns the number of strands.
     */
    public abstract int getStrandCount();

    /**
     * Sets the number of turns.
     */
    public abstract void setTurnCount(int turnCount);

    /**
     * Returns the number of turns.
     */
    public abstract int getTurnCount();

    /**
     * Sets whether the molecule has hetero group atoms.
     */
    public void setHeteroAtoms(boolean heteroAtoms)
    {
        _heteroAtoms = heteroAtoms;
    }

    /**
     * Tests whether the molecule has hetero group atoms.
     */
    public boolean hasHeteroAtoms()
    {
        return _heteroAtoms;
    }

    /**
     * Sets whether the molecule has hydrogens.
     */
    public void setHydrogens(boolean hydrogens)
    {
        _hydrogens = hydrogens;
    }

    /**
     * Tests whether the molecule has hydrogens.
     */
    public boolean hasHydrogens()
    {
        return _hydrogens;
    }

    /**
     * Sets the number of atoms selected.
     */
    public abstract void setSelectAtomCount(int selectAtomCount);

    /**
     * Returns the number of atoms selected.
     */
    public abstract int getSelectAtomCount();

    /**
     * Returns the image.
     */
    public MoleculeImage getImage()
    {
        return this;
    }

    /**
     * Returns the collection of chains.
     */
    public ArrayList getChains()
    {
        return _chains;
    }

    /**
     * Appends a chain.
     */
    public void addChain(Chain chain)
    {
        _chains.add(chain);
    }

    /**
     * Returns the number of chains.
     */
    public int numChains()
    {
        return _chains.size();
    }

    /**
     * Returns the chain at the specified index.
     */
    public Chain getChain(int index)
    {
        return (Chain)_chains.get(index);
    }

    /**
     * Returns the collection of bonds.
     */
    public ArrayList getBonds()
    {
        return _bonds;
    }

    /**
     * Appends a bond.
     */
    public void addBond(Bond bond)
    {
        _bonds.add(bond);
    }

    /**
     * Returns the number of bonds.
     */
    public int numBonds()
    {
        return _bonds.size();
    }

    /**
     * Returns the bond at the specified index.
     */
    public Bond getBond(int index)
    {
        return (Bond)_bonds.get(index);
    }

    /**
     * Returns the collection of hydrogen bonds.
     */
    public abstract ArrayList getHBonds();

    /**
     * Appends a hydrogen bond.
     */
    public abstract void addHBond(HBond bond);

    /**
     * Returns the number of hydrogen bonds.
     */
    public abstract int numHBonds();

    /**
     * Returns the hydrogen bond at the specified index.
     */
    public abstract HBond getHBond(int index);

    /**
     * Returns the collection of disulfide bonds.
     */
    public abstract ArrayList getSSBonds();

    /**
     * Appends a disulfide bond.
     */
    public abstract void addSSBond(SSBond bond);

    /**
     * Returns the number of disulfide bonds.
     */
    public abstract int numSSBonds();

    /**
     * Returns the disulfide bond at the specified index.
     */
    public abstract SSBond getSSBond(int index);

    public Bond appendBond() {
        Bond bond = createBond();
        _bonds.add(bond);
        return bond;
    }

    protected abstract Bond createBond();
}

