/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.molecules;

import java.util.*;
import org.pdbj.viewer.plus.*;
import org.pdbj.viewer.site.*;
import org.pdbj.viewer.utils.*;

/**
 * Represents molecule files.
 */
public abstract class MoleculeFile
{
    private boolean _pdbml;
    private boolean _ftp;
    private boolean _animation;
    
    private boolean _ver3;

    private String  _filePath;
    private String  _moleculeId;
    private String  _moleculeName;
    private String  _classification;
    private String  _date;
    private String  _experiment;

    private int     _fileId;
    private boolean _selected;
    private boolean _displayed;

    private String _text;
    private int _currFrame;

    private ArrayList _molecules;
    
    private Vector3f _offset;
    private Vector3f _center;
    private float    _radius;
    
    private Function _function;
    private HashMap _siteListMap;

    /**
     * Constructor.
     */
    public MoleculeFile()
    {
        _pdbml = true;
        _molecules = new ArrayList();
        _offset = new Vector3f();
        _center = new Vector3f();
        _siteListMap = new HashMap();
    }

    public abstract boolean isTrajectory();

    /**
     * Sets whether the source file is a PDBML file.
     */
    public void setPDBml(boolean pdbml)
    {
        _pdbml = pdbml;
    }

    /**
     * Tests whether the source file is a PDBML file.
     */
    public boolean isPDBml()
    {
        return _pdbml;
    }

    /**
     * Sets whether the source file is retrieved by FTP.
     */
    public void setFTP(boolean ftp)
    {
        _ftp = ftp;
    }

    /**
     * Tests whether the source file is retrieved by FTP.
     */
    public boolean isFTP()
    {
        return _ftp;
    }

    /**
     * Sets whether the source file is animation file.
     */
    public void setAnimation(boolean animation)
    {
        _animation = animation;
    }

    /**
     * Tests whether the source file is animation file.
     */
    public boolean isAnimation()
    {
        return _animation;
    }

    /**
     * Sets whether the file is PDB version 3.
     */
    public void setVer3(boolean ver3) {
        _ver3 = ver3;
    }

    /**
     * Returns whether the file is PDB version 3.
     */
    public boolean isVer3() {
        return _ver3;
    }

    /**
     * Sets the file path.
     */
    public void setFilePath(String filePath)
    {
        _filePath = filePath;
    }

    /**
     * Returns the file path.
     */
    public String getFilePath()
    {
        return _filePath;
    }

    /**
     * Sets the molecule ID (PDB code).
     */
    public void setMoleculeId(String moleculeId)
    {
        _moleculeId = moleculeId;
    }

    /**
     * Returns the molecule ID (PDB code).
     */
    public String getMoleculeId()
    {
        return _moleculeId;
    }

    /**
     * Sets the name of the molecule.
     */
    public void setMoleculeName(String moleculeName)
    {
        _moleculeName = moleculeName;
    }

    /**
     * Returns the name of the molecule.
     */
    public String getMoleculeName()
    {
        return _moleculeName;
    }

    /**
     * Sets the classification of the molecule.
     */
    public void setClassification(String classification)
    {
        _classification = classification;
    }

    /**
     * Returns the classification of the molecule.
     */
    public String getClassification()
    {
        return _classification;
    }

    /**
     * Sets the date stamp of the file.
     */
    public void setDate(String date)
    {
        _date = date;
    }

    /**
     * Returns the date stamp of the file.
     */
    public String getDate()
    {
        return _date;
    }

    /**
     * Sets the experimental description.
     */
    public void setExperiment(String experiment)
    {
        _experiment = experiment;
    }

    /**
     * Returns the experimental description.
     */
    public String getExperiment()
    {
        return _experiment;
    }

    /**
     * Returns the collection of molecules.
     */
    public ArrayList getMolecules()
    {
        return _molecules;
    }

    /**
     * Appends a molecule.
     */
    public void addMolecule(Molecule molecule)
    {
        _molecules.add(molecule);
    }

    /**
     * Returns the number of molecules.
     */
    public int numMolecules()
    {
        return _molecules.size();
    }

    /**
     * Returns the molecule at the specified index.
     */
    public Molecule getMolecule(int index)
    {
        Molecule molecule = (Molecule)_molecules.get(index);
        molecule.setFrame(index);
        return molecule;
    }

    /**
     * Returns the molecule specified by model ID.
     */
    public Molecule findMolecule(int modelId)
    {
        for(int i=0; i<_molecules.size(); i++){
            Molecule molecule = (Molecule)_molecules.get(i);
            if(molecule.getModelId() == modelId) return molecule;
        }
        return null;
    }

    /**
     * Sets the file ID.
     */
    public void setFileId(int fileId)
    {
        _fileId = fileId;
    }

    /**
     * Returns the file ID.
     */
    public int getFileId()
    {
        return _fileId;
    }

    /**
     * Sets whether this is selected.
     */
    public void setSelected(boolean selected)
    {
        _selected = selected;
    }

    /**
     * Tests whether this is selected.
     */
    public boolean isSelected()
    {
        return _selected;
    }

    /**
     * Sets whether this is displayed.
     */
    public void setDisplayed(boolean displayed)
    {
        _displayed = displayed;
    }

    /**
     * Tests whether this is displayed.
     */
    public boolean isDisplayed()
    {
        return _displayed;
    }

    /**
     * Sets the text.
     */
    public void setText(String text)
    {
        _text = text;
    }

    /**
     * Returns the text.
     */
    public String getText()
    {
        return _text;
    }

    /**
     * Sets the current frame.
     */
    public void setCurrentFrame(int currentFrame)
    {
        _currFrame = currentFrame;
    }

    /**
     * Returns the current frame.
     */
    public int getCurrentFrame()
    {
        return _currFrame;
    }
    
    /**
     * Returns the information.
     */
    public String getInformation()
    {
        StringBuffer sb = new StringBuffer();
        
        if(_molecules.size() > 0)
        {
            int cntChain   = 0;
            int cntResidue = 0;
            int cntAtom    = 0;
            int cntBond    = 0;
    
            MoleculeUtil util = new MoleculeUtil();
            Molecule molecule = (Molecule)_molecules.get(0);

            if( !molecule.builtBond() ) util.buildBonds(molecule);

            cntChain += molecule.numChains();
            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                cntResidue += chain.numResidues();
                for(int re=0; re<chain.numResidues(); re++)
                {
                    Residue residue = chain.getResidue(re);
                    cntAtom += residue.numAtoms();
                }
            }

            cntBond   += molecule.numBonds();
    
            sb.append("  Number of Chains ...... " + cntChain + "\n");
            sb.append("  Number of Groups ...... " + cntResidue + "\n");
            sb.append("  Number of Atoms ....... " + cntAtom + "\n");
            sb.append("  Number of Bonds ....... " + cntBond + "\n");
    
            Molecule firstModel = (Molecule)_molecules.get(0);
            String structure = null;
            if(firstModel.getStructure() == Molecule.STRUCT_FILE){
                structure = "PDB Data Records";
            }
            else if(firstModel.getStructure() == Molecule.STRUCT_CALCULATION){
                structure = "Calculated";
            }
            else if(firstModel.getStructure() == Molecule.STRUCT_GROMACS){
                structure = "Gromacs";
            }

            if(structure != null){
                sb.append("  Secondary Structure ... " + structure + "\n");
            }
            else{
                sb.append("  Secondary Structure ... " + "No Assignment" + "\n");
            }
        }

        return sb.toString();
    }

    /**
     * Returns the offset vector.
     */
    public Vector3f offset()
    {
        return _offset;
    }

    /**
     * Returns the center vector.
     */
    public Vector3f center()
    {
        return _center;
    }

    /**
     * Sets the radius.
     */
    public void setRadius(float radius)
    {
        _radius = radius;
    }

    /**
     * Returns the radius.
     */
    public float getRadius()
    {
        return _radius;
    }

    /**
     * Sets the function.
     */
    public void setFunction(Function function)
    {
        _function = function;
    }

    /**
     * Returns the function.
     */
    public Function getFunction()
    {
        return _function;
    }

    /**
     * Adds a site list.
     */
    public void addSiteList(String key, SiteList siteList) {
        _siteListMap.put(key, siteList);
    }

    /**
     * Returns the site list.
     */
    public SiteList getSiteList(String key) {
        return (SiteList) _siteListMap.get(key);
    }

    /**
     * Returns the number of atoms selected.
     */
    public int getSelectAtomCount()
    {
        int cnt = 0;
        for(int m=0; m<_molecules.size(); m++)
        {
            Molecule molecule = (Molecule)_molecules.get(m);
            cnt += molecule.getSelectAtomCount();
        }
        return cnt;
    }
}

