/*
 * This source code is a derived work of jV version 4.3.1 licensed as below.
 * March 2014
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.molecules;

import java.util.ArrayList;
import org.pdbj.viewer.images.*;

public class MoleculeFrameData {

    private int _frame;

    // molecule
    private boolean _builtHBond;
    private boolean _builtSSBond;
    private boolean _ribbonChecked;
    private int _structureSource;
    private int _helixCount;
    private int _strandCount;
    private int _turnCount;
    private boolean _bondInitialized;
    private boolean _backboneInitialized;
    private boolean _hbondInitialized;
    private boolean _ssbondInitialized;
    private boolean _refreshResidue;
    private float _time;

    // chain
    private boolean[] _ribbonDrawable;
    private float[][] _backboneDirectionX;
    private float[][] _backboneDirectionY;
    private float[][] _backboneDirectionZ;
    private float[][] _backboneNormalX;
    private float[][] _backboneNormalY;
    private float[][] _backboneNormalZ;
    private float[][] _backboneDistance;

    // residue
    private int[] _structure;
    private int[] _structureGromacs;
    private int[] _hbondIndex;
    private int[] _model;
    private int[] _colorIndex;
    private float[] _width;
    private int[] _listIndex;
    private int[] _backListIndex;
    private int[] _colorIndexGromacs;

    // atom
    private float[] _x;
    private float[] _y;
    private float[] _z;

    // bond
    private float[] _directionX;
    private float[] _directionY;
    private float[] _directionZ;
    private float[] _normalX;
    private float[] _normalY;
    private float[] _normalZ;
    private float[] _distance;

    private ArrayList _hbonds;
    private ArrayList _ssbonds;

    /**
     * Constructor.
     */
    public MoleculeFrameData(Molecule molecule) {
        _frame = 0;
        _builtHBond = false;
        _builtSSBond = false;
        _ribbonChecked = molecule.isRibbonChecked();
        _structureSource = molecule.getStructure();
        _helixCount = molecule.getHelixCount();
        _strandCount = molecule.getStrandCount();
        _turnCount = molecule.getTurnCount();
        _bondInitialized = molecule.bondInitialized();
        _backboneInitialized = molecule.backboneInitialized();
        _hbondInitialized = molecule.hbondInitialized();
        _ssbondInitialized = molecule.ssbondInitialized();
        _refreshResidue = molecule.refreshResidue();

        int numResidues = 0;
        int numAtoms = 0;
        for (int ch = 0; ch < molecule.numChains(); ch++) {
            Chain chain = molecule.getChain(ch);
            numResidues += chain.numResidues();
            for (int re = 0; re < chain.numResidues(); re++) {
                Residue residue = chain.getResidue(re);
                numAtoms += residue.numAtoms();
            }
        }

        // chain
        _ribbonDrawable = new boolean[molecule.numChains()];

        // residue
        _structure = new int[numResidues];
        _structureGromacs = new int[numResidues];
        _hbondIndex = new int[numResidues];
        _model = new int[numResidues];
        _colorIndex = new int[numResidues];
        _width = new float[numResidues];
        _listIndex = new int[numResidues];
        _backListIndex = new int[numResidues];
        _colorIndexGromacs = new int[numResidues];

        // atom
        _x = new float[numAtoms];
        _y = new float[numAtoms];
        _z = new float[numAtoms];

        _hbonds  = new ArrayList();
        _ssbonds = new ArrayList();
    }

    public MoleculeFrameData(int frame, MoleculeFrameData frameData) {
        _frame = frame;

        int numChains = frameData.getRibbonDrawable().length;
        int numResidues = frameData.getListIndex().length;
        int numAtoms = frameData.getX().length;
        int numBonds = frameData.getDirectionX().length;

        // chain
        _ribbonDrawable = new boolean[numChains];

        // residue
        _structure = new int[numResidues];
        _structureGromacs = new int[numResidues];
        _hbondIndex = new int[numResidues];
        _model = new int[numResidues];
        _colorIndex = new int[numResidues];
        _width = new float[numResidues];
        _listIndex = new int[numResidues];
        _backListIndex = new int[numResidues];
        _colorIndexGromacs = new int[numResidues];
        for (int i = 0; i < numResidues; i++) {
            _hbondIndex[i] = -1;
            _model[i] = ImageComponent.MODEL_NONE;
            _colorIndex[i] = -1;
            _colorIndexGromacs[i] = -1;
        }

        // atom
        _x = new float[numAtoms];
        _y = new float[numAtoms];
        _z = new float[numAtoms];

        _directionX = new float[numBonds];
        _directionY = new float[numBonds];
        _directionZ = new float[numBonds];
        _normalX = new float[numBonds];
        _normalY = new float[numBonds];
        _normalZ = new float[numBonds];
        _distance = new float[numBonds];

        _backboneDirectionX = new float[numChains][];
        _backboneDirectionY = new float[numChains][];
        _backboneDirectionZ = new float[numChains][];
        _backboneNormalX = new float[numChains][];
        _backboneNormalY = new float[numChains][];
        _backboneNormalZ = new float[numChains][];
        _backboneDistance = new float[numChains][];
        for (int ch = 0; ch < numChains; ch++) {
            int numBackbones = frameData.getBackboneDirectionX()[ch].length;
            _backboneDirectionX[ch] = new float[numBackbones];
            _backboneDirectionY[ch] = new float[numBackbones];
            _backboneDirectionZ[ch] = new float[numBackbones];
            _backboneNormalX[ch] = new float[numBackbones];
            _backboneNormalY[ch] = new float[numBackbones];
            _backboneNormalZ[ch] = new float[numBackbones];
            _backboneDistance[ch] = new float[numBackbones];
        }

        _hbonds  = new ArrayList();
        _ssbonds = new ArrayList();
    }

    public void setupBond(Molecule molecule) {
        int numBonds = molecule.numBonds();
        _directionX = new float[numBonds];
        _directionY = new float[numBonds];
        _directionZ = new float[numBonds];
        _normalX = new float[numBonds];
        _normalY = new float[numBonds];
        _normalZ = new float[numBonds];
        _distance = new float[numBonds];
    }

    public void setupBackbone(Molecule molecule) {
        int numChains = molecule.numChains();
        _backboneDirectionX = new float[numChains][];
        _backboneDirectionY = new float[numChains][];
        _backboneDirectionZ = new float[numChains][];
        _backboneNormalX = new float[numChains][];
        _backboneNormalY = new float[numChains][];
        _backboneNormalZ = new float[numChains][];
        _backboneDistance = new float[numChains][];
        for (int ch = 0; ch < molecule.numChains(); ch++) {
            Chain chain = molecule.getChain(ch);
            int numBackbones = chain.numBackbones();
            _backboneDirectionX[ch] = new float[numBackbones];
            _backboneDirectionY[ch] = new float[numBackbones];
            _backboneDirectionZ[ch] = new float[numBackbones];
            _backboneNormalX[ch] = new float[numBackbones];
            _backboneNormalY[ch] = new float[numBackbones];
            _backboneNormalZ[ch] = new float[numBackbones];
            _backboneDistance[ch] = new float[numBackbones];
        }
    }

    public int getFrame() {
        return _frame;
    }

    /**
     * Sets whether hydrogen bonds have been built.
     */
    public void setBuiltHBond(boolean builtHBond)
    {
        _builtHBond = builtHBond;
    }

    /**
     * Returns whether hydrogen bonds have been built.
     */
    public boolean builtHBond()
    {
        return _builtHBond;
    }

    /**
     * Sets whether disulfide bonds have been built.
     */
    public void setBuiltSSBond(boolean builtSSBond)
    {
        _builtSSBond = builtSSBond;
    }

    /**
     * Returns whether disulfide bonds have been built.
     */
    public boolean builtSSBond()
    {
        return _builtSSBond;
    }

    /**
     * Sets whether the ribbon model is checked for each chain.
     */
    public void setRibbonChecked(boolean ribbonChecked) {
        _ribbonChecked = ribbonChecked;
    }

    /**
     * Returns whether the ribbon model is checked for each chain.
     */
    public boolean isRibbonChecked() {
        return _ribbonChecked;
    }

    /**
     * Sets the structurre source.
     */
    public void setStructureSource(int structureSource)
    {
        _structureSource = structureSource;
    }

    /**
     * Returns the structure source.
     */
    public int getStructureSource()
    {
        return _structureSource;
    }

    /**
     * Sets the number of helices.
     */
    public void setHelixCount(int helixCount)
    {
        _helixCount = helixCount;
    }

    /**
     * Returns the number of helices.
     */
    public int getHelixCount()
    {
        return _helixCount;
    }

    /**
     * Sets the number of strands.
     */
    public void setStrandCount(int strandCount)
    {
        _strandCount = strandCount;
    }

    /**
     * Returns the number of strands.
     */
    public int getStrandCount()
    {
        return _strandCount;
    }

    /**
     * Sets the number of turns.
     */
    public void setTurnCount(int turnCount)
    {
        _turnCount = turnCount;
    }

    /**
     * Returns the number of turns.
     */
    public int getTurnCount()
    {
        return _turnCount;
    }

    /**
     * Sets whether bond images have been initialized.
     */
    public void setBondInitialized(boolean initialized)
    {
        _bondInitialized = initialized;
    }

    /**
     * Tests whether bond images have been initialized.
     */
    public boolean bondInitialized()
    {
        return _bondInitialized;
    }

    /**
     * Sets whether backbone images have been initialized.
     */
    public void setBackboneInitialized(boolean initialized)
    {
        _backboneInitialized = initialized;
    }
    
    /**
     * Tests whether backbone images have been initialized.
     */
    public boolean backboneInitialized()
    {
        return _backboneInitialized;
    }

    /**
     * Sets whether hbond images have been initialized.
     */
    public void setHBondInitialized(boolean initialized)
    {
        _hbondInitialized = initialized;
    }
    
    /**
     * Tests whether hbond images have been initialized.
     */
    public boolean hbondInitialized()
    {
        return _hbondInitialized;
    }

    /**
     * Sets whether ssbond images have been initialized.
     */
    public void setSSBondInitialized(boolean initialized)
    {
        _ssbondInitialized = initialized;
    }
    
    /**
     * Tests whether ssbond images have been initialized.
     */
    public boolean ssbondInitialized()
    {
        return _ssbondInitialized;
    }

    /**
     * Sets whether residue images should be refreshed.
     */
    public void setRefreshResidue(boolean refresh)
    {
        _refreshResidue = refresh;
    }

    /**
     * Tests whether residue images should be refreshed.
     */
    public boolean refreshResidue()
    {
        return _refreshResidue;
    }

    public void setTime(float time) {
        _time = time;
    }

    public float getTime() {
        return _time;
    }

    /**
     * Returns whether the ribbon can be drawn.
     */
    public boolean[] getRibbonDrawable()
    {
        return _ribbonDrawable;
    }

    /**
     * Returns the backbone direction vector x.
     */
    public float[][] getBackboneDirectionX()
    {
        return _backboneDirectionX;
    }

    /**
     * Returns the backbone direction vector y.
     */
    public float[][] getBackboneDirectionY()
    {
        return _backboneDirectionY;
    }

    /**
     * Returns the backbone direction vector z.
     */
    public float[][] getBackboneDirectionZ()
    {
        return _backboneDirectionZ;
    }

    /**
    * Returns the backbone normal vector x.
    */
    public float[][] getBackboneNormalX()
    {
        return _backboneNormalX;
    }

    /**
    * Returns the backbone normal vector y.
    */
    public float[][] getBackboneNormalY()
    {
        return _backboneNormalY;
    }

    /**
    * Returns the backbone normal vector z.
    */
    public float[][] getBackboneNormalZ()
    {
        return _backboneNormalZ;
    }

    /**
    * Returns the backbone distance.
    */
    public float[][] getBackboneDistance()
    {
        return _backboneDistance;
    }

    /**
     * Returns the secondary structure.
     */
    public int[] getStructure()
    {
        return _structure;
    }

    /**
     * Returns the secondary structure in Gromacs.
     */
    public int[] getStructureGromacs()
    {
        return _structureGromacs;
    }

    /**
     * Returns the index in the list of hydrogen bonds.
     */
    public int[] getHBondIndex()
    {
        return _hbondIndex;
    }

    public int[] getResidueModel() {
        return _model;
    }

    public int[] getResidueColorIndex() {
        return _colorIndex;
    }

    public float[] getResidueWidth() {
        return _width;
    }

    /**
     * Returns the list index.
     */
    public int[] getListIndex()
    {
        return _listIndex;
    }

    /**
     * Returns the back list index.
     */
    public int[] getBackListIndex()
    {
        return _backListIndex;
    }

    /**
     * Returns the color index for Gromacs.
     */
    public int[] getColorIndexGromacs()
    {
        return _colorIndexGromacs;
    }

    /**
     * Returns the x coordinate.
     */
    public float[] getX()
    {
        return _x;
    }

    /**
     * Returns the y coordinate.
     */
    public float[] getY()
    {
        return _y;
    }

    /**
     * Returns the z coordinate.
     */
    public float[] getZ()
    {
        return _z;
    }

    /**
     * Returns the direction vector x.
     */
    public float[] getDirectionX()
    {
        return _directionX;
    }

    /**
     * Returns the direction vector y.
     */
    public float[] getDirectionY()
    {
        return _directionY;
    }

    /**
     * Returns the direction vector z.
     */
    public float[] getDirectionZ()
    {
        return _directionZ;
    }

    /**
    * Returns the normal vector x.
    */
    public float[] getNormalX()
    {
        return _normalX;
    }

    /**
    * Returns the normal vector y.
    */
    public float[] getNormalY()
    {
        return _normalY;
    }

    /**
    * Returns the normal vector z.
    */
    public float[] getNormalZ()
    {
        return _normalZ;
    }

    /**
    * Returns the distance.
    */
    public float[] getDistance()
    {
        return _distance;
    }

    /**
     * Returns the collection of hydrogen bonds.
     */
    public ArrayList getHBonds()
    {
        return _hbonds;
    }

    /**
     * Appends a hydrogen bond.
     */
    public void addHBond(HBond bond)
    {
        _hbonds.add(bond);
    }

    /**
     * Returns the number of hydrogen bonds.
     */
    public int numHBonds()
    {
        return _hbonds.size();
    }

    /**
     * Returns the hydrogen bond at the specified index.
     */
    public HBond getHBond(int index)
    {
        return (HBond)_hbonds.get(index);
    }

    /**
     * Returns the collection of disulfide bonds.
     */
    public ArrayList getSSBonds()
    {
        return _ssbonds;
    }

    /**
     * Appends a disulfide bond.
     */
    public void addSSBond(SSBond bond)
    {
        _ssbonds.add(bond);
    }

    /**
     * Returns the number of disulfide bonds.
     */
    public int numSSBonds()
    {
        return _ssbonds.size();
    }

    /**
     * Returns the disulfide bond at the specified index.
     */
    public SSBond getSSBond(int index)
    {
        return (SSBond)_ssbonds.get(index);
    }
}
