/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.molecules;

import java.util.ArrayList;
import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.utils.*;

/**
 * Utility for molecules.
 */
public class MoleculeUtil
{
    private Molecule _molecule;

    /**
     * Creates bonds of the specified molecule.
     * When the molecule already has bonds, recreate them.
     */
    public void buildBonds(Molecule molecule)
    {
        _molecule = molecule;

        this.calculateBonds();
        _molecule.setBuiltBond(true);
    }

    /**
     * Creates backbones of the specified molecule.
     * When the molecule already has backbones, recreate them.
     */
    public void buildBackbones(Molecule molecule)
    {
        _molecule = molecule;

        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            this.calculateBackbones( molecule.getChain(ch) );
        }
        _molecule.setBuiltBackbone(true);
    }

    /**
     * Creates hydrogen bonds of the specified molecule.
     * When the molecule already has hydrogen bonds, recreate them.
     */
    public void buildHBonds(Molecule molecule)
    {
        _molecule = molecule;

        this.calculateHBonds();
        _molecule.setBuiltHBond(true);
    }

    /**
     * Creates disulfide bonds of the specified molecule.
     * When the molecule already has disulfide bonds, recreate them.
     */
    public void buildSSBonds(Molecule molecule)
    {
        _molecule = molecule;

        this.calculateSSBonds();
        _molecule.setBuiltSSBond(true);
    }

    /**
     * Determines secondary structure of the specified molecule.
     * When the molecule already has those information, override them.
     */
    public void determineStructure(Molecule molecule)
    {
        _molecule = molecule;

        // clear
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
                chain.getResidue(re).setStructure(Residue.STRUCT_NONE);
            }
        }

        if( !_molecule.builtHBond() )
        {
            this.calculateHBonds();
            _molecule.setBuiltHBond(true);
        }

        int helixCount  = findAlphaHelix(4);
        int strandCount = findBetaSheet();

        helixCount += findAlphaHelix(3);
        helixCount += findAlphaHelix(5);

        int turnCount   = findTurnStructure();

        _molecule.setHelixCount(helixCount);
        _molecule.setStrandCount(strandCount);
        _molecule.setTurnCount(turnCount);

        _molecule.setStructure(Molecule.STRUCT_CALCULATION);
    }

    /**
     * Checks whether the ribbon can be drawn for each chain.
     */
    public void checkRibbon(Molecule molecule)
    {
        if( !molecule.builtBackbone() )
        {
            this.buildBackbones(molecule);
        }

        _molecule = molecule;

        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            this.checkRibbon( molecule.getChain(ch) );
        }
        _molecule.setRibbonChecked(true);
    }

    //--------------------------------------------------------------------
    // Calculate bonds
    //--------------------------------------------------------------------

    private void calculateBonds()
    {
        _molecule.getBonds().clear();

        int div = 20;
        double maxDist = 600/250.0;

        double minX = _molecule.getMinX();
        double minY = _molecule.getMinY();
        double minZ = _molecule.getMinZ();
        double maxX = _molecule.getMaxX();
        double maxY = _molecule.getMaxY();
        double maxZ = _molecule.getMaxZ();

        double wx = maxX - minX;
        double wy = maxY - minY;
        double wz = maxZ - minZ;
        double dx = wx/div;
        double dy = wy/div;
        double dz = wz/div;

        ArrayList[][][] region = new ArrayList[div][div][div];
        for(int ix=0; ix<div; ix++){
        for(int iy=0; iy<div; iy++){
        for(int iz=0; iz<div; iz++){
            region[ix][iy][iz] = new ArrayList();
        }}}

        for(int ch=0; ch<_molecule.numChains(); ch++)
        {
        Chain chain = _molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
        Residue residue = chain.getResidue(re);
        for(int at1=0; at1<residue.numAtoms(); at1++)
        {
            Atom atom1 = residue.getAtom(at1);
            double x = atom1.getX() - minX;
            double y = atom1.getY() - minY;
            double z = atom1.getZ() - minZ;

            int lx = (x-maxDist<0) ? 0 : (int)((x-maxDist)/dx);
            int ly = (y-maxDist<0) ? 0 : (int)((y-maxDist)/dy);
            int lz = (z-maxDist<0) ? 0 : (int)((z-maxDist)/dz);
            int ux = (x+maxDist>=wx) ? div-1 : (int)((x+maxDist)/dx);
            int uy = (y+maxDist>=wy) ? div-1 : (int)((y+maxDist)/dy);
            int uz = (z+maxDist>=wz) ? div-1 : (int)((z+maxDist)/dz);

            for(int ix=lx; ix<=ux; ix++){
            for(int iy=ly; iy<=uy; iy++){
            for(int iz=lz; iz<=uz; iz++){
                for(int at2=0; at2<region[ix][iy][iz].size(); at2++)
                {
                    Atom atom2 = (Atom)region[ix][iy][iz].get(at2);

                    if(this.testBond(atom1, atom2)){
                        atom1.setBonded(true);
                        atom2.setBonded(true);
                        Bond bond = _molecule.appendBond();
                        bond.setSrcAtom(atom1);
                        bond.setDstAtom(atom2);
                    }
                }
            }}}

            int jx = (int)(x/dx); if(jx == div) jx--;
            int jy = (int)(y/dy); if(jy == div) jy--;
            int jz = (int)(z/dz); if(jz == div) jz--;

            region[jx][jy][jz].add(atom1);
        }}}
    }

    private boolean testBond(Atom atom1, Atom atom2)
    {
        String altLoc1 = atom1.getAltLoc();
        String altLoc2 = atom2.getAltLoc();

        if( !altLoc1.equals("") && !altLoc2.equals("")
                                && !altLoc1.equals(altLoc2) ) return false;

        double validDist
            =   ElementTable.getCovalentRadius(atom1.getAtomicNo())
              + ElementTable.getCovalentRadius(atom2.getAtomicNo())
              + 0.56;

        return testDistance(atom1, atom2, 0.16, validDist*validDist);
    }

    private boolean testDistance(
        Atom atom1, Atom atom2, double minSquared, double maxSquared)
    {
        double dx = atom1.getX() - atom2.getX();
        double dist = dx*dx;
        if(dist > maxSquared) return false;

        double dy = atom1.getY() - atom2.getY();
        dist += dy*dy;
        if(dist > maxSquared) return false;

        double dz = atom1.getZ() - atom2.getZ();
        dist += dz*dz;
        return (dist >= minSquared && dist <= maxSquared);
    }

    //--------------------------------------------------------------------
    // Calculate backbones
    //--------------------------------------------------------------------
    private void calculateBackbones(Chain chain)
    {
        chain.getBackbones().clear();

        Atom prevAtom = null;

        for(int re=0; re<chain.numResidues(); re++){
            Residue residue = chain.getResidue(re);
            if(residue.isProtein())
            {
                Atom currAtom = residue.getProteinCA();
                if(currAtom != null){
                    if(prevAtom != null){
                        // max = 17.64 = 4.2*4.2
                        if(testDistance(currAtom, prevAtom, 0, 17.64)){
                            Bond bond = chain.appendBackbone();
                            bond.setSrcAtom(currAtom);
                            bond.setDstAtom(prevAtom);
                        }
                        else{
                            residue.setChainBreak(true);
                        }
                    }
                    prevAtom = currAtom;
                }
            }
            else if(residue.isNucleotide())
            {
                Atom currAtom = residue.getNucleotideP();
                if(currAtom != null){
                    if(prevAtom != null){
                        if(testDistance(currAtom, prevAtom, 0, 225)){
                            Bond bond = chain.appendBackbone();
                            bond.setSrcAtom(prevAtom);
                            bond.setDstAtom(currAtom);
                        }
                        else{
                            residue.setChainBreak(true);
                        }
                    }
                    prevAtom = currAtom;
                }
            }            
        }
    }

    //--------------------------------------------------------------------
    // Calculate hydrogen bonds
    //--------------------------------------------------------------------
    private void calculateHBonds()
    {
        _molecule.getHBonds().clear();

        for(int ch=0; ch<_molecule.numChains(); ch++)
        {
            Chain chain = _molecule.getChain(ch);

            Residue first = chain.getResidue(0);
            if(first.isProtein())
            {
                this.calculateProteinHBonds(chain);
            }
            else if(first.isDNA())
            {
                this.calculateNucleicHBonds(chain);
            }
        }
    }

    private void calculateProteinHBonds(Chain chain)
    {
        Atom atomC = null;
        Atom atomO = null;

        for(int re1=0; re1<chain.numResidues(); re1++)
        {
            Residue residue1 = chain.getResidue(re1);

            if(atomC == null || atomO == null){
                atomC = residue1.findAtom(AtomTable.INDEX_C);
                atomO = residue1.findAtom(AtomTable.INDEX_O);
                continue;
            }

            Vector3f direction = new Vector3f(
                atomC.getX() - atomO.getX(),
                atomC.getY() - atomO.getY(),
                atomC.getZ() - atomO.getZ()
            );
            direction.normalize();

            atomC = residue1.findAtom(AtomTable.INDEX_C);
            atomO = residue1.findAtom(AtomTable.INDEX_O);

            if( !residue1.isAmino() ||
                residue1.getIndex() == ResidueTable.INDEX_PRO    )
            {
                continue;
            }

            Atom atomCA1 = residue1.getProteinCA();
            Atom atomN1  = residue1.findAtom(AtomTable.INDEX_N);
            if(atomCA1 == null || atomN1 == null) continue;

            // determine position of Hydrogen
            Vector3f coordN1 = new Vector3f(
                atomN1.getX(), atomN1.getY(), atomN1.getZ()
            );
            Vector3f coordH = new Vector3f();
            coordH.add(coordN1, direction);

            HBond best1 = null;
            HBond best2 = null;
            float energy1 = 0;
            float energy2 = 0;

            for(int re2=0; re2<chain.numResidues(); re2++)
            {
                if(re2 == re1 || re2 == re1-1) continue;

                Residue residue2 = chain.getResidue(re2);
                if(!residue2.isAmino()) continue;

                Atom atomCA2 = residue2.getProteinCA();
                if(atomCA2 == null) continue;

                if(!this.testDistance(atomCA1, atomCA2, 0, 81)) continue;

                Atom atomC2 = residue2.findAtom(AtomTable.INDEX_C);
                Atom atomO2 = residue2.findAtom(AtomTable.INDEX_O);
                if(atomC2 == null || atomO2 == null) continue;

                Vector3f coordC2 = new Vector3f(
                    atomC2.getX(), atomC2.getY(), atomC2.getZ()
                );
                Vector3f coordO2 = new Vector3f(
                    atomO2.getX(), atomO2.getY(), atomO2.getZ()
                );

                float energy
                    = getHBondEnergy(coordN1, coordH, coordO2, coordC2);

                if(energy < energy2){
                    HBond hbond = this.createHBond(
                        atomCA1, atomCA2, atomN1, atomO2, energy, re1-re2);

                    if(energy < energy1){
                        best2 = best1;
                        best1 = hbond;
                        energy2 = energy1;
                        energy1 = energy;
                    }
                    else{
                        best2 = hbond;
                        energy2 = energy;
                    }
                }
            }

            int index = _molecule.numHBonds();
            if(best1 != null){
                if(best2 != null){
                    _molecule.addHBond(best2);
                }
                _molecule.addHBond(best1);
                residue1.setHBondIndex(index);
            }
        }
    }

    private float getHBondEnergy(
        Vector3f coordN1, Vector3f coordH, Vector3f coordO2, Vector3f coordC2)
    {
        Vector3f vector = new Vector3f();
        
        vector.sub(coordN1, coordO2);
        float dno = vector.lengthSquared();
        if(dno < 0.25f) return -9900;
        vector.sub(coordN1, coordC2);
        float dnc = vector.lengthSquared();
        if(dnc < 0.25f) return -9900;
        vector.sub(coordH, coordO2);
        float dho = vector.lengthSquared();
        if(dho < 0.25f) return -9900;
        vector.sub(coordH, coordC2);
        float dhc = vector.lengthSquared();
        if(dhc < 0.25f) return -9900;

        dno = (float)Math.sqrt(dno);
        dnc = (float)Math.sqrt(dnc);
        dho = (float)Math.sqrt(dho);
        dhc = (float)Math.sqrt(dhc);

        // coupling const. = -332 * 0.42 * 0.2 * 1000 = -27888
        float energy = -27888 * (1/dho - 1/dhc - 1/dno + 1/dnc);

        if(energy < -9900){
            energy = -9900;
        }
        else if(energy > -500){
            energy = 0;
        }

        return energy;
    }

    private void calculateNucleicHBonds(Chain chain1)
    {
        for(int re1=0; re1<chain1.numResidues(); re1++)
        {
            Residue residue1 = chain1.getResidue(re1);
            if (!residue1.isDNA_A() && !residue1.isDNA_G())
            {
                continue;
            }

            Atom atomN1 = residue1.findAtom(AtomTable.INDEX_N1);
            if(atomN1 == null) continue;

            Residue nearest = null;
            Atom atomN3 = null;
            double currMin = 25; // 5.0*5.0
            for(int ch2=0; ch2<_molecule.numChains(); ch2++)
            {
                Chain chain2 = _molecule.getChain(ch2);
                if(chain2 == chain1) continue;

                Residue first = chain2.getResidue(0);
                if(!first.isDNA()) continue;

                for(int re2=0; re2<chain2.numResidues(); re2++)
                {
                    Residue residue2 = chain2.getResidue(re2);
                    if(!isComplementary(residue1, residue2)) continue;

                    Atom currAtomN3 = residue2.findAtom(AtomTable.INDEX_N3);
                    if(currAtomN3 == null) continue;

                    double dx = atomN1.getX() - currAtomN3.getX();
                    double dist = dx*dx;
                    if(dist >= currMin) continue;

                    double dy = atomN1.getY() - currAtomN3.getY();
                    dist += dy*dy;
                    if(dist >= currMin) continue;

                    double dz = atomN1.getZ() - currAtomN3.getZ();
                    dist += dz*dz;
                    if(dist < currMin){
                        nearest = residue2;
                        atomN3  = currAtomN3;
                        currMin = dist;
                    }
                }
            }

            int index = _molecule.numHBonds();
            if(nearest != null){
                Atom atomP1 = residue1.getNucleotideP();
                Atom atomP2 = nearest.getNucleotideP();
                _molecule.addHBond(
                    createHBond(atomP1, atomP2, atomN1, atomN3, 0, 0));

                if (residue1.isDNA_G()) {
                    Atom atomN2 = residue1.findAtom(AtomTable.INDEX_N2);
                    Atom atomO2 = nearest.findAtom(AtomTable.INDEX_O2);
                    if(atomN2 != null && atomO2 != null){
                        _molecule.addHBond(
                            createHBond(null, null, atomN2, atomO2, 0, 0));
                    }
                    Atom atomO6 = residue1.findAtom(AtomTable.INDEX_O6);
                    Atom atomN4 = nearest.findAtom(AtomTable.INDEX_N4);
                    if(atomO6 != null && atomN4 != null){
                        _molecule.addHBond(
                            createHBond(null, null, atomO6, atomN4, 0, 0));
                    }
                }
                else{
                    Atom atomN6 = residue1.findAtom(AtomTable.INDEX_N6);
                    Atom atomO4 = nearest.findAtom(AtomTable.INDEX_O4);
                    if(atomN6 != null && atomO4 != null){
                        _molecule.addHBond(
                            createHBond(null, null, atomN6, atomO4, 0, 0));
                    }
                }

                residue1.setHBondIndex(index);
            }
        }
    }

    private boolean isComplementary(Residue residue1, Residue residue2)
    {
        if (residue1.isDNA_A()) {
            return residue2.isDNA_T();
        }
        else{
            return residue2.isDNA_C();
        }
    }

    private HBond createHBond(
        Atom srcCA, Atom dstCA, Atom srcAtom, Atom dstAtom,
        float energy, int offset)
    {
        HBond hbond = new HBond();
        hbond.setSrcCA(srcCA);
        hbond.setDstCA(dstCA);
        hbond.setSrcAtom(srcAtom);
        hbond.setDstAtom(dstAtom);
        hbond.setEnergy(energy);
        hbond.setOffset(offset);

        return hbond;
    }

    //--------------------------------------------------------------------
    // Calculate disulfide bonds
    //--------------------------------------------------------------------
    private void calculateSSBonds()
    {
        _molecule.getSSBonds().clear();

        for(int ch1=0; ch1<_molecule.numChains(); ch1++)
        {
            Chain chain1 = _molecule.getChain(ch1);
            for(int re1=0; re1<chain1.numResidues(); re1++)
            {
                Residue residue1 = chain1.getResidue(re1);
                if(residue1.getIndex() != ResidueTable.INDEX_CYS) continue;
    
                Atom atomS1 = this.findSulfer(residue1);
                if(atomS1 == null) continue;
    
                for(int ch2=ch1; ch2<_molecule.numChains(); ch2++)
                {
                    Chain chain2 = _molecule.getChain(ch2);
                    int start = (ch2 == ch1) ? re1+1 : 0;
                    for(int re2=start; re2<chain2.numResidues(); re2++)
                    {
                        Residue residue2 = chain2.getResidue(re2);            
                        if(residue2.getIndex() != ResidueTable.INDEX_CYS) continue;
        
                        Atom atomS2 = this.findSulfer(residue2);
                        if(atomS2 == null) continue;
                        
                        if(this.testDistance(atomS1, atomS2, 0, 9))
                        {
                            Atom atomCA1 = residue1.getProteinCA();
                            Atom atomCA2 = residue2.getProteinCA();
    
                            SSBond ssbond = new SSBond();
                            ssbond.setSrcCA(atomCA2);
                            ssbond.setDstCA(atomCA1);
                            ssbond.setSrcAtom(atomS2);
                            ssbond.setDstAtom(atomS1);
    
                            _molecule.addSSBond(ssbond);
                        }
                    }
                }
            }
        }
    }

    private Atom findSulfer(Residue residue)
    {
        for(int at=0; at<residue.numAtoms(); at++)
        {
            Atom atom = residue.getAtom(at);
            if(atom.getAtomicNo() == 16) return atom;
        }
        return null;
    }

    //--------------------------------------------------------------------
    // DSSP structure determination
    //--------------------------------------------------------------------
    private int findAlphaHelix(int pitch)
    {
        int helixCount = 0;

        for(int ch=0; ch<_molecule.numChains(); ch++)
        {
            Chain chain = _molecule.getChain(ch);

            Residue first = chain.getResidue(0);
            if(!first.isProtein()) continue;

            int dstIndex = 0;
            boolean previous = false;
            int dist = 0;

            for(int re=0; re<chain.numResidues(); re++)
            {
                Residue srcResidue = chain.getResidue(re);
                Atom srcCA = srcResidue.getProteinCA();

                if( !srcResidue.isAmino() ||
                    srcCA == null )
                {
                    previous = false;
                }
                else if(dist == pitch){
                    Residue dstResidue = chain.getResidue(dstIndex);

                    if(!isHBonded(srcResidue, dstResidue)){
                        previous = false;
                    }
                    else if(previous){
                        if(dstResidue.getStructure() != Residue.STRUCT_HELIX){
                            helixCount++;
                        }
                        for(int i=dstIndex; i<re; i++){
                            Residue curr = chain.getResidue(i);
                            curr.setStructure(Residue.STRUCT_HELIX);
                        }
                    }
                    else{
                        previous = true;
                    }
                }

                if(srcResidue.getStructure() == Residue.STRUCT_HELIX){
                    dstIndex = re;
                    previous = false;
                    dist = 1;
                }
                else if(dist == pitch){
                    dstIndex++;
                }
                else{
                    dist++;
                }
            }
        }

        return helixCount;
    }

    private int findBetaSheet()
    {
        int strandCount = 0;

        Residue[] residues1 = new Residue[3];
        Residue[] residues2 = new Residue[3];

        for(int ch1=0; ch1<_molecule.numChains(); ch1++)
        {
            Chain chain1 = _molecule.getChain(ch1);

            Residue first1 = chain1.getResidue(0);
            if(!first1.isProtein()) continue;

            boolean ladder = false;
            for(int re1=0; re1<chain1.numResidues()-2; re1++)
            {
                residues1[0] = chain1.getResidue(re1);
                residues1[1] = chain1.getResidue(re1+1);
                residues1[2] = chain1.getResidue(re1+2);

                boolean found = (residues1[1].getStructure() == Residue.STRUCT_SHEET);
                boolean done = false;
                for(int ch2=ch1; !done && ch2<_molecule.numChains(); ch2++)
                {
                    Chain chain2 = _molecule.getChain(ch2);

                    Residue first2 = chain1.getResidue(0);
                    if(!first2.isProtein()) continue;
        
                    int start = (ch2 == ch1) ? re1+3 : 0;
                    for(int re2=start; re2<chain2.numResidues()-2; re2++)
                    {
                        residues2[0] = chain2.getResidue(re2);
                        residues2[1] = chain2.getResidue(re2+1);
                        residues2[2] = chain2.getResidue(re2+2);

                        if(this.testLadder(residues1, residues2)){
                            residues1[1].setStructure(Residue.STRUCT_SHEET);
                            residues2[1].setStructure(Residue.STRUCT_SHEET);
                            if(found){
                                done = true;
                                break;
                            }
                            found = true;
                        }
                    }
                }

                if(residues1[1].getStructure() == Residue.STRUCT_SHEET){
                    if(!ladder){
                        strandCount++;
                        ladder = true;
                    }
                }
                else{
                    ladder = false;
                }
            }
        }

        return strandCount;
    }

    private boolean testLadder(Residue[] residues1, Residue[] residues2)
    {
        Residue prev1 = residues1[0];
        Residue curr1 = residues1[1];
        Residue next1 = residues1[2];
        Residue prev2 = residues2[0];
        Residue curr2 = residues2[1];
        Residue next2 = residues2[2];

        // para ladder
        if( isHBonded(next1, curr2) && isHBonded(curr2, prev1)
            || isHBonded(next2, curr1) && isHBonded(curr1, prev2) )
        {
            return true;
        }

        // anti ladder
        if( isHBonded(next1, prev2) && isHBonded(next2, prev1)
            || isHBonded(curr2, curr1) && isHBonded(curr1, curr2) )
        {
            return true;
        }

        return false;
    }

    private int findTurnStructure()
    {
        int turnCount = 0;

        Residue[] residues = new Residue[5];

        for(int ch=0; ch<_molecule.numChains(); ch++)
        {
            Chain chain = _molecule.getChain(ch);

            Residue first = chain.getResidue(0);
            if(!first.isProtein()) continue;

            int cnt = 1;
            boolean found = false;
            for(int re=0; re<chain.numResidues(); re++)
            {
                Residue residue = chain.getResidue(re);
                if( residue.isChainBreak() ){
                    cnt = 1;
                    found = false;
                }

                residues[4] = residue;

                if(cnt == 5){
                    if(testTurnStructure(residues)){
                        if(!found){
                            turnCount++;
                            found = true;
                        }
                        residues[2].setStructure(Residue.STRUCT_TURN);
                    }
                    else{
                        found = false;
                    }
                }
                else{
                    cnt++;
                }

                for(int i=0; i<4; i++){
                    residues[i] = residues[i+1];
                }
            }
        }

        return turnCount;
    }

    private boolean testTurnStructure(Residue[] residues)
    {
        double COS_70_DEG = 0.34202014332567;

        Atom firstCA  = residues[0].getProteinCA();
        Atom middleCA = residues[2].getProteinCA();
        Atom lastCA   = residues[4].getProteinCA();

        if( residues[2].getStructure() != Residue.STRUCT_HELIX &&
            residues[2].getStructure() != Residue.STRUCT_SHEET &&
            firstCA != null && middleCA != null && lastCA != null)
        {
            Vector3f u = new Vector3f(
                middleCA.getX() - firstCA.getX(),
                middleCA.getY() - firstCA.getY(),
                middleCA.getZ() - firstCA.getZ()
            );
            Vector3f v = new Vector3f(
                lastCA.getX() - middleCA.getX(),
                lastCA.getY() - middleCA.getY(),
                lastCA.getZ() - middleCA.getZ()
            );

            double luv = u.length() * v.length();
            if(luv != 0){
                double cosK = (u.x*v.x + u.y*v.y + u.z*v.z)/luv;
                if(cosK < COS_70_DEG) return true;
            }
        }

        return false;
    }

    private boolean isHBonded(Residue srcResidue, Residue dstResidue)
    {
        int index = srcResidue.getHBondIndex();
        if(index != -1){
            Atom srcCA = srcResidue.getProteinCA();
            Atom dstCA = dstResidue.getProteinCA();
            while(index < _molecule.numHBonds()){
                HBond hBond = _molecule.getHBond(index);
                if(hBond.getSrcCA() != srcCA) break;
                if(hBond.getDstCA() == dstCA) return true;
                index++;
            }            
        }
        return false;
    }

    //--------------------------------------------------------------------
    // Check ribbon
    //--------------------------------------------------------------------
    private void checkRibbon(Chain chain)
    {
        Residue currResidue = null;
        Residue nextResidue = null;
        Atom currCA = null;
        Atom nextCA = null;

        currResidue = chain.getResidue(0);
        if( currResidue.isProtein() ){
            currCA = currResidue.getProteinCA();
        }
        else if( currResidue.isNucleotide() ){
            currCA = currResidue.getNucleotideP();
        }

        boolean done = false;
        for(int re=0; !done && re<chain.numResidues()-1; re++)
        {
            nextResidue = chain.getResidue(re+1);
            Atom currO = null;
            if(nextResidue.isProtein()){
                nextCA = nextResidue.getProteinCA();
                currO  = currResidue.findAtom(AtomTable.INDEX_O);
            }
            else{ // DNA
                nextCA = nextResidue.getNucleotideP();
                currO  = nextResidue.findAtom(AtomTable.INDEX_O5_ASTERISK);
            }
            if (currCA != null && nextCA != null && currO != null && !nextResidue.isChainBreak())
            {
                done = testGeometry(currResidue, currCA, currO, nextCA);
            }
            currResidue = nextResidue;
            currCA = nextCA;
        }
        if (done) {
            chain.setRibbonDrawable(true);
        }
    }

    private boolean testGeometry(Residue currResidue, Atom currCA, Atom currO,
            Atom nextCA) {
        Vector3f pointCurrCA = new Vector3f(currCA.getX(), currCA.getY(), currCA.getZ());
        Vector3f pointNextCA = new Vector3f(nextCA.getX(), nextCA.getY(), nextCA.getZ());
        Vector3f pointCurrO = new Vector3f(currO.getX(), currO.getY(), currO.getZ());

        Vector3f direction = new Vector3f();
        direction.sub(pointNextCA, pointCurrCA);

        Vector3f vCO = new Vector3f();
        if(currResidue.isProtein()){
            vCO.sub(pointCurrO, pointCurrCA);
        }
        else{
            Atom currO1P = currResidue.findAtom(AtomTable.INDEX_O1P);
            Atom currO2  = currResidue.findAtom(AtomTable.INDEX_O2_ASTERISK);
            if(currO1P != null && currO2 == null){ // DNA
                Vector3f pointCurrO1P = new Vector3f(
                    currO1P.getX(), currO1P.getY(), currO1P.getZ() );
                vCO.add(pointCurrO, pointCurrO1P);
                vCO.multiply(0.5f);
                vCO.sub(pointCurrCA);
            }
            else{ // RNA
                vCO.sub(pointCurrO, pointCurrCA);
            }
        }

        Vector3f vertical = new Vector3f();
        vertical.cross(direction, vCO);
        return vertical.lengthSquared() > 0;
    }
}

