/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.molecules;

import java.util.ArrayList;
import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.images.*;

/**
 * Represents residues.
 */
public abstract class Residue extends ResidueImage
{
    public static final int STRUCT_NONE  = 0;
    public static final int STRUCT_HELIX = 1;
    public static final int STRUCT_SHEET = 2;
    public static final int STRUCT_TURN  = 3;

    public static final int STRUCT_GROMACS_NONE = 0;
    public static final int STRUCT_GROMACS_COIL = 1;
    public static final int STRUCT_GROMACS_B_SHEET = 2;
    public static final int STRUCT_GROMACS_B_BRIDGE = 3;
    public static final int STRUCT_GROMACS_BEND = 4;
    public static final int STRUCT_GROMACS_TURN = 5;
    public static final int STRUCT_GROMACS_A_HELIX = 6;
    public static final int STRUCT_GROMACS_5_HELIX = 7;
    public static final int STRUCT_GROMACS_3_HELIX = 8;

    public static final int CHEM_COMP_TYPE_NONE = 0;
    public static final int CHEM_COMP_TYPE_AMINO = 1;
    public static final int CHEM_COMP_TYPE_DNA = 2;
    public static final int CHEM_COMP_TYPE_RNA = 3;

    private int _id;
    private char _insCode;
    private int _index;
    private int _parentIndex;
    private String _name;
    private boolean _chainBreak;

    private ArrayList _atoms;

    private static Atom _empty;
    private Atom _atomCA;
    private Atom _atomP;

    private int _chemCompType;

    /**
     * Constructor.
     */
    public Residue()
    {
        _atoms = new ArrayList();
        
        if(_empty == null) _empty = new EmptyAtom();
        _atomCA = _empty;
        _atomP  = _empty;
    }

    /**
     * Sets the residue ID.
     */
    public void setId(int id)
    {
        _id = id;
    }

    /**
     * Returns the residue ID.
     */
    public int getId()
    {
        return _id;
    }

    /**
     * Sets the insertion code.
     */
    public void setInsCode(char insCode)
    {
        _insCode = insCode;
    }

    /**
     * Returns the insertion code.
     */
    public char getInsCode()
    {
        return _insCode;
    }

    /**
     * Sets the index in ResidueTable.
     */
    public void setIndex(int index)
    {
        _index = index;
    }

    /**
     * Returns the index in ResidueTable.
     */
    public int getIndex()
    {
        return _index;
    }

    /**
     * Sets the parent index in ResidueTable.
     */
    public void setParentIndex(int parentIndex)
    {
        _parentIndex = parentIndex;
    }

    /**
     * Returns the parent index in ResidueTable.
     */
    public int getParentIndex()
    {
        return _parentIndex;
    }

    /**
     * Sets the name.
     */
    public void setName(String name)
    {
        _name = name;
    }

    /**
     * Returns the name.
     */
    public String getName()
    {
        return (_name != null) ? _name : ResidueTable.getName(_index);
    }

    /**
     * Sets the secondary structure.
     */
    public abstract void setStructure(int structure);

    /**
     * Returns the secondary structure.
     */
    public abstract int getStructure();

    /**
     * Sets the secondary structure in Gromacs.
     */
    public abstract void setStructureGromacs(int structureGromacs);

    /**
     * Returns the secondary structure in Gromacs.
     */
    public abstract int getStructureGromacs();

    /**
     * Sets the color index for Gromacs.
     */
    public abstract void setColorIndexGromacs(int colorIndexGromacs);

    /**
     * Returns the color index for Gromacs.
     */
    public abstract int getColorIndexGromacs();

    /**
     * Sets whether the chain breaks at this residue.
     */
    public void setChainBreak(boolean chainBreak)
    {
        _chainBreak = chainBreak;
    }

    /**
     * Tests whether the chain breaks at this residue.
     */
    public boolean isChainBreak()
    {
        return _chainBreak;
    }

    /**
     * Returns the image.
     */
    public ResidueImage getImage()
    {
        return this;
    }

    /**
     * Returns the collection of atoms.
     */
    public ArrayList getAtoms()
    {
        return _atoms;
    }

    /**
     * Appends an atom.
     */
    public void addAtom(Atom atom)
    {
        _atoms.add(atom);
    }

    /**
     * Returns the number of atoms.
     */
    public int numAtoms()
    {
        return _atoms.size();
    }

    /**
     * Returns the atom at the specified index.
     */
    public Atom getAtom(int index)
    {
        return (Atom)_atoms.get(index);
    }

    /**
     * Sets the index in the list of hydrogen bonds.
     */
    public abstract void setHBondIndex(int hbondIndex);

    /**
     * Returns the index in the list of hydrogen bonds.
     */
    public abstract int getHBondIndex();

    /**
     * Sets the chem-comp type.
     */
    public void setChemCompType(int chemCompType)
    {
        _chemCompType = chemCompType;
    }

    /**
     * Returns the chem-comp type.
     */
    public int getChemCompType()
    {
        return _chemCompType;
    }

    /**
     * Returns the C-alpha atom.
     * If not found, returns null.
     */
    public Atom getProteinCA()
    {
        if(_atomCA == _empty){
            _atomCA = this.findAtom(AtomTable.INDEX_CA);
        }
        return _atomCA;
    }

    /**
     * Returns the sugar phosphate.
     * If not found, returns null.
     */
    public Atom getNucleotideP()
    {
        if(_atomP == _empty){
            _atomP = this.findAtom(AtomTable.INDEX_P);
        }
        return _atomP;
    }

    /**
     * Finds atoms specified by an index defined in AtomTable class and
     * returns one that is found first.
     */
    public Atom findAtom(int atomIndex)
    {
        for(int i=0; i<_atoms.size(); i++){
            Atom atom = (Atom)_atoms.get(i);
            if(atom.getIndex() == atomIndex) return atom;
        }
        return null;
    }

    /**
     * Tests whether the residue has the property specified by an index 
     * defined in residue table.
     */
    public boolean hasProperty(int property)
    {
        return ResidueTable.hasProperty(_index, property);
    }

    /**
     * Tests whether the residue belongs to the group specified by an index 
     * defined in residue table.
     */
    public boolean isInGroup(int group)
    {
        return ResidueTable.isInGroup(_index, group);
    }

    private int getGroupIndex() {
        return (_index != ResidueTable.INDEX_OTHER) ? _index : _parentIndex;
    }

    /**
     * Tests whether the specified residue is in amino group.
     */
    public boolean isAmino() {
        return (ResidueTable.isInGroup(getGroupIndex(), ResidueTable.GROUP_AMINO)
                || _chemCompType == CHEM_COMP_TYPE_AMINO);
    }
    
    /**
     * Tests whether the specified residue is in protein group.
     */
    public boolean isProtein() {
        return (ResidueTable.isInGroup(getGroupIndex(), ResidueTable.GROUP_PROTEIN)
                || _chemCompType == CHEM_COMP_TYPE_AMINO);
    }
    
    /**
     * Tests whether the specified residue is in nucleotide group.
     */
    public boolean isNucleotide() {
        return (ResidueTable.isInGroup(getGroupIndex(), ResidueTable.GROUP_NUCLEOTIDE)
                || _chemCompType == CHEM_COMP_TYPE_DNA
                || _chemCompType == CHEM_COMP_TYPE_RNA);
    }
    
    /**
     * Tests whether the specified residue is in DNA group.
     */
    public boolean isDNA() {
        return (ResidueTable.isInGroup(getGroupIndex(), ResidueTable.GROUP_DNA)
                || _chemCompType == CHEM_COMP_TYPE_DNA);
    }
    
    /**
     * Tests whether the specified residue is in RNA group.
     */
    public boolean isRNA() {
        return (ResidueTable.isInGroup(getGroupIndex(), ResidueTable.GROUP_RNA)
                || _chemCompType == CHEM_COMP_TYPE_RNA);
    }
    
    /**
     * Tests whether the specified residue is adenine in DNA.
     */
    public boolean isDNA_A() {
        return ResidueTable.isDNA_A(_index);
    }
    
    /**
     * Tests whether the specified residue is cytosine in DNA.
     */
    public boolean isDNA_C() {
        return ResidueTable.isDNA_C(_index);
    }
    
    /**
     * Tests whether the specified residue is guanine in DNA.
     */
    public boolean isDNA_G() {
        return ResidueTable.isDNA_G(_index);
    }
    
    /**
     * Tests whether the specified residue is thymine in DNA.
     */
    public boolean isDNA_T() {
        return ResidueTable.isDNA_T(_index);
    }
    
    /**
     * Tests whether the specified residue is adenine in RNA.
     */
    public boolean isRNA_A() {
        return ResidueTable.isRNA_A(_index);
    }
    
    /**
     * Tests whether the specified residue is cytosine in RNA.
     */
    public boolean isRNA_C() {
        return ResidueTable.isRNA_C(_index);
    }
    
    /**
     * Tests whether the specified residue is guanine in RNA.
     */
    public boolean isRNA_G() {
        return ResidueTable.isRNA_G(_index);
    }
    
    /**
     * Tests whether the specified residue is uracil in RNA.
     */
    public boolean isRNA_U() {
        return ResidueTable.isRNA_U(_index);
    }
    
    /**
     * Selects the first AltLoc atoms.
     */
    public void selectFirstAltLoc()
    {
        String name = null;
        for(int at=0; at<this.numAtoms(); at++)
        {
            Atom atom = this.getAtom(at);
            boolean select = true;

            if( !atom.getAltLoc().equals("") ){
                if( atom.getName().equals(name) ){
                    select = false;
                }
                else name = atom.getName();
            }
            else name = null;

            atom.setSelected(select);
        }
    }

    private static class EmptyAtom extends Atom {
        public void setX(float x) {}
        public void setY(float y) {}
        public void setZ(float z) {}
        public float getX() {return 0;}
        public float getY() {return 0;}
        public float getZ() {return 0;}
    }
}

