/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.mouse;

import java.awt.Canvas;
import java.awt.event.*;
import java.text.NumberFormat;
import java.util.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.controller.*;

/**
 * Handles the mouse behavior.
 */
public class MouseBehavior implements MouseListener, MouseMotionListener
{
    public static final int PICKING_OFF         = 0;
    public static final int PICKING_ID          = 1;
    public static final int PICKING_COORDINATES = 2;
    public static final int PICKING_DISTANCE    = 3;
    public static final int PICKING_CENTER      = 4;
    public static final int PICKING_SELECT      = 5;

    private boolean _enabled;

    private int     _pickingMode;
    private boolean _rotation;
    private boolean _translation;
    private boolean _slabPlane;

    private boolean _usePickRadius;
    private float   _pickRadius;

    private ViewerController _viewerController;
    private MousePickListener _pickListener;
    private PickResult _pickResult;

    private Canvas _canvasL;
    private ViewStatus _viewStatus;

    private MessageArea _messageArea;
    private boolean _messageEnabled;

    private List _callbackReceivers;
    private List _mouseSettingsListeners;

    private int _width;
    private int _height;

    private int _mouseX;
    private int _mouseY;
    private int _pressX;
    private int _pressY;

    private float _tangent;
    private float _scaleXY;
    private float _scaleZ;

    private long _ticks;
    private static final long CLICK_INTERVAL = 500; // 0.5 second
    private boolean _clickModifier;

    private Vector4f _coord1;

    private static final boolean SCALE_HEIGHT = true;
    
    private NumberFormat _format;

    private Vector3f _camera;

    private Vector3f _vector1 = new Vector3f();    
    private Vector3f _vector2 = new Vector3f();    
    private Vector3f _vector3 = new Vector3f();    
    private Matrix4f _matrix1 = new Matrix4f();
    private Matrix4f _matrix2 = new Matrix4f();
    private Matrix4f _matrix3 = new Matrix4f();

    private Vector4f _vector4 = new Vector4f();    

    /**
    * Sets whether enabled.
    */
    public void setEnabled(boolean enabled)
    {
        _enabled = enabled;
    }

    /**
    * Returns whether enabled.
    */
    public boolean isEnabled()
    {
        return _enabled;
    }

    /**
     * Sets the picking mode.
     */
    public void setPickingMode(int mode) {
        if (_pickingMode != mode) {
            _pickingMode = mode;
            if (_pickingMode == PICKING_DISTANCE) {
                _coord1 = null;
            }
            notifyPickingModeChanged();
        }
    }

    /**
    * Returns the picking mode.
    */
    public int getPickingMode()
    {
        return _pickingMode;
    }

    /**
    * Sets whether or not to activate rotation.
    */
    public void setRotation(boolean enable)
    {
        _rotation = enable;
    }

    /**
    * Returns whether or not to activate rotation.
    */
    public boolean getRotation()
    {
        return _rotation;
    }

    /**
    * Sets whether or not to activate translation.
    */
    public void setTranslation(boolean enable)
    {
        _translation = enable;
    }

    /**
    * Returns whether or not to activate translation.
    */
    public boolean getTranslation()
    {
        return _translation;
    }

    /**
    * Sets whether or not to activate slab plane shift.
    */
    public void setSlabPlane(boolean enable)
    {
        _slabPlane = enable;
    }

    /**
    * Returns whether or not to activate slab plane shift.
    */
    public boolean getSlabPlane()
    {
        return _slabPlane;
    }

    /**
     * Sets whether to use the pick radius.
     */
    public void setUsePickRadius(boolean use)
    {
        _usePickRadius = use;
    }

    /**
     * Returns whether to use the pick radius.
     */
    public boolean getUsePickRadius()
    {
        return _usePickRadius;
    }

    /**
     * Sets the pick radius.
     */
    public void setPickRadius(float pickRadius)
    {
        _pickRadius = pickRadius;
    }

    /**
     * Returns the pick radius.
     */
    public float getPickRadius()
    {
        return _pickRadius;
    }

    /**
    * Sets the message area.
    */
    public void setMessageArea(MessageArea meesageArea)
    {
        _messageArea = meesageArea;
    }

    /**
    * Returns the message area.
    */
    public MessageArea getMessageArea()
    {
        return _messageArea;
    }

    /**
    * Sets whether the message is showed.
    */
    public void setMessageEnabled(boolean enable)
    {
        _messageEnabled = enable;
    }

    /**
    * Tests whether the message is showed.
    */
    public boolean messageEnabled()
    {
        return _messageEnabled;
    }

    /**
    * Constructor.
    */
    public MouseBehavior(ViewerController viewerController, MousePickListener pickListener)
    {
        _viewerController = viewerController;
        _pickListener = pickListener;
        _pickResult = new PickResult();

        _viewStatus = _viewerController.getViewStatus();

        _canvasL  = _viewerController.getCanvasL();
        
        _canvasL.addMouseListener(this);
        _canvasL.addMouseMotionListener(this);

        _canvasL.addComponentListener(
            new ComponentAdapter(){
                public void componentResized(ComponentEvent e){
                    handle_componentResized(e);
                }
            }
        );

        _tangent = (float)Math.tan( _viewStatus.getFieldAngle()/2 );

        _enabled = true;
        
        _pickingMode = PICKING_ID;
        _rotation    = true;
        _translation = true;
        _slabPlane   = true;

        _messageEnabled = true;

        this.updateSize();
        
        _format = NumberFormat.getNumberInstance();
        _format.setMaximumFractionDigits(3);
        _format.setMinimumFractionDigits(3);
        _format.setGroupingUsed(false);
        
        _callbackReceivers = new ArrayList();
        _mouseSettingsListeners = new ArrayList();

        _camera = new Vector3f();
    }

    /**
     * Initialize stereo.
     */
    public void initializeStereo()
    {
        Canvas canvasR = _viewerController.getCanvasR();
        canvasR.addMouseListener(this);
        canvasR.addMouseMotionListener(this);
    }

    /**
     * Appends a mouse callback receiver.
     */
    public void addMouseCallbackReceiver(MouseCallbackReceiver receiver)
    {
        _callbackReceivers.add(receiver);
    }

    /**
     * Returns the number of mouse callback receivers.
     */
    public int numMouseCallbackReceivers()
    {
        return _callbackReceivers.size();
    }

    /**
     * Returns the mouse callback receiver.
     */
    public MouseCallbackReceiver getMouseCallbackReceiver(int index)
    {
        return (MouseCallbackReceiver)_callbackReceivers.get(index);
    }

    /**
     * Removes a mouse callback receiver.
     */
    public void removeMouseCallbackReceiver(MouseCallbackReceiver receiver)
    {
        _callbackReceivers.remove(receiver);
    }

    /**
     * Adds mouse settings listeners.
     */
    public void addMouseSettingsListener(MouseSettingsListener mouseSettingsListener) {
        _mouseSettingsListeners.add(mouseSettingsListener);
    }

    //--------------------------------------------------------------------
    // Mouse event mask
    //--------------------------------------------------------------------
    public static boolean maskButton1(MouseEvent e)
    {
        return ( e.getModifiersEx() == MouseEvent.BUTTON1_DOWN_MASK );
    }

    public static boolean maskButton1Modifier1(MouseEvent e)
    {
        return ( e.getModifiersEx() == (MouseEvent.BUTTON1_DOWN_MASK | MouseEvent.SHIFT_DOWN_MASK) );
    }

    public static boolean maskButton1Modifier2(MouseEvent e)
    {
        return (
            e.getModifiersEx() == (MouseEvent.BUTTON1_DOWN_MASK|MouseEvent.ALT_DOWN_MASK) ||
            e.getModifiersEx() == (MouseEvent.BUTTON1_DOWN_MASK|MouseEvent.CTRL_DOWN_MASK)
        );
    }

    public static boolean maskButton3(MouseEvent e)
    {
        return (
            e.getModifiersEx() == MouseEvent.BUTTON3_DOWN_MASK ||
            e.getModifiersEx() == (MouseEvent.BUTTON1_DOWN_MASK | MouseEvent.META_DOWN_MASK)
        );
    }

    public static boolean maskButton3Modifier1(MouseEvent e)
    {
        return (
            e.getModifiersEx() == (MouseEvent.BUTTON3_DOWN_MASK | MouseEvent.SHIFT_DOWN_MASK) ||
            e.getModifiersEx() == (MouseEvent.BUTTON1_DOWN_MASK | MouseEvent.META_DOWN_MASK | MouseEvent.SHIFT_DOWN_MASK)
        );
    }

    //--------------------------------------------------------------------
    // Component event
    //--------------------------------------------------------------------
    private void handle_componentResized(ComponentEvent e)
    {
        this.updateSize();
    }

    //--------------------------------------------------------------------
    // Update parameters
    //--------------------------------------------------------------------
    private void updateSize()
    {
        _width  = _canvasL.getWidth();
        _height = _canvasL.getHeight();
    }

    private void updateScaleXY()
    {
        _viewStatus.getCameraPosition(_vector1);
        _viewStatus.getCurrentCenter(_vector2);
        
        float dz = _vector1.z - _vector2.z;
        if(dz < 0) dz *= -1;
        if(dz < 1) dz = 1;
        
        if(SCALE_HEIGHT){
            _scaleXY = (2 * _tangent/_height) * dz;
        }
        else{
            _scaleXY = (2 * _tangent/_width) * dz;
        }
    }

    private void updateScaleZ()
    {
        // empirically
        float coeff = 1.0f/200;

        _scaleZ = coeff*_viewStatus.getCurrentRadius();
    }

    //--------------------------------------------------------------------
    // Mouse event listener
    //--------------------------------------------------------------------
    public void mouseClicked(MouseEvent e)
    {
    }

    public void mouseEntered(MouseEvent e)
    {
    }

    public void mouseExited(MouseEvent e)
    {
    }

    public void mousePressed(MouseEvent e)
    {
        if(_enabled){
            this.handleMousePressed(e);
            this.callbackMousePressed(e);
        }
    }

    public void mouseReleased(MouseEvent e)
    {
        if(_enabled){
            this.handleMouseReleased(e);
        }
    }

    public void mouseMoved(MouseEvent e)
    {
    }

    public void mouseDragged(MouseEvent e)
    {
        if(_enabled){
            this.handleMouseDragged(e);
            this.callbackMouseDragged(e);
        }
    }

    //--------------------------------------------------------------------
    // Handle event
    //--------------------------------------------------------------------
    public void handleMousePressed(MouseEvent e)
    {
        _mouseX = e.getX();
        _mouseY = e.getY();
        _pressX = e.getX();
        _pressY = e.getY();
        if( maskButton1(e) )
        {
            _clickModifier = false;
            _ticks = System.currentTimeMillis();
        }
        else if( maskButton1Modifier1(e) )
        {
            _clickModifier = true;
            _ticks = System.currentTimeMillis();
        }
    }

    public void handleMouseReleased(MouseEvent e)
    {
        if( System.currentTimeMillis() - _ticks < CLICK_INTERVAL)
        {
            if (e.getX() == _pressX && e.getY() == _pressY) {
                boolean left = (e.getSource() == _canvasL);
                this.handleMousePicking(e.getX(), e.getY(), _clickModifier, left);
            }
        }
    }
    
    public void handleMouseDragged(MouseEvent e)
    {
        if( maskButton1(e) )
        {
            if(_rotation){
                float angleY = (float)(2*Math.PI*(e.getX() - _mouseX)/_width);
                float angleX = (float)(2*Math.PI*(e.getY() - _mouseY)/_height);
                _matrix1.rotationX(angleX);            
                _matrix2.rotationY(angleY);
                _matrix1.multiply(_matrix2, _matrix1);
                _viewerController.rotate(_matrix1);
            }
        }
        else if( maskButton1Modifier1(e) )
        {
            if(_translation){
                this.updateScaleZ();
                _vector1.x = 0;
                _vector1.y = 0;
                _vector1.z = (e.getY() - _mouseY)*_scaleZ;
                if (_viewerController.getViewStatus().isParallelProjection()) {
                    float dsize = - _tangent * _vector1.z;
                    _viewerController.modifyViewSize(dsize, false);
                }
                _viewerController.translate(_vector1);
            }
        }
        else if( maskButton1Modifier2(e) )
        {
            if(_slabPlane){
                this.updateScaleZ();
                float z = (e.getY() - _mouseY)*_scaleZ;
                _viewerController.moveSlabPlane(z);
            }
        }
        else if( maskButton3(e) )
        {
            if(_translation){
                this.updateScaleXY();
                _vector1.x =  (e.getX() - _mouseX)*_scaleXY;
                _vector1.y = -(e.getY() - _mouseY)*_scaleXY;
                _vector1.z = 0;
                _viewerController.translate(_vector1);
            }
        }
        else if( maskButton3Modifier1(e) )
        {
            if(_rotation){
                float angleZ = (float)(-2*Math.PI*(e.getX() - _mouseX)/_width);
                _matrix1.rotationZ(angleZ);            
                _viewerController.rotate(_matrix1);
            }
        }

        _mouseX = e.getX();
        _mouseY = e.getY();
    }

    //--------------------------------------------------------------------
    // Callback
    //--------------------------------------------------------------------
    private void callbackMousePressed(MouseEvent e)
    {
        for(int i=0; i<_callbackReceivers.size(); i++){
            MouseCallbackReceiver receiver
                = (MouseCallbackReceiver)_callbackReceivers.get(i);
            receiver.receiveMousePressed(e);
        }
    }

    private void callbackMouseDragged(MouseEvent e)
    {
        for(int i=0; i<_callbackReceivers.size(); i++){
            MouseCallbackReceiver receiver
                = (MouseCallbackReceiver)_callbackReceivers.get(i);
            receiver.receiveMouseDragged(e);
        }
    }

    //--------------------------------------------------------------------
    // Listener
    //--------------------------------------------------------------------
    private void notifyPickingModeChanged() {
        for (Iterator it = _mouseSettingsListeners.iterator(); it.hasNext();) {
            ((MouseSettingsListener) it.next()).pickingModeChanged(this);
        }
    }

    //--------------------------------------------------------------------
    // Mouse picking
    //--------------------------------------------------------------------
    private void handleMousePicking(int x, int y, boolean modifier, boolean left)
    {
        AtomPath atomPath = this.pickAtom(x, y, left);
        if(atomPath == null) return;

        if(_pickingMode == PICKING_OFF) return;

        Atom atom = atomPath.getAtom();
        Vector4f coord = new Vector4f(atom.getX(), atom.getY(), atom.getZ(), 1);
        Vector4f orig = new Vector4f(coord);
        coord.transform( atomPath.getMolecule().getImage().getTransform() );
        
        // set result
        if(_pickListener != null){
            _pickResult.setResult(atomPath, coord, orig);
            _pickListener.receiveMousePick(_pickResult);
        }

        if(_pickingMode == PICKING_ID || _pickingMode == PICKING_COORDINATES)
        {
            StringBuffer sb = new StringBuffer();
            writeId(atomPath, "", sb);
            if(_pickingMode == PICKING_COORDINATES){
                sb.append("\n   Coordinates:  " + this.toFormatString(coord) );
            }
            this.showMessage( sb.toString() );
        }
        else if(_pickingMode == PICKING_DISTANCE)
        {
            StringBuffer sb = new StringBuffer();
            if(_coord1 == null){
                writeId(atomPath, " #1", sb);
                _coord1 = coord;
            }
            else{
                Vector4f diff = new Vector4f();
                diff.sub(coord, _coord1);
                writeId(atomPath, " #2", sb);
                sb.append("\n");
                sb.append("Distance #1-#2:  " + diff.length());
                _coord1 = null;
            }
            this.showMessage( sb.toString() );
        }
        else if(_pickingMode == PICKING_CENTER)
        {
            Vector3f center = new Vector3f(coord.x, coord.y, coord.z);
            _viewerController.setRotationCenter(center);
            _viewerController.setScreenCenter(coord.x, coord.y);

            StringBuffer sb = new StringBuffer();
            sb.append("Rotating about ");
            writeId(atomPath, "", sb);
            sb.append("\n   Coordinates:  " + this.toFormatString(coord) );
            this.showMessage( sb.toString() );
        }
        else if(_pickingMode == PICKING_SELECT)
        {
            if(modifier){
                _viewerController.selectFile(
                    true, new int[]{ atomPath.getMoleculeFile().getFileId() }
                );
            }
            else{
                _viewerController.selectFile(
                    new int[]{ atomPath.getMoleculeFile().getFileId() }
                );
            }
        }
    }

    private void writeId(AtomPath atomPath, String num, StringBuffer sb) {
        //int fileId = atomPath.getMoleculeFile().getFileId();
        //int modelId = atomPath.getMolecule().getModelId();
        String chainId = atomPath.getChain().getId().trim();
        String labelAsymId = atomPath.getChain().getLabelId().trim();
        Residue residue = atomPath.getResidue();
        Atom atom = atomPath.getAtom();

        String type = null;
        if(residue.isProtein()){
            type = "Residue: ";
        }
        else if(residue.isNucleotide()){
            type = "Nucleotide: ";
        }
        else{
            type = "Group: ";
        }

        sb.append("Atom" + num + ": " + atom.getName() + " " + atom.getId() + "  ");
        sb.append(type + residue.getName() + " " + residue.getId() + "  ");
        if( !chainId.equals("") ){
            if (!labelAsymId.equals("")) {
                sb.append("Chain(auth/label): " + chainId + "/" + labelAsymId + "  ");
            } else {
                sb.append("Chain: " + chainId + "  ");
            }
        }
        /*
        if(modelId != Molecule.NO_MODEL_ID){
            sb.append("Model: " + modelId + "  ");
        }
        sb.append("File: " + fileId);
        */
    }

    private String toFormatString(Vector4f v)
    {
        StringBuffer sb = new StringBuffer();
        sb.append("( " + _format.format(v.x) + ", ");
        sb.append(       _format.format(v.y) + ", ");
        sb.append(       _format.format(v.z) + " )");
        return sb.toString();
    }

    private AtomPath pickAtom(int x, int y, boolean left)
    {
        _matrix3.setIdentity();

        // handle stereo
        if( _viewStatus.isStereo() ){
            _viewStatus.getStereoCenter(_vector1);
            _matrix3.translation(_vector1);
            _vector1.multiply(-1);
            _matrix1.translation(_vector1);
            float angle = _viewStatus.getStereoAngleRadians();
            if(!left) angle *= -1;
            _matrix2.rotationY(angle);
            _matrix2.multiply(_matrix2, _matrix1);
            _matrix3.multiply(_matrix3, _matrix2);            
        }

        // camera position        
        _viewerController.getViewStatus().getCameraPosition(_camera);

        // translation
        _vector1.set(_camera);
        _vector1.multiply(-1);
        _matrix1.translation(_vector1);

        float s = SCALE_HEIGHT ? _height : _width;
        float rx =  2*(float)(x - _width/2)/s;  // -1 <= rx <= 1
        float ry = -2*(float)(y - _height/2)/s; // -1 <= ry <= 1

        if( !_viewStatus.isParallelProjection() )
        {
            _vector1.set(1, 0, 0);
            _vector3.set(-rx, -ry, 1/_tangent);
            _vector3.normalize();
            _vector2.cross(_vector3, _vector1);
            _vector1.cross(_vector2, _vector3);

            _matrix2.setIdentity();
            _matrix2.m11 = _vector1.x; _matrix2.m12 = _vector1.y; _matrix2.m13 = _vector1.z;
            _matrix2.m21 = _vector2.x; _matrix2.m22 = _vector2.y; _matrix2.m23 = _vector2.z;
            _matrix2.m31 = _vector3.x; _matrix2.m32 = _vector3.y; _matrix2.m33 = _vector3.z;
        }
        else
        {
            float size = _viewStatus.getViewSize();
            rx *= size;
            ry *= size;
            _matrix2.translation(-rx, -ry, 0);
        }

        // transform
        _matrix1.multiply(_matrix2, _matrix1);
        _matrix1.multiply(_matrix1, _matrix3);
        
        // clip point
        float nearClip = _viewStatus.getNearClipping();

        if( !_viewStatus.isParallelProjection() ){
            _vector4.set(
                _camera.x + rx*_tangent*nearClip,
                _camera.y + ry*_tangent*nearClip,
                _camera.z - nearClip,
                1
            );
        }
        else{
            _vector4.set(
                _camera.x + rx,
                _camera.y + ry,
                _camera.z - nearClip,
                1
            );
        }
        _vector4.transform(_matrix1);

        // hit test
        AtomTester tester = new AtomTester();
        tester.clipZ = _vector4.z;
        if( _usePickRadius ){
            tester.pickRadius = _pickRadius;
        }

        ArrayList moleculeFiles = _viewerController.getMoleculeFiles();
        for(int i=0; i<moleculeFiles.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)moleculeFiles.get(i);
            if( !file.isDisplayed() ) continue;
            
            //for(int m=0; m<file.numMolecules(); m++)
            //{
                //Molecule molecule = file.getMolecule(m);
                Molecule molecule = file.getMolecule(file.getCurrentFrame());
                tester.currMolecule = molecule;
                _matrix2.multiply(_matrix1, molecule.getImage().getTransform());
                tester.transform = _matrix2;
                
                for(int ch=0; ch<molecule.numChains(); ch++)
                {
                    Chain chain = molecule.getChain(ch);
                    for(int b=0; b<chain.numBackbones(); b++)
                    {
                        tester.test( chain.getBackbone(b) );
                    }

                    for(int re=0; re<chain.numResidues(); re++)
                    {
                        Residue residue = chain.getResidue(re);
                        tester.test(residue);
                        for(int at=0; at<residue.numAtoms(); at++)
                        {
                            tester.test( residue.getAtom(at) );
                        }
                    }
                }

                for(int b=0; b<molecule.numBonds(); b++)
                {
                    tester.test( molecule.getBond(b) );
                }
                for(int b=0; b<molecule.numHBonds(); b++)
                {
                    tester.test( molecule.getHBond(b) );
                }
                for(int b=0; b<molecule.numSSBonds(); b++)
                {
                    tester.test( molecule.getSSBond(b) );
                }
            //}
        }

        if(tester.nearestAtom != null){
            return AtomPath.find(tester.nearestMolecule, tester.nearestAtom);
        }
        else{
            return null;
        }
    }

    //--------------------------------------------------------------------
    // Message
    //--------------------------------------------------------------------
    private void showMessage(String message)
    {
        if(_messageEnabled && _messageArea != null){
            _messageArea.appendMessage(message);
        }
    }
}
