/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.plus;

import java.util.*;

/**
 * Class that builds Function objects.
 */
public class FunctionBuilder
{
    private ArrayList _structSites;
    private ArrayList _structSiteGens;
    private ArrayList _goRefs;
    private LinkedHashMap _jvCommands;
    
    /**
     * Constructor.
     */
    public FunctionBuilder()
    {
        _structSites = new ArrayList();
        _structSiteGens = new ArrayList();
        _goRefs = new ArrayList();
        _jvCommands = new LinkedHashMap();
    }
    
    /**
     * Initializes.
     */
    public void initialize()
    {
        _structSites.clear();
        _structSiteGens.clear();
        _goRefs.clear();
        _jvCommands.clear();
    }

    /**
     * Appends a struct site data.
     */
    public void addStructSiteData(StructSiteData structSiteData)
    {
        StructSite structSite = new StructSite();
        structSite.setId(structSiteData.id);
        structSite.setInfoType(structSiteData.infoType);
        structSite.setInfoSubType(structSiteData.infoSubType);
        
        _structSites.add(structSite);
    }

    /**
     * Appends a struct site gen data.
     */
    public void addStructSiteGenData(StructSiteGenData structSiteGenData)
    {
        StructSiteGen structSiteGen = new StructSiteGen();
        structSiteGen.setSiteId(structSiteGenData.siteId);
        structSiteGen.setAuthAsymId(structSiteGenData.authAsymId);
        structSiteGen.setBegAuthSeqId(structSiteGenData.begAuthSeqId);
        structSiteGen.setEndAuthSeqId(structSiteGenData.endAuthSeqId);
        
        _structSiteGens.add(structSiteGen);
    }

    /**
     * Appends a GO_ref.
     */
    public void addGORefData(GORefData goRefData)
    {
        GORef goRef = this.findGORef(goRefData.goId);
        if(goRef != null) return;

        goRef = new GORef();
        goRef.setGOId(goRefData.goId);
        goRef.setAspect(goRefData.aspect);
        goRef.setContents(goRefData.contents);

        _goRefs.add(goRef);
    }

    /**
     * Appends a jV command.
     */
    public void addJvCommand(JvCommandData jvCommandData) {
        JvCommand jvCommand = new JvCommand();
        jvCommand.setName(jvCommandData.name);
        jvCommand.setDescription(jvCommandData.document);
        String[] commands = jvCommandData.command.split(";");
        for (int i = 0; i < commands.length; i++) {
            String command = commands[i].trim();
            if (command.length() > 0) {
                jvCommand.addCommand(command);
            }
        }
        if (jvCommand.getCommands().size() > 0) {
            _jvCommands.put(jvCommand.getName(), jvCommand);
        }
    }

    /**
     * Returns the result. The builder is initialized.
     */
    public Function getResult()
    {
        buildStructSites();

        Function function = new Function();
        for(int i=0; i<_structSites.size(); i++){
            function.addStructSite( (StructSite)_structSites.get(i) );
        }
        for(int i=0; i<_goRefs.size(); i++){
            function.addGORef( (GORef)_goRefs.get(i) );
        }
        for (Iterator it = _jvCommands.values().iterator(); it.hasNext();) {
            function.addJvCommand((JvCommand) it.next());
        }

        // initialize
        initialize();

        return function;
    }
    
    private void buildStructSites() {
        for (int i=0; i<_structSiteGens.size(); i++) {
            StructSiteGen structSiteGen = (StructSiteGen) _structSiteGens.get(i);
            StructSite structSite = findStructSite(structSiteGen.getSiteId());
            if (structSite != null) {
                structSite.addStructSiteGen(structSiteGen);
            }
        }
    }

    //--------------------------------------------------------------------
    // Util
    //--------------------------------------------------------------------
    private StructSite findStructSite(String id)
    {
        for(int i=0; i<_structSites.size(); i++){
            StructSite structSite = (StructSite)_structSites.get(i);
            if( structSite.getId().equals(id) ) return structSite;
        }
        return null;
    }

    private GORef findGORef(String goId)
    {
        for(int i=0; i<_goRefs.size(); i++){
            GORef goRef = (GORef)_goRefs.get(i);
            if( goRef.getGOId().equals(goId) ) return goRef;
        }
        return null;
    }
}
