/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.plus;

import java.io.*;
import java.net.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;

/**
 * Class that reads PDBMLplus files.
 */
public class FunctionReader extends DefaultHandler 
{
    private PDBmlUrl _pdbmlUrl;

    private FunctionBuilder _builder;
    private MessageListener _listener;

    private boolean _done;

    private boolean      _pickupText;
    private StringBuffer _buffer;

    private StructSiteData _structSiteData;
    private StructSiteGenData _structSiteGenData;
    private GORefData _goRefData;
    private JvCommandData _jvCommandData;

    private SAXNodeHandler _handler;

    private SAXNode _struct_site;
    private SAXNode _struct_site_gen;
    private SAXNode _auth_asym_id;
    private SAXNode _auth_seq_id;
    private SAXNode _beg_auth_seq_id;
    private SAXNode _end_auth_seq_id;
    private SAXNode _geneOntology;
    private SAXNode _goid;
    private SAXNode _namespace;
    private SAXNode _name;
    private SAXNode _jV_command;
    private SAXNode _command;
    private SAXNode _document;

    /**
     * Constructor.
     */
    public FunctionReader(MessageListener listener, PDBmlUrl pdbmlUrl)
    {
        _listener = listener;
        _pdbmlUrl = pdbmlUrl;

        _buffer = new StringBuffer();
        _structSiteData = new StructSiteData();
        _structSiteGenData = new StructSiteGenData();
        _goRefData = new GORefData();
        _jvCommandData = new JvCommandData();

        this.setupHandler();
    }

    /**
     * Reads the specified PDBMLplus file by FTP and creates the Function object.
     */
    public Function readFTP(String pdbCode) throws RecoverableException
    {
        if(_pdbmlUrl == null) return null;
        
        _builder = new FunctionBuilder();
        _builder.initialize();
        
        String name = pdbCode.toLowerCase() + "-pdbmlplus.xml";
        String path = _pdbmlUrl.getUrlFunction() + name;

        this.initialize();

        // show message
        if(_listener != null){
            _listener.begin(1);
            _listener.appendMessage("Retrieving " + name + " from " + _pdbmlUrl.getHostFunction());
        }

        try{
            this.readFile( new URL(path + ".gz") );

            if(!_done){
                this.readFile( new URL(path) );
            }
            if(!_done){
                throw new RecoverableException(ErrorMessages.CANT_FIND_FUNCTION_FILE);
            }
        }
        catch(MalformedURLException e){
            throw new RecoverableException(ErrorMessages.CANT_FIND_FUNCTION_FILE);
        }
        catch(UnknownHostException e){
            throw ExceptionUtil.handleUnknownHostException(e);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
        finally{
            if(_listener != null) _listener.end();
        }

        return _builder.getResult();
    }

    /**
     * Reads the specified PDBMLplus file and creates the Function object.
     */
    public Function read(URL url) throws RecoverableException {
        _builder = new FunctionBuilder();
        _builder.initialize();
        try {
            readFile(url);
        } catch (MalformedURLException e) {
            throw new RecoverableException(ErrorMessages.CANT_FIND_FUNCTION_FILE);
        } catch (UnknownHostException e) {
            throw ExceptionUtil.handleUnknownHostException(e);
        } catch (IOException e) {
            throw ExceptionUtil.handleIOException(e);
        }
        return _builder.getResult();
    }

    //--------------------------------------------------------------------
    // Read
    //--------------------------------------------------------------------
    private void initialize()
    {
        _done = false;
    }

    private void readFile(URL url) throws IOException, RecoverableException
    {
        ResourceRequest request = new ResourceRequest(url);

        InputStream stream = null;
        try{
            stream = request.openStream();
            this.parse(stream);            
        }
        catch(FileNotFoundException e){
        }
        finally{
            if(stream != null) stream.close();
            request.dispose();
        }
    }

    //--------------------------------------------------------------------
    // Parse
    //--------------------------------------------------------------------
    private void parse(InputStream stream) throws IOException, RecoverableException
    {
        try{
            SAXParserFactory spFactory = SAXParserFactory.newInstance();
            spFactory.setNamespaceAware(true);

            SAXParser parser = spFactory.newSAXParser();
            parser.parse(stream, this);
        }
        catch(SAXException e){
            throw ExceptionUtil.handleSAXException(e);
        }
        catch(ParserConfigurationException e){
            throw ExceptionUtil.handleParserConfigurationException(e);
        }
    }

    //-------------------------------------------------------------------------
    // SAX Handler
    //-------------------------------------------------------------------------
    /**
     * Start document.
     */
    public void startDocument() throws SAXException
    {
        _handler.initialize();
    }

    /**
     * Start element.
     */
    public void startElement(
        String uri, String localName, String qName, Attributes attributes)
        throws SAXException
    {
        _handler.enter(localName);

        _pickupText = false;

        SAXNode curr = _handler.getCurrentNode();
        if(curr != null)
        {
            if(curr == _struct_site)
            {
                _structSiteData.id = attributes.getValue("id");
                _structSiteData.infoType = attributes.getValue("info_type");
                _structSiteData.infoSubType = attributes.getValue("info_subtype");
            }
            else if(curr == _struct_site_gen)
            {
                _structSiteGenData.siteId = attributes.getValue("site_id");
            }
            else if(curr == _jV_command)
            {
                _jvCommandData.name = attributes.getValue("name");
            }

            _pickupText = true;
            _buffer.setLength(0);
        }
    }

    /**
     * Characters.
     */
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        if(_pickupText){
            _buffer.append( new String(ch, start, length) );
        }
    }

    /**
     * End element.
     */
    public void endElement(String uri, String localName, String qName)
        throws SAXException
    {
        SAXNode curr = _handler.getCurrentNode();
        if(curr != null)
        {
            if(curr == _struct_site)
            {
                _builder.addStructSiteData(_structSiteData);
                _structSiteData.clear();
            }
            else if(curr == _struct_site_gen)
            {
                _builder.addStructSiteGenData(_structSiteGenData);
                _structSiteGenData.clear();
            }
            else if(curr == _auth_asym_id)
            {
                _structSiteGenData.authAsymId = _buffer.toString();
            }
            else if(curr == _auth_seq_id)
            {
                _structSiteGenData.begAuthSeqId = _buffer.toString();
                _structSiteGenData.endAuthSeqId = _buffer.toString();
            }
            else if(curr == _beg_auth_seq_id)
            {
                _structSiteGenData.begAuthSeqId = _buffer.toString();
            }
            else if(curr == _end_auth_seq_id)
            {
                _structSiteGenData.endAuthSeqId = _buffer.toString();
            }
            else if(curr == _geneOntology)
            {
                _builder.addGORefData(_goRefData);
                _goRefData.clear();
            }
            else if(curr == _goid)
            {
                _goRefData.goId = _buffer.toString();
            }
            else if(curr == _namespace)
            {
                _goRefData.aspect = _buffer.toString();
            }
            else if(curr == _name)
            {
                _goRefData.contents = _buffer.toString();
            }
            else if(curr == _jV_command)
            {
                _builder.addJvCommand(_jvCommandData);
                _jvCommandData.clear();
            }
            else if(curr == _command)
            {
                _jvCommandData.command = _buffer.toString();
            }
            else if(curr == _document)
            {
                _jvCommandData.document = _buffer.toString();
            }
        }

        _handler.leave();
    }

    /**
     * End document.
     */
    public void endDocument() throws SAXException
    {
        _done = true;
    }

    //--------------------------------------------------------------------
    // Set up SAXNodeHandler
    //--------------------------------------------------------------------
    private void setupHandler()
    {
        SAXNode root = new SAXNode("datablock");

        _struct_site = root.createNode("struct_siteCategory/struct_site");

        _struct_site_gen = root.createNode("struct_site_genCategory/struct_site_gen");
        _auth_asym_id      = _struct_site_gen.createNode("auth_asym_id");
        _auth_seq_id       = _struct_site_gen.createNode("auth_seq_id");
        _beg_auth_seq_id   = _struct_site_gen.createNode("beg_auth_seq_id");
        _end_auth_seq_id   = _struct_site_gen.createNode("end_auth_seq_id");

        _geneOntology = root.createNode("gene_ontologyCategory/gene_ontology");
        _goid = _geneOntology.createNode("goid");
        _namespace = _geneOntology.createNode("namespace");
        _name = _geneOntology.createNode("name");

        _jV_command = root.createNode("jV_commandCategory/jV_command");
        _command = _jV_command.createNode("command");
        _document = _jV_command.createNode("document");

        _handler = new SAXNodeHandler(root);
    }
}
