/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.polygons;

import java.io.*;
import java.net.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;

/**
 * Class that reads CGO polygon files.
 */
public class CgoReader {

    private PolygonFileBuilder _builder;
    private MessageListener _listener;

    private VertexData _vertexData;
    private UserData _userData;

    private boolean _addVertex;

    /**
     * Constructor.
     */
    public CgoReader(MessageListener listener)
    {
        _listener = listener;
        _vertexData = new VertexData();
        _userData = new UserData();
    }

    /**
     * Reads the specified CGO polygon file and returns the PolygonFile object.
     */
    public PolygonFile read(File file) throws RecoverableException
    {
        _builder = new PolygonFileBuilder();
        _builder.setFilePath( file.getAbsolutePath() );
        _builder.useId(false);

        this.readFile( new ResourceRequest(file) );

        return _builder.getResult();
    }

    /**
     * Reads the specified CGO polygon file and returns the PolygonFile object.
     */
    public PolygonFile read(URL url) throws RecoverableException
    {
        _builder = new PolygonFileBuilder();
        _builder.setFilePath( url.toString() );
        _builder.useId(false);
        
        // message
        if(_listener != null){
            _listener.begin(1);
            _listener.appendMessage("Retrieving a polygon file");
        }

        try{
            this.readFile( new ResourceRequest(url) );
        }
        finally{
            if(_listener != null) _listener.end();
        }

        return _builder.getResult();
    }

    //--------------------------------------------------------------------
    // Parse
    //--------------------------------------------------------------------
    private void readFile(ResourceRequest request) throws RecoverableException
    {
        try{
            BufferedReader reader = null;
            try{
                reader = new BufferedReader(
                        new InputStreamReader(request.openStream()));
                read(reader);
            }
            finally{
                if (reader != null) reader.close();
                request.dispose();
            }
        }
        catch(FileNotFoundException e){
            throw new RecoverableException(ErrorMessages.CANT_FIND_POLYGON_FILE);
        }
        catch(UnknownHostException e){
            throw ExceptionUtil.handleUnknownHostException(e);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
        catch(java.security.AccessControlException e)
        {
            if( ExceptionUtil.isPropertyPermission(e) ){
                throw ExceptionUtil.handlePropertyPermission();
            }
            else{
                throw new RecoverableException(ErrorMessages.CANT_FIND_POLYGON_FILE);
            }
        }
    }

    private void read(BufferedReader reader) throws IOException
    {
        boolean inV = false;
        boolean inT = false;
        int id = 1;
        String s = null;
        while ((s = reader.readLine()) != null) {
            if (!inV && s.trim().equals("v = [")) {
                inV = true;
            } else if (!inT && s.trim().equals("t = [")) {
                inT = true;
            } else if (s.trim().equals("]")) {
                inV = false;
                inT = false;
            } else if (inV) {
                int idx1 = s.indexOf("[");
                int idx2 = s.indexOf("]");
                if (idx1 >=0 && idx2 >=0) {
                    processVertex(id, s.substring(idx1 + 1, idx2).trim());
                    if (_addVertex) {
                        _builder.addVertex(_vertexData, _userData);
                    } else {
                        _builder.addNullVertex(_userData);
                    }
                    id++;
                }
            } else if (inT) {
                processTriangle(s);
            }
        }
    }

    private void processVertex(int id, String values)
    {
        _addVertex = false;

        String[] tokens = values.split(",");
        if (tokens.length != 9) {
            return;
        }

        try {
            _vertexData.id = id;

            // coordinates
            _vertexData.x = Float.parseFloat(tokens[0].trim());
            _vertexData.y = Float.parseFloat(tokens[1].trim());
            _vertexData.z = Float.parseFloat(tokens[2].trim());

            // normal vector components
            _vertexData.nx = Float.parseFloat(tokens[3].trim());
            _vertexData.ny = Float.parseFloat(tokens[4].trim());
            _vertexData.nz = Float.parseFloat(tokens[5].trim());
            if(_vertexData.nx * _vertexData.nx + _vertexData.ny * _vertexData.ny + _vertexData.nz * _vertexData.nz == 0)
            {
                return;
            }

            // colors
            int r = (int)( 255 * Float.parseFloat(tokens[6].trim()) );
            int g = (int)( 255 * Float.parseFloat(tokens[7].trim()) );
            int b = (int)( 255 * Float.parseFloat(tokens[8].trim()) );
            if(r < 0 || r > 255 || g < 0 || g > 255 || b < 0 || b > 255)
            {
                return;
            }
            _vertexData.color = getColorValue(r, g, b, 255);

            _addVertex = true;

        } catch (NumberFormatException e) {
            return;
        }
    }

    private final int getColorValue(int r, int g, int b, int a) {
        return ((a & 0xFF) << 24) | ((r & 0xFF) << 16) | ((g & 0xFF) << 8) | ((b & 0xFF) << 0);
    }

    private void processTriangle(String values)
    {
        String[] tokens = values.trim().split(",");
        if (tokens.length != 3) {
            return;
        }

        try{
            int idx1 = Integer.parseInt(tokens[0].trim());
            int idx2 = Integer.parseInt(tokens[1].trim());
            int idx3 = Integer.parseInt(tokens[2].trim());
            _builder.addTriangle(idx1, idx2, idx3);
        }
        catch(NumberFormatException e){
        }
    }
}
