/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.polygons;

import java.util.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.images.*;

/**
 * Represents polygon file.
 */
public class PolygonFile
{
    private String  _filePath;
    private int     _fileId;
    private boolean _selected;
    private boolean _displayed;

    private Vertices _vertices;

    private BitSet _selectFlags;
    private BitSet _displayFlags;

    private int[] _pointIndices;
    private int[] _lineIndices;
    private int[] _triangleIndices;
    private int[] _quadIndices;

    private int[][] _polylineIndices;

    private float _pointTransparency;
    private float _lineTransparency;
    private float _triangleTransparency;
    private float _quadTransparency;
    private float _polylineTransparency;

    private float _pointSize;
    private float _lineWidth;
    private float _polylineWidth;

    private Vector3f _offset;
    private Vector3f _center;
    private float    _radius;

    private PolygonImage _image;
    
    private Hashtable _userDataTable;    

    private int _selectVertexCount;

    /**
     * Sets the file path.
     */
    public void setFilePath(String filePath)
    {
        _filePath = filePath;
    }

    /**
     * Returns the file path.
     */
    public String getFilePath()
    {
        return _filePath;
    }

    /**
     * Sets the file ID.
     */
    public void setFileId(int fileId)
    {
        _fileId = fileId;
    }

    /**
     * Returns the file ID.
     */
    public int getFileId()
    {
        return _fileId;
    }

    /**
     * Sets whether this is selected.
     */
    public void setSelected(boolean selected)
    {
        _selected = selected;
    }

    /**
     * Tests whether this is selected.
     */
    public boolean isSelected()
    {
        return _selected;
    }

    /**
     * Sets whether this is displayed.
     */
    public void setDisplayed(boolean displayed)
    {
        _displayed = displayed;
    }

    /**
     * Tests whether this is displayed.
     */
    public boolean isDisplayed()
    {
        return _displayed;
    }

    /**
     * Sets vertices.
     */
    public void setVertices(Vertices vertices)
    {
        _vertices = vertices;
        this.initializeFlags();
    }

    /**
     * Returns vertices.
     */
    public Vertices getVertices()
    {
        return _vertices;
    }

    /**
     * Returns the selection flags of vertices.
     */
    public BitSet getSelectFlags()
    {
        return _selectFlags;
    }
    
    /**
     * Returns the display flags of vertices.
     */
    public BitSet getDisplayFlags()
    {
        return _displayFlags;
    }
    
    /**
     * Sets indices of point array.
     */
    public void setPointIndices(int[] pointIndices)
    {
        _pointIndices = pointIndices;
    }

    /**
     * Returns indices of point array.
     */
    public int[] getPointIndices()
    {
        return _pointIndices;
    }

    /**
     * Sets indices of line array.
     */
    public void setLineIndices(int[] lineIndices)
    {
        _lineIndices = lineIndices;
    }

    /**
     * Returns indices of line array.
     */
    public int[] getLineIndices()
    {
        return _lineIndices;
    }

    /**
     * Sets indices of triangle array.
     */
    public void setTriangleIndices(int[] triangleIndices)
    {
        _triangleIndices = triangleIndices;
    }

    /**
     * Returns indices of triangle array.
     */
    public int[] getTriangleIndices()
    {
        return _triangleIndices;
    }

    /**
     * Sets indices of quad array.
     */
    public void setQuadIndices(int[] quadIndices)
    {
        _quadIndices = quadIndices;
    }

    /**
     * Returns indices of quad array.
     */
    public int[] getQuadIndices()
    {
        return _quadIndices;
    }

    /**
     * Sets indices of polyline array.
     */
    public void setPolylineIndices(int[][] polylineIndices)
    {
        _polylineIndices = polylineIndices;
    }

    /**
     * Returns indices of polyline array.
     */
    public int[][] getPolylineIndices()
    {
        return _polylineIndices;
    }

    /**
     * Sets the transparency of point array.
     */
    public void setPointTransparency(float pointTransparency)
    {
        _pointTransparency = pointTransparency;
    }

    /**
     * Returns the transparency of point array.
     */
    public float getPointTransparency()
    {
        return _pointTransparency;
    }

    /**
     * Sets the transparency of line array.
     */
    public void setLineTransparency(float lineTransparency)
    {
        _lineTransparency = lineTransparency;
    }

    /**
     * Returns the transparency of line array.
     */
    public float getLineTransparency()
    {
        return _lineTransparency;
    }

    /**
     * Sets the transparency of triangle array.
     */
    public void setTriangleTransparency(float triangleTransparency)
    {
        _triangleTransparency = triangleTransparency;
    }

    /**
     * Returns the transparency of triangle array.
     */
    public float getTriangleTransparency()
    {
        return _triangleTransparency;
    }

    /**
     * Sets the transparency of quad array.
     */
    public void setQuadTransparency(float quadTransparency)
    {
        _quadTransparency = quadTransparency;
    }

    /**
     * Returns the transparency of quad array.
     */
    public float getQuadTransparency()
    {
        return _quadTransparency;
    }

    /**
     * Sets the transparency of polyline array.
     */
    public void setPolylineTransparency(float polylineTransparency)
    {
        _polylineTransparency = polylineTransparency;
    }

    /**
     * Returns the transparency of polyline array.
     */
    public float getPolylineTransparency()
    {
        return _polylineTransparency;
    }

    /**
     * Sets the size of point array.
     */
    public void setPointSize(float pointSize)
    {
        _pointSize = pointSize;
    }

    /**
     * Returns the size of point array.
     */
    public float getPointSize()
    {
        return _pointSize;
    }

    /**
     * Sets the line width of line array.
     */
    public void setLineWidth(float lineWidth)
    {
        _lineWidth = lineWidth;
    }

    /**
     * Returns the line width of line array.
     */
    public float getLineWidth()
    {
        return _lineWidth;
    }

    /**
     * Sets the line width of polyline array.
     */
    public void setPolylineWidth(float polylineWidth)
    {
        _polylineWidth = polylineWidth;
    }

    /**
     * Returns the line width of polyline array.
     */
    public float getPolylineWidth()
    {
        return _polylineWidth;
    }

    /**
     * Returns the offset vector.
     */
    public Vector3f offset()
    {
        return _offset;
    }

    /**
     * Returns the center vector.
     */
    public Vector3f center()
    {
        return _center;
    }

    /**
     * Sets the radius.
     */
    public void setRadius(float radius)
    {
        _radius = radius;
    }

    /**
     * Returns the radius.
     */
    public float getRadius()
    {
        return _radius;
    }
    
    /**
     * Sets the image.
     */
    public void setImage(PolygonImage image)
    {
        _image = image;
    }
    
    /**
     * Returns the image.
     */
    public PolygonImage getImage()
    {
        return _image;
    }

    /**
     * Sets the user data table.
     */
    public void setUserDataTable(Hashtable table)
    {
        _userDataTable = table;
    }

    /**
     * Returns the number of user data.
     */
    public int numUserData()
    {
        return _userDataTable.size();
    }

    /**
     * Returns the names of user data (not ordered).
     */
    public String[] getUserDataNames()
    {
        int num = _userDataTable.size();
        String[] names = new String[num];
        Enumeration enum1 = _userDataTable.keys();
        for(int i=0; i<num; i++){
            names[i] = (String)enum1.nextElement();
        }
        return names;
    }

    /**
     * Returns the index for the specified user data.
     */
    public int getUserDataIndex(String name)
    {
        Integer index = (Integer)_userDataTable.get(name);
        return (index != null) ? index.intValue() : -1;
    }

    /**
     * Constructor.
     */
    public PolygonFile()
    {
        _selected  = true;
        _displayed = true;
        
        _offset = new Vector3f();
        _center = new Vector3f();
        _image = new PolygonImage();
    }
    
    private void initializeFlags()
    {
        int num = _vertices.length();

        _selectFlags  = new BitSet(num);
        _displayFlags = new BitSet(num);

        for (int i = 0; i < num; i++) {
            _selectFlags.set(i);
            _displayFlags.set(i);
        }
        _selectVertexCount = num;
    }

    /**
     * Sets the number of vertices selected.
     */
    public void setSelectVertexCount(int selectVertexCount) {
        _selectVertexCount = selectVertexCount;
    }

    /**
     * Returns the number of vertices selected.
     */
    public int getSelectVertexCount() {
        return _selectVertexCount;
    }
}
