/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.polygons;

import java.util.ArrayList;
import java.util.Hashtable;
import org.pdbj.viewer.utils.*;

/**
 * Creates PolygonFile object.
 */
public class PolygonFileBuilder
{
    private String _filePath;

    private ArrayList _vertices;

    private ArrayList _pointIndices;
    private ArrayList _lineIndices;
    private ArrayList _triangleIndices;
    private ArrayList _quadIndices;
    private ArrayList _polylineIndices;

    private float _pointTransparency;
    private float _lineTransparency;
    private float _triangleTransparency;
    private float _quadTransparency;
    private float _polylineTransparency;
    private float _pointSize;
    private float _lineWidth;
    private float _polylineWidth;

    private boolean _useId;
    private int _cnt;

    private Hashtable _idTable;
    private Hashtable _userDataTable;
    
    private VertexFinder _vertexFinder;

    private VertexData _nullData;

    /**
     * Constructor.
     */
    public PolygonFileBuilder()
    {
        _vertices        = new ArrayList();
        _pointIndices    = new ArrayList();
        _lineIndices     = new ArrayList();
        _triangleIndices = new ArrayList();
        _quadIndices     = new ArrayList();
        _polylineIndices = new ArrayList();

        _idTable = new Hashtable();
        _userDataTable = new Hashtable();

        _vertexFinder = new VertexFinder();
        
        _nullData = new VertexData();
        _nullData.id = Integer.MIN_VALUE;

        this.initialize();
    }

    /**
     * Initialize.
     */
    public void initialize()
    {
        _filePath = null;

        _vertices.clear();
        _pointIndices.clear();
        _lineIndices.clear();
        _triangleIndices.clear();
        _quadIndices.clear();
        _polylineIndices.clear();

        _idTable.clear();
        _userDataTable.clear();

        _vertexFinder.clear();
        
        _pointTransparency = 0;
        _lineTransparency = 0;
        _triangleTransparency = 0;
        _quadTransparency = 0;
        _polylineTransparency = 0;
        
        _pointSize = 1;
        _lineWidth = 1;
        _polylineWidth = 1;

        _useId = true;
        _cnt = 1;
    }

    /**
     * Sets the absolute path of the source file.
     */
    public void setFilePath(String filePath)
    {
        _filePath = filePath;
    }

    /**
     * Sets whether vertex ids are used.
     */
    public void useId(boolean use)
    {
        _useId = use;
    }

    /**
     * Appends a null vertex.
     */
    public void addNullVertex(UserData userData)
    {
        addVertex(_nullData, userData);
    }

    /**
     * Appends a vertex.
     */
    public void addVertex(VertexData data, UserData userData)
    {
        boolean add = true;
        int id = 0;

        if(_useId)
        {
            id = data.id;

            _idTable.put( new Integer(id), new Integer( _vertices.size() ) );
        }
        else
        {
            if(data != _nullData)
            {
                id = _cnt;
            }
            else
            {
                add = false;
            }

            _vertexFinder.addIndex(_cnt-1, add);

            _cnt++;
        }

        if(add)
        {
            Vertex vertex = new Vertex(id, data.x, data.y, data.z, data.nx, data.ny, data.nz, data.color);
            vertex.setUserData( this.processUserData(userData) );

            _vertices.add(vertex);
        }
    }

    /**
     * Appends a point.
     */
    public void addPoint(int vertex)
    {
        Integer index1 = this.findVertex(vertex);
        if(index1 != null){
            _pointIndices.add(index1);
        }
    }

    /**
     * Appends a line.
     */
    public void addLine(int vertex1, int vertex2)
    {
        Integer index1 = this.findVertex(vertex1);
        Integer index2 = this.findVertex(vertex2);
        if(index1 != null && index2 != null){
            _lineIndices.add(index1);
            _lineIndices.add(index2);
        }
    }

    /**
     * Appends a triangle.
     */
    public void addTriangle(int vertex1, int vertex2, int vertex3)
    {
        Integer index1 = this.findVertex(vertex1);
        Integer index2 = this.findVertex(vertex2);
        Integer index3 = this.findVertex(vertex3);
        if(index1 != null && index2 != null && index3 != null){
            _triangleIndices.add(index1);
            _triangleIndices.add(index2);
            _triangleIndices.add(index3);
        }
    }

    /**
     * Appends a quad.
     */
    public void addQuad(int vertex1, int vertex2, int vertex3, int vertex4)
    {
        Integer index1 = this.findVertex(vertex1);
        Integer index2 = this.findVertex(vertex2);
        Integer index3 = this.findVertex(vertex3);
        Integer index4 = this.findVertex(vertex4);
        if(index1 != null && index2 != null && index3 != null && index4 != null){
            _quadIndices.add(index1);
            _quadIndices.add(index2);
            _quadIndices.add(index3);
            _quadIndices.add(index4);
        }
    }

    /**
     * Appends a polyline.
     */
    public void addPolyline(int[] vertices)
    {
        int num = vertices.length;
        int[] indices = new int[num];
        for(int i=0; i<num; i++){
            Integer index = this.findVertex(vertices[i]);
            if(index != null){
                indices[i] = index.intValue();
            }
            else return;
        }
        _polylineIndices.add(indices);
    }

    /**
     * Set transparency of point list.
     */
    public void setPointTransparency(float transparency)
    {
        _pointTransparency = transparency;
    }

    /**
     * Set transparency of line list.
     */
    public void setLineTransparency(float transparency)
    {
        _lineTransparency = transparency;
    }

    /**
     * Set transparency of triangle list.
     */
    public void setTriangleTransparency(float transparency)
    {
        _triangleTransparency = transparency;
    }

    /**
     * Set transparency of quad list.
     */
    public void setQuadTransparency(float transparency)
    {
        _quadTransparency = transparency;
    }

    /**
     * Set transparency of polyline list.
     */
    public void setPolylineTransparency(float transparency)
    {
        _polylineTransparency = transparency;
    }

    /**
     * Sets the size of point array.
     */
    public void setPointSize(float pointSize)
    {
        _pointSize = pointSize;
    }

    /**
     * Sets the line width of line array.
     */
    public void setLineWidth(float lineWidth)
    {
        _lineWidth = lineWidth;
    }

    /**
     * Sets the line width of polyline array.
     */
    public void setPolylineWidth(float polylineWidth)
    {
        _polylineWidth = polylineWidth;
    }

    /**
     * Creates PolygonFile object.
     */
    public PolygonFile getResult()
    {
        PolygonFile polygonFile = new PolygonFile();

        polygonFile.setFilePath(_filePath);
        polygonFile.setPointTransparency(_pointTransparency);
        polygonFile.setLineTransparency(_lineTransparency);
        polygonFile.setTriangleTransparency(_triangleTransparency);
        polygonFile.setQuadTransparency(_quadTransparency);
        polygonFile.setPolylineTransparency(_polylineTransparency);
        
        polygonFile.setPointSize(_pointSize);
        polygonFile.setLineWidth(_lineWidth);
        polygonFile.setPolylineWidth(_polylineWidth);

        this.fixUserData();

        Vertices vertices = new Vertices();
        vertices.set(_vertices);
        polygonFile.setVertices(vertices);

        boolean[] used = new boolean[_vertices.size()];

        int[] pointIndices = new int[_pointIndices.size()];
        for(int i=0; i<_pointIndices.size(); i++){
            int index = ( (Integer)_pointIndices.get(i) ).intValue();
            pointIndices[i] = index;
            used[index] = true;
        }
        polygonFile.setPointIndices(pointIndices);
        
        int[] lineIndices = new int[_lineIndices.size()];
        for(int i=0; i<_lineIndices.size(); i++){
            int index = ( (Integer)_lineIndices.get(i) ).intValue();
            lineIndices[i] = index;
            used[index] = true;
        }
        polygonFile.setLineIndices(lineIndices);
        
        int[] triangleIndices = new int[_triangleIndices.size()];
        for(int i=0; i<_triangleIndices.size(); i++){
            int index = ( (Integer)_triangleIndices.get(i) ).intValue();
            triangleIndices[i] = index;
            used[index] = true;
        }
        polygonFile.setTriangleIndices(triangleIndices);
        
        int[] quadIndices = new int[_quadIndices.size()];
        for(int i=0; i<_quadIndices.size(); i++){
            int index = ( (Integer)_quadIndices.get(i) ).intValue();
            quadIndices[i] = index;
            used[index] = true;
        }
        polygonFile.setQuadIndices(quadIndices);

        int[][] polylineIndices = new int[_polylineIndices.size()][];
        for(int i=0; i<_polylineIndices.size(); i++){
            int[] indices = (int[])_polylineIndices.get(i);
            polylineIndices[i] = indices;
            for(int j=0; j<indices.length; j++){
                used[indices[j]] = true;
            }
        }
        polygonFile.setPolylineIndices(polylineIndices);

        Vector3f center = new Vector3f();
        float radius = this.determineSize(used, center);

        polygonFile.center().set(center);
        polygonFile.setRadius(radius);

        polygonFile.setUserDataTable(_userDataTable);

        return polygonFile;
    }

    //--------------------------------------------------------------------
    // Handle user data
    //--------------------------------------------------------------------
    private float[] processUserData(UserData userData)
    {
        int num = userData.numData();
        int[] indices = new int[num];

        for(int i=0; i<num; i++)
        {
            String name = userData.getName(i);
            Integer index = (Integer)_userDataTable.get(name);
            if(index == null)
            {
                index = new Integer( _userDataTable.size() );
                _userDataTable.put(name, index);
            }
            indices[i] = index.intValue();
        }

        if (_userDataTable.isEmpty()) {
            return null;
        }

        float[] data = new float[ _userDataTable.size() ];
        for(int i=0; i<data.length; i++)
        {
            data[i] = Float.NEGATIVE_INFINITY;
        }
        
        for(int i=0; i<num; i++)
        {
            data[indices[i]] = userData.getValue(i);
        }

        return data;
    }

    private void fixUserData()
    {
        int num = _userDataTable.size();
        if (num == 0) {
            return;
        }

        for(int i=0; i<_vertices.size(); i++)
        {
            Vertex vertex = (Vertex)_vertices.get(i);
            float[] userData = vertex.getUserData();
            if (userData == null)
            {
                float[] fixedData = new float[num];
                for(int j=0; j<num; j++)
                {
                    fixedData[j] = Float.NEGATIVE_INFINITY;
                }
                vertex.setUserData(fixedData);
            }
            else if (userData.length < num)
            {
                float[] fixedData = new float[num];
                for(int j=0; j<num; j++)
                {
                    fixedData[j] = (j < userData.length) ? userData[j] : Float.NEGATIVE_INFINITY;
                }
                vertex.setUserData(fixedData);
            }
        }
    }

    //--------------------------------------------------------------------
    // Find vertex
    //--------------------------------------------------------------------
    private Integer findVertex(int vertex)
    {
        if(_useId)
        {
            return (Integer)_idTable.get( new Integer(vertex) );
        }
        else
        {
            return _vertexFinder.find(vertex - 1);
        }
    }

    //--------------------------------------------------------------------
    // Determine center & radius
    //--------------------------------------------------------------------
    private float determineSize(boolean[] used, Vector3f center)
    {
        float minX, minY, minZ;
        float maxX, maxY, maxZ;

        minX = minY = minZ = Float.POSITIVE_INFINITY;
        maxX = maxY = maxZ = Float.NEGATIVE_INFINITY;

        boolean valid = false;
        int cnt = 0;
        for(int i=0; i<_vertices.size(); i++)
        {
            Vertex v = (Vertex)_vertices.get(i);
            if(used[cnt++]){
                float x = v.getX();
                float y = v.getY();
                float z = v.getZ();
                if(x<minX) minX = x;
                if(y<minY) minY = y;
                if(z<minZ) minZ = z;
                if(x>maxX) maxX = x;
                if(y>maxY) maxY = y;
                if(z>maxZ) maxZ = z;

                valid = true;
            }
        }

        if (valid) {
            center.set( (maxX + minX)/2,  (maxY + minY)/2, (maxZ + minZ)/2 );

            float radiusSq = 0;
            cnt = 0;
            Vector3f tmp = new Vector3f();
            for(int i=0; i<_vertices.size(); i++)
            {
                Vertex v = (Vertex)_vertices.get(i);
                if(used[cnt++]){
                    tmp.set(v.getX() - center.x, v.getY() - center.y, v.getZ() - center.z);
                    float curr = tmp.lengthSquared();
                    if(curr > radiusSq) radiusSq = curr;
                }
            }

            return (float)Math.sqrt(radiusSq);
        } else {
            return -1;
        }
    }
}
