/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.polygons;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;

/**
 * Class that reads polygon files.
 */
public class PolygonFileReader extends DefaultHandler
{
    private PolygonFileBuilder _builder;
    private MessageListener _listener;

    private ArrayList _currNodes;

    private boolean _inVertices;
    private boolean _inPointArray;
    private boolean _inLineArray;
    private boolean _inTriangleArray;
    private boolean _inQuadArray;
    private boolean _inPolylineArray;

    private VertexData _vertexData;
    private UserData _userData;

    private boolean _addVertex;
    private boolean _useId;

    /**
     * Constructor.
     */
    public PolygonFileReader(MessageListener listener)
    {
        _listener = listener;

        _currNodes = new ArrayList();
        _vertexData = new VertexData();
        _userData = new UserData();
        
        _useId = true;
    }

    /**
     * Reads the specified Polygon file and returns the PolygonFile object.
     */
    public PolygonFile read(File file) throws RecoverableException
    {
        _builder = new PolygonFileBuilder();
        _builder.setFilePath( file.getAbsolutePath() );

        this.readFile( new ResourceRequest(file) );

        return _builder.getResult();
    }

    /**
     * Reads the specified Polygon file and returns the PolygonFile object.
     */
    public PolygonFile read(URL url) throws RecoverableException
    {
        _builder = new PolygonFileBuilder();
        _builder.setFilePath( url.toString() );
        
        // message
        if(_listener != null){
            _listener.begin(1);
            _listener.appendMessage("Retrieving a polygon file");
        }

        try{
            this.readFile( new ResourceRequest(url) );
        }
        finally{
            if(_listener != null) _listener.end();
        }

        return _builder.getResult();
    }

    //--------------------------------------------------------------------
    // Parse
    //--------------------------------------------------------------------
    private void readFile(ResourceRequest request) throws RecoverableException
    {
        try{
            this.parse(request);
        }
        catch(FileNotFoundException e){
            throw new RecoverableException(ErrorMessages.CANT_FIND_POLYGON_FILE);
        }
        catch(UnknownHostException e){
            throw ExceptionUtil.handleUnknownHostException(e);
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
        catch(java.security.AccessControlException e)
        {
            if( ExceptionUtil.isPropertyPermission(e) ){
                throw ExceptionUtil.handlePropertyPermission();
            }
            else{
                throw new RecoverableException(ErrorMessages.CANT_FIND_POLYGON_FILE);
            }
        }
    }

    private void parse(ResourceRequest request) throws IOException, RecoverableException
    {
        InputStream stream = null;
        try{
            stream = request.openStream();

            SAXParserFactory spFactory = SAXParserFactory.newInstance();
            spFactory.setNamespaceAware(true);

            SAXParser parser = spFactory.newSAXParser();
            parser.parse(stream, this);
        }
        catch(SAXException e){
            throw ExceptionUtil.handleSAXException(e);
        }
        catch(ParserConfigurationException e){
            throw ExceptionUtil.handleParserConfigurationException(e);
        }
        finally{
            if(stream != null) stream.close();
            request.dispose();
        }
    }

    //-------------------------------------------------------------------------
    // SAX Handler
    //-------------------------------------------------------------------------
    /**
     * Start document.
     */
    public void startDocument() throws SAXException
    {
        _currNodes.clear();
        
        _inVertices = false;
        _inPointArray = false;
        _inLineArray = false;
        _inTriangleArray = false;
        _inQuadArray = false;
        _inPolylineArray = false;
    }

    /**
     * Start element.
     */
    public void startElement(
        String uri, String localName, String qName, Attributes attributes)
        throws SAXException
    {
        _currNodes.add(localName);

        if(_currNodes.size() == 4)
        {
            if( localName.equals("user_data") )
            {
                //if(_addVertex) this.processUserData(attributes);
                this.processUserData(attributes);
            }
        }
        else if(_currNodes.size() == 3)
        {
            if( _inVertices && localName.equals("vertex") )
            {
                String valueId = attributes.getValue("id");
                String valueImage = attributes.getValue("image");
                this.processVertex(valueId, valueImage);
            }
            else if( _inPointArray && localName.equals("point") ){
                String valueVertex = attributes.getValue("vertex");
                this.processPoint(valueVertex);
            }
            else if( _inLineArray && localName.equals("line") ){
                String valueVertex = attributes.getValue("vertex");
                this.processLine(valueVertex);
            }        
            else if( _inTriangleArray && localName.equals("triangle") ){
                String valueVertex = attributes.getValue("vertex");
                this.processTriangle(valueVertex);
            }        
            else if( _inQuadArray && localName.equals("quad") ){
                String valueVertex = attributes.getValue("vertex");
                this.processQuad(valueVertex);
            }
            else if( _inPolylineArray && localName.equals("polyline") ){
                String valueVertex = attributes.getValue("vertex");
                this.processPolyline(valueVertex);
            }
        }
        else if(_currNodes.size() == 2)
        {
            if( localName.equals("vertices") )
            {
                _inVertices = true;
                String use = attributes.getValue("id_numbers");
                if(use != null && use.equals("ignore"))
                {
                    _useId = false;
                    _builder.useId(false);
                }
            }
            else if( localName.equals("point_array") ){
                _inPointArray = true;
                String trans = attributes.getValue("transparency");
                _builder.setPointTransparency( handleTransparency(trans) );
                String size = attributes.getValue("size");
                _builder.setPointSize( handleSize(size) );
            }
            else if( localName.equals("line_array") ){
                _inLineArray = true;
                String trans = attributes.getValue("transparency");
                _builder.setLineTransparency( handleTransparency(trans) );
                String width = attributes.getValue("width");
                _builder.setLineWidth( handleSize(width) );
            }
            else if( localName.equals("triangle_array") ){
                _inTriangleArray = true;
                String trans = attributes.getValue("transparency");
                _builder.setTriangleTransparency( handleTransparency(trans) );
            }
            else if( localName.equals("quad_array") ){
                _inQuadArray = true;
                String trans = attributes.getValue("transparency");
                _builder.setQuadTransparency( handleTransparency(trans) );
            }
            else if( localName.equals("polyline_array") ){
                _inPolylineArray = true;
                String trans = attributes.getValue("transparency");
                _builder.setPolylineTransparency( handleTransparency(trans) );
                String width = attributes.getValue("width");
                _builder.setPolylineWidth( handleSize(width) );
            }
        }
    }

    /**
     * End element.
     */
    public void endElement(String uri, String localName, String qName)
        throws SAXException
    {
        if(_currNodes.size() == 3)
        {
            if( _inVertices && localName.equals("vertex") )
            {
                if(_addVertex)
                {
                    _builder.addVertex(_vertexData, _userData);
                }
                else if(!_useId)
                {
                    _builder.addNullVertex(_userData);
                }
            }
        }
        if(_currNodes.size() == 2){
            if( localName.equals("vertices") ){
                _inVertices = false;
            }
            else if( localName.equals("point_array") ){
                _inPointArray = false;
            }
            else if( localName.equals("line_array") ){
                _inLineArray = false;
            }
            else if( localName.equals("triangle_array") ){
                _inTriangleArray = false;
            }
            else if( localName.equals("quad_array") ){
                _inQuadArray = false;
            }
            else if( localName.equals("polyline_array") ){
                _inPolylineArray = false;
            }
        }

        _currNodes.remove( _currNodes.size() - 1 );
    }

    //--------------------------------------------------------------------
    // Process element
    //--------------------------------------------------------------------
    private void processVertex(String valueId, String valueImage)
    {
        _addVertex = false;
        _userData.clear();
        
        if(valueId == null || valueImage == null)
        {
            return;
        }

        StringTokenizer st = new StringTokenizer(valueImage);
        if(st.countTokens() != 9)
        {
            return;
        }

        try
        {
            _vertexData.id = Integer.parseInt(valueId);

            // coordinates
            _vertexData.x = Float.parseFloat( st.nextToken() );
            _vertexData.y = Float.parseFloat( st.nextToken() );
            _vertexData.z = Float.parseFloat( st.nextToken() );

            // normal vector components
            _vertexData.nx = Float.parseFloat( st.nextToken() );
            _vertexData.ny = Float.parseFloat( st.nextToken() );
            _vertexData.nz = Float.parseFloat( st.nextToken() );
            if(_vertexData.nx * _vertexData.nx + _vertexData.ny * _vertexData.ny + _vertexData.nz * _vertexData.nz == 0)
            {
                return;
            }

            // colors
            int r = (int)( Float.parseFloat( st.nextToken() ) );
            int g = (int)( Float.parseFloat( st.nextToken() ) );
            int b = (int)( Float.parseFloat( st.nextToken() ) );
            if(r < 0 || r > 255 || g < 0 || g > 255 || b < 0 || b > 255)
            {
                return;
            }
            _vertexData.color = getColorValue(r, g, b, 255);

            _addVertex = true;
        }
        catch(NumberFormatException e)
        {
            return;
        }
    }

    private final int getColorValue(int r, int g, int b, int a) {
        return ((a & 0xFF) << 24) | ((r & 0xFF) << 16) | ((g & 0xFF) << 8) | ((b & 0xFF) << 0);
    }

    private void processUserData(Attributes attributes)
    {
        int num = attributes.getLength();
        for(int i=0; i<num; i++){
            try{
                String name = attributes.getLocalName(i);
                float value = Float.parseFloat( attributes.getValue(i) );
                _userData.add(name, value);
            }
            catch(NumberFormatException e){
                continue;
            }
        }
    }

    private void processPoint(String valueVertex)
    {
        if(valueVertex != null){
            int[] indices = handleVertexAttribute(valueVertex, 1);
            if(indices != null){    
                _builder.addPoint(indices[0]);
            }
        }
    }

    private void processLine(String valueVertex)
    {
        if(valueVertex != null){
            int[] indices = handleVertexAttribute(valueVertex, 2);
            if(indices != null){    
                _builder.addLine(indices[0], indices[1]);
            }
        }
    }

    private void processTriangle(String valueVertex)
    {
        if(valueVertex != null){
            int[] indices = handleVertexAttribute(valueVertex, 3);
            if(indices != null){    
                _builder.addTriangle(indices[0], indices[1], indices[2]);
            }
        }
    }

    private void processQuad(String valueVertex)
    {
        if(valueVertex != null){
            int[] indices = handleVertexAttribute(valueVertex, 4);
            if(indices != null){    
                _builder.addQuad(indices[0], indices[1], indices[2], indices[3]);
            }
        }
    }

    private void processPolyline(String valueVertex)
    {
        if(valueVertex != null){
            int[] indices = handleVertexAttribute(valueVertex, 0);
            if(indices != null && indices.length > 1){
                _builder.addPolyline(indices);
            }
        }
    }

    //--------------------------------------------------------------------
    // Handle attribute
    //--------------------------------------------------------------------
    private int[] handleVertexAttribute(String valueVertex, int cnt)
    {
        StringTokenizer st = new StringTokenizer(valueVertex);
        if(cnt == 0){
            cnt = st.countTokens();
        }
        else if(st.countTokens() != cnt) return null;

        int[] indices = new int[cnt];

        try{
            for(int i=0; i<cnt; i++){
                indices[i] = Integer.parseInt( st.nextToken() );
            }
        }
        catch(NumberFormatException e){
            return null;
        }

        return indices;
    }
    
    private float handleTransparency(String transparency)
    {
        float tp = 0;
        if(transparency != null){
            try{
                float tmp = Float.parseFloat(transparency);
                if(tmp >= 0 && tmp <= 1){
                    tp = tmp;
                }
            }
            catch(NumberFormatException e){
            }
        }
        
        return tp;
    }

    private float handleSize(String size)
    {
        float s = 1;
        if(size != null){
            try{
                float tmp = Float.parseFloat(size);
                if(tmp > 0){
                    s = tmp;
                }
            }
            catch(NumberFormatException e){
            }
        }
        
        return s;
    }
}
