/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.renderer;

import java.awt.*;
import java.nio.*;
import java.util.ArrayList;
import javax.media.opengl.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.images.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;

/**
 * Class that renders image on a canvas.
 */
class CanvasRenderer implements GLEventListener
{
    private GLAutoDrawable _drawable;
    private RendererListener _listener;
    
    private ImageRendererImpl _renderer;

    private ViewSettings _viewSettings;
    private ColorManager _colorManager;
    private ImageBuilder _imageBuilder;

    private ArrayList _moleculeFiles;
    private ArrayList _polygonFiles;

    private boolean _changeLight;
    private boolean _changeCamera;
    private boolean _changePerspective;
    private boolean _changeBackground;
    private boolean _toImage;

    // front clip of spheres
    private float _currClip;

    private int[] _pixelData;
    
    private Vector3f _stereoCenter;
    private float    _stereoAngle;
    private Matrix4f _stereoMatrix;
    private Matrix4f _stereoMatrixL;
    private Matrix4f _stereoMatrixR;
    private Matrix4f _stereoTransform;

    // matrix
    private Vector3f _vector1 = new Vector3f();
    private Vector3f _vector2 = new Vector3f();
    private Matrix4f _matrix1 = new Matrix4f();
    private Matrix4f _matrix2 = new Matrix4f();
    private float[] _matrixArray = new float[16];
    
    private float[] _colorEmpty = {0, 0, 0, 1};
    private boolean _drawBlue;
    private boolean _drawGreen;
    
    /**
    * Sets a renderer listener.
    */
    public void setRendererListener(RendererListener listener)
    {
        _listener = listener;
    }

    /**
    * Sets the stereo center.
    */
    void setStereoCenter(Vector3f center)
    {
        _stereoCenter.set(center);
        this.setupStereo();
    }

    /**
    * Sets the stereo angle.
    */
    void setStereoAngle(float angle)
    {
        _stereoAngle = angle;
        this.setupStereo();
    }

    /**
    * Sets up anaglyph.
    */
    void setupAnaglyph(int anaglyphMode) {
        if (anaglyphMode == ViewSettings.ANAGLYPH_REDBLUE) {
            _drawBlue = true;
            _drawGreen = false;
        } else if (anaglyphMode == ViewSettings.ANAGLYPH_REDCYAN) {
            _drawBlue = true;
            _drawGreen = true;
        } else if (anaglyphMode == ViewSettings.ANAGLYPH_REDGREEN) {
            _drawBlue = false;
            _drawGreen = true;
        }
    }

    /**
     * Constructor.
     */
    CanvasRenderer(ImageRendererImpl renderer)
    {
        _renderer = renderer;

        _viewSettings = _renderer.getViewSettings();
        _colorManager = _renderer.getColorManager();
        _imageBuilder = _renderer.getImageBuilder();
        
        _moleculeFiles = _renderer.getMoleculeFiles();
        _polygonFiles  = _renderer.getPolygonFiles();
        
        _stereoCenter    = new Vector3f();
        _stereoMatrixL   = new Matrix4f();
        _stereoMatrixR   = new Matrix4f();
        _stereoTransform = new Matrix4f();
    }

    /**
     * Sets the drawable.
     */
    void setDrawable(GLAutoDrawable drawable)
    {
        _drawable = drawable;
        _drawable.addGLEventListener(this);
    }

    /**
     * Returns the drawable.
     */
    GLAutoDrawable getDrawable()
    {
        return _drawable;
    }

    /**
     * Notifies that light changed.
     */
    void setLightChanged()
    {
        _changeLight = true;
    }

    /**
     * Notifies that camera settings changed.
     */
    void setCameraChanged()
    {
        _changeCamera = true;
    }

    /**
     * Notifies that perspective changed.
     */
    void setPerspectiveChanged()
    {
        _changePerspective = true;
    }

    /**
     * Notifies that background changed.
     */
    void setBackgroundChanged()
    {
        _changeBackground = true;
    }

    /**
     * Returns the pixel data on the canvas.
     */
    int[] getPixelData()
    {
        _toImage = true;
        _drawable.display();
        _toImage = false;
        
        int[] pixelData = _pixelData;
        _pixelData = null;
        
        return pixelData;
    }

    //--------------------------------------------------------------------
    // Implements GLEventListener
    //--------------------------------------------------------------------
    /**
    * Called by the drawable immediately after the OpenGL2 context is 
    * initialized for the first time.
    */
    public void init(GLAutoDrawable drawable)
    {
        GL2 gl = drawable.getGL().getGL2();
        
        gl.glShadeModel(GL2.GL_SMOOTH);
        gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        gl.glClearDepth(1.0f);
        gl.glEnable(GL2.GL_DEPTH_TEST);
        gl.glDepthFunc(GL2.GL_LEQUAL);
        gl.glEnable(GL2.GL_LIGHTING);
        gl.glEnable(GL2.GL_NORMALIZE);
        gl.glEnable(GL2.GL_LINE_STIPPLE);
        gl.glCullFace(GL2.GL_BACK);
        gl.glEnable(GL2.GL_CULL_FACE);
        gl.glBlendFunc(GL2.GL_SRC_ALPHA, GL2.GL_ONE_MINUS_SRC_ALPHA);
        gl.glEnable(GL2.GL_POINT_SMOOTH);

        this.setupLight(gl);
        
        if(_listener != null) _listener.screenInitialized();
    }
    
    /**
    * Called by the drawable to initiate OpenGL2 rendering by the client.
    */
    public void display(GLAutoDrawable drawable)
    {
        GL2 gl = drawable.getGL().getGL2();

        if(_renderer.dispose)
        {
            if(_renderer.deleteList) _renderer.deleteList(gl);
            
            _renderer.deletePrimitives(gl);
        }
        else
        {
            if(_renderer.deleteList)      _renderer.deleteList(gl);
            if(_renderer.changeDrawlevel) _renderer.createPrimitives(gl);

            if(_changeLight)       this.setupLight(gl);
            if(_changeCamera)      this.setupCamera(gl);
            if(_changePerspective) this.setupPerspective(gl);
            if(_changeBackground)  this.setupBGColor(gl);

            // clear screen
            gl.glClear(GL2.GL_COLOR_BUFFER_BIT | GL2.GL_DEPTH_BUFFER_BIT);
    
            if (_viewSettings.getAnaglyphMode() != ViewSettings.ANAGLYPH_OFF) {
                _stereoMatrix = _stereoMatrixL;
                gl.glColorMask(true, false, false, false);
                this.renderObjects(gl);
                gl.glClear(GL2.GL_DEPTH_BUFFER_BIT);
                _stereoMatrix = _stereoMatrixR;
                gl.glColorMask(false, _drawGreen, _drawBlue, false);
                this.renderObjects(gl);
                gl.glColorMask(true, true, true, true);
            } else {
                this.renderObjects(gl);
            }

            gl.glFlush();
        }

        if(_toImage)
        {
            this.createImage(gl);
        }

        if(_listener != null) _listener.screenDrawed();
    }

    /**    
    * Called by the drawable during the first repaint after the component 
    * has been resized.
    */
    public void reshape(
        GLAutoDrawable drawable, int x, int y, int width, int height)
    {
        _viewSettings.setAspectRatio( width/(float)height );
        this.setPerspectiveChanged();

        if(_listener != null) _listener.screenResized(x, y, width, height);
    }
    
    /**
    * Notifies the listener to perform the release of all OpenGL resources per
    * GLContext, such as memory buffers and GLSL programs.
    */
    public void dispose(GLAutoDrawable drawable)
    {
    }

    //----------------------------------------------------------------------
    // Set up view
    //----------------------------------------------------------------------
    private void setupLight(GL2 gl)
    {
        float a = _viewSettings.getAmbient();
        float[] ambient = {a, a, a, 1};
        gl.glLightModelfv(GL2.GL_LIGHT_MODEL_AMBIENT, ambient, 0);

        float d1 = _viewSettings.getDiffuse1();
        if (d1 != 0) {
            float[] diffuse1 = {d1, d1, d1, 1};
            float[] dir1 = _viewSettings.getDiffuseDir1();
            float[] position1 = {-dir1[0], -dir1[1], -dir1[2], 0};
            gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_POSITION, position1, 0);
            gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_DIFFUSE, diffuse1, 0);
            gl.glEnable(GL2.GL_LIGHT0);
        } else {
            gl.glDisable(GL2.GL_LIGHT0);
        }

        float d2 = _viewSettings.getDiffuse2();
        if (d2 != 0) {
            float[] diffuse2 = {d2, d2, d2, 1};
            float[] dir2 = _viewSettings.getDiffuseDir2();
            float[] position2 = {-dir2[0], -dir2[1], -dir2[2], 0};
            gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_POSITION, position2, 0);
            gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_DIFFUSE, diffuse2, 0);
            gl.glEnable(GL2.GL_LIGHT1);
        } else {
            gl.glDisable(GL2.GL_LIGHT1);
        }

        _changeLight = false;
    }

    private void setupCamera(GL2 gl)
    {
        gl.glMatrixMode(GL2.GL_MODELVIEW);
        gl.glLoadIdentity();

        Vector3f cp = _viewSettings.getCameraPosition();
        //Vector3f ct = _viewSettings.getCameraTarget();
        //Vector3f cu = _viewSettings.getCameraUpVector();
        
        // assuming that camera direction is (0, 0, -1) and up vector is (0, 1, 0).
        gl.glTranslatef(-cp.x, -cp.y, -cp.z);

        _changeCamera = false;
    }

    private void setupPerspective(GL2 gl)
    {
        gl.glMatrixMode(GL2.GL_PROJECTION);
        gl.glLoadIdentity();

        float near = _viewSettings.getNearClipping();
        float far = _viewSettings.getFarClipping();

        if( !_viewSettings.isParallelProjection() )
        {
            float fovy = _viewSettings.getFieldAngle(); // in radians
            float ratio = _viewSettings.getAspectRatio();
            
            float top    = near*(float)Math.tan(fovy/2);
            float bottom = -top;
            float left   = ratio*bottom;
            float right  = ratio*top;
            
            gl.glFrustum(left, right, bottom, top, near, far);
        }
        else
        {
            float w = _drawable.getWidth();
            float h = _drawable.getHeight();
        
            float height = _viewSettings.getViewSize();
            float width = height*w/(float)h;

            float left   = -width;
            float right  =  width;
            float top    =  height;
            float bottom = -height;

            gl.glOrtho(left, right, bottom, top, near, far);
        }

        _changePerspective = false;
    }

    private void setupBGColor(GL2 gl)
    {
        Color bg = _viewSettings.getBGColor();
        gl.glClearColor(
            bg.getRed()/255.0f, 
            bg.getGreen()/255.0f, 
            bg.getBlue()/255.0f, 
            bg.getAlpha()/255.0f
        );

        _changeBackground = false;
    }

    private void setupStereo()
    {
        _vector1.set(_stereoCenter);
        _matrix1.translation(_vector1);
        _vector1.multiply(-1);
        _matrix2.translation(_vector1);

        _stereoMatrixL.rotationY(_stereoAngle);
        _stereoMatrixL.multiply(_stereoMatrixL, _matrix2);
        _stereoMatrixL.multiply(_matrix1, _stereoMatrixL);
        _stereoMatrixR.rotationY(-_stereoAngle);
        _stereoMatrixR.multiply(_stereoMatrixR, _matrix2);
        _stereoMatrixR.multiply(_matrix1, _stereoMatrixR);
    }

    //----------------------------------------------------------------------
    // Create image
    //----------------------------------------------------------------------
    private void createImage(GL2 gl)
    {
        int w = _drawable.getWidth();
        int h = _drawable.getHeight();

        byte[] array = new byte[4*w*h];
        ByteBuffer byteBuffer = ByteBuffer.wrap(array);
        gl.glReadPixels(0, 0, w, h, GL2.GL_RGBA, GL2.GL_UNSIGNED_BYTE, byteBuffer);

        _pixelData = new int[w*h];
        
        int index1 = 0;
        for(int y=h-1; y>=0; y--)
        {
            int index2 = 4*y*w;
            for(int x=0; x<w; x++)
            {
                int r = ( array[index2++] & 255 );
                int g = ( array[index2++] & 255 );
                int b = ( array[index2++] & 255 );
                int a = ( array[index2++] & 255 );

                _pixelData[index1++] = ( (a<<24) | (r<<16) | (g<<8) | b);
            }
        }
    }

    //----------------------------------------------------------------------
    // Render objects
    //----------------------------------------------------------------------
    private void renderObjects(GL2 gl)
    {
        gl.glMatrixMode(GL2.GL_MODELVIEW);

        _currClip
            = _viewSettings.getCameraPosition().z - _viewSettings.getNearClipping();

        if( _viewSettings.getSpecular() ){
            gl.glMaterialfv(
                GL2.GL_FRONT, GL2.GL_SPECULAR, _viewSettings.getSpecularColor(), 0
            );
            gl.glMaterialf(
                GL2.GL_FRONT, GL2.GL_SHININESS, _viewSettings.getShininess()
            );
        }
        else{
            gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_SPECULAR, _colorEmpty, 0);
        }

        for(int i=0; i<_moleculeFiles.size(); i++)
        {
            MoleculeFile moleculeFile = (MoleculeFile)_moleculeFiles.get(i);
            if( moleculeFile.isDisplayed() )
            {
                if( moleculeFile.isAnimation() ){
                    this.renderMolecule(
                        moleculeFile.getMolecule( moleculeFile.getCurrentFrame() ), gl
                    );
                }
                else{
                    for(int m=0; m<moleculeFile.numMolecules(); m++)
                    {
                        this.renderMolecule( moleculeFile.getMolecule(m), gl);
                    }
                }
            }
        }

        gl.glEnable(GL2.GL_LINE_SMOOTH);
        gl.glEnable(GL2.GL_BLEND);
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile polygonFile = (PolygonFile)_polygonFiles.get(i);
            if( polygonFile.isDisplayed() ){
                this.renderPolygon(polygonFile, false, gl);
            }
        }

        gl.glDepthMask( false );
        for(int i=0; i<_polygonFiles.size(); i++)
        {
            PolygonFile polygonFile = (PolygonFile)_polygonFiles.get(i);
            if( polygonFile.isDisplayed() ){
                this.renderPolygon(polygonFile, true, gl);
            }
        }
        gl.glDepthMask( true );
        gl.glDisable(GL2.GL_BLEND);
        gl.glDisable(GL2.GL_LINE_SMOOTH);
    }

    //----------------------------------------------------------------------
    // Render molecule
    //----------------------------------------------------------------------
    private void renderMolecule(Molecule molecule, GL2 gl)
    {
        MoleculeImage image = molecule.getImage();
        
        if( molecule.builtBond() && !image.bondInitialized() ){
            _imageBuilder.setupBond(molecule);
        }
        if( molecule.builtBackbone() && !image.backboneInitialized() ){
            _imageBuilder.setupBackbone(molecule);
        }
        if( molecule.builtHBond() && !image.hbondInitialized() ){
            _imageBuilder.setupHBond(molecule);
        }
        if( molecule.builtSSBond() && !image.ssbondInitialized() ){
            _imageBuilder.setupSSBond(molecule);
        }
        if( image.refreshResidue() ){
            _imageBuilder.buildResidueImage(molecule, gl);
        }

        Matrix4f transform = image.getTransform();
        if (_viewSettings.isStereo()) {
            _stereoTransform.multiply(_stereoMatrixL, transform);
            transform = _stereoTransform;
        } else if (_viewSettings.getAnaglyphMode() != ViewSettings.ANAGLYPH_OFF) {
            _stereoTransform.multiply(_stereoMatrix, transform);
            transform = _stereoTransform;
        }

        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            for(int re=0; re<chain.numResidues(); re++)
            {
            Residue residue = chain.getResidue(re);
            for(int at=0; at<residue.numAtoms(); at++)
            {
                Atom atom = residue.getAtom(at);
                this.renderAtom(atom, transform, false, null, 0, gl);
            }}

            for(int b=0; b<chain.numBackbones(); b++)
            {
                Bond backbone = chain.getBackbone(b);
                this.renderBond(backbone, transform, gl);
            }
        }

        for(int b=0; b<molecule.numBonds(); b++)
        {
            Bond bond = molecule.getBond(b);
            this.renderBond(bond, transform, gl);
        }

        boolean backbone = image.hbondBackbone();
        for(int b=0; b<molecule.numHBonds(); b++)
        {
            HBond hbond = molecule.getHBond(b);
            this.renderHBond(hbond, transform, backbone, gl);
        }

        backbone = image.ssbondBackbone();
        for(int b=0; b<molecule.numSSBonds(); b++)
        {
            SSBond ssbond = molecule.getSSBond(b);
            this.renderSSBond(ssbond, transform, backbone, gl);
        }

        this.renderResidue(molecule, transform, gl);
    }

    //----------------------------------------------------------------------
    // Translate 4x4 matrix into array with size 16
    //----------------------------------------------------------------------
    private void processMatrix(Matrix4f matrix)
    {
        _matrixArray[ 0] = matrix.m11;
        _matrixArray[ 1] = matrix.m21;
        _matrixArray[ 2] = matrix.m31;
        _matrixArray[ 3] = matrix.m41;
        _matrixArray[ 4] = matrix.m12;
        _matrixArray[ 5] = matrix.m22;
        _matrixArray[ 6] = matrix.m32;
        _matrixArray[ 7] = matrix.m42;
        _matrixArray[ 8] = matrix.m13;
        _matrixArray[ 9] = matrix.m23;
        _matrixArray[10] = matrix.m33;
        _matrixArray[11] = matrix.m43;
        _matrixArray[12] = matrix.m14;
        _matrixArray[13] = matrix.m24;
        _matrixArray[14] = matrix.m34;
        _matrixArray[15] = matrix.m44;
    }

    //----------------------------------------------------------------------
    // Render atoms
    //----------------------------------------------------------------------
    private void renderAtom(
        Atom atom, Matrix4f transform, boolean cap, Color capColor, float capRadius,
        GL2 gl)
    {
        if(cap || (atom.getImage().isDisplayed() && atom.getImage().getModel() == AtomImage.MODEL_SPACEFILL))
        {
            gl.glPushMatrix();

            float r = cap ? capRadius : atom.getImage().getRadius();

            _matrix1.translation( atom.getX(), atom.getY(), atom.getZ() );
            _matrix2.scale(r, r, r);
            _matrix1.multiply(_matrix1, _matrix2);

            // world transform
            _matrix1.multiply(transform, _matrix1);

            Color color = null;
            if(cap){
                color = capColor;
            }
            else{
                color = _colorManager.getColor( atom.getImage().getColorIndex() );
            }

            float[] col = {
                color.getRed()/255.0f, 
                color.getGreen()/255.0f, 
                color.getBlue()/255.0f, 
                1
            };

            this.processMatrix(_matrix1);

            gl.glMultMatrixf(_matrixArray, 0);
            gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, col, 0);
            gl.glCallList(_renderer.sphere);
            gl.glPopMatrix();

            float z = _matrix1.m34;
            if(_currClip > z - r && _currClip < z + r){
                float x = _matrix1.m14;
                float y = _matrix1.m24;
                this.renderClipCap(x, y, z, r, gl);
            }
            
            atom.getImage().setCapRadius(r);
        }
        else atom.getImage().setCapRadius(0);
    }

    private void renderClipCap(float x, float y, float z, float radius, GL2 gl)
    {
        if( _viewSettings.getSpecular() ){
            gl.glMaterialfv(
                GL2.GL_FRONT, GL2.GL_SPECULAR, _viewSettings.getClipSpecularColor(), 0
            );
        }

        gl.glPushMatrix();

        float dz = _currClip - z;
        float r = (float)Math.sqrt(radius*radius - dz*dz);

        _matrix1.translation(x, y, _currClip - 0.0001f);
        _matrix2.scale(r, r, 1);
        _matrix1.multiply(_matrix1, _matrix2);

        this.processMatrix(_matrix1);

        gl.glMultMatrixf(_matrixArray, 0);
        gl.glCallList(_renderer.clipCap);

        gl.glPopMatrix();

        if( _viewSettings.getSpecular() ){
            gl.glMaterialfv(
                GL2.GL_FRONT, GL2.GL_SPECULAR, _viewSettings.getSpecularColor(), 0
            );
        }
    }

    //----------------------------------------------------------------------
    // Render residues
    //----------------------------------------------------------------------
    private void renderResidue(Molecule molecule, Matrix4f transform, GL2 gl)
    {
        gl.glPushMatrix();

        this.processMatrix(transform);

        gl.glMultMatrixf(_matrixArray, 0);

        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);

            if (!residue.getImage().isDisplayed()) {
                continue;
            }

            int model = residue.getImage().getModel();

            if( model == ResidueImage.MODEL_RIBBON ||
                model == ResidueImage.MODEL_CARTOON ||
                model == ResidueImage.MODEL_CARTOON_TUBE ||
                model == ResidueImage.MODEL_TRACE )
            {
                int colorIndex = -1;
                Color color = null;
                float[] col = null;

                int listIndex = residue.getImage().getListIndex();
                if(listIndex != 0){
                    colorIndex = residue.getImage().getColorIndex();
                    if(colorIndex == -1){
                        Atom atomCA = residue.isProtein() ?
                            residue.getProteinCA() : residue.getNucleotideP();
                        if(atomCA != null){
                            colorIndex = atomCA.getImage().getColorIndex();
                        }
                    }
                    if(colorIndex == -1) continue;

                    color = _colorManager.getColor(colorIndex);
                    col = new float[]{
                        color.getRed()/255.0f, 
                        color.getGreen()/255.0f, 
                        color.getBlue()/255.0f, 
                        1
                    };
                    gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, col, 0);
                    gl.glCallList(listIndex);
                }
                
                if( model == ResidueImage.MODEL_RIBBON ||
                    model == ResidueImage.MODEL_CARTOON )
                {
                    int backListIndex = residue.getImage().getBackListIndex();
                    if(backListIndex != 0){
                        int backColorIndex = residue.getImage().getBackColorIndex();
                        if(backColorIndex == -1){
                            backColorIndex = molecule.getImage().getResidueBackColorIndex();
                            if (backColorIndex == -1) {
                                backColorIndex = colorIndex;
                            }
                        }
                        if(backColorIndex != colorIndex){
                            color = _colorManager.getColor(backColorIndex);
                            col = new float[]{
                                color.getRed()/255.0f, 
                                color.getGreen()/255.0f, 
                                color.getBlue()/255.0f, 
                                1
                            };
                        }
                        gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, col, 0);
                        gl.glCallList(backListIndex);
                    }
                }
            }
        }}

        gl.glPopMatrix();
    }

    //----------------------------------------------------------------------
    // Render bonds
    //----------------------------------------------------------------------
    private void renderBond(Bond bond, Matrix4f transform, GL2 gl)
    {
        BondImage image = bond.getImage();
        if(image.isDisplayed() && image.getModel() == BondImage.MODEL_WIREFRAME)
        {
            if(image.getRadius() == 0){
                this.renderLine(
                    bond.getSrcAtom(), bond.getDstAtom(), image, transform, false, gl);
            }
            else{
                this.renderStick(
                    bond.getSrcAtom(), bond.getDstAtom(), image, transform, gl);
            }
        }
    }

    private void renderLine(
        Atom src, Atom dst, BondImage image, Matrix4f transform, boolean broken,
        GL2 gl) 
    {
        float srcX, srcY, srcZ, dstX, dstY, dstZ;

        _matrix1.translation( src.getX(), src.getY(), src.getZ() );
        _matrix2.translation( dst.getX(), dst.getY(), dst.getZ() );
        _matrix1.multiply(transform, _matrix1);
        _matrix2.multiply(transform, _matrix2);

        srcX = _matrix1.m14; srcY = _matrix1.m24; srcZ = _matrix1.m34;
        dstX = _matrix2.m14; dstY = _matrix2.m24; dstZ = _matrix2.m34;

        float midX = (srcX + dstX)/2;
        float midY = (srcY + dstY)/2;
        float midZ = (srcZ + dstZ)/2;

        Color[] colors = this.getBondColor(src, dst, image);

        if(broken) gl.glLineStipple(1, (short)0x0F0F);
        else       gl.glLineStipple(1, (short)0xFFFF);

        gl.glDisable(GL2.GL_LIGHTING);

        // src
        gl.glColor3f(
            colors[0].getRed()/255.0f, 
            colors[0].getGreen()/255.0f, 
            colors[0].getBlue()/255.0f
        );
        gl.glBegin(GL2.GL_LINES);
        gl.glVertex3f(srcX, srcY, srcZ);
        gl.glVertex3f(midX, midY, midZ);
        gl.glEnd();

        // dst
        gl.glColor3f(
            colors[1].getRed()/255.0f, 
            colors[1].getGreen()/255.0f, 
            colors[1].getBlue()/255.0f
        );
        gl.glBegin(GL2.GL_LINES);
        gl.glVertex3f(dstX, dstY, dstZ);
        gl.glVertex3f(midX, midY, midZ);
        gl.glEnd();

        gl.glEnable(GL2.GL_LIGHTING);
    }

    private void renderStick(
        Atom src, Atom dst, BondImage image, Matrix4f transform, GL2 gl) 
    {
        int list = _renderer.getCylinder(image.getRadius(), gl);

        Color[] colors = this.getBondColor(src, dst, image);

        float[] srcColor = {
            colors[0].getRed()/255.0f, 
            colors[0].getGreen()/255.0f, 
            colors[0].getBlue()/255.0f,
            1
        };
        float[] dstColor = {
            colors[1].getRed()/255.0f, 
            colors[1].getGreen()/255.0f, 
            colors[1].getBlue()/255.0f,
            1
        };

        this.getBondMatrix(src, image, _matrix1, false);
        this.getBondMatrix(dst, image, _matrix2, true);

        gl.glDisable(GL2.GL_NORMALIZE);

        // src
        gl.glPushMatrix();
        _matrix1.multiply(transform, _matrix1);
        this.processMatrix(_matrix1);
        gl.glMultMatrixf(_matrixArray, 0);
        gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, srcColor, 0);
        //gl.glCallList( _renderer.cylinder );
        gl.glCallList(list);
        gl.glPopMatrix();

        // dst
        gl.glPushMatrix();
        _matrix2.multiply(transform, _matrix2);
        this.processMatrix(_matrix2);
        gl.glMultMatrixf(_matrixArray, 0);
        gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, dstColor, 0);
        //gl.glCallList( _renderer.cylinder );
        gl.glCallList(list);
        gl.glPopMatrix();

        gl.glEnable(GL2.GL_NORMALIZE);

        if( _viewSettings.getBondCap() ){
            float r = image.getRadius();
            if(src.getImage().getCapRadius() < r){
                this.renderAtom(src, transform, true, colors[0], r, gl);
            }
            if(dst.getImage().getCapRadius() < r){
                this.renderAtom(dst, transform, true, colors[1], r, gl);
            }
        }
    }

    private Color[] getBondColor(Atom src, Atom dst, BondImage image)
    {
        Color[] colors = new Color[2];
        if(image.getColorIndex() != -1){
            colors[0] = colors[1] = _colorManager.getColor(image.getColorIndex());
        }
        else{
            colors[0] = _colorManager.getColor(src.getImage().getColorIndex());
            colors[1] = _colorManager.getColor(dst.getImage().getColorIndex());
        }
        return colors;
    }

    private void getBondMatrix(
        Atom atom, BondImage image, Matrix4f matrix, boolean inverse)
    {
        int inv = inverse ? -1 : 1;

        //float radius = image.getRadius();
        float radius = 1;
        float height = image.getDistance()/2;

        image.getDirection(_vector1);
        image.getNormal(_vector2);
        Vector3f baseZ = _vector1;
        Vector3f baseX = _vector2;

        matrix.m11 = radius * baseX.x;
        matrix.m21 = radius * baseX.y;
        matrix.m31 = radius * baseX.z;
        matrix.m41 = 0;
        matrix.m12 = inv * radius * (baseZ.y*baseX.z - baseZ.z*baseX.y);
        matrix.m22 = inv * radius * (baseZ.z*baseX.x - baseZ.x*baseX.z);
        matrix.m32 = inv * radius * (baseZ.x*baseX.y - baseZ.y*baseX.x);
        matrix.m42 = 0;
        matrix.m13 = inv * height * baseZ.x;
        matrix.m23 = inv * height * baseZ.y;
        matrix.m33 = inv * height * baseZ.z;
        matrix.m43 = 0;
        matrix.m14 = atom.getX();
        matrix.m24 = atom.getY();
        matrix.m34 = atom.getZ();
        matrix.m44 = 1;
    }

    //----------------------------------------------------------------------
    // Render hbonds & ssbonds
    //----------------------------------------------------------------------
    private void renderHBond(
        HBond hbond, Matrix4f transform, boolean backbone, GL2 gl)
    {
        BondImage image = backbone ? hbond.getImageCA() : hbond.getImage();

        if(image.isDisplayed() && image.getModel() == BondImage.MODEL_WIREFRAME)
        {
            Atom src = backbone ? hbond.getSrcCA() : hbond.getSrcAtom();
            Atom dst = backbone ? hbond.getDstCA() : hbond.getDstAtom();
            if(src == null || dst == null) return;

            if(image.getRadius() == 0){
                renderLine(src, dst, image, transform, true, gl);
            }
            else{
                renderStick(src, dst, image, transform, gl);
            }
        }
    }

    private void renderSSBond(
        SSBond ssbond, Matrix4f transform, boolean backbone, GL2 gl)
    {
        BondImage image = backbone ? ssbond.getImageCA() : ssbond.getImage();

        if(image.isDisplayed() && image.getModel() == BondImage.MODEL_WIREFRAME)
        {
            Atom src = backbone ? ssbond.getSrcCA() : ssbond.getSrcAtom();
            Atom dst = backbone ? ssbond.getDstCA() : ssbond.getDstAtom();
            if(src == null || dst == null) return;
            
            if(image.getRadius() == 0){
                renderLine(src, dst, image, transform, true, gl);
            }
            else{
                renderStick(src, dst, image, transform, gl);
            }
        }
    }

    //----------------------------------------------------------------------
    // Render polygon
    //----------------------------------------------------------------------
    private void renderPolygon(PolygonFile polygonFile, boolean transparent, GL2 gl)
    {
        PolygonImage image = polygonFile.getImage();

        if( image.refresh() ){
            _imageBuilder.buildPolygonImage(polygonFile, gl);
        }

        gl.glPushMatrix();

        Matrix4f transform = image.getTransform();
        if( _viewSettings.isStereo() ){
            _stereoTransform.multiply(_stereoMatrixL, transform);
            transform = _stereoTransform;
        } else if (_viewSettings.getAnaglyphMode() != ViewSettings.ANAGLYPH_OFF) {
            _stereoTransform.multiply(_stereoMatrix, transform);
            transform = _stereoTransform;
        }

        this.processMatrix( transform );
        gl.glMultMatrixf(_matrixArray, 0);

        gl.glEnable(GL2.GL_COLOR_MATERIAL);
        gl.glColorMaterial(GL2.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE);

        if(polygonFile.getTriangleIndices().length != 0){
            if(transparent && image.getTriangleTransparency() != 0 ||
               !transparent && image.getTriangleTransparency() == 0  )
            {
                gl.glCallList( image.getTriangleListIndex() );
                gl.glFrontFace(GL2.GL_CW);
                gl.glCallList( image.getTriangleBackListIndex() );
                gl.glFrontFace(GL2.GL_CCW);
            }
        }
        if(polygonFile.getQuadIndices().length != 0){
            if(transparent && image.getQuadTransparency() != 0 ||
               !transparent && image.getQuadTransparency() == 0  )
            {
                gl.glCallList( image.getQuadListIndex() );
                gl.glFrontFace(GL2.GL_CW);
                gl.glCallList( image.getQuadBackListIndex() );
                gl.glFrontFace(GL2.GL_CCW);
            }
        }

        gl.glDisable(GL2.GL_LIGHTING);

        if(polygonFile.getPointIndices().length != 0){
            gl.glPointSize(image.getPointSize());
            if(transparent && image.getPointTransparency() != 0 ||
               !transparent && image.getPointTransparency() == 0  )
            {
                gl.glCallList( image.getPointListIndex() );
            }
        }
        if(polygonFile.getLineIndices().length != 0){
            gl.glLineWidth(image.getLineWidth());
            if(transparent && image.getLineTransparency() != 0 ||
               !transparent && image.getLineTransparency() == 0  )
            {
                gl.glCallList( image.getLineListIndex() );
            }
        }
        if(polygonFile.getPolylineIndices().length != 0){
            gl.glLineWidth(image.getPolylineWidth());
            if(transparent && image.getPolylineTransparency() != 0 ||
               !transparent && image.getPolylineTransparency() == 0  )
            {
                gl.glCallList( image.getPolylineListIndex() );
            }
        }
        
        gl.glPointSize(1);
        gl.glLineWidth(1);
        gl.glEnable(GL2.GL_LIGHTING);

        gl.glDisable(GL2.GL_COLOR_MATERIAL);
        gl.glPopMatrix();
    }
}
