/*
 * Modified at March 2014.
 * RIKEN, Japan
 */

/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.renderer;

import java.util.*;
import javax.media.opengl.*;
import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.images.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;

/**
 * Image builder.
 */
public class ImageBuilder
{
    public static final float RIBBON_HEIGHT = 0.01f;

    private ImageRendererImpl _renderer;
    private Molecule _molecule;

    private static final int POINT    = 0;
    private static final int LINE     = 1;
    private static final int TRIANGLE = 2;
    private static final int QUAD     = 3;

    private Vector3f _vector1 = new Vector3f();
    private Vector3f _vector2 = new Vector3f();
    private Matrix4f _matrix1 = new Matrix4f();
    
    private Vector3f[] _traceNormals;
    private static final int DIV = 20;

    private int[] _vertices;
    private int[] _indices;
    private int[][] _polylineIndices;

    /**
     * Constructor.
     */
    public ImageBuilder(ImageRendererImpl renderer)
    {
        _renderer = renderer;
        
        this.initializeTrace();
    }

    /**
    * Sets up transforms of bonds.
    */
    public void setupBond(Molecule molecule)
    {
        for(int b=0; b<molecule.numBonds(); b++)
        {
            Bond bond = molecule.getBond(b);
            this.setupBondImage(
                bond.getSrcAtom(), bond.getDstAtom(), bond.getImage());
        }
        molecule.getImage().setBondInitialized(true);
    }

    /**
    * Sets up transforms of backbones.
    */
    public void setupBackbone(Molecule molecule)
    {
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            for(int b=0; b<chain.numBackbones(); b++)
            {
                Bond bond = chain.getBackbone(b);
                this.setupBondImage(
                    bond.getSrcAtom(), bond.getDstAtom(), bond.getImage());
            }
        }
        molecule.getImage().setBackboneInitialized(true);
    }

    /**
    * Sets up transforms of hydrogen bonds.
    */
    public void setupHBond(Molecule molecule)
    {
        for(int b=0; b<molecule.numHBonds(); b++)
        {
            HBond bond = molecule.getHBond(b);
            this.setupBondImage(
                bond.getSrcAtom(), bond.getDstAtom(), bond.getImage());
            this.setupBondImage(
                bond.getSrcCA(), bond.getDstCA(), bond.getImageCA());
        }
        molecule.getImage().setHBondInitialized(true);
    }

    /**
    * Sets up transforms of disulfide bonds.
    */
    public void setupSSBond(Molecule molecule)
    {
        for(int b=0; b<molecule.numSSBonds(); b++)
        {
            SSBond bond = molecule.getSSBond(b);
            this.setupBondImage(
                bond.getSrcAtom(), bond.getDstAtom(), bond.getImage());
            this.setupBondImage(
                bond.getSrcCA(), bond.getDstCA(), bond.getImageCA());
        }
        molecule.getImage().setSSBondInitialized(true);
    }

    /**
    * Creates residue images.
    */
    public void buildResidueImage(Molecule molecule, GL2 gl)
    {
        _molecule = molecule;

        boolean cartoonRound = molecule.getImage().getCartoonRound();
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            this.processResidue(chain, cartoonRound, gl);
        }

        molecule.getImage().setRefreshResidue(false);

        _molecule = null;
    }

    /**
     * Sets up polygon image.
     */
    public void setupPolygonImage(PolygonFile polygonFile)
    {
        this.setupPolygonColor(polygonFile);
        
        if(polygonFile.getPointIndices().length != 0){
            setupPolygonImage(polygonFile, POINT);
        }
        if(polygonFile.getLineIndices().length != 0){
            setupPolygonImage(polygonFile, LINE);
        }
        if(polygonFile.getTriangleIndices().length != 0){
            setupPolygonImage(polygonFile, TRIANGLE);
        }
        if(polygonFile.getQuadIndices().length != 0){
            setupPolygonImage(polygonFile, QUAD);
        }
        if(polygonFile.getPolylineIndices().length != 0){
            setupPolylineImage(polygonFile);
        }

        polygonFile.getImage().setInitialized(true);        
    }

    /**
     * Creates polygon image.
     */
    public void buildPolygonImage(PolygonFile polygonFile, GL2 gl)
    {
        if(polygonFile.getPointIndices().length != 0){
            buildPolygonImage(polygonFile, POINT, gl);
        }
        if(polygonFile.getLineIndices().length != 0){
            buildPolygonImage(polygonFile, LINE, gl);
        }
        if(polygonFile.getTriangleIndices().length != 0){
            buildPolygonImage(polygonFile, TRIANGLE, gl);
        }
        if(polygonFile.getQuadIndices().length != 0){
            buildPolygonImage(polygonFile, QUAD, gl);
        }
        if(polygonFile.getPolylineIndices().length != 0){
            buildPolylineImage(polygonFile, gl);
        }

        polygonFile.getImage().setRefresh(false);
    }

    //--------------------------------------------------------------------
    // Sets up polygon image
    //--------------------------------------------------------------------
    private void setupPolygonColor(PolygonFile polygonFile)
    {
        Vertices vertices = polygonFile.getVertices();
        int num = vertices.length();

        int[] currColors = new int[num];
        for (int i = 0; i < num; i++) {
            currColors[i] = vertices.getColor(i);
        }

        polygonFile.getImage().setColors(currColors);
    }
    
    private void setupPolygonImage(PolygonFile polygonFile, int type)
    {
        PolygonImage image = polygonFile.getImage();
        if(type == POINT){
            image.setPointTransparency(polygonFile.getPointTransparency());
            image.setPointSize(polygonFile.getPointSize());
        }
        else if(type == LINE){
            image.setLineTransparency(polygonFile.getLineTransparency());
            image.setLineWidth(polygonFile.getLineWidth());
        }
        else if(type == TRIANGLE){
            image.setTriangleTransparency(polygonFile.getTriangleTransparency());
        }
        else if(type == QUAD){
            image.setQuadTransparency(polygonFile.getQuadTransparency());
        }
    }

    private void setupPolylineImage(PolygonFile polygonFile)
    {
        PolygonImage image = polygonFile.getImage();
        image.setPolylineTransparency(polygonFile.getPolylineTransparency());
        image.setPolylineWidth(polygonFile.getPolylineWidth());
    }

    //--------------------------------------------------------------------
    // Build polygon image
    //--------------------------------------------------------------------
    private void buildPolygonImage(PolygonFile polygonFile, int type, GL2 gl)
    {
        PolygonImage image = polygonFile.getImage();

        Vertices verts = polygonFile.getVertices();
        int[] colors = image.getColors();

        float alpha = 1;
        if(type == POINT){
            alpha = 1 - image.getPointTransparency();
        }
        else if(type == LINE){
            alpha = 1 - image.getLineTransparency();
        }
        else if(type == TRIANGLE){
            alpha = 1 - image.getTriangleTransparency();
        }
        else if(type == QUAD){
            alpha = 1 - image.getQuadTransparency();
        }
        else return;

        setupVertices(polygonFile, type);

        float[] interleavedArray = new float[10 * _vertices.length];
        int cnt = 0;
        for (int i = 0; i < _vertices.length; i++)
        {
            int index = _vertices[i];
            interleavedArray[cnt++] = getRed(colors[index])/255.0f;
            interleavedArray[cnt++] = getGreen(colors[index])/255.0f;
            interleavedArray[cnt++] = getBlue(colors[index])/255.0f;
            interleavedArray[cnt++] = alpha;
            interleavedArray[cnt++] = verts.getNormalX(index);
            interleavedArray[cnt++] = verts.getNormalY(index);
            interleavedArray[cnt++] = verts.getNormalZ(index);
            interleavedArray[cnt++] = verts.getX(index);
            interleavedArray[cnt++] = verts.getY(index);
            interleavedArray[cnt++] = verts.getZ(index);
        }

        if(type == POINT){
            int listIndex = image.getPointListIndex();
            image.setPointListIndex(
                _renderer.createPolygonImage(
                listIndex, GL2.GL_POINTS, interleavedArray, _indices, gl)
            );
        }
        else if(type == LINE){
            int listIndex = image.getLineListIndex();
            image.setLineListIndex(
                _renderer.createPolygonImage(
                listIndex, GL2.GL_LINES, interleavedArray, _indices, gl)
            );
        }
        else if(type == TRIANGLE){
            int listIndex = image.getTriangleListIndex();
            image.setTriangleListIndex(
                _renderer.createPolygonImage(
                listIndex, GL2.GL_TRIANGLES, interleavedArray, _indices, gl)
            );
            this.reverseNormal(interleavedArray);
            int listIndex2 = image.getTriangleBackListIndex();
            image.setTriangleBackListIndex(
                _renderer.createPolygonImage(
                listIndex2, GL2.GL_TRIANGLES, interleavedArray, _indices, gl)
            );
        }
        else if(type == QUAD){
            int listIndex = image.getQuadListIndex();
            image.setQuadListIndex(
                _renderer.createPolygonImage(
                listIndex, GL2.GL_QUADS, interleavedArray, _indices, gl)
            );
            this.reverseNormal(interleavedArray);
            int listIndex2 = image.getQuadBackListIndex();
            image.setQuadBackListIndex(
                _renderer.createPolygonImage(
                listIndex2, GL2.GL_QUADS, interleavedArray, _indices, gl)
            );
        }
        _vertices = null;
        _indices = null;
    }

    private void setupVertices(PolygonFile polygonFile, int type) {
        int[] origIndices = null;
        int unit = 0;
        if (type == POINT) {
            origIndices = polygonFile.getPointIndices();
            unit = 1;
        } else if (type == LINE) {
            origIndices = polygonFile.getLineIndices();
            unit = 2;
        } else if (type == TRIANGLE) {
            origIndices = polygonFile.getTriangleIndices();
            unit = 3;
        } else if (type == QUAD) {
            origIndices = polygonFile.getQuadIndices();
            unit = 4;
        } else {
            return;
        }

        int nV = polygonFile.getVertices().length();
        int nI = origIndices.length;

        BitSet displayFlag = polygonFile.getDisplayFlags();
        BitSet vertexFlag = new BitSet(nV);
        BitSet indexFlag = new BitSet(nI);

        int cV = 0;
        int cI = 0;
        boolean AND = true;
        int n = nI / unit;
        for (int i = 0; i < n; i++) {
            boolean use = AND;
            for (int j = 0; j < unit; j++) {
                int iV = origIndices[i * unit + j];
                if (AND) {
                    if (!displayFlag.get(iV)) {
                        use = false;
                    }
                } else {
                    if (displayFlag.get(iV)) {
                        use = true;
                    }
                }
            }
            if (use) {
                for (int j = 0; j < unit; j++) {
                    int iV = origIndices[i * unit + j];
                    if (!vertexFlag.get(iV)) {
                        vertexFlag.set(iV);
                        cV++;
                    }
                    indexFlag.set(i * unit + j);
                    cI++;
                }
            }
        }

        // vertices
        _vertices = new int[cV];
        int[] newIndices = new int[nV];
        int curr = 0;
        for (int i = 0; i < nV; i++) {
            if (vertexFlag.get(i)) {
                newIndices[i] = curr;
                _vertices[curr++] = i;
            }
        }

        // indices
        _indices = new int[cI];
        curr = 0;
        for (int i = 0; i < nI; i++) {
            if (indexFlag.get(i)) {
                _indices[curr++] = newIndices[origIndices[i]];
            }
        }
    }

    private void reverseNormal(float[] array)
    {
        int num = array.length/10;
        for(int i=0; i<num; i++){
            array[10*i+4] *= -1;
            array[10*i+5] *= -1;
            array[10*i+6] *= -1;
        }
    }

    private void buildPolylineImage(PolygonFile polygonFile,GL2 gl)
    {
        PolygonImage image = polygonFile.getImage();

        Vertices verts = polygonFile.getVertices();
        int[] colors = image.getColors();

        float alpha = 1 - image.getPolylineTransparency();

        setupPolylineVertices(polygonFile);

        float[] interleavedArray = new float[10 * _vertices.length];
        int cnt = 0;
        for (int i = 0; i < _vertices.length; i++)
        {
            int index = _vertices[i];
            interleavedArray[cnt++] = getRed(colors[index])/255.0f;
            interleavedArray[cnt++] = getGreen(colors[index])/255.0f;
            interleavedArray[cnt++] = getBlue(colors[index])/255.0f;
            interleavedArray[cnt++] = alpha;
            interleavedArray[cnt++] = verts.getNormalX(index);
            interleavedArray[cnt++] = verts.getNormalY(index);
            interleavedArray[cnt++] = verts.getNormalZ(index);
            interleavedArray[cnt++] = verts.getX(index);
            interleavedArray[cnt++] = verts.getY(index);
            interleavedArray[cnt++] = verts.getZ(index);
        }

        int listIndex = image.getPolylineListIndex();
        image.setPolylineListIndex(
            _renderer.createPolygonImage(
            listIndex, GL2.GL_LINE_STRIP, interleavedArray, _polylineIndices, gl)
        );
        _vertices = null;
        _polylineIndices = null;
    }

    private void setupPolylineVertices(PolygonFile polygonFile) {
        int[][] origIndices = polygonFile.getPolylineIndices();
        int nV = polygonFile.getVertices().length();
        
        BitSet displayFlag = polygonFile.getDisplayFlags();
        BitSet vertexFlag = new BitSet(nV);

        int cV = 0;
        List list = new ArrayList();
        for (int i = 0; i < origIndices.length; i++) {
            boolean use = false;
            int start = 0;
            for (int j = 0; j < origIndices[i].length; j++) {
                if (displayFlag.get(origIndices[i][j])) {
                    if(!use){
                        use = true;
                        start = j;
                    }
                } else if(use) {
                    int len = j - start;
                    if (len >= 2) {
                        cV = addPolylineIndice(list, origIndices[i], start, len, vertexFlag, cV);
                    }
                    use = false;
                }
            }   
            if(use){
                int len = origIndices[i].length - start;
                if(len >= 2){
                    cV = addPolylineIndice(list, origIndices[i], start, len, vertexFlag, cV);
                }
            }
        }

        // vertices
        _vertices = new int[cV];
        int[] newIndices = new int[nV];
        int curr = 0;
        for (int i = 0; i < nV; i++) {
            if (vertexFlag.get(i)) {
                newIndices[i] = curr;
                _vertices[curr++] = i;
            }
        }

        // indices
        _polylineIndices = new int[list.size()][];
        list.toArray(_polylineIndices);
        for (int i = 0; i < _polylineIndices.length; i++) {
            for (int j = 0; j < _polylineIndices[i].length; j++) {
                _polylineIndices[i][j] = newIndices[_polylineIndices[i][j]];
            }
        }
    }

    private int addPolylineIndice(List list, int[] currIndices, int start, int len,
            BitSet vertexFlag, int cV) {
        int[] curr = new int[len];
        for (int k = 0; k < len; k++) {
            int iV = currIndices[start + k];
            curr[k] = iV;
            if (!vertexFlag.get(iV)) {
                vertexFlag.set(iV);
                cV++;
            }
        }
        list.add(curr);
        return cV;
    }

    private final int getRed(int color) {
        return (color >> 16) & 0xFF;
    }

    private final int getGreen(int color) {
        return (color >> 8) & 0xFF;
    }

    private final int getBlue(int color) {
        return color & 0xFF;
    }

    //--------------------------------------------------------------------
    // Sets up bond image
    //--------------------------------------------------------------------
    private void setupBondImage(Atom srcAtom, Atom dstAtom, BondImage image)
    {
        if(srcAtom == null || dstAtom == null) return;
        
        // direction vector
        _vector1.set(
             dstAtom.getX() - srcAtom.getX(),
             dstAtom.getY() - srcAtom.getY(),
             dstAtom.getZ() - srcAtom.getZ()
        );

        float len = _vector1.length();

        if(len != 0){
            _vector1.multiply(1/len); // normalize
            
            if(_vector1.x == 0 && _vector1.y == 0){
                _vector2.set(1, 0, 0);
            }
            else{
                _vector2.set(_vector1.y, -_vector1.x, 0);
                _vector2.normalize();
            }

            image.setDistance(len);
            image.setDirection(_vector1);
            image.setNormal(_vector2);
        }
    }

    //--------------------------------------------------------------------
    // Process residue
    //--------------------------------------------------------------------
    private void processResidue(Chain chain, boolean cartoonRound, GL2 gl)
    {
        ResidueGeometry[] geometries = new ResidueGeometry[3];

        geometries[1] = new ResidueGeometry();
        geometries[2] = new ResidueGeometry();

        Residue prevResidue = null;
        Residue currResidue = null;
        Residue nextResidue = null;
        Atom currCA = null;
        Atom nextCA = null;
        SplineGeometry initSpline = null;
        SplineGeometry termSpline = null;

        currResidue = chain.getResidue(0);
        if( currResidue.isProtein() ){
            currCA = currResidue.getProteinCA();
        }
        else if( currResidue.isNucleotide() ){
            currCA = currResidue.getNucleotideP();
        }

        for(int re=0; re<chain.numResidues()+1; re++)
        {
            if(re+1 > chain.numResidues()-1){
                nextResidue = null;
                nextCA = null;
            }
            else{
                nextResidue = chain.getResidue(re+1);
            }

            if(nextResidue == null){
                termSpline = null;
            }
            else{
                Atom currO = null;
                if(nextResidue.isProtein()){
                    nextCA = nextResidue.getProteinCA();
                    currO  = currResidue.findAtom(AtomTable.INDEX_O);
                }
                else{ // DNA
                    nextCA = nextResidue.getNucleotideP();
                    currO  = nextResidue.findAtom(AtomTable.INDEX_O5_ASTERISK);
                }

                if( currCA == null || nextCA == null || currO == null
                    || nextResidue.isChainBreak() )
                {
                    termSpline = null;
                }
                else{
                    termSpline = createSplineGeometry(
                        currResidue, currCA, currO, nextCA);

                    boolean handleHelix = false;

                    // compensate for narrowing of helices
                    if(currResidue.getStructure() == Residue.STRUCT_HELIX
                       && nextResidue.getStructure() == Residue.STRUCT_HELIX)
                    {
                        Vector3f splinePoint = termSpline.getSplinePoint();
                        splinePoint.add( termSpline.getVertical() );
                        termSpline.setSplinePoint(splinePoint);
                        handleHelix = true;
                    }

                    // handle C-O flip
                    if(initSpline != null){
                        Vector3f initH = initSpline.getHorizontal();
                        Vector3f termH = termSpline.getHorizontal();
                        if( Vector3f.dot(initH, termH) < 0 ){
                            if(handleHelix){
                                Vector3f initV = initSpline.getVertical();
                                initV.multiply(-1);
                                initH.multiply(-1);
                                initSpline.setVertical(initV);
                                initSpline.setHorizontal(initH);
                            }
                            else{
                                Vector3f termV = termSpline.getVertical();
                                termV.multiply(-1);
                                termH.multiply(-1);
                                termSpline.setVertical(termV);
                                termSpline.setHorizontal(termH);
                            }
                        }
                    }

                    // unable to determine normal vector
                    if(termSpline.getHeight() == 0){
                        termSpline = null;
                    }
                }
            }

            geometries[0] = geometries[1];
            geometries[1] = geometries[2];
            geometries[2] = createResidueGeometry(
                prevResidue, currResidue, nextResidue, currCA, initSpline, termSpline);

            prevResidue = currResidue;
            currResidue = nextResidue;
            currCA = nextCA;
            initSpline = termSpline;

            Residue drawResidue = geometries[1].getResidue();
            if(geometries[1].getSplines() != null)
            {
                setNeighborWidth(geometries);
                if(drawResidue.getImage().getModel() == ResidueImage.MODEL_RIBBON)
                {
                    createRibbon(geometries[1], gl);
                }
                else if(drawResidue.getImage().getModel() == ResidueImage.MODEL_CARTOON)
                {
                    if (cartoonRound) {
                        if (drawResidue.getStructure() != Residue.STRUCT_SHEET) {
                            createCartoonRoundEdge(geometries[1], gl);
                        } else {
                            createCartoon(geometries[1], gl);
                        }
                    } else {
                        createCartoon(geometries[1], gl);
                    }
                }
                else if(drawResidue.getImage().getModel() == ResidueImage.MODEL_CARTOON_TUBE)
                {
                    createTrace(geometries[1], gl);
                }
                else if(drawResidue.getImage().getModel() == ResidueImage.MODEL_TRACE)
                {
                    createTrace(geometries[1], gl);
                }
            }
            //else if(re != 0) drawResidue.getImage().setListIndex(0);
        }
    }

    private SplineGeometry createSplineGeometry(
        Residue currResidue, Atom currCA, Atom currO, Atom nextCA)
    {
        SplineGeometry splineGeometry = new SplineGeometry();

        Vector3f pointCurrCA
            = new Vector3f( currCA.getX(), currCA.getY(), currCA.getZ() );
        Vector3f pointNextCA
            = new Vector3f( nextCA.getX(), nextCA.getY(), nextCA.getZ() );
        Vector3f pointCurrO
            = new Vector3f( currO.getX(), currO.getY(), currO.getZ() );

        Vector3f splinePoint = new Vector3f();
        splinePoint.add(pointCurrCA, pointNextCA);
        splinePoint.multiply(0.5f);
        splineGeometry.setSplinePoint(splinePoint);

        Vector3f direction = new Vector3f();
        direction.sub(pointNextCA, pointCurrCA);
        direction.multiply(0.5f);
        splineGeometry.setDirection(direction);

        Vector3f vCO = new Vector3f();
        if(currResidue.isProtein()){
            vCO.sub(pointCurrO, pointCurrCA);
        }
        else{
            Atom currO1P = currResidue.findAtom(AtomTable.INDEX_O1P);
            Atom currO2  = currResidue.findAtom(AtomTable.INDEX_O2_ASTERISK);
            if(currO1P != null && currO2 == null){ // DNA
                Vector3f pointCurrO1P = new Vector3f(
                    currO1P.getX(), currO1P.getY(), currO1P.getZ() );
                vCO.add(pointCurrO, pointCurrO1P);
                vCO.multiply(0.5f);
                vCO.sub(pointCurrCA);
            }
            else{ // RNA
                vCO.sub(pointCurrO, pointCurrCA);
            }
        }

        Vector3f vertical = new Vector3f();
        Vector3f horizontal = new Vector3f();

        vertical.cross(direction, vCO);
        if(vertical.lengthSquared() != 0){
            vertical.normalize();
            horizontal.cross(vertical, direction);
            horizontal.normalize();
        }

        splineGeometry.setVertical(vertical);
        splineGeometry.setHorizontal(horizontal);

        return splineGeometry;
    }

    private ResidueGeometry createResidueGeometry(
        Residue prevResidue, 
        Residue currResidue, Residue nextResidue, Atom currCA,
        SplineGeometry initSpline, SplineGeometry termSpline)
    {
        SplineGeometry initGeometry = null;
        SplineGeometry termGeometry = null;
        boolean spline = true;
        boolean arrow  = false;
        boolean draw   = true;

        if(initSpline == null && termSpline == null){
            draw = false;
        }
        else{
            if(initSpline == null){
                termGeometry = termSpline.copy();
                initGeometry = termSpline.copy();
                initGeometry.setSplinePoint(
                    new Vector3f( currCA.getX(), currCA.getY(), currCA.getZ() )
                );
                spline = false;
            }
            else if(termSpline == null){
                initGeometry = initSpline.copy();
                termGeometry = initSpline.copy();
                Vector3f termPoint = new Vector3f(
                    currCA.getX(), currCA.getY(), currCA.getZ() );
                if(currResidue.getStructure() == Residue.STRUCT_SHEET){
                    termPoint.add( initSpline.getDirection() );
                    arrow = true;
                }
                termGeometry.setSplinePoint(termPoint);
                spline = false;
            }
            else{
                initGeometry = initSpline.copy();
                termGeometry = termSpline.copy();
            }

            if( nextResidue != null &&
                currResidue.getStructure() == Residue.STRUCT_SHEET &&
                nextResidue.getStructure() != Residue.STRUCT_SHEET )
            {
                arrow = true;
            }

            boolean initHelix = false;
            boolean termHelix = false;
            float initHelixWidth = 0.001f;
            float termHelixWidth = 0.001f;
            if(currResidue.getStructure() == Residue.STRUCT_HELIX){
                if(prevResidue == null){
                    initHelix = true;
                }
                else if(prevResidue.getStructure() != Residue.STRUCT_HELIX){
                    initHelix = true;
                    // previous arrow head?
                    if( prevResidue.getStructure() != Residue.STRUCT_SHEET )
                    {
                        float prevWidth = prevResidue.getImage().getWidth();
                        if(prevWidth != 0) initHelixWidth = prevWidth;
                    }
                }
                if(nextResidue == null){
                    termHelix = true;
                }
                else if(nextResidue.getStructure() != Residue.STRUCT_HELIX){
                    termHelix = true;
                    float termWidth = nextResidue.getImage().getWidth();
                    if(termWidth != 0) termHelixWidth = termWidth;
                }
            }

            float width = currResidue.getImage().getWidth();
            float cartoonHeight = _molecule.getImage().getCartoonHeight();

            if(currResidue.getImage().getModel() == ResidueImage.MODEL_RIBBON)
            {
                if(initHelix) initGeometry.setWidth(initHelixWidth);
                else          initGeometry.setWidth(width);
                if(termHelix) termGeometry.setWidth(termHelixWidth);
                else          termGeometry.setWidth(width);

                initGeometry.setHeight(RIBBON_HEIGHT);
                termGeometry.setHeight(RIBBON_HEIGHT);
            }
            else if(currResidue.getImage().getModel() == ResidueImage.MODEL_CARTOON)
            {
                initGeometry.setHeight(cartoonHeight);
                termGeometry.setHeight(cartoonHeight);
                if(arrow){
                    initGeometry.setWidth(1.5f * width);
                    termGeometry.setWidth(0.001f * width);
                }
                else{
                    if(initHelix) initGeometry.setWidth(initHelixWidth);
                    else          initGeometry.setWidth(width);
                    if(termHelix) termGeometry.setWidth(termHelixWidth);
                    else          termGeometry.setWidth(width);
                }
            }
            else if(currResidue.getImage().getModel() == ResidueImage.MODEL_CARTOON_TUBE)
            {
                initGeometry.setWidth(cartoonHeight);
                termGeometry.setWidth(cartoonHeight);
                initGeometry.setHeight(cartoonHeight);
                termGeometry.setHeight(cartoonHeight);
            }
            else if(currResidue.getImage().getModel() == ResidueImage.MODEL_TRACE)
            {
                initGeometry.setWidth(width);
                termGeometry.setWidth(width);
                initGeometry.setHeight(width);
                termGeometry.setHeight(width);
            }
        }

        SplineGeometry[] splines = null;
        if(draw){
            if(spline){
                splines = createSpline(initGeometry, termGeometry);
            }
            else{
                splines = new SplineGeometry[]{initGeometry, termGeometry};
            }
        }

        ResidueGeometry rGeometry = new ResidueGeometry();
        rGeometry.setResidue(currResidue);
        rGeometry.setSplines(splines);

        return rGeometry;
    }

    private SplineGeometry[] createSpline(SplineGeometry init, SplineGeometry term)
    {
        // num = 2^n
        int num = 16;
        SplineGeometry[] splineGeometries = new SplineGeometry[num+1];

        splineGeometries[0] = init;
        splineGeometries[num] = term;

        while(num > 1)
        {
            this.createMiddleSpline(splineGeometries, num);
            num /= 2;
        }

        return splineGeometries;
    }
    
    private void createMiddleSpline(SplineGeometry[] splineGeometries, int pitch)
    {
        int num = splineGeometries.length-1;

        for(int i=0; i<num; i+=pitch)
        {
            SplineGeometry init = splineGeometries[i];
            SplineGeometry term = splineGeometries[i+pitch];

            Vector3f initP = init.getSplinePoint();
            Vector3f initD = init.getDirection();
            Vector3f initH = init.getHorizontal();
            Vector3f initV = init.getVertical();
            Vector3f termP = term.getSplinePoint();
            Vector3f termD = term.getDirection();
            Vector3f termH = term.getHorizontal();
            Vector3f termV = term.getVertical();

            int curr = i + pitch/2;

            splineGeometries[curr] = new SplineGeometry();
            splineGeometries[curr].setSplinePoint(
                new Vector3f(
                    (2*initP.x + initD.x + 2*termP.x - termD.x)/4,
                    (2*initP.y + initD.y + 2*termP.y - termD.y)/4,
                    (2*initP.z + initD.z + 2*termP.z - termD.z)/4 
                )
            );
            splineGeometries[curr].setHorizontal(
                new Vector3f(
                    (initH.x + termH.x)/2,
                    (initH.y + termH.y)/2,
                    (initH.z + termH.z)/2 
                )
            );
            splineGeometries[curr].setVertical(
                new Vector3f(
                    (initV.x + termV.x)/2,
                    (initV.y + termV.y)/2,
                    (initV.z + termV.z)/2 
                )
            );
            splineGeometries[curr].setDirection(
                new Vector3f(
                    (2*termP.x - termD.x - 2*initP.x - initD.x)/4,
                    (2*termP.y - termD.y - 2*initP.y - initD.y)/4,
                    (2*termP.z - termD.z - 2*initP.z - initD.z)/4
                )
            );

            init.getDirection().multiply(0.5f);
        }

        splineGeometries[num].getDirection().multiply(0.5f);
    }

    private void setNeighborWidth(ResidueGeometry[] geometries)
    {
        float pWidth;
        if( geometries[0].getSplines() == null ||
            geometries[1].getResidue().isChainBreak() ||
            !geometries[0].getResidue().getImage().isDisplayed())
        {
            pWidth = 0;
        }
        else
        {
            int model0 = geometries[0].getResidue().getImage().getModel();
            int model1 = geometries[1].getResidue().getImage().getModel();
            if(model1 == ResidueImage.MODEL_TRACE)
            {
                if( model0 != ResidueImage.MODEL_TRACE )
                {
                    pWidth = 0;
                }
                else{
                    SplineGeometry[] splines = geometries[0].getSplines(); 
                    pWidth = splines[ splines.length-1 ].getWidth();
                }
            }
            else if (model1 == ResidueImage.MODEL_CARTOON_TUBE) {
                if( model0 != ResidueImage.MODEL_CARTOON_TUBE )
                {
                    pWidth = 0;
                }
                else{
                    SplineGeometry[] splines = geometries[0].getSplines(); 
                    pWidth = splines[ splines.length-1 ].getWidth();
                }
            }
            else{
                if( model0 != ResidueImage.MODEL_CARTOON )
                {
                    pWidth = 0;
                }
                else{
                    SplineGeometry[] splines = geometries[0].getSplines(); 
                    pWidth = splines[ splines.length-1 ].getWidth();
                }
            }
        }
        
        float nWidth;
        if( geometries[2].getSplines() == null ||
            geometries[2].getResidue().isChainBreak() ||
            !geometries[2].getResidue().getImage().isDisplayed())
        {
            nWidth = 0;
        }
        else
        {
            int model1 = geometries[1].getResidue().getImage().getModel();
            int model2 = geometries[2].getResidue().getImage().getModel();
            if(model1 == ResidueImage.MODEL_TRACE)
            {
                if( model2 != ResidueImage.MODEL_TRACE )
                {
                    nWidth = 0;
                }
                else{
                    SplineGeometry[] splines = geometries[2].getSplines(); 
                    nWidth = splines[0].getWidth();
                }
            }
            else if (model1 == ResidueImage.MODEL_CARTOON_TUBE) {
                if( model2 != ResidueImage.MODEL_CARTOON_TUBE )
                {
                    nWidth = 0;
                }
                else{
                    SplineGeometry[] splines = geometries[2].getSplines(); 
                    nWidth = splines[0].getWidth();
                }
            }
            else{
                if( model2 != ResidueImage.MODEL_CARTOON )
                {
                    nWidth = 0;
                }
                else{
                    SplineGeometry[] splines = geometries[2].getSplines(); 
                    nWidth = splines[0].getWidth();
                }
            }
        }

        geometries[1].setPrevWidth(pWidth);
        geometries[1].setNextWidth(nWidth);
    }

    private final void setArray(float[] array, Vector3f normal, Vector3f point, int index)
    {
        array[index++] = normal.x;
        array[index++] = normal.y;
        array[index++] = normal.z;
        array[index++] = point.x;
        array[index++] = point.y;
        array[index++] = point.z;
    }

    //--------------------------------------------------------------------
    // Create ribbon geometry
    //--------------------------------------------------------------------
    private void createRibbon(ResidueGeometry geometry, GL2 gl)
    {
        SplineGeometry[] splines = geometry.getSplines();
        
        int numPoints   = splines.length;
        int numVertices = 4*numPoints;

        Vector3f[] points  = new Vector3f[numVertices];
        Vector3f[] normals = new Vector3f[numVertices];

        for(int i=0; i<numPoints; i++){
            getVertices(splines[i], points, 4*i);
            getRibbonNormals(splines[i], normals, 4*i);
        }

        int numVT = 2*numPoints;       // vertices top
        int numVB = 2*numPoints;       // vertices bottom
        int numIT = 4*(numPoints - 1); // indices top
        int numIB = 4*(numPoints - 1); // indices bottom

        boolean helix = (geometry.getResidue().getStructure() == Residue.STRUCT_HELIX);

        int numV1, numV2, numI1, numI2;

        if(helix){
            numV1 = numVT;
            numV2 = numVB;
            numI1 = numIT;
            numI2 = numIB;
        }
        else{
            numV1 = numVT + numVB;
            numV2 = 0;
            numI1 = numIT + numIB;
            numI2 = 0;
        }

        float[] ver1 = new float[6*numV1];
        float[] ver2 = new float[6*numV2];
        int[]   ind1 = new int[numI1];
        int[]   ind2 = new int[numI2];

        float[] verT, verB;
        int[]   indT, indB;
        int initVT, initVB, initIT, initIB;

        if(helix){
            verT = ver1; verB = ver2;
            indT = ind1; indB = ind2;
            initVT = 0; initVB = 0;
            initIT = 0; initIB = 0;
        }
        else{
            verT = ver1; verB = ver1;
            indT = ind1; indB = ind1;
            initVT = 0; initVB = numVT;
            initIT = 0; initIB = numIT;
        }

        // vertex
        for(int i=0; i<numPoints; i++)
        {
            int currIT = 6*initVT + 12*i;
            int currIB = 6*initVB + 12*i;

            // top
            this.setArray(verT, normals[4*i+1], points[4*i+1], currIT);
            this.setArray(verT, normals[4*i+2], points[4*i+2], currIT+6);
            // bottom
            this.setArray(verB, normals[4*i+3], points[4*i+3], currIB);
            this.setArray(verB, normals[4*i  ], points[4*i  ], currIB+6);
        }

        // index
        for(int i=0; i<numPoints-1; i++)
        {
            int currIT = initIT + 4*i;
            int currIB = initIB + 4*i;
            int currVT = initVT + 2*i;
            int currVB = initVB + 2*i;

            // top
            indT[currIT  ] = currVT + 1;
            indT[currIT+1] = currVT + 0;
            indT[currIT+2] = currVT + 2;
            indT[currIT+3] = currVT + 3;
            // bottom
            indB[currIB  ] = currVB + 1;
            indB[currIB+1] = currVB + 0;
            indB[currIB+2] = currVB + 2;
            indB[currIB+3] = currVB + 3;
        }

        ResidueImage image = geometry.getResidue().getImage(); 

        image.setListIndex( 
            _renderer.createResidueImage(image.getListIndex(), ver1, ind1, gl)
        );
        image.setBackListIndex( 
            _renderer.createResidueImage(image.getBackListIndex(), ver2, ind2, gl)
        );
    }

    private void getVertices(
        SplineGeometry splineGeometry, Vector3f[] points, int index)
    {
        for(int i=0; i<4; i++){
            points[index + i]
                = new Vector3f( splineGeometry.getSplinePoint() );
        }

        points[index  ].sub( splineGeometry.getHorizontal() );
        points[index++].sub( splineGeometry.getVertical() );

        points[index  ].sub( splineGeometry.getHorizontal() );
        points[index++].add( splineGeometry.getVertical() );

        points[index  ].add( splineGeometry.getHorizontal() );
        points[index++].add( splineGeometry.getVertical() );

        points[index  ].add( splineGeometry.getHorizontal() );
        points[index++].sub( splineGeometry.getVertical() );
    }

    private void getRibbonNormals(
        SplineGeometry splineGeometry, Vector3f[] normals, int index)
    {
        Vector3f widthVector  = new Vector3f( splineGeometry.getHorizontal() );
        Vector3f heightVector = new Vector3f( splineGeometry.getVertical() );

        float width = splineGeometry.getWidth();
        float height = splineGeometry.getHeight();

        widthVector.multiply(height/width);
        heightVector.multiply(width/height);

        for(int i=0; i<4; i++){
            normals[index + i] = new Vector3f();
        }

        normals[index].sub(widthVector);
        normals[index].sub(heightVector);
        float len = normals[index].length();
        normals[index++].multiply(1/len);

        normals[index  ].sub(widthVector);
        normals[index  ].add(heightVector);
        normals[index++].multiply(1/len);

        normals[index  ].add(widthVector);
        normals[index  ].add(heightVector);
        normals[index++].multiply(1/len);

        normals[index  ].add(widthVector);
        normals[index  ].sub(heightVector);
        normals[index++].multiply(1/len);
    }

    //--------------------------------------------------------------------
    // Create cartoon geometry
    //--------------------------------------------------------------------
    private void createCartoon(ResidueGeometry geometry, GL2 gl)
    {
        SplineGeometry[] splines = geometry.getSplines();

        int numPoints   = splines.length;
        int numVertices = 4*numPoints;

        int init = 0;
        int term = numPoints-1;

        boolean tail = ( splines[init].getWidth() > geometry.getPrevWidth() );
        boolean head = ( splines[term].getWidth() > geometry.getNextWidth() );

        Vector3f[] points  = new Vector3f[numVertices];
        Vector3f[] normals = new Vector3f[numVertices];

        for(int i=0; i<numPoints; i++){
            getVertices(splines[i], points, 4*i);
            getCartoonNormals(splines[i], normals, 4*i);
        }

        Vector3f[] tailPoints = null;
        Vector3f[] headPoints = null;
        Vector3f   tailNormal = null;
        Vector3f   headNormal = null;

        if(tail){
            SplineGeometry spline = splines[init].copy();
            spline.setWidth( geometry.getPrevWidth() );
            tailPoints = new Vector3f[4];
            getVertices(spline, tailPoints, 0);
            tailNormal = new Vector3f( spline.getDirection() );
            tailNormal.normalize();
            tailNormal.multiply(-1);
        }
        if(head){
            SplineGeometry spline = splines[term].copy();
            spline.setWidth( geometry.getNextWidth() );
            headPoints = new Vector3f[4];
            getVertices(spline, headPoints, 0);
            headNormal = new Vector3f( spline.getDirection() );
            headNormal.normalize();
        }

        int numVT = 2*numPoints;       // vertices top
        int numVB = 2*numPoints;       // vertices bottom
        int numVS = 4*numPoints;       // vertices side
        int numIT = 4*(numPoints - 1); // indices top
        int numIB = 4*(numPoints - 1); // indices bottom
        int numIS = 8*(numPoints - 1); // indices side

        if(tail){ numVS += 8; numIS += 8; }
        if(head){ numVS += 8; numIS += 8; }

        boolean helix = (geometry.getResidue().getStructure() == Residue.STRUCT_HELIX);
        boolean sheet = (geometry.getResidue().getStructure() == Residue.STRUCT_SHEET);

        int numV1, numV2, numI1, numI2;

        if(helix){
            numV1 = numVT + numVS;
            numV2 = numVB;
            numI1 = numIT + numIS;
            numI2 = numIB;
        }
        else if(sheet){
            numV1 = numVT + numVB;
            numV2 = numVS;
            numI1 = numIT + numIB;
            numI2 = numIS;
        }
        else{
            numV1 = numVT + numVS + numVB;
            numV2 = 0;
            numI1 = numIT + numIS + numIB;
            numI2 = 0;
        }

        float[] ver1 = new float[6*numV1];
        float[] ver2 = new float[6*numV2];
        int[]   ind1 = new int[numI1];
        int[]   ind2 = new int[numI2];

        float[] verT, verB, verS;
        int[]   indT, indB, indS;
        int initVT, initVB, initVS, initIT, initIB, initIS;

        if(helix){
            verT = ver1; verB = ver2; verS = ver1;
            indT = ind1; indB = ind2; indS = ind1;
            initVT = 0; initVB = 0; initVS = numVT;
            initIT = 0; initIB = 0; initIS = numIT;
        }
        else if(sheet){
            verT = ver1; verB = ver1; verS = ver2;
            indT = ind1; indB = ind1; indS = ind2;
            initVT = 0; initVB = numVT; initVS = 0;
            initIT = 0; initIB = numIT; initIS = 0;
        }
        else{
            verT = ver1; verB = ver1; verS = ver1;
            indT = ind1; indB = ind1; indS = ind1;
            initVT = 0; initVB = numVT; initVS = numVT + numVB;
            initIT = 0; initIB = numIT; initIS = numIT + numIB;
        }

        // vertex
        for(int i=0; i<numPoints; i++)
        {
            int currIT = 6*initVT + 12*i;
            int currIB = 6*initVB + 12*i;
            int currIS = 6*initVS + 24*i;

            // top
            this.setArray(verT, normals[4*i], points[4*i+1], currIT);
            this.setArray(verT, normals[4*i], points[4*i+2], currIT+6);
            // bottom
            this.setArray(verB, normals[4*i+2], points[4*i+3], currIB);
            this.setArray(verB, normals[4*i+2], points[4*i  ], currIB+6);
            // side
            this.setArray(verS, normals[4*i+3], points[4*i  ], currIS);
            this.setArray(verS, normals[4*i+3], points[4*i+1], currIS+6);
            this.setArray(verS, normals[4*i+1], points[4*i+2], currIS+12);
            this.setArray(verS, normals[4*i+1], points[4*i+3], currIS+18);
        }

        int currI = 6*initVS + 24*numPoints;
        if(tail){
            this.setArray(verS, tailNormal, tailPoints[0],    currI);
            this.setArray(verS, tailNormal, tailPoints[1],    currI+6);
            this.setArray(verS, tailNormal, tailPoints[2],    currI+12);
            this.setArray(verS, tailNormal, tailPoints[3],    currI+18);
            this.setArray(verS, tailNormal, points[4*init  ], currI+24);
            this.setArray(verS, tailNormal, points[4*init+1], currI+30);
            this.setArray(verS, tailNormal, points[4*init+2], currI+36);
            this.setArray(verS, tailNormal, points[4*init+3], currI+42);
            currI += 48;
        }
        if(head){
            this.setArray(verS, headNormal, points[4*term  ], currI);
            this.setArray(verS, headNormal, points[4*term+1], currI+6);
            this.setArray(verS, headNormal, points[4*term+2], currI+12);
            this.setArray(verS, headNormal, points[4*term+3], currI+18);
            this.setArray(verS, headNormal, headPoints[0],    currI+24);
            this.setArray(verS, headNormal, headPoints[1],    currI+30);
            this.setArray(verS, headNormal, headPoints[2],    currI+36);
            this.setArray(verS, headNormal, headPoints[3],    currI+42);
        }

        // index
        for(int i=0; i<numPoints-1; i++)
        {
            int currIT = initIT + 4*i;
            int currIB = initIB + 4*i;
            int currIS = initIS + 8*i;
            int currVT = initVT + 2*i;
            int currVB = initVB + 2*i;
            int currVS = initVS + 4*i;

            // top
            indT[currIT  ] = currVT + 1;
            indT[currIT+1] = currVT + 0;
            indT[currIT+2] = currVT + 2;
            indT[currIT+3] = currVT + 3;
            // bottom
            indB[currIB  ] = currVB + 1;
            indB[currIB+1] = currVB + 0;
            indB[currIB+2] = currVB + 2;
            indB[currIB+3] = currVB + 3;
            // side
            indS[currIS  ] = currVS + 1;
            indS[currIS+1] = currVS + 0;
            indS[currIS+2] = currVS + 4;
            indS[currIS+3] = currVS + 5;
            indS[currIS+4] = currVS + 3;
            indS[currIS+5] = currVS + 2;
            indS[currIS+6] = currVS + 6;
            indS[currIS+7] = currVS + 7;
        }

        currI = initIS + 8*(numPoints-1);
        int currV = initVS + 4*numPoints;
        if(tail){
            indS[currI  ] = currV + 1;
            indS[currI+1] = currV + 0;
            indS[currI+2] = currV + 4;
            indS[currI+3] = currV + 5;
            indS[currI+4] = currV + 3;
            indS[currI+5] = currV + 2;
            indS[currI+6] = currV + 6;
            indS[currI+7] = currV + 7;
            currI += 8;
            currV += 8;
        }
        if(head){
            indS[currI  ] = currV + 1;
            indS[currI+1] = currV + 0;
            indS[currI+2] = currV + 4;
            indS[currI+3] = currV + 5;
            indS[currI+4] = currV + 3;
            indS[currI+5] = currV + 2;
            indS[currI+6] = currV + 6;
            indS[currI+7] = currV + 7;
        }

        ResidueImage image = geometry.getResidue().getImage(); 

        image.setListIndex( 
            _renderer.createResidueImage(image.getListIndex(), ver1, ind1, gl)
        );
        image.setBackListIndex( 
            _renderer.createResidueImage(image.getBackListIndex(), ver2, ind2, gl)
        );
    }

    private void getCartoonNormals(
        SplineGeometry splineGeometry, Vector3f[] normals, int index)
    {
        normals[index  ] = new Vector3f( splineGeometry.getVertical() );
        normals[index+1] = new Vector3f( splineGeometry.getHorizontal() );

        normals[index  ].normalize();
        normals[index+1].normalize();

        normals[index+2] = new Vector3f( normals[index  ] );
        normals[index+3] = new Vector3f( normals[index+1] );

        normals[index+2].multiply(-1);
        normals[index+3].multiply(-1);
    }

    private void createCartoonRoundEdge(ResidueGeometry geometry, GL2 gl)
    {
        SplineGeometry[] splines = geometry.getSplines();

        int numPoints   = splines.length;
        int numVertices = 8*numPoints;

        int init = 0;
        int term = numPoints-1;

        boolean tail = ( splines[init].getWidth() > geometry.getPrevWidth() );
        boolean head = ( splines[term].getWidth() > geometry.getNextWidth() );

        Vector3f[] points  = new Vector3f[numVertices];
        Vector3f[] normals = new Vector3f[numVertices];

        for(int i=0; i<numPoints; i++){
            getVerticesRoundEdge(splines[i], points, 8*i);
            getCartoonNormals(splines[i], normals, 4*i);
        }

        Vector3f tailNormal = null;
        Vector3f headNormal = null;

        if(tail){
            tailNormal = new Vector3f( splines[init].getDirection() );
            tailNormal.normalize();
            tailNormal.multiply(-1);
        }
        if(head){
            headNormal = new Vector3f( splines[term].getDirection() );
            headNormal.normalize();
        }

        int numVT = 4*numPoints;        // vertices top
        int numVB = 4*numPoints;        // vertices bottom
        int numVS = 4*numPoints;        // vertices side
        int numIT = 12*(numPoints - 1); // indices top
        int numIB = 12*(numPoints - 1); // indices bottom
        int numIS = 8*(numPoints - 1);  // indices side

        if(tail){ numVS += 8; numIS += 12; }
        if(head){ numVS += 8; numIS += 12; }

        boolean helix = (geometry.getResidue().getStructure() == Residue.STRUCT_HELIX);
        boolean sheet = (geometry.getResidue().getStructure() == Residue.STRUCT_SHEET);

        int numV1, numV2, numI1, numI2;

        if(helix){
            numV1 = numVT + numVS;
            numV2 = numVB;
            numI1 = numIT + numIS;
            numI2 = numIB;
        }
        else if(sheet){
            numV1 = numVT + numVB;
            numV2 = numVS;
            numI1 = numIT + numIB;
            numI2 = numIS;
        }
        else{
            numV1 = numVT + numVS + numVB;
            numV2 = 0;
            numI1 = numIT + numIS + numIB;
            numI2 = 0;
        }

        float[] ver1 = new float[6*numV1];
        float[] ver2 = new float[6*numV2];
        int[]   ind1 = new int[numI1];
        int[]   ind2 = new int[numI2];

        float[] verT, verB, verS;
        int[]   indT, indB, indS;
        int initVT, initVB, initVS, initIT, initIB, initIS;

        if(helix){
            verT = ver1; verB = ver2; verS = ver1;
            indT = ind1; indB = ind2; indS = ind1;
            initVT = 0; initVB = 0; initVS = numVT;
            initIT = 0; initIB = 0; initIS = numIT;
        }
        else if(sheet){
            verT = ver1; verB = ver1; verS = ver2;
            indT = ind1; indB = ind1; indS = ind2;
            initVT = 0; initVB = numVT; initVS = 0;
            initIT = 0; initIB = numIT; initIS = 0;
        }
        else{
            verT = ver1; verB = ver1; verS = ver1;
            indT = ind1; indB = ind1; indS = ind1;
            initVT = 0; initVB = numVT; initVS = numVT + numVB;
            initIT = 0; initIB = numIT; initIS = numIT + numIB;
        }

        // vertex
        for(int i=0; i<numPoints; i++)
        {
            int currIT = 6*initVT + 24*i;
            int currIB = 6*initVB + 24*i;
            int currIS = 6*initVS + 24*i;

            // top
            this.setArray(verT, normals[4*i+3], points[8*i+2], currIT);
            this.setArray(verT, normals[4*i  ], points[8*i+3], currIT+6);
            this.setArray(verT, normals[4*i  ], points[8*i+4], currIT+12);
            this.setArray(verT, normals[4*i+1], points[8*i+5], currIT+18);
            // bottom
            this.setArray(verB, normals[4*i+1], points[8*i+6], currIB);
            this.setArray(verB, normals[4*i+2], points[8*i+7], currIB+6);
            this.setArray(verB, normals[4*i+2], points[8*i  ], currIB+12);
            this.setArray(verB, normals[4*i+3], points[8*i+1], currIB+18);
            // side
            this.setArray(verS, normals[4*i+3], points[8*i+1], currIS);
            this.setArray(verS, normals[4*i+3], points[8*i+2], currIS+6);
            this.setArray(verS, normals[4*i+1], points[8*i+5], currIS+12);
            this.setArray(verS, normals[4*i+1], points[8*i+6], currIS+18);
        }

        int currI = 6*initVS + 24*numPoints;
        if(tail){
            this.setArray(verS, tailNormal, points[8*init  ], currI);
            this.setArray(verS, tailNormal, points[8*init+1], currI+6);
            this.setArray(verS, tailNormal, points[8*init+2], currI+12);
            this.setArray(verS, tailNormal, points[8*init+3], currI+18);
            this.setArray(verS, tailNormal, points[8*init+4], currI+24);
            this.setArray(verS, tailNormal, points[8*init+5], currI+30);
            this.setArray(verS, tailNormal, points[8*init+6], currI+36);
            this.setArray(verS, tailNormal, points[8*init+7], currI+42);
            currI += 48;
        }
        if(head){
            this.setArray(verS, headNormal, points[8*term  ], currI);
            this.setArray(verS, headNormal, points[8*term+1], currI+6);
            this.setArray(verS, headNormal, points[8*term+2], currI+12);
            this.setArray(verS, headNormal, points[8*term+3], currI+18);
            this.setArray(verS, headNormal, points[8*term+4], currI+24);
            this.setArray(verS, headNormal, points[8*term+5], currI+30);
            this.setArray(verS, headNormal, points[8*term+6], currI+36);
            this.setArray(verS, headNormal, points[8*term+7], currI+42);
        }

        // index
        for(int i=0; i<numPoints-1; i++)
        {
            int currIT = initIT + 12*i;
            int currIB = initIB + 12*i;
            int currIS = initIS + 8*i;
            int currVT = initVT + 4*i;
            int currVB = initVB + 4*i;
            int currVS = initVS + 4*i;

            // top
            indT[currIT   ] = currVT + 1;
            indT[currIT+1 ] = currVT + 0;
            indT[currIT+2 ] = currVT + 4;
            indT[currIT+3 ] = currVT + 5;
            indT[currIT+4 ] = currVT + 2;
            indT[currIT+5 ] = currVT + 1;
            indT[currIT+6 ] = currVT + 5;
            indT[currIT+7 ] = currVT + 6;
            indT[currIT+8 ] = currVT + 3;
            indT[currIT+9 ] = currVT + 2;
            indT[currIT+10] = currVT + 6;
            indT[currIT+11] = currVT + 7;
            // bottom
            indB[currIB   ] = currVB + 1;
            indB[currIB+1 ] = currVB + 0;
            indB[currIB+2 ] = currVB + 4;
            indB[currIB+3 ] = currVB + 5;
            indB[currIB+4 ] = currVB + 2;
            indB[currIB+5 ] = currVB + 1;
            indB[currIB+6 ] = currVB + 5;
            indB[currIB+7 ] = currVB + 6;
            indB[currIB+8 ] = currVB + 3;
            indB[currIB+9 ] = currVB + 2;
            indB[currIB+10] = currVB + 6;
            indB[currIB+11] = currVB + 7;
            // side
            indS[currIS  ] = currVS + 1;
            indS[currIS+1] = currVS + 0;
            indS[currIS+2] = currVS + 4;
            indS[currIS+3] = currVS + 5;
            indS[currIS+4] = currVS + 3;
            indS[currIS+5] = currVS + 2;
            indS[currIS+6] = currVS + 6;
            indS[currIS+7] = currVS + 7;
        }

        currI = initIS + 8*(numPoints-1);
        int currV = initVS + 4*numPoints;
        if(tail){
            indS[currI   ] = currV + 0;
            indS[currI+1 ] = currV + 1;
            indS[currI+2 ] = currV + 2;
            indS[currI+3 ] = currV + 3;
            indS[currI+4 ] = currV + 0;
            indS[currI+5 ] = currV + 3;
            indS[currI+6 ] = currV + 4;
            indS[currI+7 ] = currV + 7;
            indS[currI+8 ] = currV + 7;
            indS[currI+9 ] = currV + 4;
            indS[currI+10] = currV + 5;
            indS[currI+11] = currV + 6;
            currI += 12;
            currV += 8;
        }
        if(head){
            indS[currI   ] = currV + 6;
            indS[currI+1 ] = currV + 5;
            indS[currI+2 ] = currV + 4;
            indS[currI+3 ] = currV + 7;
            indS[currI+4 ] = currV + 7;
            indS[currI+5 ] = currV + 4;
            indS[currI+6 ] = currV + 3;
            indS[currI+7 ] = currV + 0;
            indS[currI+8 ] = currV + 3;
            indS[currI+9 ] = currV + 2;
            indS[currI+10] = currV + 1;
            indS[currI+11] = currV + 0;
        }

        ResidueImage image = geometry.getResidue().getImage(); 

        image.setListIndex( 
            _renderer.createResidueImage(image.getListIndex(), ver1, ind1, gl)
        );
        image.setBackListIndex( 
            _renderer.createResidueImage(image.getBackListIndex(), ver2, ind2, gl)
        );
    }

    private void getVerticesRoundEdge(
        SplineGeometry splineGeometry, Vector3f[] points, int index)
    {
        for(int i=0; i<8; i++){
            points[index + i]
                = new Vector3f( splineGeometry.getSplinePoint() );
        }

        Vector3f h2 = new Vector3f(splineGeometry.getHorizontal());
        Vector3f v2 = new Vector3f(splineGeometry.getVertical());

        h2.multiply(0.8f);
        v2.multiply(0.5f);

        points[index  ].sub( h2 );
        points[index++].sub( splineGeometry.getVertical() );

        points[index  ].sub( splineGeometry.getHorizontal() );
        points[index++].sub( v2 );

        points[index  ].sub( splineGeometry.getHorizontal() );
        points[index++].add( v2 );

        points[index  ].sub( h2 );
        points[index++].add( splineGeometry.getVertical() );

        points[index  ].add( h2 );
        points[index++].add( splineGeometry.getVertical() );

        points[index  ].add( splineGeometry.getHorizontal() );
        points[index++].add( v2 );

        points[index  ].add( splineGeometry.getHorizontal() );
        points[index++].sub( v2 );

        points[index  ].add( h2 );
        points[index++].sub( splineGeometry.getVertical() );
    }

    //--------------------------------------------------------------------
    // Create trace geometry
    //--------------------------------------------------------------------
    private void createTrace(ResidueGeometry geometry, GL2 gl)
    {
        SplineGeometry[] splines = geometry.getSplines();

        int numPoints   = splines.length;
        int numVertices = DIV*numPoints;

        int init = 0;
        int term = numPoints-1;

        boolean tail = ( splines[init].getWidth() > geometry.getPrevWidth() );
        boolean head = ( splines[term].getWidth() > geometry.getNextWidth() );

        Vector3f[] points  = new Vector3f[numVertices];
        Vector3f[] normals = new Vector3f[numVertices];

        float radius = splines[0].getWidth();
        for(int i=0; i<numPoints; i++)
        {
            getTraceGeometry(splines[i], normals, points, DIV*i, radius);
        }

        Vector3f tailPoint = null;
        Vector3f headPoint = null;
        Vector3f tailNormal = null;
        Vector3f headNormal = null;

        if(tail){
            tailPoint  = new Vector3f( splines[init].getSplinePoint() );
            tailNormal = new Vector3f( splines[init].getDirection() );
            tailNormal.normalize();
            tailNormal.multiply(-1);
        }
        if(head){
            headPoint  = new Vector3f( splines[term].getSplinePoint() );
            headNormal = new Vector3f( splines[term].getDirection() );
            headNormal.normalize();
        }

        int numV = DIV*numPoints;         // vertices
        int numI = 4*DIV*(numPoints - 1); // indices

        if(tail){ numV += (DIV + 1); numI += 4*DIV; }
        if(head){ numV += (DIV + 1); numI += 4*DIV; }

        float[] ver = new float[6*numV];
        int[]   ind = new int[numI];

        // vertex
        for(int i=0; i<numPoints; i++)
        {
            int currIS = 6*DIV*i;
            int currVS = DIV*i;

            for(int j=0; j<DIV; j++)
            {
                this.setArray(ver, normals[currVS+j], points[currVS+j], currIS+6*j);
            }
        }

        int currI = 6*DIV*numPoints;
        if(tail){
            this.setArray(ver, tailNormal, tailPoint, currI);
            currI += 6;
            int currV = DIV*init;
            for(int j=0; j<DIV; j++)
            {
                this.setArray(ver, tailNormal, points[currV+j], currI+6*j);
            }
            currI += 6*DIV;
        }
        if(head){
            this.setArray(ver, headNormal, headPoint, currI);
            currI += 6;
            int currV = DIV*term;
            for(int j=0; j<DIV; j++)
            {
                this.setArray(ver, headNormal, points[currV+j], currI+6*j);
            }
        }

        // index
        for(int i=0; i<numPoints-1; i++)
        {
            int currIS = 4*DIV*i;
            int currV1 = DIV*i;
            int currV2 = DIV*(i+1);

            int cnt = 0;
            for(int j=0; j<DIV-1; j++)
            {
                ind[currIS+(cnt++)] = currV1 + j;
                ind[currIS+(cnt++)] = currV1 + j + 1;
                ind[currIS+(cnt++)] = currV2 + j + 1;
                ind[currIS+(cnt++)] = currV2 + j;
            }
            ind[currIS+(cnt++)] = currV1 + DIV - 1;
            ind[currIS+(cnt++)] = currV1;
            ind[currIS+(cnt++)] = currV2;
            ind[currIS+(cnt++)] = currV2 + DIV - 1;
        }

        currI = 4*DIV*(numPoints-1);
        int currV = DIV*numPoints;
        if(tail){
            int cnt = 0;
            for(int j=0; j<DIV-1; j++)
            {
                ind[currI+(cnt++)] = currV;
                ind[currI+(cnt++)] = currV;
                ind[currI+(cnt++)] = currV + j + 2;
                ind[currI+(cnt++)] = currV + j + 1;
            }
            ind[currI+(cnt++)] = currV;
            ind[currI+(cnt++)] = currV;
            ind[currI+(cnt++)] = currV + 1;
            ind[currI+(cnt++)] = currV + DIV;
            currI += 4*DIV;
            currV += (DIV + 1);
        }
        if(head){
            int cnt = 0;
            for(int j=0; j<DIV-1; j++)
            {
                ind[currI+(cnt++)] = currV;
                ind[currI+(cnt++)] = currV;
                ind[currI+(cnt++)] = currV + j + 1;
                ind[currI+(cnt++)] = currV + j + 2;
            }
            ind[currI+(cnt++)] = currV;
            ind[currI+(cnt++)] = currV;
            ind[currI+(cnt++)] = currV + DIV;
            ind[currI+(cnt++)] = currV + 1;
        }

        ResidueImage image = geometry.getResidue().getImage(); 

        image.setListIndex( 
            _renderer.createResidueImage(image.getListIndex(), ver, ind, gl)
        );
    }

    private void getTraceGeometry(
        SplineGeometry splineGeometry, Vector3f[] normals, Vector3f[] points,
        int index, float radius)
    {
        for(int i=0; i<DIV; i++){
            normals[index + i] = new Vector3f( _traceNormals[i] );
            points[index + i] = new Vector3f( splineGeometry.getSplinePoint() );
        }

        float w = splineGeometry.getWidth();
        float h = splineGeometry.getHeight();

        Vector3f baseX = new Vector3f( splineGeometry.getHorizontal() );
        Vector3f baseY = new Vector3f( splineGeometry.getVertical() );
        baseX.multiply( 1.0f/w );
        baseY.multiply( 1.0f/h );
        Vector3f baseZ = new Vector3f(
            baseX.y*baseY.z - baseX.z*baseY.y,
            baseX.z*baseY.x - baseX.x*baseY.z,
            baseX.x*baseY.y - baseX.y*baseY.x
        );

        _matrix1.m11 = baseX.x; _matrix1.m12 = baseY.x; _matrix1.m13 = baseZ.x;
        _matrix1.m21 = baseX.y; _matrix1.m22 = baseY.y; _matrix1.m23 = baseZ.y;
        _matrix1.m31 = baseX.z; _matrix1.m32 = baseY.z; _matrix1.m33 = baseZ.z;
        _matrix1.m14 = _matrix1.m24 = _matrix1.m34 = 0;
        _matrix1.m41 = _matrix1.m42 = _matrix1.m43 = 0;
        _matrix1.m44 = 1;

        for(int i=0; i<DIV; i++){
            normals[index + i].transform(_matrix1);
            _vector1.set(normals[index + i]);
            _vector1.multiply(radius);
            points[index + i].add(_vector1);
        }
    }

    private void initializeTrace()
    {
        _traceNormals = new Vector3f[DIV];

        double div = 2*Math.PI/DIV;
        double theta = 0;
        for(int i=0; i<DIV; i++){
            _traceNormals[i] = new Vector3f(
                (float)Math.cos(theta), (float)Math.sin(theta), 0
            );
            theta += div;
        }
    }
}
