/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.renderer;

import java.awt.*;
import java.awt.image.*;
import java.util.*;
import javax.media.opengl.GLAutoDrawable;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;
import org.pdbj.viewer.utils.*;

/**
 * Image renderer.
 */
public interface ImageRenderer {
    /**
     * Initialize.
     */
    public void initialize(ColorManager colorManager, RendererListener listener);

    /**
     * Initialize stereo.
     */
    public void initializeStereo();

    /**
     * Returns the left canvas.
     */
    public Canvas getCanvasL();

    /**
     * Returns the right canvas.
     */
    public Canvas getCanvasR();

    /**
     * Returns the left drawable.
     */
    public GLAutoDrawable getDrawableL();

    /**
     * Returns the right drawable.
     */
    public GLAutoDrawable getDrawableR();

    /**
     * Returns the view settings.
     */
    public ViewSettings getViewSettings();

    /**
     * Tests whether the graphics is ready.
     */
    public boolean graphicsReady();

    /**
     * Tests whether the renderer is running.
     */
    public boolean running();

    /**
     * Returns the collection of molecule files.
     */
    public ArrayList getMoleculeFiles();

    /**
     * Returns the collection of polygon files.
     */
    public ArrayList getPolygonFiles();

    /**
     * Sets the color manager.
     */
    public void setColorManager(ColorManager colorManager);

    /**
     * Returns the color manager.
     */
    public ColorManager getColorManager();

    /**
     * Starts the renderer.
     */
    public void start();

    /**
     * Stops the renderer.
     */
    public void stop();

    /**
     * Appends the molecule file.
     */
    public void add(MoleculeFile moleculeFile);

    /**
     * Appends the polygon file.
     */
    public void add(PolygonFile polygonFile);

    /**
     * Removes the molecule file.
     */
    public void remove(MoleculeFile moleculeFile);

    /**
     * Removes the polygon file.
     */
    public void remove(PolygonFile polygonFile);

    /**
     * Render images.
     */
    public void render();

    /**
     * Returns the image of the left canvas.
     */
    public BufferedImage getImage();

    /**
     * Notifies that light changed.
     */
    public void setLightChanged();

    /**
     * Notifies that camera settings changed.
     */
    public void setCameraChanged();

    /**
     * Notifies that perspective changed.
     */
    public void setPerspectiveChanged();

    /**
     * Notifies that background changed.
     */
    public void setBackgroundChanged();

    /**
     * Notifies that draw level changed.
     */
    public void setDrawlevelChanged();

    /**
     * Dispose.
     */
    public void dispose();

    /**
     * Sets whether the stereo mode.
     */
    public void setStereo(boolean enable);

    /**
     * Sets the stereo center.
     */
    public void setStereoCenter(Vector3f center);

    /**
     * Sets the stereo angle.
     */
    public void setStereoAngle(float angle);

    /**
    * Sets the anaglyph mode.
    */
    public void setAnaglyphMode(int anaglyphMode);

    /**
     * Destroy GLWindow.
     */
    public void destroyGLWindow();
}
