/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.renderer;

import java.awt.*;
import java.awt.image.*;
import java.nio.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.awt.GLCanvas;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.images.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;
import com.jogamp.newt.awt.NewtCanvasAWT;
import com.jogamp.newt.opengl.GLWindow;

/**
 * Image renderer with JSR-231.
 */
public class ImageRendererImpl implements ImageRenderer
{
    private boolean _newt;
    private ViewSettings _viewSettings;
    private ColorManager _colorManager;
    private ImageBuilder _imageBuilder;

    private boolean _ready;
    private boolean _running;

    private GLCapabilities _capabilities;
    private Canvas _canvasL;
    private Canvas _canvasR;
    private GLAutoDrawable _drawableL;
    private GLAutoDrawable _drawableR;
    
    private CanvasRenderer _rendererL;
    private CanvasRenderer _rendererR;

    private ArrayList _moleculeFiles;
    private ArrayList _polygonFiles;
    private ArrayList _deleteMoleculeFiles;
    private ArrayList _deletePolygonFiles;

    boolean changeDrawlevel;
    boolean deleteList;
    boolean dispose;

    int sphere;
    int clipCap;
    int cylinder;

    // buffer
    private FloatBuffer _floatBuffer;
    private IntBuffer _intBuffer;

    // cylinder
    private Hashtable _table;
    private float[] _x;
    private float[] _y;
    private int[] _indices;
    private IntBuffer _indicesBuffer;
    private float _prevRadius;
    private int _prevList;

    /**
    * Constructor.
    */
    public ImageRendererImpl(boolean newt)
    {
        _newt = newt;
    }

    /**
     * Initialize.
     */
    public void initialize(ColorManager colorManager, RendererListener listener)
    {
        _colorManager = colorManager;

        _viewSettings = new ViewSettings();
        _imageBuilder = new ImageBuilder(this);

        _moleculeFiles = new ArrayList();
        _polygonFiles  = new ArrayList();
        
        _deleteMoleculeFiles = new ArrayList();
        _deletePolygonFiles  = new ArrayList();
        
        _rendererL = new CanvasRenderer(this);
        _rendererR = new CanvasRenderer(this);
        
        _rendererL.setRendererListener(listener);

        this.initializeGraphics();

        this.changeDrawlevel = true;
        
        _table = new Hashtable();

        initBuffer();
    }

    /**
    * Returns the left canvas.
    */
    public Canvas getCanvasL()
    {
        return _canvasL;
    }

    /**
    * Returns the right canvas.
    */
    public Canvas getCanvasR()
    {
        return _canvasR;
    }

    /**
     * Returns the left drawable.
     */
    public GLAutoDrawable getDrawableL()
    {
        return _drawableL;
    }

    /**
     * Returns the right drawable.
     */
    public GLAutoDrawable getDrawableR()
    {
        return _drawableR;
    }

    /**
    * Returns the view settings.
    */
    public ViewSettings getViewSettings()
    {
        return _viewSettings;
    }

    /**
    * Tests whether the graphics is ready.
    */
    public boolean graphicsReady()
    {
        return _ready;
    }

    /**
    * Tests whether the renderer is running.
    */
    public boolean running()
    {
        return _running;
    }

    /**
    * Returns the collection of molecule files.
    */
    public ArrayList getMoleculeFiles()
    {
        return _moleculeFiles;
    }

    /**
    * Returns the collection of polygon files.
    */
    public ArrayList getPolygonFiles()
    {
        return _polygonFiles;
    }

    /**
    * Sets the color manager.
    */
    public void setColorManager(ColorManager colorManager)
    {
         _colorManager = colorManager;
    }

    /**
    * Returns the color manager.
    */
    public ColorManager getColorManager()
    {
        return _colorManager;
    }

    /**
    * Starts the renderer.
    */
    public void start()
    {
        if(_ready){
            _running = true;
            this.render();
        }
    }

    /**
    * Stops the renderer.
    */
    public void stop()
    {
        _running = false;
    }

    /**
     * Appends the molecule file.
     */
    public void add(MoleculeFile moleculeFile)
    {
        _moleculeFiles.add(moleculeFile);
    }

    /**
     * Appends the polygon file.
     */
    public void add(PolygonFile polygonFile)
    {
        _polygonFiles.add(polygonFile);
        _imageBuilder.setupPolygonImage(polygonFile);
    }

    /**
     * Removes the molecule file.
     */
    public void remove(MoleculeFile moleculeFile)
    {
        int index = _moleculeFiles.indexOf(moleculeFile);
        if(index != -1){
            _moleculeFiles.remove(index);
            _deleteMoleculeFiles.add(moleculeFile);
            this.deleteList = true;
        }
    }

    /**
     * Removes the polygon file.
     */
    public void remove(PolygonFile polygonFile)
    {
        int index = _polygonFiles.indexOf(polygonFile);
        if(index != -1){
            _polygonFiles.remove(index);
            _deletePolygonFiles.add(polygonFile);
            this.deleteList = true;
        }
    }

    /**
     * Render images.
     */
    public void render()
    {
        try{
            if(_running)
            {
                _drawableL.display();
                if( _viewSettings.isStereo() ) _drawableR.display();

                // enforce refresh ... a bug in JOGL?
                if( _viewSettings.shiniessChanged() ){
                    _drawableL.display();
                    if( _viewSettings.isStereo() ) _drawableR.display();
                    _viewSettings.setShininessChanged(false);                
                }
            }
        }
        catch(GLException e){
            System.out.println(e.getMessage());
        }
    }

    /**
     * Returns the image of the left canvas.
     */
    public BufferedImage getImage()
    {
        int w = _canvasL.getWidth();
        int h = _canvasL.getHeight();
        
        BufferedImage image = null;
        
        if( _viewSettings.isStereo() ){
            int[] pixelDataL = _rendererL.getPixelData();
            int[] pixelDataR = _rendererR.getPixelData();
            int[] pixelData = new int[2*w*h];
            int cnt = 0;
            int cntL = 0;
            int cntR = 0;
            for(int y=0; y<h; y++){
                for(int x=0; x<w; x++){
                    pixelData[cnt++] = pixelDataL[cntL++];
                }
                for(int x=0; x<w; x++){
                    pixelData[cnt++] = pixelDataR[cntR++];
                }
            }
            image = new BufferedImage(2*w, h, BufferedImage.TYPE_INT_RGB);
            image.setRGB(0, 0, 2*w, h, pixelData, 0, 2*w);
        }
        else{
            int[] pixelData = _rendererL.getPixelData();
            image = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
            image.setRGB(0, 0, w, h, pixelData, 0, w);
        }
        
        return image;
    }

    /**
     * Notifies that light changed.
     */
    public void setLightChanged()
    {
        _rendererL.setLightChanged();
        _rendererR.setLightChanged();
    }

    /**
     * Notifies that camera settings changed.
     */
    public void setCameraChanged()
    {
        _rendererL.setCameraChanged();
        _rendererR.setCameraChanged();
    }

    /**
     * Notifies that perspective changed.
     */
    public void setPerspectiveChanged()
    {
        _rendererL.setPerspectiveChanged();
        _rendererR.setPerspectiveChanged();
    }

    /**
     * Notifies that background changed.
     */
    public void setBackgroundChanged()
    {
        _rendererL.setBackgroundChanged();
        _rendererR.setBackgroundChanged();
    }

    /**
     * Notifies that draw level changed.
     */
    public void setDrawlevelChanged()
    {
        this.changeDrawlevel = true;
    }

    /**
     * Dispose.
     */
    public void dispose()
    {
        dispose = true;
        _drawableL.display();

        if (_newt) {
            GLWindow windowL = (GLWindow) _drawableL;
            GLWindow windowR = (GLWindow) _drawableR;
            windowL.setVisible(false);
            windowL.reparentWindow(null);
            if (windowR != null) {
                windowR.setVisible(false);
                windowR.reparentWindow(null);
            }
        }
    }

    /**
     * Destroy GLWindow.
     */
    public void destroyGLWindow()
    {
        if (_newt) {
            GLWindow windowL = (GLWindow) _drawableL;
            GLWindow windowR = (GLWindow) _drawableR;
            windowL.destroy();
            if (windowR != null) {
                windowR.destroy();
            }
        }
    }

    //--------------------------------------------------------------------
    // Stereo
    //--------------------------------------------------------------------
    /**
     * Initialize stereo.
     */
    public void initializeStereo()
    {
        if (_newt) {
            GLWindow glWindow = GLWindow.create(_capabilities);
            glWindow.setSharedContext(_drawableL.getContext());
            _drawableR = glWindow;
            _canvasR = new NewtCanvasAWT(glWindow);
        } else {
            GLCanvas glCanvas = new GLCanvas(_capabilities, null, _drawableL.getContext(), null);
            _drawableR = glCanvas;
            _canvasR = glCanvas;
        }
        _rendererR.setDrawable(_drawableR);
    }

    /**
    * Sets whether the stereo mode.
    */
    public void setStereo(boolean enable)
    {
        _viewSettings.setStereo(enable);

        this.setCameraChanged();
        this.setPerspectiveChanged();
        this.setBackgroundChanged();
    }

    /**
    * Sets the stereo center.
    */
    public void setStereoCenter(Vector3f center)
    {
        _viewSettings.setStereoCenter(center);
        _rendererL.setStereoCenter(center);
        _rendererR.setStereoCenter(center);
    }

    /**
    * Sets the stereo angle.
    */
    public void setStereoAngle(float angle)
    {
        _viewSettings.setStereoAngle(angle);
        _rendererL.setStereoAngle( angle);
        _rendererR.setStereoAngle(-angle);
    }

    /**
    * Sets the anaglyph mode.
    */
    public void setAnaglyphMode(int anaglyphMode) {
        _viewSettings.setAnaglyphMode(anaglyphMode);
        _rendererL.setupAnaglyph(anaglyphMode);
    }

    //--------------------------------------------------------------------
    // Create display lists
    //--------------------------------------------------------------------
    /**
     * Returns the image builder.
     */
    ImageBuilder getImageBuilder()
    {
        return _imageBuilder;
    }

    /**
     * Creates display list for residue images and returns its index. 
     */
    int createResidueImage(
        int listIndex, float[] interleavedArray, int[] indices, GL2 gl)
    {
        if(listIndex == 0) listIndex = gl.glGenLists(1);

        gl.glNewList(listIndex, GL2.GL_COMPILE);
            if (interleavedArray.length != 0) {
                setFloatBuffer(interleavedArray);
                setIntBuffer(indices);
                gl.glInterleavedArrays(GL2.GL_N3F_V3F, 0, _floatBuffer);
                gl.glDrawElements(
                    GL2.GL_QUADS, indices.length, GL2.GL_UNSIGNED_INT, _intBuffer);
            }
        gl.glEndList();

        return listIndex;
    }

    /**
     * Creates display list for polygon images and returns its index. 
     */
    int createPolygonImage(
        int listIndex, int mode, float[] interleavedArray, int[] indices, GL2 gl)
    {
        if(listIndex == 0) listIndex = gl.glGenLists(1);

        gl.glNewList(listIndex, GL2.GL_COMPILE);
            if (interleavedArray.length != 0) {
                setFloatBuffer(interleavedArray);
                setIntBuffer(indices);
                gl.glInterleavedArrays(GL2.GL_C4F_N3F_V3F, 0, _floatBuffer);
                gl.glDrawElements(mode, indices.length, GL2.GL_UNSIGNED_INT, _intBuffer);
            }
        gl.glEndList();

        return listIndex;
    }

    /**
     * Creates display list for polygon images and returns its index. 
     */
    int createPolygonImage(
        int listIndex, int mode, float[] interleavedArray, int[][] indices, GL2 gl)
    {
        if(listIndex == 0) listIndex = gl.glGenLists(1);

        gl.glNewList(listIndex, GL2.GL_COMPILE);
            setFloatBuffer(interleavedArray);
            gl.glInterleavedArrays(GL2.GL_C4F_N3F_V3F, 0, _floatBuffer);
            for(int i=0; i<indices.length; i++){
                setIntBuffer(indices[i]);
                gl.glDrawElements(mode, indices[i].length, GL2.GL_UNSIGNED_INT, _intBuffer);
            }
        gl.glEndList();

        return listIndex;
    }

    //--------------------------------------------------------------------
    // Delete display lists
    //--------------------------------------------------------------------
    void deleteList(GL2 gl)
    {
        for(int i=0; i<_deleteMoleculeFiles.size(); i++)
        {
            MoleculeFile file = (MoleculeFile)_deleteMoleculeFiles.get(i);
            for(int m=0; m<file.numMolecules(); m++)
            {
                Molecule molecule = file.getMolecule(m);
                for(int ch=0; ch<molecule.numChains(); ch++)
                {
                    Chain chain = molecule.getChain(ch);
                    for(int re=0; re<chain.numResidues(); re++)
                    {
                        Residue residue = chain.getResidue(re);
                        int index = residue.getImage().getListIndex();
                        if(index != 0){
                            gl.glDeleteLists(index, 1);
                            residue.getImage().setListIndex(0);
                        }
                        // delete back list
                        index = residue.getImage().getBackListIndex();
                        if(index != 0){
                            gl.glDeleteLists(index, 1);
                            residue.getImage().setBackListIndex(0);
                        }
                    }
                }
            }
        }

        for(int i=0; i<_deletePolygonFiles.size(); i++)
        {
            PolygonFile file = (PolygonFile)_deletePolygonFiles.get(i);

            PolygonImage image = file.getImage();

            if(image.getPointListIndex() != 0){
                gl.glDeleteLists(image.getPointListIndex(), 1);
                image.setPointListIndex(0);
            }
            if(image.getLineListIndex() != 0){
                gl.glDeleteLists(image.getLineListIndex(), 1);
                image.setLineListIndex(0);
            }
            if(image.getTriangleListIndex() != 0){
                gl.glDeleteLists(image.getTriangleListIndex(), 1);
                image.setTriangleListIndex(0);
            }
            if(image.getQuadListIndex() != 0){
                gl.glDeleteLists(image.getQuadListIndex(), 1);
                image.setQuadListIndex(0);
            }
            if(image.getPolylineListIndex() != 0){
                gl.glDeleteLists(image.getPolylineListIndex(), 1);
                image.setPolylineListIndex(0);
            }
            if(image.getTriangleBackListIndex() != 0){
                gl.glDeleteLists(image.getTriangleBackListIndex(), 1);
                image.setTriangleBackListIndex(0);
            }
            if(image.getQuadBackListIndex() != 0){
                gl.glDeleteLists(image.getQuadBackListIndex(), 1);
                image.setQuadBackListIndex(0);
            }
        }

        _deleteMoleculeFiles.clear();
        _deletePolygonFiles.clear();
        
        this.deleteList = false;
    }

    void deletePrimitives(GL2 gl)
    {
        if(sphere != 0){
            gl.glDeleteLists(sphere, 1);
        }
        if(clipCap != 0){
            gl.glDeleteLists(clipCap, 1);
        }
        if(cylinder != 0){
            gl.glDeleteLists(cylinder, 1);
        }        
        clearCylinder(gl);
    }

    //--------------------------------------------------------------------
    // Initialize graphics
    //--------------------------------------------------------------------
    private void initializeGraphics()
    {
        _capabilities = new GLCapabilities(GLProfile.get(GLProfile.GL2));
        if (_newt) {
            GLWindow glWindow = GLWindow.create(_capabilities);
            _drawableL = glWindow;
            _canvasL = new NewtCanvasAWT(glWindow);
        } else {
            GLCanvas glCanvas = new GLCanvas(_capabilities);
            _drawableL = glCanvas;
            _canvasL = glCanvas;
        }
        _rendererL.setDrawable(_drawableL);

        _ready = true;
        _running = true;
    }

    //----------------------------------------------------------------------
    // Create primitives
    //----------------------------------------------------------------------
    void createPrimitives(GL2 gl)
    {
        // disk
        if(clipCap == 0)
        {
            clipCap = createDiskGeometry(gl);
        }

        // sphere
        sphere = createSphereGeometry(gl);
        
        // cylinder
        clearCylinder(gl);
        setupCylinder();
        _table.clear();

        changeDrawlevel = false;
    }
    
    private int createDiskGeometry(GL2 gl)
    {
        Vector3f normal = new Vector3f(0, 0, 1);
        
        int d = 15;

        Vector3f[] v = new Vector3f[2*d+1];
        
        v[0] = new Vector3f(0, 0, 0);

        double step = Math.PI/d;
        double phi = 0;

        for(int i=1; i<=2*d; i++)
        {
            float x = (float)Math.cos(phi);
            float y = (float)Math.sin(phi);
            float z = 0;
            
            v[i] = new Vector3f(x, y, z);

            phi += step;
        }

        int list = gl.glGenLists(1);
        gl.glNewList(list, GL2.GL_COMPILE);

        float[] array = new float[6*(2*d+1)];

        int[] indices = new int[2*d+2];
        
        int cnt1 = 0;
        int cnt2 = 0;
        
        for(int i=0; i<2*d+1; i++)
        {
            array[cnt1++] = normal.x;
            array[cnt1++] = normal.y;
            array[cnt1++] = normal.z;
            array[cnt1++] = v[i].x;
            array[cnt1++] = v[i].y;
            array[cnt1++] = v[i].z;
            
            indices[cnt2] = cnt2++;
        }
        indices[cnt2] = 1;

        setFloatBuffer(array);
        setIntBuffer(indices);

        gl.glInterleavedArrays(GL2.GL_N3F_V3F, 0, _floatBuffer);
        gl.glDrawElements(
            GL2.GL_TRIANGLE_FAN, indices.length, GL2.GL_UNSIGNED_INT, _intBuffer);

        gl.glEndList();

        return list;
    }

    private int createSphereGeometry(GL2 gl)
    {
        int d = _viewSettings.getDrawlevel();
        
        Vector3f[][] v = new Vector3f[2*d][d+1];
        
        double step = Math.PI/d;
        double phi = 0;
        double theta = 0;

        for(int i=0; i<2*d; i++)
        {
            theta = 0;
            
            for(int j=0; j<d+1; j++)
            {
                float x = (float)(Math.sin(theta)*Math.cos(phi));
                float y = (float)(Math.sin(theta)*Math.sin(phi));
                float z = (float)(Math.cos(theta));
                
                v[i][j] = new Vector3f(x, y, z);
                
                theta += step;
            }
            
            phi -= step;
        }

        int list = gl.glGenLists(1);
        gl.glNewList(list, GL2.GL_COMPILE);

        float[] array = new float[12*(d+1)];

        int[] indices = new int[2*(d+1)];
        
        for(int i=0; i<2*d; i++)
        {
            int cnt1 = 0;
            int cnt2 = 0;

            int curr = i;
            int next = (i == 2*d-1) ? 0 : i+1;
            
            for(int j=0; j<d+1; j++)
            {
                array[cnt1++] = v[curr][j].x;
                array[cnt1++] = v[curr][j].y;
                array[cnt1++] = v[curr][j].z;
                array[cnt1++] = v[curr][j].x;
                array[cnt1++] = v[curr][j].y;
                array[cnt1++] = v[curr][j].z;
                array[cnt1++] = v[next][j].x;
                array[cnt1++] = v[next][j].y;
                array[cnt1++] = v[next][j].z;
                array[cnt1++] = v[next][j].x;
                array[cnt1++] = v[next][j].y;
                array[cnt1++] = v[next][j].z;

                indices[cnt2] = cnt2++;
                indices[cnt2] = cnt2++;
            }

            setFloatBuffer(array);
            setIntBuffer(indices);

            gl.glInterleavedArrays(GL2.GL_N3F_V3F, 0, _floatBuffer);
            gl.glDrawElements(
                GL2.GL_QUAD_STRIP, indices.length, GL2.GL_UNSIGNED_INT, _intBuffer);
        }

        gl.glEndList();

        return list;
    }

    int getCylinder(float radius, GL2 gl)
    {
        if(radius != _prevRadius)
        {
            Float key = new Float(radius);
            Integer value = (Integer)_table.get(key);
        
            if(value == null){
                int list = this.createCylinder(radius, gl);
                value = new Integer(list);
                _table.put(key, value);
            }

            _prevRadius = radius;
            _prevList = value.intValue();
        }
        
        return _prevList;
    }
    
    private void setupCylinder()
    {
        int div = 2*_viewSettings.getDrawlevel();

        _x = new float[div];
        _y = new float[div];

        double step = 2*Math.PI/div;
        double curr = 0;
        for(int i=0; i<div; i++){
            _x[i] = (float)Math.cos(curr);
            _y[i] = (float)Math.sin(curr);
            curr += step;
        }
        
        _indices = new int[2*div+2];
        for(int i=0; i<2*div; i++){
            _indices[i] = i;
        }
        _indices[2*div]   = 0;
        _indices[2*div+1] = 1;
        _indicesBuffer = IntBuffer.wrap(_indices);

        _prevRadius = -1;
        _prevList = 0;
    }
    
    private int createCylinder(float radius, GL2 gl)
    {
        int div = 2*_viewSettings.getDrawlevel();

        float[] array = new float[12*div];
        int cnt = 0;
        for(int i=0; i<div; i++){
            array[cnt++] = _x[i];
            array[cnt++] = _y[i];
            array[cnt++] = 0;
            array[cnt++] = radius*_x[i];
            array[cnt++] = radius*_y[i];
            array[cnt++] = 1;
            array[cnt++] = _x[i];
            array[cnt++] = _y[i];
            array[cnt++] = 0;
            array[cnt++] = radius*_x[i];
            array[cnt++] = radius*_y[i];
            array[cnt++] = 0;
        }

        setFloatBuffer(array);

        int list = gl.glGenLists(1);
        gl.glNewList(list, GL2.GL_COMPILE);
            gl.glInterleavedArrays(GL2.GL_N3F_V3F, 0, _floatBuffer);
            gl.glDrawElements(
                GL2.GL_QUAD_STRIP, _indices.length, GL2.GL_UNSIGNED_INT, _indicesBuffer);
        gl.glEndList();

        return list;
    }
    
    private void clearCylinder(GL2 gl)
    {
        Enumeration keys = _table.keys();
        while( keys.hasMoreElements() )
        {
            Float key = (Float)keys.nextElement();
            int list = ( (Integer)_table.get(key) ).intValue();
            gl.glDeleteLists(list, 1);
        }
    }

    //----------------------------------------------------------------------
    // Buffer
    //----------------------------------------------------------------------
    private void initBuffer() {
        _floatBuffer = FloatBuffer.allocate(5000);
        _intBuffer = IntBuffer.allocate(5000);
    }

    private void setFloatBuffer(float[] array) {
        if (array.length > _floatBuffer.capacity()) {
            _floatBuffer = FloatBuffer.allocate(array.length);
        } else {
            _floatBuffer.clear();
        }
        _floatBuffer.put(array).position(0);
    }

    private void setIntBuffer(int[] array) {
        if (array.length > _intBuffer.capacity()) {
            _intBuffer = IntBuffer.allocate(array.length);
        } else {
            _intBuffer.clear();
        }
        _intBuffer.put(array).position(0);
    }
}

