/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.renderer;

import java.awt.Color;
import org.pdbj.viewer.utils.*;

/**
 * View settings.
 */
public class ViewSettings
{
    public static final Color DEFAULT_BACKGROUND = Color.BLACK;
    public static final int ANAGLYPH_OFF = 0;
    public static final int ANAGLYPH_REDBLUE = 1;
    public static final int ANAGLYPH_REDCYAN = 2;
    public static final int ANAGLYPH_REDGREEN = 3;

    private Vector3f _cameraPosition;
    private Vector3f _cameraTarget;
    private Vector3f _cameraUpVector;
    private float    _fieldAngle;
    private float    _aspectRatio;
    private float    _nearClipping;
    private float    _farClipping;
    private Color    _bgColor;

    private boolean _specular;
    private float[] _specularColor;
    private float   _shininess;

    private int     _drawlevel;

    private boolean  _stereo;
    private Vector3f _stereoCenter;
    private float    _stereoAngle;
    private int      _anaglyphMode;

    private boolean _bondCap;

    private float[] _clipSpecularColor;
    private boolean _shininessChanged;
    
    private boolean _parallelProjection;
    private float   _viewSize;

    private float _ambient;
    private float _diffuse1;
    private float _diffuse2;
    private float[] _diffuseDir1;
    private float[] _diffuseDir2;

    /**
     * Sets this camera position a copy of the passed vector.
     */
    public void setCameraPosition(Vector3f cameraPosition)
    {
        _cameraPosition.set(cameraPosition);
    }

    /**
     * Returns the camera position.
     */
    public Vector3f getCameraPosition()
    {
        return _cameraPosition;
    }

    /**
     * Sets this camera target a copy of the passed vector.
     */
    public void setCameraTarget(Vector3f cameraTarget)
    {
        _cameraTarget.set(cameraTarget);
    }

    /**
     * Returns the camera target.
     */
    public Vector3f getCameraTarget()
    {
        return _cameraTarget;
    }

    /**
     * Sets this camera up vector a copy of the passed vector.
     */
    public void setCameraUpVector(Vector3f cameraUpVector)
    {
        _cameraUpVector.set(cameraUpVector);
    }

    /**
     * Returns the camera up vector.
     */
    public Vector3f getCameraUpVector()
    {
        return _cameraUpVector;
    }

    /**
     * Sets the field of view in radians.
     */
    public void setFieldAngle(float fieldAngle)
    {
        _fieldAngle = fieldAngle;
    }

    /**
     * Returns the field of view in radians.
     */
    public float getFieldAngle()
    {
        return _fieldAngle;
    }

    /**
     * Sets the aspect ratio.
     */
    public void setAspectRatio(float aspectRatio)
    {
        _aspectRatio = aspectRatio;
    }

    /**
     * Returns the aspect ratio.
     */
    public float getAspectRatio()
    {
        return _aspectRatio;
    }

    /**
     * Sets the near clipping plane distance.
     */
    public void setNearClipping(float nearClipping)
    {
        _nearClipping = nearClipping;
    }

    /**
     * Returns the near clipping plane distance.
     */
    public float getNearClipping()
    {
        return _nearClipping;
    }

    /**
     * Sets the far clipping plane distance.
     */
    public void setFarClipping(float farClipping)
    {
        _farClipping = farClipping;
    }

    /**
     * Returns the far clipping plane distance.
     */
    public float getFarClipping()
    {
        return _farClipping;
    }

    /**
     * Sets the background color.
     */
    public void setBGColor(Color bgColor)
    {
        _bgColor = bgColor;
    }

    /**
     * Returns the background color.
     */
    public Color getBGColor()
    {
        return _bgColor;
    }

    /**
     * Sets whether specular is enabled.
     */
    public void setSpecular(boolean specular)
    {
        _specular = specular;
    }

    /**
     * Returns whether specular is enabled.
     */
    public boolean getSpecular()
    {
        return _specular;
    }

    /**
     * Sets the specular color.
     */
    public void setSpecularColor(float r, float g, float b, float a)
    {
        _specularColor[0] = r;
        _specularColor[1] = g;
        _specularColor[2] = b;
        _specularColor[3] = a;

        this.adjustClipSpecularColor();
    }

    /**
     * Returns the specular color.
     */
    public float[] getSpecularColor()
    {
        return _specularColor;
    }
    
    /**
     * Sets the shininess.
     */
    public void setShininess(float shininess)
    {
        _shininess = shininess;

        this.adjustClipSpecularColor();
        _shininessChanged = true;
    }
    
    /**
     * Returns the shininess.
     */
    public float getShininess()
    {
        return _shininess;
    }

    /**
     * Sets the draw level.
     */
    public void setDrawlevel(int drawlevel)
    {
        _drawlevel = drawlevel;
    }

    /**
     * Returns the draw level.
     */
    public int getDrawlevel()
    {
        return _drawlevel;
    }

    /**
     * Sets whether bond caps are displayed.
     */
    public void setBondCap(boolean enable)
    {
        _bondCap = enable;
    }

    /**
     * Returns whether bond caps are displayed.
     */
    public boolean getBondCap()
    {
        return _bondCap;
    }

    /**
    * Returns whether the stereo mode.
    */
    public boolean isStereo()
    {
        return _stereo;
    }

    /**
    * Returns the stereo center.
    */
    public Vector3f getStereoCenter()
    {
        return _stereoCenter;
    }

    /**
    * Returns the stereo angle.
    */
    public float getStereoAngle()
    {
        return _stereoAngle;
    }

    /**
    * Returns the anaglyph mode.
    */
    public int getAnaglyphMode() {
        return _anaglyphMode;
    }

    /**
    * Sets whether parallel projection.
    */
    public void setParallelProjection(boolean enable)
    {
        _parallelProjection = enable;
    }
    
    /**
    * Returns whether parallel projection.
    */
    public boolean isParallelProjection()
    {
        return _parallelProjection;
    }

    /**
    * Sets the view size in parallel projection.
    */
    public void setViewSize(float viewSize)
    {
        _viewSize = viewSize;
    }

    /**
    * Returns the view size in parallel projection.
    */
    public float getViewSize()
    {
        return _viewSize;
    }
    
    /**
     * Sets the strength of ambient light.
     */
    public void setAmbient(float ambient) {
        _ambient = ambient;
    }

    /**
     * Returns the strength of ambient light.
     */
    public float getAmbient() {
        return _ambient;
    }

    /**
     * Sets the strength of diffuse light 1.
     */
    public void setDiffuse1(float diffuse1) {
        _diffuse1 = diffuse1;
    }

    /**
     * Returns the strength of diffuse light 1.
     */
    public float getDiffuse1() {
        return _diffuse1;
    }

    /**
     * Sets the strength of diffuse light 2.
     */
    public void setDiffuse2(float diffuse2) {
        _diffuse2 = diffuse2;
    }

    /**
     * Returns the strength of diffuse light 2.
     */
    public float getDiffuse2() {
        return _diffuse2;
    }

    /**
     * Sets the direction of diffuse light 1.
     */
    public void setDiffuseDir1(float x, float y, float z) {
        _diffuseDir1[0] = x;
        _diffuseDir1[1] = y;
        _diffuseDir1[2] = z;
    }

    /**
     * Returns the direction of diffuse light 1.
     */
    public float[] getDiffuseDir1() {
        return _diffuseDir1;
    }

    /**
     * Sets the direction of diffuse light 2.
     */
    public void setDiffuseDir2(float x, float y, float z) {
        _diffuseDir2[0] = x;
        _diffuseDir2[1] = y;
        _diffuseDir2[2] = z;
    }

    /**
     * Returns the direction of diffuse light 2.
     */
    public float[] getDiffuseDir2() {
        return _diffuseDir2;
    }

    /**
     * Constructor.
     */
    public ViewSettings()
    {
        _cameraPosition = new Vector3f(0, 0, 0);
        _cameraTarget   = new Vector3f(0, 0, 1);
        _cameraUpVector = new Vector3f(0, 1, 0);
        _fieldAngle     = (float)Math.PI/4;
        _aspectRatio    = 1;
        _nearClipping   = 0;
        _farClipping    = Float.MAX_VALUE;
        _bgColor        = DEFAULT_BACKGROUND;
        
        _specular       = false;
        _specularColor  = new float[]{1, 1, 1, 1};
        _shininess      = 64;
        
        _drawlevel = 10;
        
        _bondCap = true;

        _stereoCenter = new Vector3f();

        _clipSpecularColor = new float[]{1, 1, 1, 1};
        this.adjustClipSpecularColor();

        _ambient = 0.33f;
        _diffuse1 = 0.50f;
        _diffuse2 = 0.50f;
        _diffuseDir1 = new float[]{-1, -1, -1, 0};
        _diffuseDir2 = new float[]{ 1, -1, -1, 0};
    }
    
    float[] getClipSpecularColor()
    {
        return _clipSpecularColor;
    }
    
    void setShininessChanged(boolean changed)
    {
        _shininessChanged = changed;
    }
    
    boolean shiniessChanged()
    {
        return _shininessChanged;
    }
    
    private void adjustClipSpecularColor()
    {
        double cos_ratio = 0.90; // empirically
        
        float p = (float)Math.pow(cos_ratio, _shininess);
        
        _clipSpecularColor[0] = p * _specularColor[0];
        _clipSpecularColor[1] = p * _specularColor[1];
        _clipSpecularColor[2] = p * _specularColor[2];
    }

    void setStereo(boolean enable)
    {
        _stereo = enable;
    }

    void setStereoCenter(Vector3f center)
    {
        _stereoCenter.set(center);
    }

    void setStereoAngle(float angle)
    {
        _stereoAngle = angle;
    }

    void setAnaglyphMode(int anaglyphMode) {
        _anaglyphMode = anaglyphMode;
    }
}
