/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.site;

import java.io.*;
import java.net.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;

/**
 * SiteList reader.
 */
public class SiteListReader extends DefaultHandler {
    private SiteListBuilder _builder;
    private MessageListener _listener;
    
    private boolean _pickupText;
    private StringBuffer _buffer;

    private SAXNodeHandler _handler;
    
    private SAXNode _site_list;
    private SAXNode _site;
    private SAXNode _region;
    
    private boolean _validSite;

    /**
     * Constructs an instance.
     */
    public SiteListReader(MessageListener listener) {
        _listener = listener;
        _builder = new SiteListBuilder();
        _buffer = new StringBuffer();
        setupHandler();
    }

    /**
     * Reads.
     */
    public SiteList read(URL url) throws RecoverableException {
        if (_listener != null) {
            _listener.begin(1);
            _listener.appendMessage("Retrieving a site list.");
        }
        
        try {
            parse(url);
        } catch (SAXException e) {
            throw ExceptionUtil.handleSAXException(e);
        } catch(ParserConfigurationException e){
            throw ExceptionUtil.handleParserConfigurationException(e);
        } catch (FileNotFoundException e) {
            throw new RecoverableException("Can't Find " + url);
        } catch (UnknownHostException e) {
            throw new RecoverableException("Can't Find " + url);
        } catch (IOException e) {
            throw ExceptionUtil.handleIOException(e);
        } finally {
            if (_listener != null) {
                _listener.end();
            }
        }
        
        return _builder.getResult();
    }

    //--------------------------------------------------------------------
    // Parse
    //--------------------------------------------------------------------
    private void parse(URL url) throws IOException, SAXException,
            ParserConfigurationException {
        SAXParserFactory spFactory = SAXParserFactory.newInstance();
        spFactory.setNamespaceAware(true);
        SAXParser parser = spFactory.newSAXParser();

        ResourceRequest request = new ResourceRequest(url);
        InputStream stream = null;
        try {
            stream = request.openStream();
            parser.parse(stream, this);
        } finally {
            if (stream != null) {
                stream.close();
            }
            request.dispose();
        }
    }

    //-------------------------------------------------------------------------
    // SAX Handler
    //-------------------------------------------------------------------------
    public void startDocument() throws SAXException {
        _handler.initialize();
    }
    
    public void startElement(String uri, String localName, String qName,
            Attributes attributes) throws SAXException {
        _handler.enter(localName);
        _pickupText = false;
        SAXNode curr = _handler.getCurrentNode();
        if (curr != null) {
            if (curr == _site_list) {
                _builder.initialize();
            } else if (curr == _site) {
                Site site = new Site();
                site.setId(getValueTrim(attributes, "id"));
                site.setDb(getValueTrim(attributes, "db"));
                site.setCategory(getValueTrim(attributes, "category"));
                site.setDescription(getValueTrim(attributes, "description"));
                if (isValid(site)) {
                    _builder.addSite(site);
                    _validSite = true;
                } else {
                    _validSite = false;
                }
            } else if (curr == _region) {
                Region region = new Region();
                region.setChainId(getValueTrim(attributes, "chain_id"));
                region.setBegSeqId(getValueTrim(attributes, "beg_seq_id"));
                region.setEndSeqId(getValueTrim(attributes, "end_seq_id"));
                if (_validSite && isValid(region)) {
                    fix(region);
                    _builder.addRegion(region);
                }
            }
            _pickupText = true;
            _buffer.setLength(0);
        }
    }
    
    public void characters(char[] ch, int start, int length)
            throws SAXException {
        if (_pickupText) {
            _buffer.append(new String(ch, start, length));
        }
    }
    
    public void endElement(String uri, String localName, String qName)
            throws SAXException {
        _handler.leave();
    }
    
    public void endDocument() throws SAXException {
    }

    //-------------------------------------------------------------------------
    // Util
    //-------------------------------------------------------------------------
    private String getValueTrim(Attributes attributes, String name) {
        String value = attributes.getValue(name);
        return (value != null) ? value.trim() : null;
    }

    //-------------------------------------------------------------------------
    // Validate
    //-------------------------------------------------------------------------
    private boolean isValid(Site site) {
        if (site.getId() == null) {
            return false;
        }
        if (site.getDb() == null || site.getDb().equals("")) {
            return false;
        }
        if (site.getCategory() == null || site.getCategory().equals("")) {
            return false;
        }
        if (site.getDescription() == null) {
            return false;
        }
        return true;
    }

    private boolean isValid(Region region) {
        if (region.getChainId() == null) {
            return false;
        }
        return true;
    }
    
    private void fix(Region region) {
        String chain = region.getChainId();
        if (chain.equals("") || chain.equals("' '")) {
            region.setChainId(Region.BLANK_CHAIN_ID);
        }
        
        String beg = region.getBegSeqId();
        String end = region.getEndSeqId();
        if (beg == null) {
            beg = "";
        }
        if (end == null) {
            end = "";
        }
        if (beg.equals("") && end.equals("")) {
            return;
        } else if (!beg.equals("") && !end.equals("")) {
            if (!testSeq(beg) || !testSeq(end)) {
                beg = "";
                end = "";                
            }
        } else {
            beg = "";
            end = "";                
        }
        region.setBegSeqId(beg);
        region.setEndSeqId(end);
    }
    
    private boolean testSeq(String seq) {
        char last = seq.charAt(seq.length() - 1);
        if (last < '0' || last > '9') {
            seq = seq.substring(0, seq.length() - 1);
        }
        try {
            Integer.parseInt(seq);
        } catch (NumberFormatException e) {
            return false;
        }
        return true;
    }

    //--------------------------------------------------------------------
    // Set up SAXNodeHandler
    //--------------------------------------------------------------------
    private void setupHandler() {
        _site_list = new SAXNode("site_list");
        _site = _site_list.createNode("site");
        _region = _site.createNode("region");
        _handler = new SAXNodeHandler(_site_list);
    }
}
