/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.utils;

import java.awt.Color;
import java.util.ArrayList;
import org.pdbj.viewer.tables.*;

/**
* Class that manages colors.
*/
public class ColorManager
{
    private ArrayList _colors;

    private int[] _namedIndice;
    private int[] _cpkIndice;
    private int[] _aminoIndice;
    private int[] _shapelyIndice;

    /**
    * Constructor.
    */
    public ColorManager()
    {
        _colors = new ArrayList();

        this.SetupColors();
    }

    /**
    * Number of colors registered.
    */
    public int counts()
    {
        return _colors.size();
    }

    /**
    * Returns the index of the specified color.
    */
    public int getIndex(Color color)
    {
        int i;
        for(i=0; i<_colors.size(); i++){
            Color curr = (Color)_colors.get(i);
            if( curr.equals(color) ) return i;
        }

        _colors.add(color);
        return i;
    }

    /**
    * Returns the index of the specified color.
    */
    public int getIndex(int r, int g, int b)
    {
        return this.getIndex( new Color(r, g, b) );
    }

    /**
    * Returns the index of the predefined color specified by index
    * in ColorTable class.
    */
    public int getIndex(int namedColorIndex)
    {
        return _namedIndice[namedColorIndex];
    }

    /**
    * Returns the index of the predefined color specified by index
    * in ColorTable class.
    */
    public int getIndex(String colorName)
    {
        return _namedIndice[ColorTable.getIndex(colorName)];
    }

    /**
    * Returns the index of the CPK color of the specified atomic number.
    */
    public int getCPKIndex(int atomicNo)
    {
        return _cpkIndice[atomicNo];
    }

    /**
    * Returns the index of the amino color of the specified residue index
    * in ResidueTable class.
    */
    public int getAminoIndex(int residueIndex)
    {
        return _aminoIndice[residueIndex];
    }

    /**
    * Returns the index of the shapely color of the specified residue index
    * in ResidueTable class.
    */
    public int getShapelyIndex(int residueIndex)
    {
        return _shapelyIndice[residueIndex];
    }

    /**
    * Returns the Color object of the specified index.
    */
    public Color getColor(int index)
    {
        return (Color)_colors.get(index);
    }

    //--------------------------------------------------------------------
    // set up colors
    //--------------------------------------------------------------------
    private void SetupColors()
    {
        // colors defined in ColorTable class
        _namedIndice = new int[ColorTable.counts()];
        for(int i=0; i<ColorTable.counts(); i++){
            _namedIndice[i] = this.getIndex(ColorTable.getColor(i));
        }

        // CPK colors
        _cpkIndice = new int[ElementTable.counts()];
        for(int i=0; i<ElementTable.counts(); i++){
            _cpkIndice[i] = this.getIndex(ElementTable.getCPKColor(i));
        }

        // Amino colors
        _aminoIndice = new int[ResidueTable.counts()];
        for(int i=0; i<ResidueTable.counts(); i++){
            _aminoIndice[i] = this.getIndex(ResidueTable.getAminoColor(i));
        }

        // Shapely colors
        _shapelyIndice = new int[ResidueTable.counts()];
        for(int i=0; i<ResidueTable.counts(); i++){
            _shapelyIndice[i] = this.getIndex(ResidueTable.getShapelyColor(i));
        }
    }
}
