/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.utils;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * FTP request.
 */
public class FTPRequest
{
    public static final String USER = "anonymous";
    public static final String PASS = "pdbjviewer@protein.osaka-u.ac.jp";
    
    private static final int PORT = 21;
    private static final int TIMEOUT = 60000; //ms
    
    private URL _url;
    private boolean _passive;
    
    private Socket _controlSocket;
    private Socket _dataSocket;

    private ServerSocket _dataServerSocket;

    private BufferedWriter _controlWriter;
    private BufferedReader _controlReader;
    
    private ArrayList _responseList;

    private boolean _openControl;
    private boolean _openData;
    private boolean _readData;
    
    private boolean _dump = false;

    /**
     * Constructor.
     */
    public FTPRequest(URL url)
    {
        _url = url;
        _passive = true;
        _responseList = new ArrayList();
    }

    /**
     * Sets whether the passive mode or not.
     */
    public void setPassive(boolean passive)
    {
        _passive = passive;
    }

    /**
     * Returns whether the passive mode or not.
     */
    public boolean isPassive()
    {
        return _passive;
    }

    /**
     * Returns a response stream.
     */
    public InputStream openStream() throws UnknownHostException, IOException
    {
        String host = _url.getHost();
        String path = _url.getPath();
        String dir = null;
        String file = null;

        int index = path.lastIndexOf("/");
        if(index != -1){
            dir = path.substring(0, index);
            file = path.substring(index+1);
        }
        else{
            dir = "";
            file = path;
        }
        
        this.openControlConnection(host);
        this.login();
        _openControl = true;

        this.sendCommand("TYPE I");
        this.receiveResponse(); //200

        if( !dir.equals("") ){
            this.sendCommand("CWD " + dir);
            this.receiveResponse(); //250
        }

        if(_passive){
            this.openPassiveDataConnection();
        }
        else{
            this.openActiveDataConnection();
        }
        _openData = true;

        this.sendCommand("RETR " + file);
        this.receiveResponse(); //150
        if( this.getResponseCode() == 550 ){
            throw new FileNotFoundException(file);
        }
        this.validateResponseCode(150);
        _readData = true;

        if(!_passive){
            _dataSocket = _dataServerSocket.accept();
        }

        return _dataSocket.getInputStream();
    }

    /**
     * Close.
     */
    public void close() throws IOException
    {
        if(_openData){
            if(_passive){
                this.closePassiveDataConnection();
            }
            else{
                this.closeActiveDataConnection();
            }
            _openData = false;
        }
        if(_readData){
            this.receiveResponse(); //226
            _readData = false;
        }
        if(_openControl){
            this.logout();
            this.closeControlConnection();
            _openControl = false;
        }
    }

    //--------------------------------------------------------------------
    // Login & Logout
    //--------------------------------------------------------------------
    private void login() throws IOException
    {
        this.sendCommand("USER " + USER);
        this.receiveResponse(); //331
        this.sendCommand("PASS " + PASS);
        this.receiveResponse(); //230
    }

    private void logout() throws IOException
    {
        this.sendCommand("QUIT");
        this.receiveResponse(); //221
    }

    //--------------------------------------------------------------------
    // Command
    //--------------------------------------------------------------------
    private void sendCommand(String command) throws IOException
    {
        _controlWriter.write(command + "\r\n");
        _controlWriter.flush();

        //dump
        if(_dump) System.out.println(">" + command);
    }
    
    private void receiveResponse() throws IOException
    {
        _responseList.clear();
        
        while(true){
            String response = _controlReader.readLine();
            if(response != null){
                _responseList.add(response);
                if( response.length() >=3 && isInt( response.substring(0, 3) ) ){
                    if(response.charAt(3) != '-') break;
                }
            }
            else break;
        }
        
        //dump
        if(_dump){
            for(int i=0; i<_responseList.size(); i++){
                System.out.println( _responseList.get(i) );
            }
        }
    }

    private String getLastResponse()
    {
        if(_responseList.size() > 0){
            return (String)_responseList.get(_responseList.size()-1);
        }
        else return null;
    }

    private int getResponseCode()
    {
        int code = -1;
        String lastLine = this.getLastResponse();
        if(lastLine != null){
            try{
                code = Integer.parseInt( lastLine.substring(0, 3) );
            }
            catch(NumberFormatException e){
            }
        }
        return code;
    }

    private void validateResponseCode(int code) throws IOException
    {
        if( this.getResponseCode() != code )
        {
            String lastLine = this.getLastResponse();
            if(lastLine == null) lastLine = "";

            throw new IOException(lastLine);
        }
    }

    //--------------------------------------------------------------------
    // Connection
    //--------------------------------------------------------------------
    private void openControlConnection(String host) throws UnknownHostException, IOException
    {
        _controlSocket = new Socket(host, PORT);
        _controlSocket.setSoTimeout(TIMEOUT);

        _controlWriter = new BufferedWriter(
            new OutputStreamWriter( _controlSocket.getOutputStream() )
        );
        _controlReader = new BufferedReader(
            new InputStreamReader( _controlSocket.getInputStream() )
        );
        
        this.receiveResponse(); //220
    }
    
    private void closeControlConnection() throws IOException
    {
        if(_controlWriter != null) _controlWriter.close();
        if(_controlReader != null) _controlReader.close();
        if(_controlSocket != null) _controlSocket.close();
    }

    private void openPassiveDataConnection() throws UnknownHostException, IOException
    {
        this.sendCommand("PASV");
        this.receiveResponse(); //227
        this.validateResponseCode(227);
        
        String lastLine = (String)_responseList.get(_responseList.size()-1);
        Object[] obj = this.extractAddress(lastLine);
        if(obj == null) throw new IOException("Can't resolve server address.");
        
        String host = (String)obj[0];
        int port = ((Integer)obj[1]).intValue();
        
        _dataSocket = new Socket(host, port);
        _dataSocket.setSoTimeout(TIMEOUT);
    }

    private void openActiveDataConnection() throws UnknownHostException, IOException
    {
        _dataServerSocket = new ServerSocket(0, 1);
        int port = _dataServerSocket.getLocalPort();
        
        String address = InetAddress.getLocalHost().getHostAddress();
        address = address.replace('.', ',') + "," + (port/256) + "," + (port&0xFF);

        this.sendCommand("PORT " + address);
        this.receiveResponse(); //200
        this.validateResponseCode(200);

        _dataServerSocket.setSoTimeout(TIMEOUT);
    }

    private void closePassiveDataConnection() throws IOException
    {
        if(_dataSocket != null) _dataSocket.close();
    }

    private void closeActiveDataConnection() throws IOException
    {
        if(_dataSocket != null) _dataSocket.close();
        if(_dataServerSocket != null) _dataServerSocket.close();
    }

    //--------------------------------------------------------------------
    // Util
    //--------------------------------------------------------------------
    private Object[] extractAddress(String response)
    {
        int index1 = response.indexOf("(");
        int index4 = response.indexOf(")");
        
        int index2 = index1;
        for(int i=0; i<4; i++) index2 = response.indexOf(",", index2+1);
        
        int index3 = response.lastIndexOf(",");
        
        if(index1 < 0 || index2 < 0 || index3 < 0 || index4 < 0) return null;
        
        String address = response.substring(index1+1, index2).replace(',', '.');
        int port = 0;
        try{
            port = Integer.parseInt( response.substring(index2+1, index3) ) * 256
                 + Integer.parseInt( response.substring(index3+1, index4) );
        }
        catch(NumberFormatException e){
            return null;
        }
        
        return new Object[]{address, new Integer(port)};
    }
    
    private boolean isInt(String s)
    {
        try{
            Integer.parseInt(s);
        }
        catch(NumberFormatException e){
            return false;
        }
        return true;
    }
}

