/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.utils;

import java.net.*;

/**
 * Class that contains PDBML database URL.
 */
public class PDBmlUrl
{
    public static final String DEFAULT_HOST = "ftp.pdbj.org";
    public static final String DEFAULT_URL_NOATOM   = "ftp://ftp.pdbj.org/XML/all-noatom/";
    public static final String DEFAULT_URL_EXTATOM  = "ftp://ftp.pdbj.org/XML/all-extatom/";
    public static final String DEFAULT_URL_FUNCTION = "ftp://ftp.pdbj.org/XML/pdbmlplus/pdbml_xp_gz/";

    private String _urlNoatom;
    private String _urlExtatom;
    private String _urlFunction;
    private String _hostNoatom;
    private String _hostExtatom;
    private String _hostFunction;

    /**
     * Constructor.
     */
    public PDBmlUrl(String urlNoatom, String urlExtatom, String urlFunction)
    {
        if (!setupNoatom(urlNoatom)) {
            setupDefaultNoatom();
        }
        if (!setupExtatom(urlExtatom)) {
            setupDefaultExtatom();
        }
        if (!setupFunction(urlFunction)) {
            setupDefaultFunction();
        }
    }

    /**
     * Constructor.
     */
    public PDBmlUrl()
    {
        setupDefaultNoatom();
        setupDefaultExtatom();
        setupDefaultFunction();
    }

    /**
     * Sets the URL for no-atoms.
     */
    public boolean setUrlNoatom(String urlNoatom) {
        return setupNoatom(urlNoatom);
    }

    /**
     * Returns the URL for no-atoms.
     */
    public String getUrlNoatom()
    {
        return _urlNoatom;
    }

    /**
     * Sets the URL for ext-atoms.
     */
    public boolean setUrlExtatom(String urlExtatom) {
        return setupExtatom(urlExtatom);
    }

    /**
     * Returns the URL for ext-atoms.
     */
    public String getUrlExtatom()
    {
        return _urlExtatom;
    }

    /**
     * Sets the URL for functions.
     */
    public boolean setUrlFunction(String urlFunction) {
        return setupFunction(urlFunction);
    }

    /**
     * Returns the URL for functions.
     */
    public String getUrlFunction()
    {
        return _urlFunction;
    }

    /**
     * Returns the host name for no-atoms.
     */
    public String getHostNoatom()
    {
        return _hostNoatom;
    }

    /**
     * Returns the host name for ext-atoms.
     */
    public String getHostExtatom()
    {
        return _hostExtatom;
    }

    /**
     * Returns the host name for functions.
     */
    public String getHostFunction()
    {
        return _hostFunction;
    }

    /**
     * Returns whether the URL for no-atoms is the default one.
     */
    public boolean isNoatomDefault() {
        return _urlNoatom.equals(DEFAULT_URL_NOATOM);
    }

    /**
     * Returns whether the URL for ext-atoms is the default one.
     */
    public boolean isExtatomDefault() {
        return _urlExtatom.equals(DEFAULT_URL_EXTATOM);
    }

    /**
     * Returns whether the URL for functions is the default one.
     */
    public boolean isFunctionDefault() {
        return _urlFunction.equals(DEFAULT_URL_FUNCTION);
    }

    //--------------------------------------------------------------------
    // Set up URL
    //--------------------------------------------------------------------
    private boolean setupNoatom(String urlNoatom) {
        URL url1 = null;
        try {
            url1 = new URL(urlNoatom);
        } catch (Exception e) {
            return false;
        }

        _urlNoatom = url1.toString();
        _hostNoatom = url1.getHost();
        if (!_urlNoatom.endsWith("/")) {
            _urlNoatom += "/";
        }
        return true;
    }

    private boolean setupExtatom(String urlExtatom) {
        URL url1 = null;
        try {
            url1 = new URL(urlExtatom);
        } catch (Exception e) {
            return false;
        }

        _urlExtatom = url1.toString();
        _hostExtatom = url1.getHost();
        if (!_urlExtatom.endsWith("/")) {
            _urlExtatom += "/";
        }
        return true;
    }

    private boolean setupFunction(String urlFunction) {
        URL url1 = null;
        try {
            url1 = new URL(urlFunction);
        } catch (Exception e) {
            return false;
        }

        _urlFunction = url1.toString();
        _hostFunction = url1.getHost();
        if (!_urlFunction.endsWith("/")) {
            _urlFunction += "/";
        }
        return true;
    }

    private void setupDefaultNoatom()
    {
        _urlNoatom = DEFAULT_URL_NOATOM;
        _hostNoatom = DEFAULT_HOST;
    }

    private void setupDefaultExtatom()
    {
        _urlExtatom = DEFAULT_URL_EXTATOM;
        _hostExtatom = DEFAULT_HOST;
    }

    private void setupDefaultFunction()
    {
        _urlFunction = DEFAULT_URL_FUNCTION;
        _hostFunction = DEFAULT_HOST;
    }
}
