/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.utils;

import java.io.*;
import java.net.*;
import java.util.zip.*;

/**
 * Resource request.
 */
public class ResourceRequest
{
    private File _file;
    private URL _url;
    
    private boolean _gzip;
    private boolean _ftp;
    
    private FTPRequest _ftpRequest;
    
    /**
    * Constructor for local files.
    */
    public ResourceRequest(File file)
    {
        _file = file;
        _gzip = _file.getAbsolutePath().endsWith(".gz");
    }

    /**
    * Constructor for remote files.
    */
    public ResourceRequest(URL url)
    {
        _url = url;
        _gzip = _url.toString().endsWith(".gz");
        _ftp = _url.getProtocol().equals("ftp");
    }

    /**
    * Opens stream.
    */
    public InputStream openStream() throws IOException
    {
        InputStream stream = null;

        if(_file != null){
            stream = new FileInputStream(_file);
        }
        else if(_url != null){
            if(_ftp){
                _ftpRequest = new FTPRequest(_url);
                stream = _ftpRequest.openStream();
            }
            else{
                stream = _url.openStream();
            }
        }

        if(stream != null){
            if (_gzip) {
                // workaround for a cache problem in JRE 1.7
                try {
                    return new GZIPInputStream(stream);
                } catch (IOException e) {
                    return toByteArrayStream(stream);
                }
            }
            return stream;
            //return _gzip ? new GZIPInputStream(stream) : stream;
        }
        else return null;
    }

    private InputStream toByteArrayStream(InputStream stream) throws IOException {
        StringBuffer sb = new StringBuffer();
        BufferedReader reader = new BufferedReader(new InputStreamReader(stream));

        // fix the first line
        String s = reader.readLine();
        if (s != null) {
            if (s.startsWith("xml ")) {
                sb.append("<?");
            } else if (s.startsWith("ADER ")) {
                sb.append("HE");
            } else if (s.startsWith("MARK ")) {
                sb.append("RE");
            } else if (s.startsWith("DEL ")) {
                sb.append("MO");
            } else if (s.startsWith("OM ")) {
                sb.append("AT");
            }
            sb.append(s + "\n");
        }

        try {
            while ((s = reader.readLine()) != null) {
                sb.append(s + "\n");
            }
            
        } finally {
            reader.close();
        }

        this.dispose();

        return new ByteArrayInputStream(sb.toString().getBytes());
    }

    /**
     * Dispose.
     */
    public void dispose() throws IOException
    {
        if(_url != null && _ftp){
            if(_ftpRequest != null) {
                _ftpRequest.close();
                _ftpRequest = null;
            }
        }
    }
}
