/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.utils;

/**
 * Four dimensional vector.
 */
public class Vector4f
{
    /**
     * The x component.
     */
    public float x;

    /**
     * The y component.
     */
    public float y;

    /**
     * The z component.
     */
    public float z;

    /**
     * The w component.
     */
    public float w;

    /**
     * Constructs and initializes a Vector4f to (0,0,0,0).
     */
    public Vector4f()
    {
        x = 0;
        y = 0;
        z = 0;
        w = 0;
    }

    /**
     * Constructs and initializes a Vector4f to (x,y,z,w).
     */
    public Vector4f(float x, float y, float z, float w)
    {
        this.x = x;
        this.y = y;
        this.z = z;
        this.w = w;
    }

    /**
     * Constructs and initializes a Vector4f to v1.
     */
    public Vector4f(Vector4f v1)
    {
        this.x = v1.x;
        this.y = v1.y;
        this.z = v1.z;
        this.w = v1.w;
    }

    /**
     * Constructs and initializes a Vector4f to ({v1},w).
     */
    public Vector4f(Vector3f v1, float w)
    {
        x = v1.x;
        y = v1.y;
        z = v1.z;
        this.w = w;
    }

    /**
     * Sets this vector to the vector sum of itself and vector v1.
     */
    public void add(Vector4f v1)
    {
        x += v1.x;
        y += v1.y;
        z += v1.z;
        w += v1.w;
    }

    /**
     * Sets this vector to the vector sum of vectors v1 and v2.
     */
    public void add(Vector4f v1, Vector4f v2)
    {
        x = v1.x + v2.x;
        y = v1.y + v2.y;
        z = v1.z + v2.z;
        w = v1.w + v2.w;
    }

    /**
     * Returns the dot product of vectors v1 and v2.
     */
    public static float dot(Vector4f v1, Vector4f v2)
    {
        return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z + v1.w * v2.w;
    }

    /**
     * Returns true if all components of this vector are equal to the corresponding 
     * components in vector v1.
     */
    public boolean equals(Vector4f v1)
    {
        return ( x == v1.x && y == v1.y && z == v1.z && w == v1.w);
    }

    /**
     * Returns the length of this vector.
     */
    public float length()
    {
        return (float)Math.sqrt( x*x + y*y + z*z + w*w);
    }

    /**
     * Returns the squared length of this vector.
     */
    public float lengthSquared()
    {
        return x*x + y*y + z*z + w*w;
    }

    /**
     * Sets this vector to the scalar multiplication of the scale factor with this.
     */
    public void multiply(float s)
    {
        x *= s;
        y *= s;
        z *= s;
        w *= s;
    }

    /**
     * Sets this vector to the scalar multiplication of the scale factor with vector v1.
     */
    public void multiply(float s, Vector4f v1)
    {
        x = s * v1.x;
        y = s * v1.y;
        z = s * v1.z;
        w = s * v1.w;
    }

    /**
     * Normalizes this vector in place.
     */
    public void normalize()
    {
        float length = this.length();
        if(length != 0){
            this.multiply(1/length);
        }
    }

    /**
     * Sets the elements of this vector to the specifed values.
     */
    public void set(float x, float y, float z, float w)
    {
        this.x = x;
        this.y = y;
        this.z = z;
        this.w = w;
    }

    /**
     * Sets this vector to vector v1.
     */
    public void set(Vector4f v1)
    {
        x = v1.x;
        y = v1.y;
        z = v1.z;
        w = v1.w;
    }

    /**
     * Sets this vector to the vector difference of itself and vector v1
     * (this = this - v1).
     */
    public void sub(Vector4f v1)
    {
        x -= v1.x;
        y -= v1.y;
        z -= v1.z;
        w -= v1.w;
    }

    /**
     * Sets this vector to the vector difference of vectors v1 and v2.
     */
    public void sub(Vector4f v1, Vector4f v2)
    {
        x = v1.x - v2.x;
        y = v1.y - v2.y;
        z = v1.z - v2.z;
        w = v1.w - v2.w;
    }

    /**
     * Transforms this vector using matrix m
     * (this = m * this).
     */
    public void transform(Matrix4f m1)
    {
        float x1 = m1.m11 * x + m1.m12 * y + m1.m13 * z + m1.m14 * w;
        float y1 = m1.m21 * x + m1.m22 * y + m1.m23 * z + m1.m24 * w;
        float z1 = m1.m31 * x + m1.m32 * y + m1.m33 * z + m1.m34 * w;
        float w1 = m1.m41 * x + m1.m42 * y + m1.m43 * z + m1.m44 * w;

        x = x1;
        y = y1;
        z = z1;
        w = w1;
    }


    /**
     * Returns a string expression.
     */
    public String toString()
    {
        return "(" + x + ", " + y + ", " + z + "," + w + ")";
    }
}
