/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.writers;

import java.io.*;
import java.text.NumberFormat;
import java.util.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.tables.*;

/**
 * PDB writer.
 */
public class PDBWriter
{
    private static final int NUM_COLUMNS = 80;

    private ViewerController _controller;

    private BufferedWriter _writer;
    private StringBuffer   _buffer;

    private boolean _renumber;
    private boolean _transform;

    private int _atomCnt;

    private NumberFormat _formatCoord;
    private NumberFormat _formatTemp;

    /**
     * Sets whether the atom serial numbers are renumbered.
     */
    public void setRenumber(boolean enable)
    {
        _renumber = enable;
    }

    /**
     * Returns whether the atom serial numbers are renumbered.
     */
    public boolean getRenumber()
    {
        return _renumber;
    }

    /**
     * Sets whether the atom coordinates are transformed.
     */
    public void setTransform(boolean enable)
    {
        _transform = enable;
    }
    
    /**
     * Returns whether the atom coordinates are transformed.
     */
    public boolean getTransform()
    {
        return _transform;
    }

    /**
     * Constructor.
     */
    public PDBWriter(ViewerController controller)
    {
        _controller = controller;

        _renumber  = true;
        _transform = true;
        
        _buffer = new StringBuffer(NUM_COLUMNS);
        _buffer.setLength(NUM_COLUMNS);

        _formatCoord = NumberFormat.getNumberInstance();
        _formatCoord.setMinimumFractionDigits(3);
        _formatCoord.setGroupingUsed(false);

        _formatTemp = NumberFormat.getNumberInstance();
        _formatTemp.setMinimumFractionDigits(2);
        _formatTemp.setGroupingUsed(false);
    }

    /**
     * Write PDB file.
     */
    public void write(File file, int fileId) throws RecoverableException
    {
        MoleculeFile moleculeFile = null;

        ArrayList list = _controller.getMoleculeFiles();
        for(int i=0; i<list.size(); i++)
        {
            MoleculeFile curr = (MoleculeFile)list.get(i);
            if(curr.getFileId() == fileId){
                moleculeFile = curr;
                break;
            }
        }
        if(moleculeFile == null) return;

        try{
            _writer = new BufferedWriter( new FileWriter(file) );

            this.writeHeader(moleculeFile);
            this.writeModels(moleculeFile);

            _writer.close();
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
    }

    //--------------------------------------------------------------------
    // Write header
    //--------------------------------------------------------------------
    private void writeHeader(MoleculeFile moleculeFile) throws IOException
    {
        //fix date
        String date = moleculeFile.getDate();
        if(date.length() == 10) date = fixDate(date);

        // HEADER
        setBlankLine();
        setField("HEADER", 1, 6, false);
        setField(moleculeFile.getClassification(), 11, 50, false);
        setField(date, 51, 59, false);
        setField(moleculeFile.getMoleculeId().toUpperCase(), 63, 66, false);

        if( !_buffer.toString().trim().equals("HEADER") ) writeLine();

        // COMPND
        setBlankLine();
        setField("COMPND", 1, 6, false);
        setField(moleculeFile.getMoleculeName(), 11, 70, false);

        if( !_buffer.toString().trim().equals("COMPND") ) writeLine();

        // EXPDTA
        setBlankLine();
        setField("EXPDTA", 1, 6, false);
        setField(moleculeFile.getExperiment(), 11, 70, false);

        if( !_buffer.toString().trim().equals("EXPDTA") ) writeLine();

        // REMARK
        if (moleculeFile.isVer3()) {
            setBlankLine();
            String pdbId = moleculeFile.getMoleculeId();
            if (pdbId == null) {
                pdbId = "";
            }
            setField("REMARK", 1, 6, false);
            setField("4", 10, 10, false);
            setField(pdbId.toUpperCase(), 12, 15, false);
            setField("COMPLIES WITH FORMAT V. 3.0, 1-DEC-2006", 17, 55, false);
            writeLine();
        }
    }

    private String fixDate(String date)
    {
        StringTokenizer st = new StringTokenizer(date, "-");
        
        try{
            String year = st.nextToken();
            String month = st.nextToken();
            String day = st.nextToken();
            
            year = year.substring(2);
            if(month.equals("01"))      month = "JAN";
            else if(month.equals("02")) month = "FEB";
            else if(month.equals("03")) month = "MAR";
            else if(month.equals("04")) month = "APR";
            else if(month.equals("05")) month = "MAY";
            else if(month.equals("06")) month = "JUN";
            else if(month.equals("07")) month = "JUL";
            else if(month.equals("08")) month = "AUG";
            else if(month.equals("09")) month = "SEP";
            else if(month.equals("10")) month = "OCT";
            else if(month.equals("11")) month = "NOV";
            else if(month.equals("12")) month = "DEC";
            
            return (day + "-" + month + "-" + year);
        }
        catch(Exception e){
            return null;
        }
    }

    //--------------------------------------------------------------------
    // Write models
    //--------------------------------------------------------------------
    private void writeModels(MoleculeFile moleculeFile) throws IOException
    {
        _atomCnt = 1;

        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
            Molecule molecule = moleculeFile.getMolecule(m);
            if(molecule.getSelectAtomCount() == 0) continue;

            // MODEL
            if(molecule.getModelId() != Molecule.NO_MODEL_ID){
                setBlankLine();
                setField("MODEL", 1, 5, false);
                setField("" + molecule.getModelId(), 11, 14, true);
                writeLine();
            }

            for(int ch=0; ch<molecule.numChains(); ch++)
            {
                Chain chain = molecule.getChain(ch);
                boolean writeAtom = false;

                Residue lastResidue = null;
                Atom lastAtom = null;

                for(int re=0; re<chain.numResidues(); re++)
                {
                    Residue residue = chain.getResidue(re);
                    for(int at=0; at<residue.numAtoms(); at++)
                    {
                        Atom atom = residue.getAtom(at);
                        if( !atom.isSelected() ) continue;
    
                        writeAtom = true;
                        lastResidue = residue;
                        lastAtom = atom;
    
                        // ATOM
                        this.writeAtom(
                            chain, residue, atom, molecule.getImage().getTransform()
                        );
                        _atomCnt++;
                    }
                }

                if(writeAtom){
                    this.writeTerminate(chain, lastResidue, lastAtom);
                    _atomCnt++;
                }
            }

            // ENDMDL
            if(molecule.getModelId() != Molecule.NO_MODEL_ID){
                setBlankLine();
                setField("ENDMDL", 1, 6, false);
                writeLine();
            }
        }

        // END
        setBlankLine();
        setField("END", 1, 3, false);
        writeLine();
    }

    //--------------------------------------------------------------------
    // Write atom
    //--------------------------------------------------------------------
    private void writeAtom(
        Chain chain, Residue residue, Atom atom, Matrix4f transform)
        throws IOException
    {
        setBlankLine();

        if(atom.isHetero()) setField("HETATM", 1, 6, false);
        else                setField("ATOM", 1, 4, false);

        if(_renumber) setField("" + _atomCnt, 7, 11, true);
        else          setField("" + atom.getId(), 7, 11, true);

        setField(atom.getName(), 13, 16, false);
        setField(atom.getAltLoc(), 17, 17, false);
        setField(residue.getName(), 18, 20, false);
        setField(chain.getId(), 22, 22, false);
        setField("" + residue.getId(), 23, 26, true);
        if(residue.getInsCode() != 0){
            setField("" + residue.getInsCode(), 27, 27, false);
        }
        Vector4f coord = new Vector4f(atom.getX(), atom.getY(), atom.getZ(), 1);
        if(_transform) coord.transform(transform);

        setField(_formatCoord.format(coord.x), 31, 38, true);
        setField(_formatCoord.format(coord.y), 39, 46, true);
        setField(_formatCoord.format(coord.z), 47, 54, true);
        
        setField("1.00", 55, 60, true);
        setField(_formatTemp.format(atom.getTempFactor()), 61, 66, true);

        String symbol = ElementTable.getSymbol(atom.getAtomicNo()); 
        setField(symbol.toUpperCase(), 77, 78, true);

        writeLine();
    }

    //--------------------------------------------------------------------
    // Write terminate
    //--------------------------------------------------------------------
    private void writeTerminate(Chain chain, Residue residue, Atom atom)
        throws IOException
    {
        setBlankLine();

        setField("TER", 1, 3, false);

        if(_renumber) setField("" + _atomCnt, 7, 11, true);
        else          setField("" + (atom.getId() + 1), 7, 11, true);

        setField(residue.getName(), 18, 20, false);
        setField(chain.getId(), 22, 22, false);
        setField("" + residue.getId(), 23, 26, true);
        if(residue.getInsCode() != 0){
            setField("" + residue.getInsCode(), 27, 27, false);
        }
        writeLine();
    }
    
    //--------------------------------------------------------------------
    // Handle text
    //--------------------------------------------------------------------
    private void setBlankLine()
    {
        for(int i=0; i<_buffer.length(); i++) _buffer.setCharAt(i, ' ');
    }

    private void setField(String field, int start, int end, boolean rightJustify)
    {
        if(field == null) return;

        int size = end - start + 1;
        if(field.length() > size) field = field.substring(0, size);

        int length = field.length();
        int numSpaces = size - length;

        int curr = start - 1;
        if(rightJustify) curr += numSpaces;

        for(int i=0; i<length; i++){
            _buffer.setCharAt(curr++, field.charAt(i));
        }
    }

    private void writeLine() throws IOException
    {
        _writer.write( _buffer.toString() );
        _writer.newLine();
    }
}
