/*
* Binary and Source Code License Agreement for jV Version 4.3.1
* Copyright (c) 2003-2013, Protein Data Bank Japan
* All rights reserved.
* 
* The software, jV, and the associated documentation files (the 
* "Software") were developed by Drs. Kengo Kinoshita and Haruki Nakamura 
* in Protein Data Bank Japan at Institute for Protein Research, Osaka 
* University (the "Contributors"), which is supported by grant-in-aid 
* from National Bioscience Database Center, Japan Science and Technology
* Agancy (JST-NBDC) and Osaka University.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions 
* are met:
* 
* *  Redistributions of the source code for the Software must retain the
*    above copyright notice, this list of conditions and the following
*    disclaimer.
* 
* *  Redistributions in the binary form for the Software must reproduce
*    the above copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided with
*    the distribution.
* 
* *  Neither the name of the Institute for Protein Research, Osaka University
*    nor the names of the Contributors may be used to endorse or promote
*    products derived from this software without specific prior written
*    permission.
* 
*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
*    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*    POSSIBILITY OF SUCH DAMAGE.
*/
package org.pdbj.viewer.writers;

import java.awt.Color;
import java.io.*;
import java.util.*;
import org.pdbj.viewer.tables.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.images.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.polygons.*;
import org.pdbj.viewer.manipulators.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.mouse.*;

/**
 * Script writer.
 */
public class ScriptWriter
{
    private ViewerController _controller;
    private ColorManager _colorManager;
    private MouseBehavior _mouseBehavior;

    private BufferedWriter _writer;

    private int _fileCnt;

    private ArrayList _selectList;
    private ArrayList _displayList;

    /**
     * Constructor.
     */
    public ScriptWriter(ViewerController controller, MouseBehavior mouseBehavior)
    {
        _controller = controller;
        _mouseBehavior = mouseBehavior;

        _colorManager = _controller.getColorManager();

        _selectList  = new ArrayList();
        _displayList = new ArrayList();
    }

    /**
     * Write script file.
     */
    public void write(File file) throws RecoverableException
    {
        if( _controller.getMoleculeFiles().size() == 0
            && _controller.getPolygonFiles().size() == 0 ) return;

        try{
            _writer = new BufferedWriter( new FileWriter(file) );

            this.writeScript();

            _writer.close();
        }
        catch(IOException e){
            throw ExceptionUtil.handleIOException(e);
        }
    }

    //--------------------------------------------------------------------
    // Write script
    //--------------------------------------------------------------------
    private void writeScript() throws IOException
    {
        _fileCnt = 0;

        _selectList.clear();
        _displayList.clear();

        this.writeSettings();

        ArrayList moleculeFiles = _controller.getMoleculeFiles();
        for(int i=0; i<moleculeFiles.size(); i++)
        {
            this.writeMoleculeFile( (MoleculeFile)moleculeFiles.get(i) );
        }

        ArrayList polygonFiles = _controller.getPolygonFiles();
        for(int i=0; i<polygonFiles.size(); i++)
        {
            this.writePolygonFile( (PolygonFile)polygonFiles.get(i) );
        }

        this.writeStatus();
    }

    //--------------------------------------------------------------------
    // Write settings
    //--------------------------------------------------------------------
    private void writeSettings() throws IOException
    {
        writeLine("## Initialize");
        writeLine("zap");
        writeLine("set hetero on");
        writeLine("set hydrogen on");
        writeLine("set bondmode and");

        write("set background ");
        writeLine( getColorExpr(_controller.getViewStatus().getBGColor()) );

        writeLine("set drawlevel " + _controller.getDrawlevel() );
        writeLine("set loadcenter off");

        float a = _controller.getViewStatus().getAmbient() * 100;
        float d1 = _controller.getViewStatus().getDiffuse1() * 100;
        float d2 = _controller.getViewStatus().getDiffuse2() * 100;
        float[] dir1 = _controller.getViewStatus().getDiffuseDir1();
        float[] dir2 = _controller.getViewStatus().getDiffuseDir2();
        writeLine("set ambient " + a);
        writeLine("set diffuse1 " + d1);
        writeLine("set diffuse2 " + d2);
        writeLine("set diffuse1_direction [" + dir1[0] + "," + dir1[1] + "," + dir1[2] + "]");
        writeLine("set diffuse2_direction [" + dir2[0] + "," + dir2[1] + "," + dir2[2] + "]");

        PDBmlUrl pdbmlUrl = _controller.getPDBmlUrl();
        if (!pdbmlUrl.isNoatomDefault()) {
            writeLine("set pdbml_noatom_url " + pdbmlUrl.getUrlNoatom());
        }
        if (!pdbmlUrl.isExtatomDefault()) {
            writeLine("set pdbml_extatom_url " + pdbmlUrl.getUrlExtatom());
        }
        if (!pdbmlUrl.isFunctionDefault()) {
            writeLine("set pdbml_plus_url " + pdbmlUrl.getUrlFunction());
        }
        EfsiteUrl efsiteUrl = _controller.getEfsiteUrl();
        if (!efsiteUrl.isDefault()) {
            writeLine("set efsite_url " + efsiteUrl.getUrl());
        }
        HashMap map = _controller.getSiteUrlMap().getMap();
        Iterator it = map.keySet().iterator();
        while (it.hasNext()) {
            String key = (String) it.next();
            String val = (String) map.get(key);
            writeLine("set ext_site_url " + key + " " + val);
        }
    }

    //--------------------------------------------------------------------
    // Write molecule file
    //--------------------------------------------------------------------
    private void writeMoleculeFile(MoleculeFile moleculeFile) throws IOException
    {
        _fileCnt++;

        _selectList.add( new Boolean(moleculeFile.isSelected()) );
        _displayList.add( new Boolean(moleculeFile.isDisplayed()) );

        writeLine("");
        writeLine("## File " + _fileCnt);

        String type = moleculeFile.isPDBml() ? "pdbml" : "pdb";
        if( moleculeFile.isFTP() ) type = "ftp";
        if( moleculeFile.isAnimation() ) type = "animation";
        writeLine("load " + type + " " + moleculeFile.getFilePath());

        Molecule mol = moleculeFile.getMolecule(0);
        String hbond = mol.getImage().hbondBackbone() ? "backbone" : "sidechain";
        String ssbond = mol.getImage().ssbondBackbone() ? "backbone" : "sidechain";

        writeLine("fselect " + _fileCnt);
        writeLine("set hbond " + hbond);
        writeLine("set ssbond " + ssbond);
        writeLine("set cartoon " + mol.getImage().getCartoonHeight());
        if (mol.getImage().getResidueBackColorIndex() != -1) {
            writeLine("set ribbonback " 
                    + getColorExpr(mol.getImage().getResidueBackColorIndex()));
        } else {
            writeLine("set ribbonback none");
        }
        if (mol.getImage().getCartoonLoopTube()) {
            writeLine("set cartoon_loop_tube on");
        } else {
            writeLine("set cartoon_loop_tube off");
        }
        if (mol.getImage().getCartoonRound()) {
            writeLine("set cartoon_round on");
        } else {
            writeLine("set cartoon_round off");
        }

        Vector3f center = new Vector3f( moleculeFile.center() );
        center.add( moleculeFile.offset() );
        writeLine("set center [" + center.x + "," + center.y + "," + center.z + "]");

        this.writeTransform(mol.getImage().getTransform(), moleculeFile.center());

        this.writeAtoms(moleculeFile);
        this.writeBonds(moleculeFile);
        this.writeBackbones(moleculeFile);
        this.writeHBonds(moleculeFile);
        this.writeSSBonds(moleculeFile);
        this.writeResidues(moleculeFile);
    }

    //--------------------------------------------------------------------
    // Write atoms
    //--------------------------------------------------------------------
    private void writeAtoms(MoleculeFile moleculeFile) throws IOException
    {
        writeLine("");
        writeLine("#Atoms");

        int init = 0;
        int term = 0;
        boolean first = true;

        writeLine("select * & file = " + _fileCnt);
        writeLine("color cpk");
        writeLine("displayatom on");

        // color
        int color = -1;
        boolean cpk = false;

        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
        Molecule molecule = moleculeFile.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
        Residue residue = chain.getResidue(re);
        for(int at=0; at<residue.numAtoms(); at++)
        {
            Atom atom = residue.getAtom(at);
            if(first){
                init = term = atom.getId();
                cpk = isCPKColor(atom);
                color = atom.getImage().getColorIndex();
                first = false;
            }
            else if(cpk && isCPKColor(atom)){
                term = atom.getId();
                if(atom.getImage().getColorIndex() != color) color = -1;
            }
            else if(atom.getImage().getColorIndex() == color){
                term = atom.getId();
                cpk = false;
            }
            else{
                if(color != -1){
                    this.writeAtomRange(init, term);
                    writeLine("color " + getColorExpr(color));
                }

                init = term = atom.getId();
                cpk = isCPKColor(atom);
                color = atom.getImage().getColorIndex();
            }
        }}}}

        if(color != -1){
            this.writeAtomRange(init, term);
            writeLine("color " + getColorExpr(color));
        }

        // spacefill
        float radius = -1;
        boolean vdw = false;
        first = true;

        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
        Molecule molecule = moleculeFile.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
        Residue residue = chain.getResidue(re);
        for(int at=0; at<residue.numAtoms(); at++)
        {
            Atom atom = residue.getAtom(at);
            if(first){
                init = term = atom.getId();
                vdw = isVdWRadius(atom, molecule.hasHydrogens());
                radius = atom.getImage().getRadius();
                first = false;
            }
            else if(vdw && isVdWRadius(atom, molecule.hasHydrogens())){
                term = atom.getId();
                if(atom.getImage().getRadius() != radius) radius = -1;
            }
            else if(atom.getImage().getRadius() == radius){
                term = atom.getId();
                vdw = false;
            }
            else{
                if(radius != 0){
                    this.writeAtomRange(init, term);
                    this.writeAtomSpacefill(radius);
                }
                init = term = atom.getId();
                vdw = isVdWRadius(atom, molecule.hasHydrogens());
                radius = atom.getImage().getRadius();
            }
        }}}}

        if(radius != 0){
            this.writeAtomRange(init, term);
            this.writeAtomSpacefill(radius);
        }

        // display
        boolean display = true;

        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
        Molecule molecule = moleculeFile.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
        Residue residue = chain.getResidue(re);
        for(int at=0; at<residue.numAtoms(); at++)
        {
            Atom atom = residue.getAtom(at);
            if (atom.isDisplayed()) {
                if (!display) {
                    this.writeAtomRange(init, term);
                    writeLine("displayatom off");
                    display = true;
                }
            } else {
                if (display) {
                    init = atom.getId();
                    display = false;
                }                
                term = atom.getId();
            }
        }}}}

        if (!display) {
            this.writeAtomRange(init, term);
            writeLine("displayatom off");
        }
   }

    private void writeAtomSpacefill(float radius) throws IOException
    {
        if(radius != -1){
            writeLine("spacefill " + radius);
        }
        else{
            writeLine("spacefill on");
        }
    }

    private void writeAtomRange(int init, int term) throws IOException
    {
        if(init == term){
            write("select atomno = " + init);
        }
        else{
            write("select ( atomno >= " + init + " & atomno <= " + term + ")");
        }
        writeLine(" & file = " + _fileCnt);
    }

    private boolean isCPKColor(Atom atom)
    {
        return (
            atom.getImage().getColorIndex() == _colorManager.getCPKIndex(atom.getAtomicNo())
        );
    }

    private boolean isVdWRadius(Atom atom, boolean hydrogen)
    {
        float r;
        if(atom.getAtomicNo() == 6 && !hydrogen){
            r = ElementTable.VDW_CARBON;
        }
        else if(atom.getAtomicNo() == 7 && !hydrogen){
            r = ElementTable.VDW_NITROGEN;
        }
        else if(atom.getAtomicNo() == 8 && !hydrogen){
            r = ElementTable.VDW_OXYGEN;
        }
        else if(atom.getAtomicNo() == 10 && !hydrogen){
            r = ElementTable.VDW_SULFUR;
        }
        else{
            r = ElementTable.getVdWRadius(atom.getAtomicNo());
        }

        return ( atom.getImage().getRadius() == r );
    }

    //--------------------------------------------------------------------
    // Write bonds
    //--------------------------------------------------------------------
    private void writeBonds(MoleculeFile moleculeFile) throws IOException
    {
        FrequencyList freq1 = new FrequencyList();
        FrequencyList freq2 = new FrequencyList();

        int num = 0;
        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
            Molecule molecule = moleculeFile.getMolecule(m);
            num += molecule.numBonds();
            for(int b=0; b<molecule.numBonds(); b++)
            {
                Bond bond = molecule.getBond(b);
                freq1.add(bond.getImage().getRadius());
                freq2.add(bond.getImage().getColorIndex());
            }
        }
        if(num == 0) return;

        float defRad = freq1.getList()[0].value;
        int defCol = (int)freq2.getList()[0].value;

        writeLine("");
        writeLine("#Bonds");
        writeLine("select * & file = " + _fileCnt);

        this.writeBond("wireframe", defRad);
        this.writeColor("bonds", defCol);

        if(freq1.getList()[1] != null || freq2.getList()[1] != null)
        {
            for(int m=0; m<moleculeFile.numMolecules(); m++)
            {
            Molecule molecule = moleculeFile.getMolecule(m);
            for(int b=0; b<molecule.numBonds(); b++)
            {
                Bond bond = molecule.getBond(b);
                boolean bRadius = (bond.getImage().getRadius() != defRad);
                boolean bColor = (bond.getImage().getColorIndex() != defCol);

                if(bRadius || bColor){
                    this.writeSelectBond(bond);
                }
                if(bRadius){
                    this.writeBond("wireframe", bond.getImage().getRadius());
                }
                if(bColor){
                    this.writeColor("bonds", bond.getImage().getColorIndex());
                }
            }}
        }
    }

    private void writeSelectBond(Bond bond) throws IOException
    {
        int src = bond.getSrcAtom().getId();
        int dst = bond.getDstAtom().getId();

        write("select (atomno = " + src + " | atomno = " + dst + ")");
        writeLine(" & file = " + _fileCnt);
    }

    private void writeBond(String type, float radius) throws IOException
    {
        if(radius == -1){
            writeLine(type + " off");
        }
        else if(radius == 0){
            writeLine(type + " on");
        }
        else{
            writeLine(type + " " + radius);
        }
    }

    private void writeColor(String type, int color) throws IOException
    {
        if(color == -1){
            writeLine("color " + type + " none");
        }
        else{
            writeLine("color " + type + " " + getColorExpr(color));
        }
    }

    //--------------------------------------------------------------------
    // Write backbones
    //--------------------------------------------------------------------
    private void writeBackbones(MoleculeFile moleculeFile) throws IOException
    {
        FrequencyList freq1 = new FrequencyList();
        FrequencyList freq2 = new FrequencyList();

        int num = 0;
        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
        Molecule molecule = moleculeFile.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
            Chain chain = molecule.getChain(ch);
            num += chain.numBackbones();
            for(int b=0; b<chain.numBackbones(); b++)
            {
                Bond bond = chain.getBackbone(b);
                freq1.add(bond.getImage().getRadius());
                freq2.add(bond.getImage().getColorIndex());
            }
        }}
        if(num == 0) return;

        float defRad = freq1.getList()[0].value;
        int defCol = (int)freq2.getList()[0].value;

        writeLine("");
        writeLine("#Backbones");
        writeLine("select * & file = " + _fileCnt);

        this.writeBond("backbone", defRad);
        this.writeColor("backbone", defCol);

        if(freq1.getList()[1] != null || freq2.getList()[1] != null)
        {
            for(int m=0; m<moleculeFile.numMolecules(); m++)
            {
            Molecule molecule = moleculeFile.getMolecule(m);
            for(int ch=0; ch<molecule.numChains(); ch++)
            {
            Chain chain = molecule.getChain(ch);
            for(int b=0; b<chain.numBackbones(); b++)
            {
                Bond bond = chain.getBackbone(b);
                boolean bRadius = (bond.getImage().getRadius() != defRad);
                boolean bColor = (bond.getImage().getColorIndex() != defCol);

                if(bRadius || bColor){
                    this.writeSelectBond(bond);
                }
                if(bRadius){
                    this.writeBond("backbone", bond.getImage().getRadius());
                }
                if(bColor){
                    this.writeColor("backbone", bond.getImage().getColorIndex());
                }
            }}}
        }
    }

    //--------------------------------------------------------------------
    // Write hbonds
    //--------------------------------------------------------------------
    private void writeHBonds(MoleculeFile moleculeFile) throws IOException
    {
        FrequencyList freq1 = new FrequencyList();
        FrequencyList freq2 = new FrequencyList();

        int num = 0;
        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
            Molecule molecule = moleculeFile.getMolecule(m);
            num += molecule.numHBonds();
            for(int b=0; b<molecule.numHBonds(); b++)
            {
                Bond bond = molecule.getHBond(b);
                freq1.add(bond.getImage().getRadius());
                freq2.add(bond.getImage().getColorIndex());
            }
        }
        if(num == 0) return;

        float defRad = freq1.getList()[0].value;
        int defCol = (int)freq2.getList()[0].value;

        writeLine("");
        writeLine("#HBonds");
        writeLine("select * & file = " + _fileCnt);

        this.writeBond("hbond", defRad);
        this.writeColor("hbonds", defCol);

        if(freq1.getList()[1] != null || freq2.getList()[1] != null)
        {
            for(int m=0; m<moleculeFile.numMolecules(); m++)
            {
            Molecule molecule = moleculeFile.getMolecule(m);
            for(int b=0; b<molecule.numHBonds(); b++)
            {
                Bond bond = molecule.getHBond(b);
                boolean bRadius = (bond.getImage().getRadius() != defRad);
                boolean bColor = (bond.getImage().getColorIndex() != defCol);

                if(bRadius || bColor){
                    this.writeSelectBond(bond);
                }
                if(bRadius){
                    this.writeBond("hbond", bond.getImage().getRadius());
                }
                if(bColor){
                    this.writeColor("hbonds", bond.getImage().getColorIndex());
                }
            }}
        }
    }

    //--------------------------------------------------------------------
    // Write ssbonds
    //--------------------------------------------------------------------
    private void writeSSBonds(MoleculeFile moleculeFile) throws IOException
    {
        FrequencyList freq1 = new FrequencyList();
        FrequencyList freq2 = new FrequencyList();

        int num = 0;
        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
            Molecule molecule = moleculeFile.getMolecule(m);
            num += molecule.numSSBonds();
            for(int b=0; b<molecule.numSSBonds(); b++)
            {
                Bond bond = molecule.getSSBond(b);
                freq1.add(bond.getImage().getRadius());
                freq2.add(bond.getImage().getColorIndex());
            }
        }
        if(num == 0) return;

        float defRad = freq1.getList()[0].value;
        int defCol = (int)freq2.getList()[0].value;

        writeLine("");
        writeLine("#SSBonds");
        writeLine("select * & file = " + _fileCnt);

        this.writeBond("ssbond", defRad);
        this.writeColor("ssbonds", defCol);

        if(freq1.getList()[1] != null || freq2.getList()[1] != null)
        {
            for(int m=0; m<moleculeFile.numMolecules(); m++)
            {
            Molecule molecule = moleculeFile.getMolecule(m);
            for(int b=0; b<molecule.numSSBonds(); b++)
            {
                Bond bond = molecule.getSSBond(b);
                boolean bRadius = (bond.getImage().getRadius() != defRad);
                boolean bColor = (bond.getImage().getColorIndex() != defCol);

                if(bRadius || bColor){
                    this.writeSelectBond(bond);
                }
                if(bRadius){
                    this.writeBond("ssbond", bond.getImage().getRadius());
                }
                if(bColor){
                    this.writeColor("ssbonds", bond.getImage().getColorIndex());
                }
            }}
        }
    }

    //--------------------------------------------------------------------
    // Write residues
    //--------------------------------------------------------------------
    private void writeResidues(MoleculeFile moleculeFile) throws IOException
    {
        boolean first = true;
        for(int m=0; m<moleculeFile.numMolecules(); m++)
        {
        Molecule molecule = moleculeFile.getMolecule(m);
        for(int ch=0; ch<molecule.numChains(); ch++)
        {
        Chain chain = molecule.getChain(ch);
        for(int re=0; re<chain.numResidues(); re++)
        {
            Residue residue = chain.getResidue(re);
            if(residue.getImage().getModel() == ResidueImage.MODEL_NONE) continue;

            Atom atomCA = residue.isProtein() ?
                residue.getProteinCA() : residue.getNucleotideP();

            if(atomCA == null) continue;

            if(first){
                writeLine("");
                writeLine("#Residues");
                first = false;
            }

            this.writeSelectResidue(atomCA);

            if(residue.getImage().getModel() == ResidueImage.MODEL_RIBBON){
                writeLine("ribbons " + residue.getImage().getWidth());
            }
            else if (residue.getImage().getModel() == ResidueImage.MODEL_CARTOON
                    || residue.getImage().getModel() == ResidueImage.MODEL_CARTOON_TUBE) {
                writeLine("cartoon " + residue.getImage().getWidth());
            }

            if(residue.getImage().getColorIndex() != -1){
                this.writeColor("ribbons", residue.getImage().getColorIndex());
            }

            if (residue.getImage().getBackColorIndex() != -1) {
                this.writeColor("ribbonback", residue.getImage().getBackColorIndex());
            }
        }}}
    }

    private void writeSelectResidue(Atom atomCA) throws IOException
    {
        writeLine(
            "select atomno = " + atomCA.getId() + " & file = " + _fileCnt
        );
    }

    //--------------------------------------------------------------------
    // Write polygon file
    //--------------------------------------------------------------------
    private void writePolygonFile(PolygonFile polygonFile) throws IOException
    {
        _fileCnt++;

        _selectList.add( new Boolean(polygonFile.isSelected()) );
        _displayList.add( new Boolean(polygonFile.isDisplayed()) );

        writeLine("");
        writeLine("## File " + _fileCnt);
        writeLine("load polygon " + polygonFile.getFilePath());
        writeLine("fselect " + _fileCnt);

        if (polygonFile.getRadius() == -1) {
            return;
        }

        Vertices vertices = polygonFile.getVertices();
        int[] colors = polygonFile.getImage().getColors();
        BitSet displayFlags = polygonFile.getDisplayFlags();
        
        int[] sort = this.sortVertices(vertices);
        int init = 0;
        int term = 0;

        int num = vertices.length();
        boolean first = true;
        boolean defaultColor = false;
        Color currColor = null;        
        for (int i = 0; i < num; i++)
        {
            int k = sort[i];
            if(first){
                init = term = vertices.getId(k);
                defaultColor = (vertices.getColor(k) == colors[k]);
                currColor = new Color(colors[k]);
                first = false;
            }
            else if(defaultColor && vertices.getColor(k) == colors[k]){
                term = vertices.getId(k);
                if (!isSameColor(colors[k], currColor)) {
                    currColor = null;
                }
            }
            else if(isSameColor(colors[k], currColor)){
                term = vertices.getId(k);
                defaultColor = false;
            }
            else{
                if(currColor != null){
                    this.writeVertexRange(init, term);
                    writeLine("colorvertex " + getColorExpr(currColor));
                }
                init = term = vertices.getId(k);
                defaultColor = (vertices.getColor(k) == colors[k]);
                currColor = new Color(colors[k]);
            }
        }
        if(!defaultColor && currColor != null){
            this.writeVertexRange(init, term);
            writeLine("colorvertex " + getColorExpr(currColor));
        }

        first = true;
        boolean display = false;
        for (int i = 0; i < num; i++)
        {
            int k = sort[i];
            if(first){
                init = term = vertices.getId(k);
                display = displayFlags.get(k);
                first = false;
            }
            else if( (display && displayFlags.get(k)) || (!display && !displayFlags.get(k)) ){
                term = vertices.getId(k);
            }
            else{
                if(!display){
                    this.writeVertexRange(init, term);
                    writeLine("displayvertex off");
                }
                init = term = vertices.getId(k);
                display = displayFlags.get(k);
            }
        }
        if(!display){
            this.writeVertexRange(init, term);
            writeLine("displayvertex off");
        }
        
        // transparency
        if(polygonFile.getPointIndices().length != 0){
            float transparency = polygonFile.getImage().getPointTransparency();
            writeLine("set point_transparency " + transparency);
        }
        if(polygonFile.getLineIndices().length != 0){
            float transparency = polygonFile.getImage().getLineTransparency();
            writeLine("set line_transparency " + transparency);
        }
        if(polygonFile.getTriangleIndices().length != 0){
            float transparency = polygonFile.getImage().getTriangleTransparency();
            writeLine("set triangle_transparency " + transparency);
        }
        if(polygonFile.getQuadIndices().length != 0){
            float transparency = polygonFile.getImage().getQuadTransparency();
            writeLine("set quad_transparency " + transparency);
        }
        if(polygonFile.getPolylineIndices().length != 0){
            float transparency = polygonFile.getImage().getPolylineTransparency();
            writeLine("set polyline_transparency " + transparency);
        }

        // size
        if(polygonFile.getPointIndices().length != 0){
            float size = polygonFile.getImage().getPointSize();
            writeLine("set point_size " + size);
        }
        if(polygonFile.getLineIndices().length != 0){
            float width = polygonFile.getImage().getLineWidth();
            writeLine("set line_width " + width);
        }
        if(polygonFile.getPolylineIndices().length != 0){
            float width = polygonFile.getImage().getPolylineWidth();
            writeLine("set polyline_width " + width);
        }

        Vector3f center = new Vector3f( polygonFile.center() );
        center.add( polygonFile.offset() );
        writeLine("set center [" + center.x + "," + center.y + "," + center.z + "]");

        this.writeTransform(
            polygonFile.getImage().getTransform(), polygonFile.center()
        );
    }

    private boolean isSameColor(int rgb, Color color) {
        return color != null && color.getRGB() == rgb;
    }

    private void writeVertexRange(int init, int term) throws IOException
    {
        if(init == term){
            write("selectvertex id = " + init);
        }
        else{
            write("selectvertex ( id >= " + init + " & id <= " + term + ")");
        }
        writeLine(" & file = " + _fileCnt);
    }

    private int[] sortVertices(Vertices vertices)
    {
        int num = vertices.length();
        int[] orig = new int[num];
        int[] sort = new int[num];

        for(int i=0; i<num; i++){
            orig[i] = vertices.getId(i);
            sort[i] = i;
        }
        
        int w = 0;
        boolean change = true;
        for(int i=0; change && i<num-1; i++){
            change = false;
            for(int j=num-1; j>i; j--){
                if(orig[j] < orig[j-1]){
                    change = true;
                    w = orig[j]; orig[j] = orig[j-1]; orig[j-1] = w;
                    w = sort[j]; sort[j] = sort[j-1]; sort[j-1] = w;
                }
            }
        }

        return sort;
    }

    //--------------------------------------------------------------------
    // Write transform
    //--------------------------------------------------------------------
    private void writeTransform(Matrix4f transform, Vector3f center)
        throws IOException
    {
        if( transform.equals(Matrix4f.identity()) ) return;

        Matrix4f transCenter = new Matrix4f();
        transCenter.translation(center);
        Matrix4f m = new Matrix4f();
        m.multiply(transform, transCenter);

        float[] angles = getEulerAngles(m);
        float[] amount = {m.m14-center.x, m.m24-center.y, m.m34-center.z};

        writeLine("");
        writeLine("#Transform");

        writeLine(
            "center -r {" + center.x + "," + center.y + "," + center.z + "}"
        );

        float cutOff = 0.0001f;
        if(Math.abs(angles[2]) >= cutOff) writeLine("rotate z " + angles[2]);
        if(Math.abs(angles[1]) >= cutOff) writeLine("rotate y " + angles[1]);
        if(Math.abs(angles[0]) >= cutOff) writeLine("rotate z " + angles[0]);
        if(Math.abs(amount[0]) >= cutOff) writeLine("translate x " + amount[0]);
        if(Math.abs(amount[1]) >= cutOff) writeLine("translate y " + amount[1]);
        if(Math.abs(amount[2]) >= cutOff) writeLine("translate z " + amount[2]);
    }

    //--------------------------------------------------------------------
    // Write status
    //--------------------------------------------------------------------
    private void writeStatus() throws IOException
    {
        writeLine("");
        writeLine("## Status");

        // hetero
        if(_controller.getSelectionStatus().getHetero()){
            writeLine("set hetero on");
        }
        else{
            writeLine("set hetero off");
        }

        // hydrogen
        if(_controller.getSelectionStatus().getHydrogen()){
            writeLine("set hydrogen on");
        }
        else{
            writeLine("set hydrogen off");
        }

        // bondmode
        switch(_controller.getSelectionStatus().getBondMode())
        {
            case SelectionStatus.BOND_AND :
                writeLine("set bondmode and"); break;
            case SelectionStatus.BOND_OR :
                writeLine("set bondmode or"); break;
        }

        // picking
        switch(_mouseBehavior.getPickingMode())
        {
            case MouseBehavior.PICKING_OFF :
                writeLine("set picking off"); break;
            case MouseBehavior.PICKING_ID :
                writeLine("set picking ident"); break;
            case MouseBehavior.PICKING_COORDINATES :
                writeLine("set picking coord"); break;
            case MouseBehavior.PICKING_DISTANCE :
                writeLine("set picking distance"); break;
            case MouseBehavior.PICKING_CENTER :
                writeLine("set picking center"); break;
            case MouseBehavior.PICKING_SELECT :
                writeLine("set picking select"); break;
        }

        // loadcenter
        if(_controller.getViewStatus().isLoadCenter()){
            writeLine("set loadcenter on");
        }
        else{
            writeLine("set loadcenter off");
        }

        writeLine("");

        // select
        writeLine("fselect none");
        int cnt = 1;
        for(int i=0; i<_selectList.size(); i++)
        {
            boolean enable = ( (Boolean)_selectList.get(i) ).booleanValue();
            if(enable) writeLine("fselect -a " + cnt);
            cnt++;
        }

        // display
        writeLine("display none");
        cnt = 1;
        for(int i=0; i<_displayList.size(); i++)
        {
            boolean enable = ( (Boolean)_displayList.get(i) ).booleanValue();
            if(enable)writeLine("display -a " + cnt);
            cnt++;
        }

        writeLine("");

        // rotation center
        Vector3f v = new Vector3f();
        _controller.getViewStatus().getRotationCenter(v);
        writeLine("center -r {" + v.x + "," + v.y + "," + v.z + "}");

        // viewpoint
        _controller.getViewStatus().getCameraPosition(v);
        writeLine("set viewpoint {" + v.x + "," + v.y + "," + v.z + "}");

        /*
        // screen center
        _controller.getViewStatus().getCameraPosition(v);
        writeLine("center -s {" + v.x + "," + v.y + "}");

        // zoom
        _controller.getViewStatus().getDefaultCameraPosition(v);
        float defZ = v.z;
        _controller.getViewStatus().getCameraPosition(v);
        float camZ = v.z;
        _controller.getViewStatus().getCurrentCenter(v);
        float curZ = v.z;

        float zoomValue = 100 * (defZ - curZ)/(camZ - curZ);
        writeLine("zoom " + zoomValue);
        */

        // slab
        if (_controller.getViewStatus().isSlab()) {
            writeLine("slab -v " + _controller.getViewStatus().getSlabValue());
        } else {
            writeLine("slab off");
        }
        
        // specular
        if (_controller.getViewStatus().getSpecular()) {
            float[] specular = _controller.getViewStatus().getSpecularColor();
            writeLine("set specular " + (specular[0]*100));
            float shininess = _controller.getViewStatus().getShininess();
            writeLine("set specpower " + (100*(shininess-1)/127.0f));
        } else {
            writeLine("set specular off");
        }
        
        // projection
        if (_controller.getViewStatus().isParallelProjection()) {
            float size = _controller.getViewStatus().getViewSize();
            writeLine("set projection parallel " + (2*size));
        } else {
            writeLine("set projection perspective");
        }

        // pickradius
        if( _mouseBehavior.getUsePickRadius() ){
            float pickRadius = _mouseBehavior.getPickRadius();
            writeLine("set pickradius " + pickRadius);            
        }

        // adjustview
        if (_controller.getViewStatus().isAdjustView()) {
            writeLine("set adjustview on");
        } else {
            writeLine("set adjustview off");
        }

        writeLine("");

        // select all
        writeLine("select all");
        writeLine("selectvertex all");
    }

    //--------------------------------------------------------------------
    // Handle text
    //--------------------------------------------------------------------
    private void write(String text) throws IOException
    {
        _writer.write(text);
    }

    private void writeLine(String text) throws IOException
    {
        _writer.write(text);
        _writer.newLine();
    }

    //--------------------------------------------------------------------
    // Handle color
    //--------------------------------------------------------------------
    private String getColorExpr(int colorIndex)
    {
        Color color = _colorManager.getColor(colorIndex);
        return getColorExpr(color);
    }

    private String getColorExpr(Color color)
    {
        int r = color.getRed();
        int g = color.getGreen();
        int b = color.getBlue();

        return "[" + r + "," + g + "," + b + "]";
    }

    //--------------------------------------------------------------------
    // Euler rotation
    //--------------------------------------------------------------------
    private static float[] getEulerAngles(Matrix4f rotation)
    {
        if(rotation.m33 == 1){
            double omega = getAngle(rotation.m21, rotation.m11);
            float[] angles = {(float)(180 * omega/Math.PI), 0, 0};
            return angles;
        }
        else if(rotation.m33 == -1){
            double omega = getAngle(-rotation.m21, -rotation.m11);
            float[] angles = {(float)(180 * omega/Math.PI), 180, 0};
            return angles;
        }
        else{
            double beta = Math.acos(rotation.m33); // 0 <= acos <= PI
            double sinBeta = Math.sin(beta);
            double alpha =
                getAngle(rotation.m23/sinBeta, rotation.m13/sinBeta);
            double gamma =
                getAngle(rotation.m32/sinBeta, -rotation.m31/sinBeta);

            float[] angles = {
                (float)(180 * alpha/Math.PI),
                (float)(180 * beta/Math.PI),
                (float)(180 * gamma/Math.PI)
            };

            return angles;
        }
    }

    // returns 0 <= angle < 2PI
    private static double getAngle(double sin, double cos)
    {
        double angle = 0;

        if(cos == 0){
            angle = (sin > 0) ? Math.PI/2 : -Math.PI/2;
        }
        else{
            double atan = Math.atan(sin/cos); // -PI/2 < atan < PI/2
            if(sin >= 0){
                angle = (cos > 0) ? atan : atan + Math.PI;
            }
            else{
                angle = (cos > 0) ? atan : atan - Math.PI;
            }
        }

        // change (-PI < angle <= PI) to (0 <= angle < 2PI)
        if(angle < 0){
            angle += 2*Math.PI;
        }

        return angle;
    }    
}
